/***********************************************************************\
*
* $Revision: 319 $
* $Date: 2016-05-25 08:21:52 +0200 (Wed, 25 May 2016) $
* $Author: torsten $
* Contents: Action monitor functions
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.text.SimpleDateFormat;

import java.util.Arrays;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

// SQLite database
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

// SWT
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;

// KTable
import de.kupzog.ktable.KTable;
import de.kupzog.ktable.KTableCellEditor;
import de.kupzog.ktable.KTableCellRenderer;
import de.kupzog.ktable.KTableDefaultModel;
import de.kupzog.ktable.KTableModel;
import de.kupzog.ktable.KTableCellSelectionListener;
import de.kupzog.ktable.renderers.FixedCellRenderer;

// JNA
import com.sun.jna.Library;
import com.sun.jna.Native;
import com.sun.jna.NativeLong;
import com.sun.jna.Platform;
import com.sun.jna.Pointer;
import com.sun.jna.PointerType;
import com.sun.jna.ptr.ByReference;
import com.sun.jna.ptr.IntByReference;
import com.sun.jna.ptr.NativeLongByReference;
import com.sun.jna.ptr.PointerByReference;
import com.sun.jna.Structure;

import com.sun.jna.platform.unix.X11;
import com.sun.jna.platform.unix.X11.Atom;
import com.sun.jna.platform.unix.X11.AtomByReference;
import com.sun.jna.platform.unix.X11.WindowByReference;

import com.sun.jna.platform.win32.Kernel32;
import com.sun.jna.platform.win32.User32;
import com.sun.jna.platform.win32.WinDef;
import com.sun.jna.platform.win32.WinNT;
import com.sun.jna.platform.win32.WinDef.UINT_PTR;
import com.sun.jna.win32.StdCallLibrary;
import com.sun.jna.win32.W32APITypeMapper;
import com.sun.jna.win32.W32APIFunctionMapper;

import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.ScriptException;

/****************************** Classes ********************************/

/** tag
 */
class Tag
{
  long   startTimestamp;
  long   endTimestamp;
  String name;

  private static SimpleDateFormat TIME_FORMAT = new SimpleDateFormat("HH:mm");

  /** create tag
   * @param startTimestamp start timestamp [s]
   * @param endTimestamp end timestamp [s]
   * @param name tag name
   */
  Tag(long startTimestamp, long endTimestamp, String name)
  {
    this.startTimestamp = Math.min(startTimestamp,endTimestamp);
    this.endTimestamp   = Math.max(startTimestamp,endTimestamp);
    this.name           = name;
  }

  /** create tag
   * @param startTimestamp start timestamp [s]
   * @param endTimestamp end timestamp [s]
   */
  Tag(long startTimestamp, long endTimestamp)
  {
    this(startTimestamp,endTimestamp,"");
  }

  /** set tag range
   * @param timestamp start timestamp [s]
   */
  public void setStart(long timestamp)
  {
    if (timestamp < this.endTimestamp)
    {
      this.startTimestamp = timestamp;
    }
    else
    {
      this.startTimestamp = this.endTimestamp;
      this.endTimestamp   = timestamp;
    }
  }

  /** set tag range
   * @param timestamp end timestamp [s]
   */
  public void setEnd(long timestamp)
  {
    if (timestamp > this.startTimestamp)
    {
      this.endTimestamp = timestamp;
    }
    else
    {
      this.endTimestamp   = this.startTimestamp;
      this.startTimestamp = timestamp;
    }
  }

  /** set tag range
   * @param startTimestamp,endTimestamp start/end timestamp [s]
   */
  public void setRange(long startTimestamp, long endTimestamp)
  {
    setStart(startTimestamp);
    setEnd(endTimestamp);
  }

  /** check it tag range intersect
   * @param tag tag
   * @return true iff tag range has intersection
   */
  public boolean hasRangeIntersection(Tag tag)
  {
//TODO: round to minutes?
    return (tag.startTimestamp < this.endTimestamp) && (tag.endTimestamp > this.startTimestamp);
  }

  /** get string
   * @return string
   */
  public String toString()
  {
    return "Tag {"+TIME_FORMAT.format(startTimestamp*1000L)+", "+TIME_FORMAT.format(endTimestamp*1000L)+", "+name+"}";
  }
}

/** tag list
 */
class TagList extends LinkedList<Tag>
{
  /** get assigned tag for action
   * @param action action
   * @return tag or null
   */
  public Tag getAssignedTag(Action action)
  {
    for (Tag tag : this)
    {
//Dprintf.dprintf("tag test %s",tag);
      if ((tag.startTimestamp <= action.timestamp) && (action.timestamp < tag.endTimestamp))
      {
        return tag;
      }
    }

    return null;
  }

  /** get assigned tag for time
   * @param timestamp timestamp
   * @return tag or null
   */
  public Tag getAssignedTag(long timestamp)
  {
    for (Tag tag : this)
    {
      if ((tag.startTimestamp <= timestamp) && (timestamp <= tag.endTimestamp))
      {
        return tag;
      }
    }

    return null;
  }
}

/** action types
 */
enum ActionTypes
{
  NONE,

  WINDOW,
  APPLICATION,
  FILE,
  IDLE,
  ACTIVE,

  UNKNOWN;

  /** get string
   * @return string
   */
  public String toString()
  {
    switch (this)
    {
      case WINDOW:      return "Window";
      case APPLICATION: return "Application";
      case FILE:        return "File";
      case IDLE:        return "Idle";
      default:          return "";
    }
  }
};

/** action
 */
class Action
{
  public final ActionTypes type;
  public final long        timestamp;  // date/time [s]
  public String            text;
  public int               tagId;

  private static SimpleDateFormat TIME_FORMAT = new SimpleDateFormat("HH:mm");

  /** create action
   * @param type action type
   * @param timestamp timestamp
   * @param text text
   * @param tagId tag id
   */
  public Action(ActionTypes type, long timestamp, String text, int tagId)
  {
    this.type      = type;
    this.timestamp = timestamp;
    this.text      = text;
    this.tagId     = tagId;
  }

  /** create action
   * @param type action type
   * @param timestamp timestamp
   * @param text text
   */
  public Action(ActionTypes type, long timestamp, String text)
  {
    this(type,timestamp,text,Database.ID_NONE);
  }

  /** get text
   * @return text
   */
  public String getText()
  {
    switch (type)
    {
      case WINDOW:
      case APPLICATION:
      case FILE:
        return text;
      case IDLE:
        return "Idle";
      case ACTIVE:
        return "Active";
      default:
        return "unknown";
    }
  }

  /** get string
   * @return string
   */
  public String toString()
  {
    return "Action {"+type+", "+TIME_FORMAT.format(timestamp*1000L)+", "+text+", "+tagId+"}";
  }
}

/** time range
 */
class TimeRange
{
  long startTimestamp;  // [s]
  long endTimestamp;    // [s]

  private static SimpleDateFormat TIME_FORMAT = new SimpleDateFormat("HH:mm");

  TimeRange(long startTimestamp, long endTimestamp)
  {
    this.startTimestamp = startTimestamp;
    this.endTimestamp   = endTimestamp;
  }

  /** get string
   * @return string
   */
  public String toString()
  {
    return "TimeRange {"+TIME_FORMAT.format(startTimestamp*1000L)+" .. "+TIME_FORMAT.format(endTimestamp*1000L)+"}";
  }
}

/** action list
 */
class ActionList extends LinkedList<Action>
{
  private Calendar date;

  /** create action list
   */
  ActionList()
  {
    this.date = Calendar.getInstance();
  }

  /** check if action list is from today
   * @return true iff from today
   */
  public boolean isToday()
  {
    Calendar now = Calendar.getInstance();

    return    (date.get(Calendar.YEAR        ) == now.get(Calendar.YEAR        ))
           && (date.get(Calendar.MONTH       ) == now.get(Calendar.MONTH       ))
           && (date.get(Calendar.DAY_OF_MONTH) == now.get(Calendar.DAY_OF_MONTH));
  }

  /** get start timestamp
   * @return timestamp [s]
   */
  public long getStartTimestamp()
  {
    return !isEmpty() ? getFirst().timestamp : 0L;
  }

  /** get end timestamp
   * @return timestamp [s]
   */
  public long getEndTimestamp()
  {
    return !isEmpty() ? getLast().timestamp : 0L;
  }
}

/** action table
 */
class ActionTable extends KTable
{
  /** action listener
   */
  abstract static class ActionListener
  {
    /** called on mouse double-click
     * @param mouseEvent mouse event
     */
    public abstract void mouseDoubleClick(int row, int column);

    /** called on mouse down
     * @param mouseEvent mouse event
     */
    public abstract void mouseDown(int row, int column);

    /** called on mouse up
     * @param mouseEvent mouse event
     */
    public abstract void mouseUp(int row, int column);
  }

  /** cell renderer
   */
  private class CellRenderer implements KTableCellRenderer
  {
    public CellRenderer()
    {
    }

    @Override
    public int getOptimalWidth(GC          gc,
                               int         column,
                               int         row,
                               Object      content,
                               boolean     isFixed,
                               KTableModel tableModel
                              )
    {
      return 0;
    }

    @Override
    public void drawCell(GC          gc,
                         Rectangle   rect,
                         int         column,
                         int         row,
                         Object      content,
                         boolean     isFocused,
                         boolean     isFixed,
                         boolean     isClicked,
                         KTableModel tableModel
                        )
    {
      Color  textColor;
      Color  backColor0,backColor1;
      Color  borderColor;
      String text;
      Point  textExtend;

      assert(row >= 0);
      assert(column >= 0);
      assert(column < 3);

      text = (String)content;

      if (row == 0)
      {
        // draw title
        gc.setBackground(colorBackground);
        gc.setForeground(colorText);
        gc.setFont(fontBold);

        gc.fillRectangle(rect);
        textExtend = gc.stringExtent(text);
        gc.drawText(text,rect.x+3,rect.y+(rect.height-textExtend.y)/2);
      }
      else
      {
        // draw content
        if (isFocused)
        {
          gc.setBackground(colorBackgroundSelection);
        }
        else
        {
          gc.setBackground(((row%2) == 0) ? colorBackground1 : colorBackground0);
        }
        gc.setForeground(colorText);
        gc.setFont(fontNormal);

        gc.fillRectangle(rect);
        if (text != null)
        {
          textExtend = gc.stringExtent(text);
          switch (column)
          {
            case 0:
            case 1:
              // draw text
              gc.drawText(text,rect.x+3,rect.y+(rect.height-textExtend.y)/2);
              break;
            case 2:
              // draw brace+text
              int p[] = new int[]
              {
                rect.x+2+ 0,rect.y+3,
                rect.x+2+ 8,rect.y+3,
                rect.x+2+ 8,rect.y+(rect.height-3)/2,
                rect.x+2+12,rect.y+(rect.height-3)/2,
                rect.x+2+ 8,rect.y+(rect.height-3)/2,
                rect.x+2+ 8,rect.y+rect.height-3-3,
                rect.x+2+ 0,rect.y+rect.height-3-3
              };
              gc.drawPolyline(p);
              gc.drawText(text,rect.x+20,rect.y+(rect.height-textExtend.y)/2);
              break;
          }
        }
      }
    }
  }

  /** table model
   */
  private class TableModel extends KTableDefaultModel
  {
    public final static int COLUMNS = 3;

    private final String COLUMN_TITLES[] = new String[]{"Time","Actions","Tags"};
    private final int    COLUMN_WIDTHS[] = new int[]{44,100,0};

    private KTable       kTable;
    private CellRenderer cellRenderer;

    /**
     * Initialize the underlying model
     */
    public TableModel(KTable kTable)
    {
      // before initializing, you probably have to set some member values
      // to make all model getter methods work properly.
      initialize();
      this.kTable = kTable;
      cellRenderer = new CellRenderer();
    }

    @Override
    public Object doGetContentAt(int column, int row)
    {
      String content = null;

      assert(column < 3);

      if (row == 0)
      {
        content = COLUMN_TITLES[column];
      }
      else
      {
        Action action = actionList.get(row-1);
        switch (column)
        {
          case 0:
            content = TIME_FORMAT.format(action.timestamp*1000);
            break;
          case 1:
            content = action.text;
            break;
          case 2:
            Tag assignedTag = (tagList != null) ? tagList.getAssignedTag(action.timestamp) : null;
            if (assignedTag != null) content = assignedTag.name;
            break;
        }
      }

      return content;
    }

    @Override
    public void doSetContentAt(int column, int row, Object value)
    {
      assert(column < COLUMNS);

      // nothing to do
    }

    @Override
    public KTableCellEditor doGetCellEditor(int column, int row)
    {
      assert(column < COLUMNS);

      // no editor
      return null;
    }

    @Override
    public KTableCellRenderer doGetCellRenderer(int column, int row)
    {
      assert(column < COLUMNS);

      return cellRenderer;
    }

    @Override
    public Point doBelongsToCell(int column, int row)
    {
      assert(column < COLUMNS);

      if (   isFixedCell(column,row)
          || (column == 0)
          || (column == 1)
         )
      {
        return new Point(column,row);
      }

      column -= getFixedHeaderColumnCount()+getFixedSelectableColumnCount();
      row    -= getFixedHeaderRowCount()   +getFixedSelectableRowCount();

      // get action
      Action action = actionList.get(row);

      // find assigned tag
      Tag assignedTag = null;
      if (tagList != null)
      {
        assignedTag = tagList.getAssignedTag(action.timestamp);
//if (assignedTag != null) Dprintf.dprintf("action=%s assignedTag=%s",action,assignedTag);
      }

      // spawn rows for tag
      if (assignedTag != null)
      {
        while ((row > 0) && (tagList.getAssignedTag(actionList.get(row-1).timestamp) == assignedTag))
        {
          row--;
        }
      }

      return new Point(column+getFixedHeaderColumnCount()+getFixedSelectableColumnCount(),
                       row   +getFixedHeaderRowCount()   +getFixedSelectableRowCount()
                      );
    }

    @Override
    public int getFixedHeaderRowCount()
    {
      return 1;
    }

    @Override
    public int getFixedHeaderColumnCount()
    {
      return 0;
    }

    @Override
    public int doGetRowCount()
    {
      return getFixedRowCount()+((actionList != null) ? actionList.size() : 0);
    }

    @Override
    public int doGetColumnCount()
    {
      return 3;
    }

    @Override
    public int getFixedSelectableRowCount()
    {
      return 0;
    }

    @Override
    public int getFixedSelectableColumnCount()
    {
      return 0;
    }

    @Override
    public int getInitialColumnWidth(int column)
    {
      assert column >= 0;
      assert column < COLUMNS;

      return COLUMN_WIDTHS[column];
    }

    @Override
    public int getColumnWidth(int column)
    {
      Rectangle bounds = kTable.getClientArea();
      int       width  = 0;

      assert column >= 0;
      assert column < COLUMNS;

      switch (column)
      {
        case 0:
        case 1:
          width = super.getColumnWidth(column);
          break;
        case 2:
          width = bounds.width-super.getColumnWidth(0)-super.getColumnWidth(1);
          break;
      }

      return width;
    }

    @Override
    public int getInitialFirstRowHeight()
    {
      return margin+fontHeight+margin;
    }

    @Override
    public int getInitialRowHeight(int row)
    {
      return margin+fontHeight+margin;
    }

    @Override
    public int getRowHeightMinimum()
    {
      return margin+fontHeight+margin;
    }

    @Override
    public int getRowHeight(int row)
    {
//      assert(row >= 0);

      return margin+fontHeight+margin;
    }

    @Override
    public boolean isColumnResizable(int column)
    {
//      assert(column >= 0);
//      assert(column < COLUMNS);

      return true;
    }

    @Override
    public boolean isRowResizable(int row)
    {
//      assert(row >= 0);

      return false;
    }

/*
    @Override
    public String doGetTooltipAt(int column, int row)
    {
      assert(row >= 0);
      assert(column >= 0);
      assert(column < COLUMNS);

      return "Test-Tooltip";
    }
*/
  }

  // --------------------------- constants --------------------------------
  private final  int              DEFAULT_MARGIN = 6;
  private static SimpleDateFormat TIME_FORMAT    = new SimpleDateFormat("HH:mm");

  // --------------------------- variables --------------------------------
  private Display                 display;
  private Color                   colorBackground;
  private Color                   colorText;
  private Color                   colorBackground0,colorBackground1;
  private Color                   colorBackgroundSelection;
  private Font                    fontNormal;
  private Font                    fontBold;
  private int                     fontHeight;
  private int                     margin;
  private Menu                    menu;

  private HashSet<ActionListener> actionListenerSet  = new HashSet<ActionListener>();

  private ActionList              actionList         = null;
  private TagList                 tagList            = null;
  private boolean                 selectFlag         = false;
  private HashSet<Action>         selectedActionSet  = new HashSet<Action>();
  private Tag                     selectedTag        = null;
  private boolean                 isTagStartSelected = false;
  private boolean                 isTagEndSelected   = false;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create action table
   * @param composite parent composite widget
   * @param style style flags
   */
  ActionTable(Composite composite, int style)
  {
    super(composite,style|SWT.FULL_SELECTION|SWT.MULTI);//SWT.HIDE_SELECTION);
    setModel(new TableModel(this));

    display = Display.getCurrent();

    // get colors
    colorBackground          = display.getSystemColor(SWT.COLOR_WIDGET_BACKGROUND);
    colorText                = display.getSystemColor(SWT.COLOR_LIST_FOREGROUND);
    colorBackground0         = display.getSystemColor(SWT.COLOR_LIST_BACKGROUND);
    colorBackground1         = new Color(display,0xF8,0xF7,0xF6);
    colorBackgroundSelection = display.getSystemColor(SWT.COLOR_LIST_SELECTION);

    // get fonts
    fontNormal = getFont();
    fontHeight = 0;
    FontData[] fontData = fontNormal.getFontData();
    for (int i = 0; i < fontData.length; i++)
    {
      fontData[i].setStyle(SWT.BOLD);
      if (fontData[i].getHeight() > fontHeight) fontHeight = fontData[i].getHeight();
    }
    fontBold = new Font(display,fontData);
    margin = DEFAULT_MARGIN;

    // add listeners
    addMouseListener(new MouseListener()
    {
      @Override
      public void mouseDoubleClick(MouseEvent mouseEvent)
      {
      }
      @Override
      public void mouseDown(MouseEvent mouseEvent)
      {
        if (mouseEvent.button == 1)
        {
          // Note: use x-coordinate 0 to map to correct action row (row-span in tag column!)
          int row    = getCellForCoordinates(0,mouseEvent.y).y-1;
          int column = getCellForCoordinates(mouseEvent.x,mouseEvent.y).x;

          // clear selection
          selectedActionSet.clear();

          // get action
          Action action = ((row >= 0) && (row < actionList.size()))
                           ? actionList.get(row)
                           : null;
//Dprintf.dprintf("mouseEvent.y=%d actionList.size()=%d -> row=%d column=%d: action=%s",mouseEvent.y,actionList.size(),row,column,action);
          if (action != null)
          {
            // set selection
            selectedActionSet.add(action);
Dprintf.dprintf("set selectedActionSet.size=%d",selectedActionSet.size());

            selectedTag = tagList.getAssignedTag(action);

            // select tag if in tag-column
            if (column == 2)
            {
              isTagStartSelected = (selectedTag != null) && (selectedTag.startTimestamp == action.timestamp);
              isTagEndSelected   = (selectedTag != null) && (selectedTag.endTimestamp   == action.timestamp);
Dprintf.dprintf("%d: action=%s selectedTag=%s isTagStartSelected=%s isTagEndSelected=%s",row,action,selectedTag,isTagStartSelected,isTagEndSelected);
            }
            else
            {
Dprintf.dprintf("clear selectedTag");
//              selectedTag = null;
              for (ActionListener actionListener : actionListenerSet)
              {
                actionListener.mouseDown(row,column);
              }
            }
          }

          selectFlag = true;
        }
      }
      @Override
      public void mouseUp(MouseEvent mouseEvent)
      {
        if (mouseEvent.button == 1)
        {
          // Note: use x-coordinate 0 to map to correct action row (row-span in tag column!)
          int row    = getCellForCoordinates(0,mouseEvent.y).y-1;
          int column = getCellForCoordinates(mouseEvent.x,mouseEvent.y).x;

Dprintf.dprintf("selectFlag=%s selectedTag=%s",selectFlag,selectedTag);
          if (selectFlag)
          {
//            if (selectedTag == null)
            {
              for (ActionListener actionListener : actionListenerSet)
              {
                actionListener.mouseUp(row,column);
              }
            }

            selectFlag = false;
          }
        }
      }
    });
    addMouseMoveListener(new MouseMoveListener()
    {
      @Override
      public void mouseMove(MouseEvent mouseEvent)
      {
        if (selectFlag)
        {
          // Note: use x-coordinate 0 to map to correct action row (row-span in tag column!)
          int row    = getCellForCoordinates(0,mouseEvent.y).y-1;
          int column = getCellForCoordinates(mouseEvent.x,mouseEvent.y).x;

          // get action
          Action action = ((row >= 0) && (row < actionList.size()))
                            ? actionList.get(row)
                            : null;
          if (action != null)
          {
            // add selection
            selectedActionSet.add(action);

            if (selectedTag != null)
            {
  //Dprintf.dprintf("%d %d action=%s start=%s end=%s",row,column,action,isTagStartSelected,isTagEndSelected);
              if      (isTagStartSelected)
              {
                selectedTag.setStart(action.timestamp);
                redraw();
              }
              else if (isTagEndSelected)
              {
                selectedTag.setEnd(action.timestamp);
                redraw();
              }
            }
          }
        }
      }
    });

    // Note: required to avoid additonal context menu "input methods" added by default
    addListener(SWT.MenuDetect,new Listener()
    {
      public void handleEvent(Event event)
      {
          KTable widget = (KTable)event.widget;

          widget.setMenu(menu);
          menu.setVisible(true);
          event.doit = false;
      }
    });

/*
// Note cellSelected() is called before mouseDown!
    addCellSelectionListener(new KTableCellSelectionListener()
    {
      @Override
      public void cellSelected(int column, int row, int statemask)
      {
//Dprintf.dprintf("column=%d row=%d statemask=%x b1=%x shift=%x ctrl=%x",column,row,statemask,SWT.BUTTON1,SWT.SHIFT,SWT.CTRL);
        if ((statemask & SWT.SHIFT) != 0)
        {
          // extend/shrink tag
          if (selectedTag != null)
          {
            Action action = ((row > 0) && ((row-1) < actionList.size())) ? actionList.get(row-1) : null;
            if (action != null)
            {
Dprintf.dprintf("%d action=%s",row,action);
Dprintf.dprintf("extend/shrink %s: %s %s",selectedTag,isTagStartSelected,isTagEndSelected);
              if      (   isTagStartSelected
//                       && (action.timestamp < action.timestamp)
                      )
              {
                long startTime = action.timestamp;
                long endTime   = selectedTag.endTimestamp;

                selectedTag.startTimestamp = Math.min(startTime,endTime);
                selectedTag.endTimestamp   = Math.max(startTime,endTime);

Dprintf.dprintf("");
                redraw();
              }
              else if (   isTagEndSelected
//                       && (action.timestamp > action.timestamp)
                      )
              {
                long  startTime = selectedTag.startTimestamp;
                long  endTime   = action.timestamp;

                selectedTag.startTimestamp = Math.min(startTime,endTime);
                selectedTag.endTimestamp   = Math.max(startTime,endTime);

Dprintf.dprintf("");
                redraw();
              }

//              selectedAction = action;
            }

          }
        }
        else
        {
          // select action
//          selectedAction = ((row > 0) && ((row-1) < actionList.size())) ? actionList.get(row-1) : null;
//Dprintf.dprintf("select %s",selectedAction);
        }
      }
      @Override
      public void fixedCellSelected(int col, int row, int statemask)
      {
Dprintf.dprintf("");
      }
    });
*/
  }

  /** create action table
   * @param composite parent composite widget
   */
  ActionTable(Composite composite)
  {
    this(composite,SWT.NONE);
  }

  /** set context menu
   * @param menu context menu
   */
  public void setMenu(Menu menu)
  {
    this.menu = menu;
  }

  /** get column width
   * @return column with array
   */
  public int[] getColumnWidth()
  {
    int width[] = new int[TableModel.COLUMNS];

    for (int i = 0; i < TableModel.COLUMNS; i++)
    {
      width[i] = getModel().getColumnWidth(i);
    }

    return width;
  }

  /** set column width
   * @param width column with array
   */
  public void setColumnWidth(int width[])
  {
    for (int i = 0; i < TableModel.COLUMNS; i++)
    {
      getModel().setColumnWidth(i,width[i]);
    }
  }

  /** add action listener
   * @param actionListener action listener
   */
  public void addActionListener(ActionListener actionListener)
  {
    actionListenerSet.add(actionListener);
  }

  /** remove action listener
   * @param actionListener action listener
   */
  public void removeActionListener(ActionListener actionListener)
  {
    actionListenerSet.remove(actionListener);
  }

  /** refresh action list and redraw
   */
  public void refresh(ActionList actionList)
  {
    this.actionList = actionList;
    if (!isDisposed())
    {
      redraw();
    }
  }

  /** refresh tag list and redraw
   */
  public void refresh(TagList tagList)
  {
    this.tagList = tagList;
    if (!isDisposed())
    {
      redraw();
    }
  }

  /** get action at position
   * @param x,y position
   * @return action or null
   */
  public Action getActionAt(int x, int y)
  {
    int row = getCellForCoordinates(x,y).y-1;
    return ((row >= 0) && (row < actionList.size())) ? actionList.get(row) : null;
  }

  /** get action at mouse position
   * @param mouseEvent mouse event
   * @return action or null
   */
  public Action getActionAt(MouseEvent mouseEvent)
  {
    return getActionAt(mouseEvent.x,mouseEvent.y);
  }

  /** get selected actions
   * @return actions array
   */
  public Action[] getSelectedActions()
  {
    ArrayList<Action> actionList = new ArrayList<Action>(selectedActionSet);
    Collections.sort(actionList,new Comparator<Action>()
    {
      public int compare(Action action0, Action action1)
      {
        assert action0 != null;
        assert action1 != null;

        if      (action0.timestamp < action1.timestamp) return -1;
        else if (action0.timestamp > action1.timestamp) return  1;
        else                                            return  0;
      }
    });

    return actionList.toArray(new Action[actionList.size()]);
  }

  /** get selected tag
   * @return selected tag or null
   */
  public Tag getSelectedTag()
  {
Dprintf.dprintf("selectedTag2=%s",selectedTag);
    return selectedTag;
  }

  /** get selected actions
   * @return time ranges array
   */
  public TimeRange[] getSelectedTimeRanges()
  {
    ArrayList<TimeRange> timeRangeList = new ArrayList<TimeRange>();
    for (int i = 0; i < actionList.size()-1; i++)
    {
      Action action = actionList.get(i);

      if (selectedActionSet.contains(action))
      {
        timeRangeList.add(new TimeRange(action.timestamp,actionList.get(i+1).timestamp));
      }
    }

    Collections.sort(timeRangeList,new Comparator<TimeRange>()
    {
      public int compare(TimeRange timeRange0, TimeRange timeRange1)
      {
        assert timeRange0.startTimestamp <= timeRange0.endTimestamp;
        assert timeRange1.startTimestamp <= timeRange1.endTimestamp;
        assert (timeRange0.endTimestamp <= timeRange1.startTimestamp) || (timeRange1.endTimestamp <= timeRange0.startTimestamp);

        if      (timeRange0.endTimestamp <= timeRange1.startTimestamp) return -1;
        else if (timeRange1.endTimestamp <= timeRange0.startTimestamp) return  1;
        else                                                           return  0;
      }
    });

    return timeRangeList.toArray(new TimeRange[timeRangeList.size()]);
  }
}

/** action monitor interface
 */
interface ActionMonitorInterface
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get title of current active window
   * @return window title
   */
  abstract public String getCurrentWindowTitle();

  /** get current active application name
   * @return application name
   */
  abstract public String getCurrentApplication();

  /** check if system is idle
   * @return idle time [ms]
   */
  abstract public long getIdleTime();
}

/** action monitor interface
 */
public class ActionMonitor implements ActionMonitorInterface
{
  /** action listener
   */
  abstract static class ActionListener
  {
    /** called if action is added
     * @param action action
     */
    public void add(Action action)
    {
      // nothing to do
    }

    /** called if action is removed
     * @param action action
     */
    public void remove(Action action)
    {
      // nothing to do
    }

    /** called if all actions are cleared
     */
    public void clear()
    {
      // nothing to do
    }

    /** called after update of action list
     * @param actionList new action list
     */
    public void refresh(ActionList actionList)
    {
      // nothing to do
    }

    /** called after update of tag list
     * @param actionList new tag list
     */
    public void refresh(TagList tagList)
    {
      // nothing to do
    }
  }

  // --------------------------- constants --------------------------------
  public final static String PREFIX_WINDOW      = "WINDOW ";
  public final static String PREFIX_APPLICATION = "APPLICATION ";
  public final static String PREFIX_IDLE        = "IDLE ";

  // --------------------------- variables --------------------------------
  private static ActionMonitor    actionMonitor        = null;

  private Database                database;
  private HashSet<String>         actionFilters     = null;
  private ArrayList<Pattern>      actionFilterPatterns = new ArrayList<Pattern>();
  private HashSet<ActionListener> actionListenerSet    = new HashSet<ActionListener>();

  private HashMap<String,Integer> tagMap               = new HashMap<String,Integer>();

  private Object                  lock                 = new Object();
  private ActionList              actionList           = new ActionList();
  private TagList                 tagList              = new TagList();

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  // init action monitor instance
  static
  {
    if      (Platform.isLinux())
    {
      actionMonitor = new ActionMonitorUnix();
    }
    else if (Platform.isWindows())
    {
      actionMonitor = new ActionMonitorWindows();
    }
    else if (Platform.isMac())
    {
      actionMonitor = new ActionMonitorMacOSX();
    }

    Runtime.getRuntime().addShutdownHook(new Thread()
    {
      @Override
      public void run()
      {
        closeLocalDatabase(actionMonitor.database);
      }
    });
  }

  /** get instance
   * @return instance
   */
  public static ActionMonitor getInstance()
  {
    assert actionMonitor != null;

    return actionMonitor;
  }

  /** get current window title
   */
  public String getCurrentWindowTitle()
  {
    return "";
  }

  /** get current active application
   * @return application name
   */
  public String getCurrentApplication()
  {
    return "";
  }

  /** get system idle time
   * @return idle time [ms]
   */
  public long getIdleTime()
  {
    return 0;
  }

  /** set action filters
   * @param actionFilters action filter set
   */
  public void setActionFilters(HashSet<String> actionFilters)
  {
    // get actiom filter pattern list
    actionFilterPatterns.clear();
    for (String actionFilter : actionFilters)
    {
      actionFilterPatterns.add(Pattern.compile(StringUtils.globToRegex(actionFilter)));
    }

    // filter list
    synchronized(lock)
    {
      int i = 0;
      while (i < actionList.size())
      {
        if (isFilteredAction(actionList.get(i).text))
        {
          actionList.remove(i);
        }
        else
        {
          i++;
        }
      }
    }

    // trigger listeners
    for (ActionListener actionListener : actionListenerSet)
    {
      for (Action action : actionList)
      {
        actionListener.add(action);
      }
      actionListener.refresh(actionList);
    }
  }

  /** add action listener
   * @param actionListener action listener
   */
  public void addActionListener(ActionListener actionListener)
  {
    actionListenerSet.add(actionListener);

    actionListener.clear();
    for (Action action : actionList)
    {
      actionListener.add(action);
    }
    actionListener.refresh(actionList);
  }

  /** remove action listener
   * @param actionListener action listener
   */
  public void removeActionListener(ActionListener actionListener)
  {
    actionListener.clear();

    actionListenerSet.remove(actionListener);
  }

  /** add action entry
   * @param actionType action type
   * @param text text
   */
  public void addAction(ActionTypes actionType, String text)
    throws SQLException
  {
    final long timestamp = System.currentTimeMillis()/1000;

//TODO: happen?
assert text != null;

    // insert into database
    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = database.prepareStatement("INSERT INTO actions (type,dateTime,text) VALUES (?,DATETIME(?,'unixepoch'),?);");
      preparedStatement.setInt(1,actionType.ordinal());
      preparedStatement.setLong(2,timestamp);
      preparedStatement.setString(3,text);
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }

    // add to action list
    if (!isFilteredAction(text))
    {
      Action action = new Action(actionType,timestamp,text);
      synchronized(lock)
      {
        actionList.addLast(action);
      }

      // trigger listeners
      for (ActionListener actionListener : actionListenerSet)
      {
        actionListener.add(action);
        actionListener.refresh(actionList);
      }
    }
  }

  /** remove action entries
   * @param text text
   */
  public void removeAction(Action action)
    throws SQLException
  {
//TODO: NYI
    synchronized(lock)
    {
    }

    // trigger listeners
    for (ActionListener actionListener : actionListenerSet)
    {
      actionListener.remove(action);
      actionListener.refresh(actionList);
    }
  }

  /** add tag
   * @param tag tag
   */
  public void addTag(Tag tag)
    throws SQLException
  {
    // add tag to list and update
    synchronized(lock)
    {
      tagList.add(tag);
      updateTags(tagMap,tagList);
    }

    // trigger listeners
    for (ActionListener actionListener : actionListenerSet)
    {
      actionListener.refresh(tagList);
    }
  }

  /** update tag
   * @param tag tag to update
   */
  public void updateTag(Tag tag)
    throws SQLException
  {
    // update tag
    synchronized(lock)
    {
      updateTags(tagMap,tagList);
    }

    // trigger listeners
    for (ActionListener actionListener : actionListenerSet)
    {
      actionListener.refresh(tagList);
    }
  }

  /** remove tag
   * @param tag tag to remove
   */
  public void removeTag(Tag tag)
    throws SQLException
  {
    // remove from tag list and update
    synchronized(lock)
    {
      tagList.remove(tag);
      updateTags(tagMap,tagList);
    }

    // trigger listeners
    for (ActionListener actionListener : actionListenerSet)
    {
      actionListener.refresh(tagList);
    }
  }

//TODO: remove
  /** clear all tags
   */
  public void XclearTags()
  {
    synchronized(lock)
    {
      tagList.clear();
    }

    // update tag assignment in database
//    clearTags();

    // trigger listeners
    for (ActionListener actionListener : actionListenerSet)
    {
      actionListener.refresh(tagList);
    }
  }

  /** get tag for action
   * @param action action
   * @return tag or null
   */
  public Tag XgetAssignedTag(Action action)
  {
    for (Tag tag : tagList)
    {
//Dprintf.dprintf("tag test %s",tag);
      if ((tag.startTimestamp <= action.timestamp) && (action.timestamp <= tag.endTimestamp))
      {
        return tag;
      }
    }

    return null;
  }

  /** refresh action/tag lists
   */
  public void refresh()
  {
    // reload action/tag list
    synchronized(lock)
    {
      try
      {
        loadActions(actionList);
      }
      catch (SQLException exception)
      {
        // ignored
      }

      try
      {
        loadTags(tagMap,tagList);
      }
      catch (SQLException exception)
      {
        // ignored
      }
    }

    // trigger listeners
    for (ActionListener actionListener : actionListenerSet)
    {
      actionListener.refresh(actionList);
      actionListener.refresh(tagList);
    }
  }

  /** get nearest action
   * @param timestamp timestamp [s]
   * @return nearest action or null
   */
  public Action getNearest(long timestamp)
  {
    Action nearestAction = null;

    for (Action action : getActionList())
    {
      if (   (nearestAction == null)
          || (Math.abs(timestamp-action.timestamp) < Math.abs(timestamp-nearestAction.timestamp))
         )
      {
        nearestAction = action;
      }
    }

    return nearestAction;
  }

  /** get action list of today
   * @return action list
   */
  public ActionList getActionList()
  {
    // reload action list if day changed
    synchronized(lock)
    {
      if (!actionList.isToday())
      {
        try
        {
          loadActions(actionList);
        }
        catch (SQLException exception)
        {
          // ignored
        }

        try
        {
          loadTags(tagMap,tagList);
        }
        catch (SQLException exception)
        {
          // ignored
        }
      }
    }

    return actionList;
  }

  /** get tag list of today
   * @return tag list
   */
  public TagList getTagList()
  {
    // reload action list if day changed
    synchronized(lock)
    {
      if (!actionList.isToday())
      {
        try
        {
          loadActions(actionList);
        }
        catch (SQLException exception)
        {
          // ignored
        }

        try
        {
          loadTags(tagMap,tagList);
        }
        catch (SQLException exception)
        {
          // ignored
        }
      }
    }

    return tagList;
  }

  /** get tag name list
   * @return tag name list
   */
  public List<String> getTagNames()
  {
    return new ArrayList<String>(tagMap.keySet());
  }

  // -------------------------------------------------------------------

  /** create action monitor
   */
  protected ActionMonitor()
  {
    // initialize variables
    this.database = null;

    PreparedStatement preparedStatement = null;
    try
    {
      // open action database
      this.database = openLocalDatabase();

      // add shutdown hook for action database
      Runtime.getRuntime().addShutdownHook(new Thread()
      {
        public void run()
        {
          closeLocalDatabase(database);
        }
      });

      // load actions/tags
      loadActions(actionList);
      loadTags(tagMap,tagList);
    }
    catch (SQLException exception)
    {
      Timemine.printWarning("Cannot initialize action database (error: %s)",exception.getMessage());
      exception.printStackTrace();
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }

    // initial trigger listeners
    for (ActionListener actionListener : actionListenerSet)
    {
      actionListener.refresh(actionList);
    }
  }

  // -------------------------------------------------------------------

  /** open action database
   * @return database
   */
  private static Database openLocalDatabase()
    throws SQLException
  {
    Database database = null;
    try
    {
      // open databas
      database = new Database(Settings.TIMEMINE_DIRECTORY,"actions");
      initLocalDatabase(database);
    }
    catch (SQLException exception)
    {
      if (database != null) database.close();

      deleteLocalDatabase();

      database = new Database(Settings.TIMEMINE_DIRECTORY,"actions");
      initLocalDatabase(database);
    }

    // disable synchronous mode, journal
    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = database.prepareStatement("PRAGMA synchronous=OFF;");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("PRAGMA journal_mode=MEMROY;");
      ResultSet resultSet = preparedStatement.executeQuery();
      resultSet.close();
      preparedStatement.close(); preparedStatement = null;
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }

    return database;
  }

  /** close action database
   * @param database database
   */
  private static void closeLocalDatabase(Database database)
  {
    if (database != null)
    {
      try
      {
        database.close(); database = null;
      }
      catch (SQLException exception)
      {
        exception.printStackTrace();
      }
    }
  }

  /** init action database
   * @param database database
   */
  private static void initLocalDatabase(Database database)
    throws SQLException
  {
    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS meta ("+
                                                    "  id    INTEGER PRIMARY KEY,"+
                                                    "  name  TEXT UNIQUE,"+
                                                    "  value TEXT"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS actions ("+
                                                    "  id       INTEGER PRIMARY KEY,"+
                                                    "  type     INTEGER,"+
                                                    "  dateTime INTEGER DEFAULT (DATETIME('now')),"+
                                                    "  text     TEXT,"+
                                                    "  tagId    INTEGER DEFAULT "+Database.ID_NONE+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

//TODO: store in table?
      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS actionFilters ("+
                                                    "  id       INTEGER PRIMARY KEY,"+
                                                    "  text     TEXT"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS actionsIndex0 ON actions (dateTime,type,text)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS actionsIndex1 ON actions (type,text)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS tags ("+
                                                    "  id       INTEGER PRIMARY KEY,"+
                                                    "  name     TEXT UNIQUE,"+
                                                    "  dateTime INTEGER DEFAULT (DATETIME('now'))"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS tagsIndex0 ON tags (dateTime,name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS tagsIndex1 ON tags (name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }
  }

  /** delete database (disard all data and file!)
   */
  public static void deleteLocalDatabase()
  {
    new File(Settings.TIMEMINE_DIRECTORY,"actions.db").delete();
  }

  /** check if action is filtered
   * @return true iff filterted
   */
  private boolean isFilteredAction(String text)
  {
    Matcher matcher;
    for (Pattern actionFilterPattern : actionFilterPatterns)
    {
      if ((matcher = actionFilterPattern.matcher(text)).matches())
      {
        return true;
      }
    }

    return false;
  }

  /** load actions
   * @param actionList action list
   */
  private void loadActions(ActionList actionList)
    throws SQLException
  {
    actionList.clear();

    PreparedStatement preparedStatement = null;
    ResultSet         resultSet         = null;
    try
    {
      preparedStatement = database.prepareStatement("SELECT actions.type,"+
                                                    "       STRFTIME('%s',actions.dateTime) AS timestamp,"+
                                                    "       actions.text,"+
                                                    "       tagId "+
                                                    "FROM actions "+
                                                    "WHERE STRFTIME('%Y-%m-%d',actions.dateTime)=STRFTIME('%Y-%m-%d','now') "+
                                                    "ORDER BY actions.dateTime ASC"
                                                   );
      resultSet = preparedStatement.executeQuery();
      while (resultSet.next())
      {
        // get values
        ActionTypes type      = ActionTypes.values()[resultSet.getInt("type")];
        long        timestamp = resultSet.getLong  ("timestamp");
        String      text      = resultSet.getString("text"    );
        int         tagId     = resultSet.getInt   ("tagId"   );

        if (text != null)
        {
          if (!isFilteredAction(text))
          {
            // add action
            Action action = new Action(type,
                                       timestamp,
                                       text,
                                       tagId
                                      );
//Dprintf.dprintf("add action=%s",action);
            actionList.add(action);
          }
        }
      }
      resultSet.close(); resultSet = null;
      preparedStatement.close(); preparedStatement = null;
    }
    finally
    {
      try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }
  }

  /** load tags
   * @param tagMap tag map
   * @param tagList tag list
   */
  private void loadTags(HashMap<String,Integer> tagMap, TagList tagList)
    throws SQLException
  {
    tagMap.clear();
    tagList.clear();

    PreparedStatement preparedStatement = null;
    ResultSet         resultSet         = null;
    try
    {
      // get tag names
      preparedStatement = database.prepareStatement("SELECT id,"+
                                                    "       name "+
                                                    "FROM tags "+
                                                    "ORDER BY name ASC"
                                                   );
      resultSet = preparedStatement.executeQuery();
      while (resultSet.next())
      {
        // get values
        int    id   = resultSet.getInt   ("id"  );
        String name = resultSet.getString("name");

        // add tag name
        tagMap.put(name,id);
      }
      resultSet.close(); resultSet = null;
      preparedStatement.close(); preparedStatement = null;

      // get tag list
      preparedStatement = database.prepareStatement("SELECT STRFTIME('%s',actions.dateTime) AS timestamp,"+
                                                    "       actions.text,"+
                                                    "       tagId,"+
                                                    "       tags.name AS tagName "+
                                                    "FROM actions "+
                                                    "  LEFT JOIN tags ON tags.id=actions.tagId "+
                                                    "WHERE STRFTIME('%Y-%m-%d',actions.dateTime)=STRFTIME('%Y-%m-%d','now') "+
                                                    "ORDER BY actions.dateTime ASC"
                                                   );
      int    lastTagId      = Database.ID_NONE;
      String lastTagName    = null;
      long   startTimestamp = 0L;
      resultSet = preparedStatement.executeQuery();
      while (resultSet.next())
      {
        // get values
        long   timestamp = resultSet.getLong  ("timestamp");
        String text      = resultSet.getString("text"     );
        int    tagId     = resultSet.getInt   ("tagId"    );
        String tagName   = resultSet.getString("tagName"  );

        if (text != null)
        {
          if (lastTagId != tagId)
          {
            if ((lastTagId != Database.ID_NONE) && (lastTagName != null) && (timestamp > startTimestamp))
            {
              // add tag
              Tag tag = new Tag(startTimestamp,
                                timestamp-1,
                                lastTagName
                               );
              tagList.add(tag);
            }

            lastTagId      = tagId;
            lastTagName    = tagName;
            startTimestamp = timestamp;
          }
        }
      }
      if ((lastTagId != Database.ID_NONE) && (lastTagName != null))
      {
        // add tag
        Tag tag = new Tag(startTimestamp,
//???
                          startTimestamp+1,
                          lastTagName
                         );
Dprintf.dprintf("db2 add tag %d %d",startTimestamp,startTimestamp+60);
        tagList.add(tag);
      }
      resultSet.close(); resultSet = null;
      preparedStatement.close(); preparedStatement = null;
    }
    finally
    {
      try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }
  }

  /** store tags
   * @param tagList tag list
   */
  private void storeTags(TagList taglist)
    throws SQLException
  {
    // clear tag assignment in database
    clearTags();

    // update tag assignment in database
    PreparedStatement preparedStatement = null;
    ResultSet         resultSet         = null;
    for (Tag tag : tagList)
    {
      try
      {
        // try insert tag name (if not exists)
        preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO tags (name) VALUES (?);");
        preparedStatement.setString(1,tag.name);
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        // get tag id
        int tagId = Database.ID_NONE;
        preparedStatement = database.prepareStatement("SELECT id FROM tags WHERE name=? LIMIT 0,1;");
        preparedStatement.setString(1,tag.name);
        resultSet = preparedStatement.executeQuery();
        if (resultSet.next())
        {
          tagId = resultSet.getInt("id");
        }
        resultSet.close(); resultSet = null;
        preparedStatement.close(); preparedStatement = null;
        tagMap.put(tag.name,tagId);

        // update tag id in all actions which belongs to this tag
        if ((tagId != Database.ID_NONE) && (tag.startTimestamp > 0L) && (tag.endTimestamp > 0L))
        {
          preparedStatement = database.prepareStatement("UPDATE actions SET tagId=? WHERE (DATETIME(?,'unixepoch')<=dateTime) AND (dateTime<DATETIME(?,'unixepoch'));");
          preparedStatement.setInt(1,tagId);
          preparedStatement.setLong(2,tag.startTimestamp);
          preparedStatement.setLong(3,tag.endTimestamp);
          preparedStatement.executeUpdate();
          preparedStatement.close(); preparedStatement = null;
        }
      }
      finally
      {
        try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }
  }

  /** update tags
   * @param tagMap tag map
   * @param tagList tag list
   */
  private void updateTags(HashMap<String,Integer> tagMap, TagList tagList)
    throws SQLException
  {
    storeTags(tagList);
    loadTags(tagMap,tagList);
  }

  /** clear all tags
   */
  private void clearTags()
    throws SQLException
  {
    PreparedStatement preparedStatement = null;
    ResultSet         resultSet         = null;
    try
    {
      preparedStatement = database.prepareStatement("UPDATE actions SET tagId=? WHERE STRFTIME('%Y-%m-%d',actions.dateTime)=STRFTIME('%Y-%m-%d','now');");
      preparedStatement.setInt(1,Database.ID_NONE);
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
    }
    finally
    {
      try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }
  }
}

/** Action monitor Unix X11
 */
class ActionMonitorUnix extends ActionMonitor implements ActionMonitorInterface
{
  interface XLib extends Library
  {
//    XLib INSTANCE = (XLib) Native.loadLibrary("XLib", XLib.class);
    XLib INSTANCE = (XLib) Native.loadLibrary("X11", XLib.class);

//    int XGetInputFocus(X11.Display display, X11.Window focus_return, Pointer revert_to_return);
    int XGetInputFocus(X11.Display display, Pointer focus_return, Pointer revert_to_return);
  }

  // https://www.google.de/url?sa=t&rct=j&q=&esrc=s&source=web&cd=2&cad=rja&uact=8&ved=0ahUKEwi4_JXvoJPNAhVmB8AKHXgvAK4QFgglMAE&url=http%3A%2F%2Fsvn.chorem.org%2Fjtimer%2Ftrunk%2Fsrc%2Fmain%2Fjava%2Forg%2Fchorem%2Fjtimer%2Fui%2Fsystem%2Funix%2FXss.java&usg=AFQjCNGCPGuphUh11-KHy4c8Vx8QumtSAA&bvm=bv.123664746,d.ZGg
  /**
   * libXss JNA interface.
   *
   * Incomplete definition.
   *
   * @author chatellier
   * @version $Revision$
   *
   * Last update : $Date$
   * By : $Author$
   */
  interface Xss extends Library
  {
    /** Xss Instance */
    Xss INSTANCE = (Xss)Native.loadLibrary("Xss",Xss.class);

    /**
     * XScreenSaverInfo struct
     */
    public static class XScreenSaverInfo extends Structure
    {
      public X11.Window window; /* screen saver window - may not exist */
      public int state; /* ScreenSaverOff, ScreenSaverOn, ScreenSaverDisabled */
      public int kind; /* ScreenSaverBlanked, ...Internal, ...External */
      public NativeLong til_or_since; /* time til or since screen saver */
      public NativeLong idle; /* total time since last user input */
      public NativeLong eventMask; /* currently selected events for this client */
      @Override
      protected List getFieldOrder()
      {
        return Arrays.asList(new String[] {"window","state","kind","til_or_since","idle","eventMask"});
      }
    }

    /**
     * Information sur l'activation de l'ecran de veille de X11.
     *
     * @param display display
     * @param event_base event_base
     * @param error_base error_base
     *
     * @return true if XScreenSaver extension is loaded
     *
     * @see IntByReference
     */
    public boolean XScreenSaverQueryExtension(X11.Display display, IntByReference event_base, IntByReference error_base);

    /**
     * XScreenSaverAllocInfo.
     *
     * @return XScreenSaverInfo instance
     */
    public XScreenSaverInfo XScreenSaverAllocInfo();

    /**
     * XScreenSaverQueryInfo.
     *
     * @param display
     * @param drawable
     * @param infos
     * @return status flag
     */
    public int XScreenSaverQueryInfo(X11.Display display, X11.Drawable drawable, XScreenSaverInfo infos);
  }

interface CLibrary extends Library
{
  CLibrary INSTANCE = (CLibrary) Native.loadLibrary("c",CLibrary.class);
  int getpid();
  int getppid();
  long time(long buf[]);
}

  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------
  private final static X11         x11     = X11.INSTANCE;
  private final static XLib        xlib    = XLib.INSTANCE;
  private final static X11.Display display = x11.XOpenDisplay(null);

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get title of current active window
   * @return window title
   */
  @Override
  public String getCurrentWindowTitle()
  {
    String windowTitle;
//  System.out.println(CLibrary.INSTANCE.getpid());

    // get focused window
    final X11.Window window = getCurrentWindow();

    // get window title
    X11.XTextProperty xTextProperty = new X11.XTextProperty();
    x11.XGetWMName(display,window,xTextProperty);
    windowTitle = (String)xTextProperty.readField("value");
//if (name.isEmpty()) Dprintf.dprintf("X11.XTextProperty=%s",xTextProperty);

    return (windowTitle != null) ? windowTitle : "";
  }

  /** get current active application name
   * @return application name
   */
  @Override
  public String getCurrentApplication()
  {
    String applicatioName = "";

    // get focused window
    final X11.Window window = getCurrentWindow();

    // get active process id
    AtomByReference atomReference = new AtomByReference();
    IntByReference retFmtReference = new IntByReference();
    NativeLongByReference nItemsReference = new NativeLongByReference();
    NativeLongByReference byteAfterRetReference = new NativeLongByReference();
    PointerByReference propertyReference = new PointerByReference();
    if (  (x11.XGetWindowProperty(display,
                                     window,
                                     x11.XInternAtom(display,"_NET_WM_PID",false),
                                     new NativeLong(0),
                                     new NativeLong(1),
                                     false,
                                     X11.XA_CARDINAL,
                                     atomReference,
                                     retFmtReference,
                                     nItemsReference,
                                     byteAfterRetReference,
                                     propertyReference
                                    ) != X11.Success)
       || (atomReference.getValue() == null)
      )
    {
      return "";
    }
    // got a positive answer.
    // LOG.debug(String.format("retType[%d/%s] fmt[%d] nitem[%d] bAfter[%d] >>  val[%d]",
    // retType.getValue().intValue(), x11
    // .XGetAtomName(display, retType.getValue()),
    // retFmt.getValue(),
    // nItems.getValue().intValue(),
    // byteAfterRet.getValue().intValue(),
    // propRet.getValue().getInt(0)));
//Dprintf.dprintf("PID %d",propertyReference.getValue().getInt(0));
    int pid = propertyReference.getValue().getInt(0);

    // get process executable binary name
    try
    {
      File file = new File("/proc/"+Integer.toString(pid)+"/exe");
      applicatioName = file.getCanonicalPath();
    }
    catch (IOException exception)
    {
      return "";
    }

    return applicatioName;
  }

  /** check if system is idle
   * @return idle time [ms]
   */
  @Override
  public long getIdleTime()
  {
    long idleTime = 0;

    X11.Display display = X11.INSTANCE.XOpenDisplay(null);
    if (display != null)
    {
      // check if screensaver extension is enabled
      IntByReference eventBaseReference = new IntByReference();
      IntByReference errorBaseReference = new IntByReference();
      if (Xss.INSTANCE.XScreenSaverQueryExtension(display,eventBaseReference,errorBaseReference))
      {
        // screensaver is enabled
        Xss.XScreenSaverInfo screenSaverInfo = Xss.INSTANCE.XScreenSaverAllocInfo();
        if (screenSaverInfo != null)
        {
          // get root window
          X11.Window rootWindow = X11.INSTANCE.XDefaultRootWindow(display);
          if (rootWindow != null)
          {
            // FIX : without this, cause random NPE
            screenSaverInfo.window = rootWindow;

            Xss.INSTANCE.XScreenSaverQueryInfo(display,rootWindow,screenSaverInfo);
            if (screenSaverInfo.idle != null)
            {
              idleTime = screenSaverInfo.idle.longValue();
            }
          }
          X11.INSTANCE.XFree(screenSaverInfo.getPointer());
        }
      }
    }
    X11.INSTANCE.XCloseDisplay(display);

    return idleTime;
  }

  /** get current active window
   * @return window
   */
  private X11.Window getCurrentWindow()
  {
    // get focused window
    final WindowByReference windowReference = new WindowByReference();
    final IntByReference    intReference    = new IntByReference();
    xlib.XGetInputFocus(display,windowReference.getPointer(),intReference.getPointer());
//Dprintf.dprintf("window=%s",windowReference.getValue());
    return windowReference.getValue();
  }
}

/** Action monitor Windows
 */
class ActionMonitorWindows extends ActionMonitor implements ActionMonitorInterface
{
  interface User extends User32
  {
    User INSTANCE = (User) Native.loadLibrary("user32",User.class,new HashMap<Object,Object>()
    {
      {
        put(OPTION_TYPE_MAPPER,    W32APITypeMapper.UNICODE    );
        put(OPTION_FUNCTION_MAPPER,W32APIFunctionMapper.UNICODE);
      }
    });

    int GetWindowTextA(PointerType hWnd, byte[] lpString, int nMaxCount);

    // http://stackoverflow.com/questions/5132431/java-jna-systemparametersinfo-parameters-type
    //BOOL WINAPI SystemParametersInfo(
    // __in UINT uiAction,
    // __in UINT uiParam,
    // __in_out PVOID pvParam,
    // __in UINT fWinIni
    //);
    boolean SystemParametersInfo(
            int uiAction,
            int uiParam,
            Pointer pvParam,
            int fWinIni
    );
  }

  interface Psapi extends StdCallLibrary
  {
    Psapi INSTANCE = (Psapi)Native.loadLibrary("Psapi",Psapi.class);

    WinDef.DWORD GetModuleBaseNameA(Pointer hProcess, Pointer hModule, byte[] lpBaseName, int nSize);
    WinDef.DWORD GetModuleBaseNameW(Pointer hProcess, Pointer hModule, char[] lpBaseName, int nSize);
  }

  // --------------------------- constants --------------------------------
  private final int BUFFER_SIZE               = 512;

  private final int PROCESS_VM_READ           = 0x0010;
  private final int PROCESS_QUERY_INFORMATION = 0x0400;

  private final int SPI_GETSCREENSAVERRUNNING = 114;

  // --------------------------- variables --------------------------------
  private final Kernel32 kernel32 = Kernel32.INSTANCE;
  private final User     user     = User.INSTANCE;
  private final Psapi    psapi    = Psapi.INSTANCE;

  private long idleStartTimestamp = -1;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get title of current active window
   * @return window title
   */
  @Override
  public String getCurrentWindowTitle()
  {
    String windowTitle = "";

    byte[] buffer = new byte[BUFFER_SIZE];
    user.GetWindowTextA(user.GetForegroundWindow(), buffer, BUFFER_SIZE);
    System.out.println("Active window title: " + Native.toString(buffer));
    windowTitle = Native.toString(buffer);

    return (windowTitle != null) ? windowTitle : "";
  }

  /** get current active application name
   * @return application name
   */
  @Override
  public String getCurrentApplication()
  {
    String applicatioName = "";

    // get process executable binary name
    char[] buffer = new char[BUFFER_SIZE];
    IntByReference pid = new IntByReference();
    user.GetWindowThreadProcessId(user.GetForegroundWindow(),pid);
    WinNT.HANDLE processHandle = kernel32.OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_VM_READ,true,pid.getValue());
    if (processHandle != null)
    {
      Psapi.INSTANCE.GetModuleBaseNameW(processHandle.getPointer(),Pointer.NULL,buffer,BUFFER_SIZE);
      applicatioName = Native.toString(buffer);
      if (applicatioName.endsWith("wwahost.exe"))
      { // Metro App
        // There is no stable API to get the current Metro app
        // But you can guestimate the name form the current directory of the process
        // To query this, see:
        // http://stackoverflow.com/questions/16110936/read-other-process-current-directory-in-c-sharp
      }
//System.out.println(applicatioName);

/*
        char[] buffer = new char[MAX_TITLE_LENGTH * 2];
        GetWindowTextW(GetForegroundWindow(), buffer, MAX_TITLE_LENGTH);
        System.out.println("Active window title: " + Native.toString(buffer));

        PointerByReference pointer = new PointerByReference();
        GetWindowThreadProcessId(GetForegroundWindow(), pointer);
        Pointer process = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, false, pointer.getValue());
        GetModuleBaseNameW(process, null, buffer, MAX_TITLE_LENGTH);
        System.out.println("Active window process: " + Native.toString(buffer));
*/
    }

    return applicatioName;
  }

  /** check if system is idle
   * @return idle time [ms]
   */
  @Override
  public long getIdleTime()
  {
    if (isScreenSaverRunning())
    {
      if (idleStartTimestamp < 0)
      {
        idleStartTimestamp = System.currentTimeMillis();
      }
//Dprintf.dprintf("%d",System.currentTimeMillis()-idleStartTimestamp);

      return System.currentTimeMillis()-idleStartTimestamp;
    }
    else
    {
      idleStartTimestamp = -1;

      return 0L;
    }
  }

  boolean isScreenSaverRunning()
  {
    IntByReference pvParamReference = new IntByReference(0);

    boolean ok = user.SystemParametersInfo(SPI_GETSCREENSAVERRUNNING, 0, pvParamReference.getPointer(), 0);
    if(!ok) throw new RuntimeException("SystemParametersInfo call failed");
    return pvParamReference.getValue() != 0;
  }
}

/** Action monitor MacOSX
 */
class ActionMonitorMacOSX extends ActionMonitor implements ActionMonitorInterface
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get title of current active window
   * @return window title
   */
  @Override
  public String getCurrentWindowTitle()
  {
    String windowTitle = "";

    final String script = "tell application \"System Events\"\n" +
                          "\tname of application processes whose frontmost is tru\n" +
                          "end";
    ScriptEngine appleScript = new ScriptEngineManager().getEngineByName("AppleScript");
    try
    {
      windowTitle = (String)appleScript.eval(script);
    }
    catch (Exception exception)
    {
Dprintf.dprintf("");
    }
System.out.println(windowTitle);

    return (windowTitle != null) ? windowTitle : "";
  }

  /** get current active application name
   * @return application name
   */
  @Override
  public String getCurrentApplication()
  {
    String applicatioName = "";

    return applicatioName;
  }

  /** check if system is idle
   * @return idle time [ms]
   */
  @Override
  public long getIdleTime()
  {
    return 0L;
  }
}

/* end of file */
