/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: Backend driver
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.IOException;
import java.io.FileReader;
import java.io.File;
import java.io.InputStream;
import java.io.FileInputStream;

import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.ProtocolException;
import java.net.UnknownHostException;
import java.net.URL;

import java.security.KeyManagementException;
import java.security.NoSuchProviderException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import java.security.cert.Certificate;
//import java.security.cert.CertificateException;
//import java.security.cert.X509Certificate;
//import java.security.interfaces.RSAPublicKey;
//import java.security.InvalidKeyException;
//import java.security.Key;
//import java.security.KeyFactory;
//import java.security.KeyManagementException;
import java.security.KeyPair;
import java.security.KeyStore;
import java.security.KeyStoreException;
//import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
//import java.security.PublicKey;
import java.security.Security;
//import java.security.spec.RSAPublicKeySpec;
import java.security.UnrecoverableKeyException;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.X509TrustManager;

import java.util.Enumeration;
import java.util.Random;

import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.X509CertificateHolder;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.openssl.jcajce.JcaPEMKeyConverter;
import org.bouncycastle.openssl.jcajce.JcePEMDecryptorProviderBuilder;
import org.bouncycastle.openssl.PEMDecryptorProvider;
import org.bouncycastle.openssl.PEMEncryptedKeyPair;
import org.bouncycastle.openssl.PEMKeyPair;
import org.bouncycastle.openssl.PEMParser;
import org.bouncycastle.asn1.pkcs.PrivateKeyInfo;

// graphics
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;

// Apache commons
import org.apache.commons.codec.binary.Base64;

/****************************** Classes ********************************/

/** backend driver functions
 */
public abstract class BackendDriver
{
  // --------------------------- constants --------------------------------
  public final int READ_TIMEOUT = 60*1000;  // read timeout [ms]

  // --------------------------- variables --------------------------------
  public    final Image                       logo;

  protected final String                      serverName;
  protected final int                         serverPort;
  protected final boolean                     serverUseSSL;
  protected final String                      authorization;
  protected final Settings.CertificateTypes   certificateType;
  protected final String                      certificateFileName;
  protected final String                      keyFileName;
  protected final String                      p12FileName;
  protected final String                      keyPassword;

  // for debugging
  protected final Display                     display = Display.getDefault();
  protected final Shell                       shell   = new Shell(display,SWT.SHELL_TRIM);

  private static  JcaX509CertificateConverter certificateConverter;
  private static  byte[]                      RANDOM_DATA = new byte[64];

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  static
  {
    // add BouncyCastle as a security provider
    Security.addProvider(new BouncyCastleProvider());
    certificateConverter = new JcaX509CertificateConverter().setProvider("BC");

    // init random data
    Random random = new Random(System.currentTimeMillis());
    random.nextBytes(RANDOM_DATA);
  }

  /** initialize backend
   * @param serverName server name
   * @param serverPort server port
   * @param serverUseSSL true iff server use SSL
   * @param loginName login name
   * @param loginPasssword login password
   * @param  certificateType certificate type
   * @param certificateFileName certificate file (.cert, can be null)
   * @param keyFileName key file (.pem, can be null)
   * @param p12FileName p12 file name (.p12, can be null)
   * @param keyPassword key password
   * @param logo logo image
   */
  public BackendDriver(String                    serverName,
                       int                       serverPort,
                       boolean                   serverUseSSL,
                       String                    loginName,
                       String                    loginPassword,
                       Settings.CertificateTypes certificateType,
                       String                    certificateFileName,
                       String                    keyFileName,
                       String                    p12FileName,
                       String                    keyPassword,
                       Image                     logo
                      )
  {
    this.serverName          = serverName;
    this.serverPort          = serverPort;
    this.serverUseSSL        = serverUseSSL;
    this.authorization       = "Basic "+new String(Base64.encodeBase64((loginName+":"+loginPassword).getBytes()));
    this.certificateType     = certificateType;
    this.certificateFileName = certificateFileName;
    this.keyFileName         = keyFileName;
    this.p12FileName         = p12FileName;
    this.keyPassword         = keyPassword;
    this.logo                = logo;
  }

  /** initialize backend
   */
  public BackendDriver()
  {
    this("",  // serverName
         0,  // serverPort
         false,  // serverUseSSL
         "",  // loginName
         "",  // loginPassword
         Settings.CertificateTypes.NONE,
         (String)null,  // certificateFileName
         (String)null,  // keyFileName
         (String)null,  // p12FileName
         (String)null,  // keyPassword
         (Image)null  // logo
        );
  }

  /** reinit call-back
   */
  public void reinit()
  {
    // nothting to do
  }

  /** get backend type
   * @return backend type
   */
  abstract public Backend.Types getType();

  /** check iff epic name is supported for trackers
   * @return true iff epic name is supported for trackers
   */
  public boolean isEpicNameSupported()
  {
    return false;
  }

  /** check iff activity in time entries is supported
   * @return true iff activity in time entries is supported
   */
  public boolean isActivitySupported()
  {
    return false;
  }

  /** check iff edit progress on issues is supported
   * @return true iff setting progress on issues is supported
   */
  public boolean isEditProgressSupported()
  {
    return false;
  }

  //--------------------------------------------------------------------

  /** load issue from backend server
   * @param issueId issue backend id
   * @return issue
   */
  public abstract Issue loadIssue(Id issueId)
    throws BackendException;

  /** store issue on backend server
   * @param issue issue to store
   */
  public abstract void storeIssue(final Issue issue)
    throws BackendException;

  /** update issue on backend server
   * @param issue issue to update
   * @param originIssue origin issue
   */
  public abstract void updateIssue(final Issue issue, final Issue originIssue)
    throws BackendException;

  /** delete issue on backend server
   * @param issue issue to delete
   */
  public abstract void deleteIssue(final Issue issue)
    throws BackendException;

  //--------------------------------------------------------------------

  /** load issue from backend server
   * @param projectId project backend id
   * @param issueId issue backend id
   * @param timeEntryId time entry backend id
   * @returntie entry
   */
  public abstract TimeEntry loadTimeEntry(Id projectId, Id issueId, Id timeEntryId)
    throws BackendException;

  /** load time entry from backend server
   * @param projectId project backend id
   * @param issueId issue backend id
   * @param timeEntryId time entry backend id
   * @param clazz TimeEntry class
   */
//// TODO: remove
//  public abstract TimeEntry load(Id projectId, Id issueId, Id timeEntryId, Class<TimeEntry> clazz)
//    throws BackendException;

  /** store time entry on backend server
   * @param timeEntry time entry to store
   */
  public abstract void storeTimeEntry(final TimeEntry timeEntry)
    throws BackendException;

  /** update time entry on backend server
   * @param timeEntry time entry to update
   * @param originTimeEntry origin time entry
   */
  public abstract void updateTimeEntry(final TimeEntry timeEntry, final TimeEntry originTimeEntry)
    throws BackendException;

  /** delete time entry on backend server
   * @param timeEntry time entry to delete
   * @param originTimeEntry origin time entry
   */
  public abstract void deleteTimeEntry(final TimeEntry timeEntry)
    throws BackendException;

  //--------------------------------------------------------------------

  /** get users from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getUsers(final BackendProgress backendProgress, final EntityHandler<User> entityHandler)
    throws BackendException, InterruptedException;

  /** get current user from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   * @return current user or User.NONE
   */
  public abstract User getCurrentUser(final BackendProgress backendProgress, final EntityHandler<User> entityHandler)
    throws BackendException;

  /** get current user from backend
   * @param backendProgress progress handler or null
   * @return current user or User.NONE
   */
  public User getCurrentUser(BackendProgress backendProgress)
    throws BackendException
  {
    return getCurrentUser(backendProgress,new EntityHandler<User>()
    {
      @Override
      public void handle(User user)
      {
      }
      @Override
      public void done(User user)
        throws BackendException
      {
      }
    });
  }

  /** get current user from backend
   * @return current user or User.NONE
   */
  public User getCurrentUser()
    throws BackendException
  {
    return getCurrentUser((BackendProgress)null);
  }

  /** get roles from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getRoles(final BackendProgress backendProgress, final EntityHandler<Role> entityHandler)
    throws BackendException, InterruptedException;

  /** get trackers from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getTrackers(final BackendProgress backendProgres, final EntityHandler<Tracker> entityHandler)
    throws BackendException, InterruptedException;

  /** get epic names from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getEpicNames(final BackendProgress backendProgres, final EntityHandler<EpicName> entityHandler)
    throws BackendException, InterruptedException;

  /** get status from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getStatus(final BackendProgress backendProgress, final EntityHandler<Status> entityHandler)
    throws BackendException, InterruptedException;

  /** get priorities from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getPriorities(final BackendProgress backendProgress, final EntityHandler<Priority> entityHandler)
    throws BackendException, InterruptedException;

  /** get activities from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getActivities(final BackendProgress backendProgress, final EntityHandler<Activity> entityHandler)
    throws BackendException, InterruptedException;

  /** get projects from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getProjects(final BackendProgress backendProgress, final EntityHandler<Project> entityHandler)
    throws BackendException, InterruptedException;

  /** get members from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getMembers(final BackendProgress backendProgress, final EntityHandler<Member> entityHandler)
    throws BackendException, InterruptedException;

  /** get current member from backend
   * @param backendProgress progress handler or null
   * @param projectId project id
   * @param entityHandler entity handler
   * @return current member or Member.NONE
   */
  public abstract Member getCurrentMember(final BackendProgress backendProgress, final Id projectId, final EntityHandler<Member> entityHandler)
    throws BackendException, InterruptedException;

  /** get current member from backend
   * @param backendProgress progress handler or null
   * @param projectId project id
   * @return current member or Member.NONE
   */
  public Member getCurrentMember(BackendProgress backendProgress, Id projectId)
    throws BackendException, InterruptedException
  {
    final Member currentMember[] = new Member[1];

    getCurrentMember(backendProgress,projectId,new EntityHandler<Member>()
    {
      @Override
      public void handle(Member member)
      {
      }
      @Override
      public void done(Member member)
        throws BackendException
      {
        currentMember[0] = member;
      }
    });

    return currentMember[0];
  }

  /** get current member from backend
   * @param projectId project id
   * @return current member or Member.NONE
   */
  public Member getCurrentMember(Id projectId)
    throws BackendException, InterruptedException
  {
    return getCurrentMember((BackendProgress)null,projectId);
  }

  /** get versions from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public abstract void getVersions(final BackendProgress backendProgress, final EntityHandler<Version> entityHandler)
    throws BackendException, InterruptedException;

  /** get issues from backend
   * @param backendProgress progress handler or null
   * @param lastUpdatedOnTimestamp last updated time stamp for incremental update or 0
   * @param entityHandler entity handler
   */
  public abstract void getIssues(final BackendProgress backendProgress, long lastUpdatedOnTimestamp, final EntityHandler<Issue> entityHandler)
    throws BackendException, InterruptedException;

  /** get status transitions from backend
   * @param backendProgress progress handler or null
   * @param issueId issue id
   * @param entityHandler entity handler
   */
  public void getStatusTransitions(final BackendProgress backendProgress, final Id issueId, final EntityHandler<StatusTransition> entityHandler)
    throws BackendException, InterruptedException
  {
    // default: nothing to do
  }

  /** get time entry from backend
   * @param backendProgress progress handler or null
   * @param timeEntryId time entry id
   * @return time entry
   */
// TODO: remove or rename load -> getTimeEntry
//  public abstract TimeEntry getTimeEntry(final BackendProgress backendProgress, final long createdOnTimestamp, final EntityHandler<TimeEntry> entityHandler)
//    throws BackendException, InterruptedException;

  /** get time entries from backend
   * @param backendProgress progress handler or null
   * @param createOnTimestamp create-on timestamp
   * @param entityHandler entity handler
   */
  public abstract void getTimeEntries(final BackendProgress backendProgress, final long createdOnTimestamp, final EntityHandler<TimeEntry> entityHandler)
    throws BackendException, InterruptedException;

  /** connect to server via http/https
   * @param urlString URL string
   * @param requestMethod request method string
   * @param doOutput true for process output
   * @param arguments arguments
   */
  protected HttpURLConnection getConnection(String urlString, String requestMethod, boolean doOutput, String... arguments)
    throws BackendException
  {
    HttpURLConnection connection = null;

    try
    {
      if (serverUseSSL)
      {
        // get all possible certificate/key file names
        URL url = new URL("https://"+serverName+":"+serverPort+urlString+(((arguments != null) && (arguments.length > 0)) ? "?"+StringUtils.join(arguments,'&') : ""));
        HttpsURLConnection httpsConnection = (HttpsURLConnection)url.openConnection();
//TODO: check host

        switch (certificateType)
        {
          case NONE:
            break;
          case CERTIFICATE_KEY:
            if (   (certificateFileName != null)
                && !certificateFileName.isEmpty()
                && (keyFileName != null)
                && !keyFileName.isEmpty()
               )
            {
              httpsConnection.setSSLSocketFactory(getSocketFactory((File)null,  // certificateAuthorityFile,
                                                                   new File(certificateFileName),
                                                                   new File(keyFileName),
                                                                   keyPassword
                                                                  )
                                                 );
            }
            break;
          case P12:
            if (p12FileName != null)
            {
              httpsConnection.setSSLSocketFactory(getSocketFactory(new File(p12FileName),
                                                                   keyPassword
                                                                  )
                                                 );
            }
            break;
        }
        httpsConnection.setHostnameVerifier(Backend.ANY_HOSTNAME_VERIFIER);

        connection = httpsConnection;
      }
      else
      {
        URL url = new URL("http://"+serverName+":"+serverPort+urlString+(((arguments != null) && (arguments.length > 0)) ? "?"+StringUtils.join(arguments,'&') : ""));
        HttpURLConnection httpConnection = (HttpURLConnection)url.openConnection();
        connection = httpConnection;
      }

      connection.setReadTimeout(READ_TIMEOUT);
      connection.setRequestMethod(requestMethod);
      connection.setDoOutput(doOutput);
      connection.setRequestProperty("Authorization",authorization);
    }
    catch (ProtocolException exception)
    {
      throw new BackendException("Protocol error '"+serverName+"'");
    }
    catch (KeyManagementException exception)
    {
      throw new BackendException("Key management error '"+serverName+"'");
    }
    catch (NoSuchProviderException exception)
    {
      throw new BackendException("No such provider error '"+serverName+"'");
    }
    catch (NoSuchAlgorithmException exception)
    {
      throw new BackendException("No such algorithm error '"+serverName+"'");
    }
    catch (KeyStoreException exception)
    {
      throw new BackendException("Key store error '"+serverName+"'");
    }
    catch (UnrecoverableKeyException exception)
    {
      throw new BackendException("Key error '"+serverName+"'");
    }
    catch (CertificateException exception)
    {
      throw new BackendException("Certificate error '"+serverName+"'");
    }
    catch (UnknownHostException exception)
    {
      throw new BackendException("Unknown host '"+serverName+"'");
    }
    catch (IOException exception)
    {
      throw new BackendException("Connect to server fail: "+exception.getMessage());
    }

    return connection;
  }

  /** connect to server via http/https
   * @param urlString URL string
   * @param requestMethod request method string
   * @param doOutput true for process output
   */
  protected HttpURLConnection getConnection(String urlString, String requestMethod, boolean doOutput)
    throws BackendException
  {
    return getConnection(urlString,requestMethod,doOutput);
  }

  /** show confirm dialog
   * Note: for debugging only!
   * @param format format string
   * @param arguments optional arguments
   */
  protected boolean debugConfirm(final String format, final Object... arguments)
  {
    final boolean result[] = new boolean[]{false};
    display.syncExec(new Runnable()
    {
      public void run()
      {
        result[0] = Dialogs.confirm(shell,String.format(format,arguments));
      }
    });

    return result[0];
  }

  /** create TLS (SSL) socket factory with PEM files
   * original from: https://gist.github.com/rohanag12/07ab7eb22556244e9698
   * @param certificateAuthorityFile certificate authority PEM file
   * @param certificateFile certificate PEM file
   * @param keyFile PEM key file
   * @param password password or null
   * @return socket factory
   */
  public static SSLSocketFactory getSocketFactory(File   certificateAuthorityFile,
                                                  File   certificateFile,
                                                  File   keyFile,
                                                  String password
                                                 )
    throws KeyManagementException,
           NoSuchAlgorithmException,
           KeyStoreException,
           UnrecoverableKeyException,
           IOException,
           CertificateException
  {
    char[]                passwordChars;
    PEMParser             reader;
    X509CertificateHolder certificateHolder;
    KeyStore              keyStore;

    passwordChars = (password != null) ? password.toCharArray() : new char[0];

    // load certificate authority (CA) certificate
    if (certificateAuthorityFile != null)
    {
      reader = new PEMParser(new FileReader(certificateAuthorityFile));
      certificateHolder = (X509CertificateHolder)reader.readObject();
      reader.close();
      X509Certificate certificateAuthority = certificateConverter.getCertificate(certificateHolder);
    }

Dprintf.dprintf("certificateFile=%s",certificateFile);
    // load certificate
    reader = new PEMParser(new FileReader(certificateFile));
    certificateHolder = (X509CertificateHolder)reader.readObject();
    reader.close();
    X509Certificate certificate = certificateConverter.getCertificate(certificateHolder);

    // load private key
    reader = new PEMParser(new FileReader(keyFile));
    Object keyObject = reader.readObject();
    reader.close();
    PEMDecryptorProvider pemDecryptorProvider = new JcePEMDecryptorProviderBuilder().build(passwordChars);
    JcaPEMKeyConverter keyConverter = new JcaPEMKeyConverter().setProvider("BC");
//Dprintf.dprintf("keyObject=%s",keyObject);

    // init key pair
    PrivateKey privateKey;
    if      (keyObject instanceof PrivateKeyInfo)
    {
      privateKey = keyConverter.getPrivateKey((PrivateKeyInfo)keyObject);
    }
    else if (keyObject instanceof PEMEncryptedKeyPair)
    {
      privateKey = keyConverter.getKeyPair(((PEMEncryptedKeyPair)keyObject).decryptKeyPair(pemDecryptorProvider)).getPrivate();
    }
    else if (keyObject instanceof PEMKeyPair)
    {
      privateKey = keyConverter.getKeyPair((PEMKeyPair)keyObject).getPrivate();
    }
    else
    {
      throw new UnrecoverableKeyException();
    }

    // CA certificate used to authenticate server
/*
    keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
    keyStore.load(null,null);
    keyStore.setCertificateEntry("ca-certificate",certificateAuthority);
    TrustManagerFactory trustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
    trustManagerFactory.init(keyStore);
*/

    // certificate and client key for authentication
    keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
    keyStore.load(null,null);
    keyStore.setCertificateEntry("certificate",certificate);
    keyStore.setKeyEntry("private-key",
                         privateKey,
                         passwordChars,
                         new Certificate[]{certificate}
                        );
    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
    keyManagerFactory.init(keyStore,passwordChars);

    // create TLS (SSL) socket factory
    SSLContext sslContext = SSLContext.getInstance("TLSv1.2");
    sslContext.init(keyManagerFactory.getKeyManagers(),
null,//                    trustManagerFactory.getTrustManagers(),
                    null
                   );

    return sslContext.getSocketFactory();
  }

  public static SSLSocketFactory getSocketFactory(File   p12File,
                                                  String password
                                                 )
    throws KeyManagementException,
           NoSuchProviderException,
           NoSuchAlgorithmException,
           KeyStoreException,
           UnrecoverableKeyException,
           IOException,
           CertificateException
  {
    char[]                passwordChars;
    PEMParser             reader;
    X509CertificateHolder certificateHolder;
    KeyStore              keyStore;

    passwordChars = (password != null) ? password.toCharArray() : new char[0];

    InputStream stream = new FileInputStream(p12File);
    KeyStore store = KeyStore.getInstance("PKCS12","BC");
    store.load(stream, passwordChars);

    String alias = "";
    Enumeration<String> i = store.aliases();
    while (i.hasMoreElements())
    {
      alias = i.nextElement();
    }

    // init key pair
    PrivateKey privateKey = (PrivateKey)store.getKey(alias, passwordChars);
//Dprintf.dprintf("key=%s",privateKey);
    Certificate certificate = store.getCertificate(alias);
//Dprintf.dprintf("certificate=%s",certificate);

    // CA certificate used to authenticate server
/*
    keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
    keyStore.load(null,null);
    keyStore.setCertificateEntry("ca-certificate",certificateAuthority);
    TrustManagerFactory trustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
    trustManagerFactory.init(keyStore);
*/

    // certificate and client key for authentication
    keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
    keyStore.load(null,null);
    keyStore.setCertificateEntry("certificate",certificate);
    keyStore.setKeyEntry("private-key",
                         privateKey,
                         passwordChars,
                         new Certificate[]{certificate}
                        );
    KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
    keyManagerFactory.init(keyStore,passwordChars);

    // create TLS (SSL) socket factory
    SSLContext sslContext = SSLContext.getInstance("TLSv1.2");
    sslContext.init(keyManagerFactory.getKeyManagers(),
null,//                    trustManagerFactory.getTrustManagers(),
                    null
                   );

    return sslContext.getSocketFactory();
  }
}

/* end of file */
