/***********************************************************************\
*
* $Revision: 469 $
* $Date: 2017-02-22 09:25:36 +0100 (Wed, 22 Feb 2017) $
* $Author: torsten $
* Contents: JIRA backend implementation
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.lang.ref.SoftReference;
import java.io.BufferedReader;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.IOException;

import java.net.HttpURLConnection;
import java.net.ProtocolException;
import java.net.UnknownHostException;
import java.net.URL;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

// SSL
import javax.net.ssl.HttpsURLConnection;

// graphics
import org.eclipse.swt.graphics.Image;

// JSON
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;
import org.json.simple.parser.ContainerFactory;
//import org.json.simple.parser.ParseException;

// SQLite database
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

/****************************** Classes ********************************/

/** JIRA backend driver functions
 */
public class BackendDriverJIRA extends BackendDriver
{
  /** JIRA JSON object
   */
  private class JIRAObject extends JSONObject
  {
    JIRAObject parent;

    /** create JIRA oject
     * @param parent parent entry or null
     * @param name name
     */
    JIRAObject(JIRAObject parent, String name)
    {
      this.parent = parent;
      parent.put(name,this);
    }

    /** create JIRA oject
     * @param jiraArray JIRA array
     * @param name name
     */
    JIRAObject(JIRAArray parent, String name)
    {
      this.parent = parent.parent;
      JIRAObject entry = new JIRAObject(this);
      entry.put(name,this);
      parent.add(entry);
    }

    /** create JIRA object
     * @param parent parent entry or null
     */
    JIRAObject(JIRAObject parent)
    {
      this.parent = parent;
    }

    /** create JIRA object
     * @param parent parent entry or null
     */
    JIRAObject(JIRAArray parent)
    {
      this.parent = parent.parent;
      parent.add(this);
    }

    /** create JIRA object
     */
    JIRAObject()
    {
      this((JIRAObject)null);
    }
  };

  /** JIRA JSON array
   */
  private class JIRAArray extends JSONArray
  {
    JIRAObject parent;

    /** create JIRA array
     * @param parent parent entry or null
     */
    JIRAArray(JIRAObject parent, String name)
    {
      this.parent = parent;
      parent.put(name,this);
    }

    /** create JIRA array
     * @param jiraArray JIRA array
     * @param name name
     */
    JIRAArray(JIRAArray parent, String name)
    {
      this.parent = parent.parent;
      JIRAObject entry = new JIRAObject(parent);
      entry.put(name,this);
    }

    /** create JIRA array
     * @param parent parent entry or null
     */
    JIRAArray(JIRAObject parent)
    {
      this.parent = parent;
    }

    /** create JIRA array
     */
    JIRAArray()
    {
      this((JIRAObject)null);
    }

    /** add value
     * @param name name
     * @param value value
     */
    public void add(String name, String value)
    {
      JIRAObject jiraObject = new JIRAObject(parent);
      jiraObject.put(name,value);
      super.add(jiraObject);
    }

    /** add value
     * @param name name
     * @param value value
     */
    public void add(String name, long value)
    {
      JIRAObject jiraObject = new JIRAObject(parent);
      jiraObject.put(name,value);
      super.add(jiraObject);
    }
  };

  /** JIRA container factory
   */
  private class JIRAContainerFactory implements ContainerFactory
  {
    /** create JIRA array container
     * @return JIRA array container
     */
    public List<JIRAObject> creatArrayContainer()
    {
      return new JIRAArray();
    }

    /** create JIRA object container
     * @return JIRA object container
     */
    public Map<String,Object> createObjectContainer()
    {
      return new JIRAObject();
    }
  };

  /** element handler
   */
  private abstract class EntryHandler<T>
  {
    // --------------------------- constants --------------------------------
    // --------------------------- variables --------------------------------
//    private int     index;
public int     index;
    private int     totalCount;
    private T       result;
    private boolean done;

    // ------------------------ native functions ----------------------------

    // ---------------------------- methods ---------------------------------

    /** create entry handler
     * @param defaultResult default result value
     */
    EntryHandler(T defaultResult)
    {
      this.index      = 0;
      this.totalCount = 0;
      this.result     = defaultResult;
      this.done       = false;
    }

    /** create entry handler
     */
    EntryHandler()
    {
      this((T)null);
    }

    /** done iteration
     * @param result result value
     */
    public void done(T result)
    {
      this.result = result;
      this.done   = true;
    }

    /** done iteration
     */
    public void done()
    {
      done(null);
    }

    /** set entry index
     */
    protected void setIndex(int index)
    {
      this.index = index;
    }

    /** increment entry index
     */
    protected void incrementIndex()
    {
      this.index++;
    }

    /** set total count
     */
    protected void setTotalCount(int totalCount)
    {
      this.totalCount = totalCount;
    }

    /** process entry
     * @param entry entry
     * @param index index
     * @param totalCount total count
     * @return entity or null
     */
    protected T process(JIRAObject entry, int index, int totalCount)
      throws BackendException, InterruptedException
    {
      assert index < totalCount;

      T entity = parse(entry);
      if (entity != null)
      {
        store(entity,index,totalCount);
      }

      return entity;
    }

    /** process entry
     * @param entry entry
     * @return entity or null
     */
    protected T process(JIRAObject entry)
      throws BackendException, InterruptedException
    {
      /* Note: because JIRA is not able to deliver a correct "total"
               value in all request responses, "emulate" a missing
               value by the current index. See comment on Pagination
               in https://docs.atlassian.com/jira/REST/cloud/
      */
      int totalCount = this.totalCount; if (totalCount <= (index+1)) totalCount = index+1;
      return process(entry,index,totalCount);
    }

    /** parse entry
     * @param entry entry
     * @return entity or null
     */
    protected T parse(JIRAObject entry)
    {
//Dprintf.dprintf("-------");
//new Throwable().printStackTrace();
//throw new Error();
      return null;
    }

    /** process root entry handler
     * @param entry entry
     */
    protected void root(JIRAObject entry)
    {
    }

    /** store entity handler
     * @param entity entity to store
     * @param index index [0..n-1]
     * @param totalCount total count
     */
    protected void store(T entity, int index, int totalCount)
      throws BackendException
    {
//Dprintf.dprintf("-------");
//new Throwable().printStackTrace();
//throw new Error();
      store(entity);
    }

    /** store entity handler
     * @param index index [0..n-1]
     * @param totalCount total count
     * @param entity entity to store
     */
    protected void store(T entity)
      throws BackendException
    {
//Dprintf.dprintf("-------");
//new Throwable().printStackTrace();
//throw new Error();
    }

    /** check if done
     * @return true iff done
     */
    protected boolean isDone()
    {
      return done;
    }

    /** get result
     * @return result or null
     */
    protected T getResult()
    {
      return this.result;
    }

    /** handle error
     * @param throwable throwable
     */
    protected void error(Exception exception)
      throws BackendException
    {
      Timemine.printStackTrace(exception);
      throw new BackendException(exception);
    }
  }

  /** create entry handler
   */
  private abstract class CreateHandler<T extends Entity>
  {
    // --------------------------- constants --------------------------------

    // --------------------------- variables --------------------------------

    // ------------------------ native functions ----------------------------

    // ---------------------------- methods ---------------------------------

    /** create JIRA JSON entry
     */
    CreateHandler()
    {
    }

    /** create JSON data entry
     * @param rootEntry root element
     * @param entity entity
     */
    abstract public void data(JIRAObject rootEntry, T entity);

    // ----------------------------------------------------------------------
  }

  private class JIRAUser extends User
  {
    String key;

    /** create JIRA user
     * @param id backend id
     * @param key key
     * @param firstName first name
     * @param lastName last name
     * @param name name
     * @param mail mail address
     */
    JIRAUser(Id     id,
             String key,
             String firstName,
             String lastName,
             String name,
             String mail
            )
    {
      super(Backend.Types.JIRA,
            id,
            firstName,
            lastName,
            name,
            mail
           );
      this.key = key;
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      return "JIRAUser { id="+id+", key="+key+", firstName="+firstName+", lastName="+lastName+", name="+name+", mail="+mail+" }";
    }
  }

  // --------------------------- constants --------------------------------
  private final static String            DATABSE_NAME        = "jira-workarounds";

  private final static int               ENTRY_LIMIT         = 100; // max. 100 (see JIRA REST API)
  private final static SimpleDateFormat  DATE_FORMAT         = new SimpleDateFormat("yyyy-MM-dd");
  private final static SimpleDateFormat  DATETIME_FORMAT     = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
  private final static SimpleDateFormat  ISO_DATETIME_FORMAT = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");

  private final SoftReference<TimeEntry> TIME_ENTRY_NULL  = new SoftReference<TimeEntry>(null);

  private final String REST_API                              = "/rest/api/latest";
  private final String REST_GADGET                           = "/rest/gadget/latest";

  private final String FILTER_EQ                             = "%3D";
  private final String FILTER_GE                             = "%3E%3D";
  private final String FILTER_LE                             = "%3C%3D";

  // --------------------------- variables --------------------------------
  private static SimpleDateFormat       dateFormats[];

  private static User                   currentUser = null;
  private static SoftHashMap<Id,Member> currentMemberMap = new SoftHashMap<Id,Member>();
  private static Database               database;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  static
  {
    final String[] DATE_FORMATS = new String[]
    {
      "yyyy-MM-dd'T'HH:mm:ss'Z'",    // 2011-01-25T15:19:06Z
      "yyyy-MM-dd HH:mm:ss Z",       // 2011-01-25 15:19:06 +0900
      "yyyy-MM-dd HH:mm Z",          // 2011-01-25 15:19 +0900
      "yyyy/MM/dd HH:mm:ss",         // 2010/05/13 07:09:36
      "EEE MMM dd HH:mm:ss yyyy Z",  // Wed Dec 17 15:41:19 2008 +0100
      "yyyy-MM-dd",                  // 2011-01-25
      "dd-MMM-yy",                   // 25-Jan-11
    };

    dateFormats = new SimpleDateFormat[DATE_FORMATS.length];
    for (int i = 0; i < DATE_FORMATS.length; i++)
    {
      dateFormats[i] = new SimpleDateFormat(DATE_FORMATS[i]);
    }
  }

  /** initialize JIRA backend
   * @param serverName server name
   * @param serverPort server port
   * @param serverUseSSL true iff server use SSL
   * @param loginName login name
   * @param loginPasssword login password
   * @param certificateType certificate type
   * @param certificateFileName certificate file (.cert, can be null)
   * @param keyFileName key file (.pem, can be null)
   * @param p12FileName p12 file name (.p12, can be null)
   * @param keyPassword key password
   * @param logo logo image
   */
  public BackendDriverJIRA(String                    serverName,
                           int                       serverPort,
                           boolean                   serverUseSSL,
                           final String              loginName,
                           String                    loginPassword,
                           Settings.CertificateTypes certificateType,
                           String                    certificateFileName,
                           String                    keyFileName,
                           String                    p12FileName,
                           String                    keyPassword,
                           Image                     logo
                          )
    throws BackendException
  {
    super(serverName,
          serverPort,
          serverUseSSL,
          loginName,
          loginPassword,
          certificateType,
          certificateFileName,
          keyFileName,
          p12FileName,
          keyPassword,
          logo
         );

    // create JIRA work-around database for design lackes in JIRA, e. g. use "keys" instead of numerical ids
    try
    {
      database = new Database(Settings.TIMEMINE_DIRECTORY,DATABSE_NAME);
      initLocalDatabase();
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }

    // add shutdown hook to close work-around database on exit
    Runtime.getRuntime().addShutdownHook(new Thread()
    {
      @Override public void run()
      {
        try { database.close(); } catch (SQLException exception) { /* ignored */ }
      }
    });
  }

  /** reinit JIRA driver
   */
  public void reinit()
  {
    try
    {
      // close JIRA work-around database
      database.close();
      database = null;

      // discard JIRA work-around database
      Database.discard(Settings.TIMEMINE_DIRECTORY,DATABSE_NAME);

      // create new JIRA work-around database
      database = new Database(Settings.TIMEMINE_DIRECTORY,DATABSE_NAME);
      initLocalDatabase();
    }
    catch (SQLException exception)
    {
      // ignored
      Timemine.printInternalError(exception,"Init JIRA work-arounds database fail");
    }
  }

  /** get backend type
   * @return REDMINE
   */
  @Override
  public Backend.Types getType()
  {
    return Backend.Types.JIRA;
  }

  /** check iff epic name is supported for trackers
   * @return true iff epic name is supported for trackers
   */
  public boolean isEpicNameSupported()
  {
    return true;
  }

  //--------------------------------------------------------------------

  /** load issue from JIRA server
   * @param projectId project JIRA id
   * @param issueId issue JIRA id
   * @param clazz Issue class
   */
  @Override
  public Issue loadIssue(final Id issueId)
    throws BackendException
  {
    Issue issue = null;

    assert(!issueId.hasBackend());

    try
    {
      issue = getData(REST_API+"/issue/"+issueId.backend,"fields",new EntryHandler<Issue>()
      {
        @Override
        public Issue parse(JIRAObject entry)
        {
          return new Issue(Backend.Types.JIRA,
                           issueId,
                           getIdValue(entry,new String[]{"parent","id"}),
                           getIdValue(entry,new String[]{"project","id"}),
                           getIdValue(entry,new String[]{"tracker","id"}),
                           getIdValue(entry,new String[]{"status","id"}),
                           getIdValue(entry,new String[]{"priority","id"}),
                           userKeyToId(getValue(entry,new String[]{"reporter","key"})),  // authorId
                           getValue(entry,"summary",""),
                           getValue(entry,"description",""),
                           getDateValue(entry,"created"),
                           getDateValue(entry,"duedate"),
                           memberKeyToId(getValue(entry,new String[]{"assignee","key"}),Id.NONE),  // assignedToMemberId
                           getIntValue(entry,new String[]{"aggregateprogress","percent"},0),
                           getIdValue(entry,new String[]{"fixed_version","id"}),
                   //      = (getLongValue(entry,"timeoriginalestimate")+60-1)/60;
                           (getLongValue(entry,new String[]{"aggregateprogress","total"},0)+60-1)/60,
                           getDateValue(entry,"created"),
                           getDateValue(entry,"updated"),
                           getDateValue(entry,"closed_on")
                          );
        }
      });
    }
    catch (InterruptedException exception)
    {
      throw new BackendException(exception);
    }

    return issue;
  }

  /** store issue on JIRA server
   * @param issue issue to store
   */
  public void storeIssue(final Issue issue)
    throws BackendException
  {
    assert(issue.backendType == Backend.Types.JIRA);

    try
    {
      // create new issue
      final Tracker  tracker          = Backend.getTracker(Backend.TypeSet.JIRA,issue.trackerId);
      final Status   status           = Backend.getStatus(Backend.TypeSet.JIRA,issue.statusId);
      final Priority priority         = Backend.getPriority(Backend.TypeSet.JIRA,issue.priorityId);
      final Member   assignedToMember = Backend.getMember(Backend.TypeSet.JIRA,issue.assignedToMemberId);

      /* Note: bug in the JIRA documentation! 201 is return on success, not 200.
         https://developer.atlassian.com/static/rest/jira/6.1.html#d2e865
         Note: status cannot be set directly and must be set in second step via a transition
      */
      postData(REST_API+"/issue",HttpURLConnection.HTTP_CREATED,issue,new CreateHandler<Issue>()
      {
        @Override
        public void data(JIRAObject rootEntry, Issue issue)
        {
          JIRAArray  array;
          JIRAObject entry;

          JIRAObject fieldsEntry = new JIRAObject(rootEntry,"fields");

          entry = new JIRAObject(fieldsEntry,"project");
          entry.put("id",issue.projectId.backend);

          entry = new JIRAObject(fieldsEntry,"issuetype");
          entry.put("id",issue.trackerId.backend);

          if ((tracker != null) && tracker.requireEpicName && (issue.epicName != null))
          {
            String epicNameCustomField = getEpicNameCustomField();
            if (epicNameCustomField != null)
            {
              fieldsEntry.put(epicNameCustomField,issue.epicName);
            }
          }

          if ((priority != null) && !priority.name.isEmpty())
          {
            entry = new JIRAObject(fieldsEntry,"priority");
            entry.put("name",priority.name);
          }

          if (issue.summary != null)
          {
            fieldsEntry.put("summary",issue.summary);
          }

Dprintf.dprintf("TODOOOOOOOOOOOOOOOOOOOOOOOOOOO %s",issue.description);
/*
        if (issue.description != null)
        {
          fieldsEntry.put("description",issue.description);
        }
*/

//TODO
/*
        if (issue.targetVersionId.backend != Backend.ID_NONE)
        {
          array = new JIRAArray(fieldsEntry,"versions");
          array.add("id",issue.targetVersionId.backend);
        }
*/

          if (issue.estimatedDuration > Issue.ESTIMATED_NONE)
          {
            // Note: originalEstimateSeconds does not work - bug in JIRA?
            entry = new JIRAObject(fieldsEntry,"timetracking");
            entry.put("originalEstimate",issue.estimatedDuration);
            entry.put("remainingEstimate",0);
          }

          if (assignedToMember != null)
          {
            entry = new JIRAObject(fieldsEntry,"assignee");
            entry.put("name",(assignedToMember != Member.ANY) ? assignedToMember.name : null);
          }

  /*
          Element entry;

          if (issue.parentId != Backend.ID_NONE)
          {
            entry = document.createElement("parent_issue_id");
            entry.appendChild(document.createTextNode(Integer.toString(issue.parentId)));
            rootElement.appendChild(entry);
          }

          if (issue.statusId != Backend.ID_NONE)
          {
            entry = document.createElement("status_id");
            entry.appendChild(document.createTextNode(Integer.toString(issue.statusId)));
            rootElement.appendChild(entry);
          }

          if (issue.targetVersionId.backend != Backend.ID_NONE)
          {
            entry = document.createElement("fixed_version_id");
            entry.appendChild(document.createTextNode(Integer.toString(issue.targetVersionId)));
            rootElement.appendChild(entry);
          }

          if (issue.assignedToId.backend != Backend.ID_NONE)
          {
  assert issue.assignedToId != 0;
            entry = document.createElement("assigned_to_id");
            entry.appendChild(document.createTextNode(Integer.toString(issue.assignedToId)));
            rootElement.appendChild(entry);
          }

          entry = document.createElement("estimated_hours");
          entry.appendChild(document.createTextNode(Double.toString((double)issue.estimatedDuration/60.0)));
          rootElement.appendChild(entry);
  */
        }
      });

      // store state change
      if (!issue.statusTransitionId.isNone())
      {
        // set status transition
        postData(REST_API+"/issue/"+issue.id.backend+"/transitions",HttpURLConnection.HTTP_NO_CONTENT,issue,new CreateHandler<Issue>()
        {
          @Override
          public void data(JIRAObject rootEntry, Issue issue)
          {
            JIRAArray  array;
            JIRAObject entry;

            entry = new JIRAObject(rootEntry,"transition");
            entry.put("id",issue.statusTransitionId.backend);
          }
        });

        // get new status
        getData(REST_API+"/issue/"+issue.id.backend,new String[]{"fields","status"},new EntryHandler<Issue>()
        {
          @Override
          public Issue parse(JIRAObject entry)
          {
            issue.statusId = getIdValue(entry,"id");
            return null;
          }
        },
        "fields=status");
      }
    }
    catch (InterruptedException exception)
    {
      throw new BackendException(exception);
    }
  }

  /** update issue on JIRA server
   * @param issue issue to update
   * @param originIssue origin issue or null
   */
  public void updateIssue(final Issue issue, final Issue originIssue)
    throws BackendException
  {
Dprintf.printStackTrace();

/*
    category_id
    parent_issue_id - ID of the parent issue
    custom_fields - See Custom fields
    estimated_hours - Number of hours estimated for issue
*/
    assert(issue.backendType == Backend.Types.JIRA);
    assert(issue.id.hasBackend());

    try
    {
      final Tracker     tracker          = issue.getTracker();
      final Priority    priority         = issue.getPriority();
      final Member      assignedToMember = issue.getAssignedToMember();
      final VersionList versionList      = issue.getVersionList();
      final Version     targetVersion    = issue.getVersion();

      /* Note: bug the JIRA documentation! HTTP_NO_CONTENT (204) is return on success, not 200.
         https://developer.atlassian.com/static/rest/jira/6.1.html#d2e1209
      */
      putData(REST_API+"/issue/"+issue.id.backend,HttpURLConnection.HTTP_NO_CONTENT,issue,new CreateHandler<Issue>()
      {
        @Override
        public void data(JIRAObject rootEntry, Issue issue)
        {
          JIRAArray  array;
          JIRAObject entry;

          JIRAObject updateEntry = new JIRAObject(rootEntry,"update");
  /*
  Note: not possible in JIRA. issue have to be recreated
          if (tracker != null)
          {
            array = new JIRAArray(updateEntry,"issuetype");
            entry = new JIRAObject(array,"set");
            entry.put("name",tracker.name);
          }
  */
          if (priority != null)
          {
            array = new JIRAArray(updateEntry,"priority");
            entry = new JIRAObject(array,"set");
            entry.put("name",priority.name);
          }

  // TODO: do not overwrite possible changed issue description in ticket
  Dprintf.dprintf("TODO");
          if (issue.summary != null)
          {
            array = new JIRAArray(updateEntry,"summary");
            array.add("set",issue.summary);
          }

          if (issue.description != null)
          {
            array = new JIRAArray(updateEntry,"description");
            array.add("set",issue.description);
          }

          // Note: doneRatio (progress percentage) cannot be set in JIRA

  /*
          if (!versionList.isEmpty() && (targetVersion != Version.NONE))
          {
            array = new JIRAArray(updateEntry,"fixVersions");
            array = new JIRAArray(array,"set");
            if (targetVersion != Version.ANY)
            {
              entry = new JIRAObject(array);
              entry.put("name",targetVersion.name);
            }
          }
  */

          if (issue.estimatedDuration > Issue.ESTIMATED_NONE)
          {
            // Note: originalEstimateSeconds does not work - bug in JIRA?
            // Note: JIRA time format is odd, e. g. 60m -> 60min, 4320min (=3days) -> 1week+4days ?
            array = new JIRAArray(updateEntry,"timetracking");
            entry = new JIRAObject(array,"edit");
            entry.put("originalEstimate",String.format("%dw %dd %dm",issue.estimatedDuration/(7*24*60),(issue.estimatedDuration%(7*24*60))/(24*60),issue.estimatedDuration%60));
          }

          // Note: set assignTo with issue/<id>/assignee (see below)

  /*
          Element entry;

          if (issue.parentId != Backend.ID_NONE)
          {
            entry = document.createElement("parent_issue_id");
            entry.appendChild(document.createTextNode(Integer.toString(issue.parentId)));
            rootElement.appendChild(entry);
          }

          entry = document.createElement("tracker_id");
          entry.appendChild(document.createTextNode(Integer.toString(issue.trackerId)));
          rootElement.appendChild(entry);

          entry = document.createElement("author_id");
          entry.appendChild(document.createTextNode(Integer.toString(issue.authorId)));
          rootElement.appendChild(entry);

          entry = document.createElement("estimated_hours");
          if (issue.estimatedDuration > Issue.ESTIMATED_NONE)
          {
            entry.appendChild(document.createTextNode(Double.toString((double)issue.estimatedDuration/60.0)));
          }
          else
          {
            entry.appendChild(document.createTextNode(""));
          }
          rootElement.appendChild(entry);
  */
        }
      });

      // store state change
      if (!issue.statusTransitionId.isNone())
      {
        // set status transition
        postData(REST_API+"/issue/"+issue.id.backend+"/transitions",HttpURLConnection.HTTP_NO_CONTENT,issue,new CreateHandler<Issue>()
        {
          @Override
          public void data(JIRAObject rootEntry, Issue issue)
          {
            JIRAArray  array;
            JIRAObject entry;

            entry = new JIRAObject(rootEntry,"transition");
            entry.put("id",issue.statusTransitionId.backend);
          }
        });
        issue.statusTransitionId.backend = Id.NONE;

        // get new status
        getData(REST_API+"/issue/"+issue.id.backend,new String[]{"fields","status"},new EntryHandler<Issue>()
        {
          @Override
          public Issue parse(JIRAObject entry)
          {
            issue.statusId = getIdValue(entry,"id");

            return null;
          }
        },
        "fields=status");
      }

      // store assigned-to
  Dprintf.dprintf("TODO: do not overwrite possible changed assignee");
  /*
      if (assignedToMember != Member.NONE)
      {
        putData(REST_API+"/issue/"+issue.id.backend+"/assignee",HttpURLConnection.HTTP_NO_CONTENT,issue,new CreateHandler<Issue>()
        {
          @Override
          public void data(JIRAObject rootEntry, Issue issue)
          {
            rootEntry.put("name",(assignedToMember != Member.ANY) ? assignedToMember.name : null);
          }
        });
      }
  */
    }
    catch (InterruptedException exception)
    {
      new BackendException(exception);
    }
  }

  /** delete issue on JIRA server
   * @param issue issue to delete
   */
  public void deleteIssue(final Issue issue)
    throws BackendException
  {
    assert(issue.backendType == Backend.Types.JIRA);
    assert(issue.id.hasBackend());

throw new Error("still not implemented");
  }


  //--------------------------------------------------------------------

  /** load time entry from JIRA server
   * @param projectId project JIRA id
   * @param issueId issue JIRA id
   * @param timeEntryId time entry JIRA id
   */
  public TimeEntry loadTimeEntry(final Id projectId, final Id issueId, Id timeEntryId)
    throws BackendException
  {
Dprintf.dprintf("projectid=%s issueid=%s timeentryid=%s\n",projectId,issueId,timeEntryId);
Dprintf.printStackTrace();
    assert(timeEntryId.hasBackend());

    JIRAObject rootEntry = new JIRAObject();
    JIRAArray worklogIdArray = new JIRAArray(rootEntry,"ids");
    worklogIdArray.add(timeEntryId.backend);

    TimeEntry timeEntry = null;
    try
    {
      // Note: https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issue-worklogs/#api-rest-api-3-worklog-list-post
      timeEntry = getData(REST_API+"/worklog/list",0,1,rootEntry,new EntryHandler<TimeEntry>()
      {
        @Override
        public TimeEntry process(JIRAObject entry, int index, int totalCount)
          throws BackendException
        {
Dprintf.dprintf("entry=%s", entry);
          JIRAObject author = (JIRAObject)entry.get("author");
          assert author != null : "entry 'author' not found";

          // get project id
          Id projectId;
          try
          {
            projectId = getData(REST_API+"/issue/"+issueId.backend,"fields",new EntryHandler<Id>()
            {
              @Override
              public Id parse(JIRAObject entry)
              {
                return getIdValue(entry,new String[]{"project","id"});
              }
            });
          }
          catch (InterruptedException exception)
          {
            throw new BackendException(exception);
          }


          if (currentUser.name.equals(getValue(author,"name")))
          {
  //Dprintf.dprintf("issueId=%s dt=%s creatd=%s",getIdValue(entry,"issueId"),getLongValue(entry,"timeSpentSeconds")/60,getDateValue(entry,"created"));
            return new TimeEntry(Backend.Types.JIRA,
                                  getIdValue(entry,"id"),
                                  projectId,
                                  getIdValue(entry,"issueId"),
                                  currentUser.id,
null,//                                                  Backend.ID_NONE,  // activityId
                                  getLongValue(entry,"timeSpentSeconds")/60,
                                  getValue(entry,"comment",""),
                                  getSpentOn(entry,"started"),
                                  getDateValue(entry,"created"),
                                  getDateValue(entry,"updated")
                                 );
          }
          else
          {
            return null;
          }
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      },
  //TODO: no project data?
      "fields=project"
      );
    }
    catch (InterruptedException exception)
    {
      throw new BackendException(exception);
    }

/*
    final TimeEntry timeEntry = new TimeEntry(Backend.Types.JIRA,
                                              timeEntryId,
                                              projectId,
                                              issueId
                                             );

    try
    {
      getData("/time_entries/"+timeEntryId.backend,"time_entry",new EntryHandler<TimeEntry>()
      {
        @Override
        public TimeEntry parse(JIRAObject entry)
        {
// TODO: projectId, issueId, userId, activityId
          timeEntry.id         = getIdValue(entry,"id");
          timeEntry.projectId  = new Id();//getIntAttribute(entry,"project","id");
          timeEntry.issueId    = new Id();//getIntAttribute(entry,"issue","id");
          timeEntry.userId     = new Id();//getIntAttribute(entry,"user","id");
          timeEntry.activityId = new Id();//getIntAttribute(entry,"activity","id");
          timeEntry.duration   = (getLongValue(entry,"timespent")+60-1)/60;
          timeEntry.comments   = getValue(entry,"comments");
          timeEntry.spentOn    = getSpentOn(entry,"spent_on");
          timeEntry.createdOn  = getDateValue(entry,"created");
          timeEntry.updatedOn  = getDateValue(entry,"updated");

          return null;
        }
      });
    }
    catch (InterruptedException exception)
    {
      throw new BackendException(exception);
    }
*/

    return timeEntry;
  }

  /** load time entry from JIRA server
   * @param projectId project JIRA id
   * @param issueId issue JIRA id
   * @param timeEntryId time entry JIRA id
   * @param clazz TimeEntry class
   */
  public TimeEntry loadTimeEntry(Id projectId, Id issueId, Id timeEntryId, Class<TimeEntry> clazz)
    throws BackendException
  {
    assert(timeEntryId.hasBackend());

    try
    {
      return getData(REST_API+"/issue/"+issueId.backend,new String[]{"fields","worklog","worklogs"},new EntryHandler<TimeEntry>()
      {
        @Override
        public TimeEntry parse(JIRAObject entry)
        {
          return new TimeEntry(Backend.Types.JIRA,
                               getIdValue(entry,"id"),
                               getIdValue(entry.parent.parent,new String[]{"project","id"}),
                               getIdValue(entry,"issueId"),
//TODO
                               new Id(),//getIntAttribute(entry,"user","id"),
//TODO
                               new Id(),//getIntAttribute(entry,"activity","id"),
                               getLongValue(entry,"timeSpentSeconds")/60,
                               getValue(entry,"comment",""),
                               getSpentOn(entry,"started"),
                               getDateValue(entry,"created"),
                               getDateValue(entry,"updated")
                              );
        }
      },
      "fields=project,worklog"
      );
    }
    catch (InterruptedException exception)
    {
      throw new BackendException(exception);
    }
  }

  /** store time entry on JIRA server
   * @param timeEntry time entry to store
   */
  public void storeTimeEntry(TimeEntry timeEntry)
    throws BackendException
  {
    assert(timeEntry.backendType == Backend.Types.JIRA);

    // add new time entry
    postData(REST_API+"/issue/"+timeEntry.issueId.backend+"/worklog",HttpURLConnection.HTTP_CREATED,timeEntry,new CreateHandler<TimeEntry>()
    {
      @Override
      public void data(JIRAObject rootEntry, TimeEntry timeEntry)
      {
        rootEntry.put("comment",timeEntry.comments);
        rootEntry.put("started",timeEntry.spentOn.toString(DATETIME_FORMAT));
        rootEntry.put("timeSpentSeconds",timeEntry.duration*60);
/*
        Element element;

        element = document.createElement("issue_id");
        element.appendChild(document.createTextNode(Integer.toString(timeEntry.issueId)));
        rootElement.appendChild(element);

        element = document.createElement("spent_on");
        element.appendChild(document.createTextNode(timeEntry.spentOn.toString(DATE_FORMAT)));
        rootElement.appendChild(element);

        element = document.createElement("hours");
        element.appendChild(document.createTextNode(Double.toString(timeEntry.hours)));
        rootElement.appendChild(element);

        element = document.createElement("activity_id");
        element.appendChild(document.createTextNode(Integer.toString(timeEntry.activityId)));
        rootElement.appendChild(element);

        element = document.createElement("comments");
        element.appendChild(document.createTextNode(timeEntry.comments));
        rootElement.appendChild(element);
*/
      }
    });
  }

  /** update time entry on JIRA server
   * @param timeEntry time entry to update
   * @param originTimeEntry origin time entry
   */
  public void updateTimeEntry(TimeEntry timeEntry, final TimeEntry originTimeEntry)
    throws BackendException
  {
    assert(timeEntry.backendType == Backend.Types.JIRA);
    assert(timeEntry.id.hasBackend());

// TODO
Dprintf.dprintf("timeEntry=%s",timeEntry);
Dprintf.dprintf("originTimeEntry=%s",originTimeEntry);
    if      (!timeEntry.issueId.equals(originTimeEntry.issueId))
    {
      // issue changed => delete+add time entry on JIRA server
Dprintf.dprintf("delete origin %s",originTimeEntry);
      deleteTimeEntry(originTimeEntry);
Dprintf.dprintf("store %s",timeEntry);
      storeTimeEntry(timeEntry);
    }
    else if (timeEntry.id.isNone())
    {
      // add time entry on JIRA server
      storeTimeEntry(timeEntry);
    }
    else
    {
      // update time entry on JIRA server
      putData(REST_API+"/issue/"+timeEntry.issueId.backend+"/worklog/"+timeEntry.id.backend,HttpURLConnection.HTTP_OK,timeEntry,new CreateHandler<TimeEntry>()
      {
        @Override
        public void data(JIRAObject rootEntry, TimeEntry timeEntry)
        {
          rootEntry.put("comment",timeEntry.comments);
          rootEntry.put("timeSpentSeconds",timeEntry.duration*60);
/*
          Element element;

          element = document.createElement("issue_id");
          element.appendChild(document.createTextNode(Integer.toString(timeEntry.issueId)));
          rootElement.appendChild(element);

          element = document.createElement("spent_on");
          element.appendChild(document.createTextNode(timeEntry.spentOn.toString(DATE_FORMAT)));
          rootElement.appendChild(element);

          element = document.createElement("hours");
          element.appendChild(document.createTextNode(Double.toString(timeEntry.hours)));
          rootElement.appendChild(element);

          element = document.createElement("activity_id");
          element.appendChild(document.createTextNode(Integer.toString(timeEntry.activityId)));
          rootElement.appendChild(element);

          element = document.createElement("comments");
          element.appendChild(document.createTextNode(timeEntry.comments));
          rootElement.appendChild(element);
*/
        }
      });
    }
  }

  /** delete time entry on JIRA server
   * @param timeEntry time entry to delete
   */
  public void deleteTimeEntry(final TimeEntry timeEntry)
    throws BackendException
  {
    assert(timeEntry.backendType == Backend.Types.JIRA);
    assert(timeEntry.id.hasBackend());

    if (timeEntry.id.backend != Id.NONE)
    {
      // delete time entry from JIRA server
      deleteData(REST_API+"/issue/"+timeEntry.issueId.backend+"/worklog/"+timeEntry.id.backend);
    }
  }

  //--------------------------------------------------------------------

  /** get users from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public void getUsers(final BackendProgress backendProgress, final EntityHandler<User> entityHandler)
    throws BackendException, InterruptedException
  {
    getData(REST_API+"/user/search",new EntryHandler<User>()
    {
      @Override
      public User process(JIRAObject entry)
        throws BackendException
      {
        // get user/member key
        String key = getValue(entry,"key");

        // get first/last name
        String[] names     = getValue(entry,"displayName").split(" ",2);
        String   firstName = (names.length >= 1) ? names[0] : "";
        String   lastName  = (names.length >= 2) ? names[1] : "";

        // map key -> user id
        Id userId = userKeyToId(key);

        // create user
        User user = new User(Backend.Types.JIRA,
                             userId,
                             firstName,
                             lastName,
                             getValue(entry,"name"),
                             getValue(entry,"emailAddress")
                            );
        entityHandler.handle(user);

        // add fake backend id
        user.id.backend = user.id.local;

        entityHandler.done(user);

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    },
    "username=.");
  }

  /** get current user from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   * @return current user or User.NONE
   */
  public User getCurrentUser(final BackendProgress backendProgress, final EntityHandler<User> entityHandler)
    throws BackendException
  {
    if (currentUser == null)
    {
      // try to get current user
      try
      {
        currentUser = getData(REST_GADGET+"/currentUser",new EntryHandler<User>()
        {
          @Override
          public User process(JIRAObject entry, int index, int totalCount)
          {
            // get user key
            String key = getValue(entry,"username");

            // map key -> user id
            Id userId = userKeyToId(key);

            // get first/last name
            String[] names     = getValue(entry,"fullName").split(" ",2);
            String   firstName = (names.length >= 1) ? names[0] : "";
            String   lastName  = (names.length >= 2) ? names[1] : "";

            return new User(Backend.Types.JIRA,
                            userId,
                            firstName,
                            lastName,
                            key,  // name=key
                            getValue(entry,"email")
                           );
          }
          @Override
          protected void error(Exception exception)
          {
            // ignored
          }
        });
      } catch (InterruptedException exception)
      {
        // ignored
      }
    }

    entityHandler.handle(currentUser);
    entityHandler.done(currentUser);

    return currentUser;
  }

  /** get roles from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getRoles(final BackendProgress backendProgress, final EntityHandler<Role> entityHandler)
    throws BackendException, InterruptedException
  {
    // get project ids
    final ArrayList<Integer> projectIdList = getProjectIdList(backendProgress);

    // get role ids for all projects
    for (int i = 0; i < projectIdList.size(); i++)
    {
      getData(REST_API+"/project/"+projectIdList.get(i),"roles",new EntryHandler<Role>()
      {
        @Override
        public Role process(JIRAObject entry, int index, int totalCount)
          throws BackendException
        {
          final Pattern PATTERN = Pattern.compile(".*/(\\d+)$",Pattern.CASE_INSENSITIVE);

          Matcher matcher;
          for (Object mapEntry : entry.entrySet())
          {
            String name = ((Map.Entry<String,String>)mapEntry).getKey();
            String url  = ((Map.Entry<String,String>)mapEntry).getValue();

            if ((matcher = PATTERN.matcher(url)).matches())
            {
//Dprintf.dprintf("name %s => %s => %s",name,url,matcher.group(1));
              Role role = new Role(Backend.Types.JIRA,
                                   new Id(Integer.parseInt(matcher.group(1))),
//TODO
                                   new Id(Id.ANY),  // projectId
                                   name,
                                   ""
                                  );
              entityHandler.handle(role);
              entityHandler.done(role);
            }
          }

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
      backendProgress.setSelection(i,projectIdList.size());

      // throttle execution
      try { Thread.sleep(Settings.throttleTime); } catch (InterruptedException exception) { /* ignored */ }
    }
  }

  /** get trackers from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public void getTrackers(final BackendProgress backendProgress, final EntityHandler<Tracker> entityHandler)
    throws BackendException, InterruptedException
  {
    // get tracker data and store into local database
    getData(REST_API+"/search",new String[]{"issues"},new EntryHandler<Tracker>()
    {
      @Override
      public Tracker process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
        JIRAObject fields = (JIRAObject)entry.get("fields");
        assert fields != null : "entry 'fields' not found";
        JIRAObject project = (JIRAObject)fields.get("project");
        assert project != null : "entry 'project' not found";
        JIRAObject issueType = (JIRAObject)fields.get("issuetype");
        assert issueType != null : "entry 'issuetype' not found";

        Tracker tracker = new Tracker(Backend.Types.JIRA,
                                      getIdValue(issueType,"id"),
                                      getIdValue(project,"id"),
                                      getValue(issueType,"name",""),
                                      getValue(issueType,"description",""),
                                      getValue(issueType,"name").equalsIgnoreCase("Epic") // epic name required for "Epic" tracker
                                     );
        entityHandler.handle(tracker);
        entityHandler.done(tracker);

        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    },
    "fields=project,issuetype");
  }

  /** get epic names from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public void getEpicNames(final BackendProgress backendProgress, final EntityHandler<EpicName> entityHandler)
    throws BackendException, InterruptedException
  {
    final String epicNameCustomField = getEpicNameCustomField();

    // get epic names and store into local database
    getData(REST_API+"/search",new String[]{"issues","fields"},new EntryHandler<EpicName>()
    {
      @Override
      public EpicName process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
        String name = getValue(entry,epicNameCustomField);
        if (name != null)
        {
          // map name -> epic name id
          Id epicNameId = epicNameToId(name);

          // add epic name
          EpicName epicName = new EpicName(Backend.Types.JIRA,
                                           epicNameId,
                                           name
                                          );

          entityHandler.handle(epicName);
          entityHandler.done(epicName);
        }

        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    });
  }

  /** get status from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getStatus(final BackendProgress backendProgress, final EntityHandler<Status> entityHandler)
    throws BackendException, InterruptedException
  {
    getData(REST_API+"/status",new EntryHandler<Status>()
    {
      @Override
      public Status process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
//Dprintf.dprintf("status entry=%s",entry);
        Status status = new Status(Backend.Types.JIRA,
                                   getIdValue(entry,"id"),
//TODO
                                   new Id(Id.ANY),  // projectId
                                   getValue(entry,"name",""),
                                   getValue(entry,"description",""),
                                   false  // default
                                  );
        entityHandler.handle(status);
        entityHandler.done(status);

        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    });
  }

  /** get priorities from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getPriorities(final BackendProgress backendProgress, final EntityHandler<Priority> entityHandler)
    throws BackendException, InterruptedException
  {
    getData(REST_API+"/priority",new EntryHandler<Priority>()
    {
      @Override
      public Priority process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
        Priority priority = new Priority(Backend.Types.JIRA,
                                         getIdValue(entry,"id"),
//TODO
                                         new Id(Id.ANY),  // projectId
                                         getValue(entry,"name",""),
                                         getValue(entry,"description",""),
                                         false // isDefault
                                        );
        entityHandler.handle(priority);
        entityHandler.done(priority);

        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    });
  }

  /** get activities from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getActivities(final BackendProgress backendProgress, final EntityHandler<Activity> entityHandler)
    throws BackendException, InterruptedException
  {
/*
    getData(REST_API+"/search",new String[]{"issues","fields","issuetype"},new EntryHandler<Activity>()
    {
      @Override
      public Activity process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
        Activity activity = new Activity(Backend.Types.JIRA,
                                         getIdValue(entry,"id"),
                                         getValue(entry,"name",""),
                                         false
                                        );
        entityHandler.handle(activity);
        entityHandler.done(activity);

        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    },
    "fields=issuetype");
*/
    getData(REST_API+"/issuetype",new EntryHandler<Activity>()
    {
      @Override
      public Activity process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
        Activity activity = new Activity(Backend.Types.JIRA,
                                         getIdValue(entry,"id"),
//TODO
                                         new Id(Id.ANY),  // projectId
                                         getValue(entry,"name",""),
                                         getValue(entry,"description",""),
                                         false
                                        );
        entityHandler.handle(activity);
        entityHandler.done(activity);

        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    });
  }

  /** get projects from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  public void getProjects(final BackendProgress backendProgress, final EntityHandler<Project> entityHandler)
    throws BackendException, InterruptedException
  {
    getData(REST_API+"/project",new EntryHandler<Project>()
    {
      @Override
      public Project process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
        Project project = new Project(Backend.Types.JIRA,
                                      getIdValue(entry,"id"),
//TODO
null,//                               getIntAttribute(entry,"parent","id"),
                                      getValue(entry,"key"),
                                      getValue(entry,"name",""),
ProjectStates.ACTIVE,//                               ProjectStates.toEnum(getIntValue(entry,"status")),
                                      getValue(entry,"key"),
"",//                                     getValue(entry,"description",""),
new Date(),//                               getDateValue(entry,"created"),
new Date()//                               getDateValue(entry,"updated")
                                     );
        entityHandler.handle(project);
        entityHandler.done(project);

        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    });
  }

  /** get members from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getMembers(final BackendProgress backendProgress, final EntityHandler<Member> entityHandler)
    throws BackendException, InterruptedException
  {
    {
      backendProgress.beginSteps(2);
    }
    try
    {
      // get project ids
      backendProgress.setStep(0);
      final ArrayList<Integer> projectIdList = getProjectIdList(backendProgress);

      // get members of all projects
      backendProgress.setStep(1);
      for (int i = 0; i < projectIdList.size(); i++)
      {
        final Id projectId = new Id(projectIdList.get(i));
        try
        {
          // get project key
          String projectKey = getData(REST_API+"/project/"+projectId.backend,new EntryHandler<String>()
          {
            @Override
            public String process(JIRAObject entry)
              throws BackendException
            {
              return getValue(entry,"key");
            }
          });

          getData(REST_API+"/user/assignable/search",new EntryHandler<User>()
          {
            @Override
            public User process(JIRAObject entry)
              throws BackendException, InterruptedException
            {
              // get user/member key
              String key = getValue(entry,"key");

              // map key -> user/member id
              Id userId   = userKeyToId(key);
              Id memberId = memberKeyToId(key);

              // get first/last name
              String[] names     = getValue(entry,"displayName").split(" ",2);
              String   firstName = (names.length >= 1) ? names[0] : "";
              String   lastName  = (names.length >= 2) ? names[1] : "";

              // add member
              if (!firstName.isEmpty() && !lastName.isEmpty())
              {
                Member member = new Member(Backend.Types.JIRA,
                                           memberId,
                                           projectId,
                                           userId,
                                           firstName,
                                           lastName,
                                           key,  // name=key
                                           new Id()
                                          );
                entityHandler.handle(member);
                entityHandler.done(member);
              }

              return null;
            }
            @Override
            protected void error(Exception exception)
            {
              // ignored
            }
          },
          "project="+projectKey);
        }
        catch (BackendException exception)
        {
          // ignored
        }
        backendProgress.setSelection(i,projectIdList.size());

        // throttle execution
        try { Thread.sleep(Settings.throttleTime); } catch (InterruptedException exception) { /* ignored */ }
      }
    }
    finally
    {
      backendProgress.endSteps();
    }
  }

  /** get current member from backend
   * @param backendProgress progress handler or null
   * @param projectId project id
   * @param entityHandler entity handler
   * @return current member or Member.NONE
   */
  public Member getCurrentMember(final BackendProgress backendProgress, final Id projectId, final EntityHandler<Member> entityHandler)
    throws BackendException, InterruptedException
  {
Dprintf.dprintf("%s %s",projectId,projectId.isNone());
    assert !projectId.isNone();
Dprintf.dprintf("xxxxxxxxxxxx");

    Member currentMember = currentMemberMap.get(projectId);
    if (currentMember == null)
    {
      // get current user
      if (currentUser == null) currentUser = getCurrentUser();

      // get project key
      String projectKey = getData(REST_API+"/project/"+projectId.backend,new EntryHandler<String>()
      {
        @Override
        public String process(JIRAObject entry)
          throws BackendException
        {
          return getValue(entry,"key");
        }
      });
Dprintf.dprintf("%s %s",projectId,projectKey); System.exit(1);

      currentMember = getData(REST_API+"/user/search",new EntryHandler<Member>()
      {
        @Override
        public Member process(JIRAObject entry, int index, int totalCount)
        {
          // get key
          String key = getValue(entry,"key");

          // map key -> user/member id
          Id userId   = userKeyToId(key);
          Id memberId = memberKeyToId(key);

          // get first/last name
          String[] names     = getValue(entry,"displayName").split(" ",2);
          String   firstName = (names.length >= 1) ? names[0] : "";
          String   lastName  = (names.length >= 2) ? names[1] : "";

          return new Member(Backend.Types.JIRA,
                            memberId,
                            projectId,
                            userId,
                            firstName,
                            lastName,
                            key,  // name=key
                            new Id()
                           );
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      },
      "project="+projectKey,
      "username="+currentUser.name
      );

      currentMemberMap.put(projectId,currentMember);
    }

    entityHandler.handle(currentMember);
    entityHandler.done(currentMember);

    return (currentMember != null) ? currentMember : null;
  }

  /** get versions from backend
   * @param backendProgress progress handler or null
   * @param entityHandler entity handler
   */
  @Override
  public void getVersions(final BackendProgress backendProgress, final EntityHandler<Version> entityHandler)
    throws BackendException, InterruptedException
  {
    // get project ids
    final ArrayList<Integer> projectIdList = getProjectIdList(backendProgress);

    // get version ids for all projects
    for (int i = 0; i < projectIdList.size(); i++)
    {
      getData(REST_API+"/project/"+projectIdList.get(i)+"/versions",new EntryHandler<Version>()
      {
        @Override
        public Version process(JIRAObject entry, int index, int totalCount)
          throws BackendException
        {
          Version version = new Version(Backend.Types.JIRA,
                                        getIdValue(entry,"id"),
                                        getIdValue(entry,"projectId"),
                                        getValue(entry,"name"),
                                        "",  // description
//TODO
"",//                                        getValue(entry,"status"),
                                        getDateValue(entry,"releaseDate"),
                                        new Date(),  // created
                                        new Date()  // updated
                                       );
          entityHandler.handle(version);
          entityHandler.done(version);

          return null;
        }
/*
        @Override
        public Version parse(JIRAObject entry)
        {
          return new Version(getIdValue(entry,"id"),
                             getIdValue(entry,"projectId"),
                             getValue(entry,"name"),
                             "",  // description
//TODO
"",//                               getValue(entry,"status"),
                             getDateValue(entry,"releaseDate"),
                             new Date(),  // created
                             new Date()  // updated
                            );
        }
        @Override
        public void store(Version version, int index, int totalCount)
          throws BackendException
        {
          entityHandler.handle(version);
          entityHandler.done(version);
        }
*/
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
      backendProgress.setSelection(i,projectIdList.size());

      // throttle execution
      try { Thread.sleep(Settings.throttleTime); } catch (InterruptedException exception) { /* ignored */ }
    }
  }

  /** get issues from backend
   * @param backendProgress progress handler or null
   * @param lastUpdatedOnTimestamp last updated time stamp for incremental update or 0
   * @param entityHandler entity handler
   */
  @Override
  public void getIssues(final BackendProgress backendProgress, long lastUpdatedOnTimestamp, final EntityHandler<Issue> entityHandler)
    throws BackendException, InterruptedException
  {
    // get epic name custom field
    final String epicNameCustomField = getEpicNameCustomField();

    getData(REST_API+"/search",new String[]{"issues"},new EntryHandler<Issue>()
    {
      @Override
      public Issue process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
//Dprintf.dprintf("issue entry=%s",entry);
        JIRAObject fields = (JIRAObject)entry.get("fields");
        assert fields != null : "entry 'fields' not found";

        // get parent issue id
        Id parentId;
        JIRAArray issueLinks = (JIRAArray)fields.get("issuelinks");
        if ((issueLinks != null) && (issueLinks.size() > 0))
        {
//Dprintf.dprintf("TODO");
          parentId = new Id();
        }
        else
        {
          parentId = new Id();
        }

        // get epic name
        String epicName = getValue(fields,epicNameCustomField,"");

        // get first fix version id
        Id fixVersionId;
        JIRAArray fixVersions = (JIRAArray)fields.get("fixVersions");
        if ((fixVersions != null) && (fixVersions.size() > 0))
        {
          fixVersionId = getIdValue((JIRAObject)fixVersions.get(0),"id");
        }
        else
        {
          fixVersionId = new Id();
        }

        Issue issue = new Issue(Backend.Types.JIRA,
                                getIdValue(entry,"id"),
                                parentId,
                                getValue(entry,"key"),
                                getIdValue(fields,new String[]{"project","id"}),
                                getIdValue(fields,new String[]{"issuetype","id"}),
                                epicName,
                                getIdValue(fields,new String[]{"status","id"}),
                                getIdValue(fields,new String[]{"priority","id"}),
                                userKeyToId(getValue(fields,new String[]{"reporter","key"})),
                                getValue(fields,"summary",""),
                                getValue(fields,"description",""),
//TODO
new Date(),//                             getDateValue(entry,"start_date"),
                                getDateValue(fields,"duedate"),
                                memberKeyToId(getValue(fields,new String[]{"assignee","key"})),
                                getIntValue(fields,new String[]{"aggregateprogress","percent"},0),
                                fixVersionId,
//                                (getLongValue(fields,"timeoriginalestimate",Issue.ESTIMATED_NONE)+60-1)/60,
                                (getLongValue(entry,new String[]{"aggregateprogress","total"},0)+60-1)/60,
                                getDateValue(fields,"created"),
                                getDateValue(fields,"updated"),
//TODO
new Date()//                             getDateValue(entry,"closed_on")
                               );
        entityHandler.handle(issue);
        entityHandler.done(issue);

        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    },
    (lastUpdatedOnTimestamp > 0) ? "jql=updated"+FILTER_GE+DATE_FORMAT.format(lastUpdatedOnTimestamp*1000) : null
    );
  }

  /** get status transitions from backend
   * @param backendProgress progress handler or null
   * @param issueId issue id
   * @param entityHandler entity handler
   */
  @Override
  public void getStatusTransitions(final BackendProgress backendProgress, final Id issueId, final EntityHandler<StatusTransition> entityHandler)
    throws BackendException, InterruptedException
  {
    getData(REST_API+"/issue/"+issueId.backend+"/transitions?transitionId","transitions",new EntryHandler<StatusTransition>()
    {
      @Override
      public StatusTransition process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
        JIRAObject toEntry = (JIRAObject)entry.get("to");

        StatusTransition statusTransition = new StatusTransition(Backend.Types.JIRA,
                                                                 getIdValue(entry,"id"),
                                                                 issueId,
                                                                 getValue(entry,"name"),
                                                                 getValue(entry,new String[]{"to","description"},""),
                                                                 getIdValue(entry,new String[]{"to","id"})
                                                                );
        entityHandler.handle(statusTransition);
        entityHandler.done(statusTransition);

        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    });
  }

  /** get time entry from backend
   * @param backendProgress progress handler or null
   * @param timeEntryId time entry id
   * @return time entry
   */
/*
  public TimeEntry getTimeEntry(final BackendProgress backendProgress, Id timeEntryId)
    throws BackendException, InterruptedException
  {
Dprintf.dprintf("_");
    JIRAArray worklogIdArray = new JIRAArray(rootEntry,"ids");
    worklogIdArray.add(timeEntryId.backend);

    // Note: https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issue-worklogs/#api-rest-api-3-worklog-list-post
    TimeEntry timeEntry = getData(REST_API+"/worklog/list",worklogIdArray,new EntryHandler<TimeEntry>()
    {
      @Override
      public TimeEntry process(JIRAObject entry, int index, int totalCount)
        throws BackendException
      {
        Id          issueId = getIdValue(entry,"issueId");
        final Issue issue   = Backend.getIssue(Backend.Types.JIRA,issueId);

        JIRAObject author = (JIRAObject)entry.get("author");
        assert author != null : "entry 'author' not found";

        if (currentUser.name.equals(getValue(author,"name")))
        {
//Dprintf.dprintf("issueId=%s dt=%s creatd=%s",getIdValue(entry,"issueId"),getLongValue(entry,"timeSpentSeconds")/60,getDateValue(entry,"created"));
          return new TimeEntry(Backend.Types.JIRA,
                                              getIdValue(entry,"id"),
                                              issue.projectId,
                                              issueId,
                                              currentUser.id,
null,//                                                  Backend.ID_NONE,  // activityId
                                              getLongValue(entry,"timeSpentSeconds")/60,
                                              getValue(entry,"comment",""),
                                              getSpentOn(entry,"started"),
                                              getDateValue(entry,"created"),
                                              getDateValue(entry,"updated")
                                             );
        }
        else
        {
        return null;
      }
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    },
//TODO: no project data?
    "fields=project"
    );

    return timeEntry;
  }
*/

  /** get time entries from backend
   * @param backendProgress progress handler or null
   * @param createOnTimestamp create-on timestamp or 0
   * @param entityHandler entity handler
   */
  public void getTimeEntries(final BackendProgress backendProgress, final long createdOnTimestamp, final EntityHandler<TimeEntry> entityHandler)
    throws BackendException, InterruptedException
  {
    if (currentUser == null)
    {
      // get current user
      currentUser = getData(REST_GADGET+"/currentUser",new EntryHandler<User>()
      {
        @Override
        public User process(JIRAObject entry, int index, int totalCount)
        {
          // get user key
          String key = getValue(entry,"username");

          // map key -> user id
          Id userId = userKeyToId(key);

          // get first/last name
          String[] names     = getValue(entry,"fullName").split(" ",2);
          String   firstName = (names.length >= 1) ? names[0] : "";
          String   lastName  = (names.length >= 2) ? names[1] : "";

          return new User(Backend.Types.JIRA,
                          userId,
                          firstName,
                          lastName,
                          key,  // name=key
                          getValue(entry,"email")
                         );
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }

/*
   if (createdOnTimestamp > 0)
   {
      // get updated worklog entries
      // Note: https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issue-worklogs/#api-rest-api-3-worklog-updated-get
Dprintf.dprintf("since=%d %s",createdOnTimestamp,"worklogDate"+FILTER_GE+DATE_FORMAT.format(new Date(createdOnTimestamp*1000)));
      final ArrayList<Integer> worklogIds = new ArrayList<Integer>();
      getData(REST_API+"/worklog/updated",new String[]{"values"},new EntryHandler<TimeEntry>()
      {
        @Override
        public TimeEntry process(JIRAObject entry, int index, int totalCount)
          throws BackendException
        {
          int worklogId = getIntValue(entry,"worklogId");
Dprintf.dprintf("worklogId %d/%d: %d",index,totalCount,worklogId);
Dprintf.printStackTrace();
Dprintf.dprintf("_"); System.exit(1);

          worklogIds.add(worklogId);

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      },
      "since="+Long.toString(createdOnTimestamp*1000)
      );

      // get issues and create time entries
      for (int i = 0; i < worklogIds.size(); i += ENTRY_LIMIT)
      {
        // get POST data: worklog ids array
        JIRAObject rootEntry = new JIRAObject();
        JIRAArray worklogIdArray = new JIRAArray(rootEntry,"ids");
        for (int j = 0; ((i+j) < worklogIds.size()) && (j < ENTRY_LIMIT); j++)
        {
          worklogIdArray.add(worklogIds.get(i+j));
        }

//Dprintf.dprintf("i=%d",i);
        // get issues
        // Note: https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issue-worklogs/#api-rest-api-3-worklog-list-post
        getData(REST_API+"/worklog/list",rootEntry,new EntryHandler<Issue>()
        {
          @Override
          public Issue process(JIRAObject entry, int index, int totalCount)
            throws BackendException
          {
            Id          issueId = getIdValue(entry,"issueId");
            final Issue issue   = Backend.getIssue(Backend.Types.JIRA,issueId);

            JIRAObject author = (JIRAObject)entry.get("author");
            assert author != null : "entry 'author' not found";

            if (currentUser.name.equals(getValue(author,"name")))
            {
//Dprintf.dprintf("issueId=%s dt=%s creatd=%s",getIdValue(entry,"issueId"),getLongValue(entry,"timeSpentSeconds")/60,getDateValue(entry,"created"));
              TimeEntry timeEntry = new TimeEntry(Backend.Types.JIRA,
                                                  getIdValue(entry,"id"),
                                                  issue.projectId,
                                                  issueId,
                                                  currentUser.id,
null,//                                                  Backend.ID_NONE,  // activityId
                                                  getLongValue(entry,"timeSpentSeconds")/60,
                                                  getValue(entry,"comment",""),
                                                  getSpentOn(entry,"started"),
                                                  getDateValue(entry,"created"),
                                                  getDateValue(entry,"updated")
                                                 );
              entityHandler.handle(timeEntry);
              entityHandler.done(timeEntry);
            }

            return null;
          }
          @Override
          protected void error(Exception exception)
          {
            // ignored
          }
        },
//TODO: no project data?
        "fields=project"
        );

        backendProgress.setSelection(i,worklogIds.size());
        if (backendProgress.isAborted()) done();
      }
      backendProgress.setSelection(worklogIds.size(),worklogIds.size());
    }
    else
    {
      // get all work log entries
      getData(REST_API+"/search",new String[]{"issues"},new EntryHandler<Issue>()
      {
        @Override
        public Issue process(JIRAObject entry, int index, int totalCount)
          throws BackendException
        {
          JIRAObject fields = (JIRAObject)entry.get("fields");
          assert fields != null : "entry 'fields' not found";
//Dprintf.dprintf("issue index=%d totalCount=%d",index,totalCount);

          Id       issueId   = getIdValue(entry,"id");
          final Id projectId = getIdValue(entry,new String[]{"fields","project","id"});
//Dprintf.dprintf("------------ issueId=%s projectId=%s",issueId,projectId);

          getData(REST_API+"/issue/"+issueId.backend+"/worklog",new String[]{"worklogs"},new EntryHandler<TimeEntry>()
          {
            @Override
            public TimeEntry process(JIRAObject entry, int index, int totalCount)
              throws BackendException
            {
              JIRAObject author = (JIRAObject)entry.get("author");
              assert author != null : "entry 'author' not found";

              if (currentUser.name.equals(getValue(author,"name")))
              {
  //Dprintf.dprintf("issueId=%s dt=%s creatd=%s",getIdValue(entry,"issueId"),getLongValue(entry,"timeSpentSeconds")/60,getDateValue(entry,"created"));
                TimeEntry timeEntry = new TimeEntry(Backend.Types.JIRA,
                                                    getIdValue(entry,"id"),
                                                    projectId,
                                                    getIdValue(entry,"issueId"),
                                                    currentUser.id,
null,//                                               Backend.ID_NONE,  // activityId
                                                    getLongValue(entry,"timeSpentSeconds")/60,
                                                    getValue(entry,"comment",""),
                                                    getSpentOn(entry,"started"),
                                                    getDateValue(entry,"created"),
                                                    getDateValue(entry,"updated")
                                                   );
                entityHandler.handle(timeEntry);
                entityHandler.done(timeEntry);
              }

              return null;
            }
            @Override
            protected void error(Exception exception)
            {
              // ignored
            }
          },
//TODO: date/time limit
          "fields=worklog"//         "spent_on="+FILTER_GE+ISO_DATETIME_FORMAT.format(createdOnTimestamp*1000)
          );

          backendProgress.setSelection(index,totalCount);
          if (backendProgress.isAborted()) done();

          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      },
      "fields=project"
      );
    }
*/

//    final ArrayList<Id> issueIds = new ArrayList<Id>();
    getData(REST_API+"/search",new String[]{"issues"},new EntryHandler<Issue>()
    {
      @Override
      public Issue process(JIRAObject entry, int index, int totalCount)
        throws BackendException, InterruptedException
      {
        Id       issueId   = getIdValue(entry,"id");
        final Id projectId = getIdValue(entry,new String[]{"fields","project","id"});
//Dprintf.dprintf("issueId=%s %s %s",issueId,Thread.currentThread().hashCode(),Thread.currentThread().isInterrupted());
//Dprintf.dprintf("projectId=%s",projectId);

        // get worklog for issue and create time entries
        getData(REST_API+"/issue/"+issueId.backend+"/worklog",new String[]{"worklogs"},new EntryHandler<TimeEntry>()
        {
          @Override
          public TimeEntry process(JIRAObject entry, int index, int totalCount)
            throws BackendException
          {
            JIRAObject author = (JIRAObject)entry.get("author");
            assert author != null : "entry 'author' not found";

            if (currentUser.name.equals(getValue(author,"name")))
            {
//Dprintf.dprintf("issueId=%s dt=%s creatd=%s",getIdValue(entry,"issueId"),getLongValue(entry,"timeSpentSeconds")/60,getDateValue(entry,"created"));
              TimeEntry timeEntry = new TimeEntry(Backend.Types.JIRA,
                                                  getIdValue(entry,"id"),
                                                  projectId,
                                                  getIdValue(entry,"issueId"),
                                                  currentUser.id,
null,//                                               Backend.ID_NONE,  // activityId
                                                  getLongValue(entry,"timeSpentSeconds")/60,
                                                  getValue(entry,"comment",""),
                                                  getSpentOn(entry,"started"),
                                                  getDateValue(entry,"created"),
                                                  getDateValue(entry,"updated")
                                                 );
              entityHandler.handle(timeEntry);
              entityHandler.done(timeEntry);
            }

            if (backendProgress.isAborted()) done();

            return null;
          }
          @Override
          protected void error(Exception exception)
          {
            // ignored
          }
        },
//TODO: date/time limit
        "fields=worklog"//         "spent_on="+FILTER_GE+ISO_DATETIME_FORMAT.format(createdOnTimestamp*1000)
        );

//issueIds.add(issueId);
        backendProgress.setSelection(index,totalCount);
        if (backendProgress.isAborted()) done();

        return null;
      }
      @Override
      protected void error(Exception exception)
      {
        // ignored
      }
    },
    "fields=id,project",
    (createdOnTimestamp > 0) ? "jql=worklogDate"+FILTER_GE+DATE_FORMAT.format(new Date(createdOnTimestamp*1000)) : ""
    );
//Dprintf.dprintf("issueIds=%d",issueIds.size());
  }

  //--------------------------------------------------------------------

  /** get data
   * @param entry entry
   * @param names names
   * @param nameIndex name index
   * @param entryHandler entry handler
   * @return number of processed entries
   */
  private <T> int getData(Object          entry,
                          String[]        names,
                          int             nameIndex,
                          EntryHandler<T> entryHandler
                         )
    throws BackendException, InterruptedException
  {
    int n = 0;
//Dprintf.dprintf("getData %d/%d: %s",nameIndex,(names != null) ? names.length : 0,((names != null) && (nameIndex < names.length)) ? names[nameIndex] : "");

    if (Thread.currentThread().isInterrupted())
    {
      throw new InterruptedException();
    }

    if ((names != null) && (nameIndex < names.length))
    {
      // visit sub-entries
//Dprintf.dprintf("sub-tree %s",names[nameIndex]);
//Dprintf.dprintf("entry=%s",entry);
      Object subEntry = ((JIRAObject)entry).get(names[nameIndex]);
      if      (subEntry instanceof JIRAObject)
      {
        JIRAObject jiraObject = (JIRAObject)subEntry;

        // single element
        n = 1;

        // set parent
        jiraObject.parent = (JIRAObject)entry;

        // get data
        getData(jiraObject,
                names,
                nameIndex+1,
                entryHandler
               );
      }
      else if (subEntry instanceof JIRAArray)
      {
        JIRAArray jiraArray = (JIRAArray)subEntry;

        // array element => return size of array
        n = jiraArray.size();

        // set parent
        jiraArray.parent = (JIRAObject)entry;

        // iterate array and get data
        Iterator<JIRAObject> iterator = jiraArray.iterator();
        while (iterator.hasNext())
        {
          JIRAObject jiraObject = iterator.next();

          jiraObject.parent = jiraArray.parent;

          getData(jiraObject,
                  names,
                  nameIndex+1,
                  entryHandler
                 );
          if (entryHandler.isDone()) break;
        }
      }
      else if (subEntry != null)
      {
        entryHandler.error(new BackendException("INTERNAL ERROR: unknown JSON object type '"+subEntry.getClass()+"'"));
      }
      else
      {
        entryHandler.error(new BackendException("INTERNAL ERROR: unknown JSON object with name '"+names[nameIndex]+"'"));
      }
    }
    else if ((names == null) || (nameIndex == names.length))
    {
      // process entry
      if      (entry instanceof JIRAObject)
      {
        JIRAObject jiraObject = (JIRAObject)entry;

        // single element
        n = 1;

        T entity = entryHandler.process(jiraObject);
        if (entryHandler.result == null) entryHandler.result = entity;
        entryHandler.incrementIndex();
      }
      else if (entry instanceof JIRAArray)
      {
        JIRAArray jiraArray = (JIRAArray)entry;

        // array element => return size of array
        n = jiraArray.size();

        // process data entries
        Iterator<JIRAObject> iterator = jiraArray.iterator();
        while (iterator.hasNext())
        {
          JIRAObject jiraObject = iterator.next();

          jiraObject.parent = jiraArray.parent;

          T entity = entryHandler.process(jiraObject);
          if (entryHandler.result == null) entryHandler.result = entity;
          entryHandler.incrementIndex();

          if (entryHandler.isDone()) break;
        }
      }
      else
      {
        entryHandler.error(new BackendException("unknown JSON response data type"));
      }
    }
    else
    {
      entryHandler.error(new BackendException("Invalid JSON tree shape"));
    }

    return n;
  }

  /** get data
   * @param urlString URL string
   * @param names JSON entry names to visit
   * @param offset start offset
   * @param length max. number of entries to get
   * @param postData post data for POST or null for GET
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity or null
   */
  private <T> T getData(String          urlString,
                        String[]        names,
                        int             offset,
                        int             length,
                        JIRAObject      postData,
                        EntryHandler<T> entryHandler,
                        String...       additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    boolean done = false;
    while (!done)
    {
      HttpURLConnection connection = null;
      try
      {
        // create JSON result parser
        JSONParser parser = new JSONParser();

        // get data from JIRA server
        String[] arguments = Arrays.copyOf(additionalArguments,additionalArguments.length+2);
        arguments[additionalArguments.length+0] = "startAt="+offset;
        arguments[additionalArguments.length+1] = "maxResults="+((length > 0) ? Math.min(length,ENTRY_LIMIT) : ENTRY_LIMIT);
        if (postData != null)
        {
          connection = getConnection(urlString,"POST",true,arguments);
          connection.setRequestProperty("Content-Type","application/json; charset=utf-8");

          // output data
          OutputStreamWriter outputStream = new OutputStreamWriter(connection.getOutputStream(),"UTF-8");
          outputStream.write(postData.toJSONString());
          outputStream.close();
        }
        else
        {
          connection = getConnection(urlString,"GET",false,arguments);
        }
        Timemine.printDebug(Settings.DebugTypes.GET,"Get '%s'",connection.getURL());

        // check result
        int responseCode = connection.getResponseCode();
        if ((responseCode == 401) || (responseCode == 403))
        {
          Timemine.printDebug(Settings.DebugTypes.GET,"Get '%s' fail (response code %d)",connection.getURL(),responseCode);
          throw new BackendException("Authorization failure");
        }

        // parse JSON result
        Object result;
        InputStream inputStream = connection.getInputStream();
        if (inputStream != null)
        {
          result = parser.parse(new InputStreamReader(inputStream,"UTF-8"),
                                new JIRAContainerFactory()
                               );
        }
        else
        {
          result = null;
        }

        // close connection
        connection.disconnect(); connection = null;

        // try to get total count
        int totalCount;
        if      (result instanceof JIRAObject)
        {
//Dprintf.dprintf("entryHandler=%s offset=%d length=%d",entryHandler,offset,length);
          offset     = getIntValue((JIRAObject)result,"startAt",   offset);
          length     = getIntValue((JIRAObject)result,"maxResults",length);
          totalCount = getIntValue((JIRAObject)result,"total",     0     );
//Dprintf.dprintf("entryHandler=%s offset=%d length=%d totalCount=%d",entryHandler,offset,length,totalCount);
        }
        else if (result instanceof JIRAArray)
        {
          offset     = 0;
          length     = ((JIRAArray)result).size();
          totalCount = ((JIRAArray)result).size();
        }
        else
        {
          totalCount = 0;
        }
        entryHandler.setTotalCount(totalCount);

        // get data
        entryHandler.setIndex(offset);
        int n = 0;
        if (names != null)
        {
          n = getData(result,names,0,entryHandler);
        }
//Dprintf.dprintf("n=%d offset=%d totalCount=%d",n,offset,totalCount);

        // check if done
//Dprintf.dprintf("entry=%s",entry);
        done =    entryHandler.isDone()
               || ((offset+length) >= totalCount)
               || (n < ENTRY_LIMIT);

        // next segment
        offset += ENTRY_LIMIT;

        // throttle execution
        if (!done)
        {
          try { Thread.sleep(Settings.throttleTime); } catch (InterruptedException exception) { /* ignored */ }
        }
      }
      catch (org.json.simple.parser.ParseException exception)
      {
        Timemine.printDebug("Get data fail '%s': %s",connection,exception.getMessage());
        throw new BackendException("JSON parse fail",exception);
      }
      catch (ProtocolException exception)
      {
        Timemine.printDebug("Get data fail '%s': %s",connection,exception.getMessage());
        throw new BackendException(exception);
      }
      catch (UnknownHostException exception)
      {
        Timemine.printDebug("Get data fail '%s': %s",connection,exception.getMessage());
        throw new BackendException("Unknown host '"+serverName+"'",exception);
      }
      catch (IOException exception)
      {
        Timemine.printDebug("%s",connection);
        entryHandler.error(exception);
        break;
      }
      finally
      {
        if (connection != null) connection.disconnect();
      }
    }

    return entryHandler.getResult();
  }

  /** get data
   * @param urlString URL string
   * @param name JSON entry name to visit
   * @param offset start offset
   * @param length max. number of entries to get
   * @param postData post data for POST or null for GET
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String          urlString,
                        String          name,
                        int             offset,
                        int             length,
                        JIRAObject      postData,
                        EntryHandler<T> entryHandler,
                        String...       additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    return getData(urlString,new String[]{name},offset,length,postData,entryHandler,additionalArguments);
  }

  /** get all data
   * @param urlString URL string
   * @param names JSON entry names to visit
   * @param postData post data for POST or null for GET
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String          urlString,
                        String[]        names,
                        JIRAObject      postData,
                        EntryHandler<T> entryHandler,
                        String...       additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    return getData(urlString,names,0,-1,postData,entryHandler,additionalArguments);
  }

  /** get all data
   * @param urlString URL string
   * @param names JSON entry names to visit
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String          urlString,
                        String[]        names,
                        EntryHandler<T> entryHandler,
                        String...       additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    return getData(urlString,names,0,-1,(JIRAObject)null,entryHandler,additionalArguments);
  }

  /** get all data
   * @param urlString URL string
   * @param name JSON entry name to visit
   * @param postData post data for POST or null for GET
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String          urlString,
                        String          name,
                        JIRAObject      postData,
                        EntryHandler<T> entryHandler,
                        String...       additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    return getData(urlString,new String[]{name},postData,entryHandler,additionalArguments);
  }

  /** get all data
   * @param urlString URL string
   * @param name JSON entry name to visit
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String          urlString,
                        String          name,
                        EntryHandler<T> entryHandler,
                        String...       additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    return getData(urlString,name,(JIRAObject)null,entryHandler,additionalArguments);
  }

  /** get all data
   * @param urlString URL string
   * @param offset start offset
   * @param length max. number of entries to get
   * @param postData post data for POST or null for GET
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String          urlString,
                        int             offset,
                        int             length,
                        JIRAObject      postData,
                        EntryHandler<T> entryHandler,
                        String...       additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    return getData(urlString,new String[0],offset,length,postData,entryHandler,additionalArguments);
  }

  /** get all data
   * @param urlString URL string
   * @param postData post data for POST or null for GET
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String          urlString,
                        JIRAObject      postData,
                        EntryHandler<T> entryHandler,
                        String...       additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    return getData(urlString,0,-1,postData,entryHandler,additionalArguments);
  }

  /** get all data
   * @param urlString URL string
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return parsed entity
   */
  private <T> T getData(String          urlString,
                        EntryHandler<T> entryHandler,
                        String...       additionalArguments
                       )
    throws BackendException, InterruptedException
  {
    return getData(urlString,(JIRAObject)null,entryHandler,additionalArguments);
  }

  /** iterate over all data
   * @param urlString URL string
   * @param name XML entry name
   * @param postData post data for POST or null for GET
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return result
   */
  private <T> Object iterateData(String          urlString,
                                 String          name,
                                 JIRAObject      postData,
                                 EntryHandler<T> entryHandler,
                                 String...       additionalArguments
                                )
    throws BackendException, InterruptedException
  {
    getData(urlString,name,postData,entryHandler,additionalArguments);

    return entryHandler.getResult();
  }

  /** iterate over all data
   * @param urlString URL string
   * @param postData post data for POST or null for GET
   * @param entryHandler entry handler
   * @param additionalArguments additional arguments (optional)
   * @return result
   */
  private <T> Object iterateData(String          urlString,
                                 JIRAObject      postData,
                                 EntryHandler<T> entryHandler,
                                 String...       additionalArguments
                                )
    throws BackendException, InterruptedException
  {
    getData(urlString,(String[])null,0,-1,postData,entryHandler,additionalArguments);

    return entryHandler.getResult();
  }

  /** get length of data
   * @param urlString URL string
   * @param name XML entry name
   * @return length of data
   */
  private int getDataLength(String urlString, String name)
    throws BackendException, InterruptedException
  {
    final int[] result = new int[1];

    getData(urlString,name,0,1,(JIRAObject)null,new EntryHandler()
    {
      @Override
      public void root(JIRAObject entry)
      {
        result[0] = getIntValue(entry,"total_count");
      }
    });
    Timemine.printDebug(Settings.DebugTypes.GET,"Get length '"+urlString+"': "+result[0]);

    return result[0];
  }

  /** add data via POST
   * @param urlString URL string
   * @param entity entity for storage
   * @param createHandler create handler
   * @param additionalArguments additional arguments (optional)
   */
  private <T extends Entity> void postData(String        urlString,
                                           int           returnCode,
                                           T             entity,
                                           CreateHandler createHandler,
                                           String...     additionalArguments
                                          )
    throws BackendException
  {
    if (!Settings.readOnlyFlag)
    {
      HttpURLConnection connection = null;
      try
      {
        // create JSON
        JIRAObject rootEntry = new JIRAObject();
        createHandler.data(rootEntry,entity);

        // debug
        if (Settings.debugTypes.contains(Settings.DebugTypes.WRITE))
        {
          Timemine.printDebug("POST: %s",urlString);
          Timemine.printDebug("JSON: %s",rootEntry.toString());
          Dprintf.printStackTrace();
          if (!debugConfirm("POST: "+urlString+"\n"+
                            "JSON: "+rootEntry.toString()
                           )
             )
          {
            return;
          }
        }

        // create JSON result parser
        JSONParser parser = new JSONParser();

        // connect to JIRA server
        connection = getConnection(urlString,"POST",true,additionalArguments);
        connection.setRequestProperty("Content-Type","application/json; charset=utf-8");
        Timemine.printDebug(Settings.DebugTypes.PUT,"Post '"+connection.getURL()+"'");
        Timemine.printDebug(Settings.DebugTypes.PUT,"JSON: %s",rootEntry.toJSONString());

        // output data
        OutputStreamWriter outputStream = new OutputStreamWriter(connection.getOutputStream(),"UTF-8");
        outputStream.write(rootEntry.toJSONString());
        outputStream.close();

        // check result
        int responseCode = connection.getResponseCode();
        if ((responseCode == 401) || (responseCode == 403))
        {
          throw new BackendException("Authorization failure");
        }
        String responseMessage = connection.getResponseMessage();
        Timemine.printDebug(Settings.DebugTypes.PUT,"Result "+responseCode+": "+((responseMessage != null) ? responseMessage : ""));
        if (responseCode != returnCode)
        {
          InputStream errorStream = connection.getErrorStream();
          if (errorStream != null)
          {
            Object result        = parser.parse(new InputStreamReader(errorStream,"UTF-8"));
            Object errors        = ((JSONObject)result).get("errors");
            Object errorMessages = ((JSONObject)result).get("errorMessages");
            responseMessage = (errors != null) ? errors.toString() : ((errorMessages != null) ? errorMessages.toString() : "unknown error");
          }
          throw new BackendException("Server error "+responseCode+": "+responseMessage);
        }

        // get backend id
        InputStream inputStream = connection.getInputStream();
        if (inputStream != null)
        {
          Object result = parser.parse(new InputStreamReader(inputStream,"UTF-8"),new JIRAContainerFactory());
          entity.id.backend = getIntValue((JIRAObject)result,"id");
        }

        // close connectin
        connection.disconnect(); connection = null;
      }
      catch (org.json.simple.parser.ParseException exception)
      {
        // ignored
      }
      catch (UnknownHostException exception)
      {
        Timemine.printDebug("Post data fail '%s': %s",connection,exception.getMessage());
        throw new BackendException("Unknown host '"+serverName+"'",exception);
      }
      catch (IOException exception)
      {
        Timemine.printDebug("Post data fail '%s': %s",connection,exception.getMessage());
        throw new BackendException("Send data fail: "+exception.getMessage(),exception);
      }
      finally
      {
        if (connection != null) connection.disconnect();
      }
    }
  }

  /** update data via PUT
   * @param urlString URL string
   * @param returnCode expected return code
   * @param entity entity for storage
   * @param createHandler create handler
   * @param additionalArguments additional arguments (optional)
   */
  private <T extends Entity> void putData(String urlString, int returnCode, T entity, CreateHandler createHandler, String... additionalArguments)
    throws BackendException
  {
    if (!Settings.readOnlyFlag)
    {
      HttpURLConnection connection = null;
      try
      {
        // create JSON
        JIRAObject rootEntry = new JIRAObject();
        createHandler.data(rootEntry,entity);

        // debug
        if (Settings.debugTypes.contains(Settings.DebugTypes.WRITE))
        {
          Timemine.printDebug("PUT: %s",urlString);
          Timemine.printDebug("JSON: %s",rootEntry.toString());
          Dprintf.printStackTrace();
          if (!debugConfirm("PUT : "+urlString+"\n"+
                            "JSON: "+rootEntry.toString()
                           )
             )
          {
            return;
          }
        }

        // update data on JIRA server
        connection = getConnection(urlString,"PUT",true,additionalArguments);
        connection.setRequestProperty("Content-Type","application/json; charset=utf-8");
        if (Settings.debugTypes.contains(Settings.DebugTypes.PUT))
        {
          Timemine.printDebug("Put '"+connection.getURL()+"':");
          Timemine.printDebug("JSON: %s",rootEntry.toJSONString());
        }

        // output data
        OutputStreamWriter outputStream = new OutputStreamWriter(connection.getOutputStream(),"UTF-8");
        outputStream.write(rootEntry.toJSONString());
        outputStream.close();

        // check result
        int responseCode = connection.getResponseCode();
        if ((responseCode == 401) || (responseCode == 403))
        {
          Timemine.printDebug(Settings.DebugTypes.PUT,"Put '%s' fail (response code %d)",connection.getURL(),responseCode);
          throw new BackendException("Authorization failure");
        }
        String responseMessage = connection.getResponseMessage();
        Timemine.printDebug(Settings.DebugTypes.PUT,"Result "+responseCode+": "+((responseMessage != null) ? responseMessage : ""));
        if (responseCode != returnCode)
        {
          throw new BackendException("Server error "+responseCode+": "+((responseMessage != null) ? responseMessage : "unknown"));
        }

        // close connectin
        connection.disconnect(); connection = null;
      }
      catch (UnknownHostException exception)
      {
        throw new BackendException("Unknown host '"+serverName+"'");
      }
      catch (IOException exception)
      {
        throw new BackendException("Send data fail: "+exception.getMessage());
      }
      finally
      {
        if (connection != null) connection.disconnect();
      }
    }
  }

  /** delete data
   * @param urlString URL string
   * @param name XML entry name
   * @param additionalArguments additional arguments (optional)
   */
  private void deleteData(String urlString, String... additionalArguments)
    throws BackendException
  {
    if (!Settings.readOnlyFlag)
    {
      HttpURLConnection connection = null;
      try
      {
        // debug
        if (Settings.debugTypes.contains(Settings.DebugTypes.DELETE))
        {
          Timemine.printDebug("DELETE: %s",urlString);
          Dprintf.printStackTrace();
          if (!debugConfirm("DELETE: "+urlString
                           )
             )
          {
            return;
          }
        }

        // delete data from JIRA server
        connection = getConnection(urlString,"DELETE",true,additionalArguments);
        connection.setRequestProperty("Content-Type","application/json; charset=utf-8");
        Timemine.printDebug(Settings.DebugTypes.PUT,"Delete '"+connection.getURL()+"'");

        // check result
        int responseCode = connection.getResponseCode();
        if ((responseCode == 401) || (responseCode == 403))
        {
          throw new BackendException("Authorization failure");
        }
        String responseMessage = connection.getResponseMessage();
        Timemine.printDebug(Settings.DebugTypes.PUT,"Result "+responseCode+": "+((responseMessage != null) ? responseMessage : "OK"));
        if ((responseCode != HttpURLConnection.HTTP_OK) && (responseCode != HttpURLConnection.HTTP_NO_CONTENT) && (responseCode != HttpURLConnection.HTTP_NOT_FOUND))
        {
          throw new BackendException("Server error "+responseCode+": "+((responseMessage != null) ? responseMessage : "unknown"));
        }

        // close connectin
        connection.disconnect(); connection = null;
      }
      catch (UnknownHostException exception)
      {
        throw new BackendException("Unknown host '"+serverName+"'");
      }
      catch (IOException exception)
      {
        throw new BackendException("Send data fail: "+exception.getMessage());
      }
      finally
      {
        if (connection != null) connection.disconnect();
      }
    }
  }

  /** get string value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param names names
   * @param defaultValue default value
   * @return value
   */
  private String getValue(JIRAObject entry, String[] names, String defaultValue)
  {
    // find data
    Object data = entry;
    for (String name : names)
    {
      data = ((JIRAObject)data).get(name);
      if (data == null) break;
    }

    return (data != null) ? (String)data : defaultValue;
  }

  /** get integer value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param names names
   * @return value
   */
  private String getValue(JIRAObject entry, String... names)
  {
    return getValue(entry,names,(String)null);
  }

  /** get value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return value
   */
  private String getValue(JIRAObject entry, String name, String defaultValue)
  {
    // find data
    Object data = entry.get(name);
    assert (data == null) || (data instanceof String) : "expected 'String' type for '"+name+"', got "+data.getClass();

    return (data != null) ? (String)data : defaultValue;
  }

  /** get value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param name name
   * @return value
   */
  private String getValue(JIRAObject entry, String name)
  {
    return getValue(entry,name,(String)null);
  }

  /** get id value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param names names
   * @param defaultValue default value
   * @return id value
   */
  private Id getIdValue(JIRAObject entry, String[] names, int defaultValue)
  {
    // find data
    Object data = entry;
    for (String name : names)
    {
      data = ((JIRAObject)data).get(name);
      if (data == null) break;
    }

    // get value
    Id id = new Id(defaultValue);
    if (data != null)
    {
      if      (data instanceof Integer) id.backend = (Integer)data;
      if      (data instanceof Long   ) id.backend = (int)(long)(Long)data;
      else if (data instanceof String ) id.backend = Integer.parseInt((String)data);
      else new UnsupportedOperationException("expected 'Integer' type for '"+StringUtils.join(names,'.')+"', got "+data.getClass());
    }

    return id;
  }

  /** get id value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param names names
   * @return id value
   */
  private Id getIdValue(JIRAObject entry, String... names)
  {
    return getIdValue(entry,names,Id.NONE);
  }

  /** get id value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return id value
   */
  private Id getIdValue(JIRAObject entry, String name, int defaultValue)
  {
    // find data
    Object data = entry.get(name);

    // get value
    Id id = new Id(defaultValue);
    if (data != null)
    {
      if      (data instanceof Integer) id.backend = (Integer)data;
      if      (data instanceof Long   ) id.backend = (int)(long)(Long)data;
      else if (data instanceof String ) id.backend = Integer.parseInt((String)data);
      else new UnsupportedOperationException("expected 'Integer' type for '"+name+"', got "+data.getClass());
    }

    return id;
  }

  /** get id value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param name name
   * @return id value
   */
  private Id getIdValue(JIRAObject entry, String name)
  {
    return getIdValue(entry,name,Id.NONE);
  }

  /** get boolean value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return boolean value
   */
  private boolean getBooleanValue(JIRAObject entry, String name, boolean defaultValue)
  {
    String value = (String)entry.get(name);
    if (value != null)
    {
      try
      {
        return Integer.parseInt(value) != 0;
      }
      catch (NumberFormatException exception)
      {
        return    value.equalsIgnoreCase("true")
               || value.equalsIgnoreCase("on")
               || value.equalsIgnoreCase("yes");
      }
    }
    else
    {
      return defaultValue;
    }
  }

  /** get boolean value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @return boolean value
   */
  private boolean getBooleanValue(JIRAObject entry, String name)
  {
    return getBooleanValue(entry,name,false);
  }

  /** get integer value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return integer value
   */
  private int getIntValue(JIRAObject entry, String[] names, int defaultValue)
  {
    // find data
    Object data = entry;
    for (String name : names)
    {
      data = ((JIRAObject)data).get(name);
      if (data == null) break;
    }

    // get value
    int value = defaultValue;
    if (data != null)
    {
      if      (data instanceof Integer) value = (Integer)data;
      if      (data instanceof Long   ) value = (int)(long)(Long)data;
      else if (data instanceof String ) value = Integer.parseInt((String)data);
      else new UnsupportedOperationException("expected 'Integer' type for '"+StringUtils.join(names,'.')+"', got "+data.getClass());
    }

    return value;
  }

  /** get integer value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return integer value
   */
  private int getIntValue(JIRAObject entry, String name, int defaultValue)
  {
    // find data
    Object data = entry.get(name);

    // get value
    int value = defaultValue;
    if (data != null)
    {
      if      (data instanceof Integer) value = (Integer)data;
      if      (data instanceof Long   ) value = (int)(long)(Long)data;
      else if (data instanceof String ) value = Integer.parseInt((String)data);
      else new UnsupportedOperationException("expected 'Integer' type for '"+name+"', got "+data.getClass());
    }

    return value;
  }

  /** get integer value from JIRA JSON object
   * @param entry JIRA JSON object
   * @param name name
   * @return integer value
   */
  private int getIntValue(JIRAObject entry, String name)
  {
    return getIntValue(entry,name,0);
  }

  /** get long value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @param defaultValue default value
   * @return long value
   */
  private long getLongValue(JIRAObject entry, String[] names, long defaultValue)
  {
    // find data
    Object data = entry;
    for (String name : names)
    {
      data = ((JIRAObject)data).get(name);
      if (data == null) break;
    }

    // get value
    long value = defaultValue;
    if (data != null)
    {
      if      (data instanceof Integer) value = (long)(int)(Integer)data;
      else if (data instanceof Long   ) value = (Long)data;
      else if (data instanceof String ) value = Long.parseLong((String)data);
      else new UnsupportedOperationException("expected 'Long' type for '"+StringUtils.join(names,'.')+"', got "+data.getClass());
    }

    return value;
  }

  /** get long value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @return long value
   */
  private long getLongValue(JIRAObject entry, String... names)
  {
    return getLongValue(entry,names,0L);
  }

  /** get long value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return long value
   */
  private long getLongValue(JIRAObject entry, String name, long defaultValue)
  {
    // find data
    Object data = entry.get(name);

    // get value
    long value = defaultValue;
    if (data != null)
    {
      if      (data instanceof Integer) value = (long)(int)(Integer)data;
      else if (data instanceof Long   ) value = (Long)data;
      else if (data instanceof String ) value = Long.parseLong((String)data);
      else new UnsupportedOperationException("expected 'Long' type for '"+name+"', got "+data.getClass());
    }

    return value;
  }

  /** get long value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @return long value
   */
  private long getLongValue(JIRAObject entry, String name)
  {
    return getLongValue(entry,name,0L);
  }

  /** get double from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @param defaultValue default value
   * @return double value
   */
  private double getDoubleValue(JIRAObject entry, String[] names, double defaultValue)
  {
    // find data
    Object data = entry;
    for (String name : names)
    {
      data = ((JIRAObject)data).get(name);
      if (data == null) break;
    }

    // get value
    double value = defaultValue;
    if (data != null)
    {
      if      (data instanceof Integer) value = (double)(int)(Integer)data;
      else if (data instanceof Long   ) value = (double)(long)(Long)data;
      else if (data instanceof Double ) value = (Double)data;
      else if (data instanceof String ) value = Double.parseDouble((String)data);
      else new UnsupportedOperationException("expected 'Double' type for '"+StringUtils.join(names,'.')+"', got "+data.getClass());
    }

    return value;
  }

  /** get double from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @return double value
   */
  private double getDoubleValue(JIRAObject entry, String... names)
  {
    return getDoubleValue(entry,names,0.0);
  }

  /** get long value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return double value
   */
  private double getDoubleValue(JIRAObject entry, String name, double defaultValue)
  {
    // find data
    Object data = entry.get(name);

    // get value
    double value = defaultValue;
    if (data != null)
    {
      if      (data instanceof Integer) value = (double)(int)(Integer)data;
      else if (data instanceof Long   ) value = (double)(long)(Long)data;
      else if (data instanceof Double ) value = (Double)data;
      else if (data instanceof String ) value = Double.parseDouble((String)data);
      else new UnsupportedOperationException("expected 'Double' type for '"+name+"', got "+data.getClass());
    }

    return value;
  }

  /** get double from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @return double value
   */
  private double getDoubleValue(JIRAObject entry, String name)
  {
    return getDoubleValue(entry,name,0.0);
  }

  /** get enum from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @param clazz enum class
   * @return enum value
   */
  private <T extends Enum<T>> T getEnumValue(JIRAObject entry, String name, T defaultValue, Class<T> clazz)
  {
    String text = getValue(entry,name).toLowerCase().trim();

    if (!text.isEmpty())
    {
      // find enum value by name
      for (T value : clazz.getEnumConstants())
      {
        if (text.equals(value.name().toLowerCase()))
        {
          return value;
        }
      }

      // find enum value by value
      try
      {
        int n = Integer.parseInt(text);
        for (T value : clazz.getEnumConstants())
        {
          if (value.ordinal() == n)
          {
            return value;
          }
        }
      }
      catch (NumberFormatException exception)
      {
        // nothing to do
      }
    }

    return defaultValue;
  }

  /** get date value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @param defaultValue default value
   * @return date value
   */
  private Date getDateValue(JIRAObject entry, String[] names, Date defaultValue)
  {
    // find data
    Object data = entry;
    for (String name : names)
    {
      data = ((JIRAObject)data).get(name);
      if (data == null) break;
    }
    assert (data == null) || (data instanceof String) : "expected 'String' type for '"+StringUtils.join(names,'.')+"', got "+data.getClass();

    return parseDate((String)data);
  }

  /** get date value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @return date value
   */
  private Date getDateValue(JIRAObject entry, String... names)
  {
    return getDateValue(entry,names,new Date());
  }

  /** get date value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return date value
   */
  private Date getDateValue(JIRAObject entry, String name, Date defaultValue)
  {
    Object data = entry.get(name);
    assert (data == null) || (data instanceof String) : "expected 'String' type for '"+name+"', got "+data.getClass();

    return (data != null) ? parseDate((String)data) : defaultValue;
  }

  /** get date value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @return date value
   */
  private Date getDateValue(JIRAObject entry, String name)
  {
    return getDateValue(entry,name,new Date());
  }

  /** get spent-onvalue from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @return spent-on value
   */
  private SpentOn getSpentOn(JIRAObject entry, String name)
  {
    return new SpentOn(getDateValue(entry,name));
  }

  /** get long value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @param defaultValue default value
   * @return long value
   */
  private JIRAObject getObjectValue(JIRAObject entry, String[] names, JIRAObject defaultValue)
  {
    // find data
    Object data = entry;
    for (String name : names)
    {
      data = ((JIRAObject)data).get(name);
      if (data == null) break;
    }

    return (data != null) ? (JIRAObject)data : defaultValue;
  }

  /** get long value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @return long value
   */
  private JIRAObject getObjectValue(JIRAObject entry, String... names)
  {
    return getObjectValue(entry,names,(JIRAObject)null);
  }

  /** get long value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return long value
   */
  private JIRAObject getObjectValue(JIRAObject entry, String name, JIRAObject defaultValue)
  {
    // find data
    Object data = entry.get(name);

    return (data != null) ? (JIRAObject)data : defaultValue;
  }

  /** get long value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @return long value
   */
  private JIRAObject getObjectValue(JIRAObject entry, String name)
  {
    return getObjectValue(entry,name,(JIRAObject)null);
  }

  /** get array value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @param defaultValue default value
   * @return array value
   */
  private JIRAArray getArrayValue(JIRAObject entry, String[] names, JIRAArray defaultValue)
  {
    // find data
    Object data = entry;
    for (String name : names)
    {
      data = ((JIRAObject)data).get(name);
      if (data == null) break;
    }

    return (data != null) ? (JIRAArray)data : defaultValue;
  }

  /** get array value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param names names
   * @return array value
   */
  private JIRAArray getArrayValue(JIRAObject entry, String... names)
  {
    return getArrayValue(entry,names,(JIRAArray)null);
  }

  /** get array value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @param defaultValue default value
   * @return array value
   */
  private JIRAArray getArrayValue(JIRAObject entry, String name, JIRAArray defaultValue)
  {
    // find data
    Object data = entry.get(name);

    return (data != null) ? (JIRAArray)data : defaultValue;
  }

  /** get array value from JIRA JSON entry
   * @param entry JIRA JSON object
   * @param name name
   * @return array value
   */
  private JIRAArray getArrayValue(JIRAObject entry, String name)
  {
    return getArrayValue(entry,name,(JIRAArray)null);
  }

  //--------------------------------------------------------------------

  /** parse date
   * @param string date/time string
   * @return date/time or current date/time
   */
  private Date parseDate(String string)
  {
    Date date = null;

//string="2012-09-11T16:44:32Z";
//string="2015-03-30T15:10:23Z";
    int i = 0;
    while ((i < dateFormats.length) && (date == null))
    {
      try
      {
//Dprintf.dprintf("%s: %s",dateFormats[i],string);
        synchronized(dateFormats[i])
        {
          date = dateFormats[i].parse(string);
        }
//Dprintf.dprintf("date=%s",date);
      }
      catch (ParseException exception)
      {
        // ignored, try next format
      }
catch (Exception e)
{
e.printStackTrace();
Dprintf.dprintf("%s: %s",dateFormats[i],string);
System.exit(1);
}
      i++;
    }
//Dprintf.dprintf("date=%s",date);
//System.exit(100);

    return (date != null) ? date : new Date();
  }

  //--------------------------------------------------------------------

  /** init JIRA work-around database
   * @return database
   */
  private static void initLocalDatabase()
    throws SQLException
  {
    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS meta ( "+
                                                    "  id    INTEGER PRIMARY KEY, "+
                                                    "  name  TEXT UNIQUE, "+
                                                    "  value TEXT "+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO meta (name,value) VALUES ('version',?)");
      preparedStatement.setString(1,"1");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO meta (name,value) VALUES ('dateTime',DATE('now'))");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS users ( "+
                                                    "  id  INTEGER PRIMARY KEY, "+
                                                    "  key TEXT, "+
                                                    ""+
                                                    "  CONSTRAINT users UNIQUE (key)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS epicNames ( "+
                                                    "  id   INTEGER PRIMARY KEY, "+
                                                    "  name TEXT, "+
                                                    ""+
                                                    "  CONSTRAINT users UNIQUE (name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS members ( "+
                                                    "  id  INTEGER PRIMARY KEY, "+
                                                    "  key TEXT, "+
                                                    ""+
                                                    "  CONSTRAINT members UNIQUE (key)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }
  }

  /** get list of all project ids from JIRA server
   * @param backendProgress progress handler
   * @return project backend id list
   */
  private ArrayList<Integer> getProjectIdList(final BackendProgress backendProgress)
  {
    final ArrayList<Integer> projectIdList = new ArrayList<Integer>();

    try
    {
      getData(REST_API+"/project",new EntryHandler<Void>()
      {
        @Override
        public Void process(JIRAObject entry, int index, int totalCount)
        {
          projectIdList.add(getIntValue(entry,"id"));
          return null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }
    catch (BackendException exception)
    {
      // ignored
    }
    catch (InterruptedException exception)
    {
      // ignored
    }


    return projectIdList;
  }

  /** map user key to user id
   * Note: JIRA does not use a numeric key for a referencing users :-( Instead JIRA
   *       invented a new "wheel" called "text key" because all other wheels in the
   *       world (unique numeric database keys) are not good...
   * @param key unique user key
   * @param defaultUserId default user id
   * @return user id
   */
  private synchronized Id userKeyToId(String key, int defaultUserId)
  {
    Id id = new Id(defaultUserId);

    PreparedStatement preparedStatement = null;
    ResultSet         resultSet         = null;
    try
    {
      // get id (if exists)
      preparedStatement = database.prepareStatement("SELECT id FROM users WHERE key=?");
      preparedStatement.setString(1,key);
      resultSet = preparedStatement.executeQuery();
      if (resultSet.next())
      {
        id.backend = resultSet.getInt("id");
      }
      resultSet.close(); resultSet = null;
      preparedStatement.close(); preparedStatement = null;

      if (id.backend == defaultUserId)
      {
        // insert new mapping key -> id
        preparedStatement = database.prepareStatement("INSERT INTO users (key) VALUES (?)");
        preparedStatement.setString(1,key);

        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        // get generated backend id
        id.backend = database.getLastInsertId();
      }
    }
    catch (SQLException exception)
    {
      // ignored
      Timemine.printInternalError(exception,"Map user key '%s' to id fail",key);
    }
    finally
    {
      try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }

    return id;
  }

  /** map user name to user id (Note: create user in database if not exists)
   * @param key unique user key
   * @return user id
   */
  private synchronized Id userKeyToId(String key)
  {
    return userKeyToId(key,Id.NONE);
  }

  /** map epic name to epic name id
   * Note: JIRA does not use a numeric key for a referencing epic names :-(
   * @param epicName epic name
   * @param defaultEpicNameId default epic name id
   * @return epic name id
   */
  private synchronized Id epicNameToId(String name, int defaultEpicNameId)
  {
    Id id = new Id(defaultEpicNameId);

    PreparedStatement preparedStatement = null;
    ResultSet         resultSet         = null;
    try
    {
      // get id (if exists)
      preparedStatement = database.prepareStatement("SELECT id FROM epicNames WHERE name=?");
      preparedStatement.setString(1,name);
      resultSet = preparedStatement.executeQuery();
      if (resultSet.next())
      {
        id.backend = resultSet.getInt("id");
      }
      resultSet.close(); resultSet = null;
      preparedStatement.close(); preparedStatement = null;

      if (id.backend == defaultEpicNameId)
      {
        // insert new mapping key -> id
        preparedStatement = database.prepareStatement("INSERT INTO epicNames (name) VALUES (?)");
        preparedStatement.setString(1,name);

        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        // get generated backend id
        id.backend = database.getLastInsertId();
      }
    }
    catch (SQLException exception)
    {
      // ignored
      Timemine.printInternalError(exception,"Map epic name '%s' to id fail",name);
    }
    finally
    {
      try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }

    return id;
  }

  /** map epic name to epic name id
   * Note: JIRA does not use a numeric key for a referencing epic names :-(
   * @param epicName epic name
   * @return epic name id
   */
  private synchronized Id epicNameToId(String name)
  {
    return epicNameToId(name,Id.NONE);
  }

  /** map member key to member id
   * @param key unique member key
   * @param defaultMemberId default member id
   * @return member id
   */
  private synchronized Id memberKeyToId(String key, int defaultMemberId)
  {
    Id id = new Id(defaultMemberId);

    PreparedStatement preparedStatement = null;
    ResultSet         resultSet         = null;
    try
    {
      // get id (if exists)
      preparedStatement = database.prepareStatement("SELECT id FROM members WHERE key=?");
      preparedStatement.setString(1,key);
      resultSet = preparedStatement.executeQuery();
      if (resultSet.next())
      {
        id.backend = resultSet.getInt("id");
      }
      resultSet.close(); resultSet = null;
      preparedStatement.close(); preparedStatement = null;

      if (id.backend == defaultMemberId)
      {
        // insert new mapping key -> id
        preparedStatement = database.prepareStatement("INSERT INTO members (key) VALUES (?)");
        preparedStatement.setString(1,key);

        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        // get generated backend id
        id.backend = database.getLastInsertId();
      }
    }
    catch (SQLException exception)
    {
      // ignored
      Timemine.printInternalError(exception,"Map member key '%s' to id fail",key);
    }
    finally
    {
      try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }

    return id;
  }

  /** map member key to member id
   * @param key unique member key
   * @param defaultMemberId default member id
   * @return member id
   */
  private synchronized Id memberKeyToId(String key)
  {
    return memberKeyToId(key,Id.NONE);
  }

  /** get epic name custom field name
   * @return epic name custom field name or null
   */
  private String getEpicNameCustomField()
  {
    String epicNameCustomField = null;

    try
    {
      epicNameCustomField = getData(REST_API+"/field",new EntryHandler<String>()
      {
        @Override
        public String process(JIRAObject entry, int index, int totalCount)
        {
          return getValue(entry,"name").equals("Epic-Name") ? getValue(entry,"id") : null;
        }
        @Override
        protected void error(Exception exception)
        {
          // ignored
        }
      });
    }
    catch (BackendException exception)
    {
      // ignored
    }
    catch (InterruptedException exception)
    {
      // ignored
    }

    return epicNameCustomField;
  }

}

/* end of file */
