/***********************************************************************\
*
* $Revision: 904 $
* $Date: 2021-03-19 12:25:35 +0100 (Fri, 19 Mar 2021) $
* $Author: torsten $
* Contents: load/save program settings
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
import java.lang.annotation.Annotation;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Field;
import java.lang.reflect.Constructor;
import static java.lang.annotation.ElementType.FIELD;
import static java.lang.annotation.ElementType.TYPE;

import java.io.File;

import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.StringTokenizer;
import java.util.LinkedHashSet;
import java.util.regex.Pattern;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.RGB;

/****************************** Classes ********************************/

/** settings
 */
public class Settings
{
  /** bookmark
   */
  static class Bookmark
  {
    public Backend.Types backendType;
    public String        summary;
    public String        text;

    /** create bookmark
     * @param backendType backend type
     * @param summary summary text
     * @param text bookmark text
     */
    Bookmark(Backend.Types backendType, String summary, String text)
    {
      this.backendType = backendType;
      this.summary     = summary;
      this.text        = text;
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString()
    {
      return String.format("%s %s %s",
                           backendType.getName(),
                           StringUtils.escape(summary),
                           StringUtils.escape(text)
                          );
    }
  }

  /** config value adapter String <-> bookarmk
   */
  class SettingValueAdapterBookmark extends SettingUtils.ValueAdapter<String,Bookmark>
  {
    /** convert to value
     * @param string string
     * @return value or null
     */
    public Bookmark toValue(String string)
      throws Exception
    {
      Object[] data = new Object[3];
      if (StringParser.parse(string,"%s %\"S % S",data))
      {
        return new Bookmark(Backend.Types.toEnum((String)data[0]),
                            (String)data[1],
                            StringUtils.unescape((String)data[2])
                           );
      }
      else
      {
        // skipped
        Timemine.printWarning("Cannot parse bookmark '%s' - skipped",string);
        return null;
      }
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(Bookmark bookmark)
    {
      return bookmark.toString();
    }
  }

  /** config value adapter String <-> size
   */
  class SettingValueAdapterSize extends SettingUtils.ValueAdapter<String,Point>
  {
    /** convert to value
     * @param string string
     * @return value
     */
    public Point toValue(String string)
      throws Exception
    {
      Point point = null;

      StringTokenizer tokenizer = new StringTokenizer(string,"x");
      point = new Point(Integer.parseInt(tokenizer.nextToken()),
                        Integer.parseInt(tokenizer.nextToken())
                       );

      return point;
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(Point p)
      throws Exception
    {
      return String.format("%dx%d",p.x,p.y);
    }
  }

  /** config value adapter String <-> percentage
   */
  class SettingValueAdapterPercentage extends SettingUtils.ValueAdapter<String,Double>
  {
    /** convert to value
     * @param string string
     * @return value
     */
    public Double toValue(String string)
      throws Exception
    {
      return Double.parseDouble(StringUtils.trimEnd(string,'%'))/100.0;
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(Double value)
      throws Exception
    {
      return String.format("%f%%",value);
    }
  }

  /** column sizes
   */
  static class ColumnSizes
  {
    public final int[] width;

    /** create column sizes
     * @param width width (int array)
     */
    ColumnSizes(int... width)
    {
      int totalWidth = 0;
      this.width = new int[width.length];
      for (int z = 0; z < width.length; z++)
      {
        this.width[z] = width[z];
        totalWidth += this.width[z];
      }

      // force min. width of at least one element
      if (totalWidth == 0)
      {
        width[0] = 10;
      }
    }

    /** create column sizes
     * @param widthList with list
     */
    ColumnSizes(ArrayList<Integer> widthList)
    {
      int totalWidth = 0;
      this.width = new int[widthList.size()];
      for (int z = 0; z < widthList.size(); z++)
      {
        this.width[z] = widthList.get(z);
        totalWidth += this.width[z];
      }

      // force min. width of at least one element
      if (totalWidth == 0)
      {
        width[0] = 10;
      }
    }

    /** get width
     * @param columNb column index (0..n-1)
     * @return width or 0
     */
    public int get(int columNb)
    {
      return (columNb < width.length) ? width[columNb] : 0;
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      StringBuilder buffer = new StringBuilder();
      for (int n : width)
      {
        if (buffer.length() > 0) buffer.append(',');
        buffer.append(Integer.toString(n));
      }
      return "ColumnSizes {"+buffer.toString()+"}";
    }
  }

  /** config value adapter String <-> column width array
   */
  class SettingValueAdapterWidthArray extends SettingUtils.ValueAdapter<String,ColumnSizes>
  {
    /** convert to value
     * @param string string
     * @return value
     */
    public ColumnSizes toValue(String string)
      throws Exception
    {
      StringTokenizer tokenizer = new StringTokenizer(string,",");
      ArrayList<Integer> widthList = new ArrayList<Integer>();
      while (tokenizer.hasMoreTokens())
      {
        widthList.add(Integer.parseInt(tokenizer.nextToken()));
      }
      return new ColumnSizes(widthList);
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(ColumnSizes columnSizes)
      throws Exception
    {
      StringBuilder buffer = new StringBuilder();
      for (int width : columnSizes.width)
      {
        if (buffer.length() > 0) buffer.append(',');
        buffer.append(Integer.toString(width));
      }
      return buffer.toString();
    }
  }

  /** config value adapter String <-> hours
   */
  class SettingValueAdapterHours extends SettingUtils.ValueAdapter<String,Double>
  {
    /** convert to value
     * @param string string
     * @return value
     */
    public Double toValue(String string)
      throws Exception
    {
      double n = 0.0;

      StringTokenizer tokenizer = new StringTokenizer(string,":");
      n =   (double)Integer.parseInt(tokenizer.nextToken())
          + (double)Integer.parseInt(tokenizer.nextToken())/60.0;

      return n;
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(Double n)
      throws Exception
    {
      return String.format("%d:%d",(int)Math.floor(n),(int)Math.floor(n*60)%60);
    }
  }

  /** color
   */
  static class Color implements Cloneable
  {
    public final RGB DEFAULT_FOREGROUND;
    public final RGB DEFAULT_BACKGROUND;

    public RGB foreground;
    public RGB background;

    /** create color
     * @param foreground,background foreground/background RGB values
     */
    Color(RGB foreground, RGB background)
    {
      this.DEFAULT_FOREGROUND = foreground;
      this.DEFAULT_BACKGROUND = foreground;
      this.foreground         = foreground;
      this.background         = background;
    }

    /** create color
     * @param foreground foreground/background RGB values
     */
    Color(RGB foreground)
    {
      this(foreground,foreground);
    }

    /** clone object
     * @return cloned object
     */
    public Color clone()
    {
      return new Color(foreground,background);
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      return "Color {"+foreground+", "+background+"}";
    }
  }

  /** config value adapter String <-> Color
   */
  class SettingValueAdapterColor extends SettingUtils.ValueAdapter<String,Color>
  {
    /** convert to value
     * @param string string
     * @return value
     */
    public Color toValue(String string)
      throws Exception
    {
      Color color = null;

      Object[] data = new Object[6];
      if      (StringParser.parse(string,"%d,%d,%d:%d,%d,%d",data))
      {
        color = new Color(new RGB((Integer)data[0],(Integer)data[1],(Integer)data[2]),
                          new RGB((Integer)data[3],(Integer)data[4],(Integer)data[5])
                         );
      }
      else if (StringParser.parse(string,":%d,%d,%d",data))
      {
        color = new Color(null,new RGB((Integer)data[0],(Integer)data[1],(Integer)data[2]));
      }
      else if (StringParser.parse(string,"%d,%d,%d:",data))
      {
        color = new Color(new RGB((Integer)data[0],(Integer)data[1],(Integer)data[2]),null);
      }
      else if (StringParser.parse(string,"%d,%d,%d",data))
      {
        color = new Color(new RGB((Integer)data[0],(Integer)data[1],(Integer)data[2]));
      }
      else
      {
        Timemine.printWarning("Cannot parse color definition '%s' - skipped",string);
      }

      return color;
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(Color color)
      throws Exception
    {
      if      ((color.foreground != null) && (color.background != null))
      {
        if (   (color.foreground.red   != color.background.red  )
            || (color.foreground.green != color.background.green)
            || (color.foreground.blue  != color.background.blue )
           )
        {
          return  ((color.foreground != null) ? color.foreground.red+","+color.foreground.green+","+color.foreground.blue : "")
                 +":"
                 +((color.background != null) ? color.background.red+","+color.background.green+","+color.background.blue : "");
        }
        else
        {
          return color.foreground.red+","+color.foreground.green+","+color.foreground.blue;
        }
      }
      else if (color.foreground != null)
      {
        return color.foreground.red+","+color.foreground.green+","+color.foreground.blue+":";
      }
      else if (color.background != null)
      {
        return ":"+color.background.red+","+color.background.green+","+color.background.blue;
      }
      else
      {
        return "";
      }
    }
  }

  /** config value adapter String <-> Key
   */
  class SettingValueAdapterKey extends SettingUtils.ValueAdapter<String,Integer>
  {
    /** convert to value
     * @param string string
     * @return value
     */
    public Integer toValue(String string)
      throws Exception
    {
      int accelerator = 0;
      if (!string.isEmpty())
      {
        accelerator = Widgets.textToAccelerator(string);
        if (accelerator == 0)
        {
          Timemine.printWarning("Cannot parse key definition '%s' - skipped",string);
        }
      }

      return accelerator;
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(Integer accelerator)
      throws Exception
    {
      return Widgets.menuAcceleratorToText(accelerator);
    }
  }

  /** config value adapter String <-> id set
   */
//TODO: replace by string list
  class SettingValueAdapterIds extends SettingUtils.ValueAdapter<String,Integer>
  {
    /** convert to value
     * @param string string
     * @return value
     */
    public Integer toValue(String string)
      throws Exception
    {
      int n = 0;
      if (!string.isEmpty())
      {
        n = Integer.parseInt(string);
      }

      return n;
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(Integer value)
      throws Exception
    {
      return value.toString();
    }
  }

  /** time
   */
  static class Time extends Object implements Cloneable
  {
    private final SimpleDateFormat TIME_FORMAT = new SimpleDateFormat("HH:mm");

    private Calendar calendar;

    /** create time
     * @param time time
     */
    Time(Date time)
    {
      this();
      calendar.setTime(time);

      setHours(time.getHours());
      setMinutes(time.getMinutes());
    }

    /** create time
     * @param string time string
     */
    Time(String string)
      throws ParseException
    {
      this();
      calendar.setTime(TIME_FORMAT.parse(string));
    }

    /** create time
     * @param time time
     */
    Time()
    {
      calendar = Calendar.getInstance();
    }

    /** check if objects are equal
     * @param object object
     * @return true iff equal
     */
    public boolean equals(Object object)
    {
      return hashCode() == object.hashCode();
    }

    /** clone object
     * @return cloned object
     */
    public Time clone()
    {
      return new Time(calendar.getTime());
    }

     /** get time (with current date)
     * @return time
     */
    long get()
    {
      Calendar calendar = Calendar.getInstance();

      calendar.set(Calendar.HOUR_OF_DAY,getHours());
      calendar.set(Calendar.MINUTE,getMinutes());

      return calendar.getTimeInMillis();
    }

    /** get hours
     * @return hours
     */
    int getHours()
    {
      return calendar.get(Calendar.HOUR_OF_DAY);
    }

    /** set hours
     * @param hours hours
     */
    void setHours(int hours)
    {
      calendar.set(Calendar.HOUR_OF_DAY,hours);
    }

    /** get minutes
     * @return minutes
     */
    int getMinutes()
    {
      return calendar.get(Calendar.MINUTE);
    }

    /** set minutes
     * @param minutes minutes
     */
    void setMinutes(int minutes)
    {
      calendar.set(Calendar.MINUTE,minutes);
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      return TIME_FORMAT.format(calendar.getTime());
    }
  }

  /** config value adapter String <-> Time
   */
  class SettingValueAdapterTime extends SettingUtils.ValueAdapter<String,Time>
  {
    /** convert to value
     * @param string string
     * @return value
     */
    public Time toValue(String string)
      throws Exception
    {
      return new Time(string);
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(Time time)
      throws Exception
    {
      return time.toString();
    }
  }

  /** holiday date
   */
  static class HolidayDate extends Object implements Cloneable
  {
    private final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");

    private Calendar calendar;
    private String   description;

    /** create holiday date
     * @param date date
     * @param description or null
     */
    HolidayDate(Date date, String description)
    {
      this.calendar    = Calendar.getInstance();
      this.description = description;

      calendar.setTime(date);
    }

    /** create holiday date
     * @param date date
     */
    HolidayDate(Date date)
    {
      this(date,(String)null);
    }

    /** create holiday date
     * @param year year
     * @param month month
     * @param day day
     * @param description or null
     */
    HolidayDate(int year, int month, int day, String description)
    {
      this.calendar    = Calendar.getInstance();
      this.description = description;

      calendar.set(year,month-1,day);
    }

    /** create holiday date
     * @param year year
     * @param month month
     * @param day day
     */
    HolidayDate(int year, int month, int day)
    {
      this(year,month,day,(String)null);
    }

    /** create holiday date
     * @param string date string
     * @param description or null
     */
    HolidayDate(String string, String description)
      throws ParseException
    {
      this.calendar    = Calendar.getInstance(); this.calendar.setTime(DATE_FORMAT.parse(string));
      this.description = description;
    }

    /** create holiday date
     * @param string date string
     */
    HolidayDate(String string)
      throws ParseException
    {
      this(string,(String)null);
    }

    /** check if objects are equal
     * @param object object
     * @return true iff equal
     */
    public boolean equals(Object object)
    {
      return hashCode() == object.hashCode();
    }

    /** get hash code: day+month+year only
     * @return hash code
     */
    public int hashCode()
    {
      return   ((calendar.get(Calendar.DAY_OF_MONTH) & 0x0000001F) << 0)
             | ((calendar.get(Calendar.MONTH       ) & 0x0000000F) << 5)
             | ((calendar.get(Calendar.YEAR        ) & 0x007FFFFF) << 9);
    }

    /** clone object
     * @return cloned object
     */
    public HolidayDate clone()
    {
      return new HolidayDate(calendar.getTime(),description);
    }

    /** get date
     * @return date/time [s]
     */
    public long getDate()
    {
      return calendar.getTimeInMillis()/1000L;
    }

    /** get description
     * @return description
     */
    public String getDescription()
    {
      return description;
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      return DATE_FORMAT.format(calendar.getTime());
    }
  }

  /** config value adapter String <-> HolidayDate
   */
  class SettingValueAdapterHolidayDate extends SettingUtils.ValueAdapter<String,HolidayDate>
  {
    /** convert to value
     * @param string string
     * @return value
     */
    public HolidayDate toValue(String string)
      throws Exception
    {
      return new HolidayDate(string);
    }

    /** convert to string
     * @param value value
     * @return string
     */
    public String toString(HolidayDate holidayDate)
      throws Exception
    {
      return holidayDate.toString();
    }
  }

  // --------------------------- constants --------------------------------
  public static final String TIMEMINE_DIRECTORY = System.getProperty("user.home")+File.separator+".timemine";

  /** host system
   */
  public enum HostSystems
  {
    UNKNOWN,
    LINUX,
    SOLARIS,
    WINDOWS,
    MACOS,
    QNX
  };

  /** refresh types
   */
  public enum RefreshTypes
  {
    USERS,
    ROLES,
    TRACKERS,
    STATUS,
    PRIORITIES,
    ACTIVITIES,
    PROJECTS,
    VERSIONS,
    MEMBERS,
    ISSUES,
    TIME_ENTRIES,
    ALL;

    /** convert to string
     * @return string
     */
    public String toString()
    {
      switch (this)
      {
        case USERS:        return "users";
        case ROLES:        return "roles";
        case TRACKERS:     return "trackers";
        case STATUS:       return "status";
        case PRIORITIES:   return "priorities";
        case ACTIVITIES:   return "activities";
        case PROJECTS:     return "projects";
        case VERSIONS:     return "versions";
        case MEMBERS:      return "members";
        case ISSUES:       return "issues";
        case TIME_ENTRIES: return "time entries";
        case ALL:          return "all";
        default:           return "unknown";
      }
    }
  };

  /** debug types
   */
  public enum DebugTypes
  {
    INFO,
    WARNING,
    ERROR,

    GET,
    PUT,
    DELETE,

    REFRESH,
    LOCAL,
    SQL,
    WRITE
  };

//  public class DebugEnum extends OptionEnum<DebugTypes>
//  {
    /** parse debug type
     * @param string string
     * @return debug type
     */
//    public static OptionEnumSet<DebugTypes> parse(String string)
//    {
//return EnumSet.noneOf(DebugTypes.class);
//    }
//  };

  private static final String TIMEMINE_CONFIG_FILE_NAME = TIMEMINE_DIRECTORY+File.separator+"timemine.cfg";

  // --------------------------- variables --------------------------------

  @SettingComment(text={"Timemine configuration",""})

  // program settings
  public static HostSystems                    hostSystem                             = HostSystems.LINUX;

  public static enum CertificateTypes
  {
    NONE,
    CERTIFICATE_KEY,
    P12;

    /** parse ceritificate+key/P12
     * @param string string to parse
     * @param defaultValue default value
     * @return certificiate+key/p12
    */
    public static CertificateTypes parse(String string, CertificateTypes defaultValue)
    {
      if      (string.equalsIgnoreCase("NONEY"))           return NONE;
      else if (string.equalsIgnoreCase("CERTIFICATE+KEY")) return CERTIFICATE_KEY;
      else if (string.equalsIgnoreCase("P12")            ) return P12;
      else                                                 return defaultValue;
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      switch (this)
      {
        case NONE           : return "none";
        case CERTIFICATE_KEY: return "certificate+key";
        case P12            : return "P12";
      }

      return "CERTIFICATE+KEY";
    }
  };

  public static class Redmine
  {
    @SettingComment(text={"","Redmine server"})
    @SettingValue(name="serverNameRedmine")
    public static String                       serverName                             = "";
    @SettingValue(name="serverPortRedmine")
    public static int                          serverPort                             = 443;
    @SettingValue(name="serverUseSSLRedmine")
    public static boolean                      serverUseSSL                           = true;
    @SettingValue(name="loginNameRedmine")
    public static String                       loginName                              = "";
    // do not save password
    public static String                       loginPassword                          = "";
    @SettingValue(name="certificateKeyP12Redmine")
    public static CertificateTypes             certificateType                        = CertificateTypes.CERTIFICATE_KEY;
    @SettingValue(name="certificateFileNameRedmine")
    public static String                       certificateFileName                    = null;
    @SettingValue(name="keyFileNameRedmine")
    public static String                       keyFileName                            = null;
    @SettingValue(name="p12FileNameRedmine")
    public static String                       p12FileName                            = null;
    @SettingValue(name="keyPasswordRedmine")
    public static String                       keyPassword                            = null;
    @SettingValue(name="enabledRedmine")
    public static boolean                      enabled                                = true;

    @SettingComment(text={"","New issue"})
    @SettingValue(name="newIssueTrackerRedmine")
    public static String                       newIssueTrackerName                    = "Bug";
    @SettingValue(name="newIssueStatusRedmine")
    public static String                       newIssueStatusName                     = "New";
  };

  public static class JIRA
  {
    @SettingComment(text={"","JIRA server"})
    @SettingValue(name="serverNameJIRA")
    public static String                       serverName                             = "";
    @SettingValue(name="serverPortJIRA")
    public static int                          serverPort                             = 443;
    @SettingValue(name="serverUseSSLJIRA")
    public static boolean                      serverUseSSL                           = true;
    @SettingValue(name="loginNameJIRA")
    public static String                       loginName                              = "";
    // do not save password
    public static String                       loginPassword                          = "";
    @SettingValue(name="certificateKeyP12JIRA")
    public static CertificateTypes             certificateType                        = CertificateTypes.CERTIFICATE_KEY;
    @SettingValue(name="certificateFileNameJIRA")
    public static String                       certificateFileName                    = null;
    @SettingValue(name="keyFileNameJIRA")
    public static String                       keyFileName                            = null;
    @SettingValue(name="p12FileNameJIRA")
    public static String                       p12FileName                            = null;
    @SettingValue(name="keyPasswordJIRA")
    public static String                       keyPassword                            = null;
    @SettingValue(name="enabledJIRA")
    public static boolean                      enabled                                = true;

    @SettingComment(text={"","New issue"})
    @SettingValue(name="newIssueTrackerJIRA")
    public static String                       newIssueTrackerName                    = "Bug";
    @SettingValue(name="newIssueStatusJIRA")
    public static String                       newIssueStatusName                     = "New";
  };

//  @SettingValue(name="issueStatusIdFilter", type=SettingValueAdapterIds.class)
  @SettingValue(type=RefreshTypes.class)
  public static EnumSet<RefreshTypes>          refreshTypes                           = EnumSet.noneOf(RefreshTypes.class);

  // favorites
  @SettingComment(text={"","Favorites"})
  @SettingValue
  public static String[]                       favoriteProjects                       = new String[0];

  // issue settings
  @SettingComment(text={"","Issues"})
  @SettingValue
  public static String                         issueUpdateStatus                      = "";
  @SettingValue
  public static String                         issueCompletedStatus                   = "";
  @SettingValue(name="issueActiveStatus")
  public static HashSet<String>                issueActiveStatusSet                   = new SettingUtils.HashSet<String>();

  @SettingComment(text={"","Issue bookmarks"})
  @SettingValue(name="Bookmark",type=SettingValueAdapterBookmark.class)
  public static ArrayList<Bookmark>            issueBookmarkList                      = new SettingUtils.ArrayList<Bookmark>();

  @SettingComment(text={"","Issue status id filters"})
  @SettingValue
  public static boolean                        issueOwnFilter                         = false;
  @SettingValue(name="issueStatusIdFilter", type=SettingValueAdapterIds.class)
  public static HashSet<Integer>               issueStatusIdFilterSet                 = new SettingUtils.HashSet<Integer>();

  @SettingComment(text={"","Roles"})
  @SettingValue(name="mangerRole")
  public static HashSet<String>                managerRoles                           = new SettingUtils.HashSet<String>("Manager");
  @SettingValue(name="developerRole")
  public static HashSet<String>                developerRoles                         = new SettingUtils.HashSet<String>("Developer");
  @SettingValue(name="designerRole")
  public static HashSet<String>                designerRoles                          = new SettingUtils.HashSet<String>("Designer");

  // actions
  @SettingComment(text={"","Action filters"})
  @SettingValue(name="actionFilter")
  public static HashSet<String>                actionFilters                          = new SettingUtils.HashSet<String>();

  // time entry settings
  @SettingComment(text={"","Required hours sum per day/week [h]"})
  @SettingValue(type=SettingValueAdapterHours.class)
  public static double                         workingHoursPerDay                     = 8.0; // [h]
  @SettingValue(type=SettingValueAdapterHours.class)
  public static double                         workingHoursPerWeek                    = 5*8.0; // [h]
  @SettingComment(text={"","Pause time day [min]"})
  @SettingValue
  public static int                            pauseMinutesPerDay                     = 30; // [min]

  // vacation
  @SettingComment(text={"","Vacation time entry settings"})
  @SettingValue
//TODO: string?
  public static String                         vacationProject                        = "";
  @SettingValue
  public static String                         vacationIssue                          = "";
  @SettingValue
  public static String                         vacationActivity                       = "";
  @SettingValue
  public static String                         vacationComment                        = "";

  // timeouts
  @SettingComment(text={"","Refresh database time [min]"})
  @SettingValue
  public static int                            refreshTime                            = 120;  // [min]
  @SettingComment(text={"Refresh database issues time [min]"})
  @SettingValue
  public static int                            refreshIssuesTime                      = 120;  // [min]
  @SettingComment(text={"Increment refresh database issues time [min]"})
  @SettingValue
  public static int                            incrementalRefreshIssuesTime           = 60;  // [s]
  @SettingComment(text={"Refresh database time entries time [s]"})
  @SettingValue
  public static int                            refreshTimeEntriesTime                 = 120;  // [min]
  @SettingComment(text={"Increment refresh database time entries time [min]"})
  @SettingValue
  public static int                            incrementalRefreshTimeEntriesTime      = 60;  // [s]
  @SettingComment(text={"Max. history days for issues [day]"})
  @SettingValue
  public static int                            maxIssuesHistory                       = 60;  // [days]
  @SettingComment(text={"Max. history days for time entries [day]"})
  @SettingValue
  public static int                            maxTimeEntriesHistory                  = 60;  // [days]
  @SettingComment(text={"Synchronize database time [s]"})
  @SettingValue
  public static int                            syncTime                               = 30;  // [s]
  @SettingComment(text={"Throttle time for server requests [ms]"})
  @SettingValue
  public static int                            throttleTime                           = 2500;  // [ms]
  @SettingComment(text={"Refresh at start"})
  @SettingValue
  public static boolean                        refreshAtStart                         = false;
  @SettingComment(text={"Auto refresh time window"})
  @SettingValue
  public static boolean                        autoRefresh                            = false;
  @SettingValue(type=SettingValueAdapterTime.class)
  public static Time                           autoRefreshStartTime                   = new Time();
  @SettingValue(type=SettingValueAdapterTime.class)
  public static Time                           autoRefreshEndTime                     = new Time();

  // geometry
  @SettingComment(text={"","Geometry: <width>x<height>","Geometry columns: <width>,...","Geometry order: <name>,..."})
  @SettingValue(type=SettingValueAdapterSize.class)
  public static Point                          geometryMain                           = new Point(800,600);
  @SettingValue(type=SettingValueAdapterWidthArray.class)
  public static ColumnSizes                    geometryTodayTimeEntryColumns          = new ColumnSizes(50,70,80,100,200);
  @SettingValue(type=SettingUtils.ValueAdapterSimpleStringArray.class)
  public static SettingUtils.SimpleStringArray geometryTodayTimeEntryColumnOrder      = new SettingUtils.SimpleStringArray();
  @SettingValue(type=SettingUtils.ValueAdapterSimpleDoubleArray.class)
  public static SettingUtils.SimpleDoubleArray geometryTodayTimeSizes                 = new SettingUtils.SimpleDoubleArray(new double[]{0.8,0.2});
  @SettingValue(type=SettingValueAdapterWidthArray.class)
  public static ColumnSizes                    geometryActionTableColumns             = new ColumnSizes(45,300,0);
  @SettingValue(type=SettingValueAdapterWidthArray.class)
  public static ColumnSizes                    geometryWeekTimeEntryColumns           = new ColumnSizes(130,130,50,70,80,100,200);
  @SettingValue(type=SettingUtils.ValueAdapterSimpleStringArray.class)
  public static SettingUtils.SimpleStringArray geometryWeekTimeEntryColumnOrder       = new SettingUtils.SimpleStringArray();
  @SettingValue(type=SettingValueAdapterWidthArray.class)
  public static ColumnSizes                    geometryProjectsColumns                = new ColumnSizes(70,70,200,80,80);
  @SettingValue(type=SettingUtils.ValueAdapterSimpleStringArray.class)
  public static SettingUtils.SimpleStringArray geometryProjectsColumnOrder            = new SettingUtils.SimpleStringArray();
  @SettingValue(type=SettingValueAdapterWidthArray.class)
  public static ColumnSizes                    geometryIssuesColumns                  = new ColumnSizes(200,70,80,100,60,120,80,80);
  @SettingValue(type=SettingUtils.ValueAdapterSimpleStringArray.class)
  public static SettingUtils.SimpleStringArray geometryIssuesColumnOrder              = new SettingUtils.SimpleStringArray();
  @SettingValue(type=SettingValueAdapterSize.class)
  public static Point                          geometryEditIssue                      = new Point(600,400);
  @SettingValue(type=SettingValueAdapterSize.class)
  public static Point                          geometryEditTimeEntry                  = new Point(600,SWT.DEFAULT);
  @SettingValue(type=SettingValueAdapterSize.class)
  public static Point                          geometryPreferences                    = new Point(500,550);

  // colors
  @SettingComment(text={"","Colors: <rgb foreground>:<rgb background>, <rgb foreground>:, or :<rgb background>, or <rgb foreground+background>"})
  @SettingValue(type=SettingValueAdapterColor.class)
  public static Color                          colorTodayTimeEntries                  = new Color(null,new RGB(255,255,255));
  @SettingValue(type=SettingValueAdapterColor.class)
  public static Color                          colorTimeEntries                       = new Color(null,new RGB(255,255,255));
  @SettingValue(type=SettingValueAdapterColor.class)
  public static Color                          colorTimeEntriesUpdate                 = new Color(new RGB(0,128,255),null);
  @SettingValue(type=SettingValueAdapterColor.class)
  public static Color                          colorTimeEntriesIncomplete             = new Color(null,new RGB(255,128,128));
  @SettingValue(type=SettingValueAdapterColor.class)
  public static Color                          colorTimeEntriesWeekend                = new Color(new RGB(192,192,192),null);
  @SettingValue(type=SettingValueAdapterColor.class)
  public static Color                          colorTimeEntriesHoliday                = new Color(new RGB(160,160,160),null);
  @SettingValue(type=SettingValueAdapterColor.class)
  public static Color                          colorFuture                            = new Color(new RGB(192,192,192),null);
  @SettingValue(type=SettingValueAdapterColor.class)
  public static Color                          colorHoursExceeded                     = new Color(null,new RGB(255,196,196));

  // keyboard shortcuts
  @SettingComment(text={"","Accelerator keys"})
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyTabFavoriteProjects                 = SWT.F5;
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyTabTodayTimeEntries                 = SWT.F6;
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyTabTimeEntries                      = SWT.F7;
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyTabProjectsIssues                   = SWT.F8;
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyNewTimeEntry                        = SWT.NONE;
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyEditTimeEntry                       = SWT.NONE;
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keySaveTimeEntry                       = SWT.CTRL+'S';
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyDeleteTimeEntry                     = SWT.NONE;

  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyToggleOnlineOffline                 = SWT.CTRL+'O';
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyRefresh                             = SWT.CTRL+'R';
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyCleanUpLocalDatabase                = SWT.NONE;
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keySynchronize                         = SWT.CTRL+'W';
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyPreferences                         = SWT.CTRL+'E';
  @SettingValue(type=SettingValueAdapterKey.class)
  public static int                            keyQuit                                = SWT.CTRL+'Q';

  public static int                            keyFind                                = SWT.CTRL+'f';
  public static int                            keyFindIssue                           = SWT.CTRL+'i';
  public static int                            keyProject                             = SWT.CTRL+'p';
  public static int                            keyIssue                               = SWT.CTRL+'i';
  public static int                            keySpent                               = SWT.CTRL+'t';
  public static int                            keyActivity                            = SWT.CTRL+'y';
  public static int                            keyComments                            = SWT.CTRL+'c';
  public static int                            keyAdd                                 = SWT.CTRL+'i';
  public static int                            keySave                                = SWT.CTRL+'s';

//TODO: move to database
  // holiday dates
  @SettingComment(text={"","holiday dates [yyyy-mm-dd]"})
  @SettingValue(name="holidayDate", deprecated=true, deprecatedNames={"vacationDate"}, type=SettingValueAdapterHolidayDate.class)
  public static HashSet<HolidayDate>           holidayDateSet                         = new HashSet<HolidayDate>();

  // miscelanous
  @SettingComment(text={"","date/time formats"})
  @SettingValue
  public static String                         dateFormat                             = "yyyy-MM-dd EE";

  @SettingComment(text={"","Min. time delta [min]"})
  @SettingValue
  public static int                            minTimeDelta                           = 30; // [min]

  @SettingComment(text={"","Caching expire times [s]"})
  @SettingValue
  public static int                            cacheExpireTime                        = 30; // [s]

  @SettingComment(text={"","Snap time ranges to actions"})
  @SettingValue
  public static Boolean                        snapTimeRanges                         = new Boolean(true);

  // show flags
  @SettingComment(text={"","show dialog flags"})
  @SettingValue
  public static Boolean                        enterIssueProgress                     = new Boolean(true);
  @SettingValue
  public static Boolean                        showSplashScreen                       = new Boolean(true);
  @SettingValue
  public static Boolean                        showUpdateStatusErrors                 = new Boolean(true);
  @SettingValue
  public static Boolean                        showRestartAfterConfigChanged          = new Boolean(true);
  @SettingValue
  public static Boolean                        showUpdateIssuesWarning                = new Boolean(true);
  @SettingValue
  public static Boolean                        showUpdateTimeEntriesWarning           = new Boolean(true);
  @SettingValue
  public static Boolean                        showFutureTimeEntryWarning             = new Boolean(true);
  @SettingValue
  public static Boolean                        showNewVersionInformation              = new Boolean(true);;
  @SettingValue
  public static Boolean                        refreshAutoCloseFlag                   = false;

  @SettingValue
  public static Backend.ConnectionStates        connectionState                        = Backend.ConnectionStates.ONLINE;

  // miscelanous flags
  public static boolean                        loginDialogFlag                        = false;
  public static boolean                        actionMonitorFlag                      = false;
  public static boolean                        readOnlyFlag                           = false;
  public static EnumSet<DebugTypes>            debugTypes                             = EnumSet.noneOf(DebugTypes.class);
  public static boolean                        showAllErrorOutputFlag                 = false;
  public static boolean                        wipFlag                                = false;

  // help
  public static boolean                        helpFlag                               = false;

  // deprecated
  public static String                         serverName                             = "";
  public static int                            serverPort                             = 443;
  public static boolean                        serverUseSSL                           = true;
  public static Backend.Types                   serverType                             = Backend.Types.REDMINE;
  public static String                         loginName                              = "";
  // do not save password
  public static String                         loginPassword                          = "";

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** load program settings
   * @param fileName settings file name
   */
  public static void load(String fileName)
  {
    SettingUtils.load(new File(fileName));
  }

  /** load default program settings
   */
  public static void load()
  {
    load(TIMEMINE_CONFIG_FILE_NAME);
  }

  /** save program settings
   * @param fileName settings file name
   */
  public static void save(String fileName)
  {
    SettingUtils.save(new File(fileName));
  }

  /** save program settings with default name
   */
  public static void save()
  {
    save(TIMEMINE_CONFIG_FILE_NAME);
  }

  /** permitt program settings will be modified
   */
  public static void permitModified()
  {
    SettingUtils.permitModified();
  }

  /** check if program settings modify is permitted
   * @return true iff permitted
   */
  public static boolean isPermitModify()
  {
    return SettingUtils.isPermitModify();
  }


  /** set program settings are modified
   */
  public static void setModified()
  {
    SettingUtils.setModified();
  }

  /** check if program settings are modified
   * @return true iff modified
   */
  public static boolean isModified()
  {
    return SettingUtils.isModified();
  }

  /** check if program settings file is external modified
   * @return true iff modified
   */
  public static boolean isExternalModified()
  {
    return SettingUtils.isExternalModified(new File(TIMEMINE_CONFIG_FILE_NAME));
  }

  /** get all setting classes
   * @return classes array
   */
  public static Class[] getSettingClasses()
  {
    // get all setting classes
    ArrayList<Class> classList = new ArrayList<Class>();

    classList.add(Settings.class);
    for (Class clazz : Settings.class.getDeclaredClasses())
    {
//Dprintf.dprintf("c=%s",clazz);
      classList.add(clazz);
    }

    return classList.toArray(new Class[classList.size()]);
  }
}

/* end of file */
