/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** project database id
 */
class ProjectId extends Id
{
  /** create id
   * @param backend backend id
   */
  ProjectId(int backend)
  {
    super(backend);
  }
}

/** project
 */
class Project extends Entity
{
  public Id            parentId;
  public String        shortName;
  public String        name;
  public ProjectStates status;
  public String        identifier;
  public String        description;
  public final Date    createdOn;
  public Date          updatedOn;

  /** create project
   * @param backendType backend type
   * @param id database id
   * @param parentId parent issue id or Id.NONE
   * @param shortName short name or null
   * @param name name of project
   * @param status status
   * @param identifier name of identifier
   * @param description name of description
   * @param createdOn created-on date
   * @param updatedOn update-on date
   */
  Project(Backend.Types backendType,
          Id            id,
          Id            parentId,
          String        shortName,
          String        name,
          ProjectStates status,
          String        identifier,
          String        description,
          Date          createdOn,
          Date          updatedOn
         )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(name != null);
    assert(description != null);

    this.parentId    = parentId;
    this.shortName   = shortName;
    this.name        = name;
    this.status      = status;
    this.identifier  = identifier;
    this.description = description;
    this.createdOn   = createdOn;
    this.updatedOn   = updatedOn;
  }

  /** create project
   * @param backendType backend type
   * @param id database id
   * @param parentId parent issue id or Id.ANY
   * @param name name of project
   * @param status status
   * @param identifier name of identifier
   * @param description name of description
   * @param createdOn created-on date
   * @param updatedOn update-on date
   */
  Project(Backend.Types backendType,
          Id            id,
          Id            parentId,
          String        name,
          ProjectStates status,
          String        identifier,
          String        description,
          Date          createdOn,
          Date          updatedOn
         )
  {
    this(backendType,
         id,
         parentId,
         (String)null,  // shortName
         name,
         status,
         identifier,
         description,
         createdOn,
         updatedOn
        );
  }

  /** create project
   * @param backendType backend type
   * @param id database id
   * @param parentId parent issue id or Id.ANY
   * @param name name of project
   * @param status status
   * @param identifier name of identifier
   * @param description name of description
   */
  Project(Backend.Types backendType,
          Id            id,
          Id            parentId,
          String        shortName,
          String        name,
          ProjectStates status,
          String        identifier,
          String        description
         )
  {
    this(backendType,
         id,
         parentId,
         shortName,
         name,
         status,
         identifier,
         description,
         new Date(),
         new Date()
        );
  }

  /** create project
   * @param backendType backend type
   * @param id backend id
   */
  Project(Backend.Types backendType,
          Id            id
         )
  {
    this(backendType,
         id,
         new Id(),  // parentId
         null,  // shortName
         "",  // name
         ProjectStates.ACTIVE,
         "",  // identifier,
         ""  // description
        );
  }

  /** create project
   * @param id backend id
   */
  private Project(Id id)
  {
    super(Backend.Types.NONE,id);

    this.parentId    = new Id();
    this.shortName   = null;
    this.name        = "";
    this.status      = ProjectStates.ACTIVE;
    this.identifier  = "";
    this.description = "";
    this.createdOn   = new Date();
    this.updatedOn   = new Date();
  }

  /** create project
   */
  private Project()
  {
    this(new Id());
  }

  /** set local modified state
   * @param modifiedState modified state
   */
  public void setModified(Backend.ModifiedStates modifiedState)
    throws BackendException, InterruptedException
  {
    assert backendType != Backend.Types.NONE;

    {
      Backend.modifyLock();
    }
    try
    {
      if (modifiedState == Backend.ModifiedStates.DELETED)
      {
        Backend.removeAllLocalDatabase("projectsModified",
                                       Backend.condition("projectId=?",
                                                         id.local
                                                        )
                                      );
      }
      Backend.insertLocalDatabase("projectsModified",
                                  new Backend.Values
                                  (
                                   "projectId",    id.local,
                                   "modifiedState",modifiedState.ordinal()
  //TODO: modified values
                                  )
                                 );
      super.setModified(modifiedState);
    }
    finally
    {
      Backend.modifyUnlock();
    }
  }


  /** check iff activity in time entries is supported for project
   * @return true iff activity in time entries is supported
   */
  public boolean isActivitySupported()
  {
    return Backend.backendDrivers.get(backendType).isActivitySupported();
  }

  /** get sorted activity list
   * @return activity list
   */
  public ActivityList getActivityList()
    throws BackendException
  {
    return Backend.getActivityList(new Backend.TypeSet(backendType),new Id(Id.ANY));
  }

  /** get sorted issue list
   * @return issue list
   */
  public IssueList getIssueList()
    throws BackendException
  {
    return Backend.getIssueList(new Backend.TypeSet(backendType),id);
  }

  /** get issue by summary
   * @param summary issue summary
   * @return issue or Issue.NONE
   */
  public Issue getIssue(String summary)
  {
    Issue issue = Issue.NONE;

    try
    {
      issue = Backend.getIssue(new Backend.TypeSet(backendType),summary);
    }
    catch (BackendException exception)
    {
      // ignored
    }

    return issue;
  }

  /** get duration sum of time entries
   * @param issueId issue id or Id.ANY
   * @param userId user id or Id.ANY
   * @return duration sum of time entries or 0 if data still not available [min]
   */
  public final long getDurationSum(Id issueId, Id userId)
  {
    return Backend.getDurationSum(new Backend.TypeSet(backendType),id,issueId,userId);
  }

  /** get duration sum of time entries
   * @param userId user id or Id.ANY
   * @return duration sum of time entries or 0 if data still not available [min]
   */
  public final long getDurationSum(Id userId)
  {
    return getDurationSum(new Id(Id.ANY),userId);
  }

  /** get duration sum of time entries
   * @return duration sum of time entries or 0 if data still not available [min]
   */
  public final long getDurationSum()
  {
    return getDurationSum(new Id(Id.ANY));
  }

  /** get estimated duration sum of time entries
   * @param issueId issue id or Id.ANY
   * @return estimated duration sum of project and sub-projects or 0 if data still not available [min]
   */
  public final long getEstimatedDurationSum(Id issueId)
  {
    return Backend.getEstimatedDurationSum(new Backend.TypeSet(backendType),id,issueId);
  }

  /** get estimated duration sum of time entries
   * @return estimated duration sum of project and sub-projects or 0 if data still not available [min]
   */
  public final long getEstimatedDurationSum()
  {
    return getEstimatedDurationSum(new Id(Id.ANY));
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
    databaseParameters.put("name",name);

    return databaseParameters;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "Project { id="+id+", name="+name+" }";
  }
}

/** project list
 */
class ProjectList extends ArrayList<Project>
{
  final static ProjectList EMPTY = new ProjectList();
}

/** project comparator
 */
class ProjectComparator implements Comparator<Project>
{
  // --------------------------- constants --------------------------------
  enum SortModes
  {
    TYPE,
    ID,
    NAME,
    HOURS,
    ESTIMATED_HOURS
  };

  // --------------------------- variables --------------------------------
  private SortModes sortMode;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create time entry comparator
   */
  ProjectComparator()
  {
    this.sortMode = SortModes.TYPE;
  }

  /** get sort mode
   * @return sort mode
   */
  public SortModes getSortMode()
  {
    return sortMode;
  }

  /** set sort mode
   * @param sortMode sort mode
   */
  public void setSortMode(SortModes sortMode)
  {
    this.sortMode = sortMode;
  }

  /** compare projects
   * @param project0, project1 data to compare
   * @return -1 iff project0 < project1,
              0 iff project0 = project1,
              1 iff project0 > project1
   */
  public int compare(Project project0, Project project1)
  {
    if      (project0 == null)
    {
      return  1;
    }
    else if (project1 == null)
    {
      return -1;
    }
    else
    {
      switch (sortMode)
      {
        case TYPE:
          return project0.backendType.compareTo(project1.backendType);
        case ID:
          if      (project0.id.backend < project1.id.backend) return -1;
          else if (project0.id.backend > project1.id.backend) return  1;
          else                                                return  0;
        case NAME:
          return project0.name.compareToIgnoreCase(project1.name);
        case HOURS:
          long durationSum0 = project0.getDurationSum();
          long durationSum1 = project1.getDurationSum();
          if      (durationSum0 < durationSum1) return -1;
          else if (durationSum0 > durationSum1) return  1;
          else                                  return  0;
        case ESTIMATED_HOURS:
          long estimatedDurationSum0 = project0.getEstimatedDurationSum();
          long estimatedDurationSum1 = project1.getEstimatedDurationSum();
          if      (estimatedDurationSum0 < estimatedDurationSum1) return -1;
          else if (estimatedDurationSum0 > estimatedDurationSum1) return  1;
          else                                                    return  0;
        default:
          return 0;
      }
    }
  }
}

/** local database handler
 */
class ProjectDatabaseHandler extends DataBaseHandler<Project>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM projects"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public Project get(ValueSet valueSet)
    throws SQLException
  {
    return new Project(valueSet.getEnum   ("backendType",Backend.Types.class),
                       valueSet.getId     ("id","backendId"),
                       valueSet.getId     ("parentId"),
                       valueSet.getString ("shortName"),
                       valueSet.getString ("name"),
                       valueSet.getEnum   ("status",ProjectStates.class),
                       valueSet.getString ("identifier"),
                       valueSet.getString ("description"),
                       valueSet.getDate   ("createdOn"),
                       valueSet.getDate   ("updatedOn")
                      );
  }
}

/** modified project
 */
class ProjectModified extends Project
{
  int                  modifiedId;
  Backend.ModifiedStates modifiedState;

  /** create modified project
   * @param backendType backend type
   * @param id database id
   * @param parentId parent issue id or Id.NONE
   * @param shortName short name or null
   * @param name name of project
   * @param status status
   * @param identifier name of identifier
   * @param description name of description
   * @param createdOn created-on date
   * @param updatedOn update-on date
   */
  ProjectModified(Backend.Types         backendType,
                  Id                   id,
                  Id                   parentId,
                  String               shortName,
                  String               name,
                  ProjectStates        status,
                  String               identifier,
                  String               description,
                  Date                 createdOn,
                  Date                 updatedOn,
                  int                  modifiedId,
                  Backend.ModifiedStates modifiedState
                 )
  {
    super(backendType,
          id,
          parentId,
          shortName,
          name,
          status,
          identifier,
          description,
          createdOn,
          updatedOn
         );
    this.modifiedId    = modifiedId;
    this.modifiedState = modifiedState;
  }

  /** done project modification
   */
  protected void doneModified()
    throws BackendException, InterruptedException
  {
    {
      Backend.modifyLock();
    }
    try
    {
      Backend.removeLocalDatabase("projectsModified",
                                  modifiedId
                                 );
    }
    finally
    {
      Backend.modifyUnlock();
    }
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "ProjectModified { id="+modifiedId+", modifiedState="+modifiedState+", project="+super.toString()+" }";
  }

  /** clear all local project modifications
   */
  private void clearModified()
    throws BackendException, InterruptedException
  {
    {
      Backend.modifyLock();
    }
    try
    {
      Backend.removeAllLocalDatabase("projectsModified",
                                     Backend.condition("projectId=?",
                                                       id.local
                                                      )
                                    );
    }
    finally
    {
      Backend.modifyUnlock();
    }
  }
};

/** modified project database handler
 */
class ProjectModifiedDatabaseHandler extends DataBaseHandler<ProjectModified>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT "+
           "   projectsModified.id AS modifiedId,"+
           "   projectsModified.*,"+
           "   projects.id AS timeEntryId,"+
           "   projects.*"+
           " FROM projectsModified"+
           " LEFT JOIN projects ON projects.id=projectsModified.projectId"+
           " WHERE projects.id NOT NULL"+
           ((additionalFilter != null) ? " AND "+additionalFilter : "");
  }

  @Override
  public ProjectModified get(ValueSet valueSet)
    throws SQLException
  {
    return new ProjectModified(valueSet.getEnum   ("backendType",Backend.Types.class),
                               valueSet.getId     ("id","backendId"),
                               valueSet.getId     ("parentId"),
                               valueSet.getString ("shortName"),
                               valueSet.getString ("name"),
                               valueSet.getEnum   ("status",ProjectStates.class),
                               valueSet.getString ("identifier"),
                               valueSet.getString ("description"),
                               valueSet.getDate   ("createdOn"),
                               valueSet.getDate   ("updatedOn"),
                               valueSet.getInt    ("modifiedId"),
                               valueSet.getEnum   ("modifiedState",Backend.ModifiedStates.class)
                              );
  }
}

// project state
enum ProjectStates
{
  ACTIVE,
  CLOSED;

  /** convert ordinal value to enum
   * @param value ordinal value
   * @return enum
   */
  static ProjectStates toEnum(int value)
  {
    switch (value)
    {
      // Note: No specification for project status values available!
      case 1:  return ACTIVE;
      case 5:  return CLOSED;
      default: return ACTIVE;
    }
  }

  /** get value
   * @return value
   */
  public int getValue()
  {
    switch (this)
    {
      // Note: No specification for project status values available!
      case ACTIVE: return 1;
      case CLOSED: return 5;
      default:     return 1;
    }
  }
}

/* end of file */
