/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** tracker database id
 */
class TrackerId extends Id
{
  /** create id
   * @param backend backend id
   */
  TrackerId(int backend)
  {
    super(backend);
  }
}

/** tracker
 */
class Tracker extends Entity
{
  public final Id      projectId;
  public final String  name;
  public String        description;
  public final boolean requireEpicName;

  /** create tracker
   * @param backendType backend type
   * @param id database id
   * @param projectId project id or Id.ANY
   * @param name name
   * @param description description
   * @param requireEpicName true iff epic name is required
   */
  Tracker(Backend.Types backendType,
          Id            id,
          Id            projectId,
          String        name,
          String        description,
          boolean       requireEpicName
         )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(name != null);
    assert(description != null);

    this.projectId       = projectId;
    this.name            = name;
    this.description     = description;
    this.requireEpicName = requireEpicName;
  }

  /** create tracker
   * @param backendType backend type
   * @param id backend id
   * @param projectId project backend id
   * @param name name
   * @param description description
   */
  Tracker(Backend.Types backendType,
          Id            id,
          Id            projectId,
          String        name,
          String        description
         )
  {
    this(backendType,
         id,
         projectId,
         name,
         description,
         false
        );
  }

  /** create tracker
   * @param id backend id
   */
  private Tracker(Id id)
  {
    super(Backend.Types.NONE,id);

    this.projectId       = new Id();
    this.name            = "";
    this.description     = "";
    this.requireEpicName = false;
  }

  /** create tracker
   */
  private Tracker()
  {
    this(new Id());
  }

  /** check if tracker require epic name
   * @return true iff tracker require epic name
   */
  public boolean isRequireEpicName()
  {
    return requireEpicName;
  }

  /** get sorted epic name list for tracker
   * @return epic name list
   */
  public List<String> getEpicNameList()
  {
    List<String> epicNameList = new ArrayList<String>();

    try
    {
      epicNameList = Backend.getEpicNameList(new Backend.TypeSet(backendType),id);
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
      Timemine.printStackTrace(exception);
    }

    return epicNameList;
  }

  /** get database parameters
   * @return hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
    databaseParameters.put("projectId",id);
    databaseParameters.put("name",name);
    databaseParameters.put("description",description);
    databaseParameters.put("requireEpicName",requireEpicName);

    return databaseParameters;
  }

  /** get tracker name
   * @return name
   */
  public String getName()
  {
    return name;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "Tracker { id="+id+", projectId="+projectId+", name="+name+", description="+description+", requireEpicName="+requireEpicName+" }";
  }
}

/** tracker list
 */
class TrackerList extends ArrayList<Tracker>
{
  final static TrackerList EMPTY = new TrackerList();
}

/** tracker database handler
 */
class TrackerDatabaseHandler extends DataBaseHandler<Tracker>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM trackers"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public Tracker get(ValueSet valueSet)
    throws SQLException
  {
    return new Tracker(valueSet.getEnum   ("backendType",Backend.Types.class),
                       valueSet.getId     ("id","backendId"),
                       valueSet.getId     ("projectId"),
                       valueSet.getString ("name"),
                       valueSet.getString ("description"),
                       valueSet.getBoolean("requireEpicName")
                      );
  }
}

/* end of file */
