/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;

// SQLite database
import java.sql.ResultSet;
import java.sql.SQLException;

/****************************** Classes ********************************/

/** user database id
 */
class UserId extends Id
{
  /** create id
   * @param backend backend id
   */
  UserId(int backend)
  {
    super(backend);
  }
}

/** user
 */
class User extends Entity
{
  public final static User ANY = new User(new Id(Id.ANY));

  public String       firstName;        // first/given name
  public String       lastName;         // last name
  public final String name;             // user/login name
  public final Id     roleId;
  public final String mail;
  public final Date   createdOn;
  public final Date   lastLoginOn;

  /** create user
   * @param backendType backend type
   * @param id id
   * @param firstName first name
   * @param lastName last name
   * @param name name
   * @param roleId role id
   * @param mail mail address
   * @param createdOn create date/time
   * @param lastLoginOn last login date/time
   */
  User(Backend.Types backendType,
       Id            id,
       String        firstName,
       String        lastName,
       String        name,
       Id            roleId,
       String        mail,
       Date          createdOn,
       Date          lastLoginOn
      )
  {
    super(backendType,id);
    assert backendType != Backend.Types.NONE;
    assert !id.isNone();
    assert !id.isAny();
    assert(firstName != null);
    assert(lastName != null);
    assert(name != null);
    assert(mail != null);

    this.firstName   = firstName;
    this.lastName    = lastName;
    this.name        = name;
    this.roleId      = roleId;
    this.mail        = mail;
    this.createdOn   = createdOn;
    this.lastLoginOn = lastLoginOn;
  }

  /** create user
   * @param backendType backend type
   * @param id backend id
   * @param firstName first name
   * @param lastName last name
   * @param roleId role id
   * @param mail mail address
   * @param createdOn create date/time
   * @param lastLoginOn last login date/time
   */
  User(Backend.Types backendType,
       Id            id,
       String        firstName,
       String        lastName,
       Id            roleId,
       String        mail,
       Date          createdOn,
       Date          lastLoginOn
      )
  {
    this(backendType,
         id,
         firstName,
         lastName,
         "",
         roleId,
         mail,
         createdOn,
         lastLoginOn
        );
  }

  /** create user
   * @param backendType backend type
   * @param id backend id
   * @param firstName first name
   * @param lastName last name
   * @param roleId role id
   * @param mail mail address
   * @param createdOn create date/time
   * @param lastLoginOn last login date/time
   */
  User(Backend.Types backendType,
       Id            id,
       String        firstName,
       String        lastName,
       Id            roleId,
       String        mail
      )
  {
    this(backendType,
         id,
         firstName,
         lastName,
         roleId,
         mail,
         new Date(),
         new Date()
        );
  }

  /** create user
   * @param backendType backend type
   * @param id backend id
   * @param firstName first name
   * @param lastName last name
   * @param roleId role id
   */
  User(Backend.Types backendType,
       Id            id,
       String        firstName,
       String        lastName,
       Id            roleId
      )
  {
    this(backendType,
         id,
         firstName,
         lastName,
         roleId,
         ""
        );
  }

  /** create user
   * @param backendType backend type
   * @param id backend id
   * @param firstName first name
   * @param lastName last name
   * @param name name
   * @param mail mail address
   */
  User(Backend.Types backendType,
       Id            id,
       String        firstName,
       String        lastName,
       String        name,
       String        mail
      )
  {
    this(backendType,
         id,
         firstName,
         lastName,
         name,
         new Id(),
         mail,
         new Date(),
         new Date()
        );
  }

  /** create user
   * @param backendType backend type
   * @param id backend id
   */
  User(Backend.Types backendType,
       Id            id
      )
  {
    this(backendType,
         id,
         "",  // firstName,
         "",  // lastName,
         new Id()
        );
  }

  /** create user
   * @param id backend id
   */
  private User(Id id)
  {
    super(Backend.Types.NONE,id);

    this.firstName   = "";
    this.lastName    = "";
    this.name        = "";
    this.roleId      = new Id();
    this.mail        = "";
    this.createdOn   = new Date();
    this.lastLoginOn = new Date();
  }

  /** get database parameters
   * @returns hash map with <name,value>
   */
  HashMap<String,Object> getDatabaseParameters()
  {
    HashMap<String,Object> databaseParameters = new HashMap<String,Object>();
    databaseParameters.put("id",id);
    databaseParameters.put("firstName",firstName);
    databaseParameters.put("lastName",lastName);
    databaseParameters.put("name",name);
    databaseParameters.put("roleId",roleId);
    databaseParameters.put("mail",mail);
    databaseParameters.put("createdOn",createdOn);
    databaseParameters.put("lastLoginOn",lastLoginOn);

    return databaseParameters;
  }

  /** get full user name
   * @return name
   */
  public String getName()
  {
    StringBuilder buffer = new StringBuilder();
    buffer.append(lastName);
    if (!firstName.isEmpty())
    {
      buffer.append(", ");
      buffer.append(firstName);
    }

    return buffer.toString();
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "User { id="+id+", firstName="+firstName+", lastName="+lastName+", name="+name+", roleId="+roleId+", mail="+mail+" }";
  }
}

/** user list
 */
class UserList extends ArrayList<User>
{
  final static UserList EMPTY = new UserList();
}

/** user database handler
 */
class UserDatabaseHandler extends DataBaseHandler<User>
{
  @Override
  public String selectStatement(String additionalFilter)
  {
    return " SELECT *"+
           " FROM users"+
           ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
  }

  @Override
  public User get(ValueSet valueSet)
    throws SQLException
  {
    return new User(valueSet.getEnum  ("backendType",Backend.Types.class),
                    valueSet.getId    ("id","backendId"),
                    valueSet.getString("firstName"),
                    valueSet.getString("lastName"),
                    valueSet.getString("name"),
                    valueSet.getId    ("roleId"),
                    valueSet.getString("mail"),
                    valueSet.getDate  ("createdOn"),
                    valueSet.getDate  ("lastLoginOn")
                   );
  }
}

/* end of file */
