/***********************************************************************\
*
* $Revision: 1475 $
* $Date: 2013-09-30 22:20:25 +0200 (Mon, 30 Sep 2013) $
* $Author: trupp $
* Contents: database functions
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.File;
import java.lang.annotation.Annotation;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import static java.lang.annotation.ElementType.FIELD;
import static java.lang.annotation.ElementType.TYPE;

import java.util.AbstractCollection;

import java.sql.Array;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import java.util.ArrayList;
import java.util.StringTokenizer;

/****************************** Classes ********************************/

/** database
 */
public class Database
{
  /** database value annotation
   */
  @Target({TYPE,FIELD})
  @Retention(RetentionPolicy.RUNTIME)
  @interface Value
  {
    String name() default "";              // name of value
    Class  type() default DEFAULT.class;   // adapter class

    static final class DEFAULT
    {
    }
  }

  // --------------------------- constants --------------------------------
  public  final static int ID_NONE = -1;

  private final int        TIMEOUT = 10; // timeout [s]

  // --------------------------- variables --------------------------------
  private Connection connection;

  private String     name;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create database access
   * @param baseDirectory base directory
   * @param name name
   */
  Database(String baseDirectory, String name)
    throws SQLException
  {
    this.name = name;

    // create directory
    File directory = new File(baseDirectory);
    if ((directory != null) && !directory.exists()) directory.mkdirs();

    try
    {
      // load SQLite driver class
      Class.forName("org.sqlite.JDBC");

      // open database
      connection = null;
      try
      {
        DriverManager.setLoginTimeout(TIMEOUT);
        connection = DriverManager.getConnection("jdbc:sqlite:"+baseDirectory+File.separator+name+".db");
//        connection.setAutoCommit(false);
      }
      catch (SQLException exception)
      {
        if (connection != null) try { connection.close(); } catch (SQLException unusedException) { /* ignored */ }
        throw exception;
      }
    }
    catch (ClassNotFoundException exception)
    {
      throw new SQLException("SQLite database driver not found",exception);
    }
  }

  /** close database access
   */
  public void close()
    throws SQLException
  {
    try { connection.setAutoCommit(true); } catch (SQLException exception) { /* ignored */ }
    connection.close();
  }

  /** dicard database
   * @param baseDirectory base directory
   * @param name name
   */
  public static void discard(String baseDirectory, String name)
  {
    File file = new File(baseDirectory+File.separator+name+".db");
    file.delete();
  }

  public void setForeignKeys(boolean enabled)
    throws SQLException
  {
    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = connection.prepareStatement("PRAGMA foreign_keys="+(enabled ? "ON" : "OFF"));
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
    }
    finally
    {
      if (preparedStatement != null) preparedStatement.close();
    }
  }

  /** create new statement
   * @return statement
   */
  public Statement createStatement()
    throws SQLException
  {
    Statement statement = connection.createStatement();
    statement.setQueryTimeout(TIMEOUT);

    return statement;
  }

  /** create new prepared statement
   * @param sqlCommand SQL command
   * @return prepared statement
   */
  public PreparedStatement prepareStatement(String sqlCommand)
    throws SQLException
  {
    PreparedStatement preparedStatement = connection.prepareStatement(sqlCommand);
    preparedStatement.setQueryTimeout(TIMEOUT);

    return preparedStatement;
  }

  /** commit transaction
   */
  public void commit()
    throws SQLException
  {
    connection.commit();
  }

  /** get id of last inserted row
   * @return id or ID_NONE
   */
  public int getLastInsertId()
    throws SQLException
  {
    int id = ID_NONE;

    PreparedStatement preparedStatement;
    ResultSet         resultSet;

    // Note: Java sqlite does not support getGeneratedKeys, thus use last_insert_rowid() direct.
    preparedStatement = connection.prepareStatement("SELECT last_insert_rowid();");
    resultSet = null;
    try
    {
      resultSet = preparedStatement.executeQuery();
      if (!resultSet.next())
      {
        throw new SQLException("no result");
      }

      id = resultSet.getInt(1);

      resultSet.close(); resultSet = null;
    }
    finally
    {
      if (resultSet != null) resultSet.close();
    }

    return id;
  }

  /** create array from data
   * Note: not supported by SQLite driver
   * @param type type string
   * @param elements data
   * @return SQL array
   */
  public Array createArray(String type, Object[] elements)
    throws SQLException
  {
    return connection.createArrayOf(type,elements);
  }

  /** create array from data
   * Note: not supported by SQLite driver
   * @param type type string
   * @param collection data
   * @return SQL array
   */
  public Array createArray(String type, AbstractCollection collection)
    throws SQLException
  {
    return createArray(type,collection.toArray());
  }

  /** convert data to lines
   * @param data data (lines separated by \n) or null
   * @return lines
   */
  public static String[] dataToLines(String data)
  {
    ArrayList<String> lineList = new ArrayList<String>();

    if (data != null)
    {
      StringTokenizer stringTokenizer = new StringTokenizer(data,"\n");
      while (stringTokenizer.hasMoreTokens())
      {
        lineList.add(stringTokenizer.nextToken());
      }
    }

    return lineList.toArray(new String[lineList.size()]);
  }

  /** convert lines to data
   * @param lines lines
   * @return data (lines separated by \n)
   */
  public static String linesToData(String[] lines)
  {
    StringBuilder buffer = new StringBuilder();
    if (lines != null)
    {
      for (String line : lines)
      {
        buffer.append(line); buffer.append('\n');
      }
    }

    return buffer.toString();
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "Database {name: "+name+"}";
  }

  //-----------------------------------------------------------------------

}

/* end of file */
