;/***********************************************************************\
*
* $Revision: 895 $
* $Date: 2021-02-19 16:46:11 +0100 (Fri, 19 Feb 2021) $
* $Author: torsten $
* Contents: periodic task functions
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base

/****************************** Classes ********************************/

/** Synchronization functions
 */
public abstract class PeriodicTask
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------

  protected BackendProgress backendProgress;

  private   boolean         quitFlag;

  private   Thread          thread;
  private   Object          startTrigger = new Object();
  private   boolean         startTriggerFlag = false;
  private   Object          doneTrigger = new Object();

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create periodic task
   * @param name name
   * @param refreshTime refresh time [s]
   * @param stepCount number of progress steps or 0
   */
  PeriodicTask(String name, final long refreshTime, int stepCount)
  {
    // create backend progress
    backendProgress = new BackendProgress(stepCount)
    {
      protected void update(double n)
      {
        updateProgress(n);
      }
    };

    Runnable runnable = new Runnable()
    {
      public void run()
      {
        try
        {
          while (!quitFlag)
          {
            // sleep or wait for trigger
            synchronized(startTrigger)
            {
              if (!startTriggerFlag)
              {
                if (refreshTime > 0)
                {
                  startTrigger.wait(refreshTime*1000L);
                }
                else
                {
                  startTrigger.wait();
                }
              }
              else
              {
                startTriggerFlag = false;
              }
            }
            if (quitFlag) break;

            if (isRunnable())
            {
              // run task
              {
                Backend.updateLock();
                backendProgress.setStep(0);
                begin();
              }
              try
              {
                PeriodicTask.this.run();
              }
              finally
              {
                end();
                Backend.updateUnlock();
              }

              // signal done
              synchronized(doneTrigger)
              {
                doneTrigger.notifyAll();
              }
            }
            if (quitFlag) break;
          }
        }
        catch (InterruptedException exception)
        {
          quitFlag = true;
        }
        catch (AssertionError error)
        {
          Timemine.printInternalError(error,"assertion%s",(error.getMessage() != null) ? " "+error.getMessage() : "");
          System.exit(Timemine.EXITCODE_INTERNAL_ERROR);
        }
        catch (InternalError error)
        {
          Timemine.printInternalError(error);
          System.exit(Timemine.EXITCODE_INTERNAL_ERROR);
        }
        catch (Throwable throwable)
        {
          Timemine.printInternalError(throwable);
          System.exit(Timemine.EXITCODE_INTERNAL_ERROR);
        }
/*
        catch (Throwable throwable)
        {
          Timemine.printStackTrace(throwable);
          System.exit(Timemine.EXITCODE_INTERNAL_ERROR);
        }
*/
      }
    };

    thread = new Thread(runnable,name);
    thread.setDaemon(true);
  }

  /** create periodic task
   * @param name name
   * @param refreshTime refresh time [s]
   */
  PeriodicTask(String name, final long refreshTime)
  {
    this(name,refreshTime,0);
  }

  /** start periodic task
   */
  public void start()
  {
    thread.start();
  }

  /** set periodic task step count
   * @param stepCount step count
   */
  public void setStepCount(int stepCount)
  {
    backendProgress.setStepCount(stepCount);
  }

  /** trigger periodic task
   */
  public void trigger()
  {
    synchronized(this.startTrigger)
    {
      startTriggerFlag = true;
      startTrigger.notifyAll();
    }
  }

  /** force one run and wait until finished
   */
  public void forceRun()
  {
    synchronized(this.doneTrigger)
    {
      trigger();
      try { doneTrigger.wait(); } catch (InterruptedException exception) { /* ignored */ }
    }
  }

  /** wait for trigger or timeout
   * @param timeout timeout [ms]
   */
  public void waitTrigger(int timeout)
    throws InterruptedException
  {
    startTrigger.wait(timeout*1000L);
  }

  /** check if periodic task is interrupted
   * @return true iff interrupted
   */
  public boolean isInterrupted()
  {
    return thread.isInterrupted();
  }

  /** quit periodic task
   */
  public void quit()
  {
    quitFlag = true;
    try
    {
      thread.interrupt();
      thread.join();
    }
    catch (InterruptedException exception)
    {
      /* ignored */
    }
  }

  /** check if periodic task is runnable
   * @return true iff runnable
   */
  public boolean isRunnable()
    throws BackendException, InterruptedException
  {
    return true;
  }

  /** begin periodic task
   */
  public void begin()
    throws BackendException
  {
    // default nothing to do
  }

  /** end periodic task
   */
  public void end()
    throws BackendException
  {
    // default nothing to do
  }

  /** update progress
   * @param value progress value 0..1
   */
  public void updateProgress(double value)
  {
    // default nothing to do
  }

  /** run periodic task
   */
  public abstract void run()
    throws BackendException, InterruptedException;
}

/* end of file */
