/***********************************************************************\
*
* $Revision: 430 $
* $Date: 2016-12-12 11:33:02 +0100 (Mon, 12 Dec 2016) $
* $Author: torsten $
* Contents: backend
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.lang.ref.SoftReference;
import java.io.BufferedReader;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;

import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.ProtocolException;
import java.net.UnknownHostException;
import java.net.URL;

import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.Iterator;
import java.util.AbstractSet;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.TimeUnit;
import java.util.EnumSet;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

// SSL
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

// graphics
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;

// Apache commons
import org.apache.commons.codec.binary.Base64;

// SQLite database
import java.sql.Array;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;

/****************************** Classes ********************************/

/** project exception
 */
class BackendException extends Exception
{
  /** create backend exception
   * @param message message
   * @param cause exception cause
   */
  BackendException(String message, Exception cause)
  {
    super(message,cause);
  }

  /** create backend exception
   * @param message message
   * @param arguments optional arguments
   */
  BackendException(String message, Object... arguments)
  {
    this(String.format(message,arguments),(Exception)null);
  }

  /** create backend exception
   * @param cause exception cause
   */
  BackendException(Exception cause)
  {
    super(cause);
  }

  /** create backend exception
   */
  BackendException()
  {
    super((Exception)null);
  }
}

/** backend progress handler
 */
abstract class BackendProgress
{
  class StepLevel
  {
    double base;
    double size;
    int    step;
    int    count;
    double currentValue;

    /** create new step level
     * @param base base value
     * @param size size
     * @param count step count
     */
    StepLevel(double base, double size, int count)
    {
// TODO: activate
//      assert base >= 0.0 : String.format("%f >= 0.0",base);
      assert size >= 0.0 : String.format("%f >= 0.0",size);
      assert count > 0 : String.format("%d > 0.0",count);

      this.base         = base;
      this.size         = size;
      this.step         = 0;
      this.count        = count;
      this.currentValue = 0.0;
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      return "StepLevel { base="+base+", size="+size+", step="+step+", count="+count+", currentValue="+currentValue+" }";
    }
  }

  // none progress handler
  public final static BackendProgress NONE = new BackendProgress() { public void update(double n) {} };

  private Stack<StepLevel> stepLevels = new Stack<StepLevel>();
  private StepLevel        stepLevel  = new StepLevel(0.0,1.0,1);
  private boolean          abortFlag  = false;

  /** create backend progress
   * @param count step count
   */
  public BackendProgress(int count)
  {
    this.stepLevel.count = count;
  }

  /** create backend progress
   */
  public BackendProgress()
  {
  }

  /** begin new step level
   * @param count step count
   */
  public void beginSteps(int count)
  {
    assert count > 0;

    stepLevels.push(stepLevel);
    stepLevel = new StepLevel(stepLevel.base+(double)stepLevel.step/(double)stepLevel.count*stepLevel.size,
                              1.0/(double)stepLevel.count*stepLevel.size,
                              count
                             );
    setSelection(0);
  }

  /** end step level
   */
  public void endSteps()
  {
    stepLevel = stepLevels.pop();
    setSelection(1.0);
  }

  /** set step number
   * @param step step number
   */
  public void setStep(int step)
  {
    assert step >= 0 : String.format("%d >= 0");
    assert step <= stepLevel.count : String.format("%d <= %d",step,stepLevel.count);

    stepLevel.step = step;
    setSelection(0);
  }

  /** increment step number
   */
  public void incrementStep()
  {
    assert stepLevel.step <= stepLevel.count;

    setSelection(1.0);
    stepLevel.step++;
  }

  /** set total number of progress steps
   * @param stepCount step count (> 0)
   */
  public void setStepCount(int count)
  {
    assert count > 0;

    stepLevel.count = count;
    setSelection(0);
  }

  /** set progress update
   * @param n progress value [0..1]
   */
  public void setSelection(double n)
  {
    assert n >= 0.0;
    assert n <= 1.0;

    double value = stepLevel.base+
                   ((double)stepLevel.step/(double)stepLevel.count+
                    n/(double)stepLevel.count
                   )*stepLevel.size;
//Dprintf.dprintf("%f %f",(double)stepLevel.step/(double)stepLevel.count*stepLevel.size,n/(double)stepLevel.count*stepLevel.size);
//Dprintf.dprintf("[%f..%f]",stepLevel.base+(double)stepLevel.step/(double)stepLevel.count*stepLevel.size,stepLevel.base+((double)stepLevel.step/(double)stepLevel.count+1.0/(double)stepLevel.count)*stepLevel.size);
//Dprintf.dprintf("value=%f base=%f size=%f step=%d count=%d",value,stepLevel.base,stepLevel.size,stepLevel.step,stepLevel.count);
// TODO: activate
//    assert value >= stepLevel.base+(double)stepLevel.step/(double)stepLevel.count*stepLevel.size;
//    assert value <= stepLevel.base+((double)stepLevel.step/(double)stepLevel.count+1.0/(double)stepLevel.count)*stepLevel.size;
    if      (value > stepLevel.currentValue)
    {
      update(value);
      stepLevel.currentValue = value;
    }
/*
Note: JIRA is not able to return a correct total value of found entries. Thus the
progress value cannot be calculated reliable and may jump "back".

    else if (value < stepLevel.currentValue)
    {
for (StepLevel l : stepLevels) Dprintf.dprintf("l=%s",l);
Dprintf.dprintf("++++ ups value=%f tepLevel=%s",value,stepLevel);
Dprintf.halt();
    }
*/
  }

  /** set progress update
   * @param i current value
   * @param n total value
   */
  public void setSelection(long i, long n)
  {
    assert i >= 0;
    assert n > 0;
    assert i <= n;

    setSelection((double)i/(double)n);
  }

  /** abort progress
   */
  public void abort()
  {
     abortFlag = true;
  }

  /** check if aborted
   * @return true iff aborted
   */
  public boolean isAborted()
  {
     return abortFlag;
  }

  /** update progress
   * @param n progress value
   */
  protected abstract void update(double n);
}

/** database id
 */
class Id implements Cloneable, Comparable<Id>
{
// TODO: make private, create Id object with ANY
  public final static int NONE = 0;
  public final static int ANY  = -1;

  public int local;                      // local database id
  public int backend;                    // backend id
  public int modified;                   // modified database id

//  private boolean isNone,isAny;

  /** create id
   * @param local local database id
   * @param backend backend id
   * @param modified modified database id
   */
  Id(int local, int backend, int modified)
  {
// TODO: activate
//    assert (local != Backend.ID_NONE) || (backend != Backend.ID_NONE) || (modified != Backend.ID_NONE);

    this.local    = local;
    this.backend  = backend;
    this.modified = modified;
    //this.isNone   = false;
    //this.isAny    = false;
  }

  /** create id
   * @param local local id
   * @param backend backend id
   */
  Id(int local, int backend)
  {
    this(local,backend,NONE);
  }

  /** create id
   * @param backend backend id
   */
  Id(int backend)
  {
    this(NONE,backend);
  }

  /** create id
   */
  Id()
  {
    this(NONE);
  }

  /** clone id
   * @return cloned id
   */
  public Id clone()
  {
    return new Id(local,backend);
  }

  /** compare with other id
   * @return -1/0+1 iff lower/equals/greater
   */
  public int compareTo(Id other)
  {
/*
    if (!hasLocal())
    {
      if (!other.hasLocal())
        return  0;
      else
        return  1;
    }
    if (!other.hasLocal())
    {
      if (!hasLocal())
        return  0;
      else
        return -1;
    }

    if      (local < other.local)
      return -1;
    else if (local > other.local)
      return  1;

    if      (backend < other.backend)
      return -1;
    else if (backend > other.backend)
      return  1;

    return  0;
*/
    if      (local < other.local)
      return -1;
    else if (local > other.local)
      return  1;
    else
      return  0;
  }

  /** check if id is set
   * @return true iff set
   */
  public boolean isNone()
  {
//    return isNone;
return backend == Id.NONE;
  }

  /** check if id is set
   * @return true iff set
   */
  public boolean isAny()
  {
//    return isAny;
return backend == Id.ANY;
  }

  /** check if id is set
   * @return true iff set
   */
  boolean isSet()
  {
    return !isNone() && !isAny();
  }

  /** check if local id is set
   * @return true iff set
   */
  public boolean hasLocal()
  {
    return (local != NONE);
  }

  /** check if backend id is set
   * @return true iff set
   */
  public boolean hasBackend()
  {
    return (backend != NONE);
  }

  /** force always compare id content e. g. in hash maps
   * @return 0
   */
  public int hashCode()
  {
    return 0;
  }

  /** compare with other id
   * @return true if equals
   */
  public boolean equals(Object otherId)
  {
    return (backend == ((Id)otherId).backend);
  }

  /** format id
   * @return string
   */
  public String format()
  {
    return String.format("#%d",backend);
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    if      (isNone())
    {
      return "Id { local="+local+", backend=none"+((modified != NONE) ? ", modified="+modified:"")+" }";
    }
    else if (isAny())
    {
      return "Id { local="+local+", backend=any"+((modified != NONE) ? ", modified="+modified:"")+" }";
    }
    else
    {
      return "Id { local="+local+", backend="+backend+((modified != NONE) ? ", modified="+modified:"")+" }";
    }
  }

  /** create NONE, ANY id
   * @param isNone true iff none
   * @param isAny true iff any
   */
  private Id(boolean isNone, boolean isAny)
  {
    assert (isNone && !isAny) || (!isNone && isAny);

    this.local    = NONE;
    this.backend  = NONE;
    this.modified = NONE;
//    this.isNone   = isNone;
  //  this.isAny    = isAny;
  }
}

/** database id list
 */
class IdList extends ArrayList<Id>
{
  final static IdList EMPTY = new IdList();
}

/** database handler
 */
abstract class DataBaseHandler<T>
{
  /** query result set (Note: use that instead of plain ResultSet to parse special data types)
   */
  static class ValueSet
  {
    private ResultSet resultSet;

    /** create value set
     * @param resultSet result set
     */
    protected ValueSet(ResultSet resultSet)
    {
      this.resultSet = resultSet;
    }

    /** done value set
     */
    public void close()
      throws SQLException
    {
      resultSet.close();
    }

    /** get boolean value
     * @param name name
     * @return value
     */
    public boolean getBoolean(String name)
      throws SQLException
    {
      return resultSet.getBoolean(name);
    }

    /** get int value
     * @param name name
     * @return value
     */
    public int getInt(String name)
      throws SQLException
    {
      return resultSet.getInt(name);
    }

    /** get long value
     * @param name name
     * @return value
     */
    public long getLong(String name)
      throws SQLException
    {
      return resultSet.getLong(name);
    }

    /** get string value
     * @param name name
     * @return value
     */
    public String getString(String name)
      throws SQLException
    {
      return resultSet.getString(name);
    }

    /** get enum value
     * @param name name
     * @param enumType enum type class
     * @return value
     */
    public <T extends Enum<T>> T getEnum(String name, Class<T> enumType)
      throws SQLException
    {
      int n        = resultSet.getInt(name);
      T   values[] = enumType.getEnumConstants();

      for (T value : enumType.getEnumConstants())
      {
        if (value.ordinal() == n) return value;
      }

      return null;
    }

    /** get date value
     * @param name name
     * @return value
     */
    public Date getDate(String name)
      throws SQLException
    {
      return new Date(resultSet.getLong(name)*1000L);
    }

    /** get spent-on value
     * @param local local name
     * @param backend backend name
     * @param modified modified name
     * @return value
     */
    public Id getId(String local, String backend, String modified)
      throws SQLException
    {
      return new Id(resultSet.getInt(local),
                    resultSet.getInt(backend),
                    resultSet.getInt(modified)
                   );
    }

    /** get spent-on value
     * @param local local name
     * @param backend backend name
     * @return value
     */
    public Id getId(String local, String backend)
      throws SQLException
    {
      return new Id(resultSet.getInt(local),
                    resultSet.getInt(backend)
                   );
    }

    /** get spent-on value
     * @param backend backend name
     * @return value
     */
    public Id getId(String backend)
      throws SQLException
    {
      return new Id(resultSet.getInt(backend));
    }

    /** get spent-on value
     * @param name name
     * @return value
     */
    public SpentOn getSpentOn(String name)
      throws SQLException
    {
      return new SpentOn(new Date(resultSet.getLong(name)*1000L));
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      try
      {
        StringBuilder buffer = new StringBuilder();

        ResultSetMetaData resultSetMetaData = resultSet.getMetaData();
        for (int i = 1; i < resultSetMetaData.getColumnCount(); i++)
        {
          if (i > 1) buffer.append(',');
          buffer.append(String.format("%s=%s",resultSetMetaData.getColumnName(i),resultSet.getString(i).toString()));
        }

        return buffer.toString();
      }
      catch (SQLException exception)
      {
        return resultSet.toString();
      }
    }
  };

  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create database handler
   */
  DataBaseHandler()
  {
  }

  /** get select statement
   * @param additionalFilter additional filter
   * @return select statement
   */
  abstract public String selectStatement(String additionalFilter);

  /** process value set
   * @param valueSet value set
   */
  public void process(ValueSet valueSet)
    throws SQLException
  {
  }

  /** get database result
   * @param valueSet value set
   * @return data object
   */
  public T get(ValueSet valueSet)
    throws SQLException
  {
    return null;
  }

  /** process data
   * @param data data
   */
  public void process(T data)
    throws SQLException
  {
  }

  // ----------------------------------------------------------------------
}

/** base entity
 * Note: comparable without type
 */
abstract class Entity implements Comparable
{
  public final Backend.Types backendType;
  public Id                  id;             // database ids

  /** create entity
   * @param backendType backend type
   * @param id id
   */
  Entity(Backend.Types backendType, Id id)
  {
    this.backendType = backendType;
    this.id          = id;
  }

  /** compare with other id
   * @param otherEntity other entity
   * @return true if equals
   */
  public boolean equals(Entity otherEntity)
  {
    return compareTo(otherEntity) == 0;
  }

  /** compare with other object
   * @return -1/0+1 iff lower/equals/greater
   */
  public int compareTo(Object other)
  {
    int result = 0;

    if (   (other != null)
        && Entity.class.isAssignableFrom(other.getClass())
       )
    {
      Entity otherEntity = (Entity)other;
      if (this != otherEntity)
      {
        if (backendType != otherEntity.backendType)
        {
          result = -1;
        }
        else
        {
          result = id.compareTo(otherEntity.id);
        }
      }
    }
    else
    {
      result = -1;
    }

    return result;
  }

  /** get id
   * @return id string
   */
  String getId()
  {
    return Integer.toString(id.backend);
  }

  /** set modified state
   * @parm modifiedState modified state
   */
  protected void setModified(Backend.ModifiedStates modifiedState)
    throws BackendException, InterruptedException
  {
    Backend.triggerModifyListener(this,modifiedState);
  }
}

/** entity handler
 */
abstract class EntityHandler<T>
{
  /** handle entity
   * @param entity entity
   */
  abstract void handle(T entity)
    throws BackendException;

  void done(T entity)
    throws BackendException
  {
    // nothing to do
  }
}

/** backend functions
 */
public class Backend
{
  /** backend types
   */
  static enum Types
  {
    NONE,
    REDMINE,
    JIRA;

   /** get backend name
     * @return name
     */
    String getName()
    {
      String name = null;
      switch (this)
      {
       case NONE:    name = "NONE";    break;
       case REDMINE: name = "REDMINE"; break;
       case JIRA:    name = "JIRA";    break;
      }

      return name;
    }

    /** parse type
     * @param string string
     * @param defaultValue default value
     * @return type
     */
    static Types toEnum(String string, Types defaultValue)
    {
      if      (string.equalsIgnoreCase("REDMINE")) return REDMINE;
      else if (string.equalsIgnoreCase("JIRA"   )) return JIRA;
      else                                         return defaultValue;
    }

    /** parse type
     * @param string string
     * @return type
     */
    static Types toEnum(String string)
    {
      return toEnum(string,NONE);
    }

    /** convert ordinal value to enum
     * @param value ordinal value
     * @param defaultValue default value
     * @return enum
     */
    static Types toEnum(int value, Types defaultValue)
    {
      if      (Backend.Types.REDMINE.ordinal() == value) return Types.REDMINE;
      else if (Backend.Types.JIRA.ordinal()    == value) return Types.JIRA;
      else                                               return defaultValue;
    }

    /** convert ordinal value to enum
     * @param value ordinal value
     * @return enum or Types.NONE
     */
    static Types toEnum(int value)
    {
      return toEnum(value,NONE);
    }

    /** get logo
     * @return backend type logo image
     */
    public Image getImage()
    {
      return Backend.backendDrivers.get(this).logo;
    }

    /** convert from integer
     * @return string
     */
    public String toString()
    {
      switch (this)
      {
        case NONE:    return "none";
        case REDMINE: return "Redmine";
        case JIRA:    return "JIRA";
      }

      return "unknown";
    }
  };

  static class TypeSet extends HashSet<Types>
  {
    static final TypeSet NONE    = new TypeSet();
    static final TypeSet REDMINE = new TypeSet(Types.REDMINE);
    static final TypeSet JIRA    = new TypeSet(Types.JIRA);
    static final TypeSet ALL     = new TypeSet(Types.REDMINE,Types.JIRA);

    /** create type set
     * @param types types to add to set
     */
    TypeSet(Types... types)
    {
      for (Types type : types)
      {
        add(type);
      }
    }

    /** get ordinal string
     * @return ordinal string
     */
/*
    public Array ordinal()
    {
      Array array = null;

      try
      {
      Integer data[] = new Integer[size()];
      int i = 0;
      for (Types type : this)
      {
        data[i] = type.ordinal();
        i++;
      }

      array = database.createArray("integer",data);

    }
    catch (SQLException exception)
    {
      Dprintf.dprintf("_"); System.exit(1);
    }
    return array;
  }
*/
    public String ordinal()
    {

      StringBuilder buffer = new StringBuilder();

      if (contains(Types.REDMINE)) { if (buffer.length() > 0) buffer.append(','); buffer.append(Types.REDMINE.ordinal()); }
      if (contains(Types.JIRA   )) { if (buffer.length() > 0) buffer.append(','); buffer.append(Types.JIRA.ordinal()   ); }

      return buffer.toString();
    }
  }

  /** SQL form
   */
  private static abstract class Form
  {
    String format;
    Object arguments[];

    /** create form
     * @param format format string
     * @param arguments optional arguments
     */
    Form(String format, Object... arguments)
    {
      this.format    = format;
      this.arguments = arguments;
    }
  }

  /** SQL condition
   */
  private static class Condition extends Form
  {
    final static Condition NONE = null;

    /** create conditional
     * @param format format string
     * @param arguments optional arguments
     */
    Condition(String format, Object... arguments)
    {
      super(format,arguments);
    }
  }

  /** create condition
   * @param format format string
   * @param arguments optional arguments
   */
  protected static Condition condition(String format, Object... arguments)
  {
    return new Condition(format,arguments);
  }

  /** SQL filter
   */
  private static class Filter extends Form
  {
    final static Filter NONE = null;

    /** create filter
     * @param format format string
     * @param arguments optional arguments
     */
    Filter(String format, Object... arguments)
    {
      super(format,arguments);
    }
  }

  /** create filter
   * @param format format string
   * @param arguments optional arguments
   */
  protected static Filter filter(String format, Object... arguments)
  {
    return new Filter(format,arguments);
  }

  /** SQL values
   */
  protected static class Values
  {
    final static Values NONE = null;

    Object[] data;

    /** create values
     * @param data data (tuples name, value)
     */
    Values(Object... data)
    {
      assert data != null;
      assert (data.length % 2) == 0;

      this.data = data;
    }
  }

  /** create values
   * @param data data (tuples name, value)
   */
  protected static Values values(Object... data)
  {
    return new Values(data);
  }

  /** SQL values
   */
  protected static class Names
  {
    String[] data;

    /** create names
     * @param data names (tuples name, name)
     */
    Names(String... data)
    {
      assert data != null;
      assert (data.length % 2) == 0;

      this.data = data;
    }
  }

  /** create names
   * @param data names (tuples name, name)
   */
  protected static String[] names(String... data)
  {
    return data;
  }

  // modified state
  enum ModifiedStates
  {
    NONE,
    ADDED,
    UPDATED,
    DELETED;

    /** convert ordinal value to enum
     * @param value ordinal value
     * @param defaultValue default value
     * @return enum
     */
    static ModifiedStates toEnum(int value, ModifiedStates defaultValue)
    {
      switch (value)
      {
        case 0:  return NONE;
        case 1:  return ADDED;
        case 2:  return UPDATED;
        case 3:  return DELETED;
        default: return defaultValue;
      }
    }

    /** convert ordinal value to enum
     * @param value ordinal value
     * @return enum
     */
    static ModifiedStates toEnum(int value)
    {
      return toEnum(value,NONE);
    }

    /** get value
     * @return value
     */
    public int getValue()
    {
      return ordinal();
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      switch (this)
      {
        case NONE:     return "NONE";
        case ADDED:    return "ADDED";
        case UPDATED:  return "UPDATED";
        case DELETED:  return "DELETED";
      }

      return "";
    }
  }

  /** history filter types
   */
  static enum HistoryFilterTypes
  {
    ISSUES,
    COMMENTS
  };

  // --------------------------- constants --------------------------------
//TODO: verify host/certificate
  // accept any host name
  protected final static HostnameVerifier ANY_HOSTNAME_VERIFIER = new HostnameVerifier()
  {
    public boolean verify(String hostname, SSLSession session)
    {
      return true;
    }
  };;

// TODO: move to condition, duration
  public final static String             CONDITION_NONE  = (String)null;
  public final static double             DURATION_NONE   = 0L;
  public final static double             DURATION_UPDATE = -1L;

  /** online/offline state
   */
  public enum ConnectionStates
  {
    ONLINE,
    OFFLINE
  };

  /** online/offline state listener
   */
  static abstract class StateListener<T>
  {
    /** create state listener
     * @param userData user data
     */
    StateListener(T userData)
    {
      this.userData = userData;
    }

    /** create state listener
     */
    StateListener()
    {
    }

    /** changed callback
     * @param state changed state
     */
    abstract public void changed(ConnectionStates state);

    protected T userData;
  }

  /** modify listener
   */
  static abstract class ModifyListener<T>
  {
    /** create modify listener
     * @param userData user data
     */
    ModifyListener(T userData)
    {
      this.userData = userData;
    }

    /** create modify listener
     */
    ModifyListener()
    {
    }

    /** modified callback
     * @param entity modified entity
     * @param modifyState modified state
     */
    abstract public void modified(Entity entity, ModifiedStates modifyState);

    protected T userData;
  }

  protected final int                    READ_TIMEOUT        = 60*1000;  // read timeout [ms]

  private final static String            DATABASE_NAME       = "database";

  private final static SimpleDateFormat  DATE_FORMAT         = new SimpleDateFormat("yyyy-MM-dd");
  private final static SimpleDateFormat  DATETIME_FORMAT     = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
  private final static SimpleDateFormat  ISO_DATETIME_FORMAT = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");

  private final SoftReference<TimeEntry> TIME_ENTRY_NULL = new SoftReference<TimeEntry>(null);

  // --------------------------- variables --------------------------------
  protected static Database                          database;

// TODO: protected
  public static HashMap<Backend.Types,BackendDriver> backendDrivers = new HashMap<Backend.Types,BackendDriver>();

  private static ConnectionStates                    connectionState = ConnectionStates.OFFLINE;
  private static ArrayList<StateListener>            stateListeners = new ArrayList<StateListener>();
  private static ArrayList<ModifyListener>           modifyListeners = new ArrayList<ModifyListener>();
  private static int                                 maxUpdateTime;
  private static ReentrantLock                       modifyLock = new ReentrantLock();
  private static ReentrantLock                       updateLock = new ReentrantLock();

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  static
  {
    final BackendDriver BACKEND_DRIVER_NONE = new BackendDriver()
    {
      public Backend.Types getType()
      {
        return Backend.Types.NONE;
      }

      public Issue loadIssue(Id issueId)
        throws BackendException
      {
        // nothing to do
        return Issue.NONE;
      }

      public void storeIssue(final Issue issue)
        throws BackendException
      {
        // nothing to do
      }

      public void updateIssue(final Issue issue, final Issue originIssue)
        throws BackendException
      {
        // nothing to do
      }

      public void deleteIssue(final Issue issue)
        throws BackendException
      {
        // nothing to do
      }

      public TimeEntry loadTimeEntry(Id projectId, Id issueId, Id timeEntryId)
        throws BackendException
      {
        // nothing to do
        return null;
      }

      public void storeTimeEntry(final TimeEntry timeEntry)
        throws BackendException
      {
        // nothing to do
      }

      public void updateTimeEntry(final TimeEntry timeEntry, final TimeEntry originTimeEntry)
        throws BackendException
      {
        // nothing to do
      }

      public void deleteTimeEntry(final TimeEntry timeEntry)
        throws BackendException
      {
        // nothing to do
      }

      public void getUsers(final BackendProgress backendProgress, final EntityHandler<User> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public User getCurrentUser(final BackendProgress backendProgress, final EntityHandler<User> entityHandler)
        throws BackendException
      {
        return null;
      }

      public void getRoles(final BackendProgress backendProgress, final EntityHandler<Role> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public void getTrackers(final BackendProgress backendProgres, final EntityHandler<Tracker> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public void getEpicNames(final BackendProgress backendProgres, final EntityHandler<EpicName> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public void getStatus(final BackendProgress backendProgress, final EntityHandler<Status> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public void getPriorities(final BackendProgress backendProgress, final EntityHandler<Priority> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public void getActivities(final BackendProgress backendProgress, final EntityHandler<Activity> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public void getProjects(final BackendProgress backendProgress, final EntityHandler<Project> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public void getMembers(final BackendProgress backendProgress, final EntityHandler<Member> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public Member getCurrentMember(final BackendProgress backendProgress, final Id projectId, final EntityHandler<Member> entityHandler)
        throws BackendException
      {
        return null;
      }

      public void getVersions(final BackendProgress backendProgress, final EntityHandler<Version> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public void getIssues(final BackendProgress backendProgress, long lastUpdatedOnTimestamp, final EntityHandler<Issue> entityHandler)
        throws BackendException
      {
        // nothing to do
      }

      public TimeEntry getTimeEntry(final BackendProgress backendProgress, final long createdOnTimestamp, final EntityHandler<TimeEntry> entityHandler)
        throws BackendException, InterruptedException
      {
        return null;
      }

      public void getTimeEntries(final BackendProgress backendProgress, final long createdOnTimestamp, final EntityHandler<TimeEntry> entityHandler)
        throws BackendException
      {
        // nothing to do
      }
    };

    // default backend drivers
    backendDrivers.put(Backend.Types.NONE,   BACKEND_DRIVER_NONE);
    backendDrivers.put(Backend.Types.REDMINE,BACKEND_DRIVER_NONE);
    backendDrivers.put(Backend.Types.JIRA,   BACKEND_DRIVER_NONE);
  }

  /** initialize backend
   * @param state state
   * @param maxUpdateTime max. time to force an update [min]
   */
  public static void init(ConnectionStates connectionState,
                          int             maxUpdateTime
                         )
    throws BackendException
  {
    // initialize variables
    Backend.connectionState = ConnectionStates.OFFLINE;
    Backend.maxUpdateTime   = maxUpdateTime;

    // open local database
    PreparedStatement preparedStatement = null;
    try
    {
      // init local database
      database = new Database(Settings.TIMEMINE_DIRECTORY,DATABASE_NAME);
      initLocalDatabaseTables();
      initLocalDatabaseViews();

// Note: may be called in constructor where private member variables are still not initialized
//if (projectMap != null) projectMap.clear();
//if (membersMap != null) membersMap.clear();
//if (membersMap != null) timeEntryIdMap.clear();
//if (timeEntrySpentOnMap != null) timeEntrySpentOnMap.clear();

      // disable synchronous mode, enable journal memory
      preparedStatement = database.prepareStatement("PRAGMA synchronous=OFF;");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("PRAGMA journal_mode=MEMORY;");
      ResultSet resultSet = preparedStatement.executeQuery();
      resultSet.close();
      preparedStatement.close(); preparedStatement = null;
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }

    // add shutdown hook for local database
    Runtime.getRuntime().addShutdownHook(new Thread()
    {
      public void run()
      {
        try
        {
          Backend.database.close();
        }
        catch (SQLException exception)
        {
          // ignored
          Timemine.printInternalError(exception);
        }
      }
    });

    // install trust-manager which accept all certificates
    try
    {
      SSLContext sslContext = SSLContext.getInstance("TLS");
      sslContext.init(new KeyManager[0],new TrustManager[]{new AnyTrustManager()},new SecureRandom());
      SSLContext.setDefault(sslContext);
    }
    catch (KeyManagementException exception)
    {
      throw new BackendException("SSL fail: "+exception.getMessage());
    }
    catch (NoSuchAlgorithmException exception)
    {
      throw new BackendException("SSL fail: "+exception.getMessage());
    }

    // set connection state
    setConnectionState(connectionState);
  }

  /** init backend driver
   * @param backendType backend type
   * @param serverName server name
   * @param serverPort server port
   * @param serverUseSSL true iff server use SSL
   * @param loginName login name
   * @param loginPasssword login password
   * @param certificateType certificate type
   * @param certificateFileName certificate file (.cert, can be null)
   * @param keyFileName key file (.pem, can be null)
   * @param p12FileName p12 file name (.p12, can be null)
   * @param keyPassword key password
   */
  public static void initDriver(Display                   display,
                                Backend.Types              backendType,
                                String                    serverName,
                                int                       serverPort,
                                boolean                   serverUseSSL,
                                final String              loginName,
                                String                    loginPassword,
                                Settings.CertificateTypes certificateType,
                                String                    certificateFileName,
                                String                    keyFileName,
                                String                    p12FileName,
                                String                    keyPassword
                               )
    throws BackendException
  {
    BackendDriver backendDriver = null;

    switch (backendType)
    {
      case REDMINE:
        backendDriver = new BackendDriverRedmine(serverName,
                                                 serverPort,
                                                 serverUseSSL,
                                                 loginName,
                                                 loginPassword,
                                                 certificateType,
                                                 certificateFileName,
                                                 keyFileName,
                                                 p12FileName,
                                                 keyPassword,
                                                 Widgets.loadImage(display,"redmine.png")
                                                );
        break;
      case JIRA:
        backendDriver = new BackendDriverJIRA(serverName,
                                              serverPort,
                                              serverUseSSL,
                                              loginName,
                                              loginPassword,
                                              certificateType,
                                              certificateFileName,
                                              keyFileName,
                                              p12FileName,
                                              keyPassword,
                                              Widgets.loadImage(display,"jira.png")
                                             );
        break;
    }

    // check if login OK
    if (backendDriver.getCurrentUser(null) == null)
    {
      throw new BackendException("login fail");
    }

    backendDrivers.put(backendType,backendDriver);
  }

  /** done backend drivers
   */
  public static void doneDrivers()
  {
    for (Backend.Types backendType : Backend.Types.values())
    {
      if (backendType != Backend.Types.NONE)
      {
        backendDrivers.remove(backendType);
      }
    }
  }

  /** re-init local database (clear)
   * @param database database
   */
  public final static void reinitLocalDatabase()
    throws IOException, SQLException, InterruptedException
  {
    {
      updateLock();
    }
    try
    {
      // close databse
      database.close();
      database = null;

      // discard database
      Database.discard(Settings.TIMEMINE_DIRECTORY,DATABASE_NAME);

      // create new database
      database = new Database(Settings.TIMEMINE_DIRECTORY,DATABASE_NAME);
      initLocalDatabaseTables();

      // notice backend drivers about reinit
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        backendDriver.reinit();
      }
    }
    finally
    {
      updateUnlock();
    }
  }

  /** check iff epic name is supported for trackers
   * @param backendType backend type
   * @param issueId issue id
   * @return true iff epic name is supported for trackers
   */
  public static boolean isEpicNameSupported(Backend.Types backendType, Id issueId)
  {
    return backendDrivers.get(backendType).isEpicNameSupported();
  }

  /** check iff activity in time entries is supported
   * @param backendType backend type
   * @param projectId project id
   * @return true iff activity in time entries is supported
   */
  public static boolean isActivitySupported(Backend.Types backendType, Id projectId)
  {
//TODO: replace isSet-test by non-backend
    return !projectId.isNone() && backendDrivers.get(backendType).isActivitySupported();
  }

  /** check iff edit progress on issues is supported
   * @param backendType backend type
   * @param projectId project id
   * @return true iff setting progress on issues is supported
   */
  public static boolean isEditProgressSupported(Backend.Types backendType, Id projectId)
  {
//TODO: replace isSet-test by non-backend
    return !projectId.isNone() && backendDrivers.get(backendType).isEditProgressSupported();
  }

  /** Set online connection state
   */
  public static void online()
  {
    setConnectionState(ConnectionStates.ONLINE);
  }

  /** Set offline connection state
   */
  public static void offline()
  {
    setConnectionState(ConnectionStates.OFFLINE);
  }

  /** Set set
   * @param connectionState new connection state
   */
  public static void setConnectionState(ConnectionStates connectionState)
  {
    Backend.connectionState = connectionState;
    for (StateListener stateListener : stateListeners)
    {
      stateListener.changed(connectionState);
    }
  }

  /** add online/offline connection state listener
   * @param stateListener state listern to call on change
   */
  public static void addConnectionStatesListener(StateListener stateListener)
  {
    stateListeners.add(stateListener);
    stateListener.changed(connectionState);
  }

  /** add modify connection state listener
   * @param stateListener state listern to call on modify
   */
  public static void addModifyListener(ModifyListener modifyListener)
  {
    modifyListeners.add(modifyListener);
  }

  /** trigger modify listeners
   * @param entity modifed entity
   * @param modifyState modify state
   */
  public static void triggerModifyListener(Entity entity, ModifiedStates modifyState)
  {
     for (ModifyListener modifyListener : modifyListeners)
     {
       modifyListener.modified(entity,modifyState);
     }
  }

  /** check iff online
   * @return true iff online
   */
  public static boolean isOnline()
  {
    return connectionState == ConnectionStates.ONLINE;
  }

  /** check iff offline
   * @return true iff offline
   */
  public static boolean isOffline()
  {
    return connectionState == ConnectionStates.OFFLINE;
  }

  /** modify lock
   */
  public static void modifyLock()
    throws InterruptedException
  {
    modifyLock.lockInterruptibly();
  }

  /** modify try lock
   * @return true iff locked
   */
  public static boolean modifyTryLock(int timeout)
  {
    try
    {
      return modifyLock.tryLock(timeout,TimeUnit.MILLISECONDS);
    }
    catch (InterruptedException exception)
    {
      return false;
    }
  }

  /** modify try lock
   * @return true iff locked
   */
  public static boolean modifyTryLock()
  {
    return modifyLock.tryLock();
  }

  /** modify unlock
   */
  public static void modifyUnlock()
  {
    modifyLock.unlock();
  }

  /** update lock
   */
  public static void updateLock()
    throws InterruptedException
  {
    updateLock.lockInterruptibly();
  }

  /** update try lock
   * @return true iff locked
   */
  public static boolean updateTryLock()
  {
    return updateLock.tryLock();
  }

  /** update unlock
   */
  public static void updateUnlock()
  {
    updateLock.unlock();
  }

  /** check if users available
//TODO: bug in Redmine: non-admin user cannot get list of users
//      http://www.redmine.org/issues/7773
//      It seems the Redmine team will not fix this issue :-(

   * @return true iff users available
   */
  public static boolean hasUsers()
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("users");

    // check if users available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(new UserDatabaseHandler()
                     {
                       @Override
                       public void process(User user)
                       {
                         result[0] = true;
                       }
                     },
                     Condition.NONE,  // condition
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** get sorted user list
//TODO: bug in Redmine: non-admin user cannot get list of users
//      http://www.redmine.org/issues/7773
//      It seems the Redmine team will not fix this issue :-(

   * @return user list
   */
  public static UserList getUserList()
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("users");

    // get users from local database
    final UserList userList = new UserList();
    getLocalDatabase(new UserDatabaseHandler()
                     {
                       @Override
                       public void process(User user)
                       {
                         userList.add(user);
                       }
                     },
                     Condition.NONE,  // condition
                     filter("ORDER BY lastName,firstName,mail COLLATE NOCASE ASC")
                    );

    return userList;
  }

  /** get user
   * @param typeSet backend type set
   * @param userId user id
   * @return user or null
   */
  public static User getUser(Backend.TypeSet typeSet, Id userId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("users");

    // get user from local database
    return getLocalDatabase(new UserDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?)",
                                      userId.backend
                                     ),
                            Filter.NONE
                           );
  }

  /** get user
   * @param name user name
   * @return user or null
   */
  protected User getUser(Backend.TypeSet typeSet, String name)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("users");

    // get user id from local database
    return getLocalDatabase(new UserDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (name=?)",
                                      name
                                     ),
                            Filter.NONE
                           );
  }

  /** get user
   * @param name user name
   * @return user id
   */
  protected Id getUserId(String name)
    throws BackendException
  {
    User user = getUser(TypeSet.ALL,name);

    return (user != null) ? user.id : new Id();
  }

  /** check if roles available
   * @param typeSet backend type set
   * @return true iff roles available
   */
  public static boolean hasRoles(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("roles");

    // check if roles available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(new RoleDatabaseHandler()
                     {
                       @Override
                       public void process(Role role)
                       {
                         result[0] = true;
                       }
                     },
                     condition("backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** check if roles available
   * @return true iff roles available
   */
  public static boolean hasRoles()
    throws BackendException
  {
    return hasRoles(Backend.TypeSet.ALL);
  }

  /** get sorted role list
   * @param typeSet backend type set
   * @param projectId project id or Id.NONE
   * @return role list
   */
  public static RoleList getRoleList(Backend.TypeSet typeSet, Id projectId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("roles");

    // get roles from local database
    final RoleList roleList = new RoleList();
    getLocalDatabase(new RoleDatabaseHandler()
                     {
                       @Override
                       public void process(Role role)
                       {
                         roleList.add(role);
                       }
                     },
                     condition("(backendType IN ("+typeSet.ordinal()+")) AND (? OR (projectId=?))",
                               projectId.isAny() ? 1 : 0,
                               projectId.backend
                              ),
                     filter("ORDER BY name COLLATE NOCASE ASC")
                    );

    return roleList;
  }

  /** get sorted role list
   * @return role list
   */
  public static RoleList getRoleList()
    throws BackendException
  {
    return getRoleList(Backend.TypeSet.ALL,new Id(Id.ANY));
  }

  /** get role
   * @param typeSet backend type set
   * @param roleId role id
   * @return role or null
   */
  public final Role getRole(TypeSet typeSet, Id roleId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("roles");

    // get role from local database
    return getLocalDatabase(new RoleDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?)",
                                      roleId.backend
                                     ),
                            Filter.NONE
                           );
  }

  /** get role id list
   * @param roleIdList role id list
   * @param typeSet backend type set
   * @param projectId project id
   * @param names role names
   */
  public final static void getRoleIds(final IdList roleIdList, Backend.TypeSet typeSet, Id projectId, final Set<String> names)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("roles");

    // get role ids from local database
    getLocalDatabase(new RoleDatabaseHandler()
                     {
                       @Override
                       public void process(Role role)
                       {
                         if (names.contains(role.name))
                         {
                           roleIdList.add(role.id);
                         }
                       }
                     },
//TODO: filter by projectId?
//                       "? OR ((backendType=?) AND (? OR (projectId=?)))",
                     condition("? OR (backendType IN ("+typeSet.ordinal()+"))",
                               projectId.isAny() ? 1 : 0
                         ),
                     filter("ORDER BY name COLLATE NOCASE ASC")
                    );
  }

  /** get sortted list with role names
   * @param typeSet backend type set
   * @param projectId project id
   * @return sort role names list
   */
  public final static List<String> getRoleNames(Backend.TypeSet typeSet, Id projectId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("roles");

    // get role ids from local database
    final ArrayList<String> roleNames = new ArrayList<String>();
    getLocalDatabase(new RoleDatabaseHandler()
                     {
                       @Override
                       public void process(Role role)
                       {
                         roleNames.add(role.name);
                       }
                     },
                     condition("? OR (backendType=?)",
                               projectId.isAny() ? 1 : 0,
// TODO:
Backend.Types.JIRA.ordinal()//                               backendType.ordinal()
                              ),
                     filter("ORDER BY name COLLATE NOCASE ASC")
                    );

    return roleNames;
  }

  /** get sorted list with role names
   * @return sort role names list
   */
  public final static List<String> getRoleNames()
    throws BackendException
  {
    return getRoleNames(Backend.TypeSet.ALL,new Id(Id.ANY));
  }

  /** get role name
   * @param backendType backend type
   * @param roleId role id
   * @param defaultName default name
   * @return role name
   */
  public final String getRoleName(Backend.Types backendType, Id roleId, String defaultName)
  {
    String roleName = defaultName;

    try
    {
//// TODO:      Role role = getRole(backendType,roleId);
Role role = null;
      if (role != null) roleName = role.name;
throw new BackendException();
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return roleName;
  }

  /** check if trackers available
   * @param typeSet backend type set
  * @param projectId project id or Id.ANY
   * @return true iff trackers available
   */
  public static boolean hasTrackers(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("trackers");

    // check if trackers available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(new TrackerDatabaseHandler()
                     {
                       @Override
                       public void process(Tracker tracker)
                       {
                         result[0] = true;
                       }
                     },
                     condition("backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** check if trackers available
   * @return true iff trackers available
   */
  public static boolean hasTrackers()
    throws BackendException
  {
    return hasTrackers(Backend.TypeSet.ALL);
  }

  /** get sorted tracker list
   * @param typeSet backend type set
   * @param projectId project id or Id.ANY
   * @return tracker list
   */
  public static TrackerList getTrackerList(Backend.TypeSet typeSet, Id projectId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("trackers");

    // get trackers from local database
    final TrackerList trackerList = new TrackerList();
    getLocalDatabase(new TrackerDatabaseHandler()
                     {
                       @Override
                       public void process(Tracker tracker)
                       {
                         trackerList.add(tracker);
                       }
                     },
                     condition("(backendType IN ("+typeSet.ordinal()+")) AND (? OR (projectId=?))",
                               projectId.isAny() ? 1 : 0,
                               projectId.backend
                              ),
                     filter("GROUP BY name ORDER BY name COLLATE NOCASE ASC")
                    );

    return trackerList;
  }

  /** get sorted tracker list
   * @param typeSet backend type set
   * @return tracker list
   */
  public static TrackerList getTrackerList(Backend.TypeSet typeSet)
    throws BackendException
  {
    return getTrackerList(typeSet,new Id(Id.ANY));
  }

  /** get sorted tracker list
   * @return tracker list
   */
  public static TrackerList getTrackerList()
    throws BackendException
  {
    return getTrackerList(Backend.TypeSet.ALL);
  }

  /** get tracker
   * @param typeSet backend type set
   * @param trackerId tracker id
   * @return tracker or null
   */
  public final static Tracker getTracker(Backend.TypeSet typeSet, Id trackerId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("trackers");

    // get tracker from local database
    return getLocalDatabase(new TrackerDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?)",
                                      trackerId.backend
                                     ),
                            Filter.NONE
                           );
  }

  /** get tracker
   * @param typeSet backend type set
   * @param name tracker name
   * @return tracker or null
   */
  public final static Tracker getTracker(Backend.TypeSet typeSet, String name)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("trackers");

    // get tracker from local database
    return getLocalDatabase(new TrackerDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (name=?)",
                                      name
                                     ),
                            Filter.NONE
                           );
  }

  /** get tracker id
   * @param name tracker name
   * @return tracker id or Id.NONE
   */
  public final static Id getTrackerId(String name)
// TODO: throws BackendException
  {
    Tracker tracker = null;
    try
    {
      tracker = getTracker(TypeSet.ALL,name);
    }
    catch (BackendException exception)
    {
      // ignored
    }

    return (tracker != null) ? tracker.id : new Id();
  }

  /** get tracker name
   * @param typeSet backend type set
   * @param trackerId tracker id
   * @param defaultName default name
   * @return tracker name
   */
  public final static String getTrackerName(Backend.TypeSet typeSet, Id trackerId, String defaultName)
  {
    String trackerName = defaultName;

    try
    {
      Tracker tracker = getTracker(typeSet,trackerId);
      if (tracker != null) defaultName = tracker.name;
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return trackerName;
  }

  /** check if tracker require epic name
   * @param typeSet backend type set
   * @param trackerId tracker id
   * @return true iff tracker required epic name
   */
  public static boolean isTrackerRequireEpicName(Backend.TypeSet typeSet, Id trackerId)
  {
    boolean trackerRequireEpicName = false;

    try
    {
      Tracker tracker = getTracker(typeSet,trackerId);
      if (tracker != null) trackerRequireEpicName = tracker.isRequireEpicName();
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return trackerRequireEpicName;
  }

  /** get sorted epic name list for tracker
   * @param typeSet backend type set
   * @param trackerId tracker id
   * @return epic name list
   */
//TODO: trackerId required?
  public static List<String> getEpicNameList(Backend.TypeSet typeSet, Id trackerId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("epicNames");

    // get trackers from local database
    final ArrayList<String> epicNameList = new ArrayList<String>();
    getLocalDatabase(new EpicNameDatabaseHandler()
                     {
                       @Override
                       public void process(String name)
                       {
                         epicNameList.add(name);
                       }
                     },
                     condition("backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("ORDER BY name COLLATE NOCASE ASC")
                    );

    return epicNameList;
  }

  /** check if status available
   * @param typeSet backend type set
   * @return true iff status available
   */
  public static boolean hasStatus(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("status");

    // check if status available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(
//TODO: transition list?
                     new StatusDatabaseHandler()
                     {
                       @Override
                       public void process(Status status)
                       {
                         result[0] = true;
                       }
                     },
                     condition("backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** check if status available
   * @return true iff status available
   */
  public static boolean hasStatus()
    throws BackendException
  {
    return hasStatus(Backend.TypeSet.ALL);
  }

  /** get sorted status list
   * @param typeSet backend type set
   * @param issueId issue id or Id.ANY
   * @return status list
   */
  public static StatusList getStatusList(Backend.TypeSet typeSet, Id issueId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("status");

    // get status from local database
    final StatusList statusList = new StatusList();
    getLocalDatabase(
//TODO: transition list?
                     new StatusDatabaseHandler()
                     {
                       @Override
                       public void process(Status status)
                       {
                         statusList.add(status);
                       }
                     },
                     condition("(backendType IN ("+typeSet.ordinal()+")) AND (? OR (backendId=?))",
                               issueId.isAny() ? 1 : 0, issueId.backend
                              ),
                     filter("ORDER BY name COLLATE NOCASE ASC")
                    );

    return statusList;
  }

  /** get sorted status list
   * @param typeSet backend type set
   * @return status list
   */
  public static StatusList getStatusList(Backend.TypeSet typeSet)
    throws BackendException
  {
    return getStatusList(typeSet,new Id(Id.ANY));
  }

  /** get sorted status list
   * @return status list
   */
  public static StatusList getStatusList()
    throws BackendException
  {
    return getStatusList(Backend.TypeSet.ALL);
  }

  /** get status
   * @param typeSet backend type set
   * @param statusId status id
   * @return status or null
   */
  public final static Status getStatus(Backend.TypeSet typeSet, Id statusId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("status");

    // get status from local database
    return getLocalDatabase(new StatusDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?)",
                                      statusId.backend
                                     ),
                            Filter.NONE
                           );
  }

  /** get status
   * @param typeSet backend type set
   * @param name status name
   * @return status or null
   */
  public final static Status getStatus(Backend.TypeSet typeSet, String name)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("status");

    // get status from local database
    return getLocalDatabase(new StatusDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (name=?)",
                                      name
                                     ),
                            Filter.NONE
                           );
  }

  /** get default status id
   * @param typeSet backend type set
   * @return status id or Id.NONE
   */
  public final static Id getDefaultStatusId(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("status");

    // get new issue status name
    String newIssueStatusName = "";
    if      (typeSet.contains(Types.REDMINE)) newIssueStatusName = Settings.Redmine.newIssueStatusName;
    else if (typeSet.contains(Types.JIRA   )) newIssueStatusName = Settings.JIRA.newIssueStatusName;

    Status status = null;
    if (!newIssueStatusName.isEmpty())
    {
      status = getLocalDatabase(new StatusDatabaseHandler(),
                                condition("(backendType IN ("+typeSet.ordinal()+")) AND (name=?)",
                                          newIssueStatusName
                                         ),
                                Filter.NONE
                               );
    }
    else
    {
      status = getLocalDatabase(new StatusDatabaseHandler(),
                                condition("(backendType IN ("+typeSet.ordinal()+")) AND (isDefault=1)"
                                         ),
                                Filter.NONE
                               );
    }

    return (status != null) ? status.id : new Id();
  }

  /** get status id
   * @param name status name
   * @return status id or Id.NONE
   */
  public final static Id getStatusId(String name)
// TODO: throws BackendException
  {
    Status status = null;
    try
    {
      status = getStatus(TypeSet.ALL,name);
    }
    catch (BackendException exception)
    {
      // ignored
    }

    return (status != null) ? status.id : new Id();
  }

  /** get status name
   * @param typeSet backend type set
   * @param statusId status id
   * @param defaultName default name
   * @return status name
   */
  public final static  String getStatusName(Backend.TypeSet typeSet, Id statusId, String defaultName)
  {
    String statusName = defaultName;

    try
    {
      Status status = getStatus(typeSet,statusId);
      if (status != null) statusName = status.name;
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return statusName;
  }

  /** get status name
   * @param typeSet backend type set
   * @param statusId status id
   * @return status name
   */
  public final static  String getStatusName(Backend.TypeSet typeSet, Id statusId)
  {
    return getStatusName(typeSet,statusId,"");
  }

  /** get sorted transition list for issue
   * @param typeSet backend type set
   * @param issueId issue id
   * @return status transition list
   */
  public static StatusTransitionList getStatusTransitionList(Backend.TypeSet typeSet, Id issueId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("statusTransitions");

    // get status transitions from local database
    final StatusTransitionList statusTransitionList = new StatusTransitionList();
    getLocalDatabase(new StatusTransitionDatabaseHandler()
                     {
                       @Override
                       public void process(StatusTransition statusTransition)
                       {
                         statusTransitionList.add(statusTransition);
                       }
                     },
                     condition("(backendType IN ("+typeSet.ordinal()+") AND (issueId=?))",
                               issueId.backend
                              ),
                     filter("GROUP BY name ORDER BY name COLLATE NOCASE ASC")
                    );

    return statusTransitionList;
  }

  /** get sorted status transition list
   * @return status transition list
   */
  public static StatusTransitionList getStatusTransitionList()
    throws BackendException
  {
    return getStatusTransitionList(Backend.TypeSet.ALL,new Id(Id.ANY));
  }

  /** check if priorities available
   * @param typeSet backend type set
   * @return true iff priorities available
   */
  public static boolean hasPriorities(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("priorities");

    // check if priorities available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(new PriorityDatabaseHandler()
                     {
                       @Override
                       public void process(Priority priority)
                       {
                         result[0] = true;
                       }
                     },
                     condition("backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** check if priorities available
   * @return true iff priorities available
   */
  public static boolean hasPriorities()
    throws BackendException
  {
    return hasPriorities(Backend.TypeSet.ALL);
  }

  /** get sorted priority list
   * @param typeSet backend type set
   * @param projectId project id or Id.NONE
   * @return priority list
   */
  public static PriorityList getPriorityList(Backend.TypeSet typeSet, Id projectId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("priorities");

    // get priorities from local database
    final PriorityList priorityList = new PriorityList();
    getLocalDatabase(new PriorityDatabaseHandler()
                     {
                       @Override
                       public void process(Priority priority)
                       {
                         priorityList.add(priority);
                       }
                     },
                     condition("? OR (backendType IN ("+typeSet.ordinal()+"))",
                               projectId.isAny() ? 1 : 0
                              ),
                     filter("ORDER BY name COLLATE NOCASE ASC")
                    );

    return priorityList;
  }

  /** get sorted priority list
   * @return priority list
   */
  public static PriorityList getPriorityList()
    throws BackendException
  {
    return getPriorityList(Backend.TypeSet.ALL,new Id(Id.ANY));
  }

  /** get priority
   * @param typeSet backend type set
   * @param priorityId priority id
   * @return priority or Priority.NONE
   */
  public final static Priority getPriority(Backend.TypeSet typeSet, Id priorityId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("priorities");

    // get priority from local database
    return getLocalDatabase(new PriorityDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?)",
                                      priorityId.backend
                                     ),
                            Filter.NONE
                           );
  }

  /** get priority
   * @param typeSet backend type set
   * @param name priority name
   * @return priority or nuyll
   */
  public final static Priority getPriority(Backend.TypeSet typeSet, String name)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("priorities");

    // get priority from local database
    return getLocalDatabase(new PriorityDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (name=?)",
                                      name
                                     ),
                            Filter.NONE
                           );
  }

  /** get priority id
   * @param typeSet backend type set
   * @param name priority name
   * @return priority id or Id.NONE
   */
  public final static Id getPriorityId(Backend.TypeSet typeSet, String name)
    throws BackendException
  {
    Priority priority = getPriority(TypeSet.ALL,name);

    return (priority != null) ? priority.id : new Id();
  }

  /** get priority name
   * @param typeSet backend type set
   * @param priorityId priority id
   * @param defaultName default name
   * @return priority name
   */
  public final static String getPriorityName(Backend.TypeSet typeSet, Id priorityId, String defaultName)
    throws BackendException
  {
    Priority priority = getPriority(typeSet,priorityId);
    return (priority != null) ? priority.name : defaultName;
  }

  /** get default priority id
   * @param typeSet backend type set
   * @return priority id or Id.NONE
   */
  public final static Id getDefaultPriorityId(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("priorities");

    // get priority from local database
    Priority priority = getLocalDatabase(new PriorityDatabaseHandler(),
                                         condition("(backendType IN ("+typeSet.ordinal()+")) AND (isDefault=1)"
                                                  ),
                                         Filter.NONE
                                        );

    return (priority != null) ? priority.id : new Id();
  }

  /** check if activities available
   * @param typeSet backend type set
   * @param issueId issue id or Id.ANY (not used)
   * @return true iff activities available
   */
  public static boolean hasActivities(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("activities");

    // check if activities available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(new ActivityDatabaseHandler()
                     {
                       @Override
                       public void process(Activity activity)
                       {
                         result[0] = true;
                       }
                     },
                     condition("backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** get sorted activities list
   * @return true iff activities available
   */
  public static boolean hasActivities()
    throws BackendException
  {
    return hasActivities(Backend.TypeSet.ALL);
  }

  /** check if activities available
   * @param typeSet backend type set
   * @param issueId issue id or Id.ANY (not used)
   * @return activity list
   */
  public static ActivityList getActivityList(Backend.TypeSet typeSet, Id issueId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("activities");

    // get activities from local database
    final ActivityList activityList = new ActivityList();
    getLocalDatabase(new ActivityDatabaseHandler()
                     {
                       @Override
                       public void process(Activity activity)
                       {
                         activityList.add(activity);
                       }
                     },
                     condition("backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("ORDER BY name COLLATE NOCASE ASC")
                    );

    return activityList;
  }

  /** get sorted activities list
   * @return activity list
   */
  public static ActivityList getActivityList()
    throws BackendException
  {
    return getActivityList(Backend.TypeSet.ALL,new Id(Id.ANY));
  }

  /** get default activity id
   * @param typeSet backend type set
   * @return id or Id.NONE
   */
  public final static Id getDefaultActivityId(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("activities");

    // get activity from local database
    Activity activity = getLocalDatabase(new ActivityDatabaseHandler(),
                                         condition("isDefault=1"),
                                         Filter.NONE
                                        );

    return (activity != null) ? activity.id : new Id();
  }

  /** get activity
   * @param typeSet backend type set
   * @param id activity id
   * @return activity or null
   */
  public final static Activity getActivity(Backend.TypeSet typeSet, Id activityId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("activities");

    // get activity from local database
    return getLocalDatabase(new ActivityDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?)",
                                      activityId.backend
                                     ),
                            Filter.NONE
                           );
  }

  /** get activity by name
   * @param backendType backend type
   * @param name activity name
   * @return activity or null
   */
  public final static Activity getActivity(Backend.TypeSet typeSet, String name)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("activities");

    return getLocalDatabase(new ActivityDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (name=?)",
                                      name
                                     ),
                            Filter.NONE
                           );
  }

  /** check if projects available
   * @return true iff projects available
   */
  public static boolean hasProjects()
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("projects");

    // check if projects available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(new ProjectDatabaseHandler()
                     {
                       @Override
                       public void process(Project project)
                       {
                         result[0] = true;
                       }
                     },
                     condition("(status=?) AND ((SELECT rowid FROM projectsModified WHERE projectId=projects.id AND modifiedState=?) IS NULL)",
                               ProjectStates.ACTIVE.getValue(),
                               Backend.ModifiedStates.DELETED
                              ),
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** get sorted projects list
   * @return sorted project list
   */
  public static ProjectList getProjectList()
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("projects");

    // get projects from local database
    final ProjectList projectList = new ProjectList();
    getLocalDatabase(new ProjectDatabaseHandler()
                     {
                       @Override
                       public void process(Project project)
                       {
                         projectList.add(project);
                       }
                     },
                     condition("(status=?) AND ((SELECT rowid FROM projectsModified WHERE projectId=projects.id AND modifiedState=?) IS NULL)",
                               ProjectStates.ACTIVE.getValue(),
                               Backend.ModifiedStates.DELETED
                              ),
                     filter("ORDER BY name COLLATE NOCASE ASC")
                    );

    return projectList;
  }

  /** get project by id
   * @param backendType backend type
   * @param id project id
   * @return project or null
   */
  public final static Project getProject(Backend.TypeSet typeSet, Id projectId, DataBaseHandler<Project> databaseHandler)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("projects");

    // get project from local database
    return getLocalDatabase(databaseHandler,
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?) AND ((SELECT rowid FROM projectsModified WHERE projectId=projects.id AND modifiedState=?) IS NULL)",
                                      projectId.backend
                                     ),
                            Filter.NONE
                           );
  }

  /** get project by id
   * @param backendType backend type
   * @param id project id
   * @return project or null
   */
  public static Project getProject(Backend.TypeSet typeSet, Id projectId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("projects");

    // get project from local database
    return getLocalDatabase(new ProjectDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?) AND ((SELECT rowid FROM projectsModified WHERE projectId=projects.id AND modifiedState=?) IS NULL)",
                                      projectId.backend
                                     ),
                            Filter.NONE
                           );
  }

  /** get project by number
   * @param backendType backend type
   * @param n number
   * @return project or null
   */
  public static Project getProject(Backend.TypeSet typeSet, int n)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("projects");

    // get project from local database
    return getLocalDatabase(new ProjectDatabaseHandler(),
                            condition("(backendId IN ("+typeSet.ordinal()+")) AND ((SELECT rowid FROM projectsModified WHERE projectId=projects.id AND modifiedState=?) IS NULL)",
                                      n
                                     ),
                            Filter.NONE
                           );
  }

  /** get project by name
   * @param backendType backend type
   * @param name project name
   * @return project or null
   */
  public static Project getProject(Backend.TypeSet typeSet, String name)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("projects");

    // get project from local database
    return getLocalDatabase(new ProjectDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND ((shortName=?) OR (name=?)) AND ((SELECT rowid FROM projectsModified WHERE projectId=projects.id AND modifiedState=?) IS NULL)",
                                      name,
                                      name
                                     ),
                            Filter.NONE
                           );
  }

  /** get project id
   * @param name project name
   * @return project id or Id.NONE
   */
  public final static Id getProjectId(String name)
    throws BackendException
  {
    Project project = getProject(TypeSet.ALL,name);

    return (project != null) ? project.id : new Id();
  }

  /** get project by name
   * @param name project name
   * @return project or null
   */
  public final static Project getProject(String name)
    throws BackendException
  {
    return getProject(TypeSet.ALL,name);
  }

  /** check if local modified projects
   * @return true iff modified projects
   */
  public final static boolean hasProjectModifieds()
  {
    final boolean result[] = new boolean[]{ false };
    try
    {
      getLocalDatabase(new ProjectModifiedDatabaseHandler()
                       {
                         public boolean process(Issue issue)
                         {
                           result[0] = true;
                           return false;
                         }
                       },
                       condition("modifiedId IS NOT NULL"),
                       filter("LIMIT 0,1")
                      );
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return result[0];
  }

  /** get next local modifed project
   * @return modified project or null
   */
//TODO: get modified projects
  public final static synchronized ProjectModified getProjectModified()
  {
    final ProjectModified ProjectModified[] = new ProjectModified[]{null};
/*
    try
    {
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }
*/

    return ProjectModified[0];
  }

  /** get estimated duration sum
   * @param backendType backend type
   * @param id project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @return estimated duration sum [min]
   */
  public static long getEstimatedDurationSum(Backend.TypeSet typeSet, Id projectId, Id issueId)
  {
    long estimatedDurationSum = 0;

    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("projects");

    try
    {
      estimatedDurationSum = getLocalDatabase("issues",
                                              "WITH RECURSIVE "+
                                              "  subProjects(id) AS "+
                                              "  ( "+
                                              "    VALUES(?) "+
                                              "    UNION SELECT projects.id FROM projects,subProjects WHERE projects.parentId=subProjects.id AND ((SELECT rowid FROM projectsModified WHERE projectId=projects.id AND modifiedState=?) IS NULL)"+
                                              "  )",
                                              "SUM(estimatedDuration)",
                                              condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                                        "AND (projectId IN subProjects) "+
                                                        "AND (parentId=0) ",
// TODO: use IN operator+ANY
                                                        projectId
                                                       ),
                                              Filter.NONE,
                                              0L
                                             );
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return estimatedDurationSum;
  }

  /** check if members available
   * @param backendType backend type
   * @return true iff members available
   */
  public static boolean hasMembers(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("members");

    // check if members available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(new MemberDatabaseHandler()
                     {
                       @Override
                       public void process(Member member)
                       {
                         result[0] = true;
                       }
                     },
                     condition("backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** check if members available
   * @return true iff members available
   */
  public static boolean hasMembers()
    throws BackendException
  {
    return hasMembers(Backend.TypeSet.ALL);
  }

  /** get member list
   * @param backendType backend type
   * @param projectId project id
   * @param roleIds role filter ids or NULL
   * @return sorted member list
   */
  public static MemberList getMemberList(Backend.TypeSet typeSet, Id projectId, List<Id> roleIds)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("members");

    // get role id filter string
    StringBuilder roleIdFilter = new StringBuilder();
    for (Id roleId : roleIds)
    {
      if (roleIdFilter.length() > 0) roleIdFilter.append(',');
      roleIdFilter.append(roleId.backend);
    }

    // get members from local database
    final MemberList memberList = new MemberList();
    getLocalDatabase(new MemberDatabaseHandler()
                     {
                       @Override
                       public void process(Member member)
                       {
                         memberList.add(member);
                       }
                     },
                     condition("? OR ((backendType IN ("+typeSet.ordinal()+")) AND (projectId=?) AND ((roleId=0) OR (roleId IN ("+roleIdFilter+"))))",
                               projectId.isAny() ? 1 : 0,
                               projectId.backend
                              ),
                     filter("ORDER BY lastName,firstName ASC")
                    );

    return memberList;
  }

  /** get member
   * @param backendType backend type
   * @param memberId version id
   * @return member or null
   */
  public final static Member getMember(Backend.TypeSet typeSet, Id memberId)
    throws BackendException
  {
    Member member;

    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("members");

    if      (memberId.isAny())
    {
      member = Member.ANY;
    }
    else if (memberId.isNone())
    {
      member = null;
    }
    else
    {
      // get member from local database
      member = getLocalDatabase(new MemberDatabaseHandler(),
                                condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?)",
                                          memberId.backend
                                         ),
                                Filter.NONE
                               );
    }

    return member;
  }

  /** check if assigned-to current user
   * @param backendType backend type
   * @param projectId project id
   * @param memberId version id
   * @return true iff assigned-to current user
   */
  public static boolean isAssignedToCurrentUser(Backend.Types type, Id projectId, Id assignedToMemberId)
  {
    Member member = null;

    try
    {
      member = Backend.backendDrivers.get(type).getCurrentMember(projectId);
    }
    catch (BackendException exception)
    {
      // ignored
    }
    catch (InterruptedException exception)
    {
      // ignored
    }

    return member.id.equals(assignedToMemberId);
  }

  /** check if versions available
   * @param backendType backend type
   * @return true iff versions available
   */
  public static boolean hasVersions(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("versions");

    // check if issues available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(new VersionDatabaseHandler()
                     {
                       @Override
                       public void process(Version version)
                       {
                         result[0] = true;
                       }
                     },
                     condition("backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** check if versions available
   * @return true iff versions available
   */
  public static boolean hasVersions()
    throws BackendException
  {
    return hasVersions(Backend.TypeSet.ALL);
  }

  /** get version list
   * @param backendType backend type
   * @param projectId project id
   * @return sorted version list
   */
  public static VersionList getVersionList(Backend.TypeSet typeSet, Id projectId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("versions");

    // get versions from local database
    final VersionList versionList = new VersionList();
    getLocalDatabase(new VersionDatabaseHandler()
                     {
                       @Override
                       public void process(Version version)
                       {
                         versionList.add(version);
                       }
                     },
                     condition("? OR ((backendType IN ("+typeSet.ordinal()+")) AND (projectId=?))",
                               projectId.isAny() ? 1 : 0,
                               projectId.backend
                              ),
                     filter("ORDER BY name ASC")
                    );

    return versionList;
  }

  /** get version
   * @param backendType backend type
   * @param versionId version id
   * @return version or null
   */
  public final static Version getVersion(Backend.TypeSet typeSet, Id versionId)
    throws BackendException
  {
    Version version;

    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("versions");

    if      (versionId.isAny())
    {
      version = Version.ANY;
    }
    else if (versionId.isNone())
    {
      version = Version.NONE;
    }
    else
    {
      // get version from local database
      version = getLocalDatabase(new VersionDatabaseHandler(),
                                 condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?)",
                                           versionId.backend
                                          ),
                                  Filter.NONE
                                );
    }

    return version;
  }

  /** check if issues available
   * @param backendType backend type
   * @return true iff issues available
   */
  public static boolean hasIssues(Backend.TypeSet typeSet)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("issues");

    // check if issues available
    final boolean result[] = new boolean[]{false};
    getLocalDatabase(new IssueDatabaseHandler()
                     {
                       @Override
                       public void process(Issue issue)
                       {
                         result[0] = true;
                       }
                     },
                     condition("issues.backendType IN ("+typeSet.ordinal()+")"
                              ),
                     filter("LIMIT 0,1")
                    );

    return result[0];
  }

  /** check if issues available
   * @return true iff issues available
   */
  public static boolean hasIssues()
    throws BackendException
  {
    return hasIssues(Backend.TypeSet.ALL);
  }

  /** get issues list of project and all sub-projects
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param activeOnly true for active projects, false for all projets
   * @return sorted issue list
   */
  public static IssueList getIssueList(Backend.TypeSet typeSet, Id projectId, final boolean activeOnly)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("issues");

    // get active status ids
    final HashSet<Id> activeStatusIds = new HashSet<Id>();
    for (Status status : getStatusList())
    {
      if (Settings.issueActiveStatusSet.contains(status.name))
      {
        activeStatusIds.add(status.id);
      }
    }

    // get issues from local database
    final IssueList issueList = new IssueList();
    getLocalDatabase(new IssueDatabaseHandler()
                     {
                       @Override
                       public void process(Issue issue)
                       {
                         if (!activeOnly || activeStatusIds.contains(issue.statusId))
                         {
                           issueList.add(issue);
                         }
                       }
                     },
                     condition("    (issues.backendType IN ("+typeSet.ordinal()+")) "+
                               "AND (? OR projectId IN (WITH RECURSIVE P(backendId,parentId) AS (VALUES(?,0) UNION ALL SELECT projects.backendId,projects.parentId FROM projects,P WHERE projects.parentId=P.backendId) SELECT backendId FROM P"+
                               "                       )"+
                               "    )",
                               projectId.isAny() ? 1 : 0,
                               projectId.backend
                              ),
                     filter("ORDER BY summary COLLATE NOCASE ASC")
                    );

    return issueList;
  }

  /** get sorted issues list
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @return issue list
   */
  public final static IssueList getIssueList(Backend.TypeSet typeSet, Id projectId)
    throws BackendException
  {
    return getIssueList(typeSet,projectId,false);
  }

  /** get sorted issues list
   * @return issue list
   */
  public final static IssueList getIssueList()
    throws BackendException
  {
    return getIssueList(Backend.TypeSet.ALL,new Id(Id.ANY),false);
  }

  /** get issue by id
   * @param typeSet backend type set
   * @param issueId issue id
   * @return issue or null
   */
  public static Issue getIssue(Backend.TypeSet typeSet, Id issueId)
    throws BackendException
  {
    assert issueId.hasBackend();

    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("issues");

    // get issue from local database
    return getLocalDatabase(new IssueDatabaseHandler(),
                            condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                      "AND (backendId=?) ",
                                      issueId.backend
                                     ),
                            filter("LIMIT 0,1")
                           );
  }

  /** get origin issue by id
   * @param typeSet backend type set
   * @param issueId issue id
   * @return issue or null
   */
  public static Issue getIssueOrigin(Backend.TypeSet typeSet, Id issueId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("issues");

    if (issueId.hasBackend())
    {
      // get issue from local database
      Issue issue = getLocalDatabase(new IssueDatabaseHandler(),
                                     condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                               "AND (id=?)",
                                               issueId.local
                                              ),
                                     filter("LIMIT 0,1")
                                    );

      // load original issue from server
      return (issue != null)
               ? Backend.backendDrivers.get(issue.backendType).loadIssue(issueId)
               : null;
    }
    else
    {
      return null;
    }
  }

  /** get issue by number
   * @param typeSet backend type set
   * @param n number
   * @return issue or null
   */
  public static Issue getIssue(Backend.TypeSet typeSet, int n)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("issues");

    // get issue from local database
    return getLocalDatabase(new IssueDatabaseHandler(),
                            condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                      "AND (backendId=?)",
                                      n
                                     ),
                            Filter.NONE
                           );
  }

  /** get issue by summary
   * @param backendType backend type
   * @param summary issue summary
   * @return issue or null
   */
  public static Issue getIssue(Backend.TypeSet typeSet, String summary)
    throws BackendException
  {
// TODO:    assert backendType != Backend.Types.NONE;

    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("issues");

    // get issue from local database
    return getLocalDatabase(new IssueDatabaseHandler(),
                            condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                      "AND (summary=?)",
                                      summary
                                     ),
                            Filter.NONE
                           );
  }

  /** get issue id
   * @param backendType backend type
   * @param summary issue summary
   * @return issue id or Id.NONE
   */
  public static Id getIssueId(Backend.TypeSet typeSet, String summary)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("issues");

    // get issue from local database
    return getLocalDatabase(new IssueDatabaseHandler(),
                            condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                      "AND (summary=?)"+
                                      summary
                                     ),
                            Filter.NONE
                           ).id;
  }

  /** get pending issue status transition
   * @param backendType backend type
   * @param statusTransitionId status transition id
   * @return status transition or null
   */
  public final static StatusTransition getStatusTransition(Backend.TypeSet typeSet, Id statusTransitionId)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("status");

    // get status from local database
    return getLocalDatabase(new StatusTransitionDatabaseHandler(),
                            condition("(backendType IN ("+typeSet.ordinal()+")) AND (backendId=?)",
                                      statusTransitionId.backend
                                     ),
                            Filter.NONE
                           );
  }

  /** check if local modified issues
   * @return true iff modified issues
   */
  public final static boolean hasIssueModifieds()
  {
    final boolean result[] = new boolean[]{ false };
    try
    {
      getLocalDatabase(new IssueModifiedDatabaseHandler()
                       {
                         public boolean process(Issue issue)
                         {
                           result[0] = true;
                           return false;
                         }
                       },
                       condition("modifiedId IS NOT NULL"),
                       filter("LIMIT 0,1")
                      );
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return result[0];
  }

  /** get next local modifed issue
   * @return modified issue or null
   */
  public final static synchronized IssueModified getIssueModified()
  {
    IssueModified modfiedIssue = null;

    try
    {
      // get next modified entry
      modfiedIssue = getLocalDatabase(new IssueModifiedDatabaseHandler(),
                                      Condition.NONE,  // condition
                                      filter("ORDER BY modifiedId ASC LIMIT 0,1")
                                     );
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return modfiedIssue;
  }

  /** get today time entries
   * @return time entry list
   */
  public final static TimeEntryList getTimeEntryListToday()
    throws BackendException
  {
    return getTimeEntryList(SpentOn.today());
  }

  /** check if local modified time entires
   * @return true iff modified time entries
   */
  public final static boolean hasModifiedTimeEntries()
  {
    final boolean result[] = new boolean[]{ false };
    try
    {
      getLocalDatabase(new TimeEntryModifiedDatabaseHandler()
                       {
                         public boolean process(TimeEntry timeEntry)
                         {
                           result[0] = true;
                           return false;
                         }
                       },
                       condition("modifiedId IS NOT NULL"),
                       filter("LIMIT 0,1")
                      );
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
      Timemine.printStackTrace(exception);
    }

    return result[0];
  }

  /** get next local modified time entry
   * @return modified time entry or null
   */
  public final static synchronized TimeEntryModified getTimeEntryModified()
  {
    TimeEntryModified modifiedTimeEntry = null;

    try
    {
      // get next modified entry
      modifiedTimeEntry = getLocalDatabase(new TimeEntryModifiedDatabaseHandler(),
                                           Condition.NONE,  // condition
                                           filter("ORDER BY modifiedId ASC LIMIT 0,1")
                                          );
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return modifiedTimeEntry;
  }

  /** get time entry start date
   * @return start date
   */
  public final static Date getTimeEntryStartDate()
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("timeEntries");

    // get first time entry
    TimeEntry timeEntry = getLocalDatabase(new TimeEntryDatabaseHandler(),
                                           Condition.NONE,  // condition
                                           filter("ORDER BY spentOn ASC LIMIT 0,1")
                                          );

    return (timeEntry != null) ? timeEntry.spentOn.getDate() : new Date();
  }

  /** get duration sum of time entries
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param userId user id or Id.ANY
   * @param spentOn spent-on date or null
   * @return duration sum of time entries or 0 if data still not available [min]
   */
  public final static long getDurationSum(Backend.TypeSet typeSet, Id projectId, Id issueId, Id userId, Date from, Date to)
  {
    long durationSum = 0;

    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("timeEntries");

    try
    {
      durationSum = getLocalDatabase("timeEntries",
                                     "WITH RECURSIVE "+
                                     "  subProjects(backendId) AS "+
                                     "  ( "+
                                     "    VALUES(?) "+
                                     "    UNION SELECT projects.backendId FROM projects,subProjects WHERE (projects.parentId=subProjects.backendId)"+
                                     "  ),"+
                                     "  subIssues(backendId) AS "+
                                     "  ("+
                                     "    VALUES(?) "+
                                     "    UNION SELECT issues.backendId FROM issues,subIssues WHERE (issues.parentId=subIssues.backendId)"+
                                     "  )",
                                     "SUM(duration)",
                                     condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                               "AND (? OR (projectId IN subProjects)) "+
                                               "AND (? OR (issueId IN subIssues)) "+
                                               "AND (? OR (userId=?)) "+
                                               "AND (? OR (DATE(spentOn,'unixepoch','localtime') >= DATE(?,'unixepoch','localtime'))) "+
                                               "AND (? OR (DATE(spentOn,'unixepoch','localtime') <= DATE(?,'unixepoch','localtime'))) ",
                                               !projectId.isNone() ? projectId.backend : 0,
                                               !issueId.isNone() ? issueId.backend : 0,
                                               projectId.isAny() ? 1 : 0,
                                               issueId.isAny()   ? 1 : 0,
                                               userId.isAny()    ? 1 : 0,userId.backend,
                                               (from         == null  ) ? 1 : 0,(from   != null  ) ? from.getTime()/1000 : 0,
                                               (to           == null  ) ? 1 : 0,(to     != null  ) ? to.getTime()  /1000 : 0
                                              ),
                                     Filter.NONE,
                                     0L
                                    );
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return durationSum;
  }

  /** get duration sum of time entries
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param userId user id or Id.ANY
   * @param spentOnStart/spentOnEnd spent-on from/to date
   * @return duration sum of time entries or 0 if data still not available [min]
   */
  public final static long getDurationSum(Backend.TypeSet typeSet, Id projectId, Id issueId, Id userId, SpentOn spentOnStart, SpentOn spentOnEnd)
  {
    return getDurationSum(typeSet,projectId,issueId,userId,spentOnStart.getDate(),spentOnEnd.getDate());
  }

  /** get duration sum of time entries
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param spentOnStart/spentOnEnd spent-on from/to date
   * @return duration sum of time entries or 0 if data still not available [min]
   */
  public final static long getDurationSum(Backend.TypeSet typeSet, Id projectId, Id issueId, SpentOn spentOnStart, SpentOn spentOnEnd)
  {
    return getDurationSum(typeSet,projectId,issueId,new Id(Id.ANY),spentOnStart,spentOnEnd);
  }

  /** get duration sum of time entries
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param userId user id or Id.ANY
   * @param spentOn spent-on date
   * @return duration sum of time entries or 0 if data still not available [min]
   */
  public final static long getDurationSum(Backend.TypeSet typeSet, Id projectId, Id issueId, Id userId, SpentOn spentOn)
  {
    return getDurationSum(typeSet,projectId,issueId,userId,spentOn,spentOn);
  }

  /** get duration sum of own time entries
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param spentOn spent-on date or null
   * @return duration sum of time entries [min]
   */
  public final static long getDurationSum(Backend.TypeSet typeSet, Id projectId, Id issueId, Date from, Date to)
  {
    return getDurationSum(typeSet,projectId,issueId,new Id(Id.ANY),from,to);
  }

  /** get duration sum of own time entries
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param spentOn spent-on date
   * @return duration sum of time entries [min]
   */
  public final static long getDurationSum(Backend.TypeSet typeSet, Id projectId, Id issueId, SpentOn spentOn)
  {
    return getDurationSum(typeSet,projectId,issueId,new Id(Id.ANY),spentOn);
  }

  /** get duration sum of time entries
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param userId user id or Id.ANY
   * @return duration sum of time entries or UPDATE if data still not available [min]
   */
  public final static long getDurationSum(Backend.TypeSet typeSet, Id projectId, Id issueId, Id userId)
  {
    return getDurationSum(typeSet,projectId,issueId,userId,(Date)null,(Date)null);
  }

  /** get duration sum of own time entries
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @return duration sum of time entries [min]
   */
  public final static long getDurationSum(Backend.TypeSet typeSet, Id projectId, Id issueId)
  {
    return getDurationSum(typeSet,projectId,issueId,new Id(Id.ANY));
  }

  /** get duration sum of own time entries
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @return duration sum of time entries [min]
   */
  public final static long getDurationSum(Backend.TypeSet typeSet, Id projectId)
  {
    return getDurationSum(typeSet,projectId,new Id(Id.ANY));
  }

  /** get duration sum of own time entries
   * @param spentOn spent-on date or null
   * @return duration sum of time entries [min]
   */
  public final static long getDurationSum(SpentOn spentOnStart, SpentOn spentOnEnd)
  {
    return getDurationSum(Backend.TypeSet.ALL,new Id(Id.ANY),new Id(Id.ANY),spentOnStart,spentOnEnd);
  }

  /** get duration sum of own time entries
   * @param spentOn spent-on date or null
   * @return duration sum of time entries [min]
   */
  public final static long getDurationSum(SpentOn spentOn)
  {
    return getDurationSum(spentOn,spentOn);
  }

  /** get today duration sum of own time entries
   * @return duration sum of time entries [min]
   */
  public final static long getTimeEntryTodayDurationSum()
  {
    return getDurationSum(SpentOn.today());
  }

  /** get sorted time entry list
   * @param backendType backend type or Backend.Types.ANY
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param userId user id or Id.ANY
   * @param spentOnStart,spentOnEnd spent-on from/to date
   * @return sorted time entry list
   */
  public static TimeEntryList getTimeEntryList(Backend.TypeSet typeSet, Id projectId, Id issueId, Id userId, SpentOn spentOnStart, SpentOn spentOnEnd)
    throws BackendException
  {
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("timeEntries");

    // get time entries from local database
    final TimeEntryList timeEntryList = new TimeEntryList();
    getLocalDatabase(new TimeEntryDatabaseHandler()
                     {
                       @Override
                       public void process(TimeEntry timeEntry)
                       {
                         timeEntryList.add(timeEntry);
                       }
                     },
                     condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                               "AND (? OR (projectId=?)) "+
                               "AND (? OR (issueId=?)) "+
                               "AND (? OR (userId=?)) "+
                               "AND (? OR (DATE(spentOn,'unixepoch','localtime') >= DATE(?,'unixepoch','localtime'))) "+
                               "AND (? OR (DATE(spentOn,'unixepoch','localtime') <= DATE(?,'unixepoch','localtime'))) ",
                               projectId.isAny() ? 1 : 0,projectId.backend,
                               issueId.isAny() ? 1 : 0,issueId.backend,
                               userId.isAny() ? 1 : 0,userId.backend,
                               (spentOnStart == null) ? 1 : 0,(spentOnStart != null) ? spentOnStart.getTime()/1000 : 0,
                               (spentOnEnd   == null) ? 1 : 0,(spentOnEnd   != null) ? spentOnEnd.getTime()/1000   : 0
                              ),
                     filter("ORDER BY spentOn ASC")
                    );

    return timeEntryList;
  }

  /** get sorted time entry list
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param userId user id or Id.ANY
   * @param spentOn spent-on data
   * @return sorted time entry list
   */
  public static TimeEntryList getTimeEntryList(Backend.TypeSet typeSet, Id projectId, Id issueId, Id userId, SpentOn spentOn)
    throws BackendException
  {
    return getTimeEntryList(typeSet,projectId,issueId,userId,spentOn,spentOn);
  }

  /** get sorted time entry list
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param spentOnStart,spentOnEnd spent-on from/to date
   * @return sorted time entry list
   */
  public static TimeEntryList getTimeEntryList(Backend.TypeSet typeSet, Id projectId, Id issueId, SpentOn spentOnStart, SpentOn spentOnEnd)
    throws BackendException
  {
    return getTimeEntryList(typeSet,projectId,issueId,new Id(Id.ANY),spentOnStart,spentOnEnd);
  }

  /** get sorted time entry list
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @param spentOn spent-on date
   * @return sorted time entry list
   */
  public static TimeEntryList getTimeEntryList(Backend.TypeSet typeSet, Id projectId, Id issueId, SpentOn spentOn)
    throws BackendException
  {
    return getTimeEntryList(typeSet,projectId,issueId,spentOn,spentOn);
  }

  /** get own time entries
   * @param spentOnStart,spentOnEnd spent-on from/to date
   * @return sorted time entry list
   */
  public final static TimeEntryList getTimeEntryList(SpentOn spentOnStart, SpentOn spentOnEnd)
    throws BackendException
  {
    return getTimeEntryList(Backend.TypeSet.ALL,new Id(Id.ANY),new Id(Id.ANY),spentOnStart,spentOnEnd);
  }

  /** get own time entries as an array
   * @param spentOn spent-on date
   * @return time entry array
   */
  public final static TimeEntryList getTimeEntryList(SpentOn spentOn)
    throws BackendException
  {
    return getTimeEntryList(Backend.TypeSet.ALL,new Id(Id.ANY),new Id(Id.ANY),spentOn);
  }

  /** get own time entries as an array
   * @param spentOn spent-on date
   * @return time entry array
   */
  public final static TimeEntryList getTimeEntryList()
    throws BackendException
  {
    return getTimeEntryList(Backend.TypeSet.ALL,new Id(Id.ANY),new Id(Id.ANY),(SpentOn)null);
  }

  /** get own time entries of today
   * @param backendType backend type
   * @param projectId project id or Id.ANY
   * @param issueId issue id or Id.ANY
   * @return time entry array
   */
  public final static TimeEntryList getTimeEntryList(Backend.TypeSet typeSet, Id projectId, Id issueId)
    throws BackendException
  {
    return getTimeEntryList(typeSet,projectId,issueId,SpentOn.today());
  }

  /** get time entry
   * @param backendType backend type
   * @param timeEntryId id of time entry
   * @return time entry or TimeEntry.NONE
   */
  public final static TimeEntry getTimeEntry(Backend.TypeSet typeSet, Id timeEntryId)
    throws BackendException
  {
    assert timeEntryId.hasBackend();

    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("timeEntries");

    // get time entry from local database
    return getLocalDatabase(new TimeEntryDatabaseHandler(),
                            condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                      "AND (backendId=?)",
                                      timeEntryId.backend
                                     ),
                            filter("LIMIT 0,1")
                           );
  }

  /** get origin time entry by id
   * @param backendType backend type or Backend.Types.NONE
   * @param timeEntryId time entry id
   * @return time entry or TimeEntry.NONE
   */
  public static TimeEntry getTimeEntryOrigin(Backend.TypeSet typeSet, Id timeEntryId)
    throws BackendException
  {
Dprintf.dprintf("xxxxxxxxxxxx");
    // check if update is required (Note: usually updates are done automatically in background)
    checkLastUpdate("timeEntries");

    if (timeEntryId.hasBackend())
    {
      // get time entry from local database
      TimeEntry timeEntry = getLocalDatabase(new TimeEntryDatabaseHandler(),
                                             condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                                       "AND (id=?)",
                                                       timeEntryId.local
                                                      ),
                                             filter("LIMIT 0,1")
                                            );
Dprintf.dprintf("timeEntry=%s",timeEntry);

      // load original time entry from server
      return (timeEntry != null)
               ? Backend.backendDrivers.get(timeEntry.backendType).loadTimeEntry(timeEntry.projectId,timeEntry.issueId,timeEntryId)
               : null;
    }
    else
    {
      return null;
    }
  }

  /** get time entry at specific date (newest entry)
   * @param backendType backend type
   * @param spentOn spent-on date
   * @return time entry or TimeEntry.NONE
   */
  public final TimeEntry getTimeEntry(Backend.TypeSet typeSet, SpentOn spentOn)
    throws BackendException
  {
    TimeEntry timeEntry = null;

    timeEntry = getLocalDatabase(new TimeEntryDatabaseHandler(),
                                 condition("    (backendType IN ("+typeSet.ordinal()+")) "+
                                           "AND DATE(spentOn,'unixepoch','localtime')=DATE(?,'unixepoch','localtime')",
                                           spentOn.getTime()/1000
                                          ),
                                 filter("ORDER BY spentOn DESC LIMIT 0,1")
                                );

    return timeEntry;
  }

  /** get time entry history (last n entries)
   * @return time entry list
   */
  public final static TimeEntryList getTimeEntryHistory(int n, HistoryFilterTypes historyFilterType)
  {
    final TimeEntryList timeEntryList = new TimeEntryList();
    try
    {
      Filter historyFilter = null;
      switch (historyFilterType)
      {
        case ISSUES:
          historyFilter = filter("GROUP BY projectId,issueId,duration "+
                                 "ORDER BY spentOn DESC "+
                                 "LIMIT 0,?",
                                 n
                                );
          break;
        case COMMENTS:
          historyFilter = filter("GROUP BY comments "+
                                 "ORDER BY spentOn DESC "+
                                 "LIMIT 0,?",
                                 n
                                );
          break;
      }

      getLocalDatabase(new TimeEntryDatabaseHandler()
                       {
                         @Override
                         public void process(TimeEntry timeEntry)
                         {
                           timeEntryList.add(timeEntry);
                         }
                       },
                       Condition.NONE,
                       historyFilter
                      );
    }
    catch (BackendException exception)
    {
      // ignored
      Timemine.printInternalError(exception);
    }

    return timeEntryList;
  }

  //--------------------------------------------------------------------

  private static void discardIndizes()
    throws SQLException
  {
    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = database.prepareStatement("SELECT name FROM sqlite_master WHERE type='index' AND sql LIKE 'CREATE INDEX %';");
      ResultSet resultSet = preparedStatement.executeQuery();
      ArrayList<String> indexNameList = new ArrayList<String>();
      while (resultSet.next())
      {
        indexNameList.add(resultSet.getString(1));
      }
      preparedStatement.close(); preparedStatement = null;

      for (String indexName : indexNameList)
      {
        preparedStatement = database.prepareStatement("DROP INDEX "+indexName+";");
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;
      }
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }
  }

  private static void discardViews()
    throws SQLException
  {
    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = database.prepareStatement("SELECT name FROM sqlite_master WHERE type='view' AND sql LIKE 'CREATE VIEW %';");
      ResultSet resultSet = preparedStatement.executeQuery();
      ArrayList<String> viewNameList = new ArrayList<String>();
      while (resultSet.next())
      {
        viewNameList.add(resultSet.getString(1));
      }
      preparedStatement.close(); preparedStatement = null;

      for (String viewName : viewNameList)
      {
        preparedStatement = database.prepareStatement("DROP VIEW "+viewName+";");
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;
      }
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }
  }

  /** initialze local database tables and indizes (if not exists)
   * Note:
   *   - all id are local database ids
   *   - all *Id are backend ids
   */
  private static void initLocalDatabaseTables()
    throws SQLException
  {
//TOOD: debug only
if (System.getenv("DI") != null) { Dprintf.dprintf("discard tables"); discardIndizes(); }
if (System.getenv("DV") != null) { Dprintf.dprintf("discard views"); discardViews(); }

    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS meta ( "+
                                                    "  id    INTEGER PRIMARY KEY, "+
                                                    "  name  TEXT UNIQUE, "+
                                                    "  value TEXT "+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO meta (name,value) VALUES ('version',?)");
      preparedStatement.setString(1,"1");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO meta (name,value) VALUES ('dateTime',DATE('now'))");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS lastUpdates ( "+
                                                    "  id        INTEGER PRIMARY KEY, "+
                                                    "  name      TEXT UNIQUE, "+
                                                    "  timestamp INTEGER "+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS users ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  firstName           TEXT NOT NULL, "+
                                                    "  lastName            TEXT NOT NULL, "+
                                                    "  name                TEXT NOT NULL, "+
                                                    "  roleId              INTEGER, "+
                                                    "  mail                TEXT NOT NULL, "+
                                                    "  createdOn           INTEGER, "+
                                                    "  lastLoginOn         INTEGER, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT roles UNIQUE (backendType,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS usersIndex0 ON users(backendType,name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS usersIndex1 ON users(name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS roles ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  projectId           INTEGER, "+
                                                    "  name                TEXT NOT NULL, "+
                                                    "  description         TEXT NOT NULL, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT roles UNIQUE (backendType,projectId,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS rolesIndex0 ON roles(backendType,projectId,name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS rolesIndex1 ON roles(name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS trackers ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  projectId           INTEGER, "+
                                                    "  name                TEXT NOT NULL, "+
                                                    "  description         TEXT NOT NULL, "+
                                                    "  requireEpicName     INTEGER DEFAULT 0, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT trackers UNIQUE (backendType,projectId,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS trackersIndex0 ON trackers(backendType,projectId,name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS trackersIndex1 ON trackers(name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS epicNames ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  name                TEXT NOT NULL, "+
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT epicNames UNIQUE (backendType,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS status ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  projectId           INTEGER, "+
                                                    "  name                TEXT NOT NULL, "+
                                                    "  description         TEXT NOT NULL, "+
                                                    "  isDefault           INTEGER DEFAULT 0, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT status UNIQUE (backendType,projectId,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS statusIndex0 ON status(backendType,projectId,name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS statusIndex1 ON status(name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS statusIndex2 ON status(backendType,isDefault)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS priorities ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  projectId           INTEGER, "+
                                                    "  name                TEXT NOT NULL, "+
                                                    "  description         TEXT NOT NULL, "+
                                                    "  isDefault           INTEGER, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT priorities UNIQUE (backendType,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS prioritiesIndex0 ON priorities(backendType,projectId,name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS prioritiesIndex1 ON priorities(name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS prioritiesIndex2 ON priorities(backendType,isDefault)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS activities ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    "  projectId           INTEGER, "+
                                                    "  name                TEXT NOT NULL, "+
                                                    "  description         TEXT NOT NULL, "+
                                                    "  isDefault           INTEGER DEFAULT 0, "+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT activities UNIQUE (backendType,projectId,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS activitiesIndex0 ON activities(backendType,projectId,name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS activitiesIndex1 ON activities(name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS projectsLocal ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    "  parentId            INTEGER, "+
                                                    ""+
                                                    "  shortName           TEXT, "+
                                                    "  name                TEXT NOT NULL, "+
                                                    "  status              INTEGER, "+
                                                    "  identifier          TEXT, "+
                                                    "  description         TEXT NOT NULL, "+
                                                    "  createdOn           INTEGER, "+
                                                    "  updatedOn           INTEGER, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT projects UNIQUE (backendType,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS projectsModified ( "+
                                                    "  id            INTEGER PRIMARY KEY, "+
                                                    "  projectId     INTEGER NOT NULL REFERENCES projects(id) ON DELETE CASCADE, "+  // local issue id
                                                    "  modifiedState INTEGER DEFAULT "+Backend.ModifiedStates.NONE.ordinal()+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();

      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS projectsIndex0 ON projectsLocal(name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS projectsIndex1 ON projectsLocal(status)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS projectsIndex2 ON projectsModified(projectId,modifiedState)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS members ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  projectId           INTEGER, "+
                                                    "  userId              INTEGER, "+
                                                    "  firstName           TEXT, "+
                                                    "  lastName            TEXT, "+
                                                    "  name                TEXT, "+
                                                    "  roleId              INTEGER, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT versions UNIQUE (backendType,projectId,userId,firstName,lastName)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS membersIndex0 ON members(backendType,projectId,userId,name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS versions ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    "  projectId           INTEGER, "+
                                                    "  name                TEXT, "+
                                                    "  description         TEXT, "+
                                                    "  status              TEXT, "+
                                                    "  dueDate             INTEGER, "+
                                                    "  createdOn           INTEGER, "+
                                                    "  updatedOn           INTEGER, "+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT versions UNIQUE (backendType,projectId,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS versionsIndex0 ON versions(backendType,projectId,name)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS issuesLocal ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id (Note: no primary key/constraint because all new issues have 0)
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  parentId            INTEGER, "+
                                                    "  shortName           TEXT, "+
                                                    "  projectId           INTEGER, "+
                                                    "  trackerId           INTEGER, "+
                                                    "  epicName            TEXT DEFAULT '', "+  // not used for Redmine
                                                    "  statusId            INTEGER, "+
                                                    "  statusTransitionId  INTEGER DEFAULT "+Id.NONE+", "+
                                                    "  priorityId          INTEGER, "+
                                                    "  authorId            INTEGER, "+
                                                    "  summary             TEXT NOT NULL, "+
                                                    "  description         TEXT NOT NULL, "+
                                                    "  startDate           INTEGER, "+
                                                    "  dueDate             INTEGER, "+
                                                    "  doneRatio           INTEGER, "+
                                                    "  assignedToMemberId  INTEGER, "+
                                                    "  targetVersionId     INTEGER, "+
                                                    "  estimatedDuration   INTEGER, "+  // [min]
                                                    "  createdOn           INTEGER, "+
                                                    "  updatedOn           INTEGER, "+
                                                    "  closedOn            INTEGER, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')), "+
                                                    ""+
// TODO: remove newProjectId
                                                    "  newProjectId        INTEGER "+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS issuesModified ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+
                                                    "  backendType         INTEGER, "+
                                                    "  issueId             INTEGER, "+  // local issue id or 0
                                                    "  modifiedState       INTEGER DEFAULT "+Backend.ModifiedStates.NONE.ordinal()+","+
                                                    ""+
                                                    "  shortName           TEXT, "+
                                                    "  projectId           INTEGER, "+
                                                    "  trackerId           INTEGER, "+
                                                    "  epicName            TEXT DEFAULT '', "+  // not used for Redmine
                                                    "  statusId            INTEGER, "+
                                                    "  statusTransitionId  INTEGER DEFAULT "+Id.NONE+", "+
                                                    "  priorityId          INTEGER, "+
                                                    "  authorId            INTEGER, "+
                                                    "  summary             TEXT NOT NULL, "+
                                                    "  description         TEXT NOT NULL, "+
                                                    "  startDate           INTEGER, "+
                                                    "  dueDate             INTEGER, "+
                                                    "  doneRatio           INTEGER, "+
                                                    "  assignedToMemberId  INTEGER, "+
                                                    "  targetVersionId     INTEGER, "+
                                                    "  estimatedDuration   INTEGER, "+  // [min]
                                                    "  createdOn           INTEGER, "+
                                                    "  updatedOn           INTEGER, "+
                                                    "  closedOn            INTEGER, "+
                                                    ""+
// TODO: remove newProjectId
                                                    "  newProjectId  INTEGER DEFAULT "+Id.NONE+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS statusTransitions ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  issueId             INTEGER, "+
                                                    "  name                TEXT, "+
                                                    "  description         TEXT, "+
                                                    "  statusId            INTEGER, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')),"+
                                                    ""+
                                                    "  CONSTRAINT statusTransitions UNIQUE (backendType,issueId,name)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE INDEX IF NOT EXISTS statusTransitionsIndex0 ON statusTransitions(backendType,backendId,issueId)");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS timeEntriesLocal ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+  // local database id (Note: no primary key/constraint because all new issues have 0)
                                                    "  backendType         INTEGER, "+
                                                    "  backendId           INTEGER, "+
                                                    ""+
                                                    "  projectId           INTEGER, "+
                                                    "  issueId             INTEGER, "+
                                                    "  userId              INTEGER, "+
                                                    "  activityId          INTEGER, "+
                                                    "  duration            INTEGER, "+  // [min]
                                                    "  comments            TEXT NOT NULL, "+
                                                    "  spentOn             INTEGER, "+
                                                    "  createdOn           INTEGER, "+
                                                    "  updatedOn           INTEGER, "+
                                                    ""+
                                                    "  lastUpdateTimestamp INTEGER DEFAULT (STRFTIME('%s','now')) "+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS timeEntriesModified ( "+
                                                    "  id                  INTEGER PRIMARY KEY, "+
                                                    "  timeEntryId         INTEGER, "+  // local issue id or 0
                                                    "  modifiedState       INTEGER DEFAULT "+Backend.ModifiedStates.NONE.ordinal()+","+
                                                    ""+
                                                    "  projectId           INTEGER, "+  // new project id
                                                    "  issueId             INTEGER, "+  // new issue id
                                                    "  userId              INTEGER, "+  // new user id
                                                    "  activityId          INTEGER, "+  // new activity id
                                                    "  duration            INTEGER, "+  // new duration [min]
                                                    "  comments            TEXT NOT NULL, "+  // comment
                                                    "  spentOn             INTEGER, "+  // new spent on
                                                    "  createdOn           INTEGER, "+  // new created on
                                                    "  updatedOn           INTEGER "+   // new updated on
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }

    database.setForeignKeys(true);
  }

  /** init local database views
   */
  private static void initLocalDatabaseViews()
    throws SQLException
  {
    PreparedStatement preparedStatement = null;
    try
    {
// TODO: view from local+modified
      preparedStatement = database.prepareStatement(
" CREATE TEMPORARY VIEW projects AS "+
" SELECT * FROM projectsLocal"+
""
);
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement(" CREATE VIEW IF NOT EXISTS issues AS "+
                                                    " SELECT "+
                                                    "  issuesLocal.backendType AS backendType,"+
                                                    "  issuesLocal.id          AS id,"+
                                                    "  issuesLocal.backendId   AS backendId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.id                 ELSE "+Id.NONE+"                         END) AS modifiedId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.modifiedState      ELSE "+ModifiedStates.NONE.ordinal()+"   END) AS modifiedState,"+
                                                    "  issuesLocal.parentId AS parentId,"+
//                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.parentId           ELSE issuesLocal.parentId           END) AS parentId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.shortName          ELSE issuesLocal.shortName          END) AS shortName,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.projectId          ELSE issuesLocal.projectId          END) AS projectId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.trackerId          ELSE issuesLocal.trackerId          END) AS trackerId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.epicName           ELSE issuesLocal.epicName           END) AS epicName,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.statusId           ELSE issuesLocal.statusId           END) AS statusId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.statusTransitionId ELSE issuesLocal.statusTransitionId END) AS statusTransitionId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.priorityId         ELSE issuesLocal.priorityId         END) AS priorityId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.authorId           ELSE issuesLocal.authorId           END) AS authorId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.summary            ELSE issuesLocal.summary            END) AS summary,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.description        ELSE issuesLocal.description        END) AS description,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.startDate          ELSE issuesLocal.startDate          END) AS startDate,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.dueDate            ELSE issuesLocal.dueDate            END) AS dueDate,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.doneRatio          ELSE issuesLocal.doneRatio          END) AS doneRatio,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.assignedToMemberId ELSE issuesLocal.assignedToMemberId END) AS assignedToMemberId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.targetVersionId    ELSE issuesLocal.targetVersionId    END) AS targetVersionId,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.estimatedDuration  ELSE issuesLocal.estimatedDuration  END) AS estimatedDuration,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.createdOn          ELSE issuesLocal.createdOn          END) AS createdOn,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.updatedOn          ELSE issuesLocal.updatedOn          END) AS updatedOn,"+
                                                    "  (CASE WHEN issuesModified.id THEN issuesModified.closedOn           ELSE issuesLocal.closedOn           END) AS closedOn,"+
                                                    "  issuesLocal.lastUpdateTimestamp AS lastUpdateTimestamp"+
                                                    " FROM issuesLocal "+
                                                    "   LEFT JOIN issuesModified ON issuesModified.issueId=issuesLocal.id"+
                                                    " WHERE (issuesModified.id ISNULL) OR (issuesModified.modifiedState!="+ModifiedStates.DELETED.ordinal()+")"+
                                                    ""
                                                    );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement(" CREATE VIEW IF NOT EXISTS timeEntries AS "+
                                                    " SELECT "+
                                                    "  timeEntriesLocal.backendType AS backendType,"+
                                                    "  timeEntriesLocal.id          AS id,"+
                                                    "  timeEntriesLocal.backendId   AS backendId,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.id            ELSE "+Id.NONE+"                       END) AS modifiedId,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.modifiedState ELSE "+ModifiedStates.NONE.ordinal()+" END) AS modifiedState,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.projectId     ELSE timeEntriesLocal.projectId        END) AS projectId,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.issueId       ELSE timeEntriesLocal.issueId          END) AS issueId,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.userId        ELSE timeEntriesLocal.userId           END) AS userId,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.activityId    ELSE timeEntriesLocal.activityId       END) AS activityId,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.duration      ELSE timeEntriesLocal.duration         END) AS duration,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.comments      ELSE timeEntriesLocal.comments         END) AS comments,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.spentOn       ELSE timeEntriesLocal.spentOn          END) AS spentOn,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.createdOn     ELSE timeEntriesLocal.createdOn        END) AS createdOn,"+
                                                    "  (CASE WHEN timeEntriesModified.id THEN timeEntriesModified.updatedOn     ELSE timeEntriesLocal.updatedOn        END) AS updatedOn,"+
                                                    "  timeEntriesLocal.lastUpdateTimestamp AS lastUpdateTimestamp"+
                                                    " FROM timeEntriesLocal "+
                                                    "   LEFT JOIN timeEntriesModified ON timeEntriesModified.timeEntryId=timeEntriesLocal.id"+
                                                    " WHERE (timeEntriesModified.id ISNULL) OR (timeEntriesModified.modifiedState!="+ModifiedStates.DELETED.ordinal()+")"+
                                                    ""
                                                    );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      // debug only: logging triggers
      {
        preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS log (text TEXT);"
                                                     );
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;
      }
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }
  }

  //--------------------------------------------------------------------

  /** update users in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseUsers(final BackendProgress backendProgress)
    throws BackendException,InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get users for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getUsers(backendProgress,new EntityHandler<User>()
        {
          @Override
          public void handle(User user)
            throws BackendException
          {
            insertLocalDatabase("users",
                                backendDriver.getType(),
                                user.id,
                                condition("(backendType=?) AND (name=?)",
                                          user.backendType,
                                          user.name
                                         ),
                                new Values
                                (
                                 "firstName",  user.firstName,
                                 "lastName",   user.lastName,
                                 "name",       user.name,
                                 "roleId",     user.roleId,
                                 "mail",       user.mail,
                                 "createdOn",  user.createdOn,
                                 "lastLoginOn",user.lastLoginOn
                                )
                               );
          }
          @Override
          public void done(User user)
            throws BackendException
          {
            updateLocalDatabase("users",
                                user.backendType,
                                user.id,
                                new Values
                                (
                                 "firstName",  user.firstName,
                                 "lastName",   user.lastName,
                                 "name",       user.name,
                                 "roleId",     user.roleId,
                                 "mail",       user.mail,
                                 "createdOn",  user.createdOn,
                                 "lastLoginOn",user.lastLoginOn
                                )
                               );
            idList.add(user.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("users",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("projects");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update roles in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseRoles(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get roles for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getRoles(backendProgress,new EntityHandler<Role>()
        {
          @Override
          public void handle(Role role)
            throws BackendException
          {
            insertLocalDatabase("roles",
                                backendDriver.getType(),
                                role.id,
                                condition("(backendType=?) AND (projectId=?) AND (name=?)",
                                          role.backendType,
                                          role.projectId,
                                          role.name
                                         ),
                                new Values
                                (
                                 "projectId",  role.projectId,
                                 "name",       role.name,
                                 "description",role.description
                                )
                               );
          }
          @Override
          public void done(Role role)
            throws BackendException
          {
            updateLocalDatabase("roles",
                                role.backendType,
                                role.id,
                                new Values
                                (
                                 "projectId",  role.projectId,
                                 "name",       role.name,
                                 "description",role.description
                                )
                               );
            idList.add(role.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("roles",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("projects");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update trackers in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseTrackers(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get trackers for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getTrackers(backendProgress,new EntityHandler<Tracker>()
        {
          @Override
          public void handle(Tracker tracker)
            throws BackendException
          {
            insertLocalDatabase("trackers",
                                backendDriver.getType(),
                                tracker.id,
                                condition("(backendType=?) AND (projectId=?) AND (name=?)",
                                          tracker.backendType,
                                          tracker.projectId,
                                          tracker.name
                                         ),
                                new Values
                                (
                                 "projectId",      tracker.projectId,
                                 "name",           tracker.name,
                                 "description",    tracker.description,
                                 "requireEpicName",tracker.requireEpicName
                                )
                               );
          }
          @Override
          public void done(Tracker tracker)
            throws BackendException
          {
            updateLocalDatabase("trackers",
                                tracker.backendType,
                                tracker.id,
                                new Values
                                (
                                 "projectId",      tracker.projectId,
                                 "name",           tracker.name,
                                 "description",    tracker.description,
                                 "requireEpicName",tracker.requireEpicName
                                )
                               );
            idList.add(tracker.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("trackers",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("trackers");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update epic names in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseEpicNames(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get epic names for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getEpicNames(backendProgress,new EntityHandler<EpicName>()
        {
          @Override
          public void handle(EpicName epicName)
            throws BackendException
          {
            insertLocalDatabase("epicNames",
                                backendDriver.getType(),
                                epicName.id,
                                condition("(backendType=?) AND (name=?)",
                                          epicName.backendType,
                                          epicName.name
                                         ),
                                new Values
                                (
                                 "name",epicName.name
                                )
                               );
          }
          @Override
          public void done(EpicName epicName)
            throws BackendException
          {
            updateLocalDatabase("epicNames",
                                epicName.backendType,
                                epicName.id,
                                new Values
                                (
                                 "name",epicName.name
                                )
                               );
            idList.add(epicName.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("epicNames",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("trackers");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update status in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseStatus(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get status for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getStatus(backendProgress,new EntityHandler<Status>()
        {
          @Override
          public void handle(Status status)
            throws BackendException
          {
            insertLocalDatabase("status",
                                backendDriver.getType(),
                                status.id,
                                condition("(backendType=?) AND (projectId=?) AND (name=?)",
                                          status.backendType,
                                          status.projectId,
                                          status.name
                                         ),
                                new Values
                                (
                                 "projectId",  status.projectId,
                                 "name",       status.name,
                                 "description",status.description
                                )
                               );
          }
          @Override
          public void done(Status status)
            throws BackendException
          {
            updateLocalDatabase("status",
                                status.backendType,
                                status.id,
                                new Values
                                (
                                 "projectId",  status.projectId,
                                 "name",       status.name,
                                 "description",status.description
                                )
                               );
            idList.add(status.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("status",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("status");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update status transition in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseStatusTransition(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get status for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getStatus(backendProgress,new EntityHandler<Status>()
        {
          @Override
          public void handle(Status status)
            throws BackendException
          {
            insertLocalDatabase("status",
                                backendDriver.getType(),
                                status.id,
                                condition("(backendType=?) AND (projectId=?) AND (name=?)",
                                          status.backendType,
                                          status.name
                                         ),
                                new Values
                                (
                                 "name",       status.name,
                                 "description",status.description
                                )
                               );
          }
          @Override
          public void done(Status status)
            throws BackendException
          {
            updateLocalDatabase("status",
                                status.backendType,
                                status.id,
                                new Values
                                (
                                 "name",       status.name,
                                 "description",status.description
                                )
                               );
            idList.add(status.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("status",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("status");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update priorities in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabasePriorities(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get priorities for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getPriorities(backendProgress,new EntityHandler<Priority>()
        {
          @Override
          public void handle(Priority priority)
            throws BackendException
          {
            insertLocalDatabase("priorities",
                                backendDriver.getType(),
                                priority.id,
                                condition("(backendType=?) AND (projectId=?) AND (name=?)",
                                          priority.backendType.ordinal(),
                                          priority.projectId,
                                          priority.name
                                         ),
                                new Values
                                (
                                 "projectId",  priority.projectId,
                                 "name",       priority.name,
                                 "description",priority.description,
                                 "isDefault",  priority.isDefault
                                )
                               );
          }
          @Override
          public void done(Priority priority)
            throws BackendException
          {
            updateLocalDatabase("priorities",
                                priority.backendType,
                                priority.id,
                                new Values
                                (
                                 "projectId",  priority.projectId,
                                 "name",       priority.name,
                                 "description",priority.description,
                                 "isDefault",  priority.isDefault
                                )
                               );
            idList.add(priority.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("priorities",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("priorities");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update activities in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseActivities(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get activities for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getActivities(backendProgress,new EntityHandler<Activity>()
        {
          @Override
          public void handle(Activity activity)
            throws BackendException
          {
            insertLocalDatabase("activities",
                                backendDriver.getType(),
                                activity.id,
                                condition("(backendType=?) AND (projectId=?) AND (name=?)",
                                          activity.backendType,
                                          activity.projectId,
                                          activity.name
                                         ),
                                new Values
                                (
                                 "projectId",  activity.projectId,
                                 "name",       activity.name,
                                 "description",activity.description,
                                 "isDefault",  activity.isDefault
                                )
                               );
          }
          @Override
          public void done(Activity activity)
            throws BackendException
          {
            updateLocalDatabase("activities",
                                activity.backendType,
                                activity.id,
                                new Values
                                (
                                 "projectId",  activity.projectId,
                                 "name",       activity.name,
                                 "description",activity.description,
                                 "isDefault",  activity.isDefault
                                )
                               );
            idList.add(activity.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("activities",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("activities");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update projects in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseProjects(final BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get projects for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getProjects(backendProgress,new EntityHandler<Project>()
        {
          @Override
          public void handle(Project project)
            throws BackendException
          {
            insertLocalDatabase("projectsLocal",
                                backendDriver.getType(),
                                project.id,
                                condition("(backendType=?) AND (backendId=?)",
                                          project.backendType,
                                          project.id.backend
                                         ),
                                new Values
                                (
                                 "parentId",   project.parentId,
                                 "shortName",  project.shortName,
                                 "name",       project.name,
                                 "status",     project.status.getValue(),
                                 "identifier", project.identifier,
                                 "description",project.description,
                                 "createdOn",  project.createdOn,
                                 "updatedOn",  project.updatedOn
                                )
                               );
          }
          @Override
          public void done(Project project)
            throws BackendException
          {
            updateLocalDatabase("projectsLocal",
                                project.backendType,
                                project.id,
                                new Values
                                (
                                 "parentId",   project.parentId,
                                 "shortName",  project.shortName,
                                 "name",       project.name,
                                 "status",     project.status.getValue(),
                                 "identifier", project.identifier,
                                 "description",project.description,
                                 "createdOn",  project.createdOn,
                                 "updatedOn",  project.updatedOn
                                )
                               );
            idList.add(project.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("projectsLocal",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("projects");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update all members in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseMembers(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get members for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getMembers(backendProgress,new EntityHandler<Member>()
        {
          @Override
          public void handle(Member member)
            throws BackendException
          {
            insertLocalDatabase("members",
                                backendDriver.getType(),
                                member.id,
                                condition("(backendType=?) AND (projectId=?) AND (userId=?)",
                                          member.backendType,
                                          member.projectId,
                                          member.userId
                                         ),
                                new Values
                                (
                                 "projectId",member.projectId,
                                 "userId",   member.userId,
                                 "firstName",member.firstName,
                                 "lastName", member.lastName,
                                 "name",     member.name,
                                 "roleId",   member.roleId
                                )
                               );
          }
          @Override
          public void done(Member member)
            throws BackendException
          {
            updateLocalDatabase("members",
                                member.backendType,
                                member.id,
                                new Values
                                (
                                 "projectId",member.projectId,
                                 "userId",   member.userId,
                                 "firstName",member.firstName,
                                 "lastName", member.lastName,
                                 "roleId",   member.roleId
                                )
                               );
            idList.add(member.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("members",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("members");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update all versions in local database
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseVersions(final BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get versions for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        // insert entries
        final IdList idList = new IdList();
        backendDriver.getVersions(backendProgress,new EntityHandler<Version>()
        {
          @Override
          public void handle(Version version)
            throws BackendException
          {
            insertLocalDatabase("versions",
                                backendDriver.getType(),
                                version.id,
                                condition("(backendType=?) AND (projectId=?) AND (name=?)",
                                          version.backendType,
                                          version.projectId,
                                          version.name
                                         ),
                                new Values
                                (
                                 "projectId",  version.projectId,
                                 "name",       version.name,
                                 "description",version.description,
                                 "status",     version.status,
                                 "dueDate",    version.dueDate,
                                 "createdOn",  version.createdOn,
                                 "updatedOn",  version.updatedOn
                                )
                               );
          }
          @Override
          public void done(Version version)
            throws BackendException
          {
            updateLocalDatabase("versions",
                                version.backendType,
                                version.id,
                                new Values
                                (
                                 "projectId",  version.projectId,
                                 "name",       version.name,
                                 "description",version.description,
                                 "status",     version.status,
                                 "dueDate",    version.dueDate,
                                 "createdOn",  version.createdOn,
                                 "updatedOn",  version.updatedOn
                                )
                               );
            idList.add(version.id);
          }
        });

        // delete unknown entries in local database
        cleanLocalDatabase("versions",backendDriver.getType(),idList);

        backendProgress.incrementStep();
      }

      setLastUpdate("status");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update issues in local database
   * @param incrementalFlag true for incremental update
   * @param maxIssuesHistory max. issues history or 0
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseIssues(boolean incrementalFlag, int maxIssuesHistory, BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get issues for all backend drivers
      int i = 0;
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        final IdList issueIdList            = new IdList();
        final IdList statusTransitionIdList = new IdList();

        {
          backendProgress.beginSteps(3);
        }
        try
        {
          // get history timestamp/last update timestamp
          long lastUpdatedOnTimestamp = 0;
          backendProgress.setStep(0);
          if      (incrementalFlag)
          {
            // get last known updated timestamp
            try
            {
              lastUpdatedOnTimestamp = getLocalDatabase("issues",
                                                        null,  // withStatement
                                                        "MAX(updatedOn)",
                                                        Condition.NONE,
                                                        Filter.NONE,
                                                        0L
                                                       );
            }
            catch (BackendException exception)
            {
              // ignored
              Timemine.printInternalError(exception);
            }

            // at least 3 days history
            long nowTimestamp = System.currentTimeMillis()/1000;
            if (lastUpdatedOnTimestamp > (nowTimestamp-3*24*60*60))
            {
              lastUpdatedOnTimestamp = nowTimestamp-3*24*60*60;
            }
//Dprintf.dprintf("lastUpdatedOnTimestamp=%d",lastUpdatedOnTimestamp);
          }
          else if (maxIssuesHistory > 0)
          {
            lastUpdatedOnTimestamp = System.currentTimeMillis()/1000-maxIssuesHistory*24*60*60;
          }

          // update issues
          backendProgress.setStep(1);
          backendDriver.getIssues(backendProgress,lastUpdatedOnTimestamp,new EntityHandler<Issue>()
          {
            @Override
            public void handle(Issue issue)
              throws BackendException
            {
              // check if locally modified entry
              ModifiedStates modifiedState = getLocalDatabase(new DataBaseHandler<ModifiedStates>()
                                                              {
                                                                @Override
                                                                public String selectStatement(String additionalFilter)
                                                                {
                                                                  return " SELECT *"+
                                                                         " FROM issuesModified"+
                                                                         ((additionalFilter != null) ? " WHERE "+additionalFilter : "");
                                                                }
                                                                @Override
                                                                public ModifiedStates get(ValueSet valueSet)
                                                                  throws SQLException
                                                                {
                                                                  return valueSet.getEnum("modifiedState",Backend.ModifiedStates.class);
                                                                }
                                                              },
                                                              condition("id=?",
                                                                        issue.id.local
                                                                       ),
                                                              filter("ORDER BY id DESC LIMIT 0,1")
                                                             );
              if (modifiedState == null)
              {
                insertLocalDatabase("issuesLocal",
                                    backendDriver.getType(),
                                    issue.id,
                                    condition("(backendType=?) AND (backendId=?)",
                                              issue.backendType,
                                              issue.id.backend
                                             ),
                                    new Values
                                    (
                                     "parentId",          issue.parentId,
                                     "shortName",         issue.shortName,
                                     "projectId",         issue.projectId,
                                     "trackerId",         issue.trackerId,
                                     "epicName",          issue.epicName,
                                     "statusId",          issue.statusId,
                                     "priorityId",        issue.priorityId,
                                     "authorId",          issue.authorId,
                                     "summary",           issue.summary,
                                     "description",       issue.description,
                                     "startDate",         issue.startDate,
                                     "dueDate",           issue.dueDate,
                                     "doneRatio",         issue.doneRatio,
                                     "assignedToMemberId",issue.assignedToMemberId,
                                     "targetVersionId",   issue.targetVersionId,
                                     "estimatedDuration", issue.estimatedDuration,
                                     "createdOn",         issue.createdOn,
                                     "updatedOn",         issue.updatedOn,
                                     "closedOn",          issue.closedOn
                                    )
                                   );
              }
            }
            @Override
            public void done(Issue issue)
              throws BackendException
            {
              updateLocalDatabase("issuesLocal",
                                  issue.backendType,
                                  issue.id,
                                  new Values
                                  (
                                   "parentId",          issue.parentId,
                                   "shortName",         issue.shortName,
                                   "projectId",         issue.projectId,
                                   "trackerId",         issue.trackerId,
                                   "epicName",          issue.epicName,
                                   "statusId",          issue.statusId,
                                   "priorityId",        issue.priorityId,
                                   "authorId",          issue.authorId,
                                   "summary",           issue.summary,
                                   "description",       issue.description,
                                   "startDate",         issue.startDate,
                                   "dueDate",           issue.dueDate,
                                   "doneRatio",         issue.doneRatio,
                                   "assignedToMemberId",issue.assignedToMemberId,
                                   "targetVersionId",   issue.targetVersionId,
                                   "estimatedDuration", issue.estimatedDuration,
                                   "createdOn",         issue.createdOn,
                                   "updatedOn",         issue.updatedOn,
                                   "closedOn",          issue.closedOn
                                  )
                                 );
              issueIdList.add(issue.id);
            }
          });
//Dprintf.dprintf("issueIdList:"); for (Id issueId : issueIdList) Dprintf.dprintf("id=%s",issueId);

          // update issues status transitions
          backendProgress.setStep(2);
          int j = 0;
Dprintf.dprintf("issueIdList=%d\n",issueIdList.size());
          for (Id issueId : issueIdList)
          {
            backendDriver.getStatusTransitions(backendProgress,issueId,new EntityHandler<StatusTransition>()
            {
              @Override
              public void handle(StatusTransition statusTransition)
                throws BackendException
              {
                insertLocalDatabase("statusTransitions",
                                    backendDriver.getType(),
                                    statusTransition.id,
                                    condition("(backendType=?) AND (issueId=?) AND (name=?)",
                                              statusTransition.backendType,
                                              statusTransition.issueId.backend,
                                              statusTransition.name
                                             ),
                                    new Values
                                    (
                                     "issueId",    statusTransition.issueId,
                                     "name",       statusTransition.name,
                                     "description",statusTransition.description,
                                     "statusId",   statusTransition.statusId
                                    )
                                   );
              }
              @Override
              public void done(StatusTransition statusTransition)
                throws BackendException
              {
                updateLocalDatabase("statusTransitions",
                                    statusTransition.backendType,
                                    statusTransition.id,
                                    new Values
                                    (
                                     "issueId",    statusTransition.issueId,
                                     "name",       statusTransition.name,
                                     "description",statusTransition.description,
                                     "statusId",   statusTransition.statusId
                                    )
                                   );
                statusTransitionIdList.add(statusTransition.id);
              }
            });

            // throttle execution
            try { Thread.sleep(Settings.throttleTime); } catch (InterruptedException exception) { /* ignored */ }

            // next
            backendProgress.setSelection(j,issueIdList.size());
            j++;
          }
Dprintf.dprintf("--------------------\n");

        }
        finally
        {
          backendProgress.endSteps();
        }

        if (!incrementalFlag && (maxIssuesHistory <= 0))
        {
        // delete unknown entries in local database
//Dprintf.dprintf("issueIdList=%s",issueIdList.size());
//Dprintf.dprintf("statusTransitionIdList=%s",statusTransitionIdList.size());
          cleanLocalDatabase("issuesLocal",backendDriver.getType(),issueIdList);
          cleanLocalDatabase("statusTransitions",backendDriver.getType(),statusTransitionIdList);
        }

        backendProgress.incrementStep();
        i++;
      }

      setLastUpdate("issuesLocal");
      setLastUpdate("statusTransitions");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  /** update time entries in local database
   * @param incrementalFlag true for incremental update
   * @param maxTimeEntriesHistory max. history or 0 [days]
   * @param backendProgress progress handler or null
   */
  public static void updateLocalDatabaseTimeEntries(boolean incrementalFlag, int maxTimeEntriesHistory, BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    {
      updateLock();
      backendProgress.beginSteps(backendDrivers.size());
    }
    try
    {
      // get time entries for all backend drivers
      for (final BackendDriver backendDriver : backendDrivers.values())
      {
        final IdList idList = new IdList();

        {
          backendProgress.beginSteps(3);
        }
        try
        {
          // get history timestamp/last update timestamp
          backendProgress.setStep(0);
          long lastCreatedOnTimestamp = 0;
          if      (incrementalFlag)
          {
            // get last known updated timestamp
            try
            {
              lastCreatedOnTimestamp = getLocalDatabase("issues",
                                                        "MAX(createdOn)",
                                                        0L
                                                       );
            }
            catch (BackendException exception)
            {
              // ignored
              Timemine.printInternalError(exception);
            }
//Dprintf.dprintf("lastCreatedOnTimestam=%d",lastCreatedOnTimestamp);

            // at least 3 days history
            long nowTimestamp = System.currentTimeMillis()/1000;
            if (lastCreatedOnTimestamp > (nowTimestamp-3*24*60*60))
            {
              lastCreatedOnTimestamp = nowTimestamp-3*24*60*60;
            }
//Dprintf.dprintf("lastCreatedOnTimestamp=%d",lastCreatedOnTimestamp);
          }
          else if (maxTimeEntriesHistory > 0)
          {
            lastCreatedOnTimestamp = System.currentTimeMillis()/1000-maxTimeEntriesHistory*24*60*60;
          }

          // remove time entries to update from local database
          backendProgress.setStep(1);
          removeLocalDatabase("timeEntriesLocal",
                              condition("(backendType=?) AND (createdOn>=?)",
                                        backendDriver.getType(),
                                        lastCreatedOnTimestamp
                                       )
                             );

          // get and insert entries into local database
          backendProgress.setStep(2);
          backendDriver.getTimeEntries(backendProgress,lastCreatedOnTimestamp,new EntityHandler<TimeEntry>()
          {
            @Override
            public void handle(TimeEntry timeEntry)
              throws BackendException
            {
//Dprintf.dprintf("timeEntry.createdOn=%s",timeEntry.createdOn);
//Dprintf.dprintf("timeEntry.updatedOn=%s",timeEntry.updatedOn);
              insertLocalDatabase("timeEntriesLocal",
                                  backendDriver.getType(),
                                  timeEntry.id,
                                  condition("(backendType=?) AND (backendId=?)",
                                            timeEntry.backendType,
                                            timeEntry.id.backend
                                           ),
                                  new Values
                                  (
                                   "projectId", timeEntry.projectId,
                                   "issueId",   timeEntry.issueId,
                                   "userId",    timeEntry.userId,
                                   "activityId",timeEntry.activityId,
                                   "duration",  timeEntry.duration,
                                   "comments",  timeEntry.comments,
                                   "spentOn",   timeEntry.spentOn,
                                   "createdOn", timeEntry.createdOn,
                                   "updatedOn", timeEntry.updatedOn
                                  )
                                 );
            }
            @Override
            public void done(TimeEntry timeEntry)
              throws BackendException
            {
//Dprintf.dprintf("timeEntry.createdOn=%s",timeEntry.createdOn);
//Dprintf.dprintf("timeEntry.updatedOn=%s",timeEntry.updatedOn);
              updateLocalDatabase("timeEntriesLocal",
                                  timeEntry.backendType,
                                  timeEntry.id,
                                  new Values
                                  (
                                   "projectId", timeEntry.projectId,
                                   "issueId",   timeEntry.issueId,
                                   "userId",    timeEntry.userId,
                                   "activityId",timeEntry.activityId,
                                   "duration",  timeEntry.duration,
                                   "comments",  timeEntry.comments,
                                   "spentOn",   timeEntry.spentOn,
                                   "createdOn", timeEntry.createdOn,
                                   "updatedOn", timeEntry.updatedOn
                                  )
                                 );
              idList.add(timeEntry.id);
            }
          });
        }
        finally
        {
          backendProgress.endSteps();
        }

        if (!incrementalFlag && (maxTimeEntriesHistory <= 0))
        {
          // delete unknown entries in local database
          cleanLocalDatabase("timeEntriesLocal",backendDriver.getType(),idList);
        }

        backendProgress.incrementStep();
      }

      setLastUpdate("timeEntriesLocal");
    }
    catch (SQLException exception)
    {
      throw new BackendException(exception);
    }
    finally
    {
      backendProgress.endSteps();
      updateUnlock();
    }
  }

  //--------------------------------------------------------------------

  /** convert house/minutes fraction into duration
   * @param hourFraction spent hours fraction
   * @param minuteFraction spent minutes fraction
   * @return minutes
   */
  protected static long toDuration(int hourFraction, int minuteFraction)
  {
    return (long)hourFraction*60+(long)minuteFraction;
  }

  //--------------------------------------------------------------------

  /** set value in prepared statement
   * @param preparedStatement prepared statement
   * @param index [0..n-1]
   * @param data data
   */
  private static void setPreparedStatement(PreparedStatement preparedStatement, int index, Object data)
    throws SQLException
  {
    assert index >= 0;

    if (data != null)
    {
      if      (data.getClass() == Integer.class)
      {
        int value = ((Integer)data).intValue();
        preparedStatement.setInt(1+index,value);
      }
      else if (data.getClass() == Long.class)
      {
        long value = ((Long)data).longValue();
        preparedStatement.setLong(1+index,value);
      }
      else if (data.getClass() == Double.class)
      {
        double value = ((Double)data).doubleValue();
        preparedStatement.setDouble(1+index,value);
      }
      else if (data.getClass() == Boolean.class)
      {
        boolean value = ((Boolean)data).booleanValue();
        preparedStatement.setBoolean(1+index,value);
      }
      else if (data.getClass() == String.class)
      {
        String value = (String)data;
        preparedStatement.setString(1+index,value);
      }
      else if (data.getClass() == Date.class)
      {
        // Note: convert ms from Epoch to s from Epoch
        Date value = (Date)data;
        preparedStatement.setLong(1+index,value.getTime()/1000);
      }
      else if (data.getClass() == Array.class)
      {
        Array value = (Array)data;
        preparedStatement.setArray(1+index,value);
      }
      else if (data.getClass().isArray())
      {
throw new InternalError("still not implemented");
      }
      else if (data.getClass().isEnum())
      {
        int value = ((Enum)data).ordinal();
        preparedStatement.setInt(1+index,value);
      }
      else if (data.getClass() == Id.class)
      {
        // Note: always use backend id
        Id value = (Id)data;
        preparedStatement.setInt(1+index,!value.isAny() ? value.backend : Id.ANY);
      }
      else if (data.getClass() == SpentOn.class)
      {
        // Note: convert ms from Epoch to s from Epoch
        SpentOn value = (SpentOn)data;
        preparedStatement.setLong(1+index,value.getTime()/1000);
      }
      else
      {
        throw new IllegalArgumentException("unsupported data type '"+data.getClass()+"' for "+data+" at "+index);
      }
    }
    else
    {
      preparedStatement.setString(1+index,null);
    }
  }

  /** get last update timestamp
   * @param table table name
   * @return last timestamp [s]
   */
  private static long getLastUpdate(String tableName)
    throws BackendException
  {
    long timestamp = 0;

    synchronized(database)
    {
      PreparedStatement preparedStatement = null;
      ResultSet         resultSet         = null;
      try
      {
        preparedStatement = database.prepareStatement("SELECT timestamp FROM lastUpdates WHERE name=?");
        preparedStatement.setString(1,tableName);
        resultSet = preparedStatement.executeQuery();
        if (resultSet.next())
        {
          timestamp = resultSet.getLong("timestamp");
        }
        resultSet.close(); resultSet = null;
        preparedStatement.close(); preparedStatement = null;
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }

    return timestamp;
  }

  /** set last update timestamp
   * @param table table nametablename
   */
  private static void setLastUpdate(String tableName)
  {
    synchronized(database)
    {
      PreparedStatement preparedStatement;
      ResultSet         resultSet;

      // check if entry exists
      boolean existsFlag = false;
      preparedStatement = null;
      resultSet         = null;
      try
      {
        preparedStatement = database.prepareStatement("SELECT id FROM lastUpdates WHERE name=?");
        preparedStatement.setString(1,tableName);
        resultSet = preparedStatement.executeQuery();
        if (resultSet.next()) existsFlag = true;
        resultSet.close(); resultSet = null;
        preparedStatement.close(); preparedStatement = null;
      }
      catch (SQLException exception)
      {
        // ignored
      }
      finally
      {
        try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }

      String sqlCommand;

      // execute SQL command
      preparedStatement = null;
      try
      {
        if (existsFlag)
        {
          // create SQL update command
          preparedStatement = database.prepareStatement("UPDATE lastUpdates SET timestamp=? WHERE name=?;");
          preparedStatement.setLong(1,System.currentTimeMillis()/1000);
          preparedStatement.setString(2,tableName);
        }
        else
        {
          // create SQL insert command
          preparedStatement = database.prepareStatement("INSERT INTO lastUpdates (name,timestamp) VALUES (?,?);");
          preparedStatement.setString(1,tableName);
          preparedStatement.setLong(2,System.currentTimeMillis()/1000);
        }
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;
      }
      catch (SQLException exception)
      {
        // ignored
      }
      finally
      {
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }
  }

  /** get last update delta time
   * @param table table name
   * @param resetFlag true to reset last update timestamp
   * @return last update delta time [s]
   */
  private static long getLastUpdateDelta(String tableName, boolean resetFlag)
    throws BackendException
  {
    long lastUpdate = getLastUpdate(tableName);

    if (resetFlag)
    {
      setLastUpdate(tableName);
    }

    return System.currentTimeMillis()/1000 - lastUpdate;
  }

  /** check last update time, output info
   * @param table table name
   */
  private static void checkLastUpdate(String tableName)
  {
    if (updateLock.tryLock())
    {
      long lastUpdateDelta = 0;
      try
      {
        lastUpdateDelta = getLastUpdateDelta(tableName,true);
      }
      catch (BackendException exception)
      {
        // ignored
        Timemine.printInternalError(exception);
      }
      finally
      {
        updateLock.unlock();
      }

      if ((maxUpdateTime > 0) && (lastUpdateDelta >= maxUpdateTime*60))
      {
        Timemine.printWarning("Local '%s' database update expired (%ds)!", tableName, lastUpdateDelta);
      }
    }
  }

  /** dump SQL command and explain query plan
   * @param sqlCommand SQL command
   * @param condition condition or Form.NONE
   * @param filter filter or Filter.NONE
   */
  private static void dumpSQLCommand(String    sqlCommand,
                                     Condition condition,
                                     Filter    filter
                                    )

  {
    Dprintf.printStackTrace();

    Dprintf.dprintf("SQL command: %s",sqlCommand);
    if (condition != Condition.NONE)
    {
      for (int i = 0; i < condition.arguments.length; i++)
      {
        Dprintf.dprintf("             %d: %s",i,condition.arguments[i]);
      }
    }

    Dprintf.dprintf("SQL query plan: %s",sqlCommand);
    synchronized(database)
    {
      PreparedStatement preparedStatement = null;
      ResultSet         resultSet         = null;
      try
      {
        // prepare SQL command
        preparedStatement = database.prepareStatement("EXPLAIN QUERY PLAN "+sqlCommand);
        int n = 0;
        if (condition != null)
        {
          for (int i = 0; i < condition.arguments.length; i++)
          {
            setPreparedStatement(preparedStatement,n,condition.arguments[i]);
            n++;
          }
        }
        if (filter != null)
        {
          for (int i = 0; i < filter.arguments.length; i++)
          {
            setPreparedStatement(preparedStatement,n,filter.arguments[i]);
            n++;
          }
        }

        // execute SQL command
        resultSet = preparedStatement.executeQuery();

        // get results
        while (resultSet.next())
        {
          Dprintf.dprintf("  %s %s %s %s",
                          resultSet.getString(1),
                          resultSet.getString(2),
                          resultSet.getString(3),
                          resultSet.getString(4)
                         );
        }

        // done
        resultSet.close(); resultSet = null;
        preparedStatement.close(); preparedStatement = null;
      }
      catch (SQLException exception)
      {
        Timemine.printError("SQL command: %s",sqlCommand);
      }
      finally
      {
        try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }
  }

  /** dump SQL command and explain query plan
   * @param sqlCommand SQL command
   * @param condition condition or Form.NONE
   */
  private static void dumpSQLCommand(String    sqlCommand,
                                     Condition condition
                                    )

  {
    dumpSQLCommand(sqlCommand,condition,(Filter)null);
  }

  /** dump SQL command and explain query plan
   * @param sqlCommand SQL command
   * @param condition condition or Form.NONE
   * @param filter filter or Filter.NONE
   */
  private static void dumpSQLCommand(String sqlCommand,
                                     Values values
                                    )

  {
    Dprintf.printStackTrace();

    Dprintf.dprintf("SQL command: %s",sqlCommand);
    if (values != null)
    {
      for (int i = 0; i < values.data.length/2; i++)
      {
        Dprintf.dprintf("             %d: %s",2+i,values.data[2*i+1]);
      }
    }

    Dprintf.dprintf("SQL query plan: %s",sqlCommand);
    synchronized(database)
    {
      PreparedStatement preparedStatement = null;
      ResultSet         resultSet         = null;
      try
      {
        // prepare SQL command
        preparedStatement = database.prepareStatement("EXPLAIN QUERY PLAN "+sqlCommand);
        int n = 0;
        if (values != null)
        {
          for (int i = 0; i < values.data.length/2; i++)
          {
            setPreparedStatement(preparedStatement,n,values.data[2*i+1]);
            n++;
          }
        }

        // execute SQL command
        resultSet = preparedStatement.executeQuery();

        // get results
        while (resultSet.next())
        {
          Dprintf.dprintf("  %s %s %s %s",
                          resultSet.getString(1),
                          resultSet.getString(2),
                          resultSet.getString(3),
                          resultSet.getString(4)
                         );
        }

        // done
        resultSet.close(); resultSet = null;
        preparedStatement.close(); preparedStatement = null;
      }
      catch (SQLException exception)
      {
        Timemine.printError("SQL command: %s",sqlCommand);
      }
      finally
      {
        try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }
  }

  /** insert or update unique entity into local database
   * @param tableName tableName
   * @param backendType backend type
   * @param id database id
   * @param condition SQL condition string
   * @param values values
   * @return local database id
   */
  protected static int insertLocalDatabase(String        tableName,
                                           Backend.Types backendType,
                                           Id            id,
                                           Condition     condition,
                                           Values        values
                                          )
    throws BackendException
  {
    StringBuilder     buffer;
    PreparedStatement preparedStatement;
    ResultSet         resultSet;

    assert values != null;

    // get local id of existing entry
    if (condition != null)
    {
      buffer = new StringBuilder();
      buffer.append("SELECT rowid AS rowid FROM "+tableName);
      if (condition != Condition.NONE)
      {
        buffer.append(" WHERE ");
        buffer.append(condition.format);
      }
      if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
      {
        dumpSQLCommand(buffer.toString(),condition);
      }

      synchronized(database)
      {
        preparedStatement = null;
        resultSet         = null;
        try
        {
          preparedStatement = database.prepareStatement(buffer.toString());
          int n = 0;
          if (condition != Condition.NONE)
          {
            for (int i = 0; i < condition.arguments.length; i++)
            {
              setPreparedStatement(preparedStatement,n,condition.arguments[i]);
              n++;
            }
          }

          resultSet = preparedStatement.executeQuery();
          if (resultSet.next())
          {
            id.local = resultSet.getInt("rowid");
Dprintf.dprintf("id.local=%d",id.local);
          }
          resultSet.close(); resultSet = null;

          preparedStatement.close(); preparedStatement = null;
        }
        catch (SQLException exception)
        {
          throw new BackendException(exception);
        }
        finally
        {
          try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
          try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
        }
      }
    }
Dprintf.dprintf("id=%s",id);

    // update/insert
    if (id.local != Id.NONE)
    {
      // create SQL update command
      buffer = new StringBuilder();
      buffer.append("UPDATE "+tableName+" SET ");
      buffer.append("backendType=?,backendId=?");
      for (int i = 0; i < values.data.length/2; i++)
      {
        assert values.data[i*2+0] instanceof String;
        buffer.append(",");
        buffer.append((String)values.data[i*2+0]);
        buffer.append("=?");
      }
      buffer.append(" WHERE rowid=?");
      if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
      {
        dumpSQLCommand(buffer.toString(),values);
      }

      preparedStatement = null;
      try
      {
        preparedStatement = database.prepareStatement(buffer.toString());
        setPreparedStatement(preparedStatement,0,backendType.ordinal());
        setPreparedStatement(preparedStatement,1,id.backend);
        for (int i = 0; i < values.data.length/2; i++)
        {
          setPreparedStatement(preparedStatement,2+i,values.data[i*2+1]);
        }
        setPreparedStatement(preparedStatement,2+values.data.length/2+0,id.local);

        // execute SQL command
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        Timemine.printDebug(Settings.DebugTypes.LOCAL,"Updated entry #%d in local database '%s'",id.local,tableName);
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }
    else
    {
      // create SQL insert command
      buffer = new StringBuilder();
      buffer.append("INSERT INTO "+tableName+" (");
      buffer.append("backendType,backendId");
      for (int i = 0; i < values.data.length/2; i++)
      {
        assert values.data[i*2+0] instanceof String;
        buffer.append(",");
        buffer.append((String)values.data[i*2+0]);
      }
      buffer.append(") VALUES (");
      buffer.append("?,?");
      for (int i = 0; i < values.data.length/2; i++)
      {
        assert values.data[i*2+0] instanceof String;
        buffer.append(",?");
      }
      buffer.append(");");
      if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
      {
        dumpSQLCommand(buffer.toString(),values);
      }

      preparedStatement = null;
      try
      {
        preparedStatement = database.prepareStatement(buffer.toString());
        setPreparedStatement(preparedStatement,0,backendType.ordinal());
        setPreparedStatement(preparedStatement,1,id.backend);
        for (int i = 0; i < values.data.length/2; i++)
        {
          setPreparedStatement(preparedStatement,2+i,values.data[i*2+1]);
        }

        // execute SQL command
//Dprintf.dprintf("type=%s bckendid=%d id=%s ID_NONE=%s",getType(),id.backend,id,ID_NONE);
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        // get local id
        id.local = database.getLastInsertId();
Dprintf.dprintf("id.local=%d",id.local);
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }

      Timemine.printDebug(Settings.DebugTypes.LOCAL,"Inserted entry #%d into local database '%s'",id.local,tableName);
    }

// TODO: remove
//    triggerModifyListener(Backend.ModifiedStates.ADDED);

    return id.local;
  }

  /** insert or update unique entity into local database
   * @param tableName tableName
   * @param backendType backend type
   * @param id database id or 0
   * @param values values
   * @return local database id
   */
  protected static int insertLocalDatabase(String        tableName,
                                           Backend.Types backendType,
                                           Id            id,
                                           Values        values
                                          )
    throws BackendException
  {
    return insertLocalDatabase(tableName,backendType,id,Condition.NONE,values);
  }

  /** insert or update unique entity into local database
   * @param tableName tableName
   * @param backendType backend type
   * @param condition condition
   * @param values values
   * @return local database id
   */
  protected static int insertLocalDatabase(String        tableName,
                                           Backend.Types backendType,
                                           Condition     condition,
                                           Values        values
                                          )
    throws BackendException
  {
    return insertLocalDatabase(tableName,backendType,new Id(),condition,values);
  }

  /** insert or update unique entity into local database
   * @param tableName tableName
   * @param backendType backend type
   * @param values values
   * @return local database id
   */
  protected static int insertLocalDatabase(String        tableName,
                                           Backend.Types backendType,
                                           Values        values
                                          )
    throws BackendException
  {
    return insertLocalDatabase(tableName,backendType,Condition.NONE,values);
  }

  /** update entity in local database
   * @param tableName tableName
   * @param id local database id
   * @param values values
   * @return number of changed rows
   */
  protected static int updateLocalDatabase(String    tableName,
                                           Condition condition,
                                           Values    values
                                          )
    throws BackendException
  {
    assert values != null;

    int changedRowCount = 0;

    // create SQL update command
    StringBuilder buffer = new StringBuilder();
    buffer.append("UPDATE "+tableName+" SET ");
    for (int i = 0; i < values.data.length/2; i++)
    {
      assert values.data[i*2+0] instanceof String;
      if (i > 0) buffer.append(",");
      buffer.append((String)values.data[i*2+0]);
      buffer.append("=?");
    }
    if (condition != Condition.NONE)
    {
      buffer.append(" WHERE ");
      buffer.append(condition.format);
    }
    buffer.append(";");
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      Timemine.printDebug(Settings.DebugTypes.SQL,"SQL command: %s",buffer.toString());
      for (int i = 0; i < values.data.length/2; i++)
      {
        Timemine.printDebug(Settings.DebugTypes.SQL,"             %d: %s",2+i,values.data[2*i+1]);
      }
    }
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      dumpSQLCommand(buffer.toString(),values);
    }

    // execute SQL command
    synchronized(database)
    {
      PreparedStatement preparedStatement = null;
      try
      {
        preparedStatement = database.prepareStatement(buffer.toString());
        int n = 0;
        for (int i = 0; i < values.data.length/2; i++)
        {
          setPreparedStatement(preparedStatement,n,values.data[i*2+1]);
          n++;
        }
        if (condition != null)
        {
          for (int i = 0; i < condition.arguments.length; i++)
          {
            setPreparedStatement(preparedStatement,n,condition.arguments[i]);
            n++;
          }
        }

        changedRowCount = preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        Timemine.printDebug(Settings.DebugTypes.LOCAL,"Updated %d entries in local database '%s'",changedRowCount,tableName);
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }

    return changedRowCount;
  }

  /** update entity in local database
   * @param tableName tableName
   * @param backendType backend type
   * @param id database id
   * @param values values
   * @return number of changed rows
   */
// TODO: use updateLocalDatabase with condition? how to set backendtype?
  protected static int updateLocalDatabase(String        tableName,
                                           Backend.Types backendType,
                                           Id            id,
                                           Values        values
                                          )
    throws BackendException
  {
    assert !id.isNone();
    assert !id.isAny();
    assert values != null;

    int n = 0;

    // create SQL update command
    StringBuilder buffer = new StringBuilder();
    buffer.append("UPDATE "+tableName+" SET ");
    buffer.append("backendType=?,backendId=?");
    for (int i = 0; i < values.data.length/2; i++)
    {
      assert values.data[i*2+0] instanceof String;
      buffer.append(",");
      buffer.append((String)values.data[i*2+0]);
      buffer.append("=?");
    }
    buffer.append(" WHERE rowid=?;");
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      Timemine.printDebug(Settings.DebugTypes.SQL,"SQL command: %s",buffer.toString());
      for (int i = 0; i < values.data.length/2; i++)
      {
        Timemine.printDebug(Settings.DebugTypes.SQL,"             %d: %d",0,backendType.ordinal());
        Timemine.printDebug(Settings.DebugTypes.SQL,"             %d: %d",1,id.backend);
        Timemine.printDebug(Settings.DebugTypes.SQL,"             %d: %s",2+i,values.data[2*i+1]);
      }
      Timemine.printDebug(Settings.DebugTypes.SQL,"             %d: %d",2+values.data.length/2+0,id.local);
    }
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      dumpSQLCommand(buffer.toString(),values);
    }

    // execute SQL command
    synchronized(database)
    {
      PreparedStatement preparedStatement = null;
      try
      {
        preparedStatement = database.prepareStatement(buffer.toString());
        setPreparedStatement(preparedStatement,0,backendType.ordinal());
        setPreparedStatement(preparedStatement,1,id.backend);
        for (int i = 0; i < values.data.length/2; i++)
        {
          setPreparedStatement(preparedStatement,2+i,values.data[i*2+1]);
        }
        setPreparedStatement(preparedStatement,2+values.data.length/2+0,id.local);
        n = preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        Timemine.printDebug(Settings.DebugTypes.LOCAL,"Updated entry #%d in local database '%s'",id.local,tableName);
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }

    return n;
  }

  /** insert entity into local database
   * @param tableName tableName
   * @param values values
   * @return local database id
   */
  protected static int insertLocalDatabase(String tableName,
                                           Values values
                                          )
    throws BackendException
  {
    PreparedStatement preparedStatement;
    int               id;

    assert values != null;

    // create SQL insert command
    StringBuilder buffer = new StringBuilder();
    buffer.append("INSERT INTO "+tableName+" (");
    for (int i = 0; i < values.data.length/2; i++)
    {
      assert values.data[i*2+0] instanceof String;
      if (i > 0) buffer.append(",");
      buffer.append((String)values.data[i*2+0]);
    }
    buffer.append(") VALUES (");
    for (int i = 0; i < values.data.length/2; i++)
    {
      assert values.data[i*2+0] instanceof String;
      if (i > 0) buffer.append(",");
      buffer.append("?");
    }
    buffer.append(");");
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      Timemine.printDebug(Settings.DebugTypes.SQL,"SQL command: %s,",buffer.toString());
      for (int i = 0; i < values.data.length/2; i++)
      {
        Timemine.printDebug(Settings.DebugTypes.SQL,"             %d: %s",i,values.data[2*i+1]);
      }
    }
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      dumpSQLCommand(buffer.toString(),values);
    }

    synchronized(database)
    {
      preparedStatement = null;
      try
      {
        preparedStatement = database.prepareStatement(buffer.toString());
        for (int i = 0; i < values.data.length/2; i++)
        {
          setPreparedStatement(preparedStatement,i,values.data[i*2+1]);
        }

        // execute SQL command
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        // get id
        id = database.getLastInsertId();
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }

    return id;
  }

  /** insert entity into local database with select
   * @param tableName tableName
   * @param values values
   */
  protected static void insertLocalDatabase(String toTableName,
                                            String toNames[],
                                            String fromTableName,
                                            String fromNames[]
                                           )
    throws BackendException
  {
    PreparedStatement preparedStatement;
    int               id;

    assert toTableName != null;
    assert toNames != null;
    assert fromTableName != null;
    assert fromNames != null;
    assert toNames.length == fromNames.length;

    // create SQL insert command
    StringBuilder buffer = new StringBuilder();
    buffer.append("INSERT INTO "+toTableName+" (");
    for (int i = 0; i < toNames.length; i++)
    {
      if (i > 0) buffer.append(",");
      buffer.append(toNames[i]);
    }
    buffer.append(") SELECT ");
    for (int i = 0; i < fromNames.length; i++)
    {
      if (i > 0) buffer.append(",");
      buffer.append(fromNames[i]);
    }
    buffer.append(" FROM "+fromTableName+" ");
    buffer.append(";");
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      Timemine.printDebug(Settings.DebugTypes.SQL,"SQL command: %s,",buffer.toString());
    }
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
// TODO: remove null
      dumpSQLCommand(buffer.toString(),(Values)null);
    }

    synchronized(database)
    {
      preparedStatement = null;
      try
      {
        preparedStatement = database.prepareStatement(buffer.toString());

        // execute SQL command
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }
  }

  /** delete entity from local database
   * @param tableName tableName
   * @param condition condition or null
   * @return number of changed rows
   */
  protected static int removeLocalDatabase(String    tableName,
                                           Condition condition
                                          )
    throws BackendException
  {
    int changedRowCount = 0;

    // create SQL delete command
    StringBuilder buffer = new StringBuilder();
    buffer.append("DELETE FROM "+tableName);
    if (condition != Condition.NONE)
    {
      buffer.append(" WHERE ");
      buffer.append(condition.format);
    }
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      dumpSQLCommand(buffer.toString(),condition);
    }

    // execute SQL command
    synchronized(database)
    {
      PreparedStatement preparedStatement = null;
      try
      {
        preparedStatement = database.prepareStatement(buffer.toString());
        int n = 0;
        if (condition != null)
        {
          for (int i = 0; i < condition.arguments.length; i++)
          {
            setPreparedStatement(preparedStatement,n,condition.arguments[i]);
            n++;
          }
        }

        changedRowCount = preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        if (changedRowCount > 0) Timemine.printDebug(Settings.DebugTypes.LOCAL,"Deleted %d entries from local database '%s'",changedRowCount,tableName);
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }

// TODO: remove
//    triggerModifyListener(Backend.ModifiedStates.DELETED);

    return changedRowCount;
  }

  /** delete entity from local database
   * @param tableName tableName
   * @param id local database id
   */
  protected static void removeLocalDatabase(String tableName,
                                            int    id
                                           )
    throws BackendException
  {
    removeLocalDatabase(tableName,condition("rowid=?",id));
  }

  /** delete entity from local database
   * @param tableName table name
   * @param condition condition
   * @param arguments condition arguments
   */
  protected static void removeAllLocalDatabase(String    tableName,
                                               Condition condition,
                                               Object... arguments
                                              )
    throws BackendException
  {
    // create SQL delete command
    StringBuilder buffer = new StringBuilder();
    buffer.append("DELETE FROM "+tableName);
    if (condition != Condition.NONE)
    {
      buffer.append(" WHERE "+condition.format);
    }
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      dumpSQLCommand(buffer.toString(),condition);
    }

    // execute SQL command
    synchronized(database)
    {
      PreparedStatement preparedStatement = null;
      try
      {
        preparedStatement = database.prepareStatement(buffer.toString());
        for (int i = 0; i < arguments.length; i++)
        {
          setPreparedStatement(preparedStatement,i,arguments[i]);
        }
        int n = preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;

        if (n > 0) Timemine.printDebug(Settings.DebugTypes.LOCAL,"Deleted %d entries from local database '%s'",n,tableName);
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }

// TODO: remove
//    triggerModifyListener(Backend.ModifiedStates.DELETED);
  }

  /** delete entity from local database
   * @param tableName table name
   */
  protected static void removeAllLocalDatabase(String tableName)
    throws BackendException
  {
    removeAllLocalDatabase(tableName,Condition.NONE);
  }

  /** get data from local database
   * @param withStatement with statement
   * @param dataBaseHandler data base handler
   * @param condition condition or Form.NONE
   * @param filter filter or Filter.NONE
   * @return first entry or null
   */
  protected static <T> T getLocalDatabase(String             withStatement,
                                          DataBaseHandler<T> dataBaseHandler,
                                          Condition          condition,
                                          Filter             filter
                                         )
    throws BackendException
  {
    long t0 = System.currentTimeMillis();

    // create SQL select command
    StringBuilder buffer = new StringBuilder();
    if (withStatement != null)
    {
      buffer.append(withStatement);
      buffer.append(" ");
    }
    buffer.append(dataBaseHandler.selectStatement((condition != null) ? condition.format : null));
    if (filter != null)
    {
      buffer.append(" ");
      buffer.append(filter.format);
    }
    buffer.append(";");
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      dumpSQLCommand(buffer.toString(),condition,filter);
    }

    T data = null;
    synchronized(database)
    {
      PreparedStatement preparedStatement = null;
      ResultSet         resultSet         = null;
      try
      {
        // prepare SQL command
        preparedStatement = database.prepareStatement(buffer.toString());
        int n = 0;
        if (condition != null)
        {
          for (int i = 0; i < condition.arguments.length; i++)
          {
            setPreparedStatement(preparedStatement,n,condition.arguments[i]);
            n++;
          }
        }
        if (filter != null)
        {
          for (int i = 0; i < filter.arguments.length; i++)
          {
            setPreparedStatement(preparedStatement,n,filter.arguments[i]);
            n++;
          }
        }

        // execute SQL command
        resultSet = preparedStatement.executeQuery();

        // get results
        DataBaseHandler.ValueSet valueSet = new DataBaseHandler.ValueSet(resultSet);
        while (resultSet.next())
        {
          dataBaseHandler.process(valueSet);
          data = dataBaseHandler.get(valueSet);
          if (data != null)
          {
            dataBaseHandler.process(data);
          }
        }

        // done
        resultSet.close(); resultSet = null;
        preparedStatement.close(); preparedStatement = null;
      }
      catch (SQLException exception)
      {
        Timemine.printError("SQL command: %s",buffer.toString());
        Timemine.printInternalError(exception);
        throw new BackendException(exception);
      }
      finally
      {
        try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }

    long t1 = System.currentTimeMillis();
    if ((t1-t0) > 500)
    {
      Dprintf.dprintf("Warning: slow SQL command '%s': %dms",buffer.toString(),t1-t0);
    }

    return data;
  }

  /** get data from local database
   * @param dataBaseHandler data base handler
   * @param condition condition or Form.NONE
   * @param filter filter or Filter.NONE
   * @return first entry or null
   */
  protected static <T> T getLocalDatabase(DataBaseHandler<T> dataBaseHandler,
                                          Condition          condition,
                                          Filter             filter
                                         )
    throws BackendException
  {
    return getLocalDatabase((String)null,  // withStatement
                            dataBaseHandler,
                            condition,
                            filter
                           );
  }

  /** get string from local database
   * @param dataBaseHandler data base handler
   * @param condition condition or Form.NONE
   * @return first entry or null
   */
  protected static <T> T getLocalDatabase(DataBaseHandler<T> dataBaseHandler,
                                          Condition          condition
                                         )
    throws BackendException
  {
    return getLocalDatabase(dataBaseHandler,
                            condition,
                            Filter.NONE
                           );
  }

  /** get string from local database
   * @param dataBaseHandler data base handler
   * @return entry or null
   */
  protected static <T> T getLocalDatabase(DataBaseHandler<T> dataBaseHandler)
    throws BackendException
  {
    return getLocalDatabase(dataBaseHandler,
                            Condition.NONE
                           );
  }

  /** get data from local database
   * @param tableName tableName
   * @param withStatement with statement
   * @param selectStatement select statement
   * @param condition condition or Condition.NONE
   * @param filter filter or Filter.NONE
   * @param defaultValue default value
   * @return select result
   */
  protected static <T> T getLocalDatabase(String    tableName,
                                          String    withStatement,
                                          String    selectStatement,
                                          Condition condition,
                                          Filter    filter,
                                          T         defaultValue
                                         )
    throws BackendException
  {
    long t0 = System.currentTimeMillis();

    // create SQL select command
    StringBuilder buffer = new StringBuilder();
    if (withStatement != null)
    {
      buffer.append(withStatement);
      buffer.append(" ");
    }
    buffer.append("SELECT "+selectStatement+" FROM "+tableName);
    if (condition != null)
    {
      buffer.append(" WHERE ");
      buffer.append(condition.format);
    }
    if (filter != null)
    {
      buffer.append(" ");
      buffer.append(filter);
    }
    buffer.append(";");
    if (Settings.debugTypes.contains(Settings.DebugTypes.SQL))
    {
      dumpSQLCommand(buffer.toString(),condition,filter);
    }

    // execute SQL command
    T data = defaultValue;
    synchronized(database)
    {
      PreparedStatement preparedStatement = null;
      ResultSet         resultSet         = null;
      try
      {
        preparedStatement = database.prepareStatement(buffer.toString());
        int n = 0;
        if (condition != null)
        {
          for (int i = 0; i < condition.arguments.length; i++)
          {
            setPreparedStatement(preparedStatement,n,condition.arguments[i]);
            n++;
          }
        }
        resultSet = preparedStatement.executeQuery();
        if (resultSet.next())
        {
          if      (data instanceof Integer)
          {
            data = (T)new Integer(resultSet.getInt(1));
          }
          else if (data instanceof Long)
          {
            data = (T)new Long(resultSet.getLong(1));
          }
          else if (data instanceof Double)
          {
            data = (T)new Double(resultSet.getDouble(1));
          }
          else
          {
            data = (T)resultSet.getObject(1);
          }
        }
        resultSet.close(); resultSet = null;
        preparedStatement.close(); preparedStatement = null;
      }
      catch (SQLException exception)
      {
        throw new BackendException(exception);
      }
      finally
      {
        try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }

    long t1 = System.currentTimeMillis();
    if ((t1-t0) > 500)
    {
      Dprintf.dprintf("Warning: slow SQL command '%s': %dms",buffer.toString(),t1-t0);
    }

    return data;
  }

  /** get data from local database
   * @param tableName tableName
   * @param selectStatement select statement
   * @param condition condition or Condition.NONE
   * @param filter filter or Filter.NONE
   * @param defaultValue default value
   * @return select result
   */
  protected static <T> T getLocalDatabase(String    tableName,
                                          String    selectStatement,
                                          Condition condition,
                                          Filter    filter,
                                          T         defaultValue
                                         )
    throws BackendException
  {
    return getLocalDatabase(tableName,
                            (String)null,  // withStatement
                            selectStatement,
                            condition,
                            filter,
                            defaultValue
                           );
  }

  /** get entity from local database
   * @param tableName tableName
   * @param selectStatement select statement
   * @param defaultValue default value
   * @return select result
   */
//TODO: obsolete? because of protected static <T extends Entity> T getLocalDatabase(...)?
  protected static <T> T getLocalDatabase(String tableName,
                                          String selectStatement,
                                          T      defaultValue
                                         )
    throws BackendException
  {
    return getLocalDatabase(tableName,
                            selectStatement,
                            Condition.NONE,
                            Filter.NONE,
                            defaultValue
                           );
  }

  /** trust manager which accept any certificate
   */
  private static class AnyTrustManager implements X509TrustManager
  {
    @Override
    public void checkClientTrusted(X509Certificate[] arg0, String arg1)
      throws CertificateException
    {
    }

    @Override
    public void checkServerTrusted(X509Certificate[] arg0, String arg1)
      throws CertificateException
    {
    }

    @Override
    public X509Certificate[] getAcceptedIssuers()
    {
      return null;
    }
  }

  /** compare value in prepared statement
   * @param resultSet result set
   * @param index [0..n-1]
   * @param data data
   */
  private static boolean equalsResultSet(ResultSet resultSet, int index, Object data)
    throws SQLException
  {
    boolean equals = false;

    assert index >= 0;

    if (data != null)
    {
      if      (data.getClass() == Integer.class)
      {
        int value = ((Integer)data).intValue();
        equals = (resultSet.getInt(1+index) == value);
      }
      else if (data.getClass() == Long.class)
      {
        long value = ((Long)data).longValue();
        equals = (resultSet.getLong(1+index) == value);
      }
      else if (data.getClass() == Double.class)
      {
        double value = ((Double)data).doubleValue();
        equals = (resultSet.getDouble(1+index) == value);
      }
      else if (data.getClass() == Boolean.class)
      {
        boolean value = ((Boolean)data).booleanValue();
        equals = (resultSet.getBoolean(1+index) == value);
      }
      else if (data.getClass() == String.class)
      {
        String value = (String)data;
        equals = (resultSet.getString(1+index).equals(value));
      }
      else if (data.getClass() == Date.class)
      {
        Date value = (Date)data;
        equals = (resultSet.getLong(1+index) == value.getTime()/1000);
      }
      else if (data.getClass().isEnum())
      {
        int value = ((Enum)data).ordinal();
        equals = (resultSet.getInt(1+index) == value);
      }
      else if (data.getClass() == Id.class)
      {
        Id value = (Id)data;
        equals = (resultSet.getInt(1+index) == value.backend);
      }
      else if (data.getClass() == SpentOn.class)
      {
        SpentOn value = (SpentOn)data;
        equals = (resultSet.getLong(1+index) == value.getTime()/1000);
      }
      else
      {
        throw new IllegalArgumentException("unsupported data type '"+data.getClass()+"' for "+data+" at "+index);
      }
    }

    return equals;
  }

  /** clean all entities not in id list from local database
   * @param database database
   * @param tableName tableName
   * @param IdList id list
   * @param condition condition or null
   * @param arguments condition arguments
   */
  private static void cleanLocalDatabase(String       tableName,
                                         Backend.Types type,
                                         IdList       idList,
                                         String       condition,
                                         Object...    arguments
                                        )
    throws SQLException
  {
    // execute SQL command
    synchronized(database)
    {
      StringBuilder     buffer;
      PreparedStatement deleteStatement   = null;
      PreparedStatement preparedStatement = null;
      ResultSet         resultSet         = null;
      try
      {
        // create delete statement
        buffer = new StringBuilder("DELETE FROM "+tableName+" WHERE rowid=?");
        if (condition != null)
        {
          buffer.append(" AND ");
          buffer.append(condition);
        }
        deleteStatement = database.prepareStatement(buffer.toString());

        // delete all not existing data elements
        preparedStatement = database.prepareStatement("SELECT rowid,backendType,backendId FROM "+tableName);
        resultSet = preparedStatement.executeQuery();
        while (resultSet.next())
        {
          int          rowId       = resultSet.getInt(1);
          Backend.Types backendType = Backend.Types.toEnum(resultSet.getInt(2));
          int          backendId   = resultSet.getInt(3);

          if (type == backendType)
          {
            boolean existFlag = false;
            for (Id id : idList)
            {
              if (id.backend == backendId)
              {
                existFlag = true;
                break;
              }
            }

            if (!existFlag)
            {
              setPreparedStatement(deleteStatement,0,rowId);
              deleteStatement.addBatch();
            }
          }
        }
        resultSet.close(); resultSet = null;
        preparedStatement.close(); preparedStatement = null;
        int n = deleteStatement.executeBatch().length;
        deleteStatement.close(); deleteStatement = null;

        if (n > 0) Timemine.printDebug(Settings.DebugTypes.LOCAL,"Deleted %d '%s' entries from local database",n,tableName);
      }
      catch (SQLException exception)
      {
        Timemine.printInternalError(exception);
        throw exception;
      }
      finally
      {
        try { if (resultSet != null) resultSet.close(); } catch (SQLException exception) { /* ignored */ }
        try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
        try { if (deleteStatement != null) deleteStatement.close(); } catch (SQLException exception) { /* ignored */ }
      }
    }
  }

  /** clean all entities not in id list from local database
   * @param database database
   * @param tableName tableName
   * @param IdList id list
   */
  private static void cleanLocalDatabase(String       tableName,
                                         Backend.Types type,
                                         IdList       idList
                                        )
    throws SQLException
  {
    cleanLocalDatabase(tableName,type,idList,(String)null);
  }
}

/* end of file */
