;/***********************************************************************\
*
* $Revision: 913 $
* $Date: 2022-01-13 16:47:15 +0100 (Thu, 13 Jan 2022) $
* $Author: torsten $
* Contents: Timemine
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileDescriptor;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintStream;

import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import java.text.NumberFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.Stack;
import java.util.Timer;
import java.util.TimerTask;

// SQLite database
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

// Nebula
import org.eclipse.nebula.widgets.cdatetime.CDT;
import org.eclipse.nebula.widgets.cdatetime.CDateTime;

// graphics
import java.awt.Desktop;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.custom.TreeEditor;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.MouseWheelListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.events.ShellListener;
import org.eclipse.swt.events.TreeEvent;
import org.eclipse.swt.events.TreeListener;
import org.eclipse.swt.events.VerifyEvent;
import org.eclipse.swt.events.VerifyListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.graphics.Region;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.graphics.TextLayout;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.ColorDialog;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.DateTime;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Tray;
import org.eclipse.swt.widgets.TrayItem;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.TreeColumn;
import org.eclipse.swt.widgets.TreeItem;
import org.eclipse.swt.widgets.TypedListener;
import org.eclipse.swt.widgets.Widget;

/****************************** Classes ********************************/

/** Timemine
 */
public class Timemine
{
  /** color data
   */
  class ColorData
  {
    Color foreground;
    Color background;

    ColorData(Settings.Color color)
    {
      this.foreground = (color.foreground != null) ? new Color(display,color.foreground) : null;
      this.background = (color.background != null) ? new Color(display,color.background) : null;
    }

    public void dispose()
    {
      if (foreground != null) foreground.dispose();
      if (background != null) background.dispose();
    }
  }

  /** login data
   */
  class LoginData
  {
    Backend.Types              backendType;
    String                    serverName;
    int                       serverPort;
    boolean                   serverUseSSL;
    String                    loginName;
    String                    loginPassword;
    Settings.CertificateTypes certificateType;
    String                    certificateFileName;
    String                    keyFileName;
    String                    p12FileName;
    String                    keyPassword;
    boolean                   enabled;

    /** create login data from environment
     * @param backendType backend type
     * @param prefix environment variable prefix
     * @param defaultServerName default server name
     * @param defaultServerPort default server port
     * @param defaultServerUseSSL default true iff server use SSL
     * @param defaultLoginName default login name
     * @param defaultLoginPassword default login password
     * @param defaultCertificateKeyP12 certificate+key/P12
     * @param defaultCertificateFileName certificate file (.cert, can be null)
     * @param defaultKeyFileName key file (.pem, can be null)
     * @param defaultP12FileName p12 file name (.p12, can be null)
     * @param defaultKeyPassword key password
     * @param defaultEnabled default enabled
     */
    LoginData(Backend.Types             backendType,
              String                    defaultServerName,
              int                       defaultServerPort,
              boolean                   defaultServerUseSSL,
              String                    defaultLoginName,
              String                    defaultLoginPassword,
              Settings.CertificateTypes defaultCertificateType,
              String                    defaultCertificateFileName,
              String                    defaultKeyFileName,
              String                    defaultP12FileName,
              String                    defaultKeyPassword,
              boolean                   defaultEnabled
             )
    {
      String serverTypeString = null;
      switch (backendType)
      {
        case REDMINE: serverTypeString = "REDMINE"; break;
        case JIRA:    serverTypeString = "JIRA";    break;
      }

      this.backendType         = backendType;
      this.serverName          = getEnv(defaultServerName,         "TIMEMINE_",serverTypeString+"_SERVER",      "SERVER"      );
      this.serverPort          = getEnv(defaultServerPort,         "TIMEMINE_",serverTypeString+"_SERVER_PORT", "SERVER_PORT" );
      this.serverUseSSL        = getEnv(defaultServerUseSSL,       "TIMEMINE_",serverTypeString+"_SERVER_SSL",  "SERVER_SSL"  );
      this.loginName           = getEnv(defaultLoginName,          "TIMEMINE_",serverTypeString+"_NAME",        "NAME"        );
      this.loginPassword       = getEnv(defaultLoginPassword,      "TIMEMINE_",serverTypeString+"_PASSWORD",    "PASSWORD"    );
      this.certificateType     = getEnv(defaultCertificateType,    "TIMEMINE_",serverTypeString+"_CERTIFICATE", "CERTIFICATE" );
      this.certificateFileName = getEnv(defaultCertificateFileName,"TIMEMINE_",serverTypeString+"_CERTIFICATE", "CERTIFICATE" );
      this.keyFileName         = getEnv(defaultKeyFileName,        "TIMEMINE_",serverTypeString+"_KEY",         "KEY"         );
      this.p12FileName         = getEnv(defaultP12FileName,        "TIMEMINE_",serverTypeString+"_P12",         "P12"         );
      this.keyPassword         = getEnv(defaultKeyPassword,        "TIMEMINE_",serverTypeString+"_KEY_PASSWORD","KEY_PASSWORD");
      this.enabled             = getEnv(defaultEnabled,            "TIMEMINE_",serverTypeString+"_ENABLED",     "ENABLED"     );
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      return "LoginData { backendType="+backendType+", serverName="+serverName+", serverPort="+serverPort+", serverUseSSL="+serverUseSSL+", loginName="+loginName+", certificate="+certificateType+", certificateFileName="+certificateFileName+", keyFileName="+keyFileName+", p12FileName="+p12FileName+" }";
    }

    /** get from environment variable
     * @param defaultValue default value
     * @param prefix prefix
     * @param names names
     * @return value
     */
    private <T> T getEnv(T defaultValue, String prefix, String... names)
    {
      T value = defaultValue;

      for (String name : names)
      {
        String string = System.getenv(prefix+name);
        if (string != null)
        {
          if      (defaultValue instanceof Integer)
          {
            value = (T)new Integer(Integer.parseInt(string));
          }
          else if (defaultValue instanceof Boolean)
          {
            value = (T)new Boolean(string.equalsIgnoreCase("yes"));
          }
          else if (defaultValue instanceof String)
          {
            value = (T)string;
          }
          else if (defaultValue instanceof Enum)
          {
            try
            {
              Method parse = defaultValue.getClass().getDeclaredMethod("parse",String.class,defaultValue.getClass());
              value = (T)parse.invoke(null,string,defaultValue);
            }
            catch (NoSuchMethodException exception)
            {
              throw new InternalError(exception);
            }
            catch (IllegalAccessException exception)
            {
              throw new InternalError(exception);
            }
            catch (InvocationTargetException exception)
            {
              throw new InternalError(exception);
            }
          }
          else
          {
            throw new InternalError("unsupported type '"+((defaultValue!=null)? defaultValue.getClass() : "unknown")+"'");
          }
          break;
        }
      }

      return value;
    }
  }

  /** triggerable task
   */
  abstract class Task
  {
    private Boolean[] triggerFlag = new Boolean[]{false};
    private Thread    thread;

    /** create task
     */
    public Task(String name)
    {
      thread = new Thread(name)
      {
        public void run()
        {
          while (true)
          {
            // wait for trigger
            synchronized(triggerFlag)
            {
              while (!triggerFlag[0])
              {
                try { triggerFlag.wait(); } catch (InterruptedException exception) { /* ignored */ }
              }
              triggerFlag[0] = false;
            }

            // run
            Task.this.run();
          }
        }
      };
      thread.setDaemon(true);
      thread.start();
    }

    /** trigger task
     */
    public void trigger()
    {
      synchronized(triggerFlag)
      {
        triggerFlag[0] = true;
        triggerFlag.notifyAll();
      }
    }

    /** run method
     */
    public abstract void run();
  }

  /** favorites
   */
  class Favorite extends Canvas
  {
    private int     index;
    private Project project;
    private Image   image;
    private boolean selected;

    /** create favorite
     * @param composite composite
     * @param index index
     * @param name name
     */
    Favorite(Composite composite, int index, String name)
    {
      super(composite,SWT.BORDER);
      this.index     = index;
      this.project   = null;
      this.image     = IMAGE_EMPTY;

      // get project
      try
      {
        this.project = Backend.getProject(name);
      }
      catch (BackendException exception)
      {
        // ignored
      }

      addPaintListener(new PaintListener()
      {
        public void paintControl(PaintEvent paintEvent)
        {
          Favorite.this.paint(paintEvent);
        }
      });

      refresh();
    }

    /** get index in tab
     * @return index
     */
    public int getIndex()
    {
      return this.index;
    }

    /** get name
     * @return name
     */
    public String getName()
    {
      return this.project.name;
    }

    /** set name
     * @param name name
     */
    public void setName(String name)
    {
      // get project
      try
      {
        this.project = Backend.getProject(name);
      }
      catch (BackendException exception)
      {
        // ignored
      }

      refresh();
    }

    /** add icon
     * @param fileName file name of icon
     */
    public void addIcon(String fileName)
    {
      final File ICON_DIRECTORY = new File(Settings.TIMEMINE_DIRECTORY,"icons");

      if ((project != null) && !project.name.isEmpty())
      {
        File imageFile = new File(ICON_DIRECTORY,project.name);

        // copy file
        try
        {
          // create icon directory (if not exists)
          if (!ICON_DIRECTORY.exists())
          {
            ICON_DIRECTORY.mkdirs();
          }
          else
          {
            if (!ICON_DIRECTORY.isDirectory())
            {
              throw new IOException("not a directory");
            }
            if (!ICON_DIRECTORY.canWrite())
            {
              throw new IOException("not write access");
            }
          }

          // copy file
          FileUtils.copyFile(new File(fileName),imageFile);
        }
        catch (IOException exception)
        {
          Dialogs.error(shell,"Copy '%s' fail (error: %s)",fileName,exception.getMessage());
        }

        refresh();
      }
    }

    /** set icon
     * @param fileName file name of icon
     */
    public void setIcon(String fileName)
    {
      final File ICON_DIRECTORY = new File(Settings.TIMEMINE_DIRECTORY,"icons");

      if ((project != null) && !project.name.isEmpty())
      {
        File imageFile = new File(ICON_DIRECTORY,project.name);

        // copy file
        try
        {
          // create icon directory (if not exists)
          if (!ICON_DIRECTORY.exists())
          {
            ICON_DIRECTORY.mkdirs();
          }
          else
          {
            if (!ICON_DIRECTORY.isDirectory())
            {
              throw new IOException("not a directory");
            }
            if (!ICON_DIRECTORY.canWrite())
            {
              throw new IOException("not write access");
            }
          }

          // copy file
          FileUtils.copyFile(new File(ICON_DIRECTORY,fileName),imageFile);
        }
        catch (IOException exception)
        {
          Dialogs.error(shell,"Copy '%s' fail (error: %s)",fileName,exception.getMessage());
        }

        refresh();
      }
    }

    /** set selection
     * @param selected true iff selected
     */
    public void setSelection(boolean selected)
    {
      this.selected = selected;
      redraw();
    }

    /** get project
     * @return project or null if not found
     */
    public Project getProject()
    {
      return project;
    }

    /** get issue
     * @return issue or null if not found
     */
    public Issue getIssue()
    {
      Issue issue = null;
      try
      {
//TODO
//        issue = Backend.getIssue(Id.NONE);
throw new BackendException("NYI");
      }
      catch (BackendException exception)
      {
        // ignored
        printStackTrace(exception);
      }

      return issue;
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      return "Favorite {"+project+"}";
    }

    /** refresh
     */
    private void refresh()
    {
      final File ICON_DIRECTORY = new File(Settings.TIMEMINE_DIRECTORY,"icons");

      image = IMAGE_EMPTY;

      if ((project != null) && !project.name.isEmpty())
      {
        try
        {
          File imageFile = new File(ICON_DIRECTORY,project.name);
          if (imageFile.exists() && imageFile.canRead())
          {
            image = new Image(display,imageFile.getPath());
          }
        }
        catch (Exception exception)
        {
          // ignored
          printInternalError(exception);
        }
      }

      redraw();
    }

    /** paint favorite
     * @param paintEvent paint event
     */
    private void paint(PaintEvent paintEvent)
    {
      // background
      if (selected)
      {
        setBackground(getDisplay().getSystemColor(SWT.COLOR_LIST_SELECTION));
      }
      else
      {
        setBackground(null);
      }

      Rectangle bounds = getBounds();

      // project image
      if (image != null)
      {
        Rectangle imageBounds = image.getBounds();

        // get scaling image factor
        double f;
        if ((imageBounds.width  > bounds.width) || (imageBounds.height > bounds.height))
        {
          f = Math.min((double)(bounds.width -20)/(double)imageBounds.width,
                       (double)(bounds.height-20)/(double)imageBounds.height
                      );
        }
        else
        {
          f = 1.0;
        }

        // get width/height
        int w = (int)(imageBounds.width *f);
        int h = (int)(imageBounds.height*f);

        // image
        paintEvent.gc.drawImage(image,
                                0,0,
                                imageBounds.width,imageBounds.height,
                                (bounds.width-w)/2,(bounds.height-h)/2-20,
                                w,h
                               );
      }

      // project type, name
      if (project != null)
      {
        // draw type image
        Image     image = project.backendType.getImage();
        if (image != null)
        {
          Rectangle imageBounds = image.getBounds();
          paintEvent.gc.drawImage(image,
                                  0,0,
                                  imageBounds.width,imageBounds.height,
                                  0,0,
                                  imageBounds.width,imageBounds.height
                                 );
        }

        // draw text
        TextLayout textLayout = new TextLayout(getDisplay());
        textLayout.setAlignment(SWT.CENTER);
        textLayout.setText(project.name);
        paintEvent.gc.setForeground(getDisplay().getSystemColor(SWT.COLOR_BLACK));
        textLayout.setWidth(bounds.width-10);
        textLayout.draw(paintEvent.gc,5,bounds.height-textLayout.getBounds().height-5);
      }
    }
  }

  /** time entry tree item data
   */
  class TimeEntryTreeItemData
  {
    SpentOn spentOnStart;
    SpentOn spentOnEnd;
    long    durationSum;
    long    holidayDurationSum;

    /** create time entry tree item data
     * @param spentOnStart,spentOnEnd spent on start/end
     */
    TimeEntryTreeItemData(SpentOn spentOnStart, SpentOn spentOnEnd)
    {
      this.spentOnStart = spentOnStart;
      this.spentOnEnd = spentOnEnd;

      durationSum        = Backend.getDurationSum(spentOnStart,spentOnEnd);
      holidayDurationSum = getHolidayDurationSum(spentOnStart,spentOnEnd);
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      return "TimeEntryTreeItemData { spentOnStart="+spentOnStart+", spentOnEnd="+spentOnEnd+", durationSum="+durationSum+"}";
    }
  }

  /** time entry widgets
   */
  class AddTimeEntryWidgets extends Composite
  {
    Text          widgetFind;
    TableCombo    widgetHistory;
    TreeCombo     widgetProject;
    TableCombo    widgetIssueId;
    TableCombo    widgetIssueShortName;
    TreeCombo     widgetIssueSummary;
    Button        widgetIssueSummaryToClipboard;
    Combo         widgetIssueDoneRatio;
    Button        widgetIssueStatusOwnButton;
    Button        widgetIssueStatusFilterButton;
    Menu          widgetIssueStatusFilterMenu;
    Button        widgetIssueEdit;
    Button        widgetIssueAdd;
    Button        widgetIssueBookmark;
    Spinner       widgetSpentHourFraction;
    Spinner       widgetSpentMinuteFraction;
    Combo         widgetSpentTime;
    Label         widgetActivitiesTitle;
    Combo         widgetActivities;
    Combo         widgetComments;
    Button        widgetTimeEntryAdd;

    private Timer filterProjectsIssuesTimer;

    /** create add entry widgets
     * @param composite parent composite
     */
    AddTimeEntryWidgets(Composite composite, int style)
    {
      super(composite,SWT.NONE);

      Composite subComposite;
      Button    button;
      Label     label;
      Combo     combo;
      Spinner   spinner;
      Text      text;
      Menu      menu;
      MenuItem  menuItem;
      Control   control;

      // add new entry
      label = Widgets.newLabel(this,"Find:",SWT.NONE,Settings.keyFind);
      Widgets.layout(label,0,0,TableLayoutData.W);

      subComposite = Widgets.newComposite(this);
      subComposite.setLayout(new TableLayout(0.0,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,0,1,TableLayoutData.WE,0,2);
      {
        widgetFind = Widgets.newText(subComposite,SWT.SEARCH|SWT.ICON_SEARCH|SWT.ICON_CANCEL);
        widgetFind.setToolTipText("Enter a number or some text to find a project or an issue.");
        widgetFind.setLayoutData(new TableLayoutData(0,0,TableLayoutData.WE));
        widgetFind.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
            Text widget = (Text)selectionEvent.widget;

            String filters[] = StringUtils.splitArray(widget.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
            filterIssuesProjects(widgetProject,
                                 widgetIssueId,
                                 widgetIssueShortName,
                                 widgetIssueSummary,
                                 widgetIssueDoneRatio,
                                 widgetActivitiesTitle,
                                 widgetActivities,
                                 filters
                                );
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Text widget = (Text)selectionEvent.widget;

            String filters[] = StringUtils.splitArray(widget.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
            filterIssuesProjects(widgetProject,
                                 widgetIssueId,
                                 widgetIssueShortName,
                                 widgetIssueSummary,
                                 widgetIssueDoneRatio,
                                 widgetActivitiesTitle,
                                 widgetActivities,
                                 filters
                                );
          }
        });
        widgetFind.addKeyListener(new KeyListener()
        {
          @Override
          public void keyPressed(KeyEvent keyEvent)
          {
          }
          @Override
          public void keyReleased(KeyEvent keyEvent)
          {
            Text widget = (Text)keyEvent.widget;

            if      (Widgets.isAccelerator(keyEvent,Settings.keyProject))
            {
              Widgets.setFocus(widgetProject);
              return;
            }
            else if (Widgets.isAccelerator(keyEvent,Settings.keyIssue))
            {
              Widgets.setFocus(widgetIssueSummary);
              return;
            }
            else if (Widgets.isAccelerator(keyEvent,Settings.keySpent))
            {
              Widgets.setFocus(widgetSpentHourFraction);
              return;
            }
            else if (widgetActivities.getEnabled() && (Widgets.isAccelerator(keyEvent,Settings.keyActivity)))
            {
              Widgets.setFocus(widgetActivities);
              return;
            }
            else if (Widgets.isAccelerator(keyEvent,Settings.keyComments))
            {
              Widgets.setFocus(widgetComments);
              return;
            }
            else if (Widgets.isAccelerator(keyEvent,Settings.keyAdd))
            {
              Widgets.setFocus(widgetTimeEntryAdd);
              return;
            }
            else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'x'))
            {
              widget.setText("");
            }

            // restart filter projects/issues timer
            if (filterProjectsIssuesTimer != null)
            {
              filterProjectsIssuesTimer.cancel();
              filterProjectsIssuesTimer = null;
            }
            if (widgetFind.getText().trim().length() >= 3)
            {
              filterProjectsIssuesTimer = new Timer();
              filterProjectsIssuesTimer.schedule(new TimerTask()
              {
                public void run()
                {
                  display.syncExec(new Runnable()
                  {
                    public void run()
                    {
                      String filters[] = StringUtils.splitArray(widgetFind.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
                      filterIssuesProjects(widgetProject,
                                           widgetIssueId,
                                           widgetIssueShortName,
                                           widgetIssueSummary,
                                           widgetIssueDoneRatio,
                                           widgetActivitiesTitle,
                                           widgetActivities,
                                           filters
                                          );
                    }
                  });
                }
              },
              250  // delay [ms]
              );
            }
          }
        });

        widgetHistory = new TableCombo(subComposite,SWT.CENTER,SWT.CENTER,300);
        widgetHistory.setToolTipText("Time entries history.");
        widgetHistory.addTableColumn(0,"Project",         SWT.LEFT,100,true);
        widgetHistory.addTableColumn(1,"Issue short name",SWT.LEFT, 80,true);
        widgetHistory.addTableColumn(1,"Issue",           SWT.LEFT,150,true);
        widgetHistory.addTableColumn(2,"Duration",        SWT.LEFT, 50,true);
        widgetHistory.setText("History");
        Widgets.layout(widgetHistory,0,1,TableLayoutData.E,0,0,0,0,80,SWT.DEFAULT);
        widgetHistory.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TableCombo widget    = (TableCombo)selectionEvent.widget;
            TimeEntry  timeEntry = widget.getSelection(TimeEntry.class);
            Project    project   = timeEntry.getProject();
            Issue      issue     = timeEntry.getIssue();

            if (project != null)
            {
              setSelectedProject(widgetProject,
                                 widgetIssueId,
                                 widgetIssueShortName,
                                 widgetIssueSummary,
                                 widgetIssueDoneRatio,
                                 widgetActivitiesTitle,
                                 widgetActivities,
                                 project
                                );
            }
            if (project != null)
            {
              setSelectedIssue(widgetIssueId,
                               widgetIssueShortName,
                               widgetIssueSummary,
                               widgetIssueDoneRatio,
                               issue
                              );
            }
            widgetSpentHourFraction.setSelection((int)(timeEntry.duration/60));
            widgetSpentMinuteFraction.setSelection((int)(timeEntry.duration%60));
            widgetComments.setText(timeEntry.comments);
          }
        });
      }

      label = Widgets.newLabel(this,"Project:",SWT.NONE,Settings.keyProject);
      Widgets.layout(label,1,0,TableLayoutData.W);

      widgetProject = new TreeCombo(this,SWT.LEFT,SWT.NONE,100,TreeCombo.DEFAULT_HEIGHT);
      Widgets.layout(widgetProject,1,1,TableLayoutData.WE,0,2);
      widgetProject.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          TreeCombo widget = (TreeCombo)selectionEvent.widget;

          Project project = widget.getSelection(Project.class);
          if (project != null)
          {
            String filters[] = StringUtils.splitArray(widgetFind.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
            setSelectedProject(widgetProject,
                               widgetIssueId,
                               widgetIssueShortName,
                               widgetIssueSummary,
                               widgetIssueDoneRatio,
                               widgetActivitiesTitle,
                               widgetActivities,
                               project,
                               filters
                              );
          }
        }
      });
      widgetProject.addModifyListener(new ModifyListener()
      {
        @Override
        public void modifyText(ModifyEvent modifyEvent)
        {
          widgetTimeEntryAdd.setEnabled(   (widgetProject.getSelection(Project.class) != null)
                                        && (widgetIssueSummary.getSelection(Issue.class) != null)
                                        && ((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0))
                                        && (!widgetActivities.getEnabled() || (widgetActivities.getSelectionIndex() >= 0))
                                        && !widgetComments.getText().trim().isEmpty()
                                       );
// TODO: remove
//if (!widgetTimeEntryAdd.getEnabled())
Dprintf.dprintf("not ena p=%s s=%s h=%s a=%s c=%s",
(widgetProject.getSelection(Project.class) != null),
(widgetIssueSummary.getSelection(Issue.class) != null),
((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0)),
(!widgetActivities.getEnabled() || (widgetActivities.getSelectionIndex() >= 0)),
!widgetComments.getText().trim().isEmpty()
);
        }
      });

      label = Widgets.newLabel(this,"Issue:",SWT.NONE,Settings.keyIssue);
      Widgets.layout(label,2,0,TableLayoutData.W);

      subComposite = Widgets.newComposite(this);
      subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,1.0,0.0,0.0,0.0,0.0,0.0}));
      Widgets.layout(subComposite,2,1,TableLayoutData.WE,0,2);
      {
        widgetIssueId = new TableCombo(subComposite,SWT.LEFT,SWT.NONE,150);
        widgetIssueId.addTableColumn(0,SWT.RIGHT,80,false);
        widgetIssueId.addTableColumn(1,SWT.LEFT, 70,false);
        Widgets.layout(widgetIssueId,0,0,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);
        widgetIssueId.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TableCombo widget = (TableCombo)selectionEvent.widget;

            Issue issue = widget.getSelection(Issue.NONE);
            if (issue != Issue.NONE)
            {
              setSelectedIssue(widgetIssueId,
                               widgetIssueShortName,
                               widgetIssueSummary,
                               widgetIssueDoneRatio,
                               issue
                              );
            }
          }
        });

        widgetIssueShortName = new TableCombo(subComposite,SWT.LEFT,SWT.NONE,150,EnumSet.of(TableCombo.FilterTypes.TEXT));
        widgetIssueShortName.addTableColumn(0,SWT.LEFT,120,false);
        Widgets.layout(widgetIssueShortName,0,1,TableLayoutData.W,0,0,0,0,120,SWT.DEFAULT);
        widgetIssueShortName.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TableCombo widget = (TableCombo)selectionEvent.widget;

            Issue issue = widget.getSelection(Issue.NONE);
            if (issue != Issue.NONE)
            {
              setSelectedIssue(widgetIssueId,
                               widgetIssueShortName,
                               widgetIssueSummary,
                               widgetIssueDoneRatio,
                               issue
                              );
            }
          }
        });

        widgetIssueSummary = new TreeCombo(subComposite,SWT.LEFT,SWT.NONE);
        widgetIssueSummary.addTreeColumn("#",    SWT.RIGHT, 80,false);
        widgetIssueSummary.addTreeColumn("Name", SWT.LEFT, 120,false);
        widgetIssueSummary.addTreeColumn("Issue",SWT.LEFT,   0,true );
        widgetIssueSummary.setLayout(new double[]{0.0,1.0});
        Widgets.layout(widgetIssueSummary,0,2,TableLayoutData.WE);
        widgetIssueSummary.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TreeCombo widget = (TreeCombo)selectionEvent.widget;

            Issue issue = widget.getSelection();
            if (issue != null)
            {
              setSelectedIssue(widgetIssueId,
                               widgetIssueShortName,
                               widgetIssueSummary,
                               widgetIssueDoneRatio,
                               issue
                              );
              Widgets.setFocus(widgetSpentTime);
            }
          }
        });
        widgetIssueSummary.addModifyListener(new ModifyListener()
        {
          @Override
          public void modifyText(ModifyEvent modifyEvent)
          {
            widgetTimeEntryAdd.setEnabled(   (widgetProject.getSelection() != null)
                                          && (widgetIssueSummary.getSelection() != null)
                                          && ((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0))
                                          && (!widgetActivities.getEnabled() || (widgetActivities.getSelectionIndex() >= 0))
                                          && !widgetComments.getText().trim().isEmpty()
                                         );
// TODO: remove
//if (!widgetTimeEntryAdd.getEnabled())
Dprintf.dprintf("not ena p=%s s=%s h=%s a=%s c=%s",
(widgetProject.getSelection() != null),
(widgetIssueSummary.getSelection() != null),
((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0)),
(!widgetActivities.getEnabled() || (widgetActivities.getSelectionIndex() >= 0)),
!widgetComments.getText().trim().isEmpty()
);
          }
        });

        widgetIssueSummaryToClipboard = Widgets.newButton(subComposite,IMAGE_CLIPBOARD);
        widgetIssueSummaryToClipboard.setToolTipText("Copy issue summary to clipboard.");
        Widgets.layout(widgetIssueSummaryToClipboard,0,3,TableLayoutData.NONE);
        widgetIssueSummaryToClipboard.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Issue issue = widgetIssueSummary.getSelection();
            if (issue != null)
            {
              Widgets.setClipboard(clipboard,String.format("%s %s %s",issue.id.format(),issue.shortName,issue.summary));
            }
          }
        });

        widgetIssueDoneRatio = Widgets.newOptionMenu(subComposite);
        widgetIssueDoneRatio.setToolTipText("Done percentage");
//TODO
//        widgetIssueDoneRatio.setEnabled(Backend.isEditProgressSupported());
        Widgets.layout(widgetIssueDoneRatio,0,4,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);
        Widgets.setOptionMenuItems(widgetIssueDoneRatio,
                                   new Object[]{"0%" ,   0,
                                                "10%",  10,
                                                "20%",  20,
                                                "30%",  30,
                                                "40%",  40,
                                                "50%",  50,
                                                "60%",  60,
                                                "70%",  70,
                                                "80%",  80,
                                                "90%",  90,
                                                "100%",100
                                               }
                                  );
        widgetIssueDoneRatio.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Combo       widget = (Combo)selectionEvent.widget;
            int         index  = widget.getSelectionIndex();
            final Issue issue  = widgetIssueSummary.getSelection();

            if ((index >= 0) && (issue != null))
            {
              int doneRatio = (Integer)Widgets.getComboItem(widget,index);

              // get issue
//TODO: remove
//              final Issue issue = Backend.getIssue(issueId);

              try
              {
                // set done ratio
                issue.setDoneRatio(doneRatio);

                // set status
                if      (issue.doneRatio >= 100)
                {
                  if (!Settings.issueCompletedStatus.isEmpty())
                  {
                    Id completedStatusId = Backend.getStatusId(Settings.issueCompletedStatus);
                    if (!completedStatusId.isNone() && (issue.statusId != completedStatusId))
                    {
                      switch (Dialogs.select(shell,
                                             "Select issue status",
                                             String.format("The issue does not have completed status '%s'. Change status?",Settings.issueCompletedStatus),
                                             new String[]{"Change","Keep","Cancel"},
                                             0
                                            )
                             )
                      {
                        case 0:
                          issue.setStatusTransitionId(completedStatusId);
                          break;
                        case 1:
                          break;
                        case 2:
                          Widgets.setFocus(widgetComments);
                          return;
                      }
                    }
                  }
                }
                else if (issue.doneRatio > 0)
                {
                  if (!Settings.issueUpdateStatus.isEmpty())
                  {
                    Id updateStatusId = Backend.getStatusId(Settings.issueUpdateStatus);
                    if (issue.statusId != updateStatusId)
                    {
                      switch (Dialogs.select(shell,
                                             "Select issue status",
                                             String.format("The issue does not have update status '%s'. Change status?",Settings.issueUpdateStatus),
                                             new String[]{"Change","Keep","Cancel"},
                                             0
                                            )
                             )
                      {
                        case 0:
                          issue.setStatusTransitionId(updateStatusId);
                          break;
                        case 1:
                          break;
                        case 2:
                          Widgets.setFocus(widgetComments);
                          return;
                      }
                    }
                  }
                }

                // save changes
                issue.setModified(Backend.ModifiedStates.UPDATED);
                Widgets.notify(shell,USER_EVENT_UPDATE_ISSUE,issue);
              }
              catch (BackendException exception)
              {
                printStackTrace(exception);
                Dialogs.error(shell,"Cannot set issue done ratio (error: "+exception.getMessage()+")");
              }
              catch (InterruptedException exception)
              {
                // ignored
              }
            }
          }
        });

        widgetIssueStatusOwnButton = Widgets.newCheckbox(subComposite,"Own");
        widgetIssueStatusOwnButton.setSelection(Settings.issueOwnFilter);
        widgetIssueStatusOwnButton.setToolTipText("Show own issues only");
        Widgets.layout(widgetIssueStatusOwnButton,0,5,TableLayoutData.E);
        widgetIssueStatusOwnButton.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button widget = (Button)selectionEvent.widget;

            Settings.issueOwnFilter = widget.getSelection();
            Widgets.notify(shell,USER_EVENT_UPDATE_OWN_FILTER);

            // get and update issues for current selected project
            Project project   = widgetProject.getSelection();
            String  filters[] = StringUtils.splitArray(widgetFind.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
            updateIssues(widgetIssueId,
                         widgetIssueShortName,
                         widgetIssueSummary,
                         project,
                         filters
                        );
          }
        });
        widgetIssueStatusOwnButton.addListener(USER_EVENT_UPDATE_OWN_FILTER,new Listener()
        {
          @Override
          public void handleEvent(Event event)
          {
            widgetIssueStatusOwnButton.setSelection(Settings.issueOwnFilter);
          }
        });

        widgetIssueStatusFilterButton = Widgets.newButton(subComposite,IMAGE_FILTER);
        widgetIssueStatusFilterButton.setToolTipText("Edit status filter");
        Widgets.layout(widgetIssueStatusFilterButton,0,6,TableLayoutData.E);
        widgetIssueStatusFilterButton.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            widgetIssueStatusFilterMenu.setVisible(true);
          }
        });

        widgetIssueStatusFilterMenu = Widgets.newPopupMenu(shell);
        {
          menuItem = Widgets.addMenuItemCheckbox(widgetIssueStatusFilterMenu,"Any");
          menuItem.setSelection(true);
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              // disable all other filters
              MenuItem[] menuItems = widgetIssueStatusFilterMenu.getItems();
              if (menuItems[MENU_ITEM_INDEX_ANY].getSelection())
              {
                for (int i = MENU_ITEM_INDEX_FILTER_START; i < menuItems.length; i++)
                {
                  menuItems[i].setSelection(false);
                }
                Settings.issueStatusIdFilterSet.clear();
                widgetIssueStatusFilterButton.setBackground(null);
              }
              else
              {
                boolean selectedSome = false;
                int i = MENU_ITEM_INDEX_FILTER_START;
                while ((i < menuItems.length) && !selectedSome)
                {
                  selectedSome = menuItems[i].getSelection();
                }
                menuItems[MENU_ITEM_INDEX_ANY].setSelection(!selectedSome);
                widgetIssueStatusFilterButton.setBackground(selectedSome ? COLOR_CYAN : null);
              }

              // get and update issues for current selected project
              Project project   = widgetProject.getSelection();
              String  filters[] = StringUtils.splitArray(widgetFind.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
              updateIssues(widgetIssueId,
                           widgetIssueShortName,
                           widgetIssueSummary,
                           project,
                           filters
                          );
            }
          });
        }
        widgetIssueStatusFilterButton.setMenu(widgetIssueStatusFilterMenu);

        widgetIssueEdit = Widgets.newButton(subComposite,IMAGE_EDIT);
        widgetIssueEdit.setToolTipText("Edit issue");
        Widgets.layout(widgetIssueEdit,0,7,TableLayoutData.E);
        widgetIssueEdit.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            final Issue issue = widgetIssueSummary.getSelection();
            if (issue != null)
            {
              if (editIssue(issue,"Edit issue","Save"))
              {
                try
                {
                  issue.setModified(Backend.ModifiedStates.UPDATED);
                  Widgets.notify(shell,USER_EVENT_UPDATE_ISSUE,issue);
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot update issue (error: "+exception.getMessage()+")");
                  return;
                }
                catch (InterruptedException exception)
                {
                  // ignored
                }
              }
            }
          }
        });

        widgetIssueAdd = Widgets.newButton(subComposite,IMAGE_ADD);
        widgetIssueAdd.setToolTipText("Add new issue");
        Widgets.layout(widgetIssueAdd,0,8,TableLayoutData.E);
        widgetIssueAdd.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Project project = widgetProject.getSelection();
            if (project != null)
            {
              // create new issue
              final Issue issue = new Issue(project.backendType,
                                            project.id
                                           );
              switch (issue.backendType)
              {
                case REDMINE:
                  issue.statusId  = Backend.getStatusId(Settings.Redmine.newIssueStatusName);
                  issue.trackerId = Backend.getTrackerId(Settings.Redmine.newIssueTrackerName);
                  break;
                case JIRA:
                  issue.statusId  = Backend.getStatusId(Settings.JIRA.newIssueStatusName);
                  issue.trackerId = Backend.getTrackerId(Settings.JIRA.newIssueTrackerName);
                  break;
              }
              if (editIssue(issue,"Add issue","Add"))
              {
                try
                {
                  issue.setModified(Backend.ModifiedStates.ADDED);
                  Widgets.notify(shell,USER_EVENT_ADD_ISSUE,issue);
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot add issue (error: "+exception.getMessage()+")");
                  return;
                }
                catch (InterruptedException exception)
                {
                  // ignored
                }
              }
            }
          }
        });

        widgetIssueBookmark = Widgets.newButton(subComposite,IMAGE_BOOKMARK);
        widgetIssueBookmark.setToolTipText("Add issue bookmark");
        Widgets.layout(widgetIssueBookmark,0,9,TableLayoutData.E);
        widgetIssueBookmark.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Issue issue = widgetIssueSummary.getSelection();
            if (issue != null)
            {
              // add issue bookmark
              Settings.Bookmark bookmark = new Settings.Bookmark(issue.backendType,issue.summary,issue.summary);
              if (editBookmark(bookmark,"Add bookmark","Add"))
              {
                // add bookmark
                Settings.issueBookmarkList.add(bookmark);
                Settings.setModified();
                updateBookmarksMenu(widgetBookmarksMenu,Settings.issueBookmarkList);
              }
            }
          }
        });
      }

      label = Widgets.newLabel(this,"Time spent:",SWT.NONE,Settings.keySpent);
      Widgets.layout(label,3,0,TableLayoutData.W);

      subComposite = Widgets.newComposite(this);
      subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,0.0,0.0,0.0,0.0,1.0,0.0}));
      Widgets.layout(subComposite,3,1,TableLayoutData.WE);
      {
        widgetSpentHourFraction = Widgets.newSpinner(subComposite,SWT.NONE,0,Integer.MAX_VALUE);
        widgetSpentHourFraction.setToolTipText("New time entry spent hours.");
        widgetSpentHourFraction.setTextLimit(2);
        widgetSpentHourFraction.setIncrement(1);
        widgetSpentHourFraction.setSelection(Settings.minTimeDelta/60);
        Widgets.layout(widgetSpentHourFraction,0,0,TableLayoutData.WE,0,0,4,0,90,SWT.DEFAULT);
        widgetSpentHourFraction.addModifyListener(new ModifyListener()
        {
          @Override
          public void modifyText(ModifyEvent modifyEvent)
          {
            widgetTimeEntryAdd.setEnabled(   (widgetProject.getSelection() != null)
                                          && (widgetIssueSummary.getSelection() != null)
                                          && ((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0))
                                          && (!widgetActivities.getEnabled() || (widgetActivities.getSelectionIndex() >= 0))
                                          && !widgetComments.getText().trim().isEmpty()
                                         );
//if (!widgetTimeEntryAdd.getEnabled())
Dprintf.dprintf("1 not ena project=%s issuesum=%s spent=%s activities=%s:%s comment=%s",
(widgetProject.getSelection() != null),
(widgetIssueSummary.getSelection() != null),
((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0)),
!widgetActivities.getEnabled(),(widgetActivities.getSelectionIndex() >= 0),
!widgetComments.getText().trim().isEmpty()
);
          }
        });

        label = Widgets.newLabel(subComposite,"h");
        Widgets.layout(label,0,1,TableLayoutData.W);
        widgetSpentMinuteFraction = Widgets.newSpinner(subComposite,SWT.NONE,-60,60);
        widgetSpentMinuteFraction.setToolTipText("New time entry spent minutes.");
        widgetSpentMinuteFraction.setSelection(Settings.minTimeDelta%60);
        widgetSpentMinuteFraction.setTextLimit(2);
        widgetSpentMinuteFraction.setIncrement(Settings.minTimeDelta%60);
        Widgets.layout(widgetSpentMinuteFraction,0,2,TableLayoutData.WE,0,0,4,0,90,SWT.DEFAULT);
        label = Widgets.newLabel(subComposite,"min");
        Widgets.layout(label,0,3,TableLayoutData.W);
        widgetSpentMinuteFraction.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            int hourFraction   = widgetSpentHourFraction.getSelection();
            int minuteFraction = widgetSpentMinuteFraction.getSelection();

            // Note: sometimes the minute value is not a multiple of the increment. Correct this.
            if (minuteFraction >= 0)
            {
              minuteFraction = ((minuteFraction+Settings.minTimeDelta-1)/Settings.minTimeDelta)*Settings.minTimeDelta;
            }
            else
            {
              minuteFraction = -((-minuteFraction+Settings.minTimeDelta-1)/Settings.minTimeDelta)*Settings.minTimeDelta;
            }

            if      (minuteFraction >= 60)
            {
              // increment hours
              hourFraction   += minuteFraction/60;
              minuteFraction %= 60;
            }
            else if (minuteFraction < 0)
            {
              if (hourFraction > 0)
              {
                // decrement hours, reset
                hourFraction   -= (-minuteFraction+60-1)/60;
                minuteFraction = 60-Settings.minTimeDelta;
              }
              else
              {
                // reset
                minuteFraction = 0;
              }
            }

            widgetSpentHourFraction.setSelection(hourFraction);
            widgetSpentMinuteFraction.setSelection(minuteFraction);
          }
        });
        widgetSpentMinuteFraction.addModifyListener(new ModifyListener()
        {
          @Override
          public void modifyText(ModifyEvent modifyEvent)
          {
            widgetTimeEntryAdd.setEnabled(   (widgetProject.getSelection() != null)
                                          && (widgetIssueSummary.getSelection() != null)
                                          && ((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0))
                                          && (!widgetActivities.getEnabled() || (widgetActivities.getSelectionIndex() >= 0))
                                          && !widgetComments.getText().trim().isEmpty()
                                         );
/*
Dprintf.dprintf("2 not ena project=%s issuesum=%s spent=%s activities=%s:%s comment=%s",
(widgetProject.getSelection() != null),
(widgetIssueSummary.getSelection() != null),
((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0)),
!widgetActivities.getEnabled(),(widgetActivities.getSelectionIndex() >= 0),
!widgetComments.getText().trim().isEmpty()
);*/
          }
        });

        widgetSpentTime = Widgets.newOptionMenu(subComposite);
        widgetSpentTime.setToolTipText("Spent time");
        Widgets.layout(widgetSpentTime,0,4,TableLayoutData.W,0,0,0,0,64,SWT.DEFAULT);
        Widgets.addOptionMenuItem(widgetSpentTime,new Long(0L),"");
        for (int i = 1; i <= 8*60/Settings.minTimeDelta; i++)
        {
          Widgets.addOptionMenuItem(widgetSpentTime,new Long(i*Settings.minTimeDelta),formatHours(i*Settings.minTimeDelta));
        }
        widgetSpentTime.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Combo widget = (Combo)selectionEvent.widget;
            long  value  = Widgets.getSelectedOptionMenuItem(widget,0L);
            if (value > 0L)
            {
              int hourFraction   = (int)(value/60);
              int minuteFraction = (int)(value%60);

              widgetSpentHourFraction.setSelection(hourFraction);
              widgetSpentMinuteFraction.setSelection(minuteFraction);
            }
            Widgets.setSelectedOptionMenuItem(widget,0);

            Widgets.setFocus(widgetComments);
          }
        });

        widgetActivitiesTitle = Widgets.newLabel(subComposite,"Activity");
        Widgets.layout(widgetActivitiesTitle,0,5,TableLayoutData.W);
        widgetActivities = Widgets.newOptionMenu(subComposite);
        Widgets.layout(widgetActivities,0,6,TableLayoutData.WE);
        widgetActivities.addModifyListener(new ModifyListener()
        {
          @Override
          public void modifyText(ModifyEvent modifyEvent)
          {
            widgetTimeEntryAdd.setEnabled(   (widgetProject.getSelection() != null)
                                          && (widgetIssueSummary.getSelection() != null)
                                          && ((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0))
                                          && (!widgetActivities.getEnabled() || (widgetActivities.getSelectionIndex() >= 0))
                                          && !widgetComments.getText().trim().isEmpty()
                                         );
/*
Dprintf.dprintf("not ena p=%s s=%s h=%s a=%s c=%s",
(widgetProject.getSelection() != null),
(widgetIssueSummary.getSelection() != null),
((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0)),
(!widgetActivities.getEnabled() || (widgetActivities.getSelectionIndex() >= 0)),
!widgetComments.getText().trim().isEmpty()
);*/
          }
        });
      }

      label = Widgets.newLabel(this,"Comments:",SWT.NONE,Settings.keyComments);
      Widgets.layout(label,4,0,TableLayoutData.W);

      widgetComments = Widgets.newCombo(this);
      widgetComments.setToolTipText("New time entry comment line.");
      Widgets.layout(widgetComments,4,1,TableLayoutData.WE);
      widgetComments.addModifyListener(new ModifyListener()
      {
        @Override
        public void modifyText(ModifyEvent modifyEvent)
        {
          widgetTimeEntryAdd.setEnabled(   (widgetProject.getSelection() != null)
                                        && (widgetIssueSummary.getSelection(null) != null)
                                        && ((widgetSpentHourFraction.getSelection() > 0) || (widgetSpentMinuteFraction.getSelection() > 0))
                                        && (!widgetActivities.getEnabled() || (widgetActivities.getSelectionIndex() >= 0))
                                        && !widgetComments.getText().trim().isEmpty()
                                       );
        }
      });
      widgetComments.addKeyListener(new KeyListener()
      {
        public void keyPressed(KeyEvent keyEvent)
        {
          Combo widget = (Combo)keyEvent.widget;

          if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'c'))
          {
            Widgets.setClipboard(clipboard,widget.getText());
          }
        }
        public void keyReleased(KeyEvent keyEvent)
        {
        }
      });
      widgetComments.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Widgets.setFocus(widgetTimeEntryAdd);
        }
      });

      widgetTimeEntryAdd = Widgets.newButton(this,"Add new\ntime entry",Settings.keyAdd);
      widgetTimeEntryAdd.setEnabled(false);
      Widgets.layout(widgetTimeEntryAdd,3,2,TableLayoutData.NSE,2,0);
      widgetTimeEntryAdd.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Project  project   = widgetProject.getSelection();
          Issue    issue     = widgetIssueSummary.getSelection();
          int      doneRatio = 0;
          Activity activity  = widgetActivities.getEnabled() ? Widgets.getSelectedOptionMenuItem(widgetActivities,Activity.class) : null;
          long     duration  = Backend.toDuration(widgetSpentHourFraction.getSelection(),widgetSpentMinuteFraction.getSelection());
          String   comments  = widgetComments.getText().trim();

          if (!comments.isEmpty())
          {
            // check parameters
            if (project == null)
            {
              Dialogs.error(shell,"Please select a project for the time entry.");
              Widgets.setFocus(widgetProject);
              return;
            }
            if (project.id.isNone())
            {
//TOOD: sync?
              Dialogs.error(shell,"Project still not synchronized - cannot add a time entry.");
              Widgets.setFocus(widgetTimeEntryAdd);
              return;
            }
            if (issue == null)
            {
              Dialogs.error(shell,"Please select an issue for the time entry.");
              widgetIssueSummary.setFocus();
              return;
            }
            if (issue.id.isNone())
            {
//TOOD: sync?
              Dialogs.error(shell,"Issue still not synchronized - cannot add a time entry.");
              Widgets.setFocus(widgetTimeEntryAdd);
              return;
            }
            if (duration <= 0)
            {
              Dialogs.error(shell,"Please select some duration for the time entry.");
              Widgets.setFocus(widgetSpentMinuteFraction);
              return;
            }
            if (widgetActivities.getEnabled() && (activity == null))
            {
              Dialogs.error(shell,"Please select an activity for the time entry.");
              Widgets.setFocus(widgetActivities);
              return;
            }

            if (Settings.enterIssueProgress)
            {
              // change done ratio
              doneRatio = editDoneRatio(issue.doneRatio);
              if      (doneRatio >= 100)
              {
                if (!Settings.issueCompletedStatus.isEmpty())
                {
                  Id completedStatusId = Backend.getStatusId(Settings.issueUpdateStatus);
                  if (issue.statusId != completedStatusId)
                  {
                    switch (Dialogs.select(shell,
                                           "Select issue status",
                                           String.format("The issue does not have completed status '%s'. Change status?",Settings.issueCompletedStatus),
                                           new String[]{"Change","Keep","Cancel"},
                                           0
                                          )
                           )
                    {
                      case 0:
                        issue.statusId = completedStatusId;
                        break;
                      case 1:
                        break;
                      case 2:
                        Widgets.setFocus(widgetComments);
                        return;
                    }
                  }
                }
              }
              else if (doneRatio > 0)
              {
                if (!Settings.issueUpdateStatus.isEmpty())
                {
                  Id updateStatusId = Backend.getStatusId(Settings.issueUpdateStatus);
                  if (issue.statusId != updateStatusId)
                  {
                    switch (Dialogs.select(shell,
                                           "Select issue status",
                                           String.format("The issue does not have update status '%s'. Change status?",Settings.issueUpdateStatus),
                                           new String[]{"Change","Keep","Cancel"},
                                           0
                                          )
                           )
                    {
                      case 0:
                        issue.statusId = updateStatusId;
                        break;
                      case 1:
                        break;
                      case 2:
                        Widgets.setFocus(widgetComments);
                        return;
                    }
                  }
                }
              }
            }

            // create new time entry
            TimeEntry timeEntry = new TimeEntry(issue.backendType,
                                                new Id(),
                                                project.id,
                                                issue.id,
                                                widgetActivities.getEnabled() ? activity.id : new Id(),
                                                duration,
                                                comments,
                                                SpentOn.today()
                                               );
            {
              setStatus("Add new time entry\u2026");
            }
            try
            {
              // update issue
              if (Settings.enterIssueProgress)
              {
                issue.setDoneRatio(doneRatio);
                Widgets.notify(shell,USER_EVENT_UPDATE_ISSUE,issue);
              }

              // add new time entry
              timeEntry.setModified(Backend.ModifiedStates.ADDED);
              Widgets.notify(shell,USER_EVENT_ADD_TIME_ENTRY,timeEntry);
            }
            catch (BackendException exception)
            {
              printStackTrace(exception);
              Dialogs.error(shell,"Cannot add new time entry (error: "+exception.getMessage()+")");
            }
            catch (InterruptedException exception)
            {
              // ignored
            }
            finally
            {
              clearStatus();
            }

            // reset input fields
            widgetSpentHourFraction.setSelection(Settings.minTimeDelta/60);
            widgetSpentMinuteFraction.setSelection(Settings.minTimeDelta%60);
            widgetComments.setText("");
          }
        }
      });

      // add listeners
      widgetHistory.addListener(USER_EVENT_ADD_TIME_ENTRY,new Listener()
      {

        @Override
        public void handleEvent(Event event)
        {
          TableCombo widget = (TableCombo)event.widget;

          TimeEntryList historyTimeEntryList = Backend.getTimeEntryHistory(MAX_HISTORY_SIZE, Backend.HistoryFilterTypes.COMMENTS);
          updateHistory(widget,historyTimeEntryList);
        }
      });
      widgetHistory.addListener(USER_EVENT_UPDATE_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          TableCombo widget = (TableCombo)event.widget;

          TimeEntryList historyTimeEntryList = Backend.getTimeEntryHistory(MAX_HISTORY_SIZE, Backend.HistoryFilterTypes.COMMENTS);
          updateHistory(widget,historyTimeEntryList);
        }
      });
      widgetHistory.addListener(USER_EVENT_UPDATE_TIME_ENTRIES,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          TableCombo widget = (TableCombo)event.widget;

          TimeEntryList historyTimeEntryList = Backend.getTimeEntryHistory(MAX_HISTORY_SIZE, Backend.HistoryFilterTypes.COMMENTS);
          updateHistory(widget,historyTimeEntryList);
        }
      });
      widgetHistory.addListener(USER_EVENT_REMOVE_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          TableCombo widget = (TableCombo)event.widget;

          TimeEntryList historyTimeEntryList = Backend.getTimeEntryHistory(MAX_HISTORY_SIZE, Backend.HistoryFilterTypes.COMMENTS);
          updateHistory(widget,historyTimeEntryList);
        }
      });
      widgetProject.addListener(USER_EVENT_UPDATE_PROJECTS,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          TreeCombo widget = (TreeCombo)event.widget;

          ProjectList projectList = (ProjectList)event.data;
          updateProjects(widget,projectList);
        }
      });
      widgetIssueId.addListener(USER_EVENT_ADD_ISSUE,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Project project   = widgetProject.getSelection();
          String  filters[] = StringUtils.splitArray(widgetFind.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
          Issue   issue     = widgetIssueSummary.getSelection();
          updateIssues(widgetIssueId,
                       widgetIssueShortName,
                       widgetIssueSummary,
                       project,
                       filters,
                       issue
                      );
        }
      });
      widgetIssueId.addListener(USER_EVENT_UPDATE_ISSUE,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Project project   = widgetProject.getSelection();
          String  filters[] = StringUtils.splitArray(widgetFind.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
          Issue   issue     = widgetIssueSummary.getSelection();
          updateIssues(widgetIssueId,
                       widgetIssueShortName,
                       widgetIssueSummary,
                       project,
                       filters,
                       issue
                      );
        }
      });
      widgetIssueId.addListener(USER_EVENT_REMOVE_ISSUE,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          resetIssues(widgetIssueId,widgetIssueShortName,widgetIssueSummary);
        }
      });
      widgetIssueDoneRatio.addListener(USER_EVENT_ADD_ISSUE,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Issue issue = (Issue)event.data;
          updateIssueDoneRatio(widgetIssueDoneRatio,issue.doneRatio);
        }
      });
      widgetIssueDoneRatio.addListener(USER_EVENT_UPDATE_ISSUE,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Issue issue = (Issue)event.data;
          updateIssueDoneRatio(widgetIssueDoneRatio,issue.doneRatio);
        }
      });
      widgetIssueDoneRatio.addListener(USER_EVENT_REMOVE_ISSUE,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          resetIssueDoneRatio(widgetIssueDoneRatio);
        }
      });
      widgetIssueStatusFilterButton.addListener(USER_EVENT_UPDATE_STATUS_TRANSITION,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          widgetIssueStatusFilterMenu.notifyListeners(event.type,event);
        }
      });
      widgetIssueStatusFilterMenu.addListener(USER_EVENT_UPDATE_STATUS_TRANSITION,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          StatusTransitionList statusTransitionList = (StatusTransitionList)event.data;
          updateIssueStatusFilterMenu(widgetProject,
                                      widgetIssueId,
                                      widgetIssueShortName,
                                      widgetIssueSummary,
                                      widgetIssueStatusFilterMenu,
                                      statusTransitionList
                                     );
        }
      });
      widgetActivities.addListener(USER_EVENT_UPDATE_ACTIVITIES,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          ActivityList activityList = (ActivityList)event.data;
          updateActivities(widgetActivitiesTitle,widgetActivities,activityList);
        }
      });
      widgetComments.addListener(USER_EVENT_UPDATE_TIME_ENTRIES,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          TimeEntryList historyTimeEntryList = Backend.getTimeEntryHistory(MAX_HISTORY_SIZE, Backend.HistoryFilterTypes.COMMENTS);
          updateHistory(widgetComments,historyTimeEntryList);
        }
      });

      Widgets.setNextFocus(widgetProject,
                           widgetIssueId,
                           widgetIssueSummary,
//                           widgetIssueDoneRatio,
//                           widgetIssueStatusOwnButton,
//                           widgetIssueStatusFilterButton,
//                           widgetIssueEdit,
//                           widgetIssueAdd,
//                           widgetSpentHourFraction,
//                           widgetSpentMinuteFraction,
                           widgetSpentTime,
//                           widgetActivities,
                           widgetComments,
                           widgetTimeEntryAdd
                          );
    }
  }

  // --------------------------- constants --------------------------------

  // exit codes
  public final static int              EXITCODE_OK             =   0;
  public final static int              EXITCODE_FAIL           =   1;
  public final static int              EXITCODE_UNKNOWN_OPTION =   2;
  public final static int              EXITCODE_RESTART        =  64;
  public final static int              EXITCODE_INTERNAL_ERROR = 127;
  public final static int              EXITCODE_UNKNOWN        = 255;

  public static String                 ALL_FILE_EXTENSION      = (Settings.hostSystem == Settings.HostSystems.WINDOWS) ? "*.*" : "*";

  // colors
  public static Color                  COLOR_BLACK;
  public static Color                  COLOR_GRAY;
  public static Color                  COLOR_GREEN;
  public static Color                  COLOR_LIGHT_RED;
  public static Color                  COLOR_ORANGE;
  public static Color                  COLOR_CYAN;
  public static Color                  COLOR_WHITE;
  public static Color                  COLOR_BACKGROUND;
  public static Color                  COLOR_BACKGROUND_ODD;
  public static Color                  COLOR_BACKGROUND_EVEN;
  public static ColorData              COLOR_TODAY_TIME_ENTRIES;
  public static ColorData              COLOR_TIME_ENTRIES;
  public static ColorData              COLOR_TIME_ENTRIES_UPDATE;
  public static ColorData              COLOR_TIME_ENTRIES_INCOMPLETE;
  public static ColorData              COLOR_TIME_ENTRIES_WEEKEND;
  public static ColorData              COLOR_TIME_ENTRIES_HOLIDAY;
  public static ColorData              COLOR_FUTURE;
  public static ColorData              COLOR_HOURS_EXCEEDED;

  // images
  public static Image                  IMAGE_SPLASH_SCREEN;
  public static Image                  IMAGE_EMPTY;
  public static Image                  IMAGE_OFFLINE;
  public static Image                  IMAGE_PROGRAM_ICON;
  public static Image                  IMAGE_CLIPBOARD;
  public static Image                  IMAGE_FILTER;
  public static Image                  IMAGE_EDIT;
  public static Image                  IMAGE_ADD;
  public static Image                  IMAGE_BOOKMARK;
  public static Image                  IMAGE_DIRECTORY;

  // cursors
  public static Cursor                 CURSOR_WAIT;

  // date/time format
  public static SimpleDateFormat       DATE_FORMAT;
  public static SimpleDateFormat       TIME_FORMAT;
  public static SimpleDateFormat       DEBUG_TIME_FORMAT;

  // user events
  public final static int              USER_EVENT_QUIT                     = 0xFFFF+ 0;
  public final static int              USER_EVENT_UPDATE_USERS             = 0xFFFF+ 1;
  public final static int              USER_EVENT_UPDATE_ROLES             = 0xFFFF+ 2;
  public final static int              USER_EVENT_UPDATE_TRACKERS          = 0xFFFF+ 3;
  public final static int              USER_EVENT_UPDATE_EPIC_NAMES        = 0xFFFF+ 4;
  public final static int              USER_EVENT_UPDATE_STATUS_TRANSITION = 0xFFFF+ 5;
  public final static int              USER_EVENT_UPDATE_PRIORITIES        = 0xFFFF+ 6;
  public final static int              USER_EVENT_UPDATE_ACTIVITIES        = 0xFFFF+ 7;
  public final static int              USER_EVENT_UPDATE_PROJECTS          = 0xFFFF+ 8;
  public final static int              USER_EVENT_UPDATE_VERSIONS          = 0xFFFF+ 9;
  public final static int              USER_EVENT_UPDATE_MEMBERS           = 0xFFFF+10;
  public final static int              USER_EVENT_UPDATE_ISSUES            = 0xFFFF+11;
  public final static int              USER_EVENT_UPDATE_TIME_ENTRIES      = 0xFFFF+12;
  public final static int              USER_EVENT_ADD_ISSUE                = 0xFFFF+13;
  public final static int              USER_EVENT_UPDATE_ISSUE             = 0xFFFF+14;
  public final static int              USER_EVENT_REMOVE_ISSUE             = 0xFFFF+15;
  public final static int              USER_EVENT_ADD_TIME_ENTRY           = 0xFFFF+16;
  public final static int              USER_EVENT_UPDATE_TIME_ENTRY        = 0xFFFF+17;
  public final static int              USER_EVENT_REMOVE_TIME_ENTRY        = 0xFFFF+18;
  public final static int              USER_EVENT_UPDATE_OWN_FILTER        = 0xFFFF+19;

  // version, email, homepage URL
  private final static String          VERSION          = Config.VERSION_MAJOR+"."+Config.VERSION_MINOR+" ("+Config.VERSION_REVISION+")";
  private final static char            MAIL_AT          = '@';  // use MAIL_AT to avoid spam
  private final static String          EMAIL            = "torsten.rupp"+MAIL_AT+"gmx.net";
  private final static String          URL              = "http://www.kigen.de/projects/timemine";
  private final static String          URL_VERSION_FILE = URL+"/version";

  // command line options
  private static final OptionEnumeration[] OPTION_ENUMERATION_STATE =
  {
    new OptionEnumeration("online", Backend.ConnectionStates.ONLINE ),
    new OptionEnumeration("offline",Backend.ConnectionStates.OFFLINE),
  };
  private static final Option[] OPTIONS =
  {
    new Option("--host-redmine",         null,Options.Types.STRING,         "Redmine.serverName"                                   ),
    new Option("--port-redmine",         null,Options.Types.INTEGER,        "Redmine.serverPort"                                   ),
    new Option("--ssl-redmine",          null,Options.Types.BOOLEAN,        "Redmine.serverUseSSL"                                 ),
    new Option("--name-redmine",         null,Options.Types.STRING,         "Redmine.loginName"                                    ),
    new Option("--password-redmine",     null,Options.Types.STRING,         "Redmine.loginPassword"                                ),

    new Option("--host-jira",            null,Options.Types.STRING,         "JIRA.serverName"                                      ),
    new Option("--port-jira",            null,Options.Types.INTEGER,        "JIRA.serverPort"                                      ),
    new Option("--ssl-jira",             null,Options.Types.BOOLEAN,        "JIRA.serverUseSSL"                                    ),
    new Option("--name-jira",            null,Options.Types.STRING,         "JIRA.loginName"                                       ),
    new Option("--password-jira",        null,Options.Types.STRING,         "JIRA.loginPassword"                                   ),

    new Option("--host",                 null,Options.Types.STRING,         "serverName"                                           ),
    new Option("--port",                 "-p",Options.Types.INTEGER,        "serverPort"                                           ),
    new Option("--ssl",                  null,Options.Types.BOOLEAN,        "serverUseSSL"                                         ),
    new Option("--name",                 "-n",Options.Types.STRING,         "loginName"                                            ),
    new Option("--password",             null,Options.Types.STRING,         "loginPassword"                                        ),

    new Option("--login-dialog",         null,Options.Types.BOOLEAN,        "loginDialogFlag"                                      ),

    new Option("--refresh-at-start",     null,Options.Types.BOOLEAN,        "refreshAtStart"                                       ),

    new Option("--online",               null,Options.Types.ENUMERATION,    "connectionState",      Backend.ConnectionStates.ONLINE ),
    new Option("--offline",              null,Options.Types.ENUMERATION,    "connectionState",      Backend.ConnectionStates.OFFLINE),
    new Option("--read-only",            null,Options.Types.BOOLEAN,        "readOnlyFlag"                                         ),
    new Option("--help",                 "-h",Options.Types.BOOLEAN,        "helpFlag"                                             ),

    new Option("--debug",                null,Options.Types.ENUMERATION_SET,"debugTypes",           Settings.DebugTypes.class      ),
    new Option("--action-monitor",       null,Options.Types.BOOLEAN,        "actionMonitorFlag"                                    ),
    new Option("--work-in-progress",     null,Options.Types.BOOLEAN,        "wipFlag"                                              ),
    new Option("--show-all-error-output",null,Options.Types.BOOLEAN,        "showAllErrorOutputFlag"                               ),

    // ignored
    new Option("--swing",                null,Options.Types.BOOLEAN,        null                                                   ),
  };

  // menu indizes
  private final int MENU_ITEM_INDEX_ANY          = 0;
  private final int MENU_ITEM_INDEX_FILTER_START = 1;

  private final int MAX_HISTORY_SIZE = 20;

  // --------------------------- variables --------------------------------
  private static PrintStream      errorStream = System.err;
  private static Display          display;
  private static Shell            shell;
  private static Tray             tray;
  private static TrayItem         trayItem;
  private static Clipboard        clipboard;

  private int                     exitcode;

  private Database                database;

  private Favorite                selectedFavorite = null;
  private Id                      lastParentId     = new Id();
  private Project                 selectedProject  = null;

  private Stack<StatusText>       statusTextStack = new Stack<StatusText>();

  private Shell                   widgetSplashScreen = null;
  private Label                   widgetSplashStatus;

  private TabFolder               widgetTabFolder;
  private Label                   widgetTodayTime;
  private Label                   widgetTodayDurationSum;
  private Button                  widgetTodayRestTime;
  private Composite               widgetTabFavoriteProjects;
  private Composite               widgetTabDay;
  private Composite               widgetTabWeeks;
  private Composite               widgetTabProjectsIssues;

  private Favorite                widgetFavoriteProjects[] = new Favorite[3*2];
  private Menu                    widgetBookmarksMenu;
  private Menu                    widgetFavoriteProjectsMenu;
  private Menu                    widgetFavoriteProjectsSelectMenu;
  private SelectionListener       widgetFavoriteProjectsSelectMenuSelectionListener;

  private Menu                    widgetAllProjectsMenu1,widgetAllProjectsMenu2;
  private SelectionListener       widgetAllProjectsMenuSelectionListener1,widgetAllProjectsMenuSelectionListener2;

  private ScrolledComposite       widgetTimelineComposite;
  private Timeline                widgetTodayWindowTimeLine;
  private Pane                    widgetTodayPane;
  private Table                   widgetTodayTimeEntryTable;
  private Menu                    widgetTodayTimeEntryTableHeaderMenu,widgetTodayTimeEntryTableBodyMenu;
  private ActionTable             widgetTodayActionTable;

  private AddTimeEntryWidgets     addTimeEntry1,addTimeEntry2;

  private Tree                    widgetWeeksTimeEntryTree;
  private Menu                    widgetWeeksTimeEntryTreeHeaderMenu,widgetWeeksTimeEntryTreeBodyMenu;
  private Button                  widgetEditTimeEntries;
  private Button                  widgetAddTimeEntries;
  private Button                  widgetDeleteTimeEntries;

  private Tree                    widgetProjectTree;
  private Text                    widgetProjectTreeFind;
  private Timer                   filterProjectsTimer;
  private Button                  widgetEditProjectIssue;
  private Button                  widgetAddProjectIssue;
  private Button                  widgetDeleteProjectIssue;
  private Tree                    widgetIssuesTree;
  private Text                    widgetIssuesTreeFind;
  private Timer                   filterIssuesTimer;

  private Label                   widgetStatus;
  private SynchronizeAnimation    synchronizeAnimation;
  private Button                  widgetSynchronizeAnimation;

  private MenuItem                menuItemRefresh;
  private MenuItem                menuItemReinit;
  private MenuItem                menuItemSynchronize;

  private Menu                    menuTodaySetHours;
  private Menu                    menuTodayHoursPlus;
  private Menu                    menuTodayHoursMinus;

  private Menu                    menuWeekSetHours;
  private Menu                    menuWeekHoursPlus;
  private Menu                    menuWeekHoursMinus;

  private Object                  todayTimeLock = new Object();
  private Calendar                todayTimeStartDateTime = null;
  private WidgetVariable<Integer> todayTime = new WidgetVariable<Integer>(0);  // today time sum [min]
  private WidgetVariable<Long>    todayDurationSum = new WidgetVariable<Long>(0L);  // today working duration [s]
  private Timer                   updateTodayTimer = new Timer("updateTodayTimer",true);

  // selected start/end action in action table
  private Action                  startAction = null;
  private Action                  endAction   = null;

  private boolean                 quitFlag;

  // refresh server -> local database entries
  private PeriodicTask            refreshTask;
  private PeriodicTask            refreshIssuesTask;
  private Thread                  refreshTimeEntriesThread;
  private PeriodicTask            refreshTimeEntriesTask;
  private PeriodicTask            incrementalRefreshTask;

  // update time entries in today/week view
  private PeriodicTask            updateTimeEntriesTask;

  // synchronize local database entries -> server
  private PeriodicTask            synchronizeTask;

  // refresh server -> local database entries
  private PeriodicTask            autoRefreshTask;
  private Thread                  actionMonitorThread;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** print error to stderr
   * @param format format string
   * @param arguments optional arguments
   */
  public static void printError(String format, Object... arguments)
  {
    errorStream.println("ERROR: "+String.format(format,arguments));
  }

  /** print internal error to stderr
   * @param throwable throwable or null
   * @param format format string
   * @param arguments optional arguments
   */
  public static void printInternalError(Throwable throwable, String format, Object... arguments)
  {
    errorStream.println("INTERNAL ERROR: "+String.format(format,arguments));
    if (throwable != null)
    {
      printStackTrace(throwable,true);
    }
  }

  /** print internal error to stderr
   * @param throwable throwable
   * @param args optional arguments
   */
  public static void printInternalError(String format, Object... arguments)
  {
    printInternalError((Throwable)null,format,arguments);
  }

  /** print internal error to stderr
   * @param throwable throwable or null
   * @param format format string
   * @param arguments optional arguments
   */
  public static void printInternalError(Throwable throwable)
  {
    errorStream.println("INTERNAL ERROR: "+throwable.getMessage());
    printStackTrace(throwable,true);
  }

  /** print warning to stderr
   * @param format format string
   * @param arguments optional arguments
   */
  public static void printWarning(String format, Object... arguments)
  {
    if (Settings.debugTypes.contains(Settings.DebugTypes.WARNING))
    {
      errorStream.print("Warning: "+String.format(format,arguments));
      StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
      errorStream.print(" at "+stackTrace[2].getFileName()+", "+stackTrace[2].getLineNumber());
    }
    errorStream.println();
  }

  /** print debug to stderr
   * @param debugType debug type
   * @param format format string
   * @param arguments optional arguments
   */
  public static void printDebug(Settings.DebugTypes debugType, String format, Object... arguments)
  {
    if (Settings.debugTypes.contains(debugType))
    {
      errorStream.println("DEBUG "+DEBUG_TIME_FORMAT.format(new Date())+": "+String.format(format,arguments));
    }
  }

  /** print debug to stderr
   * @param debugType debug type
   * @param format format string
   * @param arguments optional arguments
   */
  public static void printDebug(String format, Object... arguments)
  {
    printDebug(Settings.DebugTypes.INFO,format,arguments);
  }


  /** print exception stack trace to stderr
   * @param exception exception to print
   * @param force true to force print
   */
  public static synchronized void printStackTrace(Throwable throwable, boolean force)
  {
    if (Settings.debugTypes.contains(Settings.DebugTypes.ERROR) || force)
    {
      errorStream.println("Stack trace:");
      for (StackTraceElement stackTraceElement : throwable.getStackTrace())
      {
        errorStream.println("  "+stackTraceElement);
      }
      Throwable cause = throwable.getCause();
      while (cause != null)
      {
        errorStream.println("Caused by "+cause.getClass().getName()+":");
        for (StackTraceElement stackTraceElement : cause.getStackTrace())
        {
          errorStream.println("  "+stackTraceElement);
        }

        cause = cause.getCause();
      }
    }
  }

  /** print stack trace to stderr (only in debug mode)
   * @param throwable throwable to print
   */
  public static synchronized void printStackTrace(Throwable throwable)
  {
    printStackTrace(throwable,false);
  }

  /** renice i/o exception (remove java.io.IOExcpetion text from exception)
   * @param exception i/o exception to renice
   * @return reniced exception
   */
  public static IOException reniceIOException(IOException exception)
  {
    final Pattern PATTERN = Pattern.compile("^(.*?)\\s*java.io.IOException: error=\\d+,\\s*(.*)$",Pattern.CASE_INSENSITIVE);

    Matcher matcher;
    if ((matcher = PATTERN.matcher(exception.getMessage())).matches())
    {
      exception = new IOException(matcher.group(1)+" "+matcher.group(2));
    }

    return exception;
  }

  /** main
   * @param args command line arguments
   */
  public static void main(String[] arguments)
  {
    new Timemine(arguments);
  }

  /** timemine main
   * @param arguments command line arguments
   */
  Timemine(String[] arguments)
  {
    final char MAIL_AT = '@';

    exitcode = EXITCODE_UNKNOWN;
    try
    {
      // load settings
      Settings.load();
      ActionMonitor.getInstance().setActionFilters(Settings.actionFilters);

      // parse arguments
      parseArguments(arguments);

      if (Settings.actionMonitorFlag)
      {
        // run as action monitor
        try
        {
          ActionMonitor actionMonitor = ActionMonitor.getInstance();

          BufferedWriter output = new BufferedWriter(new OutputStreamWriter(System.out,"UTF-8"));
          while (true)
          {
            // get active window
            System.out.println(ActionMonitor.PREFIX_WINDOW+actionMonitor.getCurrentWindowTitle());

            // get active application
            System.out.println(ActionMonitor.PREFIX_APPLICATION+actionMonitor.getCurrentApplication());

            // get idle
            System.out.println(ActionMonitor.PREFIX_IDLE+actionMonitor.getIdleTime());

            // sleep a short time
            try { Thread.sleep(30*1000); } catch (InterruptedException exception) { /* ignored */ }
          }
        }
        catch (Throwable throwable)
        {
          printInternalError(throwable);
        }
      }
      else
      {
        // run as main application

        // init
        initAll();

        if (exitcode == EXITCODE_UNKNOWN)
        {
          // run
          run();
        }

        // done
        doneAll();
      }
    }
    catch (org.eclipse.swt.SWTException exception)
    {
      errorStream.println("ERROR graphics: "+exception.getCause());
      printStackTrace(exception);
      errorStream.println("Version "+VERSION);
      errorStream.println("Please report this error to "+EMAIL+".");
    }
    catch (AssertionError error)
    {
      printInternalError(error);
      errorStream.println("Version "+VERSION);
      errorStream.println("Please report this error to "+EMAIL+".");
    }
    catch (InternalError error)
    {
      printInternalError(error);
      errorStream.println("Version "+VERSION);
      errorStream.println("Please report this error to "+EMAIL+".");
    }
    catch (Throwable throwable)
    {
      printInternalError(throwable);
      errorStream.println("Version "+VERSION);
      errorStream.println("Please report this error to "+EMAIL+".");
    }

    System.exit(exitcode);
  }

  //-----------------------------------------------------------------------

  /** static initializer
   */
  {
    // initialize file associations (Windows only)
    if (isWindowsSystem())
    {
    }
  }

  /** print program usage
   */
  private void printUsage()
  {
    System.out.println("timemine usage: <options> [--]");
    System.out.println("");
    System.out.println("Options: ");
    System.out.println("");
    System.out.println("  --host-[redmine|jira]=<name>                    - server name Redmine/JIRA");
    System.out.println("  -p|--port=<n>|--port-[redmine|jira]=<n>         - server port Redmine/JIRA");
    System.out.println("  --ssl|--ssl-[redmine|jira]                      - enable server SSL Redmine/JIRA");
    System.out.println("  -n|--name=<name>|--name-[redmine|jira]=<name>   - login name Redmine/JIRA");
    System.out.println("  --password|--password-[redmine|jira]=<password> - login password Redmine/JIRA (use with care!)");
    System.out.println("  --login-dialog                                  - show login dialog");
    System.out.println("  --refresh-at-start                              - refresh of local database at start");
    System.out.println("  --offline                                       - start offline");
    System.out.println("  -h|--help                                       - print this help");
    System.out.println("  --debug=[<value>,...]                           - enable debug output:");
    System.out.println("                                                      none");
    for (Enum enumeration : EnumSet.allOf(Settings.DebugTypes.class))
    {
      System.out.println("                                                      "+enumeration.name().toLowerCase());
    }
    System.out.println("                                                      all");
    System.out.println("  --work-in-progress                              - enable work-in-progress features");
    System.out.println("  --show-all-error-output                         - show all error output on stderr");
  }

  /** parse arguments
   * @param arguments arguments
   */
  private void parseArguments(String[] arguments)
  {
    // parse arguments
    int z = 0;
    boolean endOfOptions = false;
    while (z < arguments.length)
    {
      if      (!endOfOptions && arguments[z].equals("--"))
      {
        endOfOptions = true;
        z++;
      }
      else if (!endOfOptions && (arguments[z].startsWith("--") || arguments[z].startsWith("-")))
      {
        int i = Options.parse(OPTIONS,arguments,z,Settings.class);
        if (i < 0)
        {
          Timemine.printError("unknown option '"+arguments[z]+"'!");
          System.exit(EXITCODE_UNKNOWN_OPTION);
        }
        z = i;
      }
      else
      {
        printWarning("Ignored argument '%s'",arguments[z]);
        z++;
      }
    }

    // help
    if (Settings.helpFlag)
    {
      printUsage();
      System.exit(EXITCODE_OK);
    }

    // check arguments
  }

  /** check if system is Windows system
   * @return TRUE iff Windows, FALSE otherwise
   */
  private static boolean isWindowsSystem()
  {
    String osName = System.getProperty("os.name").toLowerCase();

    return (osName.indexOf("win") >= 0);
  }

  /** get favorite icon file names
   * @return icon file names
   */
  public static List<String> getFavoriteIconFilenames()
  {
    final File ICON_DIRECTORY = new File(Settings.TIMEMINE_DIRECTORY,"icons");

    ArrayList<String> iconNameList = new ArrayList<String>();

    File files[] = ICON_DIRECTORY.listFiles();
    if (files != null)
    {
      for (File file : files)
      {
        iconNameList.add(file.getName());
      }
    }
    Collections.sort(iconNameList);

    return iconNameList;
  }

  /** set selected favorite
   * @param favorite favorite
   */
  private void setSelectedFavorite(Favorite favorite)
  {
    if (selectedFavorite != null)
    {
      selectedFavorite.setSelection(false);
    }
    selectedFavorite = favorite;
    selectedFavorite.setSelection(true);
  }

  /** get selected favorite
   * @return selected favorite or null
   */
  private Favorite getSelectedFavorite()
  {
    return selectedFavorite;
  }

  /** set selected project
   * @param widgetProject projects widget
   * @param widgetIssueId issue ids widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary issue summary widget
   * @param widgetIssueDoneRatio issue done ratio widget
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param project project
   * @param filters filters or null
   */
  public static void setSelectedProject(TreeCombo  widgetProject,
                                        TableCombo widgetIssueId,
                                        TableCombo widgetIssueShortName,
                                        TreeCombo  widgetIssueSummary,
                                        Combo      widgetIssueDoneRatio,
                                        Label      widgetActivitiesTitle,
                                        Combo      widgetActivities,
                                        Project    project,
                                        String[]   filters
                                       )
  {
    // select project
    widgetProject.setSelection(project,project.name);

    // update issue
    updateIssues(widgetIssueId,widgetIssueShortName,widgetIssueSummary,project,filters);

    // update activities
    updateActivities(widgetActivitiesTitle,widgetActivities,project);
  }

  /** set selected project
   * @param widgetProject projects widget
   * @param widgetIssueId issue ids widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary issue summary widget
   * @param widgetIssueDoneRatio issue done ratio widget
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param project project
   */
  public static void setSelectedProject(TreeCombo  widgetProject,
                                        TableCombo widgetIssueId,
                                        TableCombo widgetIssueShortName,
                                        TreeCombo  widgetIssueSummary,
                                        Combo      widgetIssueDoneRatio,
                                        Label      widgetActivitiesTitle,
                                        Combo      widgetActivities,
                                        Project    project
                                       )
  {
    setSelectedProject(widgetProject,
                       widgetIssueId,
                       widgetIssueShortName,
                       widgetIssueSummary,
                       widgetIssueDoneRatio,
                       widgetActivitiesTitle,
                       widgetActivities,
                       project,
                       (String[])null  // filters
                      );
  }

  /** set selected project
   * @param widgetProject projects widget
   * @param widgetIssueId issue ids widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary issue summary widget
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param project project
   * @param issue issue
   */
  public static void setSelectedProject(TreeCombo  widgetProject,
                                        TableCombo widgetIssueId,
                                        TableCombo widgetIssueShortName,
                                        TreeCombo  widgetIssueSummary,
                                        Label      widgetActivitiesTitle,
                                        Combo      widgetActivities,
                                        Project    project
                                       )
  {
    setSelectedProject(widgetProject,
                       widgetIssueId,
                       widgetIssueShortName,
                       widgetIssueSummary,
                       (Combo)null,  // widgetIssueDoneRatio,
                       widgetActivitiesTitle,
                       widgetActivities,
                       project
                      );
  }

  /** set selected project
   * @param widgetProject projects widget
   * @param widgetIssueSummary issue summary widget
   * @param widgetIssueDoneRatio issue done ratio widget
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param project project
   */
  public static void setSelectedProject(TreeCombo widgetProject,
                                        TreeCombo widgetIssueSummary,
                                        Combo     widgetIssueDoneRatio,
                                        Label     widgetActivitiesTitle,
                                        Combo     widgetActivities,
                                        Project   project
                                       )
  {
    setSelectedProject(widgetProject,
                       (TableCombo)null,  // widgetIssueId
                       (TableCombo)null,  // widgetIssueShortName
                       widgetIssueSummary,
                       widgetIssueDoneRatio,
                       widgetActivitiesTitle,
                       widgetActivities,
                       project
                      );
  }

  /** set selected project
   * @param widgetProject projects widget
   * @param widgetIssueSummary issue summary widget
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param project project
   */
  public static void setSelectedProject(TreeCombo widgetProject,
                                        TreeCombo widgetIssueSummary,
                                        Label     widgetActivitiesTitle,
                                        Combo     widgetActivities,
                                        Project   project
                                       )
  {
    setSelectedProject(widgetProject,
                       widgetIssueSummary,
                       (Combo)null,  // widgetIssueDoneRatio
                       widgetActivitiesTitle,
                       widgetActivities,
                       project
                      );
  }

  /** set selected issue
   * @param widgetIssueId issue ids widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary issue summary widget
   * @param widgetIssueDoneRatio issue done ratio widget
   * @param issue issue
   */
  public static void setSelectedIssue(TableCombo widgetIssueId,
                                      TableCombo widgetIssueShortName,
                                      TreeCombo  widgetIssueSummary,
                                      Combo      widgetIssueDoneRatio,
                                      Issue      issue
                                     )
  {
// TODO:issue null?
    if (widgetIssueId != null)
    {
      // select issue id
      widgetIssueId.setSelection(issue,issue.id.format());
    }

    if (widgetIssueShortName != null)
    {
      // select issue short name
      widgetIssueShortName.setSelection(issue,(issue.shortName != null) ? issue.shortName : "");
    }

    // select summary
    widgetIssueSummary.setSelection(issue,"%s",(issue != Issue.NONE) ? issue.summary : "");

    // select done ratio
    if (widgetIssueDoneRatio != null)
    {
      if (issue.isEditProgressSupported())
      {
        widgetIssueDoneRatio.setEnabled(true);
        Widgets.setSelectedComboItem(widgetIssueDoneRatio,(issue != Issue.NONE) ? (Integer)issue.doneRatio : 0);
      }
      else
      {
        Widgets.setSelectedComboItem(widgetIssueDoneRatio,0);
        widgetIssueDoneRatio.setEnabled(false);
      }
    }
  }

  /** set selected issue
   * @param widgetIssueId issue ids widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary issue summary widget
   * @param issue issue
   */
  public static void setSelectedIssue(TableCombo widgetIssueId,
                                      TableCombo widgetIssueShortName,
                                      TreeCombo  widgetIssueSummary,
                                      Issue      issue
                                     )
  {
Dprintf.dprintf("TODO: required?");
    setSelectedIssue(widgetIssueId,widgetIssueShortName,widgetIssueSummary,(Combo)null,issue);
  }

  /** set selected issue
   * @param widgetIssueSummary issue summary widget
   * @param issue issue
   */
  public static void setSelectedIssue(TreeCombo widgetIssueSummary,
                                      Issue     issue
                                     )
  {
    setSelectedIssue((TableCombo)null,(TableCombo)null,widgetIssueSummary,issue);
  }

  /** init display variables
   */
  private void initDisplay()
  {
    // init display, clipboard
    display   = new Display();
//for (Monitor m : display.getMonitors()) { Dprintf.dprintf("m=%s %s",m,m.getBounds()); } System.exit(1);
    clipboard = new Clipboard(display);

    // create sync animation
    synchronizeAnimation = new SynchronizeAnimation(display);

    // get colors
    COLOR_BLACK                   = display.getSystemColor(SWT.COLOR_BLACK);
    COLOR_GRAY                    = display.getSystemColor(SWT.COLOR_GRAY);
    COLOR_GREEN                   = display.getSystemColor(SWT.COLOR_GREEN);
    COLOR_LIGHT_RED               = new Color(display,255,196,196);
    COLOR_ORANGE                  = display.getSystemColor(SWT.COLOR_YELLOW);
    COLOR_CYAN                    = display.getSystemColor(SWT.COLOR_CYAN);
    COLOR_WHITE                   = display.getSystemColor(SWT.COLOR_WHITE);
    COLOR_BACKGROUND              = display.getSystemColor(SWT.COLOR_WIDGET_BACKGROUND);
    COLOR_BACKGROUND_ODD          = new Color(display,0xF8,0xF7,0xF6);
    COLOR_BACKGROUND_EVEN         = display.getSystemColor(SWT.COLOR_WHITE);
    COLOR_TODAY_TIME_ENTRIES      = new ColorData(Settings.colorTodayTimeEntries);
    COLOR_TIME_ENTRIES            = new ColorData(Settings.colorTimeEntries);
    COLOR_TIME_ENTRIES_UPDATE     = new ColorData(Settings.colorTimeEntriesUpdate);
    COLOR_TIME_ENTRIES_INCOMPLETE = new ColorData(Settings.colorTimeEntriesIncomplete);
    COLOR_TIME_ENTRIES_WEEKEND    = new ColorData(Settings.colorTimeEntriesWeekend);
    COLOR_TIME_ENTRIES_HOLIDAY    = new ColorData(Settings.colorTimeEntriesHoliday);
    COLOR_FUTURE                  = new ColorData(Settings.colorFuture);
    COLOR_HOURS_EXCEEDED          = new ColorData(Settings.colorHoursExceeded);

    // get images
    IMAGE_SPLASH_SCREEN           = Widgets.loadImage(display,"splash-screen.png");
    IMAGE_EMPTY                   = Widgets.loadImage(display,"empty.png");
    IMAGE_PROGRAM_ICON            = Widgets.loadImage(display,"program-icon.png");
    IMAGE_CLIPBOARD               = Widgets.loadImage(display,"clipboard.png");
    IMAGE_FILTER                  = Widgets.loadImage(display,"filter.png");
    IMAGE_EDIT                    = Widgets.loadImage(display,"edit.png");
    IMAGE_ADD                     = Widgets.loadImage(display,"add.png");
    IMAGE_BOOKMARK                = Widgets.loadImage(display,"bookmark.png");
    IMAGE_DIRECTORY               = Widgets.loadImage(display,"directory.png");

    // get cursors
    CURSOR_WAIT                   = new Cursor(display,SWT.CURSOR_WAIT);

    // date/time formats
    DATE_FORMAT                   = new SimpleDateFormat(Settings.dateFormat);
    TIME_FORMAT                   = new SimpleDateFormat("HH:mm");
    DEBUG_TIME_FORMAT             = new SimpleDateFormat("HH:mm:ss.SSS");
  }

  /** init holiday days database
   */
  private void initHolidayDatabase()
  {
    final Settings.HolidayDate holidayDates[] =
    {
      // 2015
      // 2016
      // 2017
      new Settings.HolidayDate(2017, 1, 1),
      new Settings.HolidayDate(2017, 4,14),
      new Settings.HolidayDate(2017, 4,17),
      new Settings.HolidayDate(2017, 5, 1),
      new Settings.HolidayDate(2017, 5,25),
      new Settings.HolidayDate(2017, 6, 5),
      new Settings.HolidayDate(2017,10, 3),
      new Settings.HolidayDate(2017,10,31),
      new Settings.HolidayDate(2017,12,25),
      new Settings.HolidayDate(2017,12,26),
      // 2018
      new Settings.HolidayDate(2018, 1, 1),
      new Settings.HolidayDate(2018, 3,30),
      new Settings.HolidayDate(2018, 4, 2),
      new Settings.HolidayDate(2018, 5, 1),
      new Settings.HolidayDate(2018, 5,10,"Fronleichnam"),
      new Settings.HolidayDate(2018, 5,21),
      new Settings.HolidayDate(2018,10, 3),
      new Settings.HolidayDate(2018,12,25),
      new Settings.HolidayDate(2018,12,26),
      // 2019
      new Settings.HolidayDate(2019, 1, 1),
      new Settings.HolidayDate(2019, 4,19),
      new Settings.HolidayDate(2019, 4,22),
      new Settings.HolidayDate(2019, 5, 1),
      new Settings.HolidayDate(2019, 5,30),
      new Settings.HolidayDate(2019, 6,10),
      new Settings.HolidayDate(2019,10, 3),
      new Settings.HolidayDate(2019,12,25),
      new Settings.HolidayDate(2019,12,26),
      // 2020
      new Settings.HolidayDate(2020, 1, 1),
      new Settings.HolidayDate(2020, 4,10),
      new Settings.HolidayDate(2020, 4,13),
      new Settings.HolidayDate(2020, 5, 1),
      new Settings.HolidayDate(2020, 5,21),
      new Settings.HolidayDate(2020, 6, 1),
      new Settings.HolidayDate(2020,10, 3),
      new Settings.HolidayDate(2020,12,25),
      new Settings.HolidayDate(2020,12,26),
    };

    PreparedStatement preparedStatement = null;
    try
    {
      // open database
      database = new Database(Settings.TIMEMINE_DIRECTORY,"holidays");

      // create tables
      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS meta ( "+
                                                    "  id    INTEGER PRIMARY KEY, "+
                                                    "  name  TEXT UNIQUE, "+
                                                    "  value TEXT "+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO meta (name,value) VALUES ('version',?)");
      preparedStatement.setString(1,"1");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
      preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO meta (name,value) VALUES ('dateTime',DATE('now'))");
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS days ( "+
                                                    "  id          INTEGER, "+
                                                    "  dateTime    INTEGER UNIQUE, "+
                                                    "  description TEXT, "+
                                                    ""+
                                                    "  CONSTRAINT users UNIQUE (id)"+
                                                    ");"
                                                   );
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;

      // insert known holidays
      for (Settings.HolidayDate holidayDate : holidayDates)
      {
        preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO days (dateTime,description) VALUES (DATE(?,'unixepoch','localtime'),?)");
        preparedStatement.setLong(1,holidayDate.getDate());
        preparedStatement.setString(2,holidayDate.getDescription());
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;
      }

      // add deprecated holidays from settings
      for (Settings.HolidayDate holidayDate : Settings.holidayDateSet)
      {
        preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO days (dateTime,description) VALUES (DATE(?,'unixepoch','localtime'),?)");
        preparedStatement.setLong(1,holidayDate.getDate());
        preparedStatement.setString(2,holidayDate.getDescription());
        preparedStatement.executeUpdate();
        preparedStatement.close(); preparedStatement = null;
      }
    }
    catch (SQLException exception)
    {
      // ignored
      printStackTrace(exception);
    }
    finally
    {
      try { if (preparedStatement != null) preparedStatement.close(); } catch (SQLException exception) { /* ignored */ }
    }
  }

  /** done holiday days database
   */
  private void doneHolidayDatabase()
  {
    try
    {
      database.close();
    }
    catch (SQLException exception)
    {
      // ignored
    }
  }

  /** init loaded classes/JARs watchdog
   */
  private void initActionMonitor()
  {
    ActionMonitor.getInstance();
  }

  /** init loaded classes/JARs watchdog
   */
  private void initClassesWatchDog()
  {
    // get timestamp of all classes/JAR files
    final HashMap<File,Long> classModifiedMap = new HashMap<File,Long>();
    LinkedList<File> directoryList = new LinkedList<File>();
    for (String name : System.getProperty("java.class.path").split(File.pathSeparator))
    {
      File file = new File(name);
      if (file.isDirectory())
      {
        directoryList.add(file);
      }
      else
      {
        classModifiedMap.put(file,new Long(file.lastModified()));
      }
    }
    while (directoryList.size() > 0)
    {
      File directory = directoryList.removeFirst();
      File[] files = directory.listFiles();
      if (files != null)
      {
        for (File file : files)
        {
          if (file.isDirectory())
          {
            directoryList.add(file);
          }
          else
          {
            classModifiedMap.put(file,new Long(file.lastModified()));
          }
        }
      }
    }

    // periodically check timestamp of classes/JAR files and version file on server
    Thread classWatchDogThread = new Thread("Watch dog")
    {
      private String                  homepageVersionMajor    = null;
      private String                  homepageVersionMinor    = null;
      private String                  homepageVersionRevision = null;
      private final ArrayList<String> homepageChangeLog       = new ArrayList<String>();

      public void run()
      {
        final long REMINDER_TIME       =  5*60*1000; // [ms]
        final long VERSION_CHECK_TIME  = 60*60*1000; // [ms]

        long            lastRemindedTimestamp       = 0L;
        long            lastVersionCheckedTimestamp = 0L;
        final boolean[] reminderFlag                = new boolean[]{true};

        for (;;)
        {
          // check timestamps of classes/JAR files, show warning dialog
          for (final File file : classModifiedMap.keySet())
          {
            if (   reminderFlag[0]
                && (file.lastModified() > classModifiedMap.get(file))
                && (System.currentTimeMillis() > (lastRemindedTimestamp+REMINDER_TIME))
               )
            {
              display.syncExec(new Runnable()
              {
                public void run()
                {
                  switch (Dialogs.select(shell,
                                         "Warning","Class/JAR file '"+file.getName()+"' changed. It is recommended to restart Timemine now.",
                                         new String[]{"Restart","Remind me again in 5min","Ignore"},
                                         0
                                        )
                         )
                  {
                    case 0:
                      // send close event with restart
                      Widgets.notify(shell,USER_EVENT_QUIT,EXITCODE_RESTART);
                      break;
                    case 1:
                      break;
                    case 2:
                      reminderFlag[0] = false;
                      break;
                  }
                }
              });
              lastRemindedTimestamp = System.currentTimeMillis();
              break;
            }
          }

          if (Settings.showNewVersionInformation && (System.currentTimeMillis() > (lastVersionCheckedTimestamp+VERSION_CHECK_TIME)))
          {
            // get version on homepage, show warning dialog
            getServerVersionInfo();

            // check if newer version is available
            if (   ((homepageVersionMajor != null) && (homepageVersionMinor != null))
                && (   (homepageVersionMajor.compareTo(Config.VERSION_MAJOR) > 0)
                    || (homepageVersionMinor.compareTo(Config.VERSION_MINOR) > 0)
                   )
               )
            {
              final String homepageVersion = String.format("%s.%s%s",
                                                           homepageVersionMajor,
                                                           homepageVersionMinor,
                                                           (homepageVersionRevision != null) ? " (revision "+homepageVersionRevision+")" : ""
                                                          );
              display.syncExec(new Runnable()
              {
                public void run()
                {
                  showNewVersionInfo(shell);
                }
              });
            }
            lastVersionCheckedTimestamp = System.currentTimeMillis();
          }

          // sleep a short time
          try { Thread.sleep(30*1000); } catch (InterruptedException exception) { /* ignored */ }
        }
      }

      /** get servers version info
       */
      private void getServerVersionInfo()
      {
        final Pattern PATTERN_MAJOR    = Pattern.compile("MAJOR=(.*)");
        final Pattern PATTERN_MINOR    = Pattern.compile("MINOR=(.*)");
        final Pattern PATTERN_REVISION = Pattern.compile("REVISION=(.*)");

        BufferedReader input = null;
        try
        {
          input = new BufferedReader(new InputStreamReader(new URL(URL_VERSION_FILE).openStream(),"UTF-8"));

          // get version/change log available on server
          String  line;
          Matcher matcher;
          while (((line = input.readLine()) != null) && !line.isEmpty())
          {
//Dprintf.dprintf("homepage1 %s",line);
            if      ((matcher = PATTERN_MAJOR.matcher(line)).matches())
            {
              homepageVersionMajor    = matcher.group(1);
            }
            else if ((matcher = PATTERN_MINOR.matcher(line)).matches())
            {
              homepageVersionMinor    = matcher.group(1);
            }
            else if ((matcher = PATTERN_REVISION.matcher(line)).matches())
            {
              homepageVersionRevision = matcher.group(1);
            }
          }
          homepageChangeLog.clear();
          while ((line = input.readLine()) != null)
          {
//Dprintf.dprintf("homepage2 %s",line);
            homepageChangeLog.add(line);
          }

          input.close(); input = null;
        }
        catch (Exception exception)
        {
          // nothing to do
        }
        finally
        {
          try { input.close(); } catch (Exception exception) { /* nothing to do */ }
        }
      }

      /** show version info, change log
       * @param shell parent shell
       */
      private void showNewVersionInfo(final Shell shell)
      {
        Composite composite;
        Label     label;
        Text      text;
        Button    button;

        String version = String.format("%s.%s%s",
                                       homepageVersionMajor,
                                       homepageVersionMinor,
                                       (homepageVersionRevision != null) ? " (revision "+homepageVersionRevision+")" : ""
                                      );

        StringBuilder changeLog = new StringBuilder();
        for (String line : homepageChangeLog)
        {
          changeLog.append(line+Text.DELIMITER);
        }

        final Shell dialog = Dialogs.openModal(shell,"Confirmation",new double[]{1.0,0.0},1.0);
        dialog.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));

        final Button widgetShowAgain;
        composite = new Composite(dialog,SWT.NONE);
        composite.setLayout(new TableLayout(new double[]{0.0,1.0,0.0},new double[]{0.0,1.0},4));
        composite.setLayoutData(new TableLayoutData(0,0,TableLayoutData.NSWE));
        {
          // message
          label = Widgets.newImage(composite,Widgets.loadImage(shell.getDisplay(),"question.png"));
          Widgets.layout(label,0,0,TableLayoutData.W,0,0,10);

          label = Widgets.newLabel(composite,
                                   String.format("A newer version %s of Timemine is available. You\ncan download it from the Timemine homepage:\n\n%s\n\nChangeLog:",
                                                 version,
                                                 URL
                                                ),
                                   SWT.LEFT|SWT.WRAP
                                  );
          Widgets.layout(label,0,1,TableLayoutData.W,0,0,4);

          // change log
          text = Widgets.newStringView(composite,SWT.BORDER|SWT.WRAP);
          text.setText(changeLog.toString());
          Widgets.layout(text,1,1,TableLayoutData.NSWE,0,0,4);

          // show again
          widgetShowAgain = Widgets.newCheckbox(composite,"show again");
          widgetShowAgain.setSelection(true);
          Widgets.layout(widgetShowAgain,2,1,TableLayoutData.W,0,0,4);
        }

        // buttons
        composite = new Composite(dialog,SWT.NONE);
        composite.setLayout(new TableLayout(0.0,1.0));
        composite.setLayoutData(new TableLayoutData(1,0,TableLayoutData.WE,0,0,4));
        {
          button = Widgets.newButton(composite,"Open browser");
          button.setFocus();
          Widgets.layout(button,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,120,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              Dialogs.close(dialog,true);
            }
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
          });

          button = Widgets.newButton(composite,"Cancel");
          Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,120,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              Dialogs.close(dialog,false);
            }
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
          });
        }

        Dialogs.run(dialog,
                    true,
                    new DialogRunnable()
                    {
                      public void done(Object result)
                      {
                        if ((Boolean)result)
                        {
                          try
                          {
                            if (Desktop.isDesktopSupported())
                            {
                              Desktop.getDesktop().browse(new URI(URL));
                            }
                          }
                          catch (URISyntaxException exception)
                          {
                            Dialogs.error(shell,"Cannot open default web browser:\n\n"+exception.getMessage());
                          }
                          catch (IOException exception)
                          {
                            Dialogs.error(shell,"Cannot open default web browser:\n\n"+reniceIOException(exception).getMessage());
                          }
                          catch (Exception exception)
                          {
                            Dialogs.error(shell,"Cannot open default web browser:\n\n"+exception.getMessage());
                          }
                        }
                        else
                        {
                          // nothting to do
                        }
                        Settings.showNewVersionInformation = widgetShowAgain.getSelection();
                      }
                    }
                   );
      }
    };
    classWatchDogThread.setDaemon(true);
    classWatchDogThread.setPriority(Thread.MIN_PRIORITY);
    classWatchDogThread.start();
  }

  /** create main window
   */
  private void createWindow()
  {
    Composite   tab;
    TableColumn tableColumn;
    Composite   composite,subComposite;
    Button      button;
    Label       label;
    Combo       combo;
    Spinner     spinner;
    Text        text;
    Menu        menu,subMenu;
    MenuItem    menuItem;
    Control     control;
    TreeColumn  treeColumn;
    Pane        pane;

    // create window
    shell = new Shell(display,SWT.SHELL_TRIM);
    shell.setVisible(false);
    shell.setText("Timemine");
    shell.setImage(IMAGE_PROGRAM_ICON);
    shell.setLayout(new TableLayout(new double[]{1.0,0.0,0.0},1.0));
    shell.addShellListener(new ShellListener()
    {
      public void shellClosed(ShellEvent shellEvent)
      {
        shell.setVisible(false);
        shellEvent.doit = false;
      }
      public void shellActivated(ShellEvent e)
      {
      };
      public void shellDeactivated(ShellEvent e)
      {
      };
      public void shellDeiconified(ShellEvent e)
      {
      };
      public void shellIconified(ShellEvent e)
      {
      };
    });

    // create tabs
    widgetTabFolder        = Widgets.newTabFolder(shell);
    widgetTodayTime        = Widgets.newLabel(widgetTabFolder);
    widgetTodayDurationSum = Widgets.newLabel(widgetTabFolder);
    widgetTodayRestTime    = Widgets.newButton(widgetTabFolder);
    Widgets.layout(widgetTabFolder,0,0,TableLayoutData.NSWE);
    widgetTabFolder.addListener(USER_EVENT_ADD_TIME_ENTRY,new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        TabFolder widget = (TabFolder)event.widget;
        widget.redraw();
      }
    });
    widgetTabFolder.addListener(USER_EVENT_UPDATE_TIME_ENTRY,new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        TabFolder widget = (TabFolder)event.widget;
        widget.redraw();
      }
    });
    widgetTabFolder.addListener(USER_EVENT_REMOVE_TIME_ENTRY,new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        TabFolder widget = (TabFolder)event.widget;
        widget.redraw();
      }
    });
    widgetTabFolder.addListener(SWT.Resize,new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        TabFolder widget = (TabFolder)event.widget;
        GC        gc     = new GC(widget);
        Rectangle bounds = widget.getBounds();

        Point todayTimeSumSize     = Widgets.getTextSize(gc,widgetTodayTime.getText());
        Point todayDurationSumSize = Widgets.getTextSize(gc,widgetTodayDurationSum.getText());
        Point todayRestTextSize    = Widgets.getTextSize(gc,widgetTodayRestTime.getText());

        int w0 = todayTimeSumSize.x    +  4;
        int w1 = todayDurationSumSize.x+  4;
        int w2 = todayRestTextSize.x   +8+4;
        int h0 = todayTimeSumSize.y    +0;
        int h1 = todayDurationSumSize.y+0;
        int h2 = todayRestTextSize.y   +8;
        int h = 0;
        if (h0 > h) h = h0;
        if (h1 > h) h = h1;
        if (h2 > h) h = h2;

        int x0 = bounds.width-4-todayTimeSumSize.x-8-todayDurationSumSize.x-8-todayRestTextSize.x-8;
        int x1 = bounds.width-4-                     todayDurationSumSize.x-4-todayRestTextSize.x-8;
        int x2 = bounds.width-4-                                              todayRestTextSize.x-8;
        int y0 = (h-h0)/2;
        int y1 = (h-h1)/2;
        int y2 = (h-h2)/2;

        widget.redraw(bounds.x,bounds.y,bounds.width,bounds.height,true);
        widgetTodayTime.setBounds (x0,y0,w0,h0);
        widgetTodayDurationSum.setBounds(x1,y1,w1,h1);
        widgetTodayRestTime.setBounds(x2,y2,w2,h2);

        gc.dispose();
      }
    });
    Widgets.addModifyListener(new WidgetModifyListener(widgetTodayTime,todayTime)
    {
      @Override
      public void modified(Label label)
      {
        int n = todayTime.get();

        Color color = COLOR_GRAY;
        if (todayTimeStartDateTime != null)
        {
          if ((double)(n-Settings.pauseMinutesPerDay)/60.0 >= Settings.workingHoursPerDay)
          {
            color = COLOR_GREEN;
          }
          else
          {
            color = COLOR_ORANGE;
          }
        }
        label.setText(String.format("\u231a %s",formatHours(n)));
        label.setBackground(color);
      }
    });
    Widgets.addModifyListener(new WidgetModifyListener(widgetTodayDurationSum,todayDurationSum)
    {
      @Override
      public void modified(Label label)
      {
        long n = todayDurationSum.get();

        label.setText(String.format("\u03a3 %s",formatHours(n)));

        long restTime = (long)(Settings.workingHoursPerDay*60.0)-n;
        widgetTodayRestTime.setText(String.format("Rest %s",formatHours((restTime > 0L) ? restTime : 0L)));
        widgetTodayRestTime.setEnabled(restTime > 0.0);
      }
    });
    widgetTodayRestTime.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        long restTime = (long)(Settings.workingHoursPerDay*60.0)-todayDurationSum.get();
        if (restTime > 0L)
        {
          int hourFraction   = (int)(restTime/60);
          int minuteFraction = (int)(restTime%60);

          addTimeEntry1.widgetSpentHourFraction.setSelection(hourFraction);
          addTimeEntry1.widgetSpentMinuteFraction.setSelection(minuteFraction);
          addTimeEntry2.widgetSpentHourFraction.setSelection(hourFraction);
          addTimeEntry2.widgetSpentMinuteFraction.setSelection(minuteFraction);
          switch (widgetTabFolder.getSelectionIndex())
          {
            case 0: Widgets.setFocus(addTimeEntry1.widgetProject); break;
            case 1: Widgets.setFocus(addTimeEntry2.widgetProject); break;
          }
        }
      }
    });

    widgetTabFavoriteProjects = Widgets.addTab(widgetTabFolder,"Favorite projects ("+Widgets.acceleratorToText(Settings.keyTabFavoriteProjects)+")");
    widgetTabFavoriteProjects.setLayout(new TableLayout(new double[]{1.0,0.0},1.0,2));
    Widgets.layout(widgetTabFavoriteProjects,0,0,TableLayoutData.NSWE);
    {
      composite = Widgets.newComposite(widgetTabFavoriteProjects);
      composite.setLayout(new TableLayout(1.0,1.0));
      Widgets.layout(composite,0,0,TableLayoutData.NSWE);
      {
        widgetFavoriteProjectsMenu = Widgets.newPopupMenu(shell);
        {
          widgetFavoriteProjectsSelectMenu = Widgets.addMenu(widgetFavoriteProjectsMenu,"Select project");

          subMenu = Widgets.addMenu(widgetFavoriteProjectsMenu,"Select image");
          {
            menuItem = Widgets.addMenuItem(subMenu,"From file"+"\u2026",SWT.NONE);
            menuItem.addSelectionListener(new SelectionListener()
            {
              @Override
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              @Override
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                Favorite selectedFavorite = getSelectedFavorite();
                if (selectedFavorite != null)
                {
                  String fileName = Dialogs.fileOpen(shell,
                                                     "Select image",
                                                     "",
                                                     new String[]
                                                     {
                                                       "PNG","*.png",
                                                       "All","*"
                                                     }
                                                    );
                  if (fileName != null)
                  {
                    selectedFavorite.addIcon(fileName);
                  }
                }
              }
            });

            Widgets.addMenuItemSeparator(subMenu);

            for (final String fileName : getFavoriteIconFilenames())
            {
              menuItem = Widgets.addMenuItem(subMenu,fileName,SWT.NONE);
              menuItem.setData(fileName);
              menuItem.addSelectionListener(new SelectionListener()
              {
                @Override
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                @Override
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  selectedFavorite.setIcon(fileName);
                }
              });
            }
          }

          Widgets.addMenuItemSeparator(widgetFavoriteProjectsMenu);

          menuItem = Widgets.addMenuItem(widgetFavoriteProjectsMenu,"New issue"+"\u2026",SWT.NONE);
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              Favorite selectedFavorite = getSelectedFavorite();
              if (selectedFavorite != null)
              {
                Project project = selectedFavorite.getProject();
                if (project != null)
                {
                  try
                  {
                    // add new issue
                    addIssue(project);
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    Dialogs.error(shell,"Cannot add new issue (error: "+exception.getMessage()+")");
                    return;
                  }
                }
                else
                {
                  Dialogs.error(shell,"Cannot add new issue (error: invalid project)");
                  return;
                }
              }
            }
          });
        }
        widgetFavoriteProjectsMenu.addListener(USER_EVENT_UPDATE_PROJECTS,new Listener()
        {
          @Override
          public void handleEvent(Event event)
          {
            ProjectList projectList = (ProjectList)event.data;
            updateProjectsMenu(widgetFavoriteProjectsSelectMenu,widgetFavoriteProjectsSelectMenuSelectionListener,projectList);
            updateProjectsMenu(widgetAllProjectsMenu1,widgetAllProjectsMenuSelectionListener1,projectList);
            updateProjectsMenu(widgetAllProjectsMenu2,widgetAllProjectsMenuSelectionListener2,projectList);
          }
        });
        widgetFavoriteProjectsSelectMenuSelectionListener = new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem menuItem = (MenuItem)selectionEvent.widget;

            Favorite selectedFavorite = getSelectedFavorite();
            if (selectedFavorite != null)
            {
              int index = selectedFavorite.getIndex();
              Project project = (Project)menuItem.getData();

              if (index >= Settings.favoriteProjects.length)
              {
                Settings.favoriteProjects = Arrays.copyOf(Settings.favoriteProjects,Settings.favoriteProjects.length+1);
              }
              if (index >= Settings.favoriteProjects.length)
              {
                Settings.favoriteProjects = Arrays.copyOf(Settings.favoriteProjects,index+1);
              }
              Settings.favoriteProjects[index] = project.name;
// TODO: check if external modified
              Settings.save();

              selectedFavorite.setName(project.name);
            }
          }
        };

        MouseListener mouseListener = new MouseListener()
        {
          @Override
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseDown(MouseEvent mouseEvent)
          {
            // select favorite
            Favorite favorite = (Favorite)mouseEvent.widget;
            setSelectedFavorite(favorite);

            switch (mouseEvent.button)
            {
              case 1:
                // select project (if any)
                Project project = favorite.getProject();
                if (project != null)
                {
                  setSelectedProject(addTimeEntry1.widgetProject,
                                     addTimeEntry1.widgetIssueId,
                                     addTimeEntry1.widgetIssueShortName,
                                     addTimeEntry1.widgetIssueSummary,
                                     addTimeEntry1.widgetIssueDoneRatio,
                                     addTimeEntry1.widgetActivitiesTitle,
                                     addTimeEntry1.widgetActivities,
                                     project
                                    );
                  setSelectedProject(addTimeEntry2.widgetProject,
                                     addTimeEntry2.widgetIssueId,
                                     addTimeEntry2.widgetIssueShortName,
                                     addTimeEntry2.widgetIssueSummary,
                                     addTimeEntry2.widgetIssueDoneRatio,
                                     addTimeEntry2.widgetActivitiesTitle,
                                     addTimeEntry2.widgetActivities,
                                     project
                                    );
                  Widgets.setFocus(addTimeEntry1.widgetComments);
                }
                break;
              case 3:
                // show context menu
                Point p = selectedFavorite.toDisplay(mouseEvent.x,mouseEvent.y);
                widgetFavoriteProjectsMenu.setLocation(p);
                widgetFavoriteProjectsMenu.setVisible(true);
                break;
              default:
                break;
            }
          }
          @Override
          public void mouseUp(MouseEvent mouseEvent)
          {
            widgetFavoriteProjectsMenu.setVisible(false);
          }
        };

        widgetFavoriteProjects[0] = new Favorite(composite,0,(Settings.favoriteProjects.length > 0) ? Settings.favoriteProjects[0] : "");
        Widgets.layout(widgetFavoriteProjects[0],0,0,TableLayoutData.NSWE);
        widgetFavoriteProjects[0].addMouseListener(mouseListener);

        widgetFavoriteProjects[1] = new Favorite(composite,1,(Settings.favoriteProjects.length > 1) ? Settings.favoriteProjects[1] : "");
        Widgets.layout(widgetFavoriteProjects[1],0,1,TableLayoutData.NSWE);
        widgetFavoriteProjects[1].addMouseListener(mouseListener);

        widgetFavoriteProjects[2] = new Favorite(composite,2,(Settings.favoriteProjects.length > 2) ? Settings.favoriteProjects[2] : "");
        Widgets.layout(widgetFavoriteProjects[2],0,2,TableLayoutData.NSWE);
        widgetFavoriteProjects[2].addMouseListener(mouseListener);

        widgetFavoriteProjects[3] = new Favorite(composite,3,(Settings.favoriteProjects.length > 3) ? Settings.favoriteProjects[3] : "");
        Widgets.layout(widgetFavoriteProjects[3],1,0,TableLayoutData.NSWE);
        widgetFavoriteProjects[3].addMouseListener(mouseListener);

        widgetFavoriteProjects[4] = new Favorite(composite,4,(Settings.favoriteProjects.length > 4) ? Settings.favoriteProjects[4] : "");
        Widgets.layout(widgetFavoriteProjects[4],1,1,TableLayoutData.NSWE);
        widgetFavoriteProjects[4].addMouseListener(mouseListener);

        widgetFavoriteProjects[5] = new Favorite(composite,5,(Settings.favoriteProjects.length > 5) ? Settings.favoriteProjects[5] : "");
        Widgets.layout(widgetFavoriteProjects[5],1,2,TableLayoutData.NSWE);
        widgetFavoriteProjects[5].addMouseListener(mouseListener);
      }
      composite.addListener(USER_EVENT_UPDATE_PROJECTS,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          widgetFavoriteProjectsMenu.notifyListeners(event.type,event);
        }
      });

      // add new entry
      addTimeEntry1 = new AddTimeEntryWidgets(widgetTabFavoriteProjects,SWT.NONE);
      addTimeEntry1.setLayout(new TableLayout(0.0,new double[]{0.0,1.0},2));
      Widgets.layout(addTimeEntry1,1,0,TableLayoutData.WE);
    }

    widgetTabDay = Widgets.addTab(widgetTabFolder,"Day ("+Widgets.acceleratorToText(Settings.keyTabTodayTimeEntries)+")");
    widgetTabDay.setLayout(new TableLayout(new double[]{0.0,1.0},1.0,2));
    Widgets.layout(widgetTabDay,0,0,TableLayoutData.NSWE);
    {
      // today time line
      widgetTimelineComposite = Widgets.newScrolledComposite(widgetTabDay,SWT.H_SCROLL);
      widgetTimelineComposite.setLayout(new TableLayout(0.0,new double[]{1.0,0.0}));
      Widgets.layout(widgetTimelineComposite,0,0,TableLayoutData.WE);
      {
        widgetTodayWindowTimeLine = new Timeline(widgetTimelineComposite,1024);
        widgetTodayWindowTimeLine.setToolTipText("Click to add tag.");
        Widgets.layout(widgetTodayWindowTimeLine,0,0,TableLayoutData.WE,0,0,0,0,SWT.DEFAULT,65);
        ActionMonitor.getInstance().addActionListener(new ActionMonitor.ActionListener()
        {
          @Override
          public void refresh(final ActionList actionList)
          {
            final Calendar now = Calendar.getInstance();

            if (!display.isDisposed())
            {
              display.asyncExec(new Runnable()
              {
                public void run()
                {
                  // refresh time line
                  if (!widgetTodayActionTable.isDisposed())
                  {
                    widgetTodayWindowTimeLine.refresh(actionList);
                  }
                }
              });
            }
          }
        });
        widgetTodayWindowTimeLine.addMouseListener(new MouseListener()
        {
          @Override
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseDown(MouseEvent mouseEvent)
          {
            // set start time
            long timestamp = widgetTodayWindowTimeLine.toTimestamp(mouseEvent.x);
            if (Settings.snapTimeRanges)
            {
              // get nearest action
              Action nearestAction = ActionMonitor.getInstance().getNearest(timestamp);
              if (nearestAction != null)
              {
                timestamp = nearestAction.timestamp;
              }
              else
              {
                return;
              }
            }
            widgetTodayWindowTimeLine.setTagStartTime(timestamp);
          }
          @Override
          public void mouseUp(MouseEvent mouseEvent)
          {
            // set end time
            long timestamp = widgetTodayWindowTimeLine.toTimestamp(mouseEvent.x);
            if (Settings.snapTimeRanges)
            {
              // get nearest action
              Action nearestAction = ActionMonitor.getInstance().getNearest(timestamp);
              if (nearestAction != null)
              {
                timestamp = nearestAction.timestamp;
              }
              else
              {
                return;
              }
            }
            widgetTodayWindowTimeLine.setTagEndTime(timestamp);

            // get time range
            long startTimestamp = widgetTodayWindowTimeLine.getTagStartTime();
            long endTimestamp   = widgetTodayWindowTimeLine.getTagEndTime();

            // add new tag
            Tag tag = new Tag(startTimestamp,endTimestamp);
            if (editTag(tag,"Add tag","Add"))
            {
              try
              {
                ActionMonitor.getInstance().addTag(tag);
              }
              catch (SQLException exception)
              {
                Dialogs.error(shell,"Cannot add tag (error: "+exception.getMessage()+")");
              }
            }

            // clear tag
            widgetTodayWindowTimeLine.clearTag();
          }
        });
        widgetTodayWindowTimeLine.addMouseMoveListener(new MouseMoveListener()
        {
          @Override
          public void mouseMove(MouseEvent mouseEvent)
          {
            long timestamp = widgetTodayWindowTimeLine.toTimestamp(mouseEvent.x);
            if (Settings.snapTimeRanges)
            {
              // get nearst action
              Action nearestAction = ActionMonitor.getInstance().getNearest(timestamp);
              if (nearestAction != null)
              {
                timestamp = nearestAction.timestamp;
              }
              else
              {
                return;
              }
            }
            widgetTodayWindowTimeLine.setTagEndTime(timestamp);
          }
        });
        widgetTodayWindowTimeLine.addMouseWheelListener(new MouseWheelListener()
        {
          @Override
          public void mouseScrolled(MouseEvent mouseEvent)
          {
            double increment = 0;
            if      (mouseEvent.count > 0) increment =  0.1;
            else if (mouseEvent.count < 0) increment = -0.1;

            Point origin = widgetTimelineComposite.getOrigin();
            int x0 = origin.x+mouseEvent.x;
            long timestamp = widgetTodayWindowTimeLine.toTimestamp(x0);
//Dprintf.dprintf("w=%d orig=%s x0=%d datetime=%d",w,origin,x0,timestamp);

            widgetTodayWindowTimeLine.incrementZoomFactor(increment,10.0);

            int x1 = widgetTodayWindowTimeLine.toX(timestamp)-mouseEvent.x;
            if (x1 < 0) x1 = 0;

            origin.x = x1;
//Dprintf.dprintf("xd=%d x1=%d orig=%s",widgetTodayWindowTimeLine.toX(dateTime),x1,origin);
            widgetTimelineComposite.setOrigin(origin);
          }
        });
//TODO: remove?
/*
        TreeCombo widgetActivities = new TreeCombo(composite,100,TreeCombo.DEFAULT_HEIGHT,SWT.NONE);
        widgetActivities.addTreeColumn("Time",    SWT.RIGHT, 80,false);
        widgetActivities.addTreeColumn("Activity",SWT.LEFT,  0, true );
        widgetActivities.setLayout(new double[]{0.0,1.0});
//        widgetActivities.setToolTipText("Parent issue.");
        Widgets.layout(widgetActivities,0,1,TableLayoutData.WE);*/
      }

      // create pane
      widgetTodayPane = Widgets.newPane(widgetTabDay,2,SWT.VERTICAL);
      Widgets.layout(widgetTodayPane,1,0,TableLayoutData.NSWE);

      // today time entry table
      composite = widgetTodayPane.getComposite(0);
      composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));

      final String TIME_ENTRY_COLUMN_NAMES[] = new String[]{"Id","Hours","Activity","Project","Issue short name","Issue","Comments"};

      SelectionListener selectionListener = new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          TableColumn         tableColumn         = (TableColumn)selectionEvent.widget;
          TimeEntryComparator timeEntryComparator = new TimeEntryComparator();

          if      (tableColumn == widgetTodayTimeEntryTable.getColumn(0)) timeEntryComparator.setSortMode(TimeEntryComparator.SortModes.SPENT_ON);
          else if (tableColumn == widgetTodayTimeEntryTable.getColumn(1)) timeEntryComparator.setSortMode(TimeEntryComparator.SortModes.ID      );
          else if (tableColumn == widgetTodayTimeEntryTable.getColumn(2)) timeEntryComparator.setSortMode(TimeEntryComparator.SortModes.HOURS   );
          else if (tableColumn == widgetTodayTimeEntryTable.getColumn(3)) timeEntryComparator.setSortMode(TimeEntryComparator.SortModes.ACTIVITY);
          else if (tableColumn == widgetTodayTimeEntryTable.getColumn(4)) timeEntryComparator.setSortMode(TimeEntryComparator.SortModes.PROJECT );
          else if (tableColumn == widgetTodayTimeEntryTable.getColumn(5)) timeEntryComparator.setSortMode(TimeEntryComparator.SortModes.ISSUE   );
          else if (tableColumn == widgetTodayTimeEntryTable.getColumn(6)) timeEntryComparator.setSortMode(TimeEntryComparator.SortModes.ISSUE   );
          else if (tableColumn == widgetTodayTimeEntryTable.getColumn(7)) timeEntryComparator.setSortMode(TimeEntryComparator.SortModes.COMMENTS);
          Widgets.sortTableColumn(widgetTodayTimeEntryTable,tableColumn,timeEntryComparator);
        }
      };

      widgetTodayTimeEntryTable = Widgets.newTable(composite,SWT.MULTI);
      widgetTodayTimeEntryTable.setToolTipText("Time entries of today.\nENTER/RETURN to edit entry.\nDEL/BACKSPACE to delete entries.");
      widgetTodayTimeEntryTable.setLayout(new TableLayout(null,1.0));
      Widgets.layout(widgetTodayTimeEntryTable,0,0,TableLayoutData.NSWE);
      tableColumn = Widgets.addTableColumn(widgetTodayTimeEntryTable,0,TIME_ENTRY_COLUMN_NAMES[0],SWT.RIGHT, 50,true );
      tableColumn.setMoveable(true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetTodayTimeEntryTable,1,TIME_ENTRY_COLUMN_NAMES[1],SWT.RIGHT, 50,true );
      tableColumn.setMoveable(true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetTodayTimeEntryTable,2,TIME_ENTRY_COLUMN_NAMES[2],SWT.LEFT,  70,true );
      tableColumn.setMoveable(true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetTodayTimeEntryTable,3,TIME_ENTRY_COLUMN_NAMES[3],SWT.LEFT,  80,true );
      tableColumn.setMoveable(true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetTodayTimeEntryTable,4,TIME_ENTRY_COLUMN_NAMES[4],SWT.LEFT,  80,true );
      tableColumn.setMoveable(true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetTodayTimeEntryTable,5,TIME_ENTRY_COLUMN_NAMES[5],SWT.LEFT, 100,true );
      tableColumn.setMoveable(true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetTodayTimeEntryTable,6,TIME_ENTRY_COLUMN_NAMES[6],SWT.LEFT, 200,true );
      tableColumn.setMoveable(true);
      tableColumn.addSelectionListener(selectionListener);
      Widgets.setTableColumnWidth(widgetTodayTimeEntryTable,Settings.geometryTodayTimeEntryColumns.width);
      widgetTodayTimeEntryTable.setColumnOrder(Settings.geometryTodayTimeEntryColumnOrder.getMap(TIME_ENTRY_COLUMN_NAMES));
      Listener todayTimeEntryTableColumnResizeListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          if (Settings.isPermitModify())
          {
            Settings.geometryTodayTimeEntryColumns = new Settings.ColumnSizes(Widgets.getTableColumnWidth(widgetTodayTimeEntryTable));
            Settings.setModified();
          }
        }
      };
      Listener todayTimeEntryTableColumnMoveListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          int[]    columnOrder = widgetTodayTimeEntryTable.getColumnOrder();
          String[] names       = new String[columnOrder.length];
          for (int i = 0; i < columnOrder.length; i++)
          {
            names[i] = TIME_ENTRY_COLUMN_NAMES[columnOrder[i]];
          }
          if (Settings.isPermitModify())
          {
            Settings.geometryTodayTimeEntryColumnOrder = new SettingUtils.SimpleStringArray(names);
            Settings.setModified();
          }
        }
      };
      for (TableColumn orderTableColumn : widgetTodayTimeEntryTable.getColumns())
      {
        orderTableColumn.setMoveable(true);
        orderTableColumn.addListener(SWT.Move,todayTimeEntryTableColumnMoveListener);
        orderTableColumn.addListener(SWT.Resize,todayTimeEntryTableColumnResizeListener);
      }

      widgetTodayTimeEntryTableHeaderMenu = Widgets.newPopupMenu(shell);
      {
        menuItem = Widgets.addMenuItemCheckbox(widgetTodayTimeEntryTableHeaderMenu,TIME_ENTRY_COLUMN_NAMES[0]);
        menuItem.setSelection(Settings.geometryTodayTimeEntryColumns.get(0) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetTodayTimeEntryTable,0,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetTodayTimeEntryTableHeaderMenu,TIME_ENTRY_COLUMN_NAMES[1]);
        menuItem.setSelection(Settings.geometryTodayTimeEntryColumns.get(1) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetTodayTimeEntryTable,1,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetTodayTimeEntryTableHeaderMenu,TIME_ENTRY_COLUMN_NAMES[2]);
        menuItem.setSelection(Settings.geometryTodayTimeEntryColumns.get(2) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetTodayTimeEntryTable,2,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetTodayTimeEntryTableHeaderMenu,TIME_ENTRY_COLUMN_NAMES[3]);
        menuItem.setSelection(Settings.geometryTodayTimeEntryColumns.get(3) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetTodayTimeEntryTable,3,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetTodayTimeEntryTableHeaderMenu,TIME_ENTRY_COLUMN_NAMES[4]);
        menuItem.setSelection(Settings.geometryTodayTimeEntryColumns.get(4) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetTodayTimeEntryTable,4,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetTodayTimeEntryTableHeaderMenu,TIME_ENTRY_COLUMN_NAMES[5]);
        menuItem.setSelection(Settings.geometryTodayTimeEntryColumns.get(5) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetTodayTimeEntryTable,5,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetTodayTimeEntryTableHeaderMenu,TIME_ENTRY_COLUMN_NAMES[6]);
        menuItem.setSelection(Settings.geometryTodayTimeEntryColumns.get(6) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetTodayTimeEntryTable,6,widget.getSelection());
          }
        });
      }

      widgetTodayTimeEntryTableBodyMenu = Widgets.newPopupMenu(shell);
      {
        menuTodaySetHours = Widgets.addMenu(widgetTodayTimeEntryTableBodyMenu,"Set hours"+"\u2026");
        for (int i = 1; i <= 8*60/Settings.minTimeDelta; i++)
        {
          menuItem = Widgets.addMenuItem(menuTodaySetHours,new Integer(i*Settings.minTimeDelta),formatHours(i*Settings.minTimeDelta));
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              MenuItem menuItem = (MenuItem)selectionEvent.widget;
              int      duration = (Integer)menuItem.getData();

              TableItem[] tableItems = widgetTodayTimeEntryTable.getSelection();
              if (tableItems.length > 0)
              {
                if (tableItems[0].getData() instanceof TimeEntry)
                {
                  TimeEntry timeEntry = (TimeEntry)tableItems[0].getData();
                  assert timeEntry != null;

                  // set duration
                  timeEntry.duration = duration;

                  try
                  {
                    // update time entry
                    timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                    Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    final String message = exception.getMessage();
                    Dialogs.error(shell,"Cannot update time entry (error: "+message+")");
                    return;
                  }
                  catch (InterruptedException exception)
                  {
                    // ignored
                  }
                }
              }
            }
          });
        }
        menuTodayHoursPlus = Widgets.addMenu(widgetTodayTimeEntryTableBodyMenu,"+");
        for (int i = 1; i <= 8*60/Settings.minTimeDelta; i++)
        {
          menuItem = Widgets.addMenuItem(menuTodayHoursPlus,new Integer(i*Settings.minTimeDelta),formatHours(i*Settings.minTimeDelta));
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              MenuItem menuItem = (MenuItem)selectionEvent.widget;
              int      delta    = (Integer)menuItem.getData();

              TableItem[] tableItems = widgetTodayTimeEntryTable.getSelection();
              if (tableItems.length > 0)
              {
                if (tableItems[0].getData() instanceof TimeEntry)
                {
                  TimeEntry timeEntry = (TimeEntry)tableItems[0].getData();
                  assert timeEntry != null;
Dprintf.dprintf("timeentry=%s",timeEntry);


                  // increment duration
                  timeEntry.duration += delta;

                  try
                  {
                    // update time entry
                    timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                    Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    final String message = exception.getMessage();
                    Dialogs.error(shell,"Cannot update time entry (error: "+message+")");
                    return;
                  }
                  catch (InterruptedException exception)
                  {
                    // ignored
                  }
/**/
                }
              }
            }
          });
        }
        menuTodayHoursMinus = Widgets.addMenu(widgetTodayTimeEntryTableBodyMenu,"-");
        for (int i = 1; i <= 8*60/Settings.minTimeDelta; i++)
        {
          menuItem = Widgets.addMenuItem(menuTodayHoursMinus,new Integer(i*Settings.minTimeDelta),formatHours(i*Settings.minTimeDelta));
          menuItem.setData(new Integer(i*Settings.minTimeDelta));
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              MenuItem menuItem = (MenuItem)selectionEvent.widget;
              int      delta    = (Integer)menuItem.getData();

              TableItem[] tableItems = widgetTodayTimeEntryTable.getSelection();
              if (tableItems.length > 0)
              {
                if (tableItems[0].getData() instanceof TimeEntry)
                {
                  TimeEntry timeEntry = (TimeEntry)tableItems[0].getData();
                  assert timeEntry != null;
Dprintf.dprintf("timeentry=%s",timeEntry);

                  if      (timeEntry.duration > delta)
                  {
                    // decrement duration
                    timeEntry.duration -= delta;

                    // update time entry
                    try
                    {
                      timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                      Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
                    }
                    catch (BackendException exception)
                    {
                      printStackTrace(exception);
                      final String message = exception.getMessage();
                      Dialogs.error(shell,"Cannot update time entry (error: "+message+")");
                      return;
                    }
                    catch (InterruptedException exception)
                    {
                      // ignored
                    }
                  }
                  else if (Dialogs.confirm(shell,"Confirmation","Spent hours will become zero - delete time entry?","Delete","Cancel"))
                  {
                    // remove time entry
                    try
                    {
                      removeTimeEntry(timeEntry);
                    }
                    catch (BackendException exception)
                    {
                      printStackTrace(exception);
                      final String message = exception.getMessage();
                      Dialogs.error(shell,"Cannot remove time entry (error: "+message+")");
                      return;
                    }
                  }
                }
              }
            }
          });
        }

        Widgets.addMenuItemSeparator(widgetTodayTimeEntryTableBodyMenu);

        menuItem = Widgets.addMenuItem(widgetTodayTimeEntryTableBodyMenu,"Add new time entry"+"\u2026",Settings.keyNewTimeEntry);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetAddTimeEntries);
          }
        });
        menuItem = Widgets.addMenuItem(widgetTodayTimeEntryTableBodyMenu,"Edit time entry"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TableItem[] tableItems = widgetTodayTimeEntryTable.getSelection();
            if (tableItems.length > 0)
            {
              final TimeEntry timeEntry = (TimeEntry)tableItems[0].getData();

              TimeEntry oldTimeEntry = timeEntry.clone();
              if (editTimeEntry(timeEntry,"Edit time entry","update","Save"))
              {
                try
                {
                  timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                  Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  final String message = exception.getMessage();
                  display.syncExec(new Runnable()
                  {
                    public void run()
                    {
                      Dialogs.error(shell,"Cannot update time entry (error: "+message+")");
                    }
                  });
                  return;
                }
                catch (InterruptedException exception)
                {
                  // ignored
                }
              }
            }
          }
        });
        menuItem = Widgets.addMenuItem(widgetTodayTimeEntryTableBodyMenu,"Delete time entries"+"\u2026",Settings.keyDeleteTimeEntry);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TableItem[] tableItems = widgetTodayTimeEntryTable.getSelection();
            if (tableItems.length > 0)
            {
              if (Dialogs.confirm(shell,"Delete "+tableItems.length+" time entries?"))
              {
                try
                {
                  for (TableItem tableItem : tableItems)
                  {
                    if (tableItem != null)
                    {
                      TimeEntry timeEntry = (TimeEntry)tableItem.getData();
                      removeTimeEntry(timeEntry);
                    }
                  }
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot delete time entries (error: "+exception.getMessage()+")");
                  return;
                }
              }
            }
          }
        });
        menuItem = Widgets.addMenuItem(widgetTodayTimeEntryTableBodyMenu,"Delete local time entries"+"\u2026",Settings.keyDeleteTimeEntry);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TableItem[] tableItems = widgetTodayTimeEntryTable.getSelection();
            if (tableItems.length > 0)
            {
              if (Dialogs.confirm(shell,"Delete "+tableItems.length+" local time entries?"))
              {
/*
                try
                {
                  for (TableItem tableItem : tableItems)
                  {
                    if (tableItem != null)
                    {
                      TimeEntry timeEntry = (TimeEntry)tableItem.getData();
Dprintf.dprintf("");
                      removeTimeEntry(timeEntry);
                    }
                  }
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot delete time entries (error: "+exception.getMessage()+")");
                  return;
                }
*/
              }
            }
          }
        });

        Widgets.addMenuItemSeparator(widgetTodayTimeEntryTableBodyMenu);

        menuItem = Widgets.addMenuItem(widgetTodayTimeEntryTableBodyMenu,"Edit issue"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TableItem[] tableItems = widgetTodayTimeEntryTable.getSelection();
            if (tableItems.length > 0)
            {
              // get time entry
              TimeEntry timeEntry = (TimeEntry)tableItems[0].getData();

              // edit issue
              Issue issue = timeEntry.getIssue();
              if (editIssue(issue,"Edit issue","Save"))
              {
                try
                {
                  issue.setModified(Backend.ModifiedStates.UPDATED);
                  Widgets.notify(shell,USER_EVENT_UPDATE_ISSUE,issue);
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot update issue (error: "+exception.getMessage()+")");
                  return;
                }
                catch (InterruptedException exception)
                {
                  // ignored
                }
              }
            }
          }
        });
      }
      widgetTodayTimeEntryTable.setMenu(widgetTodayTimeEntryTableBodyMenu);

      widgetTodayTimeEntryTable.addListener(SWT.MenuDetect, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Table widget = (Table)event.widget;

          Point     p          = display.map(null, widget, new Point(event.x, event.y));
          Rectangle clientArea = widget.getClientArea();
          boolean   isInHeader =    (clientArea.y <= p.y)
                                 && (p.y < (clientArea.y + widget.getHeaderHeight()));
          widget.setMenu(isInHeader ? widgetTodayTimeEntryTableHeaderMenu : widgetTodayTimeEntryTableBodyMenu);
        }
      });
      widgetTodayTimeEntryTable.addListener(SWT.Dispose, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          widgetTodayTimeEntryTableHeaderMenu.dispose();
          widgetTodayTimeEntryTableBodyMenu.dispose();
        }
      });
      widgetTodayTimeEntryTable.addListener(SWT.EraseItem, new Listener()
      {
        public void handleEvent(Event event)
        {
          TableItem tableItem  = (TableItem)event.item;
          int       index      = widgetTodayTimeEntryTable.indexOf(tableItem);

          Rectangle clientArea = widgetTodayTimeEntryTable.getBounds();
          Rectangle bounds     = event.getBounds();
          GC        gc         = event.gc;

/*
          // enable alpha blending if possible
          gc.setAdvanced(true);
          if (gc.getAdvanced()) gc.setAlpha(127);
*/
          // draw background
          Color background = gc.getBackground();
          gc.setBackground(((index % 2)==0) ? COLOR_BACKGROUND_EVEN : COLOR_BACKGROUND_ODD);
          gc.fillRectangle(bounds.x, bounds.y, clientArea.width, bounds.height);
          gc.setBackground(background);

          event.detail &= ~SWT.BACKGROUND;
        }
      });
      widgetTodayTimeEntryTable.addMouseListener(new MouseListener()
      {
        @Override
        public void mouseDoubleClick(MouseEvent mouseEvent)
        {
          Table       widget             = (Table)mouseEvent.widget;
          TableItem[] selectedTableItems = widget.getSelection();

          if (selectedTableItems.length > 0)
          {
            final TimeEntry timeEntry = (TimeEntry)selectedTableItems[0].getData();

            TimeEntry oldTimeEntry = timeEntry.clone();
            if (editTimeEntry(timeEntry,"Edit time entry","update","Save"))
            {
              try
              {
                timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
              }
              catch (final BackendException exception)
              {
                printStackTrace(exception);
                display.syncExec(new Runnable()
                {
                  public void run()
                  {
                    Dialogs.error(shell,"Cannot update time entry (error: "+exception.getMessage()+")");
                  }
                });
                return;
              }
              catch (InterruptedException exception)
              {
                // ignored
              }
            }
          }
        }
        @Override
        public void mouseDown(MouseEvent mouseEvent)
        {
        }
        @Override
        public void mouseUp(MouseEvent mouseEvent)
        {
        }
      });
      widgetTodayTimeEntryTable.addKeyListener(new KeyListener()
      {
        @Override
        public void keyPressed(KeyEvent keyEvent)
        {
          Table       widget             = (Table)keyEvent.widget;
          TableItem[] selectedTableItems = widget.getSelection();

          if      (Widgets.isAccelerator(keyEvent,SWT.INSERT))
          {
            Widgets.setFocus(addTimeEntry2.widgetProject);
          }
          else if (Widgets.isAccelerator(keyEvent,SWT.DEL) || Widgets.isAccelerator(keyEvent,SWT.BS))
          {
            if (selectedTableItems.length > 0)
            {
              if (Dialogs.confirm(shell,"Delete "+selectedTableItems.length+" time entries?"))
              {
                try
                {
                  for (TableItem tableItem : selectedTableItems)
                  {
                    TimeEntry timeEntry = (TimeEntry)tableItem.getData();
                    removeTimeEntry(timeEntry);
                  }
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot delete data from server (error: "+exception.getMessage()+")");
                  return;
                }
              }
            }
          }
        }
        @Override
        public void keyReleased(KeyEvent keyEvent)
        {
        }
      });
      widgetTodayTimeEntryTable.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Table       widget             = (Table)selectionEvent.widget;
          TableItem[] selectedTableItems = widget.getSelection();

          if (selectedTableItems.length > 0)
          {
            int index;
            try
            {
              ProjectList projectList = Backend.getProjectList();
              TimeEntry   timeEntry   = (TimeEntry)selectedTableItems[0].getData();
              Activity    activity    = timeEntry.getActivity();
              Project     project     = timeEntry.getProject ();
              Issue       issue       = timeEntry.getIssue   ();

              // update projects, issues, activies
              updateProjects(addTimeEntry2.widgetProject,projectList,project);
              updateIssues(addTimeEntry2.widgetIssueId,
                           addTimeEntry2.widgetIssueShortName,
                           addTimeEntry2.widgetIssueSummary,
                           project
                          );
              updateActivities(addTimeEntry2.widgetActivitiesTitle,
                               addTimeEntry2.widgetActivities,
                               project
                              );

              // select matching issue
              setSelectedIssue(addTimeEntry2.widgetIssueId,
                               addTimeEntry2.widgetIssueShortName,
                               addTimeEntry2.widgetIssueSummary,
                               addTimeEntry2.widgetIssueDoneRatio,
                               issue
                              );

              // select matching activity
              if (addTimeEntry2.widgetActivities != null)
              {
                if (activity != null)
                {
                  Widgets.setSelectedOptionMenuItem(addTimeEntry2.widgetActivities,activity);
                }
              }

              // set spent hours, minutes fraction
              addTimeEntry2.widgetSpentHourFraction.setSelection(timeEntry.getHourFraction());
              addTimeEntry2.widgetSpentMinuteFraction.setSelection(timeEntry.getMinuteFraction());

              // set comments
              addTimeEntry2.widgetComments.setText(timeEntry.comments);
            }
            catch (BackendException exception)
            {
              // ignored
              printStackTrace(exception);
            }
          }
        }
      });
      widgetTodayTimeEntryTable.addListener(USER_EVENT_ADD_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Table     widget    = (Table)event.widget;
          TimeEntry timeEntry = (TimeEntry)event.data;
//Dprintf.dprintf("widgetTodayTimeEntryTable timeEntry=%s",timeEntry);

          if (timeEntry.spentOn.isToday())
          {
            // add today table item
            TableItem tableItem = addTableItem(widget,timeEntry);
            widget.setSelection(tableItem);

            // update today duration sum
            todayDurationSum.set(Backend.getTimeEntryTodayDurationSum());
          }
        }
      });
      widgetTodayTimeEntryTable.addListener(USER_EVENT_UPDATE_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Table     widget    = (Table)event.widget;
          TimeEntry timeEntry = (TimeEntry)event.data;

          // refresh/remove today table item
          if (timeEntry.spentOn.isToday())
          {
            refreshTableItem(widget,timeEntry);
          }
          else
          {
            removeTableItem(widget,timeEntry);
          }

          // update today duration sum
          todayDurationSum.set(Backend.getTimeEntryTodayDurationSum());
        }
      });
      widgetTodayTimeEntryTable.addListener(USER_EVENT_REMOVE_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Table     widget    = (Table)event.widget;
          TimeEntry timeEntry = (TimeEntry)event.data;

          // remove table item
          removeTableItem(widget,timeEntry);

          // update today duration sum
          todayDurationSum.set(Backend.getTimeEntryTodayDurationSum());
        }
      });

      // action list
      composite = widgetTodayPane.getComposite(1);
      composite.setLayout(new TableLayout(1.0,1.0));

      widgetTodayActionTable = new ActionTable(composite,SWT.MULTI|SWT.V_SCROLL);
      widgetTodayActionTable.setLayout(new TableLayout(null,new double[]{0.0,0.8,0.2}));
      Widgets.layout(widgetTodayActionTable,0,0,TableLayoutData.NSWE);
      widgetTodayActionTable.setColumnWidth(Settings.geometryActionTableColumns.width);
      ActionMonitor.getInstance().addActionListener(new ActionMonitor.ActionListener()
      {
        @Override
        public void refresh(final ActionList actionList)
        {
          if (!display.isDisposed())
          {
            display.asyncExec(new Runnable()
            {
              public void run()
              {
                if (!widgetTodayActionTable.isDisposed())
                {
                  widgetTodayActionTable.refresh(actionList);
                }
              }
            });
          }
        }
        @Override
        public void refresh(final TagList tagList)
        {
          if (!display.isDisposed())
          {
            display.asyncExec(new Runnable()
            {
              public void run()
              {
                if (!widgetTodayActionTable.isDisposed())
                {
                  widgetTodayActionTable.refresh(tagList);
                }
              }
            });
          }
        }
      });
      widgetTodayActionTable.addActionListener(new ActionTable.ActionListener()
      {
        @Override
        public void mouseDoubleClick(int row, int column)
        {
          Action selectedActions[] = widgetTodayActionTable.getSelectedActions();

          if (selectedActions.length > 0)
          {
            Tag tag = ActionMonitor.getInstance().getTagList().getAssignedTag(selectedActions[0]);
            if (tag != null)
            {
              try
              {
                addTimeEntry((((tag.endTimestamp-tag.startTimestamp)/60+Settings.minTimeDelta-1)/Settings.minTimeDelta)*Settings.minTimeDelta);
              }
              catch (BackendException exception)
              {
                printStackTrace(exception);
                Dialogs.error(shell,"Cannot add new time entry (error: "+exception.getMessage()+")");
                return;
              }
            }
          }
        }
        @Override
        public void mouseDown(int row, int column)
        {
          startAction = widgetTodayActionTable.getActionAt(row,column);
Dprintf.dprintf("startAction=%s",startAction);
        }
        @Override
        public void mouseUp(int row, int column)
        {
          if (column == 2)
          {
            TimeRange selectedTimeRanges[] = widgetTodayActionTable.getSelectedTimeRanges();
            if (selectedTimeRanges.length > 0)
            {
Dprintf.dprintf("selectedTimeRanges0=%s",selectedTimeRanges[0]);
Dprintf.dprintf("selectedTimeRangesend=%s",selectedTimeRanges[selectedTimeRanges.length-1]);
              // get time range
              long startTimestamp = selectedTimeRanges[0].startTimestamp;
              long endTimestamp   = selectedTimeRanges[selectedTimeRanges.length-1].endTimestamp;
              if (endTimestamp <= startTimestamp) endTimestamp = startTimestamp+1;

              // add new tag
              Tag tag = new Tag(startTimestamp,endTimestamp);
              if (editTag(tag,"Add tag","Add"))
              {
                try
                {
                  ActionMonitor.getInstance().addTag(tag);
                }
                catch (SQLException exception)
                {
                  Dialogs.error(shell,"Cannot add tag (error: "+exception.getMessage()+")");
                }
              }
            }
          }
        }
      });

      menu = Widgets.newPopupMenu(shell);
      {
        menuItem = Widgets.addMenuItem(menu,"Add tag"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TimeRange selectedTimeRanges[] = widgetTodayActionTable.getSelectedTimeRanges();

Dprintf.dprintf("%d",selectedTimeRanges.length);
            if (selectedTimeRanges.length > 0)
            {
Dprintf.dprintf("%s",selectedTimeRanges[0]);
Dprintf.dprintf("%s",selectedTimeRanges[selectedTimeRanges.length-1]);
              // get time range
              long startTimestamp = selectedTimeRanges[0].startTimestamp;
              long endTimestamp   = selectedTimeRanges[selectedTimeRanges.length-1].endTimestamp;
              if (endTimestamp <= startTimestamp) endTimestamp = startTimestamp+1;

              // add new tag
              Tag tag = new Tag(startTimestamp,endTimestamp);
              if (editTag(tag,"Add tag","Add"))
              {
                try
                {
                  ActionMonitor.getInstance().addTag(tag);
                }
                catch (SQLException exception)
                {
                  Dialogs.error(shell,"Cannot add tag (error: "+exception.getMessage()+")");
                }
              }
            }
          }
        });

        menuItem = Widgets.addMenuItem(menu,"Edit tag"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Tag selectedTag = widgetTodayActionTable.getSelectedTag();
            if (selectedTag != null)
            {
              // edit tag
              if (editTag(selectedTag,"Edit tag","Save"))
              {
                try
                {
                  ActionMonitor.getInstance().updateTag(selectedTag);
                }
                catch (SQLException exception)
                {
                  Dialogs.error(shell,"Cannot update tag (error: "+exception.getMessage()+")");
                }
              }
            }
          }
        });

        menuItem = Widgets.addMenuItem(menu,"Remove tag"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Tag selectedTag = widgetTodayActionTable.getSelectedTag();
            if (selectedTag != null)
            {
              try
              {
                ActionMonitor.getInstance().removeTag(selectedTag);
              }
              catch (SQLException exception)
              {
                Dialogs.error(shell,"Cannot remove tag (error: "+exception.getMessage()+")");
              }
            }
          }
        });

        Widgets.addMenuItemSeparator(menu);

        menuItem = Widgets.addMenuItem(menu,"Add filter"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Action selectedActions[] = widgetTodayActionTable.getSelectedActions();

            String actionFilter = Dialogs.string(shell,
                                                 "Add action filter",
                                                 "Filter",
                                                 (selectedActions.length > 0) ? selectedActions[0].text : "",
                                                 "Add",
                                                 "Cancel"
                                                );
            if (actionFilter != null)
            {
              // add filter
              Settings.actionFilters.add(actionFilter);

              // update list
Dprintf.dprintf("filter ");
              ActionMonitor.getInstance().setActionFilters(Settings.actionFilters);
            }
          }
        });
      }
      widgetTodayActionTable.setMenu(menu);

      // add new time entry
      addTimeEntry2 = new AddTimeEntryWidgets(widgetTabDay,SWT.NONE);
      addTimeEntry2.setLayout(new TableLayout(0.0,new double[]{0.0,1.0},2));
      Widgets.layout(addTimeEntry2,2,0,TableLayoutData.WE);
    }
    widgetTabDay.addListener(SWT.Show,new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        Widgets.setFocus(widgetTodayTimeEntryTable);
      }
    });

    widgetTabWeeks = Widgets.addTab(widgetTabFolder,"Weeks ("+Widgets.acceleratorToText(Settings.keyTabTimeEntries)+")");
    widgetTabWeeks.setLayout(new TableLayout(new double[]{1.0,0.0},1.0,2));
    Widgets.layout(widgetTabWeeks,0,0,TableLayoutData.NSWE);
    {
      final String WEEKS_TIME_ENTRY_COLUMN_NAMES[] = new String[]{"Spent on","Id","Hours","Activity","Project","Issue short name","Issue","Comments"};

      // time entry tree
      widgetWeeksTimeEntryTree = Widgets.newTree(widgetTabWeeks,SWT.VIRTUAL|SWT.MULTI);
      widgetWeeksTimeEntryTree.setToolTipText("All time entries of user.\nENTER/RETURN to open/edit entry.\nDEL/BACKSPACE to delete entries.");
      widgetWeeksTimeEntryTree.setLayout(new TableLayout(null,null));
      Widgets.layout(widgetWeeksTimeEntryTree,0,0,TableLayoutData.NSWE);
      treeColumn = Widgets.addTreeColumn(widgetWeeksTimeEntryTree,WEEKS_TIME_ENTRY_COLUMN_NAMES[0],SWT.LEFT,  80,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetWeeksTimeEntryTree,WEEKS_TIME_ENTRY_COLUMN_NAMES[1],SWT.RIGHT, 50,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetWeeksTimeEntryTree,WEEKS_TIME_ENTRY_COLUMN_NAMES[2],SWT.RIGHT, 70,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetWeeksTimeEntryTree,WEEKS_TIME_ENTRY_COLUMN_NAMES[3],SWT.LEFT,  70,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetWeeksTimeEntryTree,WEEKS_TIME_ENTRY_COLUMN_NAMES[4],SWT.LEFT,  80,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetWeeksTimeEntryTree,WEEKS_TIME_ENTRY_COLUMN_NAMES[5],SWT.LEFT,  80,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetWeeksTimeEntryTree,WEEKS_TIME_ENTRY_COLUMN_NAMES[6],SWT.LEFT, 100,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetWeeksTimeEntryTree,WEEKS_TIME_ENTRY_COLUMN_NAMES[7],SWT.LEFT, 200,true );
      treeColumn.setMoveable(true);
      Widgets.setTreeColumnWidth(widgetWeeksTimeEntryTree,Settings.geometryWeekTimeEntryColumns.width);
      widgetWeeksTimeEntryTree.setColumnOrder(Settings.geometryWeekTimeEntryColumnOrder.getMap(WEEKS_TIME_ENTRY_COLUMN_NAMES));

      Listener timeEntryTreeColumnResizeListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          if (Settings.isPermitModify())
          {
            Settings.geometryWeekTimeEntryColumns = new Settings.ColumnSizes(Widgets.getTreeColumnWidth(widgetWeeksTimeEntryTree));
            Settings.setModified();
          }
        }
      };
      Listener timeEntryTreeColumnMoveListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          int[]    columnOrder = widgetWeeksTimeEntryTree.getColumnOrder();
          String[] names       = new String[columnOrder.length];
          for (int i = 0; i < columnOrder.length; i++)
          {
            names[i] = WEEKS_TIME_ENTRY_COLUMN_NAMES[columnOrder[i]];
          }
          Settings.geometryWeekTimeEntryColumnOrder = new SettingUtils.SimpleStringArray(names);
        }
      };
      for (TreeColumn orderTreeColumn : widgetWeeksTimeEntryTree.getColumns())
      {
        orderTreeColumn.setMoveable(true);
        orderTreeColumn.addListener(SWT.Move,timeEntryTreeColumnMoveListener);
        orderTreeColumn.addListener(SWT.Resize,timeEntryTreeColumnResizeListener);
      }

      widgetWeeksTimeEntryTreeHeaderMenu = Widgets.newPopupMenu(shell);
      {
        menuItem = Widgets.addMenuItemCheckbox(widgetWeeksTimeEntryTreeHeaderMenu,WEEKS_TIME_ENTRY_COLUMN_NAMES[0]);
        menuItem.setSelection(Settings.geometryWeekTimeEntryColumns.get(0) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetWeeksTimeEntryTree,0,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetWeeksTimeEntryTreeHeaderMenu,WEEKS_TIME_ENTRY_COLUMN_NAMES[1]);
        menuItem.setSelection(Settings.geometryWeekTimeEntryColumns.get(1) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetWeeksTimeEntryTree,1,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetWeeksTimeEntryTreeHeaderMenu,WEEKS_TIME_ENTRY_COLUMN_NAMES[2]);
        menuItem.setSelection(Settings.geometryWeekTimeEntryColumns.get(2) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetWeeksTimeEntryTree,2,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetWeeksTimeEntryTreeHeaderMenu,WEEKS_TIME_ENTRY_COLUMN_NAMES[3]);
        menuItem.setSelection(Settings.geometryWeekTimeEntryColumns.get(3) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetWeeksTimeEntryTree,3,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetWeeksTimeEntryTreeHeaderMenu,WEEKS_TIME_ENTRY_COLUMN_NAMES[4]);
        menuItem.setSelection(Settings.geometryWeekTimeEntryColumns.get(4) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetWeeksTimeEntryTree,4,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetWeeksTimeEntryTreeHeaderMenu,WEEKS_TIME_ENTRY_COLUMN_NAMES[5]);
        menuItem.setSelection(Settings.geometryWeekTimeEntryColumns.get(5) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetWeeksTimeEntryTree,5,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetWeeksTimeEntryTreeHeaderMenu,WEEKS_TIME_ENTRY_COLUMN_NAMES[6]);
        menuItem.setSelection(Settings.geometryWeekTimeEntryColumns.get(6) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetWeeksTimeEntryTree,6,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetWeeksTimeEntryTreeHeaderMenu,WEEKS_TIME_ENTRY_COLUMN_NAMES[7]);
        menuItem.setSelection(Settings.geometryWeekTimeEntryColumns.get(7) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetWeeksTimeEntryTree,7,widget.getSelection());
          }
        });
      }

      widgetWeeksTimeEntryTreeBodyMenu = Widgets.newPopupMenu(shell);
      {
        menuWeekSetHours = Widgets.addMenu(widgetWeeksTimeEntryTreeBodyMenu,"Set hours"+"\u2026");
        for (int i = 1; i <= 8*60/Settings.minTimeDelta; i++)
        {
          menuItem = Widgets.addMenuItem(menuWeekSetHours,new Integer(i*Settings.minTimeDelta),formatHours(i*Settings.minTimeDelta));
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              MenuItem menuItem = (MenuItem)selectionEvent.widget;
              int      duration = (Integer)menuItem.getData();

              TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
              if (treeItems.length > 0)
              {
                if (treeItems[0].getData() instanceof TimeEntry)
                {
                  TimeEntry timeEntry = (TimeEntry)treeItems[0].getData();
                  assert timeEntry != null;

                  // set duration
                  timeEntry.duration = duration;

                  // update time entry
                  try
                  {
                    timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                    Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    final String message = exception.getMessage();
                    Dialogs.error(shell,"Cannot update time entry (error: "+message+")");
                    return;
                  }
                  catch (InterruptedException exception)
                  {
                    // ignored
                  }
                }
              }
            }
          });
        }
        menuWeekHoursPlus = Widgets.addMenu(widgetWeeksTimeEntryTreeBodyMenu,"+");
        for (int i = 1; i <= 8*60/Settings.minTimeDelta; i++)
        {
          menuItem = Widgets.addMenuItem(menuWeekHoursPlus,new Integer(i*Settings.minTimeDelta),formatHours(i*Settings.minTimeDelta));
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              MenuItem menuItem = (MenuItem)selectionEvent.widget;
              int      delta    = (Integer)menuItem.getData();

              TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
              if (treeItems.length > 0)
              {
                if (treeItems[0].getData() instanceof TimeEntry)
                {
                  TimeEntry timeEntry = (TimeEntry)treeItems[0].getData();
                  assert timeEntry != null;
Dprintf.dprintf("inc timeentry=%s",timeEntry);


                  // increment duration
                  timeEntry.duration += delta;

                  // update time entry
                  try
                  {
                    timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                    Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    final String message = exception.getMessage();
                    Dialogs.error(shell,"Cannot update time entry (error: "+message+")");
                    return;
                  }
                  catch (InterruptedException exception)
                  {
                    // ignored
                  }
/**/
                }
              }
            }
          });
        }
        menuWeekHoursMinus = Widgets.addMenu(widgetWeeksTimeEntryTreeBodyMenu,"-");
        for (int i = 1; i <= 8*60/Settings.minTimeDelta; i++)
        {
          menuItem = Widgets.addMenuItem(menuWeekHoursMinus,new Integer(i*Settings.minTimeDelta),formatHours(i*Settings.minTimeDelta));
          menuItem.setData(new Integer(i*Settings.minTimeDelta));
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              MenuItem menuItem = (MenuItem)selectionEvent.widget;
              int      delta    = (Integer)menuItem.getData();

              TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
              if (treeItems.length > 0)
              {
                if (treeItems[0].getData() instanceof TimeEntry)
                {
                  TimeEntry timeEntry = (TimeEntry)treeItems[0].getData();
                  assert timeEntry != null;
Dprintf.dprintf("dec timeentry=%s",timeEntry);

                  if      (timeEntry.duration > delta)
                  {
                    // decrement duration
                    timeEntry.duration -= delta;

                    // update time entry
                    try
                    {
                      timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                      Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
                    }
                    catch (BackendException exception)
                    {
                      printStackTrace(exception);
                      final String message = exception.getMessage();
                      Dialogs.error(shell,"Cannot update time entry (error: "+message+")");
                      return;
                    }
                    catch (InterruptedException exception)
                    {
                      // ignored
                    }
                  }
                  else if (Dialogs.confirm(shell,"Confirmation","Spent hours will become zero - delete time entry?","Delete","Cancel"))
                  {
                    // remove time entry
                    try
                    {
                      removeTimeEntry(timeEntry);
                    }
                    catch (BackendException exception)
                    {
                      printStackTrace(exception);
                      final String message = exception.getMessage();
                      Dialogs.error(shell,"Cannot update time entry (error: "+message+")");
                      return;
                    }
                  }
                }
              }
            }
          });
        }

        Widgets.addMenuItemSeparator(widgetWeeksTimeEntryTreeBodyMenu);

        menuItem = Widgets.addMenuItem(widgetWeeksTimeEntryTreeBodyMenu,"Add time entry"+"\u2026",Settings.keyNewTimeEntry);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetAddTimeEntries);
          }
        });

        widgetAllProjectsMenu1 = Widgets.addMenu(widgetWeeksTimeEntryTreeBodyMenu,"Add time entry for");

        menuItem = Widgets.addMenuItem(widgetWeeksTimeEntryTreeBodyMenu,"Edit time entry"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetEditTimeEntries);
          }
        });

        menuItem = Widgets.addMenuItem(widgetWeeksTimeEntryTreeBodyMenu,"Delete time entries"+"\u2026",Settings.keyDeleteTimeEntry);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetDeleteTimeEntries);
          }
        });

        Widgets.addMenuItemSeparator(widgetWeeksTimeEntryTreeBodyMenu);

        widgetAllProjectsMenu2 = Widgets.addMenu(widgetWeeksTimeEntryTreeBodyMenu,"Add issue for");

        menuItem = Widgets.addMenuItem(widgetWeeksTimeEntryTreeBodyMenu,"Edit issue"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
            if (treeItems.length > 0)
            {
              if (treeItems[0].getData() instanceof TimeEntry)
              {
                // get time entry
                TimeEntry timeEntry = (TimeEntry)treeItems[0].getData();

                // edit issue
                try
                {
                  Issue issue = timeEntry.getIssue();
                  if (editIssue(issue,"Edit issue","Save"))
                  {
                    issue.setModified(Backend.ModifiedStates.UPDATED);
                    Widgets.notify(shell,USER_EVENT_UPDATE_ISSUE,issue);
                  }
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot edit issue (error: "+exception.getMessage()+")");
                  return;
                }
                catch (InterruptedException exception)
                {
                  // ignored
                }
              }
            }
          }
        });

        Widgets.addMenuItemSeparator(widgetWeeksTimeEntryTreeBodyMenu);

        menuItem = Widgets.addMenuItem(widgetWeeksTimeEntryTreeBodyMenu,"Mark as holiday days");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
            if (treeItems.length > 0)
            {
              for (TreeItem treeItem : treeItems)
              {
                if (treeItem.getData() instanceof SpentOn)
                {
                  // add holiday day
                  addHoliday((SpentOn)treeItem.getData());

                  // refresh tree item
                  Event event = new Event();
                  event.item = treeItem;
                  widgetWeeksTimeEntryTree.notifyListeners(SWT.SetData,event);
                }
              }
            }
          }
        });
        menuItem = Widgets.addMenuItem(widgetWeeksTimeEntryTreeBodyMenu,"Mark as vaction days");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
            if (treeItems.length > 0)
            {
              for (TreeItem treeItem : treeItems)
              {
                if (treeItem.getData() instanceof SpentOn)
                {
                  // add vaction day
                  try
                  {
                    addVacation((SpentOn)treeItem.getData());
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    Dialogs.error(shell,"Cannot add new vacation time entry (error: "+exception.getMessage()+")");
                    return;
                  }

                  // refresh tree item
                  Event event = new Event();
                  event.item = treeItem;
                  widgetWeeksTimeEntryTree.notifyListeners(SWT.SetData,event);
                }
              }
            }
          }
        });
        menuItem = Widgets.addMenuItem(widgetWeeksTimeEntryTreeBodyMenu,"Mark as working days");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
            if (treeItems.length > 0)
            {
              for (TreeItem treeItem : treeItems)
              {
                if (treeItem.getData() instanceof SpentOn)
                {
                  // remove holiday day
                  removeHolidayDate((SpentOn)treeItem.getData());

                  // refresh tree item
                  Event event = new Event();
                  event.item = treeItem;
                  widgetWeeksTimeEntryTree.notifyListeners(SWT.SetData,event);
                }
              }
            }
          }
        });
      }
      widgetWeeksTimeEntryTree.setMenu(widgetWeeksTimeEntryTreeBodyMenu);

      widgetWeeksTimeEntryTree.addListener(SWT.MenuDetect, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree widget = (Tree)event.widget;

          Point p = display.map(null, widget, new Point(event.x, event.y));
          Rectangle clientArea = widget.getClientArea();
          boolean isInHeader =    (clientArea.y <= p.y)
                               && (p.y < (clientArea.y + widget.getHeaderHeight()));
          widget.setMenu(isInHeader ? widgetWeeksTimeEntryTreeHeaderMenu : widgetWeeksTimeEntryTreeBodyMenu);
        }
      });
      widgetWeeksTimeEntryTree.addListener(SWT.Dispose, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          widgetWeeksTimeEntryTreeHeaderMenu.dispose();
          widgetWeeksTimeEntryTreeBodyMenu.dispose();
        }
      });
      widgetWeeksTimeEntryTree.addTreeListener(new TreeListener()
      {
        public void treeCollapsed(TreeEvent treeEvent)
        {
          TreeItem treeItem = (TreeItem)treeEvent.item;

          collapseTimeEntryTree(treeItem);
        }
        public void treeExpanded(TreeEvent treeEvent)
        {
          TreeItem treeItem = (TreeItem)treeEvent.item;

          if      (treeItem.getData() instanceof TimeEntryTreeItemData)
          {
            TimeEntryTreeItemData itemData = (TimeEntryTreeItemData)treeItem.getData();

            expandTimeEntryTree(treeItem,itemData.spentOnStart,itemData.spentOnEnd);
          }
          else if (treeItem.getData() instanceof SpentOn)
          {
            SpentOn spentOn  = (SpentOn)treeItem.getData();

            expandTimeEntryTree(treeItem,spentOn);
          }
else {
//TODO: should not happeN?
Dprintf.dprintf("xxxxxxxxxxxxxxxxxxxxxxx");
}
        }
      });
      widgetWeeksTimeEntryTree.addListener(SWT.EraseItem, new Listener()
      {
        public void handleEvent(Event event)
        {
          TreeItem  treeItem   = (TreeItem)event.item;

          Rectangle clientArea = widgetWeeksTimeEntryTree.getBounds();
          Rectangle bounds     = event.getBounds();
          GC        gc         = event.gc;

          // get index tree (including sub-trees)
//TODO: get size of open sub-trees
          int      index          = 0;
          TreeItem subTreeItem    = treeItem;
          TreeItem parentTreeItem = subTreeItem.getParentItem();
          while (parentTreeItem != null)
          {
//Dprintf.dprintf("t=%s: %d %d",t,parentTreeItem.indexOf(subTreeItem),widgetWeeksTimeEntryTree.indexOf(subTreeItem));
            index += 1+parentTreeItem.indexOf(subTreeItem);
            subTreeItem    = parentTreeItem;
            parentTreeItem = parentTreeItem.getParentItem();
          }
          index += widgetWeeksTimeEntryTree.indexOf(subTreeItem);

          // enable alpha blending if possible
//          gc.setAdvanced(true);
//          if (gc.getAdvanced()) gc.setAlpha(127);

          // draw background
          Color backgroundSave = gc.getBackground();
          Color background;
          if (treeItem.getData() instanceof TimeEntryTreeItemData)
          {
            TimeEntryTreeItemData itemData = (TimeEntryTreeItemData)treeItem.getData();
            if ((itemData.durationSum+itemData.holidayDurationSum) < (long)(Settings.workingHoursPerWeek*60.0))
            {
              background = COLOR_TIME_ENTRIES_INCOMPLETE.background;
            }
            else
            {
              background = ((index % 2)==0) ? COLOR_BACKGROUND_EVEN : COLOR_BACKGROUND_ODD;
            }
          }
          else
          {
            background = ((index % 2)==0) ? COLOR_BACKGROUND_EVEN : COLOR_BACKGROUND_ODD;
          }
          gc.setBackground(background);
          gc.fillRectangle(bounds.x, bounds.y, clientArea.width, bounds.height);
          gc.setBackground(backgroundSave);

          event.detail &= ~SWT.BACKGROUND;
        }
      });
      widgetWeeksTimeEntryTree.addListener(SWT.SetData,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree     tree     = widgetWeeksTimeEntryTree;//(Tree)event.widget;
          TreeItem treeItem = (TreeItem)event.item;

          int index = tree.indexOf(treeItem);
          if (index >= 0)
          {
            Calendar week = Calendar.getInstance();
            week.add(Calendar.DAY_OF_WEEK,-index*7);

            Calendar weekStart = (Calendar)week.clone();
            weekStart.add(Calendar.DAY_OF_WEEK,-(weekStart.get(Calendar.DAY_OF_WEEK)-Calendar.MONDAY));

            Calendar weekEnd = (Calendar)weekStart.clone();
            weekEnd.add(Calendar.DAY_OF_YEAR,6);

            SpentOn spentOnStart = new SpentOn(weekStart.getTime());
            SpentOn spentOnEnd   = new SpentOn(weekEnd.getTime()  );
//Dprintf.dprintf("fromSpentOn=%s .. %s",spentOnStart,spentOnEnd);

            int weekNumber = week.get(Calendar.WEEK_OF_YEAR);

            TimeEntryTreeItemData itemData = new TimeEntryTreeItemData(spentOnStart,spentOnEnd);
//Dprintf.dprintf("update treeItem=%s index=%d itemData=%s",treeItem,index,itemData);

            treeItem.setData(itemData);
            treeItem.setText(0,String.format("W%d",weekNumber));
            treeItem.setText(1,"");
            treeItem.setText(2,(itemData.durationSum != Backend.DURATION_UPDATE) ? formatHours(itemData.durationSum) : "");
            if      ((itemData.durationSum+itemData.holidayDurationSum) < (long)(Settings.workingHoursPerWeek*60.0))
            {
              treeItem.setForeground(COLOR_TIME_ENTRIES_INCOMPLETE.foreground);
            }
            else
            {
              treeItem.setForeground(COLOR_TIME_ENTRIES.foreground);
            }

            if (treeItem.getExpanded())
            {
              int i = 0;
              for (TreeItem weekDayTreeItem : treeItem.getItems())
              {
//Dprintf.dprintf("spentOn=%s spentOnStart=%s spentOnEnd=%s",spentOn,spentOnStart,spentOnEnd);
                SpentOn spentOn = new SpentOn(spentOnEnd.getDate(-i));

                weekDayTreeItem.setData(spentOn);
                setTreeEntry(weekDayTreeItem,spentOn);

                if (weekDayTreeItem.getExpanded())
                {
                  TreeItem timeEntryTreeItems[] = weekDayTreeItem.getItems();
                  for (TreeItem timeEntryTreeItem : timeEntryTreeItems)
                  {
                    TimeEntry timeEntry = (TimeEntry)timeEntryTreeItem.getData();
                    if (timeEntry.spentOn.equals(spentOn))
                    {
//Dprintf.dprintf("timeEntryTreeItem=%s",timeEntryTreeItem);
                    setTreeEntry(timeEntryTreeItem,timeEntry);
                    }
                    else
                    {
                      timeEntryTreeItem.dispose();
                    }
                  }
                }

                i++;
              }
            }
            else
            {
              new TreeItem(treeItem,SWT.NONE);
            }
          }
          else if (treeItem.getData() instanceof SpentOn)
          {
            SpentOn spentOn = (SpentOn)treeItem.getData();
//Dprintf.dprintf("treeItem=%s",treeItem);

            setTreeEntry(treeItem,spentOn);
            if (treeItem.getExpanded())
            {
              for (TreeItem timeEntryTreeItem : treeItem.getItems())
              {
                TimeEntry timeEntry = (TimeEntry)timeEntryTreeItem.getData();
                setTreeEntry(timeEntryTreeItem,timeEntry);
              }
            }
          }
        }
      });
      widgetWeeksTimeEntryTree.addMouseListener(new MouseListener()
      {
        @Override
        public void mouseDoubleClick(MouseEvent mouseEvent)
        {
          final Tree widget = (Tree)mouseEvent.widget;

          // get tree item at mouse position (Note: use first column which is required on Windows)
          TreeItem treeItem = widget.getItem(new Point(widget.getColumns()[0].getWidth()/2,mouseEvent.y));

          if (treeItem != null)
          {
            if      (   (treeItem.getData() instanceof TimeEntryTreeItemData)
                     || (treeItem.getData() instanceof SpentOn              )
                    )
            {
              /* On Linux a double-click does not open the entry by default. Send a
                 event to initiate this behavior on Linux.
              */
              if (System.getProperty("os.name").toLowerCase().matches("linux"))
              {
                Event event = new Event();
                event.item = treeItem;
                if (treeItem.getExpanded())
                {
                  widget.notifyListeners(SWT.Collapse,event);
                }
                else
                {
                  widget.notifyListeners(SWT.Expand,event);
                }
              }
            }
            else if (treeItem.getData() instanceof TimeEntry)
            {
              final TimeEntry timeEntry = (TimeEntry)treeItem.getData();

              TimeEntry oldTimeEntry = timeEntry.clone();
              if (editTimeEntry(timeEntry,"Edit time entry","update","Save"))
              {
                try
                {
                  timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                  Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot update time entry (error: "+exception.getMessage()+")");
                  return;
                }
                catch (InterruptedException exception)
                {
                  // ignored
                }
              }
            }
          }
        }
        @Override
        public void mouseDown(MouseEvent mouseEvent)
        {
          final Tree widget = (Tree)mouseEvent.widget;

          // get tree item at mouse position (Note: use first column which is required on Windows)
          TreeItem treeItem = widget.getItem(new Point(widget.getColumns()[0].getWidth()/2,mouseEvent.y));
          if (treeItem != null)
          {
            for (MenuItem menuItem : menuWeekSetHours.getItems())
            {
              menuItem.setEnabled(treeItem.getData() instanceof TimeEntry);
            }
            for (MenuItem menuItem : menuWeekHoursPlus.getItems())
            {
              menuItem.setEnabled(treeItem.getData() instanceof TimeEntry);
            }
            for (MenuItem menuItem : menuWeekHoursMinus.getItems())
            {
              if (treeItem.getData() instanceof TimeEntry)
              {
                TimeEntry timeEntry = (TimeEntry)treeItem.getData();
                int       duration  = (Integer)menuItem.getData();
                menuItem.setEnabled(timeEntry.duration >= duration);
              }
              else
              {
                menuItem.setEnabled(false);
              }
            }
          }
        }
        @Override
        public void mouseUp(MouseEvent mouseEvent)
        {
        }
      });
      widgetWeeksTimeEntryTree.addKeyListener(new KeyListener()
      {
        @Override
        public void keyPressed(KeyEvent keyEvent)
        {
          final Tree widget = (Tree)keyEvent.widget;

          if      (Widgets.isAccelerator(keyEvent,SWT.SPACE))
          {
            TreeItem[] selectedTreeItems = widget.getSelection();
            if (selectedTreeItems.length > 0)
            {
              Event event = new Event();
              event.item = selectedTreeItems[0];
              if (selectedTreeItems[0].getExpanded())
              {
                widget.notifyListeners(SWT.Collapse,event);
              }
              else
              {
                widget.notifyListeners(SWT.Expand,event);
              }
            }
          }
          else if (   Widgets.isAccelerator(keyEvent,Settings.keyEditTimeEntry)
                   || Widgets.isAccelerator(keyEvent,SWT.CR)
                   || Widgets.isAccelerator(keyEvent,SWT.KEYPAD_CR)
                  )
          {
            Widgets.invoke(widgetEditTimeEntries);
          }
          else if (   Widgets.isAccelerator(keyEvent,Settings.keyNewTimeEntry)
                   || Widgets.isAccelerator(keyEvent,SWT.INSERT)
                  )
          {
            Widgets.invoke(widgetAddTimeEntries);
          }
          else if (   Widgets.isAccelerator(keyEvent,Settings.keyDeleteTimeEntry)
                   || Widgets.isAccelerator(keyEvent,SWT.DEL)
                   || Widgets.isAccelerator(keyEvent,SWT.BS)
                  )
          {
            Widgets.invoke(widgetDeleteTimeEntries);
          }
        }
        @Override
        public void keyReleased(KeyEvent keyEvent)
        {
        }
      });
      widgetWeeksTimeEntryTree.addListener(USER_EVENT_ADD_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree      widget    = (Tree)event.widget;
          TimeEntry timeEntry = (TimeEntry)event.data;

          addTreeItem(widget,timeEntry);
        }
      });
      widgetWeeksTimeEntryTree.addListener(USER_EVENT_UPDATE_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree      widget    = (Tree)event.widget;
          TimeEntry timeEntry = (TimeEntry)event.data;

          refreshTreeItems(widget,timeEntry.spentOn);
        }
      });
      widgetWeeksTimeEntryTree.addListener(USER_EVENT_REMOVE_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree      widget    = (Tree)event.widget;
          TimeEntry timeEntry = (TimeEntry)event.data;

          removeTreeItem(widget,timeEntry);
          refreshTreeItems(widget,timeEntry.spentOn);
        }
      });
      widgetWeeksTimeEntryTree.addListener(USER_EVENT_UPDATE_TIME_ENTRIES,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          updateTimeEntriesTask.trigger();
        }
      });
      widgetAllProjectsMenuSelectionListener1 = new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          MenuItem menuItem = (MenuItem)selectionEvent.widget;

          Project project = (Project)menuItem.getData();

          // get spent-on of selected days or date of today
          List<SpentOn> spentOnList = new ArrayList<SpentOn>();

          TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
          if (treeItems.length > 0)
          {
            for (TreeItem treeItem : treeItems)
            {
              while (   (treeItem != null)
                     && (treeItem.getData() != null)
                     && !(treeItem.getData() instanceof SpentOn)
                    )
              {
                treeItem = treeItem.getParentItem();
              }
              if ((treeItem != null) && (treeItem.getData() != null))
              {
                spentOnList.add((SpentOn)(((SpentOn)treeItem.getData()).clone()));
              }
            }
          }
          else
          {
            spentOnList.add(SpentOn.today());
          }

          if (spentOnList.size() >= 1)
          {
            try
            {
              // add new time entry
              addTimeEntry(project,spentOnList);
            }
            catch (BackendException exception)
            {
Dprintf.dprintf("_");
              printStackTrace(exception);
              Dialogs.error(shell,"Cannot add new time entry (error: "+exception.getMessage()+")");
              return;
            }
          }
        }
      };
      widgetAllProjectsMenuSelectionListener2 = new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          MenuItem menuItem = (MenuItem)selectionEvent.widget;

          Project project = (Project)menuItem.getData();
          try
          {
            // add new issue
            addIssue(project);
          }
          catch (BackendException exception)
          {
            printStackTrace(exception);
            Dialogs.error(shell,"Cannot add new issue (error: "+exception.getMessage()+")");
            return;
          }
        }
      };

      composite = Widgets.newComposite(widgetTabWeeks);
      composite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,1.0,0.0}));
      Widgets.layout(composite,1,0,TableLayoutData.WE);
      {
        button = Widgets.newButton(composite,"Open selected");
        Widgets.layout(button,0,0,TableLayoutData.W,0,0,0,0,120,SWT.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {

            for (TreeItem treeItem : widgetWeeksTimeEntryTree.getSelection())
            {
              if (!treeItem.getExpanded())
              {
                Event event = new Event();
                event.item = treeItem;
                widgetWeeksTimeEntryTree.notifyListeners(SWT.Expand,event);
                display.update();
              }
            }
          }
        });

        button = Widgets.newButton(composite,"Close all");
        Widgets.layout(button,0,1,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {

            for (TreeItem treeItem : widgetWeeksTimeEntryTree.getItems())
            {
              if (treeItem.getExpanded())
              {
                Event event = new Event();
                event.item = treeItem;
                widgetWeeksTimeEntryTree.notifyListeners(SWT.Collapse,event);
                display.update();
              }
            }
          }
        });

        widgetAddTimeEntries = Widgets.newButton(composite,"Add time entry"+"\u2026",Settings.keyNewTimeEntry);
        Widgets.layout(widgetAddTimeEntries,0,3,TableLayoutData.E,0,0,0,0,170,SWT.DEFAULT);
        widgetAddTimeEntries.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            // get spent-on of selected days or date of today
            List<SpentOn> spentOnList = new ArrayList<SpentOn>();

            TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
            if (treeItems.length > 0)
            {
              for (TreeItem treeItem : treeItems)
              {
                while (   (treeItem != null)
                       && (   (treeItem.getData() == null)
                           || !(treeItem.getData() instanceof SpentOn)
                          )
                      )
                {
                  treeItem = treeItem.getParentItem();
                }
                if ((treeItem != null) && (treeItem.getData() != null))
                {
                  spentOnList.add((SpentOn)(((SpentOn)treeItem.getData()).clone()));
                }
                else
                {
                  spentOnList.add(SpentOn.today());
                }
              }
            }
            else
            {
              spentOnList.add(SpentOn.today());
            }
            assert spentOnList.size() >= 1;

            try
            {
              // add new time entries
              addTimeEntry(spentOnList);
            }
            catch (BackendException exception)
            {
              printStackTrace(exception);
              Dialogs.error(shell,"Cannot add new time entries (error: "+exception.getMessage()+")");
              return;
            }
          }
        });

        widgetEditTimeEntries = Widgets.newButton(composite,"Edit time entry"+"\u2026",Settings.keyEditTimeEntry);
        Widgets.layout(widgetEditTimeEntries,0,2,TableLayoutData.E,0,0,0,0,170,SWT.DEFAULT);
        widgetEditTimeEntries.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
            if (treeItems.length > 0)
            {
              if (treeItems[0].getData() instanceof TimeEntry)
              {
                // get time entry
                TimeEntry timeEntry = (TimeEntry)treeItems[0].getData();

                // edit time entry
                if (editTimeEntry(timeEntry,"Edit time entry","update","Save"))
                {
                  try
                  {
                    timeEntry.setModified(Backend.ModifiedStates.UPDATED);
                    Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntry);
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    Dialogs.error(shell,"Cannot update time entry (error: "+exception.getMessage()+")");
                    return;
                  }
                  catch (InterruptedException exception)
                  {
                    // ignored
                  }
                }
              }
            }
          }
        });

        widgetDeleteTimeEntries = Widgets.newButton(composite,"Delete time entries"+"\u2026",Settings.keyDeleteTimeEntry);
        Widgets.layout(widgetDeleteTimeEntries,0,4,TableLayoutData.E,0,0,0,0,170,SWT.DEFAULT);
        widgetDeleteTimeEntries.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TreeItem[] treeItems = widgetWeeksTimeEntryTree.getSelection();
            if (treeItems.length > 0)
            {
              // get time entries to delete
              List<TimeEntry> timeEntryList = new ArrayList<TimeEntry>();
              try
              {
                for (TreeItem treeItem : treeItems)
                {
                  if       (treeItem.getData() instanceof SpentOn)
                  {
                    for (TimeEntry timeEntry : Backend.getTimeEntryList((SpentOn)treeItem.getData()))
                    {
                      assert timeEntry != null;
                      timeEntryList.add(timeEntry);
                    }
                  }
                  else if  (treeItem.getData() instanceof TimeEntry)
                  {
                    TimeEntry timeEntry = (TimeEntry)treeItem.getData();
                    assert timeEntry != null;
                    timeEntryList.add(timeEntry);
                  }
                }
              }
              catch (BackendException exception)
              {
                printStackTrace(exception);
                Dialogs.error(shell,"Cannot get time entries to delete (error: "+exception.getMessage()+")");
                return;
              }

              if (timeEntryList.size() > 0)
              {
                if (Dialogs.confirm(shell,"Delete "+timeEntryList.size()+" time entries?"))
                {
                  try
                  {
                    for (TimeEntry timeEntry : timeEntryList)
                    {
                      removeTimeEntry(timeEntry);
                    }
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    Dialogs.error(shell,"Cannot delete time entries (error: "+exception.getMessage()+")");
                    return;
                  }
                }
              }
            }
          }
        });
      }
    }
    widgetTabWeeks.addListener(SWT.Show,new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        Widgets.setFocus(widgetWeeksTimeEntryTree);
      }
    });

    widgetTabProjectsIssues = Widgets.addTab(widgetTabFolder,"Projects && Issues ("+Widgets.acceleratorToText(Settings.keyTabProjectsIssues)+")");
    widgetTabProjectsIssues.setLayout(new TableLayout(1.0,1.0));
    Widgets.layout(widgetTabProjectsIssues,0,0,TableLayoutData.NSWE);
    {
      // create pane
      pane = Widgets.newPane(widgetTabProjectsIssues,2,SWT.HORIZONTAL);
      Widgets.layout(pane,0,0,TableLayoutData.NSWE);

      // project/issue tree
      composite = pane.getComposite(0);
      composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0,2));

      final String PROJECTS_COLUMN_NAMES[] = new String[]{"Type","Id","Name","Hours","Estimated"};

      SelectionListener selectionListener = new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          TreeColumn        treeColumn        = (TreeColumn)selectionEvent.widget;
          ProjectComparator projectComparator = new ProjectComparator();

          if      (treeColumn == widgetProjectTree.getColumn(0)) projectComparator.setSortMode(ProjectComparator.SortModes.TYPE           );
          else if (treeColumn == widgetProjectTree.getColumn(1)) projectComparator.setSortMode(ProjectComparator.SortModes.ID             );
          else if (treeColumn == widgetProjectTree.getColumn(2)) projectComparator.setSortMode(ProjectComparator.SortModes.NAME           );
          else if (treeColumn == widgetProjectTree.getColumn(3)) projectComparator.setSortMode(ProjectComparator.SortModes.HOURS          );
          else if (treeColumn == widgetProjectTree.getColumn(4)) projectComparator.setSortMode(ProjectComparator.SortModes.ESTIMATED_HOURS);
          Widgets.sortTreeColumn(widgetProjectTree,treeColumn,projectComparator);
        }
      };

      widgetProjectTree = Widgets.newTree(composite,SWT.MULTI);
      widgetProjectTree.setHeaderVisible(true);
      widgetProjectTree.setToolTipText("All projects.\nENTER/RETURN to open/edit an project.");
      widgetProjectTree.setLayout(new TableLayout(1.0,new double[]{0.0,1.0,0.0,0.0}));
      Widgets.layout(widgetProjectTree,0,0,TableLayoutData.NSWE);
      treeColumn = Widgets.addTreeColumn(widgetProjectTree,PROJECTS_COLUMN_NAMES[0],SWT.LEFT,  45,true );
      treeColumn.setMoveable(true);
      treeColumn.addSelectionListener(selectionListener);
      treeColumn = Widgets.addTreeColumn(widgetProjectTree,PROJECTS_COLUMN_NAMES[1],SWT.RIGHT, 70,true );
      treeColumn.setMoveable(true);
      treeColumn.addSelectionListener(selectionListener);
      treeColumn = Widgets.addTreeColumn(widgetProjectTree,PROJECTS_COLUMN_NAMES[2],SWT.LEFT, 200,true );
      treeColumn.setMoveable(true);
      treeColumn.addSelectionListener(selectionListener);
      treeColumn = Widgets.addTreeColumn(widgetProjectTree,PROJECTS_COLUMN_NAMES[3],SWT.RIGHT, 80,false);
      treeColumn.setMoveable(true);
      treeColumn.addSelectionListener(selectionListener);
      treeColumn = Widgets.addTreeColumn(widgetProjectTree,PROJECTS_COLUMN_NAMES[4],SWT.RIGHT, 80,false);
      treeColumn.setMoveable(true);
      treeColumn.addSelectionListener(selectionListener);
      Widgets.setTreeColumnWidth(widgetProjectTree,Settings.geometryProjectsColumns.width);
      widgetProjectTree.setColumnOrder(Settings.geometryProjectsColumnOrder.getMap(PROJECTS_COLUMN_NAMES));
      composite.setData(widgetProjectTree);

      Listener projectTreeColumnResizeListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          if (Settings.isPermitModify())
          {
            Settings.geometryProjectsColumns = new Settings.ColumnSizes(Widgets.getTreeColumnWidth(widgetProjectTree));
Dprintf.dprintf("%s",Settings.geometryProjectsColumns);
            Settings.setModified();
          }
        }
      };
      Listener projectTreeColumnMoveListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          int[]    columnOrder = widgetProjectTree.getColumnOrder();
          String[] names       = new String[columnOrder.length];
          for (int i = 0; i < columnOrder.length; i++)
          {
            names[i] = PROJECTS_COLUMN_NAMES[columnOrder[i]];
          }
          if (Settings.isPermitModify())
          {
            Settings.geometryProjectsColumnOrder = new SettingUtils.SimpleStringArray(names);
            Settings.setModified();
          }
        }
      };
      for (TreeColumn orderTreeColumn : widgetProjectTree.getColumns())
      {
        orderTreeColumn.setMoveable(true);
        orderTreeColumn.addListener(SWT.Move,projectTreeColumnMoveListener);
        orderTreeColumn.addListener(SWT.Resize,projectTreeColumnResizeListener);
      }

      final Menu projectTreeHeaderMenu = Widgets.newPopupMenu(shell);
      {
        menuItem = Widgets.addMenuItemCheckbox(projectTreeHeaderMenu,PROJECTS_COLUMN_NAMES[0]);
        menuItem.setSelection(Settings.geometryProjectsColumns.get(0) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetProjectTree,0,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(projectTreeHeaderMenu,PROJECTS_COLUMN_NAMES[1]);
        menuItem.setSelection(Settings.geometryProjectsColumns.get(1) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetProjectTree,1,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(projectTreeHeaderMenu,PROJECTS_COLUMN_NAMES[2]);
        menuItem.setSelection(Settings.geometryProjectsColumns.get(2) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetProjectTree,2,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(projectTreeHeaderMenu,PROJECTS_COLUMN_NAMES[3]);
        menuItem.setSelection(Settings.geometryProjectsColumns.get(3) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetProjectTree,3,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(projectTreeHeaderMenu,PROJECTS_COLUMN_NAMES[4]);
        menuItem.setSelection(Settings.geometryProjectsColumns.get(4) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetProjectTree,4,widget.getSelection());
          }
        });
      }

      // add listeners
      widgetProjectTree.addListener(SWT.MenuDetect, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree widget = (Tree)event.widget;

          Point p = display.map(null, widget, new Point(event.x, event.y));
          Rectangle clientArea = widget.getClientArea();
          boolean isInHeader =    (clientArea.y <= p.y)
                               && (p.y < (clientArea.y + widget.getHeaderHeight()));
          widget.setMenu(isInHeader ? projectTreeHeaderMenu : null);
        }
      });
      widgetProjectTree.addListener(SWT.Dispose, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          projectTreeHeaderMenu.dispose();
        }
      });
      widgetProjectTree.addListener(SWT.EraseItem, new Listener()
      {
        public void handleEvent(Event event)
        {
          TreeItem  treeItem   = (TreeItem)event.item;
          int       index      = widgetProjectTree.indexOf(treeItem);

          Rectangle clientArea = widgetProjectTree.getBounds();
          Rectangle bounds     = event.getBounds();
          GC        gc         = event.gc;

/*
          // enable alpha blending if possible
          gc.setAdvanced(true);
          if (gc.getAdvanced()) gc.setAlpha(127);
*/
          // draw background
          Color background = gc.getBackground();
          gc.setBackground(((index % 2)==0) ? COLOR_BACKGROUND_EVEN : COLOR_BACKGROUND_ODD);
          gc.fillRectangle(bounds.x, bounds.y, clientArea.width, bounds.height);
          gc.setBackground(background);

          event.detail &= ~SWT.BACKGROUND;
        }
      });
      widgetProjectTree.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          TreeItem treeItem = (TreeItem)selectionEvent.item;

          {
            shell.setCursor(CURSOR_WAIT);
          }
          try
          {
            selectedProject = (Project)treeItem.getData();
            if (selectedProject != null)
            {
              IssueList issueList = selectedProject.getIssueList();
              setTreeEntries(widgetIssuesTree,issueList);
            }
          }
          catch (BackendException exception)
          {
            printStackTrace(exception);
            String message = exception.getMessage();
            if (message == null) message = exception.getCause().getMessage();
            Dialogs.error(shell,"Cannot get issue data from server (error: "+message+")");
            return;
          }
          finally
          {
            shell.setCursor(null);
          }
        }
      });
      widgetProjectTree.addMouseListener(new MouseListener()
      {
        @Override
        public void mouseDoubleClick(MouseEvent mouseEvent)
        {
          final Tree widget = (Tree)mouseEvent.widget;

          // get tree item at mouse position (Note: use first column which is required on Windows)
          TreeItem treeItem = widget.getItem(new Point(mouseEvent.x,mouseEvent.y));
          if (treeItem != null)
          {
            /* On Linux a double-click does not open the entry by default. Send a
               event to initiate this behavior on Linux.
            */
            if (System.getProperty("os.name").toLowerCase().matches("linux"))
            {
              Event event = new Event();
              event.item = treeItem;
              if (treeItem.getExpanded())
              {
                widget.notifyListeners(SWT.Collapse,event);
              }
              else
              {
                widget.notifyListeners(SWT.Expand,event);
              }
            }
          }
        }
        @Override
        public void mouseDown(MouseEvent mouseEvent)
        {
        }
        @Override
        public void mouseUp(MouseEvent mouseEvent)
        {
        }
      });
      widgetProjectTree.addListener(USER_EVENT_UPDATE_PROJECTS,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree        widget      = (Tree)event.widget;
          ProjectList projectList = (ProjectList)event.data;

          // get project id set
          HashSet<Id> removeIdSet = new HashSet<Id>();
          for (TreeItem treeItem : Widgets.getAllTreeItems(widget))
          {
            if (treeItem.getData() instanceof Project)
            {
              removeIdSet.add(((Project)treeItem.getData()).id);
            }
          }

          // update/add projects
          for (Project project : projectList)
          {
            if (project.parentId.isNone())
            {
//TODO: use updateInsertTreeItem?
              if (Widgets.updateTreeItem(widget,project,"",project.name) == null)
              {
                TreeItem treeItem = Widgets.addTreeItem(widget,
                                                        project,
                                                        project.backendType.getImage(),
                                                        Widgets.TREE_ITEM_FLAG_NONE
                                                       );
                setTreeEntry(treeItem,project);
                treeItem.setExpanded(true);
                setTreeEntries(treeItem,project.id,projectList);
              }
              else
              {
                removeIdSet.remove(project.id);
              }
            }
          }

          // remove not existing projects
// TODO: use remove obsolete
          for (TreeItem treeItem : Widgets.getAllTreeItems(widget))
          {
            if (!treeItem.isDisposed())
            {
              if (treeItem.getData() instanceof Project)
              {
                Project project = (Project)treeItem.getData();
                if (removeIdSet.contains(project.id))
                {
                  disposeComposites(treeItem);
                  Widgets.removeTreeItem(treeItem);
                }
              }
            }
          }
        }
      });
      widgetProjectTree.addListener(USER_EVENT_UPDATE_ISSUES,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree                   widget     = (Tree)event.widget;
//          Project projects[] = (Project[])event.data;
Dprintf.dprintf("TODO widgetProjectTree USER_EVENT_UPDATE_ISSUES");
        }
      });
      widgetProjectTree.addListener(USER_EVENT_ADD_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree      widget    = (Tree)event.widget;
          TimeEntry timeEntry = (TimeEntry)event.data;

          // get project
          Project project = timeEntry.getProject();

          // update project tree item
          for (TreeItem treeItem : widgetProjectTree.getItems())
          {
            if (project.equals((Project)treeItem.getData()))
            {
              setTreeEntry(treeItem,project);
              break;
            }
          }
        }
      });
      widgetProjectTree.addListener(USER_EVENT_UPDATE_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree      widget    = (Tree)event.widget;
          TimeEntry timeEntry = (TimeEntry)event.data;

          // get project
          Project project = timeEntry.getProject();

          // update project tree item
          for (TreeItem treeItem : widgetProjectTree.getItems())
          {
            if (project.equals((Project)treeItem.getData()))
            {
              setTreeEntry(treeItem,project);
              break;
            }
          }
        }
      });
      widgetProjectTree.addListener(USER_EVENT_REMOVE_TIME_ENTRY,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree      widget    = (Tree)event.widget;
          TimeEntry timeEntry = (TimeEntry)event.data;

          // get project
          Project project = timeEntry.getProject();

          // update project tree item
          for (TreeItem treeItem : widgetProjectTree.getItems())
          {
            if (project.equals((Project)treeItem.getData()))
            {
              setTreeEntry(treeItem,project);
              break;
            }
          }
        }
      });

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0}));
      Widgets.layout(subComposite,1,0,TableLayoutData.WE);
      {
        label = Widgets.newLabel(subComposite,"Find:",SWT.NONE,Settings.keyFind);
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetProjectTreeFind = Widgets.newText(subComposite,SWT.SEARCH|SWT.ICON_SEARCH|SWT.ICON_CANCEL);
        widgetProjectTreeFind.setToolTipText("Enter a number or some text to filter projects.");
        widgetProjectTreeFind.setLayoutData(new TableLayoutData(0,1,TableLayoutData.WE));
        widgetProjectTreeFind.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
            Text widget = (Text)selectionEvent.widget;

            String filters[] = StringUtils.splitArray(widget.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
            filterProjects(widgetProjectTree,filters);
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Text widget = (Text)selectionEvent.widget;

            String filters[] = StringUtils.splitArray(widget.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
            filterProjects(widgetProjectTree,filters);
          }
        });
        widgetProjectTreeFind.addKeyListener(new KeyListener()
        {
          @Override
          public void keyPressed(KeyEvent keyEvent)
          {
          }
          @Override
          public void keyReleased(KeyEvent keyEvent)
          {
            Text widget = (Text)keyEvent.widget;

            if      (Widgets.isAccelerator(keyEvent,Settings.keyFind))
            {
              Widgets.setFocus(widgetProjectTreeFind);
              return;
            }
            else if (Widgets.isAccelerator(keyEvent,Settings.keyFindIssue))
            {
              Widgets.setFocus(widgetIssuesTreeFind);
              return;
            }
            else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'x'))
            {
              widget.setText("");
            }

            // restart filter projects timer
            if (filterProjectsTimer != null)
            {
              filterProjectsTimer.cancel();
            }
            filterProjectsTimer = new Timer();
            filterProjectsTimer.schedule(new TimerTask()
            {
              public void run()
              {
                display.syncExec(new Runnable()
                {
                  public void run()
                  {
                    String filters[] = StringUtils.splitArray(widgetProjectTreeFind.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
                    filterProjects(widgetProjectTree,filters);
                  }
                });
              }
            },
            250  // delay [ms]
            );
          }
        });
      }

      // issues tree
      composite = pane.getComposite(1);
      composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));

      final String ISSUES_COLUMN_NAMES[] = new String[]{"Id","Summary","Tracker","Status","Prioritry","Done","Assigned to","Hours","Estimated"};

      widgetIssuesTree = Widgets.newTree(composite,SWT.MULTI);
      widgetIssuesTree.setToolTipText("All issues.\nENTER/RETURN to open/edit an issue.\nDEL/BACKSPACE to delete an issue.");
      widgetIssuesTree.setHeaderVisible(true);
      widgetIssuesTree.setLayout(new TableLayout(1.0,new double[]{0.0,1.0,1.0,1.0,1.0,1.0,1.0,0.0,0.0}));
      Widgets.layout(widgetIssuesTree,0,0,TableLayoutData.NSWE);
      treeColumn = Widgets.addTreeColumn(widgetIssuesTree,ISSUES_COLUMN_NAMES[0],SWT.LEFT,  70,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetIssuesTree,ISSUES_COLUMN_NAMES[1],SWT.LEFT, 200,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetIssuesTree,ISSUES_COLUMN_NAMES[2],SWT.LEFT,  70,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetIssuesTree,ISSUES_COLUMN_NAMES[3],SWT.LEFT,  80,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetIssuesTree,ISSUES_COLUMN_NAMES[4],SWT.LEFT, 100,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetIssuesTree,ISSUES_COLUMN_NAMES[5],SWT.RIGHT, 60,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetIssuesTree,ISSUES_COLUMN_NAMES[6],SWT.LEFT, 120,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetIssuesTree,ISSUES_COLUMN_NAMES[7],SWT.RIGHT, 70,true );
      treeColumn.setMoveable(true);
      treeColumn = Widgets.addTreeColumn(widgetIssuesTree,ISSUES_COLUMN_NAMES[8],SWT.RIGHT, 90,true );
      treeColumn.setMoveable(true);
      Widgets.setTreeColumnWidth(widgetIssuesTree,Settings.geometryIssuesColumns.width);
      widgetIssuesTree.setColumnOrder(Settings.geometryIssuesColumnOrder.getMap(ISSUES_COLUMN_NAMES));
      composite.setData(widgetIssuesTree);

      Listener issueTreeColumnResizeListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          if (Settings.isPermitModify())
          {
            Settings.geometryIssuesColumns = new Settings.ColumnSizes(Widgets.getTreeColumnWidth(widgetIssuesTree));
            Settings.setModified();
          }
        }
      };
      Listener issueTreeColumnMoveListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          int[]    columnOrder = widgetIssuesTree.getColumnOrder();
          String[] names       = new String[columnOrder.length];
          for (int i = 0; i < columnOrder.length; i++)
          {
            names[i] = ISSUES_COLUMN_NAMES[columnOrder[i]];
          }
          Settings.geometryIssuesColumnOrder = new SettingUtils.SimpleStringArray(names);
        }
      };
      for (TreeColumn orderTreeColumn : widgetIssuesTree.getColumns())
      {
        orderTreeColumn.setMoveable(true);
        orderTreeColumn.addListener(SWT.Move,issueTreeColumnMoveListener);
        orderTreeColumn.addListener(SWT.Resize,issueTreeColumnResizeListener);
      }

      final Menu issueTreeHeaderMenu = Widgets.newPopupMenu(shell);
      {
        menuItem = Widgets.addMenuItemCheckbox(issueTreeHeaderMenu,ISSUES_COLUMN_NAMES[0]);
        menuItem.setSelection(Settings.geometryIssuesColumns.get(0) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetIssuesTree,0,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(issueTreeHeaderMenu,ISSUES_COLUMN_NAMES[1]);
        menuItem.setSelection(Settings.geometryIssuesColumns.get(1) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetIssuesTree,1,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(issueTreeHeaderMenu,ISSUES_COLUMN_NAMES[2]);
        menuItem.setSelection(Settings.geometryIssuesColumns.get(2) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetIssuesTree,2,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(issueTreeHeaderMenu,ISSUES_COLUMN_NAMES[3]);
        menuItem.setSelection(Settings.geometryIssuesColumns.get(3) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetIssuesTree,3,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(issueTreeHeaderMenu,ISSUES_COLUMN_NAMES[4]);
        menuItem.setSelection(Settings.geometryIssuesColumns.get(4) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetIssuesTree,4,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(issueTreeHeaderMenu,ISSUES_COLUMN_NAMES[5]);
        menuItem.setSelection(Settings.geometryIssuesColumns.get(5) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetIssuesTree,5,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(issueTreeHeaderMenu,ISSUES_COLUMN_NAMES[6]);
        menuItem.setSelection(Settings.geometryIssuesColumns.get(6) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetIssuesTree,6,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(issueTreeHeaderMenu,ISSUES_COLUMN_NAMES[7]);
        menuItem.setSelection(Settings.geometryIssuesColumns.get(7) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetIssuesTree,7,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(issueTreeHeaderMenu,ISSUES_COLUMN_NAMES[8]);
        menuItem.setSelection(Settings.geometryIssuesColumns.get(8) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetIssuesTree,8,widget.getSelection());
          }
        });
      }

      final Menu issueTreeBodyMenu = Widgets.newPopupMenu(shell);
      {
        menuItem = Widgets.addMenuItem(issueTreeBodyMenu,"Edit"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetEditProjectIssue);
          }
        });
        menuItem = Widgets.addMenuItem(issueTreeBodyMenu,"Add"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetAddProjectIssue);
          }
        });
        menuItem = Widgets.addMenuItem(issueTreeBodyMenu,"Delete"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetDeleteProjectIssue);
          }
        });
      }
      widgetIssuesTree.setMenu(issueTreeBodyMenu);

      // add listeners
      widgetIssuesTree.addListener(SWT.MenuDetect, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree widget = (Tree)event.widget;

          Point p = display.map(null, widget, new Point(event.x, event.y));
          Rectangle clientArea = widget.getClientArea();
          boolean isInHeader =    (clientArea.y <= p.y)
                               && (p.y < (clientArea.y + widget.getHeaderHeight()));
          widget.setMenu(isInHeader ? issueTreeHeaderMenu : issueTreeBodyMenu);
        }
      });
      widgetIssuesTree.addListener(SWT.Dispose, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          issueTreeHeaderMenu.dispose();
          issueTreeBodyMenu.dispose();
        }
      });
      widgetIssuesTree.addListener(SWT.EraseItem, new Listener()
      {
        public void handleEvent(Event event)
        {
          TreeItem  treeItem   = (TreeItem)event.item;
          int       index      = widgetIssuesTree.indexOf(treeItem);

          Rectangle clientArea = widgetIssuesTree.getBounds();
          Rectangle bounds     = event.getBounds();
          GC        gc         = event.gc;

/*
          // enable alpha blending if possible
          gc.setAdvanced(true);
          if (gc.getAdvanced()) gc.setAlpha(127);
*/
          // draw background
          Color background = gc.getBackground();
          gc.setBackground(((index % 2)==0) ? COLOR_BACKGROUND_EVEN : COLOR_BACKGROUND_ODD);
          gc.fillRectangle(bounds.x, bounds.y, clientArea.width, bounds.height);
          gc.setBackground(background);

          event.detail &= ~SWT.BACKGROUND;
        }
      });
      widgetIssuesTree.addMouseListener(new MouseListener()
      {
        @Override
        public void mouseDoubleClick(MouseEvent mouseEvent)
        {
          Widgets.invoke(widgetEditProjectIssue);
        }
        @Override
        public void mouseDown(MouseEvent mouseEvent)
        {
        }
        @Override
        public void mouseUp(MouseEvent mouseEvent)
        {
        }
      });
      widgetIssuesTree.addListener(USER_EVENT_ADD_ISSUE,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree  widget = (Tree)event.widget;
          Issue issue  = (Issue)event.data;
Dprintf.dprintf("TODO");
        }
      });
      widgetIssuesTree.addListener(USER_EVENT_UPDATE_ISSUE,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree                 widget = (Tree)event.widget;
          Issue issue  = (Issue)event.data;

          TreeItem treeItem = Widgets.getTreeItem(widget,issue);
          if (treeItem != null)
          {
            setTreeEntry(treeItem,issue);
          }
        }
      });

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0,0.0,0.0}));
      Widgets.layout(subComposite,1,0,TableLayoutData.WE);
      {
        label = Widgets.newLabel(subComposite,"Find:",SWT.NONE,Settings.keyFindIssue);
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetIssuesTreeFind = Widgets.newText(subComposite,SWT.SEARCH|SWT.ICON_SEARCH|SWT.ICON_CANCEL);
        widgetIssuesTreeFind.setToolTipText("Enter a number or some text to filter issues.");
        widgetIssuesTreeFind.setLayoutData(new TableLayoutData(0,1,TableLayoutData.WE));
        widgetIssuesTreeFind.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
            Text widget = (Text)selectionEvent.widget;

            String filters[] = StringUtils.splitArray(widget.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
            filterIssues(widgetIssuesTree,selectedProject,filters);
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Text widget = (Text)selectionEvent.widget;

            String filters[] = StringUtils.splitArray(widget.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
            filterIssues(widgetIssuesTree,selectedProject,filters);
          }
        });
        widgetIssuesTreeFind.addKeyListener(new KeyListener()
        {
          @Override
          public void keyPressed(KeyEvent keyEvent)
          {
          }
          @Override
          public void keyReleased(KeyEvent keyEvent)
          {
            Text widget = (Text)keyEvent.widget;

            if      (Widgets.isAccelerator(keyEvent,Settings.keyFind))
            {
              Widgets.setFocus(widgetProjectTreeFind);
              return;
            }
            else if (Widgets.isAccelerator(keyEvent,Settings.keyFindIssue))
            {
              Widgets.setFocus(widgetIssuesTreeFind);
              return;
            }
            else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'x'))
            {
              widget.setText("");
            }

            // restart filter issues timer
            if (filterIssuesTimer != null)
            {
              filterIssuesTimer.cancel();
            }
            filterIssuesTimer = new Timer();
            filterIssuesTimer.schedule(new TimerTask()
            {
              public void run()
              {
                display.syncExec(new Runnable()
                {
                  public void run()
                  {
                    String filters[] = StringUtils.splitArray(widgetIssuesTreeFind.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
                    filterIssues(widgetIssuesTree,selectedProject,filters);
                  }
                });
              }
            },
            250  // delay [ms]
            );
          }
        });

        widgetEditProjectIssue = Widgets.newButton(subComposite,"Edit"+"\u2026",Settings.keyEditTimeEntry);
        Widgets.layout(widgetEditProjectIssue,0,2,TableLayoutData.E,0,0,0,0,80,SWT.DEFAULT);
        widgetEditProjectIssue.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Issue issues[] = getSelectedProjectIssueTreeEntries(widgetIssuesTree);
            if (issues.length > 0)
            {
              // edit issue
              if (editIssue(issues[0],"Edit issue","Save"))
              {
                // update issue
                try
                {
                  issues[0].setModified(Backend.ModifiedStates.UPDATED);
                  Widgets.notify(shell,USER_EVENT_UPDATE_ISSUE,issues[0]);
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot update issue (error: "+exception.getMessage()+")");
                  return;
                }
                catch (InterruptedException exception)
                {
                  // ignored
                }
              }
            }
          }
        });

        widgetAddProjectIssue = Widgets.newButton(subComposite,"Add"+"\u2026",Settings.keyNewTimeEntry);
        Widgets.layout(widgetAddProjectIssue,0,3,TableLayoutData.E,0,0,0,0,80,SWT.DEFAULT);
        widgetAddProjectIssue.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TreeItem treeItems[] = widgetIssuesTree.getSelection();
            if (treeItems.length > 0)
            {
              Issue issue = (Issue)treeItems[0].getData();

              // create new issue
              final Issue newIssue = new Issue(issue.backendType,
                                               issue.projectId
                                              );
              switch (issue.backendType)
              {
                case REDMINE:
                  issue.statusId  = Backend.getStatusId(Settings.Redmine.newIssueStatusName);
                  issue.trackerId = Backend.getTrackerId(Settings.Redmine.newIssueTrackerName);
                  break;
                case JIRA:
                  issue.statusId  = Backend.getStatusId(Settings.JIRA.newIssueStatusName);
                  issue.trackerId = Backend.getTrackerId(Settings.JIRA.newIssueTrackerName);
                  break;
              }
              if (editIssue(newIssue,"Add issue","Add"))
              {
                try
                {
                  newIssue.setModified(Backend.ModifiedStates.ADDED);
                  Widgets.notify(shell,USER_EVENT_ADD_ISSUE,issue);
                }
                catch (BackendException exception)
                {
                  printStackTrace(exception);
                  Dialogs.error(shell,"Cannot add issue (error: "+exception.getMessage()+")");
                  return;
                }
                catch (InterruptedException exception)
                {
                  // ignored
                }
              }
            }
          }
        });

        widgetDeleteProjectIssue = Widgets.newButton(subComposite,"Delete"+"\u2026",Settings.keyDeleteTimeEntry);
        Widgets.layout(widgetDeleteProjectIssue,0,4,TableLayoutData.E,0,0,0,0,80,SWT.DEFAULT);
        widgetDeleteProjectIssue.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            TreeItem[] treeItems = widgetIssuesTree.getSelection();
            if (treeItems.length > 0)
            {
              // get time entries to delete
              IssueList     issueList     = new IssueList();
              TimeEntryList timeEntryList = new TimeEntryList();
              try
              {
                for (TreeItem treeItem : treeItems)
                {
                  if       (treeItem.getData() instanceof Issue)
                  {
                    Issue issue = (Issue)treeItem.getData();

                    issueList.add(issue);
                    for (TimeEntry timeEntry : issue.getTimeEntryList())
                    {
                      assert timeEntry != null;
                      timeEntryList.add(timeEntry);
                    }
                  }
                  else if (treeItem.getData() instanceof SpentOn)
                  {
                    for (TimeEntry timeEntry : Backend.getTimeEntryList((SpentOn)treeItem.getData()))
                    {
                      assert timeEntry != null;
                      timeEntryList.add(timeEntry);
                    }
                  }
                  else if  (treeItem.getData() instanceof TimeEntry)
                  {
                    TimeEntry timeEntry = (TimeEntry)treeItem.getData();
                    assert timeEntry != null;
                    timeEntryList.add(timeEntry);
                  }
                }
              }
              catch (BackendException exception)
              {
                printStackTrace(exception);
                Dialogs.error(shell,"Cannot get time entries to delete (error: "+exception.getMessage()+")");
                return;
              }

              if (issueList.size() > 0)
              {
                if (Dialogs.confirm(shell,"Delete "+issueList.size()+" issues?"))
                {
                  try
                  {
                    for (Issue issue : issueList)
                    {
                      removeIssue(issue);
                    }
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    Dialogs.error(shell,"Cannot delete issues (error: "+exception.getMessage()+")");
                    return;
                  }
                }
              }
              if (timeEntryList.size() > 0)
              {
                if (Dialogs.confirm(shell,"Delete "+timeEntryList.size()+" time entries?"))
                {
                  try
                  {
                    for (TimeEntry timeEntry : timeEntryList)
                    {
                      removeTimeEntry(timeEntry);
                    }
                  }
                  catch (BackendException exception)
                  {
                    printStackTrace(exception);
                    Dialogs.error(shell,"Cannot delete time entries (error: "+exception.getMessage()+")");
                    return;
                  }
                }
              }
            }
          }
        });
      }
    }

    // add shortcut listeners
    Listener keyListener = new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        if      (Widgets.isAccelerator(event,Settings.keyFind))
        {
          switch (widgetTabFolder.getSelectionIndex())
          {
            case 0: Widgets.setFocus(addTimeEntry1.widgetFind); break;
            case 1: Widgets.setFocus(addTimeEntry2.widgetFind); break;
            case 2: Widgets.setFocus(widgetProjectTreeFind);      break;
          }
          event.doit = false;
        }
        else if (Widgets.isAccelerator(event,Settings.keyFindIssue))
        {
          switch (widgetTabFolder.getSelectionIndex())
          {
            case 3: Widgets.setFocus(widgetIssuesTreeFind); break;
          }
          event.doit = false;
        }
        else if (Widgets.isAccelerator(event,Settings.keyProject))
        {
          switch (widgetTabFolder.getSelectionIndex())
          {
            case 0: Widgets.setFocus(addTimeEntry1.widgetProject); break;
            case 1: Widgets.setFocus(addTimeEntry2.widgetProject); break;
          }
          event.doit = false;
        }
        else if (Widgets.isAccelerator(event,Settings.keyIssue))
        {
          switch (widgetTabFolder.getSelectionIndex())
          {
            case 0: Widgets.setFocus(addTimeEntry1.widgetIssueSummary); break;
            case 1: Widgets.setFocus(addTimeEntry2.widgetIssueSummary); break;
          }
          event.doit = false;
        }
        else if (Widgets.isAccelerator(event,Settings.keySpent))
        {
          switch (widgetTabFolder.getSelectionIndex())
          {
            case 0: Widgets.setFocus(addTimeEntry1.widgetSpentHourFraction); break;
            case 1: Widgets.setFocus(addTimeEntry2.widgetSpentHourFraction); break;
          }
          event.doit = false;
        }
        else if ((addTimeEntry1.widgetActivities != null) && (addTimeEntry2.widgetActivities != null) && Widgets.isAccelerator(event,Settings.keyActivity))
        {
          switch (widgetTabFolder.getSelectionIndex())
          {
            case 0: Widgets.setFocus(addTimeEntry1.widgetActivities); break;
            case 1: Widgets.setFocus(addTimeEntry2.widgetActivities); break;
          }
          event.doit = false;
        }
        else if (Widgets.isAccelerator(event,Settings.keyComments))
        {
          switch (widgetTabFolder.getSelectionIndex())
          {
            case 0: Widgets.setFocus(addTimeEntry1.widgetComments); break;
            case 1: Widgets.setFocus(addTimeEntry2.widgetComments); break;
          }
          event.doit = false;
        }
        else if (Widgets.isAccelerator(event,Settings.keyAdd))
        {
          switch (widgetTabFolder.getSelectionIndex())
          {
            case 0: Widgets.setFocus(addTimeEntry1.widgetTimeEntryAdd); break;
            case 1: Widgets.setFocus(addTimeEntry2.widgetTimeEntryAdd); break;
          }
          event.doit = false;
        }
      }
    };
    display.addFilter(SWT.KeyDown,keyListener);

    // create status line
    composite = Widgets.newComposite(shell);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0}));
    Widgets.layout(composite,2,0,TableLayoutData.WE);
    {
      label = Widgets.newLabel(composite,"Status:");
      Widgets.layout(label,0,0,TableLayoutData.W,0,0,2);

      widgetStatus = Widgets.newView(composite,"",SWT.NONE);
      Widgets.layout(widgetStatus,0,1,TableLayoutData.WE);

      widgetSynchronizeAnimation = synchronizeAnimation.newButton(composite);
      Widgets.layout(widgetSynchronizeAnimation,0,2,TableLayoutData.E,0,0,2,2);
      widgetSynchronizeAnimation.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          if (!Backend.isOnline())
          {
            Backend.online();
          }

          synchronizeTask.trigger();
        }
      });
    }

    Widgets.showTab(widgetTabFolder,widgetTabDay);
    Widgets.setFocus(widgetTodayTimeEntryTable);
  }

  /** create menu
   */
  private void createMenu()
  {
    Menu     menuBar;
    Menu     menu,subMenu;
    MenuItem menuItem;

    // create menu
    menuBar = Widgets.newMenuBar(shell);

    menu = Widgets.addMenu(menuBar,"Program");
    {
      subMenu = Widgets.addMenu(menu,"Start working"+"\u2026");
      {
        for (int time = 7*60; time <= 15*60; time += 15)
        {
          String timeString = String.format("%02d:%02d",time/60,time%60);
          menuItem = Widgets.addMenuItem(subMenu,timeString,SWT.NONE);
          menuItem.setData(time);
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              MenuItem menuItem = (MenuItem)selectionEvent.widget;
              int time  = (Integer)menuItem.getData();

              Calendar start = Calendar.getInstance();
              start.set(Calendar.HOUR_OF_DAY,time/60);
              start.set(Calendar.MINUTE,time%60);
              start.set(Calendar.SECOND,0);
              restartTodayTimer(start.getTimeInMillis()/1000);
            }
          });
        }
      }

      Widgets.addMenuItemSeparator(menu);

      subMenu = Widgets.addMenu(menu,"Database");
      {
        menuItem = Widgets.addMenuItem(subMenu,"Incremental refresh");
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            if (!Backend.isOnline())
            {
              if (!Dialogs.confirm(shell,"Switch to online mode for database incremental refresh?"))
              {
                return;
              }
              Backend.online();
            }

            incrementalRefreshTask.trigger();
          }
        });

        menuItemRefresh = Widgets.addMenuItem(subMenu,"Refresh"+"\u2026",Settings.keyRefresh);
        menuItemRefresh.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            // check if online
            if (!Backend.isOnline())
            {
              if (!Dialogs.confirm(shell,"Switch to online mode for database refresh?"))
              {
                return;
              }
              Backend.online();
            }

            // force synchronize
            synchronizeTask.forceRun();

            refreshDatabase();
          }
        });

        menuItemRefresh = Widgets.addMenuItem(subMenu,"Clean-up local"+"\u2026",Settings.keyCleanUpLocalDatabase);
        menuItemRefresh.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            cleanUpLocalDatabase();
          }
        });

        Widgets.addMenuItemSeparator(subMenu);

        menuItemReinit = Widgets.addMenuItem(subMenu,"Re-init"+"\u2026");
        menuItemReinit.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            if (!Backend.isOnline())
            {
              if (!Dialogs.confirm(shell,"Switch to online mode for database reinit?"))
              {
                return;
              }
              Backend.online();
            }

            reinitDatabase();
          }
        });
      }

      menuItemSynchronize = Widgets.addMenuItem(menu,"Synchronize",Settings.keySynchronize);
      menuItemSynchronize.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          if (!Backend.isOnline())
          {
            if (!Dialogs.confirm(shell,"Switch to online mode for database sychronize?"))
            {
              return;
            }
            Backend.online();
          }

          synchronizeTask.trigger();
        }
      });

      menuItem = Widgets.addMenuItem(menu,"Preferences"+"\u2026",Settings.keyPreferences);
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          new Preferences(shell);
        }
      });

      Widgets.addMenuItemSeparator(menu);

      menuItem = Widgets.addMenuItemCheckbox(menu,"Online",Settings.keyToggleOnlineOffline);
      menuItem.setSelection(Backend.isOnline());
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          MenuItem menuItem = (MenuItem)selectionEvent.widget;
          if (menuItem.getSelection())
          {
            Backend.online();
          }
          else
          {
            Backend.offline();
          }
        }
      });
      Backend.addConnectionStatesListener(new Backend.StateListener(menuItem)
      {
        public void changed(Backend.ConnectionStates connectionState)
        {
          MenuItem menuItem = (MenuItem)userData;

          menuItem.setSelection(connectionState == Backend.ConnectionStates.ONLINE);
        }
      });

      Widgets.addMenuItemSeparator(menu);

      menuItem = Widgets.addMenuItem(menu,"Save settings");
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Settings.save();
        }
      });

      menuItem = Widgets.addMenuItem(menu,"Restart");
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          // send close event with restart
          Widgets.notify(shell,USER_EVENT_QUIT,EXITCODE_RESTART);
        }
      });

      menuItem = Widgets.addMenuItem(menu,"Quit",Settings.keyQuit);
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          // send close-event to shell
          Widgets.notify(shell,USER_EVENT_QUIT,EXITCODE_OK);
        }
      });
    }

    widgetBookmarksMenu = Widgets.addMenu(menuBar,"Bookmarks");
    {
      menuItem = Widgets.addMenuItem(widgetBookmarksMenu,"Edit\u2026");
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          editBookmarks();
          updateBookmarksMenu(widgetBookmarksMenu,Settings.issueBookmarkList);
        }
      });

      Widgets.addMenuItemSeparator(widgetBookmarksMenu);

      updateBookmarksMenu(widgetBookmarksMenu,Settings.issueBookmarkList);
    }

    menu = Widgets.addMenu(menuBar,"Help");
    {
      menuItem = Widgets.addMenuItem(menu,"About");
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.info(shell,
                       "About",
                         "Timemine "+Config.VERSION_MAJOR+"."+Config.VERSION_MINOR+" (revision "+Config.VERSION_REVISION+").\n"
                       + "\n"
                       + "SWT: "+SWT.getPlatform()+" "+Integer.toString(SWT.getVersion())+"\n"
                       + "\n"
                       + "Written by Torsten Rupp.");
        }
      });
    }
  }

  /** create tray item
   */
  private void createTrayItem()
  {
    final Menu menu;
    MenuItem   menuItem;

    tray = display.getSystemTray();
    if (tray != null)
    {
      trayItem = new TrayItem (tray, SWT.NONE);
      trayItem.setToolTipText("Timemine - project time utility");
      trayItem.setImage(IMAGE_PROGRAM_ICON);
      trayItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          if (!shell.isVisible())
          {
            shell.setVisible(true);
            shell.setActive();
          }
          else
          {
            shell.setVisible(false);
          }
        }
      });

      menu = new Menu(shell,SWT.POP_UP);
      menuItem = Widgets.addMenuItem(menu,"Restart");
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          // send close-event to shell
          Widgets.notify(shell,USER_EVENT_QUIT,EXITCODE_RESTART);
        }
      });
      menuItem = Widgets.addMenuItem(menu,"Quit");
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          // send close-event to shell
          Widgets.notify(shell,USER_EVENT_QUIT,EXITCODE_OK);
        }
      });
      trayItem.addListener(SWT.MenuDetect,new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          menu.setVisible(true);
        }
      });
    }
    else
    {
      trayItem = null;
    }
  }

  /** create event handlers
   */
  private void createEventHandlers()
  {
  }

  /** open splash screen
   */
  private void openSplashScreen()
  {
    final int SPLASH_WIDTH  = 280;
    final int SPLASH_HEIGHT = 180;

    Control control;

    widgetSplashScreen = new Shell(SWT.ON_TOP);
    Rectangle bounds = display.getPrimaryMonitor().getBounds();
    widgetSplashScreen.setLocation((bounds.width-SPLASH_WIDTH)/2,(bounds.height-SPLASH_HEIGHT)/2);
    widgetSplashScreen.setSize(SPLASH_WIDTH,SPLASH_HEIGHT);
    widgetSplashScreen.setLayout(new TableLayout(new double[]{1.0,0.0},new double[]{1.0,0.0}));
    {
      control = Widgets.newImage(widgetSplashScreen,IMAGE_SPLASH_SCREEN,SWT.RIGHT);
      Widgets.layout(control,0,0,TableLayoutData.NSWE,0,0,5,5);

      widgetSplashStatus = Widgets.newLabel(widgetSplashScreen);
      Widgets.layout(widgetSplashStatus,1,0,TableLayoutData.WE,2,0,5,5,SWT.DEFAULT,30);
    }
    widgetSplashScreen.open();
    display.update();
  }

  /** close splash screen
   */
  private void closeSplashScreen()
  {
    widgetSplashScreen.close();
    widgetSplashScreen = null;
  }

  /** set splash screen status/status line
   * @param format format
   * @param arguments optional arguments
   */
  private StatusText setStatus(String format, Object... arguments)
  {
    StatusText statusText = new StatusText(format,arguments);

    synchronized(statusTextStack)
    {
      statusTextStack.push(statusText);
      updateStatus(statusText.get());
    }

    return statusText;
  }

  /** set splash screen status/status line
   * @param statusText status text
   */
  private void setStatus(StatusText statusText)
  {
    synchronized(statusTextStack)
    {
      statusTextStack.push(statusText);
      updateStatus(statusText.get());
    }
  }

  /** clear splash screen status/status line
   */
  private void clearStatus()
  {
    synchronized(statusTextStack)
    {
      assert !statusTextStack.isEmpty();
      statusTextStack.pop();
      StatusText statusText = !statusTextStack.isEmpty() ? statusTextStack.peek() : (StatusText)null;
      if (statusText != null)
      {
        updateStatus(statusText);
      }
      else
      {
        updateStatus("");
      }
    }
  }

  /** update splash screen status/status line
   * @param format format
   * @param arguments optional arguments
   */
  private void updateStatus(final String text)
  {
    if (!display.isDisposed())
    {
      if (widgetSplashScreen != null)
      {
        display.asyncExec(new Runnable()
        {
          public void run()
          {
            if (!widgetSplashStatus.isDisposed()) widgetSplashStatus.setText(text);
            display.update();
          }
        });
      }
      else
      {
        display.asyncExec(new Runnable()
        {
          public void run()
          {
            if (!widgetStatus.isDisposed()) widgetStatus.setText(text);
            display.update();
          }
        });
      }
    }
  }

  /** update splash screen status/status line
   * @param statusText status text
   */
  private void updateStatus(StatusText statusText)
  {
    synchronized(statusTextStack)
    {
      if (statusTextStack.peek() == statusText)
      {
        updateStatus(statusText.get());
      }
    }
  }

  /** init all
   */
  private void initAll()
  {
    Runnable runnable;

    // work-around for annoying GTK warnings on stderr: close stderr and use replacement
    if (!Settings.showAllErrorOutputFlag)
    {
      System.err.close();
      errorStream = new PrintStream(new FileOutputStream(FileDescriptor.out),true);
      Dprintf.setOutputStream(errorStream);
    }
    else
    {
      errorStream = System.err;
    }

    // init display
    initDisplay();

    // init holiday days database
    initHolidayDatabase();

    // connect to server
    LoginData loginData[] = new LoginData[]
                            {
                              new LoginData(Backend.Types.REDMINE,
                                            Settings.Redmine.serverName,
                                            Settings.Redmine.serverPort,
                                            Settings.Redmine.serverUseSSL,
                                            Settings.Redmine.loginName,
                                            Settings.Redmine.loginPassword,
                                            Settings.Redmine.certificateType,
                                            Settings.Redmine.certificateFileName,
                                            Settings.Redmine.keyFileName,
                                            Settings.Redmine.p12FileName,
                                            Settings.Redmine.keyPassword,
                                            Settings.Redmine.enabled
                                           ),
                              new LoginData(Backend.Types.JIRA,
                                            Settings.JIRA.serverName,
                                            Settings.JIRA.serverPort,
                                            Settings.JIRA.serverUseSSL,
                                            Settings.JIRA.loginName,
                                            Settings.JIRA.loginPassword,
                                            Settings.JIRA.certificateType,
                                            Settings.JIRA.certificateFileName,
                                            Settings.JIRA.keyFileName,
                                            Settings.JIRA.p12FileName,
                                            Settings.JIRA.keyPassword,
                                            Settings.JIRA.enabled
                                           ),
                            };

    Backend.ConnectionStates connectionState[]   = {Settings.connectionState};
    boolean                 connectOkFlags[]    = new boolean[]{false,false};
    boolean                 refreshDatabaseFlag = false;
    if (!Settings.loginDialogFlag)
    {
      // try to connect to servers with given credentials
      for (int i = 0; i < loginData.length; i++)
      {
        if (   loginData[i].enabled
            && !loginData[i].loginName.isEmpty()
            && !loginData[i].loginPassword.isEmpty()
           )
        {
          try
          {
            Backend.initDriver(display,
                               loginData[i].backendType,
                               loginData[i].serverName,
                               loginData[i].serverPort,
                               loginData[i].serverUseSSL,
                               loginData[i].loginName,
                               loginData[i].loginPassword,
                               loginData[i].certificateType,
                               loginData[i].certificateFileName,
                               loginData[i].keyFileName,
                               loginData[i].p12FileName,
                               loginData[i].keyPassword
                              );
            connectOkFlags[i] = true;
          }
          catch (BackendException exception)
          {
            // ignored
            if (Settings.debugTypes.contains(Settings.DebugTypes.ERROR))
            {
              printStackTrace(exception);
            }
          }
        }
      }
    }

    while (   (   (loginData[0].enabled && !connectOkFlags[0])
               || (loginData[1].enabled && !connectOkFlags[1])
              )
           && (connectionState[0] != Backend.ConnectionStates.OFFLINE)
          )
    {
      // get login data
      if (!getLoginData(loginData,connectionState))
      {
        System.exit(EXITCODE_OK);
      }

      switch (connectionState[0])
      {
        case ONLINE:
          // try to connect to server
          boolean retryFlag = false;
          do
          {
            Backend.doneDrivers();
            for (int i = 0; i < loginData.length; i++)
            {
              if (   loginData[i].enabled
                  && !loginData[i].loginName.isEmpty()
                  && !loginData[i].loginPassword.isEmpty()
                 )
              {
                try
                {
                  Backend.initDriver(display,
                                     loginData[i].backendType,
                                     loginData[i].serverName,
                                     loginData[i].serverPort,
                                     loginData[i].serverUseSSL,
                                     loginData[i].loginName,
                                     loginData[i].loginPassword,
                                     loginData[i].certificateType,
                                     loginData[i].certificateFileName,
                                     loginData[i].keyFileName,
                                     loginData[i].p12FileName,
                                     loginData[i].keyPassword
                                    );
                  connectOkFlags[i] = true;
                }
                catch (BackendException exception)
                {
                  if (Settings.debugTypes.contains(Settings.DebugTypes.ERROR))
                  {
                    printStackTrace(exception);
                  }
                  switch (Dialogs.select(new Shell(display,SWT.SHELL_TRIM),
                                         "Connection fail",
                                         "Error: "+exception.getMessage()+((exception.getCause() != null) ? ":\n\n"+exception.getCause().getMessage() : ""),
                                         new String[]{"Try again","Login dialog","Offline","Cancel"},
                                         0
                                        )
                     )
                  {
                    case 0:
                      // retry
                      retryFlag = true;
                      break;
                    case 1:
                      // show login dialog
                      break;
                    case 2:
                      // work offline
                      connectionState[0] = Backend.ConnectionStates.OFFLINE;
                      break;
                    case 3:
                      // terminate
                      System.exit(EXITCODE_FAIL);
                      break;
                    default:
                      System.exit(EXITCODE_FAIL);
                      break;
                  }
                }
              }
            }
          }
          while (retryFlag);
          break;
        case OFFLINE:
          retryFlag = false;
          break;
      }
    }
    Settings.connectionState = connectionState[0];

    // add watchdog for loaded classes/JARs
    initClassesWatchDog();

    // init backend
    try
    {
      Backend.init(Settings.connectionState,
                   Settings.refreshTime*60*3   // 3x: usually let the background task do the update
                  );
    }
    catch (BackendException exception)
    {
      printStackTrace(exception);
      Dialogs.error(new Shell(display,SWT.SHELL_TRIM),
                    "Initialize backend fail",
                    "Error: "+exception.getMessage()+((exception.getCause() != null) ? ":\n\n"+exception.getCause().getMessage() : "")
                   );
      System.exit(EXITCODE_FAIL);
    }
    Backend.addConnectionStatesListener(new Backend.StateListener()
    {
      public void changed(Backend.ConnectionStates connectionState)
      {
        Settings.connectionState = connectionState;
      }
    });
        // create main window and menu, set layout
    createWindow();
    createMenu();
    createTrayItem();
    createEventHandlers();

    // listener
    shell.addListener(USER_EVENT_QUIT,new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        // store geometry
        Settings.geometryMain                  = shell.getSize();
        Settings.geometryTodayTimeSizes        = new SettingUtils.SimpleDoubleArray(widgetTodayPane.getSizes());
        Settings.geometryTodayTimeEntryColumns = new Settings.ColumnSizes(Widgets.getTableColumnWidth(widgetTodayTimeEntryTable));
        Settings.geometryActionTableColumns    = new Settings.ColumnSizes(widgetTodayActionTable.getColumnWidth());
        Settings.geometryWeekTimeEntryColumns  = new Settings.ColumnSizes(Widgets.getTreeColumnWidth(widgetWeeksTimeEntryTree));
        Settings.geometryProjectsColumns       = new Settings.ColumnSizes(Widgets.getTreeColumnWidth(widgetProjectTree));
        Settings.geometryIssuesColumns         = new Settings.ColumnSizes(Widgets.getTreeColumnWidth(widgetIssuesTree));

        // save settings
        boolean saveSettings = true;
        if (Settings.isExternalModified())
        {
          switch (Dialogs.select(shell,
                                 "Confirmation","Settings were modified externally.",
                                 new String[]{"Overwrite","Just quit","Cancel"},
                                 0
                                )
                 )
          {
            case 0:
              break;
            case 1:
              saveSettings = false;
              break;
            case 2:
              event.doit = false;
              return;
          }
        }
        if (saveSettings)
        {
          Settings.save();
        }

        // store exitcode
        exitcode = event.index;

        // close
        shell.dispose();
        display.dispose();
      }
    });

    display.addFilter(SWT.KeyDown,new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        if      (Widgets.isAccelerator(event,Settings.keyTabFavoriteProjects))
        {
          Widgets.showTab(widgetTabFolder,widgetTabFavoriteProjects);
          event.doit = false;
        }
        else if (Widgets.isAccelerator(event,Settings.keyTabTodayTimeEntries))
        {
          Widgets.showTab(widgetTabFolder,widgetTabDay);
          event.doit = false;
        }
        else if (Widgets.isAccelerator(event,Settings.keyTabTimeEntries))
        {
          Widgets.showTab(widgetTabFolder,widgetTabWeeks);
//          Widgets.setFocus(widgetWeeksTimeEntryTree);
          event.doit = false;
        }
        else if (Widgets.isAccelerator(event,Settings.keyTabProjectsIssues))
        {
          Widgets.showTab(widgetTabFolder,widgetTabProjectsIssues);
          event.doit = false;
        }
/*
        else if (Widgets.isAccelerator(event,Settings.keyNewTimeEntry))
        {
Dprintf.dprintf("");
          TimeEntry timeEntry = new TimeEntry(Id.NONE,
                                              Id.NONE,
                                              Backend.getDefaultActivityId(),
                                              Settings.minTimeDelta,
                                              ""
                                             );

          if (editTimeEntry(timeEntry))
          {
Dprintf.dprintf("");
          }
        }
*/
      }
    });

    // start today timer
    updateTodayTimer.scheduleAtFixedRate(new TimerTask()
      {
        public void run()
        {
          synchronized(todayTimeLock)
          {
            if (todayTimeStartDateTime != null)
            {
              // reset/increment today time
              final int newTodayTime[] = new int[1];
              final Calendar now = Calendar.getInstance();
              if (   (todayTimeStartDateTime.get(Calendar.YEAR        ) != now.get(Calendar.YEAR        ))
                  || (todayTimeStartDateTime.get(Calendar.MONTH       ) != now.get(Calendar.MONTH       ))
                  || (todayTimeStartDateTime.get(Calendar.DAY_OF_MONTH) != now.get(Calendar.DAY_OF_MONTH))
                 )
              {
                // day changed (Note: asyncExec() to avoid dead-lock!)
                display.asyncExec(new Runnable()
                {
                  public void run()
                  {
//                    widgetTodayWindowTimeLine.resetActionList();
//TODO
//                    widgetTodayActionTable.removeAll();
                  }
                });

                // reset today sum
                todayTimeStartDateTime = null;
                newTodayTime[0] = 0;
              }
              else
              {
                // increment today sum [min]
                newTodayTime[0] = todayTime.get()+1;
              }

              // redraw (Note: asyncExec() to avoid dead-lock!)
              if (!display.isDisposed())
              {
                display.asyncExec(new Runnable()
                {
                  public void run()
                  {
                    todayTime.set(newTodayTime[0]);
                  }
                });
              }
            }
          }
        }
      },
      0L,
      60*1000L
    );

    // add restart today timer listeners
    Listener restartTodayTimerListener = new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        restartTodayTimer();
      }
    };
    display.addFilter(SWT.KeyDown,restartTodayTimerListener);
    display.addFilter(SWT.MouseDown,restartTodayTimerListener);
    display.addFilter(SWT.MouseMove,restartTodayTimerListener);

    // set initial today start time
    long todayStartTimestamp = 0L;
    try
    {
      todayStartTimestamp = ActionMonitor.getInstance().getActionList().getStartTimestamp();
    }
    catch (Exception exception)
    {
      printInternalError(exception);
      System.exit(EXITCODE_INTERNAL_ERROR);
    }
    restartTodayTimer((todayStartTimestamp > 0L) ? todayStartTimestamp : System.currentTimeMillis()/1000L);

    // refresh local database thread
    refreshTask = new PeriodicTask("refresh",Settings.refreshTime*60,10)
    {
      final StatusText statusText = new StatusText();

      @Override
      public boolean isRunnable()
        throws BackendException
      {
        return Backend.isOnline();
      }
      @Override
      public void begin()
        throws BackendException
      {
        printDebug("Refresh local database");

        statusText.set("Refresh local database\u2026");
        setStatus(statusText);
        synchronizeAnimation.start();
      }
      @Override
      public void end()
        throws BackendException
      {
        printDebug("Refresh local database done");
        synchronizeAnimation.stop();
        clearStatus();
      }
      @Override
      public void updateProgress(double value)
      {
        statusText.set("Refresh local database\u2026 [%.1f%%]",value*100.0);
        updateStatus(statusText);
      }
      @Override
      public void run()
        throws BackendException, InterruptedException
      {
        updateLocalDatabaseRoles(backendProgress);
        backendProgress.setStep(1);
        updateLocalDatabaseTrackers(backendProgress);
        backendProgress.setStep(2);
        updateLocalDatabaseEpicNames(backendProgress);
        backendProgress.setStep(3);
        updateLocalDatabaseStatus(backendProgress);
        backendProgress.setStep(4);
        updateLocalDatabasePriorities(backendProgress);
        backendProgress.setStep(5);
        updateLocalDatabaseActivities(backendProgress);
        backendProgress.setStep(6);
        updateLocalDatabaseProjects(backendProgress);
        backendProgress.setStep(7);
        updateLocalDatabaseVersions(backendProgress);
        backendProgress.setStep(8);
        updateLocalDatabaseMembers(backendProgress);
        backendProgress.setStep(9);
        updateLocalDatabaseUsers(backendProgress); // Note: Workaround must be done _after_ update members. Bug in Backendn (#7773)
        backendProgress.setStep(10);
      }
    };

    // create refresh local database issues thread
    refreshIssuesTask = new PeriodicTask("refreshIssues",Settings.refreshIssuesTime*60)
    {
      final StatusText statusText = new StatusText();

      @Override
      public boolean isRunnable()
        throws BackendException
      {
        return Backend.isOnline();
      }
      @Override
      public void begin()
        throws BackendException
      {
        printDebug("Refresh local database issues");

        statusText.set("Refresh local database issues\u2026");
        setStatus(statusText);
        synchronizeAnimation.start();
      }
      @Override
      public void end()
        throws BackendException
      {
        printDebug("Refresh local database issues done");
        synchronizeAnimation.stop();
        clearStatus();
      }
      @Override
      public void updateProgress(double value)
      {
        statusText.set("Refresh local database issues\u2026 [%.1f%%]",value*100.0);
        updateStatus(statusText);
      }
      @Override
      public void run()
        throws BackendException, InterruptedException
      {
        updateLocalDatabaseIssues(Settings.maxIssuesHistory,backendProgress);
      }
    };

// TODO
    refreshTimeEntriesTask = new PeriodicTask("refreshTimeEntries",Settings.refreshTimeEntriesTime)
//    refreshTimeEntriesTask = new PeriodicTask("refreshTimeEntries",5)
    {
      final StatusText statusText = new StatusText();

      @Override
      public boolean isRunnable()
        throws BackendException
      {
        return Backend.isOnline();
      }
      @Override
      public void begin()
        throws BackendException
      {
        printDebug("Refresh local database time entries");

        statusText.set("Refresh local database time entries\u2026");
        setStatus(statusText);
        synchronizeAnimation.start();
      }
      @Override
      public void end()
        throws BackendException
      {
        printDebug("Refresh local database time entries done");
        synchronizeAnimation.stop();
        clearStatus();
      }
      @Override
      public void updateProgress(double value)
      {
        statusText.set("Refresh local database time entries\u2026 [%.1f%%]",value*100.0);
        updateStatus(statusText);
      }
      @Override
      public void run()
        throws BackendException, InterruptedException
      {
        updateLocalDatabaseTimeEntries(Settings.maxTimeEntriesHistory,backendProgress);
      }
    };

    incrementalRefreshTask = new PeriodicTask("incrementalRefresh",Settings.incrementalRefreshIssuesTime)
    {
      final StatusText statusText = new StatusText();

      @Override
      public boolean isRunnable()
        throws BackendException
      {
        return Backend.isOnline();
      }
      @Override
      public void begin()
        throws BackendException
      {
        printDebug("Incremental refresh local database");

        statusText.set("Incremental refresh local database\u2026");
        setStatus(statusText);
        synchronizeAnimation.start();
        backendProgress.beginSteps(2);
      }
      @Override
      public void end()
        throws BackendException
      {
        printDebug("Incremental refresh local database done");

        backendProgress.endSteps();
        synchronizeAnimation.stop();
        clearStatus();
      }
      @Override
      public void updateProgress(double value)
      {
        statusText.set("Incremental refresh local database\u2026 [%.1f%%]",value*100.0);
        updateStatus(statusText);
      }
      @Override
      public void run()
        throws BackendException, InterruptedException
      {
// TODO: working?
//        updateLocalDatabaseIssues(true,backendProgress);
        backendProgress.setStep(1);
        updateLocalDatabaseTimeEntries(true,backendProgress);
        backendProgress.setStep(2);
      }
    };

// TODO: revert
//    updateTimeEntriesTask = new PeriodicTask("updateTimeEntries",30,3)
updateTimeEntriesTask = new PeriodicTask("updateTimeEntries",50,3)
    {
      final StatusText statusText = new StatusText();

      @Override
      public void begin()
        throws BackendException
      {
        printDebug("Update time entries");
        setStatus(statusText);
        synchronizeAnimation.start();
      }
      @Override
      public void end()
        throws BackendException
      {
        printDebug("Update time entries done");
        synchronizeAnimation.stop();
        clearStatus();
      }
      @Override
      public void updateProgress(double value)
      {
        statusText.set("Update time entries\u2026 [%.1f%%]",value*100.0);
        updateStatus(statusText);
      }
      @Override
      public void run()
        throws BackendException, InterruptedException
      {
        // set number of time entry weeks in time entries tree
        try
        {
          final int timeEntryWeeks = (int)(((System.currentTimeMillis()-Backend.getTimeEntryStartDate().getTime())/(24*60*60*1000))+7-1)/7;
          if (!display.isDisposed())
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                if (!widgetWeeksTimeEntryTree.isDisposed())
                {
                  widgetWeeksTimeEntryTree.setItemCount(timeEntryWeeks);
                }
              }
            });
          }
        }
        catch (BackendException exception)
        {
          // ignored
          printStackTrace(exception);
        }
        backendProgress.setStep(1);

        if (!display.isDisposed())
        {
          display.syncExec(new Runnable()
          {
            public void run()
            {
              // update all tree items (Note: clearAll refresh visible items in virtual tree)
              if (!widgetWeeksTimeEntryTree.isDisposed())
              {
                widgetWeeksTimeEntryTree.clearAll(true);
              }
            }
          });
        }
        backendProgress.setStep(2);

        try
        {
          // get today time entries
          final TimeEntryList timeEntryList = Backend.getTimeEntryListToday();

          // update today time entry table
          if (!display.isDisposed())
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                // update entries, get not existing entries
                if (!widgetTodayTimeEntryTable.isDisposed())
                {
                  ArrayList<TableItem> existingTableItemList = new ArrayList<TableItem>();
                  for (TimeEntry timeEntry : timeEntryList)
                  {
                    existingTableItemList.add(updateInsertTableItem(widgetTodayTimeEntryTable,timeEntry));
                  }
                  Widgets.removeObsoleteTableItems(widgetTodayTimeEntryTable,existingTableItemList);

                  // update today duration sum
                  todayDurationSum.set(Backend.getTimeEntryTodayDurationSum());
                }
              }
            });
          }
        }
        catch (BackendException exception)
        {
          // ignored
          printStackTrace(exception);
        }
        backendProgress.setStep(3);
      }
    };

    // synchronize changed issues/time entries
    synchronizeTask = new PeriodicTask("synchronize",Settings.syncTime,2)
    {
      final StatusText statusText = new StatusText();

      @Override
      public boolean isRunnable()
        throws BackendException
      {
        return Backend.isOnline();
      }
      @Override
      public void begin()
        throws BackendException
      {
        printDebug("Synchronize");

        statusText.set("Auto refresh local database\u2026");
        setStatus(statusText);
        synchronizeAnimation.start();
      }
      @Override
      public void end()
        throws BackendException
      {
        printDebug("Synchronize done");
        synchronizeAnimation.stop();
        clearStatus();
      }
      @Override
      public void updateProgress(double value)
      {
        statusText.set("Incremental refresh local database\u2026 [%.1f%%]",value*100.0);
        updateStatus(statusText);
      }
      @Override
      public void run()
        throws BackendException, InterruptedException
      {
        final boolean doneFlag[] = new boolean[1];

        // synchronize issues
        doneFlag[0] = false;
        do
        {
          {
            if (!Backend.modifyTryLock(5000)) break;
          }
          try
          {
            final IssueModified issueModified = Backend.getIssueModified();
            if (issueModified != null)
            {
Dprintf.dprintf("IssueModified=%s\n",issueModified);
              final Issue issueOrigin = Backend.getIssueOrigin(Backend.TypeSet.ALL,issueModified.id);
Dprintf.dprintf("issueOrigin=%s\n",issueOrigin);

              if (issueOrigin != null)
              {
                // check changes
                if (!issueModified.trackerId.equals(issueOrigin.trackerId))
                {
                  final Tracker trackerModified = Backend.getTracker(new Backend.TypeSet(issueModified.backendType),issueModified.trackerId);
                  final Tracker trackerOrigin   = Backend.getTracker(new Backend.TypeSet(issueOrigin.backendType),issueOrigin.trackerId);
                  if (!display.isDisposed())
                  {
                    display.syncExec(new Runnable()
                    {
                      public void run()
                      {
                        switch (Dialogs.select(shell,
                                               "Conflict",
                                               "Conflict on issue tracker.",
                                               new String[]
                                               {
                                                 "keep own: "+trackerModified.getName(),
                                                 "server: "+trackerOrigin.getName()
                                               },
                                               "OK",
                                               "Cancel",
                                               1
                                              )
                              )
                        {
                          case 0: break;
                          case 1: issueModified.trackerId = issueOrigin.trackerId; break;
                          default: doneFlag[0] = true; break;
                        }
                      }
                    });
                  };
                  if (doneFlag[0]) break;
                }

                if (   Backend.getTracker(new Backend.TypeSet(issueModified.backendType),issueModified.trackerId).isRequireEpicName()
                    && !issueModified.epicName.equals(issueOrigin.epicName)
                   )
                {
                  if (!display.isDisposed())
                  {
                    display.syncExec(new Runnable()
                    {
                      public void run()
                      {
                        switch (Dialogs.select(shell,
                                               "Conflict",
                                               "Conflict on issue epic changes.",
                                               new String[]
                                               {
                                                 "keep own: "+issueModified.epicName,
                                                 "server: "+issueOrigin.epicName
                                               },
                                               "OK",
                                               "Cancel",
                                               1
                                              )
                              )
                        {
                          case 0: break;
                          case 1: issueModified.epicName = issueOrigin.epicName; break;
                          default: doneFlag[0] = true; break;
                        }
                      }
                    });
                  };
                  if (doneFlag[0]) break;
                }

                if (!issueModified.summary.equals(issueOrigin.summary))
                {
                  if (!display.isDisposed())
                  {
                    display.syncExec(new Runnable()
                    {
                      public void run()
                      {
                        switch (Dialogs.select(shell,
                                               "Conflict",
                                               "Conflict on issue summary changes.",
                                               new String[]
                                               {
                                                 "keep own: "+issueModified.summary,
                                                 "server: "+issueOrigin.summary,
                                                 "both: "+issueOrigin.summary+issueModified.summary
                                               },
                                               "OK",
                                               "Cancel",
                                               1
                                              )
                              )
                        {
                          case 0: break;
                          case 1: issueModified.summary = issueOrigin.summary; break;
                          case 2: issueModified.summary = issueOrigin.summary+issueModified.summary; break;
                          default: doneFlag[0] = true; break;
                        }
                      }
                    });
                  };
                  if (doneFlag[0]) break;
                }

                if (!issueModified.description.equals(issueOrigin.description))
                {
                  if (!display.isDisposed())
                  {
                    display.syncExec(new Runnable()
                    {
                      public void run()
                      {
                        switch (Dialogs.select(shell,
                                               "Conflict",
                                               "Conflict on issue description changes.",
                                               new String[]
                                               {
                                                 "keep own",
                                                 "server"
                                               },
                                               "OK",
                                               "Cancel",
                                               1
                                              )
                              )
                        {
                          case 0: break;
                          case 1: issueModified.description = issueOrigin.description; break;
                          default: doneFlag[0] = true; break;
                        }
                      }
                    });
                  };
                  if (doneFlag[0]) break;
                }

                if (!issueModified.assignedToMemberId.equals(issueOrigin.assignedToMemberId))
                {
                  final Member memberModified = Backend.getMember(new Backend.TypeSet(issueModified.backendType),issueModified.assignedToMemberId);
                  final Member memberOrigin   = Backend.getMember(new Backend.TypeSet(issueOrigin.backendType),issueOrigin.assignedToMemberId);
                  if (!display.isDisposed())
                  {
                    display.syncExec(new Runnable()
                    {
                      public void run()
                      {
                        switch (Dialogs.select(shell,
                                               "Conflict",
                                               "Conflict on issue assignment.",
                                               new String[]
                                               {
                                                 "keep own: "+memberModified.getName(),
                                                 "server: "+memberOrigin.getName()
                                               },
                                               "OK",
                                               "Cancel",
                                               1
                                              )
                              )
                        {
                          case 0: break;
                          case 1: issueModified.assignedToMemberId = issueOrigin.assignedToMemberId; break;
                          default: doneFlag[0] = true; break;
                        }
                      }
                    });
                  };
                  if (doneFlag[0]) break;
                }

                // TODO: status
                issueModified.statusId = issueOrigin.statusId;

                if (!issueModified.priorityId.equals(issueOrigin.priorityId))
                {
                  final Priority priorityModified = Backend.getPriority(new Backend.TypeSet(issueModified.backendType),issueModified.priorityId);
                  final Priority priorityOrigin   = Backend.getPriority(new Backend.TypeSet(issueOrigin.backendType),issueOrigin.priorityId);
                  if (!display.isDisposed())
                  {
                    display.syncExec(new Runnable()
                    {
                      public void run()
                      {
                        switch (Dialogs.select(shell,
                                               "Conflict",
                                               "Conflict on priority.",
                                               new String[]
                                               {
                                                 "keep own: "+priorityModified.getName(),
                                                 "server: "+priorityOrigin.getName()
                                               },
                                               "OK",
                                               "Cancel",
                                               1
                                              )
                              )
                        {
                          case 0: break;
                          case 1: issueModified.priorityId = issueOrigin.priorityId; break;
                          default: doneFlag[0] = true; break;
                        }
                      }
                    });
                  };
                  if (doneFlag[0]) break;
                }

                // TODO: target version
                issueModified.targetVersionId = issueOrigin.targetVersionId;

                if (issueModified.estimatedDuration != issueOrigin.estimatedDuration)
                {
                  if (!display.isDisposed())
                  {
                    display.syncExec(new Runnable()
                    {
                      public void run()
                      {
                        switch (Dialogs.select(shell,
                                               "Conflict",
                                               "Conflict on estimated duraction.",
                                               new String[]
                                               {
                                                 "keep own: "+Long.toString(issueModified.estimatedDuration/60)+"h:"+Long.toString(issueModified.estimatedDuration%60)+"m",
                                                 "server: "+Long.toString(issueOrigin.estimatedDuration/60)+"h:"+Long.toString(issueOrigin.estimatedDuration%60)+"m"
                                               },
                                               "OK",
                                               "Cancel",
                                               1
                                              )
                              )
                        {
                          case 0: break;
                          case 1: issueModified.priorityId = issueOrigin.priorityId; break;
                          default: doneFlag[0] = true; break;
                        }
                      }
                    });
                  };
                  if (doneFlag[0]) break;
                }
              }

              // sync with server
Dprintf.dprintf("issueModified=%s",issueModified);
              try
              {
                if      ((issueModified.modifiedState == Backend.ModifiedStates.ADDED) || (issueOrigin == null))
                {
                  issueModified.store();
                }
                else if (issueModified.modifiedState == Backend.ModifiedStates.UPDATED)
                {
                  issueModified.update(issueOrigin);
                }
                else if (issueModified.modifiedState == Backend.ModifiedStates.DELETED)
                {
                  issueModified.delete();
                }
                Widgets.notify(shell,USER_EVENT_UPDATE_ISSUE,issueModified);

                issueModified.doneModified();
              }
              catch (BackendException exception)
              {
                final String message = exception.getMessage();

                printDebug("Synchronize issue fail: %s",message);
                printStackTrace(exception);

                if (!display.isDisposed())
                {
                  display.syncExec(new Runnable()
                  {
                    public void run()
                    {
                      switch (Dialogs.select(shell,
                                             "Synchronize issue",
                                             "Synchronize issue\n\n"+issueModified.id.format()+": '"+issueModified.summary+"'\n\nfail with error:\n\n"+message,
                                             new String[]{"Try again",
                                                          "Discard local changes",
                                                          "Delete local issue",
                                                          "Work offline"
                                                         },
                                             0
                                            )
                             )
                      {
                        case 0:
                          // nothing to do
                          break;
                        case 1:
                          // discard local changes
                          try
                          {
                            issueModified.revertLocalDatabase();
                          }
                          catch (BackendException exception)
                          {
                            // ignored
                            printStackTrace(exception);
                          }
                          catch (InterruptedException exception)
                          {
                            // ignored
                          }
                          break;
                        case 2:
                          // delete local issue
                          try
                          {
                            issueModified.removeLocalDatabase();
                          }
                          catch (BackendException exception)
                          {
                            // ignored
                            printStackTrace(exception);
                          }
                          catch (InterruptedException exception)
                          {
                            // ignored
                          }
                          break;
                        case 3:
                          // work offline
                          Backend.offline();
                          doneFlag[0] = true;
                          break;
                      }
                    }
                  });
                }
              }
            }
            else
            {
              doneFlag[0] = true;
            }
          }
          finally
          {
            Backend.modifyUnlock();
          }
        }
        while (!doneFlag[0]);

        // synchronize time entries
        doneFlag[0] = false;
        do
        {
          {
            if (!Backend.modifyTryLock(5000)) break;
          }
          try
          {
            final TimeEntryModified timeEntryModified = Backend.getTimeEntryModified();
            if (timeEntryModified != null)
            {
Dprintf.dprintf("TimeEntryModified=%s",timeEntryModified);
              final TimeEntry timeEntryOrigin = Backend.getTimeEntryOrigin(Backend.TypeSet.ALL,timeEntryModified.id);
Dprintf.dprintf("timeEntryOrigin=%s\n",timeEntryOrigin);

              try
              {
                if      ((timeEntryModified.modifiedState == Backend.ModifiedStates.ADDED) || (timeEntryOrigin == null))
                {
                  timeEntryModified.store();
                }
                else if (timeEntryModified.modifiedState == Backend.ModifiedStates.UPDATED)
                {
                  timeEntryModified.update(timeEntryOrigin);
                }
                else if (timeEntryModified.modifiedState == Backend.ModifiedStates.DELETED)
                {
                  timeEntryModified.delete();
                }
                Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRY,timeEntryModified);

                timeEntryModified.doneModified();
              }
              catch (BackendException exception)
              {
                final String message = exception.getMessage();

                printDebug("Synchronize time entry fail: %s",message);
                printStackTrace(exception);

                final int result[] = new int[1];
                if (!display.isDisposed())
                {
                  display.syncExec(new Runnable()
                  {
                    public void run()
                    {
                      result[0] = Dialogs.select(shell,
                                                 "Synchronize time entry",
                                                 "Synchronize time entry\n\n"+timeEntryModified.id.format()+": '"+timeEntryModified.comments+"'\n\nfail with error:\n\n"+message,
                                                 new String[]{"Try again",
                                                              "Discard local changes",
                                                              "Delete local time entry",
                                                              "Work offline"
                                                             },
                                                 0
                                                );
                    }
                  });
                }
                switch (result[0])
                {
                  case 0:
                    // nothing to do
                    break;
                  case 1:
                    // discard local changes
                    try
                    {
                      timeEntryModified.revertLocalDatabase();
                    }
                    catch (BackendException exception2)
                    {
                      // ignored
                      printStackTrace(exception2);
                    }
                    catch (InterruptedException exception2)
                    {
                      // ignored
                    }
                    break;
                  case 2:
                    // delete local time entry
                    try
                    {
Dprintf.dprintf("_");
                      timeEntryModified.removeLocalDatabase();
Dprintf.dprintf("_");
                    }
                    catch (BackendException exception2)
                    {
                      // ignored
                      printStackTrace(exception2);
                    }
                    catch (InterruptedException exception2)
                    {
                      // ignored
                    }
                    break;
                  case 3:
                    // work offline
                    Backend.offline();
                    doneFlag[0] = true;
                    break;
                }
              }
            }
            else
            {
              doneFlag[0] = true;
            }
          }
          finally
          {
            Backend.modifyUnlock();
          }
        }
        while (!doneFlag[0]);
      }
    };

    // auto refresh local database
    autoRefreshTask = new PeriodicTask("autoRefresh",5*60,11)
    {
      final StatusText statusText = new StatusText();

      @Override
      public boolean isRunnable()
        throws BackendException, InterruptedException
      {
        // wait for auto refresh time window
        if  (   !Settings.autoRefresh
             || (System.currentTimeMillis() <= Settings.autoRefreshStartTime.get())
             || (System.currentTimeMillis() >= Settings.autoRefreshEndTime.get())
            )
        {
          return false;
        }

        // random wait time auto refresh time window
        long n = (long)((double)((Settings.autoRefreshEndTime.get()-Settings.autoRefreshStartTime.get())/1000)*Math.random());
        while (n > 0)
        {
          Thread.sleep(10*1000);
          n = (n > 10) ? n-10 : 0;
        }

        return true;
      }
      @Override
      public void begin()
        throws BackendException
      {
        printDebug("Auto refresh local database");

        statusText.set("Auto refresh local database\u2026");
        setStatus(statusText);
        synchronizeAnimation.start();
      }
      @Override
      public void end()
        throws BackendException
      {
        printDebug("Auto refresh local database done");
        synchronizeAnimation.stop();
        clearStatus();
      }
      @Override
      public void updateProgress(double value)
      {
        statusText.set("Incremental refresh local database\u2026 [%.1f%%]",value*100.0);
        updateStatus(statusText);
      }
      @Override
      public void run()
        throws BackendException, InterruptedException
      {
        Backend.updateLocalDatabaseRoles(backendProgress);
        backendProgress.setStep(1);
        Backend.updateLocalDatabaseTrackers(backendProgress);
        backendProgress.setStep(2);
        Backend.updateLocalDatabaseEpicNames(backendProgress);
        backendProgress.setStep(3);
        Backend.updateLocalDatabaseStatus(backendProgress);
        backendProgress.setStep(4);
        Backend.updateLocalDatabasePriorities(backendProgress);
        backendProgress.setStep(5);
        Backend.updateLocalDatabaseActivities(backendProgress);
        backendProgress.setStep(6);
        Backend.updateLocalDatabaseProjects(backendProgress);
        backendProgress.setStep(7);
        Backend.updateLocalDatabaseVersions(backendProgress);
        backendProgress.setStep(8);
        Backend.updateLocalDatabaseMembers(backendProgress);
        backendProgress.setStep(9);
        Backend.updateLocalDatabaseUsers(backendProgress); // Note: Workaround must be done _after_ update members. Bug in Redmine backend (#7773)
        backendProgress.setStep(10);
        updateLocalDatabaseIssues(backendProgress);
        backendProgress.setStep(11);
      }
    };

    // create action monitor thread
    runnable = new Runnable()
    {
      public void run()
      {
        ActionMonitor actionMonitor   = ActionMonitor.getInstance();
        String        lastWindowTitle = "";
        String        lastApplication = "";
        long          lastIdleTime    = -1;

        while (!quitFlag)
        {
          try
          {
            // start external action monitor as separated process
            String javaVM = System.getProperty("java.home")+File.separator+"bin"+File.separator+"java";
            ProcessBuilder processBuilder = new ProcessBuilder(javaVM,
                                                               "-Djava.library.path=",
                                                               "-classpath",
                                                               System.getProperty("java.class.path"),
                                                               Timemine.class.getCanonicalName(),
                                                               "--action-monitor"
                                                              );
//Dprintf.dprintf("comm=%s",StringUtils.join(processBuilder.command()));
            Process process = processBuilder.start();

            BufferedReader input = new BufferedReader(new InputStreamReader(process.getInputStream(),"UTF-8"));
            String line;
            while (   !quitFlag
                   && ((line = input.readLine()) != null)
                  )
            {
//Dprintf.dprintf("line=%s",line);
              if      (line.startsWith(ActionMonitor.PREFIX_WINDOW))
              {
                String windowTitle = line.substring(ActionMonitor.PREFIX_WINDOW.length()).trim();
//Dprintf.dprintf("windowTitle=%s",windowTitle);

                if (!windowTitle.isEmpty())
                {
                  if (!windowTitle.equals(lastWindowTitle))
                  {
                    // add action
                    actionMonitor.addAction(ActionTypes.WINDOW,windowTitle);
                    lastWindowTitle = windowTitle;
                  }
                }
              }
              else if (line.startsWith(ActionMonitor.PREFIX_APPLICATION))
              {
                String application = line.substring(ActionMonitor.PREFIX_APPLICATION.length()).trim();
//Dprintf.dprintf("application=%s",application);

                if (!application.isEmpty())
                {
                  if (!application.equals(lastApplication))
                  {
                    // add action
                    actionMonitor.addAction(ActionTypes.APPLICATION,application);
                    lastApplication = application;
                  }
                }
              }
              else if (line.startsWith(ActionMonitor.PREFIX_IDLE))
              {
                long idleTime = Long.parseLong(line.substring(ActionMonitor.PREFIX_IDLE.length()).trim());
//Dprintf.dprintf("idleTime=%d",idleTime);

                if      ((lastIdleTime < 0) || (idleTime < lastIdleTime))
                {
                  // add action
                  actionMonitor.addAction(ActionTypes.ACTIVE,"active");

                  // restart today timer if needed
                  if (!display.isDisposed())
                  {
                    display.syncExec(new Runnable()
                    {
                      public void run()
                      {
                        restartTodayTimer();
                      }
                    });
                  }

                  lastIdleTime = 0L;
                }
                else if ((lastIdleTime > 0) && (idleTime > lastIdleTime))
                {
                  lastIdleTime = idleTime;
                }
                else if (idleTime > 60*1000)
                {
                  // add action
                  actionMonitor.addAction(ActionTypes.IDLE,"idle");
                  lastIdleTime = idleTime;
                }
              }
              else
              {
                printWarning("Unknown action '%s'",line);
              }
              if (quitFlag) break;
            }
            process.getOutputStream().close();
            process.getInputStream().close();
            process.getErrorStream().close();
            process.destroy();// } catch (InterruptedException exception) { /* ignored */ }
            process.waitFor();
            int exitCode = process.exitValue();
            if (exitCode != 0)
            {
              printInternalError("Action monitor process returned exitcode %d",exitCode);
            }
          }
          catch (InterruptedException exception)
          {
            quitFlag = true;
          }
          catch (Throwable throwable)
          {
            printInternalError(throwable);
          }
          if (quitFlag) break;

          // sleep a short time
          try
          {
            Thread.sleep(60*1000);
          }
          catch (InterruptedException exception)
          {
            quitFlag = true;
          }
        }
Dprintf.dprintf("aciton theread done");
      }
    };
    actionMonitorThread = new Thread(runnable,"actionMonitorThread");
    actionMonitorThread.setDaemon(true);

    // show splash screen/open main window
    if (Settings.showSplashScreen)
    {
      // open spash screen
      openSplashScreen();
    }
    else
    {
      // set window size, manage window (Note: for some reason the layout may change after the Window is open)
      shell.setSize(Settings.geometryMain.x,Settings.geometryMain.y);
      widgetTodayPane.setSizes(new double[]{0.85,0.15});
      shell.open();
      shell.setSize(Settings.geometryMain.x,Settings.geometryMain.y);
      widgetTodayPane.setSizes(new double[]{0.85,0.15});
    }

    // start and get initial data
    quitFlag = false;
    refreshTask.start();
    refreshIssuesTask.start();
    incrementalRefreshTask.start();
    refreshTimeEntriesTask.start();
    updateTimeEntriesTask.start();
    actionMonitorThread.start();

    if (Settings.refreshAtStart)
    {
      // get initial data from server
      refreshTask.trigger();
      refreshIssuesTask.trigger();
      refreshTimeEntriesTask.trigger();
    }

    updateTimeEntriesTask.trigger();

    // start synchronize, auto refresh
    synchronizeTask.start();
    autoRefreshTask.start();

    // close splash screen/open main window
    if (Settings.showSplashScreen)
    {
      closeSplashScreen();

      // set window size, manage window (Note: for some reason the layout may change after the Windows is open)
      shell.setSize(Settings.geometryMain.x,Settings.geometryMain.y);
      shell.open();
      shell.setSize(Settings.geometryMain.x,Settings.geometryMain.y);
    }

    // set sizes
    widgetTodayPane.setSizes(Settings.geometryTodayTimeSizes.get());
    Widgets.setTableColumnWidth(widgetTodayTimeEntryTable,Settings.geometryTodayTimeEntryColumns.width);
    Widgets.setTreeColumnWidth(widgetWeeksTimeEntryTree,Settings.geometryWeekTimeEntryColumns.width);
    Widgets.setTreeColumnWidth(widgetProjectTree,Settings.geometryProjectsColumns.width);
    Widgets.setTreeColumnWidth(widgetIssuesTree,Settings.geometryIssuesColumns.width);

    // permit modification of settings
    Settings.permitModified();

    // store stettings
    Settings.Redmine.serverName          = loginData[0].serverName;
    Settings.Redmine.serverPort          = loginData[0].serverPort;
    Settings.Redmine.serverUseSSL        = loginData[0].serverUseSSL;
    Settings.Redmine.loginName           = loginData[0].loginName;
    Settings.Redmine.certificateType     = loginData[0].certificateType;
    Settings.Redmine.certificateFileName = loginData[0].certificateFileName;
    Settings.Redmine.keyFileName         = loginData[0].keyFileName;
    Settings.Redmine.p12FileName         = loginData[0].p12FileName;
    Settings.Redmine.keyPassword         = loginData[0].keyPassword;
    Settings.JIRA.serverName             = loginData[1].serverName;
    Settings.JIRA.serverPort             = loginData[1].serverPort;
    Settings.JIRA.serverUseSSL           = loginData[1].serverUseSSL;
    Settings.JIRA.loginName              = loginData[1].loginName;
    Settings.JIRA.certificateType        = loginData[1].certificateType;
    Settings.JIRA.certificateFileName    = loginData[1].certificateFileName;
    Settings.JIRA.keyFileName            = loginData[1].keyFileName;
    Settings.JIRA.p12FileName            = loginData[1].p12FileName;
    Settings.JIRA.keyPassword            = loginData[1].keyPassword;
    Settings.connectionState             = connectionState[0];
    Settings.setModified();

    // trigger initial update
    if (refreshDatabaseFlag)
    {
      refreshDatabase(EnumSet.allOf(Settings.RefreshTypes.class),true);
    }
    else
    {
      EnumSet<Settings.RefreshTypes> refreshTypes = EnumSet.noneOf(Settings.RefreshTypes.class);

      try
      {
        if (!Backend.hasUsers())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No users - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.USERS);
          }
        }
        if (!Backend.hasRoles())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No roles - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.ROLES);
          }
        }
        if (!Backend.hasTrackers())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No trackers - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.TRACKERS);
          }
        }
        if (!Backend.hasStatus())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No status transitions - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.STATUS);
          }
        }
        if (!Backend.hasPriorities())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No priorities - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.PRIORITIES);
          }
        }
        if (!Backend.hasActivities())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No activities - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.ACTIVITIES);
          }
        }
        if (!Backend.hasProjects())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No projects - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.PROJECTS);
          }
        }
        if (!Backend.hasMembers())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No members - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.MEMBERS);
          }
        }
        if (!Backend.hasVersions())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No versions - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.VERSIONS);
          }
        }
        if (!Backend.hasIssues())
        {
          if (!refreshTypes.isEmpty() || Dialogs.confirm(shell,"No issues - refresh database?"))
          {
            refreshTypes.add(Settings.RefreshTypes.ISSUES);
          }
        }
      }
      catch (BackendException exception)
      {
        // ignored
        printStackTrace(exception);
      }

      if (!refreshTypes.isEmpty())
      {
        if (!Backend.isOnline())
        {
          Backend.online();
        }
        refreshDatabase(refreshTypes);
      }
    }

    Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRIES);

    try
    {
      RoleList roleList = Backend.getRoleList();
      Widgets.notify(shell,USER_EVENT_UPDATE_ROLES,roleList);
    }
    catch (BackendException exception)
    {
      // ignored
      printStackTrace(exception);
    }

    try
    {
      TrackerList trackerList = Backend.getTrackerList();
      Widgets.notify(shell,USER_EVENT_UPDATE_TRACKERS,trackerList);
    }
    catch (BackendException exception)
    {
      // ignored
      printStackTrace(exception);
    }
//TODO: remove
/*
    try
    {
      StatussTransitionList statussTransitionList = Backend.getStatusTransitionList();
      Widgets.notify(shell,USER_EVENT_UPDATE_STATUS_TRANSITION,statussTransitionList);
    }
    catch (BackendException exception)
    {
      // ignored
      printStackTrace(exception);
    }
*/
//TODO: remove
/*
    try
    {
      PriorityList priorityList = Backend.getPriorityList();
      Widgets.notify(shell,USER_EVENT_UPDATE_PRIORITIES,priorityList);
    }
    catch (BackendException exception)
    {
      // ignored
      printStackTrace(exception);
    }
    try
    {
      ActivityList activityList = Backend.getActivityList();
Dprintf.dprintf("");
      Widgets.notify(shell,USER_EVENT_UPDATE_ACTIVITIES,activityList);
    }
    catch (BackendException exception)
    {
      // ignored
      printStackTrace(exception);
    }
*/
    try
    {
      ProjectList projectList = Backend.getProjectList();
      Widgets.notify(shell,USER_EVENT_UPDATE_PROJECTS,projectList);
    }
    catch (BackendException exception)
    {
      // ignored
      printStackTrace(exception);
    }
  }

  /** done all
   */
  private void doneAll()
  {
    // shutdown running background threads
    quitFlag = true;
    try
    {
      actionMonitorThread.interrupt();
      actionMonitorThread.join();
    }
    catch (InterruptedException exception)
    {
      // ignored
    }
    autoRefreshTask.quit();
    synchronizeTask.quit();
    updateTimeEntriesTask.quit();
    refreshTimeEntriesTask.quit();
    incrementalRefreshTask.quit();
    refreshIssuesTask.quit();
    refreshTask.quit();

    // close holiday days database
    doneHolidayDatabase();
  }

  /** run application
   * @return exit code
   */
  private void run()
  {
    // SWT event loop
    while (!shell.isDisposed())
    {
//errorStream.print(".");
      if (!display.readAndDispatch())
      {
        display.sleep();
      }
    }
  }

  /** server/password dialog
   * @param loginData server login data
   * @param state state
   * @return true iff login data ok, false otherwise
   */
  private boolean getLoginData(final LoginData loginData[], final Backend.ConnectionStates connectionState[])
  {
    Composite       composite,subComposite,subSubComposite,subSubSubComposite;
    TableLayout     tableLayout;
    TableLayoutData tableLayoutData;
    Group           group;
    Separator       separator;
    Label           label;
    Button          button;

    connectionState[0] = Backend.ConnectionStates.ONLINE;

    final Shell dialog = Dialogs.openModal(shell,"Login servers",600,SWT.DEFAULT);

    final Button[]    widgetEnabled                       = new Button[loginData.length];
    final Text[]      widgetServerName                    = new Text[loginData.length];
    final Spinner[]   widgetServerPort                    = new Spinner[loginData.length];
    final Button[]    widgetServerUseSSL                  = new Button[loginData.length];
    final Text[]      widgetLoginName                     = new Text[loginData.length];
    final Text[]      widgetLoginPassword                 = new Text[loginData.length];
    final Button[]    widgetCertificateTypeCertificateKey = new Button[loginData.length];
    final Button[]    widgetCertificateTypeP12            = new Button[loginData.length];
    final Composite[] widgetCertificateKeyComposite       = new Composite[loginData.length];
    final Text[]      widgetCertificateFileName           = new Text[loginData.length];
    final Text[]      widgetKeyFileName                   = new Text[loginData.length];
    final Composite[] widgetP12Composite                  = new Composite[loginData.length];
    final Text[]      widgetP12FileName                   = new Text[loginData.length];
    final Text[]      widgetKeyPassword                   = new Text[loginData.length];
    final Button      widgetLoginButton;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{0.0,1.0,0.0,1.0},1.0,2));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE);
    {
      for (int i = 0; i < loginData.length; i++)
      {
        final int index = i;

        separator = Widgets.newSeparator(composite,loginData[i].backendType.toString());
        separator.setLayout(new TableLayout(0.0,1.0));
        Widgets.layout(separator,i*2+0,0,TableLayoutData.NSWE,0,2);

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0}));
        Widgets.layout(subComposite,i*2+1,0,TableLayoutData.NSWE,0,2);
        {
          widgetEnabled[i] = Widgets.newCheckbox(subComposite);
          widgetEnabled[i].setSelection(loginData[i].enabled);
          Widgets.layout(widgetEnabled[i],0,0,TableLayoutData.W);

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(0.0,new double[]{1.0,0.0,0.0,0.0}));
          Widgets.layout(subSubComposite,0,1,TableLayoutData.WE,0,2);
          {
            widgetServerName[i] = Widgets.newText(subSubComposite,SWT.LEFT|SWT.BORDER);
            if (loginData[i].serverName != null) widgetServerName[i].setText(loginData[i].serverName);
            Widgets.layout(widgetServerName[i],0,0,TableLayoutData.WE);

            label = Widgets.newLabel(subSubComposite,"Port:");
            Widgets.layout(label,0,1,TableLayoutData.W);

            widgetServerPort[i] = Widgets.newSpinner(subSubComposite,1,65535);
            widgetServerPort[i].setTextLimit(5);
            widgetServerPort[i].setSelection(loginData[i].serverPort);
            Widgets.layout(widgetServerPort[i],0,2,TableLayoutData.E,0,0,0,0,60,SWT.DEFAULT);

            widgetServerUseSSL[i] = Widgets.newCheckbox(subSubComposite,"SSL");
            widgetServerUseSSL[i].setSelection(loginData[i].serverUseSSL);
            Widgets.layout(widgetServerUseSSL[i],0,3,TableLayoutData.E);
          }

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0}));
          Widgets.layout(subSubComposite,1,1,TableLayoutData.WE,0,2);
          {
            label = Widgets.newLabel(subSubComposite,"Login name:");
            Widgets.layout(label,0,0,TableLayoutData.W);
            widgetLoginName[i] = Widgets.newText(subSubComposite);
            if (loginData[i].loginName != null) widgetLoginName[i].setText(loginData[i].loginName);
            Widgets.layout(widgetLoginName[i],0,1,TableLayoutData.WE);

            label = Widgets.newLabel(subSubComposite,"Login password:");
            Widgets.layout(label,1,0,TableLayoutData.W);
            widgetLoginPassword[i] = Widgets.newPassword(subSubComposite);
            if (loginData[i].loginPassword != null) widgetLoginPassword[i].setText(loginData[i].loginPassword);
            Widgets.layout(widgetLoginPassword[i],1,1,TableLayoutData.WE);

            subSubSubComposite = Widgets.newComposite(subSubComposite);
            subSubSubComposite.setLayout(new TableLayout(0.0,0.0));
            Widgets.layout(subSubSubComposite,2,1,TableLayoutData.WE);
            {
              widgetCertificateTypeCertificateKey[i] = Widgets.newRadio(subSubSubComposite,"Certificate+Key");
              widgetCertificateTypeCertificateKey[i].setSelection(loginData[i].certificateType == Settings.CertificateTypes.CERTIFICATE_KEY);
              Widgets.layout(widgetCertificateTypeCertificateKey[i],0,0,TableLayoutData.W);
              widgetCertificateTypeP12[i] = Widgets.newRadio(subSubSubComposite,"P12");
              widgetCertificateTypeP12[i].setSelection(loginData[i].certificateType == Settings.CertificateTypes.P12);
              Widgets.layout(widgetCertificateTypeP12[i],0,1,TableLayoutData.W);
            }

            widgetCertificateKeyComposite[i] = Widgets.newComposite(subSubComposite);
            widgetCertificateKeyComposite[i].setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0}));
            widgetCertificateKeyComposite[i].setVisible(loginData[i].certificateType == Settings.CertificateTypes.CERTIFICATE_KEY);
            Widgets.layout(widgetCertificateKeyComposite[i],3,1,TableLayoutData.NWE);
            {
              label = Widgets.newLabel(widgetCertificateKeyComposite[i],"Certificate:");
              Widgets.layout(label,0,0,TableLayoutData.W);

              widgetCertificateFileName[i] = Widgets.newText(widgetCertificateKeyComposite[i]);
              if (loginData[i].certificateFileName != null) widgetCertificateFileName[i].setText(loginData[i].certificateFileName);
              Widgets.layout(widgetCertificateFileName[i],0,1,TableLayoutData.WE);

              button = Widgets.newButton(widgetCertificateKeyComposite[i],Timemine.IMAGE_DIRECTORY);
              Widgets.layout(button,0,2,TableLayoutData.E);
              button.addSelectionListener(new SelectionListener()
              {
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  String fileName = Dialogs.fileOpen(dialog,
                                                     "Select certificate",
                                                     widgetCertificateFileName[index].getText(),
                                                     new String[]{"Certificates","*.crt",
                                                                  "All files",   Timemine.ALL_FILE_EXTENSION
                                                                 }
                                                    );
                  if (fileName != null)
                  {
                    widgetCertificateFileName[index].setText(fileName);
                  }
                }
              });

              label = Widgets.newLabel(widgetCertificateKeyComposite[i],"Key:");
              Widgets.layout(label,1,0,TableLayoutData.W);

              widgetKeyFileName[i] = Widgets.newText(widgetCertificateKeyComposite[i]);
              if (loginData[i].keyFileName != null) widgetKeyFileName[i].setText(loginData[i].keyFileName);
              Widgets.layout(widgetKeyFileName[i],1,1,TableLayoutData.WE);

              button = Widgets.newButton(widgetCertificateKeyComposite[i],Timemine.IMAGE_DIRECTORY);
              Widgets.layout(button,1,2,TableLayoutData.E);
              button.addSelectionListener(new SelectionListener()
              {
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  String fileName = Dialogs.fileOpen(dialog,
                                                     "Select key",
                                                     widgetKeyFileName[index].getText(),
                                                     new String[]{"PEM keys", "*.pem",
                                                                  "Keys",     "*.key",
                                                                  "All files",Timemine.ALL_FILE_EXTENSION
                                                                 }
                                                    );
                  if (fileName != null)
                  {
                    widgetKeyFileName[index].setText(fileName);
                  }
                }
              });
            }

            widgetP12Composite[i] = Widgets.newComposite(subSubComposite);
            widgetP12Composite[i].setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0}));
            widgetP12Composite[i].setVisible(loginData[i].certificateType == Settings.CertificateTypes.P12);
            Widgets.layout(widgetP12Composite[i],3,1,TableLayoutData.NWE);
            {
              label = Widgets.newLabel(widgetP12Composite[i],"P12:");
              Widgets.layout(label,0,0,TableLayoutData.W);

              widgetP12FileName[i] = Widgets.newText(widgetP12Composite[i]);
              if (loginData[i].p12FileName != null) widgetP12FileName[i].setText(loginData[i].p12FileName);
              Widgets.layout(widgetP12FileName[i],0,1,TableLayoutData.WE);

              button = Widgets.newButton(widgetP12Composite[i],Timemine.IMAGE_DIRECTORY);
              Widgets.layout(button,0,2,TableLayoutData.E);
              button.addSelectionListener(new SelectionListener()
              {
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  String fileName = Dialogs.fileOpen(dialog,
                                                     "Select P12",
                                                     widgetCertificateFileName[index].getText(),
                                                     new String[]{"P12","*.p12",
                                                                  "All files",   Timemine.ALL_FILE_EXTENSION
                                                                 }
                                                    );
                  if (fileName != null)
                  {
                    widgetP12FileName[index].setText(fileName);
                  }
                }
              });
            }

            label = Widgets.newLabel(subSubComposite,"Key password:");
            Widgets.layout(label,4,0,TableLayoutData.W);
            widgetKeyPassword[i] = Widgets.newPassword(subSubComposite);
            if (loginData[i].keyPassword != null) widgetKeyPassword[i].setText(loginData[i].keyPassword);
            Widgets.layout(widgetKeyPassword[i],4,1,TableLayoutData.WE);
          }
        }
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0,2));
    Widgets.layout(composite,1,0,TableLayoutData.WE);
    {
      widgetLoginButton = Widgets.newButton(composite,"Login");
      Widgets.layout(widgetLoginButton,0,0,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
      widgetLoginButton.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          for (int i = 0; i < loginData.length; i++)
          {
            loginData[i].enabled             = widgetEnabled[i].getSelection();
            loginData[i].serverName          = widgetServerName[i].getText();
            loginData[i].serverPort          = widgetServerPort[i].getSelection();
            loginData[i].serverUseSSL        = widgetServerUseSSL[i].getSelection();
            loginData[i].loginName           = widgetLoginName[i].getText();
            loginData[i].loginPassword       = widgetLoginPassword[i].getText();
            if      (widgetCertificateTypeCertificateKey[i].getSelection())
            {
              loginData[i].certificateType = Settings.CertificateTypes.CERTIFICATE_KEY;
            }
            else if (widgetCertificateTypeP12[i].getSelection())
            {
              loginData[i].certificateType = Settings.CertificateTypes.P12;
            }
            else
            {
              loginData[i].certificateType = Settings.CertificateTypes.NONE;
            }
            loginData[i].certificateFileName = widgetCertificateFileName[i].getText();
            loginData[i].keyFileName         = widgetKeyFileName[i].getText();
            loginData[i].p12FileName         = widgetP12FileName[i].getText();
            loginData[i].keyPassword         = widgetKeyPassword[i].getText();
          }
          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Work offline");
      Widgets.layout(button,0,1,TableLayoutData.NONE);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          connectionState[0] = Backend.ConnectionStates.OFFLINE;
          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,2,TableLayoutData.E,0,0,0,0,60,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // add listeners
    for (int i = 0; i < loginData.length; i++)
    {
      final int index = i;
      widgetEnabled[i].addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          widgetServerName[index].setEnabled(widget.getSelection());
          widgetServerPort[index].setEnabled(widget.getSelection());
          widgetServerUseSSL[index].setEnabled(widget.getSelection());
          widgetLoginName[index].setEnabled(widget.getSelection());
          widgetLoginPassword[index].setEnabled(widget.getSelection());

          widgetLoginButton.setEnabled(widgetEnabled[0].getSelection() || widgetEnabled[1].getSelection());
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          widgetServerName[index].setEnabled(widget.getSelection());
          widgetServerPort[index].setEnabled(widget.getSelection());
          widgetServerUseSSL[index].setEnabled(widget.getSelection());
          widgetLoginName[index].setEnabled(widget.getSelection());
          widgetLoginPassword[index].setEnabled(widget.getSelection());

          widgetLoginButton.setEnabled(widgetEnabled[0].getSelection() || widgetEnabled[1].getSelection());
        }
      });

      widgetCertificateTypeCertificateKey[index].addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          widgetCertificateKeyComposite[index].setVisible(true);
          widgetP12Composite[index].setVisible(false);
        }
      });
      widgetCertificateTypeP12[i].addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          widgetCertificateKeyComposite[index].setVisible(false);
          widgetP12Composite[index].setVisible(true);
        }
      });
    }

    // set next focus
    Widgets.setNextFocus(widgetServerName[0],
                         widgetServerPort[0],
                         widgetServerUseSSL[0],
                         widgetLoginName[0],
                         widgetLoginPassword[0],
                         widgetCertificateFileName[0],
                         widgetKeyFileName[0],
                         widgetKeyPassword[0],
                         widgetServerName[1],
                         widgetServerPort[1],
                         widgetServerUseSSL[1],
                         widgetLoginName[1],
                         widgetLoginPassword[1],
                         widgetCertificateFileName[1],
                         widgetKeyFileName[1],
                         widgetP12FileName[1],
                         widgetKeyPassword[1],
                         widgetLoginButton
                        );

    // run
    if      (loginData[0].enabled && (loginData[0].serverName != null) && !loginData[0].serverName.isEmpty() && widgetLoginName[0].getText().isEmpty()) Widgets.setFocus(widgetLoginName[0]);
    else if (loginData[0].enabled && (loginData[0].serverName != null) && !loginData[0].serverName.isEmpty() && widgetLoginPassword[0].getText().isEmpty()) Widgets.setFocus(widgetLoginPassword[0]);
    else if (loginData[1].enabled && (loginData[1].serverName != null) && !loginData[1].serverName.isEmpty() && widgetLoginName[1].getText().isEmpty()) Widgets.setFocus(widgetLoginPassword[1]);
    else if (loginData[1].enabled && (loginData[1].serverName != null) && !loginData[1].serverName.isEmpty() && widgetLoginPassword[1].getText().isEmpty()) Widgets.setFocus(widgetLoginPassword[1]);
    Boolean result = (Boolean)Dialogs.run(dialog);

    return (result != null) ? result : false;
  }

  /** get index in integer array
   * @param array integer array
   * @param value value to find
   * @return index or -1
   */
  private int getIndex(int[] array, int value)
  {
    for (int i = 0; i < array.length; i++)
    {
      if (array[i] == value) return i;
    }
    return -1;
  }

  /** format hours/minutes into string
   * @param minutes minutes
   * @return string
   */
  private String formatHours(long minutes)
  {
    return String.format("%02d:%02d",minutes/60,minutes%60);
  }

  /** restart today timer if day changed or forced
   * @param dateTime date/time or 0 for current date/time [s]
   */
  private void restartTodayTimer(long dateTime)
  {
    final Calendar now = Calendar.getInstance();

//Dprintf.dprintf("todayTimeStartDateTime=null:%s dateTime=%d today=%d now=%d hang=%s",todayTimeStartDateTime==null,dateTime,(todayTimeStartDateTime != null) ? todayTimeStartDateTime.get(Calendar.DAY_OF_MONTH) : -1,now.get(Calendar.DAY_OF_MONTH),(todayTimeStartDateTime == null) || ((todayTimeStartDateTime.get(Calendar.YEAR        ) != now.get(Calendar.YEAR        ))          || (todayTimeStartDateTime.get(Calendar.MONTH       ) != now.get(Calendar.MONTH       ))          || (todayTimeStartDateTime.get(Calendar.DAY_OF_MONTH) != now.get(Calendar.DAY_OF_MONTH))));
    synchronized(todayTimeLock)
    {
      if (   (todayTimeStartDateTime == null)
          || (dateTime != 0)
          || (todayTimeStartDateTime.get(Calendar.YEAR        ) != now.get(Calendar.YEAR        ))
          || (todayTimeStartDateTime.get(Calendar.MONTH       ) != now.get(Calendar.MONTH       ))
          || (todayTimeStartDateTime.get(Calendar.DAY_OF_MONTH) != now.get(Calendar.DAY_OF_MONTH))
         )
      {
        // set start time
//Dprintf.dprintf("new todayTimeStartDateTime");
        todayTimeStartDateTime = Calendar.getInstance();
        if (dateTime != 0)
        {
          todayTimeStartDateTime.setTimeInMillis(dateTime*1000L);
        }
        todayTime.set(0);
//Dprintf.dprintf("new todayTimeStartDateTime: %s",todayTimeStartDateTime);

        // force refresh action/tag list
        ActionMonitor.getInstance().refresh();

        // update time line
        long startDateTime = todayTimeStartDateTime.getTimeInMillis()/1000L;
        widgetTodayWindowTimeLine.setDate(startDateTime);
        widgetTodayWindowTimeLine.setStartTime(startDateTime);

        // update today start time, today time sum
        if (dateTime == 0)
        {
          long actionStartTimestamp = ActionMonitor.getInstance().getActionList().getStartTimestamp()*1000L;
          todayTimeStartDateTime.setTimeInMillis((actionStartTimestamp > 0) ? actionStartTimestamp : now.getTimeInMillis());
        }
        todayTime.set((int)((now.getTimeInMillis()-todayTimeStartDateTime.getTimeInMillis())/(1000L*60L)));
      }
    }
  }

  /** force restart today timer with current date/time
   */
  private void restartTodayTimer()
  {
    restartTodayTimer(0L);
  }

  /** update local data base roles
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseRoles(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseRoles(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_ROLES);
  }

  /** update local data base trackers
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseTrackers(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseTrackers(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_TRACKERS);
  }

  /** update local data base epic names
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseEpicNames(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseEpicNames(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_EPIC_NAMES);
  }

  /** update local data base status
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseStatus(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseStatus(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_STATUS_TRANSITION);
  }

  /** update local data base priorities
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabasePriorities(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabasePriorities(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_PRIORITIES);
  }

  /** update local data base activities
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseActivities(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseActivities(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_ACTIVITIES);
  }

  /** update local data base projects
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseProjects(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseProjects(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_PROJECTS);
  }

  /** update local data base versions
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseVersions(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseVersions(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_VERSIONS);
  }

  /** update local data base members
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseMembers(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseMembers(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_MEMBERS);
  }

  /** update local data base users
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseUsers(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseUsers(backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_USERS);
  }

  /** update local data base issues
   * @param incrementalFlag true for incremental update
   * @param maxIssuesHistory max. issues history or 0
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseIssues(boolean incrementalFlag, int maxIssuesHistory, BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseIssues(incrementalFlag,maxIssuesHistory,backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_ISSUES);
  }

  /** update local data base issues
   * @param maxIssuesHistory max. issues history or 0
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseIssues(int maxIssuesHistory, BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    updateLocalDatabaseIssues(false,maxIssuesHistory,backendProgress);
  }

  /** update local data base issues
   * @param incrementalFlag true for incremental update
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseIssues(boolean incrementalFlag, BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    updateLocalDatabaseIssues(incrementalFlag,0,backendProgress);
  }

  /** update local data base issues
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseIssues(BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    updateLocalDatabaseIssues(false,backendProgress);
  }

  /** update local data base time entries
   * @param incrementalFlag true for incremental update
   * @param maxTimeEntriesHistory max. time entries history or 0
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseTimeEntries(boolean incrementalFlag, int maxTimeEntriesHistory, BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    Backend.updateLocalDatabaseTimeEntries(incrementalFlag,maxTimeEntriesHistory,backendProgress);

    Widgets.notify(shell,USER_EVENT_UPDATE_TIME_ENTRIES);
  }

  /** update local data base time entries
   * @param maxTimeEntriesHistory max. time entries history or 0
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseTimeEntries(int maxTimeEntriesHistory, BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    updateLocalDatabaseTimeEntries(false,maxTimeEntriesHistory,backendProgress);
  }

  /** update local data base time entries
   * @param incrementalFlag true for incremental update
   * @param backendProgress progress handler or null
   */
  private void updateLocalDatabaseTimeEntries(boolean incrementalFlag, BackendProgress backendProgress)
    throws BackendException, InterruptedException
  {
    updateLocalDatabaseTimeEntries(incrementalFlag,0,backendProgress);
  }

  /** refresh local database
   * @param refreshTypes pre-selected refresh types (can be null for default)
   * @param immediateFlag true to start refresh entries immediately
   */
  private void refreshDatabase(EnumSet<Settings.RefreshTypes> refreshTypes, boolean immediateFlag)
  {
    class Data
    {
      EnumSet<Settings.RefreshTypes> refreshTypes;
      int                            maxIssuesHistory;
      int                            maxTimeEntriesHistory;
      boolean                        runFlag;
      boolean                        abortFlag;

      Data(EnumSet<Settings.RefreshTypes> refreshTypes)
      {
        this.refreshTypes          = (refreshTypes != null) ? refreshTypes : Settings.refreshTypes;
        this.maxIssuesHistory      = Settings.maxIssuesHistory;
        this.maxTimeEntriesHistory = Settings.maxTimeEntriesHistory;
        this.runFlag               = false;
        this.abortFlag             = false;
      }
    };

    final Shell dialog;
    Composite   composite,subComposite;
    Label       label;
    Combo       combo;
    Button      button;

    final Data  data = new Data(refreshTypes);

    dialog = Dialogs.open(shell,"Refresh");
    dialog.setLayout(new TableLayout(0.0,1.0));

    final Button      widgetUsers;
    final Button      widgetRoles;
    final Button      widgetTracker;
    final Button      widgetStatus;
    final Button      widgetPriorities;
    final Button      widgetActivities;
    final Button      widgetProject;
    final Button      widgetMembers;
    final Button      widgetVersions;
    final Button      widgetIssues;
    final Combo       widgetMaxIssuesHistory;
    final Button      widgetTimeEntries;
    final Combo       widgetMaxTimeEntriesHistory;
    final Button      widgetAll;
    final ProgressBar widgetProgressBar;
    final Text        widgetInfo;
    final Button      widgetAutoClose;
    final Button      widgetStartStop;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0,4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      widgetUsers = Widgets.newCheckbox(composite,"Users");
      widgetUsers.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.USERS));
      Widgets.layout(widgetUsers,0,0,TableLayoutData.W);
      widgetUsers.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.USERS);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.USERS);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetUsers,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
          }
        }
      });

      widgetRoles = Widgets.newCheckbox(composite,"Roles");
      widgetRoles.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.ROLES));
      Widgets.layout(widgetRoles,1,0,TableLayoutData.W);
      widgetRoles.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.ROLES);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.ROLES);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetRoles,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
          }
        }
      });

      widgetTracker = Widgets.newCheckbox(composite,"Trackers");
      widgetTracker.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.TRACKERS));
      Widgets.layout(widgetTracker,2,0,TableLayoutData.W);
      widgetTracker.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.TRACKERS);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.TRACKERS);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetTracker,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
          }
        }
      });

      widgetStatus = Widgets.newCheckbox(composite,"Status");
      widgetStatus.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.STATUS));
      Widgets.layout(widgetStatus,3,0,TableLayoutData.W);
      widgetStatus.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.STATUS);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.STATUS);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetStatus,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
          }
        }
      });

      widgetPriorities = Widgets.newCheckbox(composite,"Priorities");
      widgetPriorities.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.PRIORITIES));
      Widgets.layout(widgetPriorities,4,0,TableLayoutData.W);
      widgetPriorities.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.PRIORITIES);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.PRIORITIES);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetPriorities,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
          }
        }
      });

      widgetActivities = Widgets.newCheckbox(composite,"Activities");
      widgetActivities.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.ACTIVITIES));
      Widgets.layout(widgetActivities,5,0,TableLayoutData.W);
      widgetActivities.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.ACTIVITIES);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.ACTIVITIES);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetActivities,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
          }
        }
      });

      widgetProject = Widgets.newCheckbox(composite,"Projects");
      widgetProject.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.PROJECTS));
      Widgets.layout(widgetProject,6,0,TableLayoutData.W);
      widgetProject.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.PROJECTS);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.PROJECTS);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetProject,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
          }
        }
      });

      widgetMembers = Widgets.newCheckbox(composite,"Members");
      widgetMembers.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.MEMBERS));
      Widgets.layout(widgetMembers,7,0,TableLayoutData.W);
      widgetMembers.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.MEMBERS);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.MEMBERS);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetMembers,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
          }
        }
      });

      widgetVersions = Widgets.newCheckbox(composite,"Versions");
      widgetVersions.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.VERSIONS));
      Widgets.layout(widgetVersions,8,0,TableLayoutData.W);
      widgetVersions.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.VERSIONS);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.VERSIONS);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetVersions,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
          }
        }
      });

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,9,0,TableLayoutData.WE);
      {
        widgetIssues = Widgets.newCheckbox(subComposite,"Issues");
        widgetIssues.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.ISSUES));
        Widgets.layout(widgetIssues,0,0,TableLayoutData.W);
        widgetIssues.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button widget = (Button)selectionEvent.widget;
            if (widget.getSelection())
            {
              data.refreshTypes.add(Settings.RefreshTypes.ISSUES);
            }
            else
            {
              data.refreshTypes.remove(Settings.RefreshTypes.ISSUES);
            }
            Widgets.modified(data);
          }
        });
        Widgets.addModifyListener(new WidgetModifyListener(widgetIssues,data)
        {
          @Override
          public void modified(Button button)
          {
            if (!button.isDisposed())
            {
              button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
            }
          }
        });

        widgetMaxIssuesHistory = Widgets.newOptionMenu(subComposite);
        widgetMaxIssuesHistory.setToolTipText("Number of days to refresh time entries history.");
        Widgets.setOptionMenuItems(widgetMaxIssuesHistory,
                                   new Object[]{"1days",     1,
                                                "3days",     3,
                                                "7days",     7,
                                                "14days",   14,
                                                "30days",   30,
                                                "60days",   60,
                                                "90days",   90,
                                                "unlimited", 0
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetMaxIssuesHistory,new Integer(Settings.maxIssuesHistory));
        Widgets.layout(widgetMaxIssuesHistory,0,1,TableLayoutData.W);
        widgetMaxIssuesHistory.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Combo widget = (Combo)selectionEvent.widget;

            int n = Widgets.getSelectedOptionMenuItem(widget,Settings.maxIssuesHistory);
            if (n == 0)
            {
              if (!Dialogs.confirm(dialog,
                                   Dialogs.booleanFieldUpdater(Settings.class,"showUpdateIssuesWarning"),
                                   "Update full issues history may take a long time and cause a high server load.\nReally update full history?"
                                  )
                 )
              {
                Widgets.setSelectedOptionMenuItem(widget,0);
                return;
              }
            }
            data.maxIssuesHistory = n;
            Widgets.modified(data);
          }
        });
        Widgets.addModifyListener(new WidgetModifyListener(widgetMaxIssuesHistory,data)
        {
          @Override
          public void modified(Combo combo)
          {
            if (!combo.isDisposed())
            {
              combo.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL) && data.refreshTypes.contains(Settings.RefreshTypes.ISSUES));
            }
          }
        });
      }

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,10,0,TableLayoutData.WE);
      {
        widgetTimeEntries = Widgets.newCheckbox(subComposite,"Time entries");
        widgetTimeEntries.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.TIME_ENTRIES));
        Widgets.layout(widgetTimeEntries,0,0,TableLayoutData.W);
        widgetTimeEntries.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button widget = (Button)selectionEvent.widget;
            if (widget.getSelection())
            {
              data.refreshTypes.add(Settings.RefreshTypes.TIME_ENTRIES);
            }
            else
            {
              data.refreshTypes.remove(Settings.RefreshTypes.TIME_ENTRIES);
            }
            Widgets.modified(data);
          }
        });
        Widgets.addModifyListener(new WidgetModifyListener(widgetTimeEntries,data)
        {
          @Override
          public void modified(Button button)
          {
            if (!button.isDisposed())
            {
              button.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL));
            }
          }
        });

        widgetMaxTimeEntriesHistory = Widgets.newOptionMenu(subComposite);
        widgetMaxTimeEntriesHistory.setToolTipText("Number of days to refresh time entries history.");
        Widgets.setOptionMenuItems(widgetMaxTimeEntriesHistory,
                                   new Object[]{"1days",     1,
                                                "3days",     3,
                                                "7days",     7,
                                                "14days",   14,
                                                "30days",   30,
                                                "60days",   60,
                                                "90days",   90,
                                                "unlimited", 0
                                               }
                                  );
        Widgets.setSelectedOptionMenuItem(widgetMaxTimeEntriesHistory,new Integer(Settings.maxTimeEntriesHistory));
        Widgets.layout(widgetMaxTimeEntriesHistory,0,1,TableLayoutData.W);
        widgetMaxTimeEntriesHistory.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Combo widget = (Combo)selectionEvent.widget;

            int n = Widgets.getSelectedOptionMenuItem(widget,Settings.maxTimeEntriesHistory);
            if (n == 0)
            {
              if (!Dialogs.confirm(dialog,
                                   Dialogs.booleanFieldUpdater(Settings.class,"showUpdateTimeEntriesWarning"),
                                   "Update full time entry history may take a long time and cause a high server load.\nReally update full history?"
                                  )
                 )
              {
                Widgets.setSelectedOptionMenuItem(widget,0);
                return;
              }
            }
            data.maxTimeEntriesHistory = n;
            Widgets.modified(data);
          }
        });
        Widgets.addModifyListener(new WidgetModifyListener(widgetMaxTimeEntriesHistory,data)
        {
          @Override
          public void modified(Combo combo)
          {
            if (!combo.isDisposed())
            {
              combo.setEnabled(!data.runFlag && !data.refreshTypes.contains(Settings.RefreshTypes.ALL) && data.refreshTypes.contains(Settings.RefreshTypes.TIME_ENTRIES));
            }
          }
        });
      }

      widgetAll = Widgets.newCheckbox(composite,"All");
      widgetAll.setSelection(data.refreshTypes.contains(Settings.RefreshTypes.ALL));
      Widgets.layout(widgetAll,11,0,TableLayoutData.W);
      widgetAll.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;
          if (widget.getSelection())
          {
            data.refreshTypes.add(Settings.RefreshTypes.ALL);
          }
          else
          {
            data.refreshTypes.remove(Settings.RefreshTypes.ALL);
          }
          Widgets.modified(data);
        }
      });
      Widgets.addModifyListener(new WidgetModifyListener(widgetAll,data)
      {
        @Override
        public void modified(Button button)
        {
          if (!button.isDisposed())
          {
            button.setEnabled(!data.runFlag);
          }
        }
      });

      widgetProgressBar = Widgets.newProgressBar(composite,0.0,100.0);
      Widgets.layout(widgetProgressBar,12,0,TableLayoutData.WE,0,0,0,0,150,SWT.DEFAULT);

      widgetInfo = Widgets.newStringView(composite);
      Widgets.layout(widgetInfo,16,0,TableLayoutData.WE);

      widgetAutoClose = Widgets.newCheckbox(composite,"Auto close");
      widgetAutoClose.setSelection(Settings.refreshAutoCloseFlag);
      Widgets.layout(widgetAutoClose,17,0,TableLayoutData.W);
    }

    final BackendProgress backendProgress = new BackendProgress()
                                            {
                                              @Override
                                              protected void update(final double value)
                                              {
                                                if (!display.isDisposed())
                                                {
                                                  display.syncExec(new Runnable()
                                                  {
                                                    public void run()
                                                    {
                                                      if (!widgetProgressBar.isDisposed())
                                                      {
                                                        widgetProgressBar.setSelection(value*100.0);
                                                      }
                                                    }
                                                  });
                                                }
                                              }
                                            };

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetStartStop = Widgets.newButton(composite,"Start");
      Widgets.layout(widgetStartStop,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);

      button = Widgets.newButton(composite,"Close");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          synchronized(data)
          {
            if (data.runFlag)
            {
              if (!Dialogs.confirm(dialog,"Abort running refresh?"))
              {
                return;
              }
            }
            Settings.refreshAutoCloseFlag = widgetAutoClose.getSelection();
            Dialogs.close(dialog,false);
          }
        }
      });
    }

    // add listeners
    widgetStartStop.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        if (Backend.isOnline())
        {
          synchronized(data)
          {
            if (!data.runFlag)
            {
              // store selected refresh types
              Settings.refreshTypes = data.refreshTypes;

              // run background refresh thread
              Background.run(new BackgroundRunnable(dialog,data.refreshTypes)
              {
                public void run(final Shell dialog, EnumSet<Settings.RefreshTypes> refreshTypes)
                  throws InterruptedException
                {
                  final StatusText statusText = new StatusText();

                  statusText.set("Refresh local database\u2026");
                  {
                    Backend.updateLock();
                    printDebug("Refresh local database");
                    setStatus(statusText);
                    synchronizeAnimation.start();
                  }
                  try
                  {
                    // start
                    synchronized(data)
                    {
                      data.runFlag = true;

                      if (!display.isDisposed())
                      {
                        display.syncExec(new Runnable()
                        {
                          public void run()
                          {
                            // reset to stop
                            if (!widgetProgressBar.isDisposed())
                            {
                              widgetProgressBar.setSelection(0);
                            }
                            if (!widgetInfo.isDisposed())
                            {
                              widgetInfo.setText("");
                            }
                            if (!widgetStartStop.isDisposed())
                            {
                              widgetStartStop.setText("Stop");
                            }
                            Widgets.modified(data);
                          }
                        });
                      }
                    }

                    // get refresh types
                    if (refreshTypes.contains(Settings.RefreshTypes.ALL))
                    {
                      refreshTypes = EnumSet.allOf(Settings.RefreshTypes.class);
                    }
//Dprintf.dprintf("--------------------------------------------------------------");
//Dprintf.dprintf("refreshTypes=%s",refreshTypes);

                    if (!refreshTypes.isEmpty() )
                    {
                      // run refresh
                      try
                      {
                        // progress handler
                        backendProgress.setStepCount(refreshTypes.size());

                        // run refresh steps
                        if (!display.isDisposed())
                        {
                          for (final Settings.RefreshTypes refreshType : refreshTypes)
                          {
                            // check aborted
                            if (data.abortFlag || dialog.isDisposed())
                            {
                              break;
                            }

//Dprintf.dprintf("refreshType=%s",refreshType);
                            // update info
                            display.syncExec(new Runnable()
                            {
                              public void run()
                              {
                                if (!widgetInfo.isDisposed())
                                {
                                  widgetInfo.setText(String.format("Refresh '%s'\u2026",refreshType.toString()));
                                }
                              }
                            });

                            try
                            {
Dprintf.dprintf("refresty=%s\n",refreshType);
                              switch (refreshType)
                              {
                                case USERS:
                                  Backend.updateLocalDatabaseUsers(backendProgress);
                                  break;
                                case ROLES:
                                  Backend.updateLocalDatabaseRoles(backendProgress);
                                  break;
                                case TRACKERS:
                                  backendProgress.beginSteps(2);
                                  {
                                    backendProgress.setStep(0);
                                    Backend.updateLocalDatabaseTrackers(backendProgress);
                                    backendProgress.setStep(1);
                                    Backend.updateLocalDatabaseEpicNames(backendProgress);
                                  }
                                  backendProgress.endSteps();
                                  break;
                                case STATUS:
                                  Backend.updateLocalDatabaseStatus(backendProgress);
                                  break;
                                case PRIORITIES:
                                  Backend.updateLocalDatabasePriorities(backendProgress);
                                  break;
                                case ACTIVITIES:
                                  Backend.updateLocalDatabaseActivities(backendProgress);
                                  break;
                                case PROJECTS:
                                  Backend.updateLocalDatabaseProjects(backendProgress);
                                  break;
                                case VERSIONS:
                                  Backend.updateLocalDatabaseVersions(backendProgress);
                                  break;
                                case MEMBERS:
                                  Backend.updateLocalDatabaseMembers(backendProgress);
                                  break;
                                case ISSUES:
                                  updateLocalDatabaseIssues(refreshTypes.contains(Settings.RefreshTypes.ALL)
                                                              ? Settings.maxIssuesHistory
                                                              : data.maxIssuesHistory,
                                                            backendProgress
                                                           );
                                  break;
                                case TIME_ENTRIES:
Dprintf.dprintf("data.maxTimeEntriesHistory=%d",data.maxTimeEntriesHistory);
                                  updateLocalDatabaseTimeEntries(refreshTypes.contains(Settings.RefreshTypes.ALL)
                                                                   ? Settings.maxTimeEntriesHistory
                                                                   : data.maxTimeEntriesHistory,
                                                                 backendProgress
                                                                );
                                  break;
                                default:
                                  break;
                              }
Dprintf.dprintf("refresty=%s done\n",refreshType);
                            }
                            catch (final BackendException exception)
                            {
                              printStackTrace(exception);
                              display.syncExec(new Runnable()
                              {
                                public void run()
                                {
                                  Dialogs.error(shell,"Cannot update local database (error: "+exception.getMessage()+")");
                                }
                              });
                              break;
                            }

                            backendProgress.incrementStep();
                          }
                        }
                      }
                      catch (Throwable throwable)
                      {
                        Timemine.printInternalError(throwable);
                        printStackTrace(throwable);
                        System.exit(EXITCODE_INTERNAL_ERROR);
                      }
                    }

                    // done refresh
                    if (!display.isDisposed())
                    {
                      display.syncExec(new Runnable()
                      {
                        public void run()
                        {
                          if (!dialog.isDisposed())
                          {
                            // set text
                            widgetInfo.setText("OK");
                            widgetStartStop.setText("Start");

                            // auto-close if requested
                            Settings.refreshAutoCloseFlag = widgetAutoClose.getSelection();
                            if (Settings.refreshAutoCloseFlag)
                            {
                              Dialogs.close(dialog,false);
                            }
                          }
                        }
                      });
                    }

                    // stop
                    synchronized(data)
                    {
                      data.runFlag   = false;
                      data.abortFlag = false;
                      if (!display.isDisposed())
                      {
                        display.syncExec(new Runnable()
                        {
                          public void run()
                          {
                            // reset
                            if (!widgetInfo.isDisposed())
                            {
                              widgetInfo.setText("");
                            }
                            if (!widgetStartStop.isDisposed())
                            {
                              widgetStartStop.setText("Start");
                              widgetStartStop.setEnabled(true);
                            }
                            Widgets.modified(data);
                          }
                        });
                      }
                    }
                  }
                  finally
                  {
                    synchronizeAnimation.stop();
                    clearStatus();
                    Backend.updateUnlock();
                  }
                }
              });
            }
            else
            {
              // abort
              if (Dialogs.confirm(dialog,"Abort running refresh?"))
              {
                // stop background thread
                if (!widgetStartStop.isDisposed())
                {
                  widgetStartStop.setEnabled(false);
                }
                backendProgress.abort();
                data.abortFlag = true;
                Widgets.modified(data);
              }
            }
          }
        }
        else
        {
          Dialogs.error(dialog,"Offline mode - cannot refresh database!");
        }
      }
    });
    Widgets.addModifyListener(new WidgetModifyListener(widgetStartStop,data)
    {
      @Override
      public void modified(Button button)
      {
        if (!button.isDisposed())
        {
          button.setEnabled(!data.abortFlag && !data.refreshTypes.isEmpty());
        }
      }
    });

    // set next focus
    Widgets.setNextFocus(widgetUsers,
                         widgetRoles,
                         widgetTracker,
                         widgetStatus,
                         widgetPriorities,
                         widgetActivities,
                         widgetProject,
                         widgetMembers,
                         widgetVersions,
                         widgetIssues,
                         widgetTimeEntries,
                         widgetAll,
                         widgetStartStop
                        );

    // show dialog
    Dialogs.show(dialog);

    if (immediateFlag)
    {
      data.refreshTypes.add(Settings.RefreshTypes.ALL);
      Widgets.modified(data);
      Widgets.invoke(widgetStartStop);
    }

    // run
    Widgets.setFocus(widgetStartStop);
    Dialogs.run(dialog);
  }

  /** refresh local database
   * @param refreshTypes pre-selected refresh types (can be null for default)
   */
  private void refreshDatabase(EnumSet<Settings.RefreshTypes> refreshTypes)
  {
    refreshDatabase(refreshTypes,false);
  }

  /** refresh local database
   */
  private void refreshDatabase()
  {
    refreshDatabase((EnumSet<Settings.RefreshTypes>)null);
  }

  /** clean-up local database
   */
  private void cleanUpLocalDatabase()
  {
    // clean-up time entries
    try
    {
      for (TimeEntry timeEntry : Backend.getTimeEntryList())
      {
        try
        {
          if (timeEntry.projectId.isNone())
          {
            if (Dialogs.confirm(shell,String.format("No project for time entry:\n"+
                                                    "\n"+
                                                    "  %s\n"+
                                                    "  %s, %02dh:%02dmin\n"+
                                                    "\n"+
                                                    "Delete local time entry?",
                                                    timeEntry.comments,
                                                    timeEntry.spentOn,
                                                    (int)(timeEntry.duration/60),(int)(timeEntry.duration%60)
                                                   )
                               )
               )
            {
              removeTimeEntry(timeEntry);
            }
          }
          if (timeEntry.issueId.isNone())
          {
            if (Dialogs.confirm(shell,String.format("No issue for time entry:\n"+
                                                    "\n"+
                                                    "  %s\n"+
                                                    "  %s, %02dh:%02dmin\n"+
                                                    "\n"+
                                                    "Delete local time entry?",
                                                    timeEntry.comments,
                                                    timeEntry.spentOn,
                                                    (int)(timeEntry.duration/60),(int)(timeEntry.duration%60)
                                                   )
                               )
               )
            {
              removeTimeEntry(timeEntry);
            }
          }
        }
        catch (BackendException exception)
        {
          // ignored
        }
      }
    }
    catch (BackendException exception)
    {
      // ignored
    }

    // clean-up issues
    try
    {
      for (Issue issue : Backend.getIssueList())
      {
        try
        {
          if (issue.projectId.isNone())
          {
            if (Dialogs.confirm(shell,String.format("No project for issue:\n"+
                                                    "\n"+
                                                    "  %s\n"+
                                                    "\n"+
                                                    "Delete local issue?",
                                                    issue.description
                                                   )
                               )
               )
            {
              Backend.removeLocalDatabase("issues",
                                          issue.id.local
                                         );
              Widgets.notify(shell,USER_EVENT_REMOVE_ISSUE,issue);
            }
          }
        }
        catch (BackendException exception)
        {
          // ignored
        }
      }
    }
    catch (BackendException exception)
    {
      // ignored
    }
  }

  /** re-init local database
   */
  private void reinitDatabase()
  {
    if (Dialogs.confirm(shell,"Confirmation","Really delete and re-init local database?\n\nNote: all local data will be lost!"))
    {
      printDebug("Reinit local database");
      try
      {
        Backend.reinitLocalDatabase();
        refreshDatabase(EnumSet.of(Settings.RefreshTypes.ALL),true);
      }
      catch (IOException exception)
      {
        Dialogs.error(shell,"Cannot re-init local database (error: "+exception.getMessage()+")");
      }
      catch (SQLException exception)
      {
        Dialogs.error(shell,"Cannot re-init local database (error: "+exception.getMessage()+")");
      }
      catch (InterruptedException exception)
      {
        // ignored
      }
    }
  }

  /** add issue child Summaries to tree item
   * @param widgetProjectMenu menu
   * @param widgetProjectMenuSelectionListener selection listenre
   * @param projectsList project list
   * @param parentProjectsIdSet ids of projects with sub-projects
   * @param parentId parent id
   */
 private void addProjectSubMenu(Menu widgetProjectMenu, SelectionListener widgetProjectMenuSelectionListener, ProjectList projectList, HashSet<Id> parentProjectsIdSet, Id parentId)
  {
    for (Project project : projectList)
    {
      if (project.parentId == parentId)
      {
        Menu     subMenu;
        MenuItem menuItem;

        assert(!project.id.isNone() || project.name.isEmpty());

        if (parentProjectsIdSet.contains(project.id))
        {
          // project with sub-projects
          subMenu = Widgets.addMenu(widgetProjectMenu,project.name.replaceAll("&","&&"));
          menuItem = Widgets.addMenuItem(subMenu,project.name.replaceAll("&","&&"));
          menuItem.setData(project);
          menuItem.addSelectionListener(widgetProjectMenuSelectionListener);
          addProjectSubMenu(subMenu,widgetProjectMenuSelectionListener,projectList,parentProjectsIdSet,project.id);
        }
        else
        {
          // simple project
          menuItem = Widgets.addMenuItem(widgetProjectMenu,project.name.replaceAll("&","&&"));
          menuItem.setData(project);
          menuItem.addSelectionListener(widgetProjectMenuSelectionListener);
        }
      }
    }
  }

  /** sort bookmarks
   * @param bookmarkList bookmark list
   */
  private void sortBookmarks(ArrayList<Settings.Bookmark> bookmarkList)
  {
    // sort
    Collections.sort(bookmarkList,new Comparator<Settings.Bookmark>()
    {
      public int compare(Settings.Bookmark bookmark0, Settings.Bookmark bookmark1)
      {
        int result = 0;

        if ((bookmark0 != null) && (bookmark1 != null))
        {
          try
          {
            Issue   issue0   = Backend.getIssue(new Backend.TypeSet(bookmark0.backendType),bookmark0.summary);
            Project project0 = (issue0 != null) ? issue0.getProject() : null;
            Issue   issue1   = Backend.getIssue(new Backend.TypeSet(bookmark1.backendType),bookmark1.summary);
            Project project1 = (issue1 != null) ? issue1.getProject() : null;

            if ((project0 != null) && (project1 != null))
            {
              result = project0.name.compareTo(project1.name);
              if (result == 0)
              {
                result = bookmark0.text.compareTo(bookmark1.text);
              }
            }
            else
            {
              result = bookmark0.text.compareTo(bookmark1.text);
            }
          }
          catch (BackendException exception)
          {
            // ignored
          }
        }
        else if (bookmark0 != null)
        {
          result = -1;
        }
        else // bookmark1 != null
        {
          result = 1;
        }

        return result;
      }
    });
  }

  /** update bookmarks menu
   * @param widgetBookmarksMenu menu
   * @param projectsList project list
   * @param bookmarks bookmarks
   */
  private void updateBookmarksMenu(Menu widgetBookmarksMenu, ArrayList<Settings.Bookmark> bookmarkList)
  {
    Menu     subMenu;
    MenuItem menuItem,subMenuItem;

    // sort
    sortBookmarks(bookmarkList);

    // update menu items
    Widgets.removeMenuItems(widgetBookmarksMenu,2,-1);
    subMenu = null;
    for (Settings.Bookmark bookmark : bookmarkList)
    {
      try
      {
        Issue   issue   = Backend.getIssue(new Backend.TypeSet(bookmark.backendType),bookmark.summary);
        Project project = (issue != null) ? issue.getProject() : null;

        // add menu
        if (   ((subMenu == null) && (project != null))
            || ((subMenu != null) && (project != null) && !project.name.equals((String)subMenu.getData()))
           )
        {
          subMenu = Widgets.addMenu(widgetBookmarksMenu,project.name.replaceAll("&","&&"));
          subMenu.setData(project.name);
        }

        if (subMenu != null)
        {
          // add menu item
          subMenuItem = Widgets.addMenuItem(subMenu,bookmark.text.replaceAll("&","&&"));
          subMenuItem.setData(bookmark);
          subMenuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              MenuItem          widget   = (MenuItem)selectionEvent.widget;
              Settings.Bookmark bookmark = (Settings.Bookmark)widget.getData();
              try
              {
                Issue issue = Backend.getIssue(new Backend.TypeSet(bookmark.backendType),bookmark.summary);
                if (issue != Issue.NONE)
                {
                  Project project = issue.getProject();

                  setSelectedProject(addTimeEntry1.widgetProject,
                                     addTimeEntry1.widgetIssueId,
                                     addTimeEntry1.widgetIssueShortName,
                                     addTimeEntry1.widgetIssueSummary,
                                     addTimeEntry1.widgetIssueDoneRatio,
                                     addTimeEntry1.widgetActivitiesTitle,
                                     addTimeEntry1.widgetActivities,
                                     project
                                    );
                  setSelectedIssue(addTimeEntry1.widgetIssueId,
                                   addTimeEntry1.widgetIssueShortName,
                                   addTimeEntry1.widgetIssueSummary,
                                   addTimeEntry1.widgetIssueDoneRatio,
                                   issue
                                  );

                  setSelectedProject(addTimeEntry2.widgetProject,
                                     addTimeEntry2.widgetIssueId,
                                     addTimeEntry2.widgetIssueShortName,
                                     addTimeEntry2.widgetIssueSummary,
                                     addTimeEntry2.widgetIssueDoneRatio,
                                     addTimeEntry2.widgetActivitiesTitle,
                                     addTimeEntry2.widgetActivities,
                                     project
                                    );
                  setSelectedIssue(addTimeEntry2.widgetIssueId,
                                   addTimeEntry2.widgetIssueShortName,
                                   addTimeEntry2.widgetIssueSummary,
                                   addTimeEntry2.widgetIssueDoneRatio,
                                   issue
                                  );
                }
                else
                {
                  Dialogs.error(shell,"Issue %s: %s not found",bookmark.backendType,bookmark.summary);
                  return;
                }
              }
              catch (BackendException exception)
              {
                printStackTrace(exception);
                Dialogs.error(shell,"Cannot select issue (error: "+exception.getMessage()+")");
                return;
              }
            }
          });
        }
      }
      catch (BackendException exception)
      {
        // ignored
      }
    }
  }

  /** update projects favorite menu
   * @param widgetProjectMenu menu
   * @param widgetProjectMenuSelectionListener selection listenre
   * @param projectsList project list
   */
  private void updateProjectsMenu(Menu              widgetProjectMenu,
                                  SelectionListener widgetProjectMenuSelectionListener,
                                  ProjectList       projectList
                                 )
  {
    // find projects which have sub-projects
    HashSet<Id> parentProjectsIdSet = new HashSet<Id>();
    for (Project project : projectList)
    {
      if (!project.parentId.isNone())
      {
        parentProjectsIdSet.add(project.parentId);
      }
    }
//for (Project p : projectList) Dprintf.dprintf("p=%s",p.name); System.exit(1);

    // group projects
    LinkedHashMap<String,ProjectList> projectMap = new LinkedHashMap<String,ProjectList>();
    String      prevNamePrefix = "";
    ProjectList subProjectList = new ProjectList();
    for (int i = 0; i < projectList.size(); i++)
    {
      Project project = projectList.get(i);

      // get name prefix
      int j = 0;
      while (   (j < project.name.length())
             && (   Character.isLetterOrDigit(project.name.charAt(j))
                 || Character.isSpace(project.name.charAt(j))
                )
            )
      {
        j++;
      }
      String namePrefix = project.name.substring(0,j).trim();

      subProjectList = projectMap.get(namePrefix);
      if (subProjectList == null)
      {
        subProjectList = new ProjectList();
        projectMap.put(namePrefix,subProjectList);
      }
      subProjectList.add(project);

      prevNamePrefix = namePrefix;
    }

    // create menu
    Widgets.removeMenuItems(widgetProjectMenu,0,-1);
    for (String namePrefix : projectMap.keySet())
    {
      Menu subMenu = Widgets.addMenu(widgetProjectMenu,namePrefix.replaceAll("&","&&"));

      for (Project project : projectMap.get(namePrefix))
      {
//Dprintf.dprintf("project=%s",project);
        Menu     subSubMenu;
        MenuItem menuItem;

        if (project.parentId.isNone())
        {
          if (parentProjectsIdSet.contains(project.id))
          {
            // project with sub-projects
            subSubMenu = Widgets.addMenu(subMenu,project.name.replaceAll("&","&&"));
            menuItem = Widgets.addMenuItem(subSubMenu,project.name.replaceAll("&","&&"));
            addProjectSubMenu(subMenu,widgetProjectMenuSelectionListener,projectList,parentProjectsIdSet,project.id);
          }
          else
          {
            // simple project
            menuItem = Widgets.addMenuItem(subMenu,project.name.replaceAll("&","&&"));
          }
          menuItem.setData(project);
          menuItem.setImage(project.backendType.getImage());
          menuItem.addSelectionListener(widgetProjectMenuSelectionListener);
        }
      }
    }
  }

  /** update history combo box
   * @param widgetHistory widget
   * @param historyTimeEntryList history time entry list
   */
  private void updateHistory(Combo widgetHistory, TimeEntryList historyTimeEntryList)
  {
    String text = widgetHistory.getText();
    widgetHistory.removeAll();
    for (TimeEntry historyTimeEntry : historyTimeEntryList)
    {
      widgetHistory.add(historyTimeEntry.comments);
    }
    widgetHistory.setText(text);
  }

  /** update history table combo box
   * @param widgetHistory widget
   * @param historyTimeEntryList history time entry list
   */
  private void updateHistory(TableCombo widgetHistory, TimeEntryList historyTimeEntryList)
  {
    Comparator<TimeEntry> comparator = new Comparator<TimeEntry>()
    {
       public int compare(TimeEntry timeEntry0, TimeEntry timeEntry1)
       {
         if (timeEntry0.id.equals(timeEntry1.id))
         {
           return 0;
         }
         else
         {
           return 1;
         }
       }
    };

    for (TimeEntry historyTimeEntry : historyTimeEntryList)
    {
      Project project = historyTimeEntry.getProject();
      Issue   issue   = historyTimeEntry.getIssue  ();

      if ((project != null) && (issue != null))
      {
        widgetHistory.updateInsertTableItem(comparator,
                                            historyTimeEntry,
                                            project.name,
                                            issue.shortName,
                                            issue.summary,
                                            String.format("%02d:%02d",historyTimeEntry.duration/60,historyTimeEntry.duration%60)
                                          );
      }

      while (widgetHistory.getItemCount() > MAX_HISTORY_SIZE)
      {
        widgetHistory.removeTableItem(MAX_HISTORY_SIZE);
      }
    }
    widgetHistory.setText("History");
  }

  /** add issue child Summaries to tree item
   * @param widgetProject project combo tree
   * @param issues project array
   * @param parentId parent id
   * @param treeItem tree item
   */
  private static void addProjectSubTree(TreeCombo widgetProject, final ProjectList projectList, Id parentId, TreeCombo.Item parentTreeComboItem)
  {
    for (Project project : projectList)
    {
      if (project.parentId == parentId)
      {
        assert(!project.id.isNone() || project.name.isEmpty());
        TreeCombo.Item treeComboItem = widgetProject.addTreeItem(project,
                                                                  parentTreeComboItem,
                                                                  project.name
                                                                 );
        addProjectSubTree(widgetProject,projectList,project.id,treeComboItem);
      }
    }
  }

  /** update projects tree combo box
   * @param widgetProject widget
   * @param projectList project data list
   * @param selectedProject selected project or null
   * @param addNone true to add none-project
   */
  private static void updateProjectTree(TreeCombo widgetProject, final ProjectList projectList, Project selectedProject, boolean addNone)
  {
    widgetProject.removeAllTreeItems();
    if (addNone)
    {
      widgetProject.addTreeItem((Project)null,"");
    }
    for (Project project : projectList)
    {
      // check if parent of project exists
      boolean parentProjectExists = false;
      for (int j = 0; j < projectList.size(); j++)
      {
        if (project.parentId == projectList.get(j).id)
        {
          parentProjectExists = true;
          break;
        }
      }

      // add root project and all sub-projects if no parent (=real root element) or parent project is not in list
      if (   project.parentId.isNone()
          || !parentProjectExists
         )
      {
        // add root item
        assert(!project.id.isNone() || project.name.isEmpty());
        TreeCombo.Item treeComboItem = widgetProject.addTreeItem(project,
                                                                  project.backendType.getImage(),
                                                                  project.name
                                                                 );

        // add sub-items
        addProjectSubTree(widgetProject,projectList,project.id,treeComboItem);
      }
    }

    // select project
    if (selectedProject != null)
    {
      widgetProject.setSelection(selectedProject,selectedProject.name);
    }
    else
    {
      widgetProject.setSelection((Project)null,"");
    }
  }

  /** update projects tree combo box
   * @param widgetProject widget
   * @param projectList project data list
   * @param selectedProject selected project or null
   * @param addNone true to add none-project
   */
  public static void updateProjects(TreeCombo   widgetProject,
                                    ProjectList projectList,
                                    Project     selectedProject,
                                    boolean     addNone
                                   )
  {
    updateProjectTree(widgetProject,projectList,selectedProject,addNone);
  }

  /** update projects tree combo box
   * @param widgetProject widget
   * @param selectedProjectId selected project or null
   */
  public static void updateProjects(TreeCombo widgetProject,
                                    Project   selectedProject,
                                    boolean   addNone
                                   )
  {
    try
    {
      // get issues for project and all sub-projects
      ProjectList projectList = Backend.getProjectList();

      // update
      updateProjectTree(widgetProject,projectList,selectedProject,addNone);
    }
    catch (BackendException exception)
    {
      printStackTrace(exception);
      String message = exception.getMessage();
      if (message == null) message = exception.getCause().getMessage();
      Dialogs.error(shell,"Cannot get issue data (error: "+message+")");
      return;
    }
  }

  /** update projects tree combo box
   * @param widgetProject widget
   * @param projectList project data list
   * @param selectedProjectId selected project or null
   */
  public static void updateProjects(TreeCombo   widgetProject,
                                    ProjectList projectList,
                                    Project     selectedProject
                                   )
  {
    updateProjectTree(widgetProject,projectList,selectedProject,false);
  }

  /** update projects tree combo box
   * @param widgetProject widget
   * @param projectList project data list
   * @param selectedProjectName selected project name
   * @param addNone true to add none-project
   */
  public static void updateProjects(TreeCombo         widgetProject,
                                    final ProjectList projectList,
                                    String            selectedProjectName,
                                    boolean           addNone
                                   )
  {
    Project selectedProject = null;
    for (Project project : projectList)
    {
      if (project.name.equals(selectedProjectName))
      {
        selectedProject = project;
        break;
      }
    }

    updateProjects(widgetProject,projectList,selectedProject,addNone);
  }

  /** update projects tree combo box
   * @param widgetProject widget
   * @param projectList project data list
   * @param selectedProjectName selected project name
   */
  public static void updateProjects(TreeCombo         widgetProject,
                                    final ProjectList projectList,
                                    String            selectedProjectName
                                   )
  {
    updateProjects(widgetProject,projectList,selectedProjectName,false);
  }

  /** update projects tree combo box
   * @param widgetProject widget
   * @param projectList project data list
   */
  public static void updateProjects(TreeCombo   widgetProject,
                                    ProjectList projectList
                                   )
  {
    Project selectedProject = widgetProject.getSelection();
    updateProjects(widgetProject,projectList,selectedProject);
  }

  /** reset project combo boxes
   * @param widgetProject project combo widget
   */
  private void resetProjects(TreeCombo widgetProject)
  {
    widgetProject.removeAllTreeItems();
  }
  /** update issues ids combo box
   * @param widgetIssueId combo box widget
   * @param issueList issue data list
   * @param selectedIssue selected issue or null
   * @param addAny true to add any-id
   * @param addNone true to add none-id
   */
  private static void updateIssueIds(TableCombo widgetIssueId,
                                     IssueList  issueList,
                                     Issue      selectedIssue,
                                     boolean    addAny,
                                     boolean    addNone
                                    )
  {
    // sort issues by id
    Collections.sort(issueList,new Comparator<Issue>()
    {
      public int compare(Issue issue0, Issue issue1)
      {
        assert issue0 != null;
        assert issue1 != null;

        if      (issue0.id.backend < issue1.id.backend) return -1;
        else if (issue0.id.backend > issue1.id.backend) return  1;
        else                                            return  0;
      }
    });

    if (widgetIssueId != null)
    {
      // update id combo box
      widgetIssueId.removeAllTableItems();
      if (addAny)
      {
        widgetIssueId.addTableItem(Issue.ANY,"*");
      }
      if (addNone)
      {
        widgetIssueId.addTableItem(Issue.NONE,"");
      }
      for (Issue issue : issueList)
      {
//Dprintf.dprintf("issue=%s -> %s",issue,issue.id);
        widgetIssueId.addTableItem(issue,
                                    !issue.id.isNone() ? Integer.toString(issue.id.backend) : "-",
                                    issue.shortName
                                   );
      }

      // select selected issue
      if      ((selectedIssue != null) && (selectedIssue != Issue.NONE) && (selectedIssue.shortName != null))
      {
        widgetIssueId.setSelection(selectedIssue,"%s",selectedIssue.id.format());
      }
      else if (addNone)
      {
        widgetIssueId.setSelection(Issue.NONE,"");
      }
      else if (!addAny && !issueList.isEmpty())
      {
        widgetIssueId.setSelection(issueList.get(0),"%s",issueList.get(0).id.format());
      }
    }
  }

  /** update issues ids combo box
   * @param widgetIssueId id combo widget
   * @param issueList new issue data list
   * @param addAny true to add any-id
   * @param addNone true to add none-id
   */
  private static void updateIssueIds(TableCombo widgetIssueId,
                                     IssueList  issueList,
                                     boolean    addAny,
                                     boolean    addNone
                                    )
  {
    updateIssueIds(widgetIssueId,issueList,(Issue)null,addAny,addNone);
  }

  /** reset issues ids combo box
   * @param widgetIssueId issue id combo widget
   */
  private void resetIssueIds(TableCombo widgetIssueId)
  {
    widgetIssueId.removeAllTableItems();
  }

  /** update issues short names combo box
   * @param widgetIssueShortName short names widget
   * @param issueList issue data list
   * @param selectedIssue selected issue or null
   * @param addAny true to add any-id
   * @param addNone true to add none-id
   */
  private static void updateIssueShortNames(TableCombo widgetIssueShortName,
                                            IssueList  issueList,
                                            Issue      selectedIssue,
                                            boolean    addAny,
                                            boolean    addNone
                                           )
  {
    // sort filtered issues by id
    Collections.sort(issueList,new Comparator<Issue>()
    {
      public int compare(Issue issue0, Issue issue1)
      {
        assert issue0 != null;
        assert issue1 != null;

        if      (issue0.id.backend < issue1.id.backend) return -1;
        else if (issue0.id.backend > issue1.id.backend) return  1;
        else                                            return  0;
      }
    });

    if (widgetIssueShortName != null)
    {
      // update id combo box
      widgetIssueShortName.removeAllTableItems();
      if (addAny)
      {
        widgetIssueShortName.addTableItem(Issue.ANY,"*");
      }
      if (addNone)
      {
        widgetIssueShortName.addTableItem(Issue.NONE,"");
      }
      for (Issue issue : issueList)
      {
//Dprintf.dprintf("issue=%s -> %s",issue,issue.id);
        widgetIssueShortName.addTableItem(issue,
                                          (issue.shortName != null) ? issue.shortName : ""
                                         );
      }

      // select selected issue
      if      ((selectedIssue != null) && (selectedIssue != Issue.NONE) && (selectedIssue.shortName != null))
      {
        widgetIssueShortName.setSelection(selectedIssue,"%s",(selectedIssue.shortName != null) ? selectedIssue.shortName : "");
      }
      else if (addNone)
      {
        widgetIssueShortName.setSelection(Issue.NONE,"");
      }
      else if (!addAny && !issueList.isEmpty())
      {
        widgetIssueShortName.setSelection(issueList.get(0),"%s",issueList.get(0).shortName);
      }
    }
  }

  /** reset issues short names combo box
   * @param widgetIssueShortName issue short combo widget
   */
  private void resetIssueShortNames(TableCombo widgetIssueShortName)
  {
    widgetIssueShortName.removeAllTableItems();
  }

  /** add issue child summaries to tree item
   * @param widgetIssueSummary issue combo tree
   * @param issueList issue list
   * @param parentId parent id
   * @param treeItem tree item
   */
  private void addIssueSubTree(TreeCombo widgetIssueSummary, final IssueList issueList, Id parentId, TreeCombo.Item parentTreeComboItem)
  {
    for (Issue issue : issueList)
    {
      if (issue.parentId == parentId)
      {
        TreeCombo.Item treeComboItem = widgetIssueSummary.addTreeItem(issue,
                                                                        parentTreeComboItem,
                                                                        !issue.id.isNone() ? issue.id : "-",
                                                                        issue.shortName,
                                                                        issue.summary
                                                                       );
        if (!issue.id.isNone())
        {
          addIssueSubTree(widgetIssueSummary,issueList,issue.id,treeComboItem);
        }
      }
    }
  }

  /** update issue tree
   * @param widgetIssueSummary issue combo tree
   * @param issueList issue list
   * @param selectedIssue selected issue or null
   * @param addAny true to add any-id
   * @param addNone true to add none-id
   */
  private static void updateIssueTree(TreeCombo             widgetIssueSummary,
                                      AbstractMap<Id,Issue> issueIdMap,
                                      IssueList             issueList,
                                      Issue                 selectedIssue,
                                      boolean               addAny,
                                      boolean               addNone
                                     )
  {
    // get issue depth map
    final HashMap<Id,Integer> depthMap = new HashMap<Id,Integer>();
    for (Issue issue : issueList)
    {
      int depth    = 0;
      Id  parentId = issue.parentId;
      while (!parentId.isNone())
      {
        Issue parentIssue = issueIdMap.get(parentId);
        depth++;
        parentId = (parentIssue != null) ? parentIssue.parentId : new Id();
Dprintf.dprintf("depth=%d parentId=%s",depth,parentId);
      }
      depthMap.put(issue.id,depth);
    }
//Dprintf.dprintf("depthMap.size=%d",depthMap.size());

    // get hierachical sorted issue list
    IssueList sortedIssueList = (IssueList)issueList.clone();
    Collections.sort(sortedIssueList,new Comparator<Issue>()
                     {
                       public int compare(Issue issue0, Issue issue1)
                       {
                         int depth0 = depthMap.get(issue0.id);
                         int depth1 = depthMap.get(issue1.id);
                         if      (depth0 < depth1)
                         {
                           return -1;
                         }
                         else if (depth0 > depth1)
                         {
                           return +1;
                         }
                         else
                         {
                           return issue0.summary.compareTo(issue1.summary);
                         }
                       }
                     }
                    );

    // add tree items
    widgetIssueSummary.removeAllTreeItems();
    if (addAny)
    {
      widgetIssueSummary.addTreeItem(Issue.ANY,"*");
    }
    if (addNone)
    {
      widgetIssueSummary.addTreeItem(Issue.NONE,"");
    }
    HashMap<Id,TreeCombo.Item> parentItemMap = new HashMap<Id,TreeCombo.Item>();
    for (Issue issue : sortedIssueList)
    {
      TreeCombo.Item treeComboItem = widgetIssueSummary.addTreeItem(issue,
                                                                    parentItemMap.get(issue.parentId),
                                                                    !issue.id.isNone() ? Integer.toString(issue.id.backend) : "-",
                                                                    issue.shortName,
                                                                    issue.summary
                                                                   );
      if (!issue.id.isNone()) parentItemMap.put(issue.id,treeComboItem);
    }
    widgetIssueSummary.filterItems();

    // select issue
    if      ((selectedIssue != null) && (selectedIssue != Issue.NONE))
    {
      widgetIssueSummary.setSelection(selectedIssue,"%s",selectedIssue.summary);
    }
    else if (addNone)
    {
      widgetIssueSummary.setSelection(Issue.NONE,"");
    }
    else if (!addAny && !sortedIssueList.isEmpty())
    {
      widgetIssueSummary.setSelection(sortedIssueList.get(0),"%s",sortedIssueList.get(0).summary);
    }
    else
    {
      widgetIssueSummary.setSelection(Issue.NONE,"");
    }
  }

  /** update issue Summaries tree combo box
   * @param widgetIssueSummary summary tree combo widget
   * @param issues issue data array
   * @param selectedIssue selected issue or null
   * @param addAny true to add any-id
   * @param addNone true to add none-id
   */
  public static void updateIssueSummaries(TreeCombo widgetIssueSummary,
                                          IssueList issueList,
                                          Issue     selectedIssue,
                                          boolean   addAny,
                                          boolean   addNone
                                         )
  {
    // get issue id map
    HashMap<Id,Issue> issueIdMap = new HashMap<Id,Issue>();
    for (Issue issue : issueList)
    {
      issueIdMap.put(issue.id,issue);
    }

    // add all parent issues of issues
//Dprintf.dprintf("filteredIssueList.size=%d",filteredIssueList.size());
    Issue parentIssue;
    do
    {
      // find new parent issue
      parentIssue = null;
      for (Issue issue : issueList)
      {
        if (   !issue.parentId.isNone()
            && !issueIdMap.containsKey(issue.parentId)
           )
        {
          // find parent issue
          for (Issue findIssue : issueList)
          {
            if (issue.parentId.equals(findIssue.id))
            {
              parentIssue = findIssue;
              break;
            }
          }
          if (parentIssue != null)
          {
            break;
          }
        }
      }

      // add parent issue
      if (parentIssue != null)
      {
        issueList.add(parentIssue);
        issueIdMap.put(parentIssue.id,parentIssue);
      }
    }
    while (parentIssue != null);
//Dprintf.dprintf("issueList.size=%d",issueList.size());

    // sort filtered issues
    Collections.sort(issueList,new Comparator<Issue>()
    {
      public int compare(Issue issue0, Issue issue1)
      {
        assert issue0 != null;
        assert issue1 != null;

        return issue0.summary.compareTo(issue1.summary);
      }
    });
//for (Issue i : filteredIssueList) Dprintf.dprintf("filtered %s",i);

    // update issue tree
    updateIssueTree(widgetIssueSummary,issueIdMap,issueList,selectedIssue,addAny,addNone);
  }

  /** update issue Summaries tree combo box
   * @param widgetIssueSummary summary tree combo widget
   * @param issueList issue data list
   * @param selectedIssueId selected issue or null
   */
  public static void updateIssueSummaries(TreeCombo widgetIssueSummary,
                                          IssueList issueList,
                                          Issue     selectedIssue
                                         )
  {
    updateIssueSummaries(widgetIssueSummary,issueList,selectedIssue,false,false);
  }

  /** update issues Summaries tree combo box
   * @param widgetIssueSummary summary tree combo widget
   * @param issueList issue data list
   * @param addAny true to add any-id
   * @param addNone true to add none-id
   */
  public static void updateIssueSummaries(TreeCombo widgetIssueSummary,
                                          IssueList issueList,
                                          boolean   addAny,
                                          boolean   addNone
                                         )
  {
    updateIssueSummaries(widgetIssueSummary,issueList,(Issue)null,addAny,addNone);
  }

  /** reset issues Summaries tree combo box
   * @param widgetIssueSummary summary tree combo widget
   */
  private void resetIssueSummaries(TreeCombo widgetIssueSummary)
  {
    widgetIssueSummary.removeAllTreeItems();
  }

  /** update issues combo boxes
   * @param widgetIssueId id combo widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary summary tree combo widget
   * @param issueList issue data list
   * @param selectedIssue selected issue or null
   * @param addAny true to add any-id
   * @param addNone true to add none-id
   */
  public static void updateIssues(TableCombo widgetIssueId,
                                  TableCombo widgetIssueShortName,
                                  TreeCombo  widgetIssueSummary,
                                  IssueList  issueList,
                                  Issue      selectedIssue,
                                  boolean    addAny,
                                  boolean    addNone
                                 )
  {
    updateIssueIds(widgetIssueId,issueList,selectedIssue,addAny,addNone);
    updateIssueShortNames(widgetIssueShortName,issueList,selectedIssue,addAny,addNone);
//Dprintf.dprintf("%s %s %s %s %s",widgetIssueSummary,issueList,selectedIssue,addAny,addNone);
    updateIssueSummaries(widgetIssueSummary,issueList,selectedIssue,addAny,addNone);
  }

  /** update issues combo boxes
   * @param widgetIssueId id combo widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary summary tree combo widget
   * @param issueList issue data list
   * @param selectedIssueId selected issue or null
   */
  public static void updateIssues(TableCombo widgetIssueId,
                                  TableCombo widgetIssueShortName,
                                  TreeCombo  widgetIssueSummary,
                                  IssueList  issueList,
                                  Issue      selectedIssue
                                 )
  {
    updateIssues(widgetIssueId,widgetIssueShortName,widgetIssueSummary,issueList,selectedIssue,false,false);
  }


  /** update issues combo boxes
   * @param widgetIssueId id combo widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary summary combo widget
   * @param issueList issue data list
   */
  public static void updateIssues(TableCombo widgetIssueId,
                                  TableCombo widgetIssueShortName,
                                  TreeCombo  widgetIssueSummary,
                                  IssueList  issueList
                                 )
  {
    Issue selectedIssue = widgetIssueSummary.getSelection();
    updateIssues(widgetIssueId,widgetIssueShortName,widgetIssueSummary,issueList,selectedIssue);
  }

  /** update issues combo boxes
   * @param widgetIssueId id combo widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary summary combo widget
   * @param project project
   * @param filters filter strings or null
   * @param selectedIssue selected issue or null
   */
  public static void updateIssues(TableCombo widgetIssueId,
                                  TableCombo widgetIssueShortName,
                                  TreeCombo  widgetIssueSummary,
                                  Project    project,
                                  String     filters[],
                                  Issue      selectedIssue
                                 )
  {
    if (project != null)
    {
      try
      {
        int i;

        // get issues for project and all sub-projects
        IssueList issueList = project.getIssueList();

        // filter issue list
        if (Settings.issueOwnFilter)
        {
          i = 0;
          while (i < issueList.size())
          {
            Issue issue = issueList.get(i);

            if (issue.isAssignedToCurrentUser())
            {
              // match filter => keep issue
              i++;
            }
            else
            {
              // filter issue
              issueList.remove(i);
            }
          }
        }
        if (filters != null)
        {
          for (String filter : filters)
          {
            filter = filter.toLowerCase();

            i = 0;
            while (i < issueList.size())
            {
              Issue issue = issueList.get(i);

              if (issue.summary.toLowerCase().contains(filter))
              {
                // match filter => keep issue
                i++;
              }
              else
              {
                // filter issue
                issueList.remove(i);
              }
            }
          }
        }

        // update
        updateIssues(widgetIssueId,widgetIssueShortName,widgetIssueSummary,issueList,selectedIssue);
      }
      catch (BackendException exception)
      {
        printStackTrace(exception);
        String message = exception.getMessage();
        if (message == null) message = exception.getCause().getMessage();
        Dialogs.error(shell,"Cannot get issue data (error: "+message+")");
        return;
      }
    }
  }

  /** update issues combo boxes
   * @param widgetIssueId id combo widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary summary combo widget
   * @param project project
   * @param filters filter strings or null
   */
  public static void updateIssues(TableCombo widgetIssueId,
                                  TableCombo widgetIssueShortName,
                                  TreeCombo  widgetIssueSummary,
                                  Project    project,
                                  String     filters[]
                                 )
  {
    updateIssues(widgetIssueId,widgetIssueShortName,widgetIssueSummary,project,filters,(Issue)null);
  }

  /** update issues combo boxes
   * @param widgetIssueId id combo widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary summary combo widget
   * @param project project
   * @param selectedIssue selected issue or null
   */
  public static void updateIssues(TableCombo widgetIssueId,
                                  TableCombo widgetIssueShortName,
                                  TreeCombo  widgetIssueSummary,
                                  Project    project,
                                  Issue      selectedIssue
                                 )
  {
    updateIssues(widgetIssueId,widgetIssueShortName,widgetIssueSummary,project,(String[])null,selectedIssue);
  }

  /** update issues combo boxes
   * @param widgetIssueId id combo widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary summary combo widget
   * @param project project
   */
  public static void updateIssues(TableCombo widgetIssueId,
                                  TableCombo widgetIssueShortName,
                                  TreeCombo  widgetIssueSummary,
                                  Project    project
                                 )
  {
    updateIssues(widgetIssueId,widgetIssueShortName,widgetIssueSummary,project,(String[])null);
  }

  /** update issues combo boxes
   * @param widgetIssueSummary summary combo widget
   * @param project project
   * @param filters filter strings or null
   * @param selectedIssue selected issue or null
   */
  public static void updateIssues(TreeCombo widgetIssueSummary,
                                  Project   project,
                                  String    filters[],
                                  Issue     selectedIssue
                                 )
  {
    updateIssues((TableCombo)null,(TableCombo)null,widgetIssueSummary,project,filters,selectedIssue);
  }

  /** update issues combo boxes
   * @param widgetIssueSummary summary combo widget
   * @param project project
   * @param selectedIssue selected issue or null
   */
  public static void updateIssues(TreeCombo widgetIssueSummary,
                                  Project   project,
                                  Issue     selectedIssue
                                 )
  {
    updateIssues(widgetIssueSummary,project,(String[])null,selectedIssue);
  }

  /** update issues combo boxes
   * @param widgetIssueSummary summary combo widget
   * @param project project
   * @param filters filter strings or null
   */
  public static void updateIssues(TreeCombo widgetIssueSummary,
                                  Project   project,
                                  String    filters[]
                                 )
  {
    updateIssues((TableCombo)null,(TableCombo)null,widgetIssueSummary,project,filters);
  }

  /** reset issues combo boxes
   * @param widgetIssueId id combo widget
   * @param widgetIssueSummary summary combo widget
   * @param projectId project id
   */
  private void resetIssues(TableCombo widgetIssueId,
                           TableCombo widgetIssueShortName,
                           TreeCombo  widgetIssueSummary
                          )
  {
    resetIssueIds(widgetIssueId);
    resetIssueShortNames(widgetIssueShortName);
    resetIssueSummaries(widgetIssueSummary);
  }

  /** update issues done ratio combo box
   * @param widgetIssues combo widget
   * @param doneRatio issue done ratio (0..100)
   */
  private void updateIssueDoneRatio(Combo widgetIssueDoneRatio, int doneRatio)
  {
    Widgets.setSelectedComboItem(widgetIssueDoneRatio,(Integer)doneRatio);
  }

  /** reset issues done ratio combo box
   * @param issue issue
   * @param widgetIssues combo widget
   */
  private void resetIssueDoneRatio(Combo widgetIssueDoneRatio)
  {
    Widgets.setSelectedComboItem(widgetIssueDoneRatio,new Id());
  }

  /** update tracker combo box
   * @param widgetTracker issue tracker combo widget
   * @param trackerList tracker data list
   * @param selectedTracker selected tracker
   */
  private void updateTrackers(final Combo widgetTracker, TrackerList trackerList, Tracker selectedTracker)
  {
    Widgets.removeAllOptionMenuItems(widgetTracker);
    for (Tracker tracker : trackerList)
    {
      Widgets.addOptionMenuItem(widgetTracker,tracker,tracker.name);
    }

    // select tracker
    if (selectedTracker != null)
    {
      Widgets.setSelectedOptionMenuItem(widgetTracker,selectedTracker);
    }
  }

  /** update issue status transition menu
   * @param widgetIssueStatus status combo widget
   * @param statusTransitionList status transition data list
   * @param selectedStatus selected status
   * @param selectedStatusTransition selected status transition
   */
  private void updateIssueStatusTransition(final Combo          widgetIssueStatus,
                                           StatusTransitionList statusTransitionList,
                                           Status               selectedStatus,
                                           StatusTransition     selectedStatusTransition
                                          )
  {
    // clear status list
    Widgets.removeAllOptionMenuItems(widgetIssueStatus);
    int selectStatusIndex = -1;

    // add current status
    if (selectedStatus != null)
    {
      Widgets.addOptionMenuItem(widgetIssueStatus,StatusTransition.NONE,selectedStatus.name);
      selectStatusIndex = 0;
    }

    // add status transitions/current status, get selection index
    for (int i = 0; i < statusTransitionList.size(); i++)
    {
      StatusTransition statusTransition = statusTransitionList.get(i);

      // add status transition
      Widgets.addOptionMenuItem(widgetIssueStatus,statusTransition,statusTransition.name);

      // get selection index of pending status transition
      if (   (selectedStatusTransition != StatusTransition.NONE)
          && statusTransition.equals(selectedStatusTransition)
         )
      {
        selectStatusIndex = i;
      }
    }

    // select status transition
    if (selectStatusIndex >= 0)
    {
      Widgets.setSelectedOptionMenuItem(widgetIssueStatus,selectStatusIndex);
    }
  }

  /** update issue status menu
   * @param widgetIssuePriorities priority combo widget
   * @param priorityList new priority data list
   * @param selectedPriority selected priority
   */
  private void updateIssuePriorities(final Combo widgetIssuePriorities, PriorityList priorityList, Priority selectedPriority)
  {
    Widgets.removeAllOptionMenuItems(widgetIssuePriorities);
    for (Priority priority : priorityList)
    {
      Widgets.addOptionMenuItem(widgetIssuePriorities,priority,priority.name);
    }

    // select priority
    if (selectedPriority != null)
    {
      Widgets.setSelectedOptionMenuItem(widgetIssuePriorities,selectedPriority);
    }
  }

  /** update issue assign-to member menu
   * @param widgetIssueAssignedToMember member combo widget
   * @param memberList new member data list
   * @param selectedAssignedToMember selected assigned to user
   */
  private void updateIssueAssignedToMember(final Combo widgetIssueAssignedToMember, MemberList memberList, Member selectedAssignedToMember)
  {
    Widgets.removeAllOptionMenuItems(widgetIssueAssignedToMember);
    Widgets.addOptionMenuItem(widgetIssueAssignedToMember,Member.ANY,"");
    for (Member member : memberList)
    {
      Widgets.addOptionMenuItem(widgetIssueAssignedToMember,member,member.getName());
    }

    // select assign-to member
    if (selectedAssignedToMember != Member.ANY)
    {
      Widgets.setSelectedOptionMenuItem(widgetIssueAssignedToMember,selectedAssignedToMember);
    }
  }

  /** update issue version menu
   * @param widgetIssueTargetVersion versions combo widget
   * @param versionList new versions list
   * @param selectedTargetVersion selected target version
   */
  private void updateIssueVersion(final Combo widgetIssueTargetVersion, VersionList versionList, Version selectedTargetVersion)
  {
    Widgets.removeAllOptionMenuItems(widgetIssueTargetVersion);
    Widgets.addOptionMenuItem(widgetIssueTargetVersion,Version.ANY,"");
    for (Version version : versionList)
    {
      Widgets.addOptionMenuItem(widgetIssueTargetVersion,version,version.name);
    }

    // select version
    if (selectedTargetVersion != Version.NONE)
    {
      Widgets.setSelectedOptionMenuItem(widgetIssueTargetVersion,selectedTargetVersion);
    }
  }

  /** update issue status filter menu
   * @param widgetProject projects wiodget
   * @param widgetIssueId issue ids widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary issue summary widget
   * @param widgetIssueStatusFilterMenu widget issue status filter menu
   * @param statusTransitionList new status transition list
   */
  private void updateIssueStatusFilterMenu(final TreeCombo      widgetProject,
                                           final TableCombo     widgetIssueId,
                                           final TableCombo     widgetIssueShortName,
                                           final TreeCombo      widgetIssueSummary,
                                           final Menu           widgetIssueStatusFilterMenu,
                                           StatusTransitionList statusTransitionList
                                          )
  {
    // enable "any" status filter, discard other status filters
    MenuItem[] menuItems = widgetIssueStatusFilterMenu.getItems();
    menuItems[MENU_ITEM_INDEX_ANY].setSelection(true);
    for (int i = MENU_ITEM_INDEX_FILTER_START; i < menuItems.length; i++)
    {
      menuItems[i].dispose();
    }

    // add new status filters menu items
    boolean selectedSome = false;
    for (StatusTransition statusTransition : statusTransitionList)
    {
//Dprintf.dprintf("status=%s %s",status,Settings.issueStatusIdFilterSet.contains(status.id));
      // check if status is selected
//TODO: use names
      boolean selectedFlag = Settings.issueStatusIdFilterSet.contains(statusTransition.id);

      // add menu item
      MenuItem menuItem = Widgets.addMenuItemCheckbox(widgetIssueStatusFilterMenu,statusTransition.name);
      menuItem.setData(statusTransition);
      menuItem.setSelection(selectedFlag);
      menuItem.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          MenuItem         menuItem         = (MenuItem)selectionEvent.widget;
          StatusTransition statusTransition = (StatusTransition)menuItem.getData();

          // update issue status filter set
          if (menuItem.getSelection())
          {
            Settings.issueStatusIdFilterSet.add(statusTransition.id.backend);
          }
          else
          {
            Settings.issueStatusIdFilterSet.remove(statusTransition.id.backend);
          }
          Settings.setModified();

          // update issues for current selected project
          Project project = widgetProject.getSelection();
          updateIssues(widgetIssueId,
                       widgetIssueShortName,
                       widgetIssueSummary,
                       project
                      );

          // select/deselect "any" status filter
          widgetIssueStatusFilterMenu.getItem(MENU_ITEM_INDEX_ANY).setSelection(Settings.issueStatusIdFilterSet.size() == 0);
        }
      });

      if (selectedFlag) selectedSome = true;
    }

    // disable "any" status filter if some other status is selected
    menuItems[MENU_ITEM_INDEX_ANY].setSelection(!selectedSome);
    addTimeEntry1.widgetIssueStatusFilterButton.setBackground(selectedSome ? COLOR_CYAN : null);
    addTimeEntry2.widgetIssueStatusFilterButton.setBackground(selectedSome ? COLOR_CYAN : null);
  }

  /** update activities combo box
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param isActivitySupported true iff activity supported
   * @param activityList new activity data list
   * @param selectedActivity selected activity or null
   */
  public static void updateActivities(Label        widgetActivitiesTitle,
                                      Combo        widgetActivities,
                                      boolean      isActivitySupported,
                                      ActivityList activityList,
                                      Activity     selectedActivity
                                     )
  {
    Widgets.removeAllOptionMenuItems(widgetActivities);
    if (isActivitySupported)
    {
      // set list
Dprintf.dprintf("activityList=%d",activityList.size());
Dprintf.printStackTrace();
      for (Activity activity : activityList)
      {
        Widgets.addOptionMenuItem(widgetActivities,activity,activity.name);
        if (activity.isDefault) Widgets.setSelectedOptionMenuItem(widgetActivities,activity);
      }

      // select activity
      if (selectedActivity != null)
      {
        Widgets.setSelectedOptionMenuItem(widgetActivities,selectedActivity);
      }

      widgetActivitiesTitle.setEnabled(true);
      widgetActivities.setEnabled(true);
    }
    else
    {
      widgetActivitiesTitle.setEnabled(false);
      widgetActivities.setEnabled(false);
    }

    // send modified event
    Event event = new Event();
    event.widget = widgetActivities;
    widgetActivities.notifyListeners(SWT.Modify,event);
  }

  /** update activities combo box
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param activityList new activity data list
   * @param selectedActivityName selected activity name
   */
  public static void updateActivities(Label        widgetActivitiesTitle,
                                      Combo        widgetActivities,
                                      ActivityList activityList,
                                      String       selectedActivityName
                                     )
  {
    Activity selectedActivity = null;
    for (Activity activity : activityList)
    {
      if (activity.name.equals(selectedActivityName))
      {
        selectedActivity = activity;
        break;
      }
    }

    updateActivities(widgetActivitiesTitle,widgetActivities,true,activityList,selectedActivity);
  }

  /** update activities combo box
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param activityList new activity data list
   */
  public static void updateActivities(Label        widgetActivitiesTitle,
                                      Combo        widgetActivities,
                                      ActivityList activityList
                                     )
  {
    boolean  activitySupported = widgetActivities.getEnabled();
    Activity selectedActivity  = Widgets.getSelectedOptionMenuItem(widgetActivities,null);
    updateActivities(widgetActivitiesTitle,widgetActivities,activitySupported,activityList,selectedActivity);
  }

  /** update activities combo box
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param project project
   * @param selectedActivity selected activity or null
   */
  public static void updateActivities(Label    widgetActivitiesTitle,
                                      Combo    widgetActivities,
                                      Project  project,
                                      Activity selectedActivity
                                     )
  {
    if (project != null)
    {
      try
      {
        // get activity supported, activities for project
        boolean      activitySupported = project.isActivitySupported();
        ActivityList activityList      = project.getActivityList();

        // update
        updateActivities(widgetActivitiesTitle,widgetActivities,activitySupported,activityList,selectedActivity);
      }
      catch (BackendException exception)
      {
        printStackTrace(exception);
        String message = exception.getMessage();
        if (message == null) message = exception.getCause().getMessage();
        Dialogs.error(shell,"Cannot get activity data from server (error: "+message+")");
        return;
      }
    }
  }

  /** update activities combo box
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param project project
   */
  public static void updateActivities(Label   widgetActivitiesTitel,
                                      Combo   widgetActivities,
                                      Project project
                                     )
  {
    Activity selectedActivity = Widgets.getSelectedOptionMenuItem(widgetActivities,null);
    updateActivities(widgetActivitiesTitel,widgetActivities,project,selectedActivity);
  }

  /** filter issues/projects
   * @param widgetProject projects widget
   * @param widgetIssueId issue ids widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary issue summary widget
   * @param widgetIssueDoneRatio issue done ratio widget
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param filters filters or null
   */
  private void filterIssuesProjects(TreeCombo  widgetProject,
                                    TableCombo widgetIssueId,
                                    TableCombo widgetIssueShortName,
                                    TreeCombo  widgetIssueSummary,
                                    Combo      widgetIssueDoneRatio,
                                    Label      widgetActivitiesTitle,
                                    Combo      widgetActivities,
                                    String     filters[]
                                   )
  {
    try
    {
      IssueList   filteredIssueList;
      ProjectList filteredProjectList;

      // filter issue list by issue number
      filteredIssueList = new IssueList();
      for (String filter : filters)
      {
        try
        {
          Issue issue = Backend.getIssue(Backend.TypeSet.ALL,Integer.parseInt(filter));
          if (issue != Issue.NONE)
          {
            filteredIssueList.add(issue);
          }
        }
        catch (NumberFormatException exception)
        {
          // nothing to do: not a number
        }
      }

      // filter issue list by issue short name/summary
      if (filteredIssueList.isEmpty())
      {
        filteredIssueList = Backend.getIssueList();
        for (String filter : filters)
        {
          filter = filter.toLowerCase();

          int i = 0;
          while (i < filteredIssueList.size())
          {
            Issue issue = filteredIssueList.get(i);

            if (   issue.shortName.toLowerCase().contains(filter)
                || issue.summary.toLowerCase().contains(filter)
               )
            {
              // match filter => keep issue
              i++;
            }
            else
            {
              // filter issue
              filteredIssueList.remove(i);
            }
          }
        }
      }

      // filter project list by project number
      filteredProjectList = new ProjectList();
      for (String filter : filters)
      {
        try
        {
          Project project = Backend.getProject(Backend.TypeSet.ALL,Integer.parseInt(filter));
          if (project != null)
          {
            filteredProjectList.add(project);
          }
        }
        catch (NumberFormatException exception)
        {
          // nothing to do: not a number
        }
      }

      // filter project list by name
      if (filteredProjectList.isEmpty())
      {
        filteredProjectList = Backend.getProjectList();
        for (String filter : filters)
        {
          filter = filter.toLowerCase();

          int i = 0;
          while (i < filteredProjectList.size())
          {
            Project project = filteredProjectList.get(i);

            if (project.name.toLowerCase().contains(filter))
            {
              // match filter => keep project
              i++;
            }
            else
            {
              boolean keepFlag = false;
              for (Issue filteredIssue : filteredIssueList)
              {
                if (filteredIssue.projectId.equals(project.id))
                {
                  keepFlag = true;
                  break;
                }
              }
              if (keepFlag)
              {
                // used by issue => keep project
                i++;
              }
              else
              {
                // filter project
                filteredProjectList.remove(i);
              }
            }
          }
        }
      }

      // update projects
      updateProjects(widgetProject,filteredProjectList);

      // select first project/issue
      if      (!filteredIssueList.isEmpty())
      {
        Issue   issue   = filteredIssueList.get(0);
        Project project = issue.getProject();

        setSelectedProject(widgetProject,
                           widgetIssueId,
                           widgetIssueShortName,
                           widgetIssueSummary,
                           widgetIssueDoneRatio,
                           widgetActivitiesTitle,
                           widgetActivities,
                           project
                          );
        setSelectedIssue(widgetIssueId,
                         widgetIssueShortName,
                         widgetIssueSummary,
                         widgetIssueDoneRatio,
                         issue
                        );
      }
      else if (!filteredProjectList.isEmpty())
      {
        Project project = filteredProjectList.get(0);

        setSelectedProject(widgetProject,
                           widgetIssueId,
                           widgetIssueShortName,
                           widgetIssueSummary,
                           widgetIssueDoneRatio,
                           widgetActivitiesTitle,
                           widgetActivities,
                           project
                          );
      }
      else
      {
        resetProjects(widgetProject);
        resetIssues(widgetIssueId,
                    widgetIssueShortName,
                    widgetIssueSummary
                   );
      }
    }
    catch (BackendException exception)
    {
      // nothing to do
      printStackTrace(exception);
    }
  }

  /** filter issues/projects
   * @param widgetProject projects widget
   * @param widgetIssueId issue ids widget
   * @param widgetIssueShortName short names widget
   * @param widgetIssueSummary issue summary widget
   * @param widgetActivitiesTitle activities title widget
   * @param widgetActivities activities widget
   * @param filters filters
   */
  private void filterIssuesProjects(TreeCombo  widgetProject,
                                    TableCombo widgetIssueId,
                                    TableCombo widgetIssueShortName,
                                    TreeCombo  widgetIssueSummary,
                                    Label      widgetActivitiesTitle,
                                    Combo      widgetActivities,
                                    String     filters[]
                                   )
  {
    filterIssuesProjects(widgetProject,
                         widgetIssueShortName,
                         widgetIssueId,
                         widgetIssueSummary,
                         (Combo)null,  // widgetIssueDoneRatio
                         widgetActivitiesTitle,
                         widgetActivities,
                         filters
                        );
  }

  /** add new time entry to table
   * @param table table widget
   * @param timeEntry time entry to add
   */
  private TableItem addTableItem(Table table, TimeEntry timeEntry)
  {
    Activity activity = timeEntry.getActivity();
    Project  project  = timeEntry.getProject ();
    Issue    issue    = timeEntry.getIssue   ();

    return Widgets.addTableItem(table,
                                timeEntry,
                                !timeEntry.id.isNone() ? Integer.toString(timeEntry.id.backend) : "",
                                formatHours(timeEntry.duration),
                                (activity != null) ? activity.name : "",
                                (project != null) ? project.name : "",
                                (issue != null) ? issue.shortName : "",
                                (issue != null) ? issue.summary : "",
                                timeEntry.comments
                               );
  }

  /** update/insert time entry in table
   * @param table table widget
   * @param timeEntry time entry to update or insert
   */
  private TableItem updateInsertTableItem(Table table, TimeEntry timeEntry)
  {
    Activity activity = timeEntry.getActivity();
    Project  project  = timeEntry.getProject ();
    Issue    issue    = timeEntry.getIssue   ();

    return Widgets.updateInsertTableItem(table,
                                         timeEntry,
                                         !timeEntry.id.isNone() ? Integer.toString(timeEntry.id.backend) : "",
                                         formatHours(timeEntry.duration),
                                         (activity != null) ? activity.name : "",
                                         (project != null) ? project.name : "",
                                         (issue != null) ? issue.shortName : "",
                                         (issue != null) ? issue.summary : "",
                                         timeEntry.comments
                                        );
  }

  /** remove time entry from table
   * @param table table widget
   * @param timeEntry time entry to remove
   */
  private void removeTableItem(Table table, TimeEntry timeEntry)
  {
    Widgets.removeTableItem(table,timeEntry);
  }

  /** refresh time entry in table
   * @param table table widget
   * @param timeEntry time entry to refresh
   */
  private void refreshTableItem(Table table, TimeEntry timeEntry)
  {
    Activity activity = timeEntry.getActivity();
    Project  project  = timeEntry.getProject ();
    Issue    issue    = timeEntry.getIssue   ();

    Widgets.updateTableItem(table,
                            timeEntry,
                            !timeEntry.id.isNone() ? Integer.toString(timeEntry.id.backend) : "",
                            formatHours(timeEntry.duration),
                            (activity != null) ? activity.name : "",
                            (project != null) ? project.name : "",
                            (issue != null) ? issue.shortName : "",
                            (issue != null) ? issue.summary : "",
                            timeEntry.comments
                           );
  }

  /** dispose composite in tree item and in all sub-tree items
   * @param treeItem tree item
   */
  private void disposeComposites(TreeItem treeItem)
  {
    if (treeItem.getData() instanceof Composite)
    {
      Composite composite = (Composite)treeItem.getData();
      composite.dispose();
    }
    for (TreeItem subTreeItem : treeItem.getItems())
    {
      disposeComposites(subTreeItem);
    }
  }

  /** set tree item project
   * @param treeItem tree item
   * @param project project
   */
  private void setTreeEntry(TreeItem treeItem, Project project)
  {
    long durationSum          = project.getDurationSum(new Id(Id.ANY),new Id(Id.ANY));
    long estimatedDurationSum = project.getEstimatedDurationSum(new Id(Id.ANY));

    treeItem.setData(project);
    treeItem.setImage(project.backendType.getImage());
    treeItem.setText(0,"");
    treeItem.setText(1,Integer.toString(project.id.backend));
    treeItem.setText(2,project.name);
    treeItem.setText(3,formatHours(durationSum));
    treeItem.setText(4,formatHours(estimatedDurationSum));
  }

  /** set sub-tree projects
   * @param treeItem parent tree item
   * @param parentId parent issue id
   * @param projectList project list
   */
  private void setTreeEntries(TreeItem treeItem, Id parentId, ProjectList projectList)
  {
    for (Project project : projectList)
    {
      if (project.parentId == parentId)
      {
        TreeItem subTreeItem = new TreeItem(treeItem,SWT.NONE);
        setTreeEntry(subTreeItem,project);
        treeItem.setExpanded(true);

        setTreeEntries(subTreeItem,project.id,projectList);
      }
    }
  }

  /** filter projects
   * @param widgetProject projects widget
   * @param filters filters
   */
  private void filterProjects(Tree   widgetProjectTree,
                              String filters[]
                             )
  {
    try
    {
      ProjectList projectList = Backend.getProjectList();
      int         i;

      widgetProjectTree.removeAll();

      for (Project project : projectList)
      {
        boolean showFlag = true;

        for (String filter : filters)
        {
          filter = filter.toLowerCase();

          // filter by project number
          try
          {
            if (Integer.parseInt(filter) != project.id.backend)
            {
              showFlag = false;
              break;
            }
          }
          catch (NumberFormatException exception)
          {
            // nothing to do
          }

          // filter by project name
          if (!project.name.toLowerCase().contains(filter))
          {
            showFlag = false;
            break;
          }
        }

        if (showFlag)
        {
          TreeItem treeItem = Widgets.addTreeItem(widgetProjectTree,
                                                  project,
                                                  project.backendType.getImage(),
                                                  Widgets.TREE_ITEM_FLAG_NONE
                                                 );
          setTreeEntry(treeItem,project);
          treeItem.setExpanded(true);

          setTreeEntries(treeItem,project.id,projectList);
        }
      }
    }
    catch (BackendException exception)
    {
      // nothing to do
      printStackTrace(exception);
    }
  }

  /** get selected issues from sub-tree item
   * @param treeItem tree item
   * @param issueList issue list
   */
  private void getSelectedProjectIssueTreeEntries(TreeItem treeItem, IssueList issueList)
  {
//Dprintf.dprintf("treeItem.getData()=%s",treeItem.getData());
    if      (treeItem.getData() instanceof Project)
    {
      for (TreeItem subTreeItem : treeItem.getItems())
      {
        getSelectedProjectIssueTreeEntries(subTreeItem,issueList);
      }
    }
    else if (treeItem.getData() instanceof Issue)
    {
      issueList.add((Issue)treeItem.getData());
      for (TreeItem subTreeItem : treeItem.getItems())
      {
        getSelectedProjectIssueTreeEntries(subTreeItem,issueList);
      }
    }
    else if (treeItem.getData() instanceof Composite)
    {
      Tree issueTree = (Tree)((Composite)treeItem.getData()).getData();
      getSelectedProjectIssueTreeEntries(issueTree,issueList);
    }
  }

  /** get selected issues from sub-tree
   * @param tree sub-tree
   * @param issueList issue list
   */
  private void getSelectedProjectIssueTreeEntries(Tree tree, IssueList issueList)
  {
    for (TreeItem treeItem : tree.getSelection())
    {
      getSelectedProjectIssueTreeEntries(treeItem,issueList);
    }
  }

  /** get selected issues from tree
   * @param tree tree
   */
  private Issue[] getSelectedProjectIssueTreeEntries(Tree tree)
  {
    IssueList issueList = new IssueList();

    getSelectedProjectIssueTreeEntries(tree,issueList);

    return issueList.toArray(new Issue[issueList.size()]);
  }

  /** set tree item issue data
   * @param treeItem tree item
   * @param issue issue
   */
  private void setTreeEntry(TreeItem treeItem, Issue issue)
  {
    Tracker  tracker          = issue.getTracker         ();
    Status   status           = issue.getStatus          ();
    Priority priority         = issue.getPriority        ();
    Member   assignedToMember = issue.getAssignedToMember();
    long     durationSum      = issue.getDurationSum     (new Id(Id.ANY));

    treeItem.setData(issue);
    treeItem.setText(0,Integer.toString(issue.id.backend));
    treeItem.setText(1,issue.shortName);
    treeItem.setText(2,issue.summary);
    treeItem.setText(3,(tracker != null) ? tracker.name : "");
    treeItem.setText(4,(status != null) ? status.name : "");
    treeItem.setText(5,(priority != null) ? priority.name : "");
    treeItem.setText(6,String.format("%d%%",issue.doneRatio));
    treeItem.setText(7,(assignedToMember != null) ? assignedToMember.getName() : "");
    treeItem.setText(8,formatHours(durationSum));
    treeItem.setText(9,formatHours(issue.estimatedDuration));
    if (durationSum > issue.estimatedDuration)
    {
      treeItem.setBackground(COLOR_HOURS_EXCEEDED.background);
    }
  }

  /** set sub-tree issues
   * @param treeItem parent tree item
   * @param parentId parent issue id
   * @param issueList issue list
   */
  private void setTreeEntries(TreeItem treeItem, Id parentId, IssueList issueList)
    throws BackendException
  {
    for (Issue issue : issueList)
    {
      if (issue.parentId == parentId)
      {
        TreeItem subTreeItem = new TreeItem(treeItem,SWT.NONE);
        setTreeEntry(subTreeItem,issue);
        treeItem.setExpanded(true);

        setTreeEntries(subTreeItem,issue.id,issueList);
      }
    }
  }

  /** set tree issues
   * @param tree tree
   * @param issues issues
   */
  private void setTreeEntries(Tree tree, IssueList issueList)
  {
    tree.removeAll();
    try
    {
      for (Issue issue : issueList)
      {
  //Dprintf.dprintf("issue=%s",issue);
        if (issue.parentId.isNone())
        {
          TreeItem treeItem = new TreeItem(tree,SWT.NONE);
          setTreeEntry(treeItem,issue);
          setTreeEntries(treeItem,issue.id,issueList);
        }
      }
    }
    catch (BackendException exception)
    {
      // ignored
      printStackTrace(exception);
    }
  }

  /** filter projects
   * @param widgetIssuesTree issues widget
   * @param project project
   * @param filters filters
   */
  private void filterIssues(Tree    widgetIssuesTree,
                            Project project,
                            String  filters[]
                           )
  {
    try
    {
      IssueList issueList = project.getIssueList();
      int       i;

      widgetIssuesTree.removeAll();

      for (Issue issue : issueList)
      {
        boolean showFlag = true;

        for (String filter : filters)
        {
          filter = filter.toLowerCase();

          // filter by number
          try
          {
            if (Integer.parseInt(filter) != issue.id.backend)
            {
              showFlag = false;
              break;
            }
          }
          catch (NumberFormatException exception)
          {
            // nothing to do
          }

          // filter by text
          if (!issue.summary.toLowerCase().contains(filter))
          {
            showFlag = false;
            break;
          }
        }

        if (showFlag)
        {
          TreeItem treeItem = new TreeItem(widgetIssuesTree,SWT.NONE);
          setTreeEntry(treeItem,issue);
          treeItem.setExpanded(true);

          setTreeEntries(treeItem,issue.id,issueList);
        }
      }
    }
    catch (BackendException exception)
    {
      // nothing to do
      printStackTrace(exception);
    }
  }

  /** set tree time entry
   * @param treeItem parent tree item
   * @param spentOn spent-on date
   * @param id id or Id.NONE
   */
  private void setTreeEntry(TreeItem treeItem, SpentOn spentOn, Id id)
  {
    if (!treeItem.isDisposed())
    {
      int  spentOnWeekday = spentOn.getWeekday();
      Date midnight       = new Date((System.currentTimeMillis()/(24*60*60*1000))*(24*60*60*1000) + 24*60*60*1000);

      // get duration sum
      long durationSum = Backend.getDurationSum(spentOn);

      treeItem.setText(0,spentOn.toString(DATE_FORMAT));
      treeItem.setText(1,(id.backend != Id.NONE) ? Integer.toString(id.backend) : "");
      treeItem.setText(2,(durationSum != Backend.DURATION_UPDATE) ? formatHours(durationSum) : "");
      if      ((spentOnWeekday == Calendar.SATURDAY) || (spentOnWeekday == Calendar.SUNDAY))
      {
        treeItem.setForeground(COLOR_TIME_ENTRIES_WEEKEND.foreground);
      }
      else if (isHolidayDate(spentOn.getDate()))
      {
        treeItem.setText(3,getHolidayDescription(spentOn.getDate()));
        treeItem.setForeground(COLOR_TIME_ENTRIES_HOLIDAY.foreground);
      }
      else if (!spentOn.isBefore(midnight))
      {
        treeItem.setForeground(COLOR_FUTURE.foreground);
      }
      else if (durationSum == Backend.DURATION_UPDATE)
      {
        treeItem.setForeground(COLOR_TIME_ENTRIES_UPDATE.foreground);
      }
      else if (durationSum < (long)(Settings.workingHoursPerDay*60.0))
      {
        treeItem.setForeground(COLOR_TIME_ENTRIES_INCOMPLETE.foreground);
      }
      else
      {
        treeItem.setForeground(COLOR_TIME_ENTRIES.foreground);
      }
    }
  }

  /** set tree time entry
   * @param treeItem parent tree item
   * @param spentOn spent-on date
   */
  private void setTreeEntry(TreeItem treeItem, SpentOn spentOn)
  {
    setTreeEntry(treeItem,spentOn,new Id());
  }

  /** set tree time entry
   * @param treeItem parent tree item
   * @param timeEntry time entry
   */
  private void setTreeEntry(TreeItem treeItem, TimeEntry timeEntry)
  {
    final String EMPTY = StringUtils.repeat(' ',255);

    Activity activity = timeEntry.getActivity();
    Project  project  = timeEntry.getProject ();
    Issue    issue    = timeEntry.getIssue   ();

    treeItem.setText(0,EMPTY);
    treeItem.setText(1,timeEntry.id.isNone() ? Integer.toString(timeEntry.id.backend) : "");
    treeItem.setText(2,formatHours(timeEntry.duration));
    treeItem.setText(3,(activity != null) ? activity.name : "");
    treeItem.setText(4,(project != null) ? project.name : "");
    treeItem.setText(5,(issue != null) ? issue.shortName : "");
    treeItem.setText(6,(issue != null) ? issue.summary : "");
    treeItem.setText(7,timeEntry.comments);
    treeItem.setForeground(COLOR_TIME_ENTRIES.foreground);
  }

  /** expand time entries sub-tree
   * @param treeItem parent tree item
   * @param spentOnStart,spentOnEnd spent end from/to
   */
  private void expandTimeEntryTree(TreeItem treeItem, SpentOn spentOnStart, SpentOn spentOnEnd)
  {
    treeItem.removeAll();
    Date date = spentOnEnd.getDate();
    while (date.compareTo(spentOnStart.getDate()) >= 0)
    {
      SpentOn spentOn = new SpentOn(date);

      TreeItem subTreeItem = new TreeItem(treeItem,SWT.NONE);
      subTreeItem.setData(spentOn);
      setTreeEntry(subTreeItem,spentOn);

      if (subTreeItem.getExpanded())
      {
        expandTimeEntryTree(subTreeItem,spentOn);
      }
      else
      {
        collapseTimeEntryTree(subTreeItem);
      }

//TODO: use calendar
      // previous day
      date = new Date(date.getTime()-24*60*60*1000);
    }
    treeItem.setExpanded(true);
  }

  /** expand time entries sub-tree
   * @param treeItem parent tree item
   * @param spentOn spent-on date
   */
  private void expandTimeEntryTree(TreeItem treeItem, SpentOn spentOn)
  {
    treeItem.removeAll();
    try
    {
      for (TimeEntry timeEntry : Backend.getTimeEntryList(spentOn))
      {
        TreeItem subTreeItem = new TreeItem(treeItem,SWT.NONE);
        subTreeItem.setData(timeEntry);
        setTreeEntry(subTreeItem,timeEntry);
      }
    }
    catch (BackendException exception)
    {
      // ignored
      printStackTrace(exception);
    }
    treeItem.setExpanded(true);
  }

  /** collapse time entries sub-tree
   * @param treeItem parent tree item
   */
  private void collapseTimeEntryTree(TreeItem treeItem)
  {
    treeItem.setExpanded(false);
    treeItem.removeAll();
    new TreeItem(treeItem,SWT.NONE);
  }

  private void addTreeItem(Tree tree, TimeEntry timeEntry)
  {
    // find weekday tree item
    TreeItem weekdayTreeItem = null;
    for (TreeItem treeItem : Widgets.getAllTreeItems(tree))
    {
      if     (treeItem.getData() instanceof SpentOn)
      {
        SpentOn spentOn = (SpentOn)treeItem.getData();

        if (spentOn.equals(timeEntry.spentOn))
        {
          weekdayTreeItem = treeItem;
          break;
        }
      }
    }

    if (weekdayTreeItem != null)
    {
      if (weekdayTreeItem.getExpanded())
      {
        // add time entry
        TreeItem treeItem = new TreeItem(weekdayTreeItem,SWT.NONE);
        treeItem.setData(timeEntry);
        setTreeEntry(treeItem,timeEntry);
        setTreeEntry(weekdayTreeItem,timeEntry.spentOn);
      }

      // update weekday tree item
      tree.clearAll(true);
    }
  }

  /** remove time entry and collaps sub-tree if needed
   * @param tree tree
   * @param timeEntry time entry to remove
   */
  private void removeTreeItem(Tree tree, TimeEntry timeEntry)
  {
    // get tree item
    TreeItem treeItem = Widgets.getTreeItem(tree,timeEntry);
    if (treeItem != null)
    {
      // get weekday tree item
      TreeItem weekdayTreeItem = treeItem.getParentItem();

      // remove tree item
      Widgets.removeTreeItem(treeItem);
      setTreeEntry(weekdayTreeItem,timeEntry.spentOn);

      // update weekday tree item
      tree.clearAll(true);
    }
  }

  /** refresh time entries in tree
   * @param tree tree widget
   * @param spentOnRefresh spent-on refresh date
   */
  private void refreshTreeItems(Tree tree, TreeItem treeItem, SpentOn spentOnRefresh)
  {
    for (TreeItem subTreeItem : treeItem.getItems())
    {
      if (subTreeItem.getExpanded())
      {
        if     (subTreeItem.getData() instanceof SpentOn)
        {
          SpentOn spentOn = (SpentOn)subTreeItem.getData();

          if (!spentOnRefresh.isBefore(spentOn) && !spentOnRefresh.isAfter(spentOn))
          {
            expandTimeEntryTree(subTreeItem,spentOn);
          }
        }
        else if (subTreeItem.getData() instanceof TimeEntry)
        {
          TimeEntry timeEntry = (TimeEntry)subTreeItem.getData();

          if (!spentOnRefresh.isBefore(timeEntry.spentOn) && !spentOnRefresh.isAfter(timeEntry.spentOn))
          {
            Event event = new Event();
            event.item = subTreeItem;
            tree.notifyListeners(SWT.SetData,event);
          }
        }
      }
    }
  }

  /** refresh time entries in tree
   * @param tree tree widget
   * @param spentOnRefresh spent-on refresh date
   */
  private void refreshTreeItems(Tree tree, SpentOn spentOnRefresh)
  {
    for (TreeItem treeItem : tree.getItems())
    {
      TimeEntryTreeItemData itemData = (TimeEntryTreeItemData)treeItem.getData();

      if ((itemData != null) && !spentOnRefresh.isBefore(itemData.spentOnStart) && !spentOnRefresh.isAfter(itemData.spentOnEnd))
      {
        refreshTreeItems(tree,treeItem,spentOnRefresh);

        Event event = new Event();
        event.item = treeItem;
        tree.notifyListeners(SWT.SetData,event);
      }
    }
  }

  /** edit issue
   * @param issue issue to edit
   * @return true if issue edited, false otherwise
   */
  private boolean editIssue(final Issue issue, String title, String okText)
  {
    // key accelerators

    final Shell dialog;
    Composite   composite,subComposite;
    Label       label;
    Combo       combo;
    Button      button;
    MenuItem    menuItem;

    // repository edit dialog
    dialog = Dialogs.openModal(shell,title+(!issue.id.isNone() ? ": "+issue.id.format() : ""),800,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Label      widgetProject;
    final TableCombo widgetParentIssueId;
    final TableCombo widgetParentIssueShortName;
    final TreeCombo  widgetParentIssueSummary;
    final Combo      widgetTracker;
    final Label      widgetEpicNameTitle;
    final Combo      widgetEpicName;
    final Text       widgetSummary;
    final StyledText widgetDescription;
    final Combo      widgetIssueStatusTransition;
    final Combo      widgetIssuePriority;
    final Combo      widgetIssueAssignedToMember;
    final Combo      widgetIssueTargetVersion;
    final Spinner    widgetEstimatedHourFraction,widgetEstimatedMinuteFraction;
    final Label      widgetIssueDoneRatioTitle;
    final Combo      widgetIssueDoneRatio;
    final Button     widgetSave;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,0.0,0.0},new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Project:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      widgetProject = Widgets.newView(composite);
      Widgets.layout(widgetProject,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Parent:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,1.0}));
      Widgets.layout(subComposite,1,1,TableLayoutData.WE);
      {
        widgetParentIssueId = new TableCombo(subComposite);
        widgetParentIssueId.setToolTipText("Parent issue.");
        Widgets.layout(widgetParentIssueId,0,0,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);

        widgetParentIssueShortName = new TableCombo(subComposite);
        widgetParentIssueShortName.setToolTipText("Parent issue.");
        widgetParentIssueShortName.addTableColumn(0,SWT.LEFT,80,false);
        Widgets.layout(widgetParentIssueShortName,0,1,TableLayoutData.W,0,0,0,0,120,SWT.DEFAULT);

        widgetParentIssueSummary = new TreeCombo(subComposite,SWT.LEFT,SWT.NONE,100,TreeCombo.DEFAULT_HEIGHT);
        widgetParentIssueSummary.addTreeColumn("#",               SWT.RIGHT,80,false);
        widgetParentIssueSummary.addTreeColumn("Issue short name",SWT.LEFT, 80,true );
        widgetParentIssueSummary.addTreeColumn("Issue",           SWT.LEFT,  0,true );
        widgetParentIssueSummary.setLayout(new double[]{0.0,0.0,1.0});
        widgetParentIssueSummary.setToolTipText("Parent issue.");
        Widgets.layout(widgetParentIssueSummary,0,2,TableLayoutData.WE);
      }

      label = Widgets.newLabel(composite,"Tracker:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,1.0}));
      Widgets.layout(subComposite,2,1,TableLayoutData.WE);
      {
        widgetTracker = Widgets.newOptionMenu(subComposite);
        Widgets.layout(widgetTracker,0,0,TableLayoutData.W);

        widgetEpicNameTitle = Widgets.newLabel(subComposite,"Epic name:");
        Widgets.layout(widgetEpicNameTitle,0,1,TableLayoutData.W);
        widgetEpicName = Widgets.newCombo(subComposite);
        Widgets.layout(widgetEpicName,0,2,TableLayoutData.WE);
      }

      label = Widgets.newLabel(composite,"Summary:");
      Widgets.layout(label,3,0,TableLayoutData.W);
      widgetSummary = Widgets.newText(composite);
      widgetSummary.setToolTipText("Summary text of issue.");
      widgetSummary.setText(issue.summary);
      Widgets.layout(widgetSummary,3,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Description:");
      Widgets.layout(label,4,0,TableLayoutData.NW);
      widgetDescription = Widgets.newStyledText(composite,SWT.LEFT|SWT.BORDER|SWT.V_SCROLL|SWT.H_SCROLL|SWT.MULTI|SWT.WRAP);
      widgetSummary.setToolTipText("Detailed description text of issue.");
      widgetDescription.setText(issue.description);
      Widgets.layout(widgetDescription,4,1,TableLayoutData.NSWE);

      label = Widgets.newLabel(composite,"Status:");
      Widgets.layout(label,5,0,TableLayoutData.W);
      widgetIssueStatusTransition = Widgets.newOptionMenu(composite);
      //TODO: sync with server changes
      widgetIssueStatusTransition.setEnabled(false);
      Widgets.layout(widgetIssueStatusTransition,5,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Priority:");
      Widgets.layout(label,6,0,TableLayoutData.W);
      widgetIssuePriority = Widgets.newOptionMenu(composite);
      Widgets.layout(widgetIssuePriority,6,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Assigned to:");
      Widgets.layout(label,7,0,TableLayoutData.W);
      widgetIssueAssignedToMember = Widgets.newOptionMenu(composite);
      Widgets.layout(widgetIssueAssignedToMember,7,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Target version:");
      Widgets.layout(label,8,0,TableLayoutData.W);
      widgetIssueTargetVersion = Widgets.newOptionMenu(composite);
      //TODO: sync with server changes
      widgetIssueTargetVersion.setEnabled(false);
      Widgets.layout(widgetIssueTargetVersion,8,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Estimated time:");
      Widgets.layout(label,9,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{1.0,0.0,0.0}));
      Widgets.layout(subComposite,9,1,TableLayoutData.W);
      {
        widgetEstimatedHourFraction = Widgets.newSpinner(subComposite,SWT.NONE,0,Integer.MAX_VALUE);
        widgetEstimatedHourFraction.setToolTipText("Time entry spent duration.");
        widgetEstimatedHourFraction.setTextLimit(3);
        widgetEstimatedHourFraction.setIncrement(1);
        widgetEstimatedHourFraction.setSelection(0);
        Widgets.layout(widgetEstimatedHourFraction,0,0,TableLayoutData.WE,0,0,4,0,90,SWT.DEFAULT);
        label = Widgets.newLabel(subComposite,"h");
        Widgets.layout(label,0,1,TableLayoutData.W);
        widgetEstimatedMinuteFraction = Widgets.newSpinner(subComposite,SWT.NONE,-60,60);
        widgetEstimatedMinuteFraction.setToolTipText("Time entry spent minutes.");
        widgetEstimatedMinuteFraction.setSelection(Settings.minTimeDelta);
        widgetEstimatedMinuteFraction.setTextLimit(2);
        widgetEstimatedMinuteFraction.setIncrement(Settings.minTimeDelta);
        Widgets.layout(widgetEstimatedMinuteFraction,0,2,TableLayoutData.WE,0,0,4,0,90,SWT.DEFAULT);
        label = Widgets.newLabel(subComposite,"min");
        Widgets.layout(label,0,3,TableLayoutData.W);

        combo = Widgets.newOptionMenu(subComposite);
        combo.setToolTipText("Estimated time");
        Widgets.layout(combo,0,4,TableLayoutData.W,0,0,0,0,64,SWT.DEFAULT);
        Widgets.addOptionMenuItem(combo,new Long(0L),"");
        for (int i = 1; i <= 8*60/Settings.minTimeDelta; i++)
        {
          Widgets.addOptionMenuItem(combo,new Long(i*Settings.minTimeDelta),formatHours(i*Settings.minTimeDelta));
        }
        Widgets.addOptionMenuItem(combo,new Long(16*60),"16:00");
        Widgets.addOptionMenuItem(combo,new Long(1*8*60),"1 day");
        Widgets.addOptionMenuItem(combo,new Long(2*8*60),"2 days");
        Widgets.addOptionMenuItem(combo,new Long(3*8*60),"3 days");
        Widgets.addOptionMenuItem(combo,new Long(4*8*60),"4 days");
        Widgets.addOptionMenuItem(combo,new Long(1*5*8*60),"1 week");
        Widgets.addOptionMenuItem(combo,new Long(2*5*8*60),"2 weeks");
        Widgets.addOptionMenuItem(combo,new Long(3*5*8*60),"3 weeks");
        Widgets.addOptionMenuItem(combo,new Long(4*5*8*60),"4 weeks");
        combo.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Combo widget = (Combo)selectionEvent.widget;
            long  value  = Widgets.getSelectedOptionMenuItem(widget,0L);
            if (value > 0L)
            {
              int hourFraction   = (int)(value/60);
              int minuteFraction = (int)(value%60);

              widgetEstimatedHourFraction.setSelection(hourFraction);
              widgetEstimatedMinuteFraction.setSelection(minuteFraction);
            }

            Widgets.setSelectedOptionMenuItem(widget,0);
          }
        });
      }

      widgetIssueDoneRatioTitle = Widgets.newLabel(composite,"% done:");
      widgetIssueDoneRatioTitle.setEnabled(issue.isEditProgressSupported());
      Widgets.layout(widgetIssueDoneRatioTitle,10,0,TableLayoutData.W);
      widgetIssueDoneRatio = Widgets.newOptionMenu(composite);
      widgetIssueDoneRatio.setToolTipText("Done percentage");
      widgetIssueDoneRatio.setEnabled(Backend.isEditProgressSupported(issue.backendType,issue.id));
      Widgets.layout(widgetIssueDoneRatio,10,1,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);
      Widgets.setOptionMenuItems(widgetIssueDoneRatio,
                                 new Object[]{"0%" ,   0,
                                              "10%",  10,
                                              "20%",  20,
                                              "30%",  30,
                                              "40%",  40,
                                              "50%",  50,
                                              "60%",  60,
                                              "70%",  70,
                                              "80%",  80,
                                              "90%",  90,
                                              "100%",100
                                             }
                                );
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetSave = Widgets.newButton(composite,okText);
      Widgets.layout(widgetSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetSave.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Tracker tracker  = Widgets.getSelectedOptionMenuItem(widgetTracker,Tracker.class);
          String  epicName = widgetEpicName.getEnabled() ? widgetEpicName.getText().trim() : "";
          String  summary  = widgetSummary.getText().trim();

          // check parameters
          if (tracker == null)
          {
            Dialogs.error(shell,"Please select a tracker for the issue.");
            Widgets.setFocus(widgetTracker);
            return;
          }
          if ((tracker != null) && tracker.isRequireEpicName() && ((widgetEpicName == null) || widgetEpicName.getText().trim().isEmpty()))
          {
            Dialogs.error(shell,"Please enter a epic name for the issue.");
            Widgets.setFocus(widgetEpicName);
            return;
          }
          if (summary.isEmpty())
          {
            Dialogs.error(shell,"Please enter a summary for the issue.");
            Widgets.setFocus(widgetSummary);
            return;
          }

          StatusTransition statusTransition = Widgets.getSelectedOptionMenuItem(widgetIssueStatusTransition,StatusTransition.class);
          Priority         priority         = Widgets.getSelectedOptionMenuItem(widgetIssuePriority,Priority.class);
          Member           assignedToMember = Widgets.getSelectedOptionMenuItem(widgetIssueAssignedToMember,Member.class);
          Version          targetVersion    = Widgets.getSelectedOptionMenuItem(widgetIssueTargetVersion,Version.class);

          issue.parentId           = widgetParentIssueId.getSelection(Issue.NONE).id;
          issue.trackerId          = tracker.id;
          issue.epicName           = epicName;
          issue.summary            = summary;
          issue.description        = widgetDescription.getText().trim();
//TODO: id
          issue.statusTransitionId = (statusTransition != null) ? statusTransition.id : new Id();
          issue.priorityId         = (priority != null) ? priority.id : new Id();
          issue.assignedToMemberId = (assignedToMember != null) ? assignedToMember.id : new Id();
          issue.targetVersionId    = (targetVersion != null) ? targetVersion.id : new Id();
          issue.setEstimatedDuration(widgetEstimatedHourFraction.getSelection(),widgetEstimatedMinuteFraction.getSelection());
          issue.doneRatio          = Widgets.getSelectedOptionMenuItem(widgetIssueDoneRatio,0);

          lastParentId = issue.parentId;

          Settings.geometryEditIssue = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    TrackerList          trackerList          = null;
    StatusTransitionList statusTransitionList = null;
    PriorityList         priorityList         = null;
    VersionList          versionList          = null;
    MemberList           memberList           = null;
    IssueList            issueList            = null;

    Project              project              = null;
    Issue                parentIssue          = null;
    Tracker              tracker              = null;
    Status               status               = null;
    StatusTransition     statusTransition     = null;
    Priority             priority             = null;
    Member               assignedToMember     = null;
    Version              version              = null;
    try
    {
      // get role ids for member list
      IdList roleIdList = new IdList();
      issue.getRoleIds(roleIdList,Settings.managerRoles  );
      issue.getRoleIds(roleIdList,Settings.developerRoles);
      issue.getRoleIds(roleIdList,Settings.designerRoles );

      // get trackers, status transition, priorities, version, users, issues lists
      trackerList          = issue.getTrackerList();
      statusTransitionList = issue.getStatusTransitionList();
      priorityList         = issue.getPriorityList();
      memberList           = issue.getMemberList(roleIdList);
      versionList          = issue.getVersionList();
      issueList            = issue.getIssueList();

      // get project, parent issue, tracker, status, priority, assigned-to, version
      project              = issue.getProject();
      if ((parentIssue == null) && !issue.parentId.isNone()) parentIssue = Backend.getIssue(new Backend.TypeSet(issue.backendType),issue.parentId);
      if ((parentIssue == null) && !lastParentId.isNone()) parentIssue = Backend.getIssue(new Backend.TypeSet(issue.backendType),lastParentId);
      tracker              = issue.getTracker();
      status               = issue.getStatus();
      statusTransition     = issue.getStatusTransition();
      priority             = issue.getPriority();
      assignedToMember     = issue.getAssignedToMember();
      version              = issue.getVersion();
    }
    catch (BackendException exception)
    {
      printStackTrace(exception);
      Dialogs.error(shell,"Cannot get data (error: "+exception.getMessage()+")");
      return false;
    }
//Dprintf.dprintf("-------------------------------------------------");
//Dprintf.dprintf("issue=%s",issue);
//Dprintf.dprintf("trackerList=%d",trackerList.size());
//Dprintf.dprintf("versionList=%d",versionList.size()); for (Version v : versionList) Dprintf.dprintf("v=%s",v);
//Dprintf.dprintf("memberList=%d",memberList.size()); for (Member m : memberList) Dprintf.dprintf("m=%s",m);
//Dprintf.dprintf("assignedToMember=%s",assignedToMember);
//Dprintf.dprintf("statusId=%s -> status=%s",issue.statusId,status);
//Dprintf.dprintf("statusTransitionId=%s -> statusTransition=%s",issue.statusTransitionId,statusTransition);

    widgetProject.setText(project.name);
    updateIssues(widgetParentIssueId,
                 widgetParentIssueShortName,
                 widgetParentIssueSummary,
                 issueList,
                 parentIssue,
                 false,
                 true
                );
    updateTrackers(widgetTracker,
                   trackerList,
                   ((tracker == null) && (trackerList.size() == 1)) ? trackerList.get(0) : tracker
                  );
    if (issue.isEpicNameSupported())
    {
      widgetEpicNameTitle.setEnabled(true);
      if (issue.isTrackerRequireEpicName())
      {
        List<String> epicNameList = tracker.getEpicNameList();

        widgetEpicName.setEnabled(true);

        for (String epicName : epicNameList)
        {
          Widgets.addOptionMenuItem(widgetEpicName,epicName,epicName);
        }
        if (issue.epicName != null) widgetEpicName.setText(issue.epicName);
      }
      else
      {
        widgetEpicName.setEnabled(false);
      }
    }
    else
    {
      widgetEpicNameTitle.setEnabled(false);
      widgetEpicName.setEnabled(false);
    }
//Dprintf.dprintf("status=%s",status);
//Dprintf.dprintf("statusTransition=%s",statusTransition);
    updateIssueStatusTransition(widgetIssueStatusTransition,statusTransitionList,status,statusTransition);
    updateIssuePriorities      (widgetIssuePriority,        priorityList,        priority               );
    updateIssueAssignedToMember(widgetIssueAssignedToMember,memberList,          assignedToMember       );
    updateIssueVersion         (widgetIssueTargetVersion,   versionList,         version                );
    widgetEstimatedHourFraction.setSelection(issue.getEstimatedHourFraction());
    widgetEstimatedMinuteFraction.setSelection(issue.getEstimatedMinuteFraction());
    updateIssueDoneRatio(widgetIssueDoneRatio,issue.doneRatio);

    // add listeners
    widgetParentIssueId.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TableCombo widget = (TableCombo)selectionEvent.widget;

        Issue issue = widget.getSelection(Issue.NONE);
        if (issue != Issue.NONE)
        {
          setSelectedIssue(widgetParentIssueId,widgetParentIssueShortName,widgetParentIssueSummary,issue);
        }
      }
    });
    widgetParentIssueSummary.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TreeCombo widget = (TreeCombo)selectionEvent.widget;

        Issue issue = widget.getSelection(Issue.NONE);
        if (issue != Issue.NONE)
        {
          setSelectedIssue(widgetParentIssueId,widgetParentIssueShortName,widgetParentIssueSummary,issue);
        }
      }
    });
    widgetTracker.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Combo widget = (Combo)selectionEvent.widget;

        if (widgetEpicName.getEnabled())
        {
          Tracker tracker = Widgets.getSelectedOptionMenuItem(widget,Tracker.class);
          if (tracker != null)
          {
            if (tracker.isRequireEpicName())
            {
              List<String> epicNameList = tracker.getEpicNameList();

              widgetEpicName.setEnabled(true);

              String currentEpicName = widgetEpicName.getText();
              Widgets.removeAllOptionMenuItems(widgetEpicName);
              for (String epicName : epicNameList)
              {
                Widgets.addOptionMenuItem(widgetEpicName,epicName,epicName);
              }
              widgetEpicName.setText(currentEpicName);
            }
            else
            {
              widgetEpicName.setEnabled(false);
            }
          }
          else
          {
            widgetEpicName.setEnabled(false);
          }
        }
      }
    });
    widgetEstimatedMinuteFraction.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        int hourFraction   = widgetEstimatedHourFraction.getSelection();
        int minuteFraction = widgetEstimatedMinuteFraction.getSelection();

        // Note: sometimes the minute value is not a multiple of the increment. Correct this.
        if (minuteFraction >= 0)
        {
          minuteFraction = ((minuteFraction+Settings.minTimeDelta-1)/Settings.minTimeDelta)*Settings.minTimeDelta;
        }
        else
        {
          minuteFraction = -((((-minuteFraction)+Settings.minTimeDelta-1)/Settings.minTimeDelta)*Settings.minTimeDelta);
        }

        if      (minuteFraction >= 60)
        {
          // increment hours
          hourFraction   += minuteFraction/60;
          minuteFraction %= 60;
        }
        else if (minuteFraction < 0)
        {
          if (hourFraction > 0)
          {
            // decrement hours, reset
            hourFraction   -= ((-minuteFraction)+60-1)/60;
            minuteFraction = minuteFraction+((((-minuteFraction)+60-1)/60)*60);
          }
          else
          {
            // reset
            minuteFraction = 0;
          }
        }

        widgetEstimatedHourFraction.setSelection(hourFraction);
        widgetEstimatedMinuteFraction.setSelection(minuteFraction);
      }
    });
    widgetTracker.addModifyListener(new ModifyListener()
    {
      @Override
      public void modifyText(ModifyEvent modifyEvent)
      {
        Tracker tracker = Widgets.getSelectedOptionMenuItem(widgetTracker,Tracker.class);
        widgetEpicName.setEnabled((tracker != null) && tracker.isRequireEpicName());
      }
    });

    // set next focus
    Widgets.setNextFocus(widgetParentIssueId,
                         widgetParentIssueSummary,
                         widgetTracker,
                         widgetSummary,
                         widgetDescription,
                         widgetIssueStatusTransition,
                         widgetIssuePriority,
                         widgetIssueAssignedToMember,
                         widgetIssueTargetVersion,
                         widgetEstimatedHourFraction,
                         widgetEstimatedMinuteFraction,
                         widgetIssueDoneRatio,
                         widgetSave
                        );

    // show dialog
    Dialogs.show(dialog,Settings.geometryEditIssue);

    // run
    Widgets.setFocus(widgetSummary);
    return (Boolean)Dialogs.run(dialog,false);
  }

  /** add new issue
   * @param projectId project id
   */
  private void addIssue(Project project)
    throws BackendException
  {
    final Issue issue = new Issue(project.backendType,
                                  project.id
                                 );
    if (editIssue(issue,"Add issue","Add"))
    {
      try
      {
        issue.setModified(Backend.ModifiedStates.ADDED);
        Widgets.notify(shell,USER_EVENT_ADD_ISSUE,issue);
      }
      catch (BackendException exception)
      {
        printStackTrace(exception);
        Dialogs.error(shell,"Cannot add issue (error: "+exception.getMessage()+")");
        return;
      }
      catch (InterruptedException exception)
      {
        // ignored
      }
    }
  }

  /** remove issue
   * @param issue issue to remove
   */
  private void removeIssue(Issue issue)
    throws BackendException
  {
    try
    {
      issue.setModified(Backend.ModifiedStates.DELETED);
      Widgets.notify(shell,USER_EVENT_REMOVE_ISSUE,issue);
    }
    catch (InterruptedException exception)
    {
      // ignored
    }
  }

  /** edit time entry
   * @param timeEntry time entry to edit
   * @param title window title text
   * @param actionText action text
   * @param okText ok-button text
   * @return true if time entry edited, false otherwise
   */
  private boolean editTimeEntry(final TimeEntry timeEntry,
                                String          title,
                                final String    actionText,
                                String          okText
                               )
  {
    /** dialog data
     */
    class Data
    {
      Data()
      {
      }
    };

    final Data  data = new Data();
    final Shell dialog;
    Composite   composite,subComposite,subSubComposite;
    Label       label;
    Combo       combo;
    Button      button;
    MenuItem    menuItem;

    // get current issue id
    final Id prevIssueId = timeEntry.issueId;

    // repository edit dialog
    dialog = Dialogs.openModal(shell,title+(!timeEntry.id.isNone() ? ": "+timeEntry.id.format() : ""),700,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

Dprintf.dprintf("timeEntry=%s",timeEntry);

    final Text       widgetFind;
    final TableCombo widgetHistory;
    final TreeCombo  widgetProject;
    final TableCombo widgetIssueId;
    final TableCombo widgetIssueShortName;
    final TreeCombo  widgetIssueSummary;
    final Combo      widgetIssueDoneRatio;
    final Button     widgetIssueStatusOwnButton;
    final Button     widgetIssueStatusFilterButton;
    final Menu       widgetIssueStatusFilterMenu;
    final CDateTime  widgetSpentOn;
    final Spinner    widgetSpentHourFraction;
    final Spinner    widgetSpentMinuteFraction;
    final Label      widgetActivitiesTitle;
    final Combo      widgetActivities;
    final Combo      widgetComments;
    final Button     widgetSave;
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Find:",SWT.NONE,Settings.keyFind);
      Widgets.layout(label,0,0,TableLayoutData.W);

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,0,1,TableLayoutData.WE);
      {
        widgetFind = Widgets.newText(subComposite,SWT.SEARCH|SWT.ICON_SEARCH|SWT.ICON_CANCEL);
        widgetFind.setToolTipText("Enter a number or some text to find a project or an issue.");
        widgetFind.setLayoutData(new TableLayoutData(0,0,TableLayoutData.WE));

        widgetHistory = new TableCombo(subComposite,SWT.CENTER,SWT.CENTER,300);
        widgetHistory.setToolTipText("Time entries history.");
        widgetHistory.addTableColumn(0,"Project",         SWT.LEFT,100,true);
        widgetHistory.addTableColumn(1,"Issue short name",SWT.LEFT, 80,true);
        widgetHistory.addTableColumn(1,"Issue",           SWT.LEFT,150,true);
        widgetHistory.addTableColumn(2,"Duration",        SWT.LEFT, 50,true);
        widgetHistory.setText("History");
        Widgets.layout(widgetHistory,0,1,TableLayoutData.E,0,0,0,0,80,SWT.DEFAULT);
      }

      label = Widgets.newLabel(composite,"Project:",SWT.NONE,Settings.keyProject);
      Widgets.layout(label,1,0,TableLayoutData.W);

      widgetProject = new TreeCombo(composite,SWT.LEFT,SWT.NONE,100,TreeCombo.DEFAULT_HEIGHT);
      Widgets.layout(widgetProject,1,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Issue:",SWT.NONE,Settings.keyIssue);
      Widgets.layout(label,2,0,TableLayoutData.W);

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,1.0,0.0,0.0,0.0}));
      Widgets.layout(subComposite,2,1,TableLayoutData.WE);
      {
        widgetIssueId = new TableCombo(subComposite,SWT.LEFT,SWT.NONE,150);
        widgetIssueId.addTableColumn(0,SWT.RIGHT,80,false);
        widgetIssueId.addTableColumn(1,SWT.LEFT, 70,false);
        Widgets.layout(widgetIssueId,0,0,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);

        widgetIssueShortName = new TableCombo(subComposite,SWT.LEFT,SWT.NONE,150);
        widgetIssueShortName.addTableColumn(0,SWT.LEFT,80,false);
        Widgets.layout(widgetIssueShortName,0,1,TableLayoutData.W,0,0,0,0,120,SWT.DEFAULT);

        widgetIssueSummary = new TreeCombo(subComposite,SWT.LEFT,SWT.NONE,100,TreeCombo.DEFAULT_HEIGHT);
        widgetIssueSummary.addTreeColumn("#",    SWT.RIGHT, 80,false);
        widgetIssueSummary.addTreeColumn("Name", SWT.LEFT, 120,false);
        widgetIssueSummary.addTreeColumn("Issue",SWT.LEFT,   0,true );
        Widgets.layout(widgetIssueSummary,0,2,TableLayoutData.WE);

        widgetIssueDoneRatio = Widgets.newOptionMenu(subComposite);
        widgetIssueDoneRatio.setToolTipText("Done percentage");
//TODO
//        widgetIssueDoneRatio.setEnabled(Backend.isEditProgressSupported());
        Widgets.layout(widgetIssueDoneRatio,0,3,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);
        Widgets.setOptionMenuItems(widgetIssueDoneRatio,
                                   new Object[]{"0%" ,   0,
                                                "10%",  10,
                                                "20%",  20,
                                                "30%",  30,
                                                "40%",  40,
                                                "50%",  50,
                                                "60%",  60,
                                                "70%",  70,
                                                "80%",  80,
                                                "90%",  90,
                                                "100%",100
                                               }
                                  );

        label = Widgets.newLabel(subComposite,"Filter:");
        Widgets.layout(label,0,4,TableLayoutData.W);

        widgetIssueStatusOwnButton = Widgets.newCheckbox(subComposite,"Own");
        widgetIssueStatusOwnButton.setSelection(Settings.issueOwnFilter);
        widgetIssueStatusOwnButton.setToolTipText("Show own issues only");
        Widgets.layout(widgetIssueStatusOwnButton,0,5,TableLayoutData.E);
        widgetIssueStatusOwnButton.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button widget = (Button)selectionEvent.widget;

            Settings.issueOwnFilter = widget.getSelection();
            Widgets.notify(shell,USER_EVENT_UPDATE_OWN_FILTER);

            // get and update issues for current selected project
            Project project = widgetProject.getSelection();
            updateIssues(widgetIssueId,
                         widgetIssueShortName,
                         widgetIssueSummary,
                         project
                        );
          }
        });
        widgetIssueStatusOwnButton.addListener(USER_EVENT_UPDATE_OWN_FILTER,new Listener()
        {
          @Override
          public void handleEvent(Event event)
          {
            widgetIssueStatusOwnButton.setSelection(Settings.issueOwnFilter);
          }
        });

        widgetIssueStatusFilterButton = Widgets.newButton(subComposite,IMAGE_FILTER);
        Widgets.layout(widgetIssueStatusFilterButton,0,6,TableLayoutData.E);

        widgetIssueStatusFilterMenu = Widgets.newPopupMenu(dialog);
        {
          menuItem = Widgets.addMenuItemCheckbox(widgetIssueStatusFilterMenu,"Any");
          menuItem.setSelection(true);
          menuItem.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              // disable all other filters
              MenuItem[] menuItems = widgetIssueStatusFilterMenu.getItems();
              if (menuItems[MENU_ITEM_INDEX_ANY].getSelection())
              {
                for (int i = MENU_ITEM_INDEX_FILTER_START; i < menuItems.length; i++)
                {
                  menuItems[i].setSelection(false);
                }
                Settings.issueStatusIdFilterSet.clear();
              }
              else
              {
                boolean selectedSome = false;
                int i = MENU_ITEM_INDEX_FILTER_START;
                while ((i < menuItems.length) && !selectedSome)
                {
                  selectedSome = menuItems[i].getSelection();
                }
                menuItems[MENU_ITEM_INDEX_ANY].setSelection(!selectedSome);
              }

              // get and update issues for current selected project
              Project project = widgetProject.getSelection();
              updateIssues(widgetIssueId,
                           widgetIssueShortName,
                           widgetIssueSummary,
                           project
                          );
            }
          });
        }
        widgetIssueStatusFilterButton.setMenu(widgetIssueStatusFilterMenu);
      }

      label = Widgets.newLabel(composite,"Time spent:",SWT.NONE,Settings.keySpent);
      Widgets.layout(label,3,0,TableLayoutData.W);

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,1.0}));
      Widgets.layout(subComposite,3,1,TableLayoutData.WE);
      {
        subSubComposite = Widgets.newComposite(subComposite);
        subSubComposite.setLayout(new TableLayout(0.0,0.0));
        Widgets.layout(subSubComposite,0,0,TableLayoutData.WE);
        {
          widgetSpentOn     = new CDateTime(subSubComposite,CDT.BORDER|CDT.COMPACT|CDT.DROP_DOWN|CDT.DATE_LONG|CDT.TIME_MEDIUM);
          widgetSpentOn.setToolTipText("Spent-on date.");
          widgetSpentOn.setPattern(Settings.dateFormat);
          widgetSpentOn.setSelection(timeEntry.spentOn.getDate());
          Widgets.layout(widgetSpentOn,0,0,TableLayoutData.W,0,0,0,4);
          widgetSpentOn.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              CDateTime widget = (CDateTime)selectionEvent.widget;
              widget.setOpen(false);
            }
          });

          widgetSpentHourFraction = Widgets.newSpinner(subSubComposite,SWT.NONE,0,Integer.MAX_VALUE);
          widgetSpentHourFraction.setToolTipText("Spent time duration.");
          widgetSpentHourFraction.setTextLimit(2);
          widgetSpentHourFraction.setIncrement(1);
          widgetSpentHourFraction.setSelection(timeEntry.getHourFraction());
          Widgets.layout(widgetSpentHourFraction,0,1,TableLayoutData.WE,0,0,4,0,90,SWT.DEFAULT);

          label = Widgets.newLabel(subSubComposite,"h");
          Widgets.layout(label,0,2,TableLayoutData.W);

          widgetSpentMinuteFraction = Widgets.newSpinner(subSubComposite,SWT.NONE,-60,60);
          widgetSpentMinuteFraction.setToolTipText("Spent time minutes.");
          widgetSpentMinuteFraction.setTextLimit(2);
          widgetSpentMinuteFraction.setIncrement(Settings.minTimeDelta);
          widgetSpentMinuteFraction.setSelection(timeEntry.getMinuteFraction());
          Widgets.layout(widgetSpentMinuteFraction,0,3,TableLayoutData.WE,0,0,4,0,90,SWT.DEFAULT);

          label = Widgets.newLabel(subSubComposite,"min");
          Widgets.layout(label,0,4,TableLayoutData.W);

          combo = Widgets.newOptionMenu(subSubComposite);
          Widgets.layout(combo,0,5,TableLayoutData.W);
          Widgets.addOptionMenuItem(combo,new Long(0L),"");
          for (int i = 1; i <= (8*60)/Settings.minTimeDelta; i++)
          {
            Widgets.addOptionMenuItem(combo,new Long(i*Settings.minTimeDelta),formatHours(i*Settings.minTimeDelta));
          }
          combo.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              Combo widget = (Combo)selectionEvent.widget;
              long  value  = (long)Widgets.getSelectedOptionMenuItem(widget,0L);
              if (value > 0L)
              {
                int hourFraction   = (int)(value/60);
                int minuteFraction = (int)(value%60);

                widgetSpentHourFraction.setSelection(hourFraction);
                widgetSpentMinuteFraction.setSelection(minuteFraction);
              }

              Widgets.setSelectedOptionMenuItem(widget,0L);
            }
          });
        }

        widgetActivitiesTitle = Widgets.newLabel(subComposite,"Activity:");
        widgetActivitiesTitle.setEnabled(timeEntry.isActivitySupported());
        Widgets.layout(widgetActivitiesTitle,0,1,TableLayoutData.W);
        widgetActivities = Widgets.newOptionMenu(subComposite);
        widgetActivities.setEnabled(timeEntry.isActivitySupported());
        Widgets.layout(widgetActivities,0,2,TableLayoutData.WE);

//TODO: required?
Dprintf.dprintf("xxxxxxxxxxxxxxxxxxxxxxxxxxx");
        Event event = new Event();
        event.widget = widgetActivities;
        widgetActivities.notifyListeners(SWT.Modify,event);
      }

      label = Widgets.newLabel(composite,"Comments:",SWT.NONE,Settings.keyComments);
      Widgets.layout(label,4,0,TableLayoutData.W);

      widgetComments = Widgets.newCombo(composite);
      widgetComments.setToolTipText("Time entry comment line.");
Dprintf.dprintf("%s",timeEntry.comments);
      widgetComments.setText(timeEntry.comments);
      Widgets.layout(widgetComments,4,1,TableLayoutData.WE);
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetSave = Widgets.newButton(composite,okText,Settings.keySave);
      Widgets.layout(widgetSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetSave.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Project  project  = widgetProject.getSelection();
          Issue    issue    = widgetIssueSummary.getSelection();
          Activity activity = widgetActivities.isEnabled() ? Widgets.getSelectedOptionMenuItem(widgetActivities,Activity.class) : null;
          long     duration = Backend.toDuration(widgetSpentHourFraction.getSelection(),widgetSpentMinuteFraction.getSelection());
          String   comments = widgetComments.getText().trim();

          if (project == null)
          {
            Dialogs.error(shell,"Please select a project for the time entry.");
            Widgets.setFocus(widgetProject);
            return;
          }
          if (project.id.isNone())
          {
//TOOD: force sync?
            Dialogs.error(shell,"Project still not synchronized - cannot "+actionText+" time entry.");
            Widgets.setFocus(widgetSave);
            return;
          }
          if (issue == null)
          {
            Dialogs.error(shell,"Please select an issue for the time entry.");
            Widgets.setFocus(widgetIssueSummary);
            return;
          }
          if (issue.id.isNone())
          {
//TOOD: force sync?
            Dialogs.error(shell,"Issue still not synchronized - cannot "+actionText+" time entry.");
            Widgets.setFocus(widgetSave);
            return;
          }
          if (duration <= 0)
          {
            Dialogs.error(shell,"Please select some duration for the time entry.");
            Widgets.setFocus(widgetSpentMinuteFraction);
            return;
          }
          if (widgetActivities.isEnabled() && (activity == null))
          {
            Dialogs.error(shell,"Please select an activity for the time entry.");
            Widgets.setFocus(widgetActivities);
            return;
          }

          timeEntry.projectId = project.id;
          timeEntry.issueId   = issue.id;
//Dprintf.dprintf("prevIssueId=%s",prevIssueId);
//Dprintf.dprintf("issue.id=%s",issue.id);
          if (!prevIssueId.equals(issue.id))
          {
// TODO: set new issue?
            timeEntry.issueId = issue.id;
//Dprintf.dprintf("timeEntry.newIssueId=%s",timeEntry.newIssueId);
          }
          timeEntry.spentOn    = new SpentOn(widgetSpentOn.getSelection());
          timeEntry.duration   = duration;
          timeEntry.activityId = widgetActivities.isEnabled() ? activity.id : new Id();
          timeEntry.comments   = comments;

          if (timeEntry.spentOn.isAfter(SpentOn.today()) && Settings.showFutureTimeEntryWarning)
          {
            if (!Dialogs.confirm(dialog,
                                 Dialogs.booleanFieldUpdater(Settings.class,"showFutureTimeEntryWarning"),
                                 "Time spent "+timeEntry.spentOn.toString()+" is in future.\nReally use this time?"
                                )
               )
            {
              Widgets.setFocus(widgetSpentOn);
              return;
            }
          }

          Settings.geometryEditTimeEntry = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // update history, projects, issues, status, activities
    try
    {
      TimeEntryList        historyTimeEntryList = Backend.getTimeEntryHistory(MAX_HISTORY_SIZE, Backend.HistoryFilterTypes.COMMENTS);
      ProjectList          projectList          = Backend.getProjectList();
      Project              project              = timeEntry.getProject             ();
      Issue                issue                = timeEntry.getIssue               ();
      boolean              activitySupported    = timeEntry.isActivitySupported    ();
      StatusTransitionList statusTransitionList = timeEntry.getStatusTransitionList();
      ActivityList         activityList         = timeEntry.getActivityList        ();
      Activity             activity             = timeEntry.getActivity            ();

      updateHistory(widgetHistory,historyTimeEntryList);
      updateProjects(widgetProject,projectList,project);
      updateIssues(widgetIssueId,widgetIssueShortName,widgetIssueSummary,project,issue);
      updateIssueStatusFilterMenu(widgetProject,widgetIssueShortName,widgetIssueId,widgetIssueSummary,widgetIssueStatusFilterMenu,statusTransitionList);
      updateActivities(widgetActivitiesTitle,widgetActivities,activitySupported,activityList,activity);
      updateHistory(widgetComments,historyTimeEntryList);
    }
    catch (BackendException exception)
    {
      printStackTrace(exception);
      Dialogs.error(shell,"Cannot get data from server (error: "+exception.getMessage()+")");
      return false;
    }

    // add listeners
    widgetFind.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        Text widget = (Text)selectionEvent.widget;
Dprintf.dprintf("");
//          updateStorageTreeTableThread.triggerUpdateStorageName(widget.getText());
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Text widget = (Text)selectionEvent.widget;
Dprintf.dprintf("");
//          updateStorageTreeTableThread.triggerUpdateStorageName(widget.getText());
      }
    });
    widgetFind.addKeyListener(new KeyListener()
    {
      @Override
      public void keyPressed(KeyEvent keyEvent)
      {
      }
      @Override
      public void keyReleased(KeyEvent keyEvent)
      {
        Text widget = (Text)keyEvent.widget;

        if      (Widgets.isAccelerator(keyEvent,Settings.keyProject))
        {
          Widgets.setFocus(widgetProject);
          return;
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyIssue))
        {
          Widgets.setFocus(widgetIssueSummary);
          return;
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keySpent))
        {
          Widgets.setFocus(widgetSpentHourFraction);
          return;
        }
        else if ((widgetActivities != null) && (Widgets.isAccelerator(keyEvent,Settings.keyActivity)))
        {
          Widgets.setFocus(widgetActivities);
          return;
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyComments))
        {
          Widgets.setFocus(widgetComments);
          return;
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keySave))
        {
          Widgets.setFocus(widgetSave);
          return;
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'x'))
        {
          widget.setText("");
        }

        // filter projects/issues
        String filters[] = StringUtils.splitArray(widget.getText(),StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS);
        filterIssuesProjects(widgetProject,
                             widgetIssueShortName,
                             widgetIssueId,
                             widgetIssueSummary,
                             widgetActivitiesTitle,
                             widgetActivities,
                             filters
                            );
      }
    });
    widgetHistory.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TableCombo widget    = (TableCombo)selectionEvent.widget;
        TimeEntry  timeEntry = widget.getSelection(TimeEntry.class);
        Project    project   = timeEntry.getProject();
        Issue      issue     = timeEntry.getIssue();

        if (project != null)
        {
          setSelectedProject(widgetProject,
                             widgetIssueId,
                             widgetIssueShortName,
                             widgetIssueSummary,
                             widgetIssueDoneRatio,
                             widgetActivitiesTitle,
                             widgetActivities,
                             project
                            );
        }
        if (project != null)
        {
          setSelectedIssue(widgetIssueId,
                           widgetIssueShortName,
                           widgetIssueSummary,
                           widgetIssueDoneRatio,
                           issue
                          );
        }
        widgetSpentHourFraction.setSelection((int)(timeEntry.duration/60));
        widgetSpentMinuteFraction.setSelection((int)(timeEntry.duration%60));
        widgetComments.setText(timeEntry.comments);
      }
    });
    widgetProject.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TreeCombo widget = (TreeCombo)selectionEvent.widget;

        Project project = widget.getSelection();
        if (project != null)
        {
          setSelectedProject(widgetProject,
                             widgetIssueId,
                             widgetIssueShortName,
                             widgetIssueSummary,
                             widgetActivitiesTitle,
                             widgetActivities,
                             project
                            );
        }
      }
    });
    widgetIssueId.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TableCombo widget = (TableCombo)selectionEvent.widget;

        Issue issue = widget.getSelection(Issue.NONE);
        if (issue != Issue.NONE)
        {
          setSelectedIssue(widgetIssueId,widgetIssueShortName,widgetIssueSummary,issue);
        }
      }
    });
    widgetIssueShortName.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TableCombo widget = (TableCombo)selectionEvent.widget;

        Issue issue = widget.getSelection(Issue.NONE);
        if (issue != Issue.NONE)
        {
          setSelectedIssue(widgetIssueId,widgetIssueShortName,widgetIssueSummary,issue);
        }
      }
    });
    widgetIssueSummary.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TreeCombo widget = (TreeCombo)selectionEvent.widget;

        Issue issue = widget.getSelection();
        if (issue != null)
        {
          setSelectedIssue(widgetIssueId,widgetIssueShortName,widgetIssueSummary,issue);
Dprintf.dprintf("kkkkkkkkkkkkkkkkkkkkkkkkk");
//          Widgets.setSelectedOptionMenuItem(widgetIssueId,(Integer)issueId);
        }
      }
    });
    widgetIssueDoneRatio.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Combo       widget = (Combo)selectionEvent.widget;
        int         index  = widget.getSelectionIndex();
        final Issue issue  = widgetIssueSummary.getSelection();

        if ((index >= 0) && (issue != null))
        {
          int doneRatio = (Integer)Widgets.getComboItem(widget,index);

          // get issue
//TODO: remove
//              final Issue issue = Backend.getIssue(issueId);

          try
          {
            // set done ratio
            issue.setDoneRatio(doneRatio);

            // set status
            if      (issue.doneRatio >= 100)
            {
              if (!Settings.issueCompletedStatus.isEmpty())
              {
                Id completedStatusId = Backend.getStatusId(Settings.issueCompletedStatus);
                if (!completedStatusId.isNone() && (issue.statusId != completedStatusId))
                {
                  switch (Dialogs.select(shell,
                                         "Select issue status",
                                         String.format("The issue does not have completed status '%s'. Change status?",Settings.issueCompletedStatus),
                                         new String[]{"Change","Keep","Cancel"},
                                         0
                                        )
                         )
                  {
                    case 0:
                      issue.setStatusTransitionId(completedStatusId);
                      break;
                    case 1:
                      break;
                    case 2:
                      Widgets.setFocus(widgetComments);
                      return;
                  }
                }
              }
            }
            else if (issue.doneRatio > 0)
            {
              if (!Settings.issueUpdateStatus.isEmpty())
              {
                Id updateStatusId = Backend.getStatusId(Settings.issueUpdateStatus);
                if (issue.statusId != updateStatusId)
                {
                  switch (Dialogs.select(shell,
                                         "Select issue status",
                                         String.format("The issue does not have update status '%s'. Change status?",Settings.issueUpdateStatus),
                                         new String[]{"Change","Keep","Cancel"},
                                         0
                                        )
                         )
                  {
                    case 0:
                      issue.setStatusTransitionId(updateStatusId);
                      break;
                    case 1:
                      break;
                    case 2:
                      Widgets.setFocus(widgetComments);
                      return;
                  }
                }
              }
            }

            // save changes
            issue.setModified(Backend.ModifiedStates.UPDATED);
            Widgets.notify(shell,USER_EVENT_UPDATE_ISSUE,issue);
          }
          catch (BackendException exception)
          {
            printStackTrace(exception);
            Dialogs.error(shell,"Cannot set issue done ratio (error: "+exception.getMessage()+")");
          }
          catch (InterruptedException exception)
          {
            // ignored
          }
        }
      }
    });
    widgetIssueStatusFilterButton.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        //widgetIssueStatusFilterMenu.setLocation(point.x, point.y);
        widgetIssueStatusFilterMenu.setVisible(true);
      }
    });
    widgetSpentMinuteFraction.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        int hourFraction   = widgetSpentHourFraction.getSelection();
        int minuteFraction = widgetSpentMinuteFraction.getSelection();

        // Note: sometimes the minute value is not a multiple of the increment. Correct this.
        if (minuteFraction >= 0)
        {
          minuteFraction = ((minuteFraction+Settings.minTimeDelta-1)/Settings.minTimeDelta)*Settings.minTimeDelta;
        }
        else
        {
          minuteFraction = -((((-minuteFraction)+Settings.minTimeDelta-1)/Settings.minTimeDelta)*Settings.minTimeDelta);
        }

        if      (minuteFraction >= 60)
        {
          // increment hours
          hourFraction   += minuteFraction/60;
          minuteFraction %= 60;
        }
        else if (minuteFraction < 0)
        {
          if (hourFraction > 0)
          {
            // decrement hours, reset
            hourFraction   -= ((-minuteFraction)+60-1)/60;
            minuteFraction = minuteFraction+((((-minuteFraction)+60-1)/60)*60);
          }
          else
          {
            // reset
            minuteFraction = 0;
          }
        }

        widgetSpentHourFraction.setSelection(hourFraction);
        widgetSpentMinuteFraction.setSelection(minuteFraction);
      }
    });
    widgetComments.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        Combo widget = (Combo)keyEvent.widget;

        if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'c'))
        {
          Widgets.setClipboard(clipboard,widget.getText());
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    });

    // add shortcut listener
    Listener keyListener = new Listener()
    {
      @Override
      public void handleEvent(Event event)
      {
        if (Widgets.isChildOf(dialog,event.widget))
        {
          if      (Widgets.isAccelerator(event,Settings.keyFind))
          {
            Widgets.setFocus(widgetFind);
            event.doit = false;
          }
          else if (Widgets.isAccelerator(event,Settings.keyProject))
          {
            Widgets.setFocus(widgetProject);
            event.doit = false;
          }
          else if (Widgets.isAccelerator(event,Settings.keyIssue))
          {
            Widgets.setFocus(widgetIssueSummary);
            event.doit = false;
          }
          else if (Widgets.isAccelerator(event,Settings.keySpent))
          {
            Widgets.setFocus(widgetSpentOn);
            event.doit = false;
          }
          else if ((widgetActivities != null) && (Widgets.isAccelerator(event,Settings.keyActivity)))
          {
            Widgets.setFocus(widgetActivities);
            event.doit = false;
          }
          else if (Widgets.isAccelerator(event,Settings.keyComments))
          {
            Widgets.setFocus(widgetComments);
            event.doit = false;
          }
          else if (Widgets.isAccelerator(event,Settings.keySave))
          {
            Widgets.invoke(widgetSave);
            event.doit = false;
          }
        }
      }
    };
    display.addFilter(SWT.KeyDown,keyListener);

    // set next focus
    Widgets.setNextFocus(widgetProject,
                         widgetIssueId,
                         widgetIssueSummary,
                         widgetIssueStatusOwnButton,
                         widgetIssueStatusFilterButton,
  //                       (spentOnList != null) ? widgetSpentOnList : widgetSpentOn,
                         widgetSpentHourFraction,
                         widgetSpentMinuteFraction,
                         widgetActivities,
                         widgetComments,
                         widgetSave
                        );

    // show dialog
    Dialogs.show(dialog,Settings.geometryEditTimeEntry);

    // run
    if      (timeEntry.projectId.isNone() ) Widgets.setFocus(widgetProject);
    else if (timeEntry.issueId.isNone()   ) Widgets.setFocus(widgetIssueSummary);
    else                                    Widgets.setFocus(widgetSpentMinuteFraction);
    boolean result = (Boolean)Dialogs.run(dialog,false);

    // remove listeners, free resources
    display.removeFilter(SWT.KeyDown,keyListener);

    return result;
  }

  /** add new time entries
   * @param backendTypes backend type
   * @param projectId project id or Id.NONE
   * @param spentOnList spent-on date list or null
   */
  private void addTimeEntry(Backend.Types backendType,
                            Id            projectId,
                            List<SpentOn> spentOnList
                           )
    throws BackendException
  {
    assert spentOnList.size() >= 1;

    if (spentOnList.size() > 1)
    {
      // create new time entry
      TimeEntry timeEntry = new TimeEntry(backendType,
                                          new Id(),
                                          projectId,
                                          new Id(),
                                          Backend.getDefaultActivityId(Backend.TypeSet.NONE),
                                          Settings.minTimeDelta,
                                          "",
                                          (SpentOn)null
                                         );
      if (editTimeEntry(timeEntry,"Add new time entry","add","Add"))
      {
        try
        {
          timeEntry.setModified(Backend.ModifiedStates.ADDED);
          Widgets.notify(shell,USER_EVENT_ADD_TIME_ENTRY,timeEntry);
        }
        catch (InterruptedException exception)
        {
          // ignored
        }
      }
    }
    else
    {
      addTimeEntry(backendType,projectId,spentOnList.get(0));
    }
  }

  /** add new time entries
   * @param project project
   * @param spentOnList spent-on date list or null
   */
  private void addTimeEntry(Project       project,
                            List<SpentOn> spentOnList
                           )
    throws BackendException
  {
    addTimeEntry(project.backendType,project.id,spentOnList);
  }

  /** add new time entry
   * @param backendTypes backend type
   * @param projectId project id or Id.NONE
   * @param duration duration
   * @param spentOn spent-on
   */
  private void addTimeEntry(Backend.Types backendType,
                            Id            projectId,
                            long          duration,
                            SpentOn       spentOn
                           )
    throws BackendException
  {
    // create new time entry
    TimeEntry timeEntry = new TimeEntry(backendType,
                                        new Id(),
                                        projectId,
                                        new Id(),
                                        Backend.getDefaultActivityId(Backend.TypeSet.NONE),
                                        duration,
                                        "",
                                        spentOn
                                       );

    // special case: pre-select issue if only one issue in project
    IssueList issueList = timeEntry.getIssueList(true);
    if (issueList.size() == 1)
    {
      timeEntry.issueId = issueList.get(0).id;
    }

    // edit time entry
    if (editTimeEntry(timeEntry,"Add new time entry","add","Add"))
    {
      try
      {
        timeEntry.setModified(Backend.ModifiedStates.ADDED);
        Widgets.notify(shell,USER_EVENT_ADD_TIME_ENTRY,timeEntry);
      }
      catch (InterruptedException exception)
      {
        // ignored
      }
    }
  }

  /** add new time entry
   * @param backendTypes backend type
   * @param projectId project id or Id.NONE
   * @param spentOn spent-on
   */
  private void addTimeEntry(Backend.Types backendType,
                            Id            projectId,
                            SpentOn       spentOn
                           )
    throws BackendException
  {
    addTimeEntry(backendType,projectId,Settings.minTimeDelta,spentOn);
  }

  /** add new time entry
   * @param project project
   * @param spentOn spent-on
   */
  private void addTimeEntry(Project project, SpentOn spentOn)
    throws BackendException
  {
    addTimeEntry(project.backendType,project.id,spentOn);
  }

  /** add new time entry
   * @param spentOn spent-on
   */
  private void addTimeEntry(List<SpentOn> spentOnList)
    throws BackendException
  {
    addTimeEntry(Backend.Types.NONE,new Id(),spentOnList);
  }

  /** add new time entry
   * @param spentOn spent-on
   */
  private void addTimeEntry(SpentOn spentOn)
    throws BackendException
  {
    addTimeEntry(Backend.Types.NONE,new Id(),spentOn);
  }

  /** add new time entry
   * @param duration duration
   */
  private void addTimeEntry(long duration)
    throws BackendException
  {
    addTimeEntry(Backend.Types.NONE,new Id(),duration,new SpentOn());
  }

  /** remove time entry
   * @param timeEntry time entry to remove
   */
  private void removeTimeEntry(TimeEntry timeEntry)
    throws BackendException
  {
    try
    {
      timeEntry.setModified(Backend.ModifiedStates.DELETED);
      Widgets.notify(shell,USER_EVENT_REMOVE_TIME_ENTRY,timeEntry);
    }
    catch (InterruptedException exception)
    {
      // ignored
    }
  }

  /** edit done ratio
   * @param doneRatio done ration
   * @return new done ratio or -1
   */
  private int editDoneRatio(int doneRatio)
  {
    /** dialog data
     */
    class Data
    {
      int doneRatio;

      Data()
      {
        doneRatio = 0;
      }
    };

    Composite composite,subComposite;
    Label     label;
    Button    button;

    final Data  data   = new Data();
    final Shell dialog = Dialogs.openModal(shell,"Select issue progress",300,SWT.DEFAULT,new double[]{0.0,1.0},1.0);

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{0.0,1.0},new double[]{0.0,1.0},2));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE);
    {
      label = Widgets.newLabel(composite,"Issue progress:");
      Widgets.layout(label,0,0,TableLayoutData.W);

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(0.0,1.0));
      Widgets.layout(subComposite,0,1,TableLayoutData.WE);
      {
        Color doneBackgroundColor = display.getSystemColor(SWT.COLOR_GREEN);
        int column = 0;
        for (int newDoneRatio = 0; newDoneRatio <= 100; newDoneRatio += 10)
        {
          button = Widgets.newButton(subComposite,String.format("%d",newDoneRatio));
          button.setData(newDoneRatio);
          if (doneRatio >= newDoneRatio) button.setBackground(doneBackgroundColor);
          Widgets.layout(button,0,column,TableLayoutData.WE,0,0,0,0,40,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            @Override
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            @Override
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              Button widget = (Button)selectionEvent.widget;

              data.doneRatio = (Integer)widget.getData();
              Dialogs.close(dialog,true);
            }
          });
          column++;
        }
      }

      button = Widgets.newCheckbox(composite,"show again");
      button.setSelection(Settings.enterIssueProgress);
      Widgets.layout(button,1,0,TableLayoutData.E,0,0,0,2);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          Settings.enterIssueProgress = widget.getSelection();
        }
      });
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(1.0,1.0,2));
    Widgets.layout(composite,1,0,TableLayoutData.NSWE);
    {
      button = Widgets.newButton(composite,"Skip");
      Widgets.layout(button,0,0,TableLayoutData.NSWE);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // add listeners

    // set next focus
//    Widgets.setNextFocus(widgetServerName,widgetServerPort,widgetServerUseSSL,widgetLoginName,widgetLoginPassword,widgetLoginButton);

    // run
    Boolean result = (Boolean)Dialogs.run(dialog);

    return ((result != null) && result.booleanValue()) ? data.doneRatio : doneRatio;
  }

  /** edit bookmark
   * @param bookmark bookmark to edit
   * @param issue issue
   * @return true if bookmark edited, false otherwise
   */
  private boolean editBookmark(final Settings.Bookmark bookmark, String title, String okText)
  {
    final Shell dialog;
    Composite   composite,subComposite;
    Label       label;
    Combo       combo;
    Button      button;
    MenuItem    menuItem;

    // repository edit dialog
    dialog = Dialogs.openModal(shell,title,400,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Label  widgetProject;
    final Label  widgetIssue;
    final Text   widgetText;
    final Button widgetSave;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Project:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      widgetProject = Widgets.newView(composite);
      Widgets.layout(widgetProject,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Issue:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      widgetIssue = Widgets.newView(composite);
      Widgets.layout(widgetIssue,1,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Summary:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      widgetText = Widgets.newText(composite);
      widgetText.setToolTipText("Summary text of issue.");
      Widgets.layout(widgetText,2,1,TableLayoutData.WE);
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetSave = Widgets.newButton(composite,okText);
      Widgets.layout(widgetSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetSave.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          String text = widgetText.getText().trim();

          bookmark.text = text;

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // try to set issue, project, fill-in text
    try
    {
      Issue   issue   = Backend.getIssue(new Backend.TypeSet(bookmark.backendType),bookmark.summary);
      Project project = (issue != null) ? issue.getProject() : null;

      if (project != null) widgetProject.setText(project.name);
      if (issue != null) widgetIssue.setText(issue.summary);

Dprintf.dprintf("%s",bookmark.text);

      if (bookmark.text.isEmpty()) bookmark.text = issue.summary;
    }
    catch (BackendException exception)
    {
      // ignored
    }

    // add listeners

    // set next focus
    Widgets.setNextFocus(widgetText,
                         widgetSave
                        );

    // show dialog
    Dialogs.show(dialog);

    // run
    Widgets.setFocus(widgetText);
    return (Boolean)Dialogs.run(dialog,false);
  }

  /** edit bookmark
   * @param issue issue
   * @param bookmark bookmark to edit
   * @return true if bookmark edited, false otherwise
   */
  private void editBookmarks()
  {
    final Shell dialog;
    Composite   composite,subComposite;
    Label       label;
    Combo       combo;
    Button      button;
    MenuItem    menuItem;

    // repository edit dialog
    dialog = Dialogs.openModal(shell,"Edit bookmarks",600,400,new double[]{1.0,0.0},1.0);

    final Table  widgetIssueBookmarkTable;
    final Button widgetClose;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0,4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      widgetIssueBookmarkTable = Widgets.newTable(composite,SWT.MULTI);
      widgetIssueBookmarkTable.setToolTipText("ENTER/RETURN to edit entry.\nDEL/BACKSPACE to delete entries.");
      widgetIssueBookmarkTable.setLayout(new TableLayout(null,1.0));
      Widgets.layout(widgetIssueBookmarkTable,0,0,TableLayoutData.NSWE);
      Widgets.addTableColumn(widgetIssueBookmarkTable,0,"Project",SWT.LEFT, 80,true );
      Widgets.addTableColumn(widgetIssueBookmarkTable,1,"Issue",SWT.LEFT, 80,true );
      Widgets.addTableColumn(widgetIssueBookmarkTable,2,"Text",SWT.LEFT, 80,true );

      subComposite = Widgets.newComposite(composite,SWT.NONE,4);
      Widgets.layout(subComposite,1,0,TableLayoutData.W);
      {
        button = Widgets.newButton(subComposite,"Edit"+"\u2026");
        Widgets.layout(button,0,0,TableLayoutData.DEFAULT,0,0,0,0,110,SWT.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
// TODO:
Dprintf.dprintf("");
            int index = widgetIssueBookmarkTable.getSelectionIndex();
            if (index >= 0)
            {
              TableItem tableItem = widgetIssueBookmarkTable.getItem(index);
              if (!tableItem.isDisposed())
              {
                Settings.Bookmark bookmark = (Settings.Bookmark)tableItem.getData();
                if (editBookmark(bookmark,"Edit bookmark","Save"))
                {
Dprintf.dprintf("_");

                }
              }
            }
          }
        });

        button = Widgets.newButton(subComposite,"Delete"+"\u2026");
        Widgets.layout(button,0,1,TableLayoutData.DEFAULT,0,0,0,0,110,SWT.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            int index = widgetIssueBookmarkTable.getSelectionIndex();
            if (index >= 0)
            {
              TableItem tableItem = widgetIssueBookmarkTable.getItem(index);
              if (!tableItem.isDisposed())
              {
                Settings.issueBookmarkList.remove((Settings.Bookmark)tableItem.getData());
                Widgets.removeTableItem(widgetIssueBookmarkTable,tableItem);
              }
            }
          }
        });
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetClose = Widgets.newButton(composite,"Close");
      Widgets.layout(widgetClose,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetClose.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // add listeners
    widgetIssueBookmarkTable.addMouseListener(new MouseListener()
    {
      @Override
      public void mouseDoubleClick(MouseEvent mouseEvent)
      {
        Table widget = (Table)mouseEvent.widget;

        int index = widget.getSelectionIndex();
        if (index >= 0)
        {
          TableItem tableItem = widgetIssueBookmarkTable.getItem(index);
          if (!tableItem.isDisposed())
          {
            Settings.Bookmark bookmark = (Settings.Bookmark)tableItem.getData();
            if (editBookmark(bookmark,"Edit bookmark","Save"))
            {
Dprintf.dprintf("_");
              Settings.setModified();
              updateBookmarksMenu(widgetBookmarksMenu,Settings.issueBookmarkList);
            }
          }
        }
      }
      @Override
      public void mouseDown(MouseEvent mouseEvent)
      {
      }
      @Override
      public void mouseUp(MouseEvent mouseEvent)
      {
      }
    });

    // set next focus
    Widgets.setNextFocus(widgetIssueBookmarkTable,
                         widgetClose
                        );

    // show dialog
    Dialogs.show(dialog);

    // insert bookmarks
    sortBookmarks(Settings.issueBookmarkList);
    for (Settings.Bookmark bookmark : Settings.issueBookmarkList)
    {
      try
      {
        Issue   issue   = Backend.getIssue(new Backend.TypeSet(bookmark.backendType),bookmark.summary);
        Project project = (issue != null) ? issue.getProject() : null;

        if ((issue != null) && (project != null))
        {
          TableItem tableItem = Widgets.addTableItem(widgetIssueBookmarkTable,
                                                     bookmark,
                                                     (project != null) ? project.name : "",
                                                     (issue != null) ? issue.shortName : "",
                                                     (issue != null) ? issue.summary : "",
                                                     bookmark.text
                                                    );
          tableItem.setData(bookmark);
        }
      }
      catch (BackendException exception)
      {
        // ignored
      }
    }

    // run
    Widgets.setFocus(widgetIssueBookmarkTable);
    Dialogs.run(dialog);
  }

  /** edit tag
   * @param tag tag to edit
   * @param title window title text
   * @param okText ok-button text
   * @return true iff edited
   */
  private boolean editTag(final Tag tag, String title, String okText)
  {
    Composite composite,subComposite;
    Label     label;
    Button    button;

    final Shell  dialog = Dialogs.openModal(shell,title,300,SWT.DEFAULT,new double[]{0.0,1.0},1.0);

    final Combo  widgetTagName;
    final Button widgetOk;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0},2));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE);
    {
      label = Widgets.newLabel(composite,"Time:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      label = Widgets.newLabel(composite,TIME_FORMAT.format(tag.startTimestamp*1000L)+".."+TIME_FORMAT.format(tag.endTimestamp*1000L));
      Widgets.layout(label,0,1,TableLayoutData.W);

      label = Widgets.newLabel(composite,"Duration:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      label = Widgets.newLabel(composite,String.format("%02d:%02d",(tag.endTimestamp-tag.startTimestamp)/(60*60),((tag.endTimestamp-tag.startTimestamp+60-1)%(60*60))/60));
      Widgets.layout(label,1,1,TableLayoutData.W);

      label = Widgets.newLabel(composite,"Name:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      widgetTagName = Widgets.newCombo(composite);
      widgetTagName.setText(tag.name);
      for (String tagName : ActionMonitor.getInstance().getTagNames())
      {
        widgetTagName.add(tagName);
      }
      Widgets.layout(widgetTagName,2,1,TableLayoutData.WE);
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(1.0,1.0,2));
    Widgets.layout(composite,1,0,TableLayoutData.NSWE);
    {
      widgetOk = Widgets.newButton(composite,okText);
      Widgets.layout(widgetOk,0,0,TableLayoutData.W,0,0,0,0,120,SWT.DEFAULT);
      widgetOk.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          tag.name = widgetTagName.getText();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,120,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // add listeners

    // set next focus
    Widgets.setNextFocus(widgetTagName,
                         widgetOk
                        );

    // run
    return (Boolean)Dialogs.run(dialog,false);
  }

  /** add holiday day
   * @param spentOn spent-on date
   */
  private void addHoliday(SpentOn spentOn)
  {
    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = database.prepareStatement("INSERT OR IGNORE INTO days (dateTime) VALUES (DATE(?,'unixepoch','localtime'))");
      preparedStatement.setLong(1,spentOn.getTime()/1000);
      preparedStatement.executeUpdate();
      preparedStatement.close(); preparedStatement = null;
    }
    catch (SQLException exception)
    {
      printStackTrace(exception);
    }
  }

  /** remove holiday date
   * @param spentOn spent-on date
   */
  private void removeHolidayDate(SpentOn spentOn)
  {
    PreparedStatement preparedStatement = null;
    try
    {
      preparedStatement = database.prepareStatement("DELETE FROM days WHERE STRFTIME('%Y-%m-%d',dateTime)=STRFTIME('%Y-%m-%d',DATE(?,'unixepoch','localtime'))");
      preparedStatement.setLong(1,spentOn.getTime()/1000);
      preparedStatement.close(); preparedStatement = null;
    }
    catch (SQLException exception)
    {
      printStackTrace(exception);
    }
  }

  /** check if date is holiday date
   * @param date date
   */
  private boolean isHolidayDate(Date date)
  {
    boolean result = false;

    PreparedStatement preparedStatement = null;
    ResultSet         resultSet         = null;
    try
    {
      preparedStatement = database.prepareStatement("SELECT id FROM days WHERE STRFTIME('%Y-%m-%d',dateTime)=STRFTIME('%Y-%m-%d',DATE(?,'unixepoch','localtime')) LIMIT 0,1");
      preparedStatement.setLong(1,date.getTime()/1000);
      resultSet = preparedStatement.executeQuery();
      if (resultSet.next())
      {
        result = true;
      }
      resultSet.close(); resultSet = null;
      preparedStatement.close(); preparedStatement = null;
    }
    catch (SQLException exception)
    {
      printStackTrace(exception);
    }
//Dprintf.dprintf("isHolidayDate %s: %s",date,result);

    return result;
  }

  /** get holiday description
   * @param date date
   * @return description or null
   */
  private String getHolidayDescription(Date date)
  {
    String result = "";

    PreparedStatement preparedStatement = null;
    ResultSet         resultSet         = null;
    try
    {
      preparedStatement = database.prepareStatement("SELECT description FROM days WHERE STRFTIME('%Y-%m-%d',dateTime)=STRFTIME('%Y-%m-%d',DATE(?,'unixepoch','localtime')) LIMIT 0,1");
      preparedStatement.setLong(1,date.getTime()/1000);
      resultSet = preparedStatement.executeQuery();
      if (resultSet.next())
      {
        result = resultSet.getString(1);
        if (result == null) result = "";
      }
      resultSet.close(); resultSet = null;
      preparedStatement.close(); preparedStatement = null;
    }
    catch (SQLException exception)
    {
      printStackTrace(exception);
    }

    return result;
  }

  /** get holiday duration
   * @param spendOnStart,spendOnEnd spend on start/end date
   * @return holiday duration in specified range [min]
   */
  private long getHolidayDurationSum(SpentOn spentOnStart, SpentOn spentOnEnd)
  {
    long durationSum = 0L;

    int n = 0;
    while (spentOnStart.getDate(n).before(spentOnEnd.getDate()))
    {
      if (isHolidayDate(spentOnStart.getDate(n))) durationSum += (long)(Settings.workingHoursPerDay*60.0);
      n++;
    }

    return durationSum;
  }

  /** add vacation day
   * @param spentOn spent-on date
   */
  private boolean addVacation(SpentOn spentOn)
    throws BackendException
  {
    Project  vacationProject  = Backend.getProject(Settings.vacationProject);
    if (vacationProject == null)
    {
      Dialogs.error(shell,"Can not add vaction day: no vacation project");
      return false;
    }

    Issue    vacationIssue    = vacationProject.getIssue(Settings.vacationIssue);
    if (vacationIssue == null)
    {
      Dialogs.error(shell,"Can not add vaction day: no vacation issue");
      return false;
    }

    Activity vacationActivity = vacationIssue.getActivity(Settings.vacationActivity);

    TimeEntry timeEntry = new TimeEntry(vacationProject.backendType,
                                        new Id(),
                                        vacationProject.id,
                                        vacationIssue.id,
                                        (vacationActivity != null) ? vacationActivity.id : new Id(),
                                        (long)(Settings.workingHoursPerDay*60.0),
                                        Settings.vacationComment,
                                        spentOn
                                       );

    try
    {
      timeEntry.setModified(Backend.ModifiedStates.ADDED);
      Widgets.notify(shell,USER_EVENT_ADD_TIME_ENTRY,timeEntry);
    }
    catch (InterruptedException exception)
    {
      // ignored
    }

    return true;
  }

  /** add colors to colors list
   */
  private void addColors(Table widgetColors)
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor         = Settings.SettingValueAdapterColor.class.getDeclaredConstructor(Settings.class);
      SettingUtils.ValueAdapter settingValueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance(new Settings());

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.SettingValueAdapterColor.class.isAssignableFrom(configValue.type()))
              {
                // get color
                Settings.Color color = (Settings.Color)field.get(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // add entry
                color = color.clone();
                Widgets.addTableItem(widgetColors,color,name.substring(5));   // Note: remove prefix "Color"
                if (color.foreground != null) Widgets.setTableItemColor(widgetColors,color,1,new Color(null,color.foreground));
                if (color.background != null) Widgets.setTableItemColor(widgetColors,color,2,new Color(null,color.background));
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Timemine.printInternalError(exception);
    }
  }

  /** save colors
   * @param widgetColors color widget
   * @return
   */
  private void saveColors(Table widgetColors)
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor         = Settings.SettingValueAdapterColor.class.getDeclaredConstructor(Settings.class);
      SettingUtils.ValueAdapter settingValueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance(new Settings());

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.SettingValueAdapterColor.class.isAssignableFrom(configValue.type()))
              {
                // get color
                Settings.Color color = (Settings.Color)field.get(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // set color
                name = name.substring(5);
                boolean found = false;
                for (TableItem tableItem : widgetColors.getItems())
                {
                  if (name.equals(tableItem.getText(0)))
                  {
                    field.set(null,tableItem.getData());
                    found = true;
                    break;
                  }
                }
                if (!found)
                {
                  Timemine.printInternalError("Color %s not found in table!",name);
                }
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Timemine.printInternalError(exception);
    }
  }

  /** edit color
   * @param name name
   * @param color color
   * @return true if edit OK, false on cancel
   */
  private boolean editColor(String name, final Settings.Color color)
  {
    Composite composite,subComposite;
    Label     label;
    Text      text;
    Button    button;

    // add editor dialog
    final Shell dialog = Dialogs.openModal(shell,"Edit color",100,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Canvas widgetColorForeground;
    final Canvas widgetColorBackground;
    final Label  widgetValueForeground;
    final Label  widgetValueBackground;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Name:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      text = Widgets.newStringView(composite);
      text.setText(name);
      Widgets.layout(text,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Foreground:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,1,1,TableLayoutData.WE);
      {
        widgetColorForeground = Widgets.newCanvas(subComposite,SWT.BORDER);
        widgetColorForeground.setForeground((color.foreground != null) ? new Color(null,color.foreground) : null);
        widgetColorForeground.setBackground((color.foreground != null) ? new Color(null,color.foreground) : null);
        Widgets.layout(widgetColorForeground,0,0,TableLayoutData.WE,0,0,0,0,60,20);
        String colorName = (color.foreground != null)
                             ? String.format("#%02x%02x%02x",color.foreground.red,color.foreground.green,color.foreground.blue)
                             : "";
        widgetValueForeground = Widgets.newLabel(subComposite,colorName);
        Widgets.layout(widgetValueForeground,0,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
        widgetColorForeground.addMouseListener(new MouseListener()
        {
          @Override
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseUp(MouseEvent mouseEvent)
          {
            ColorDialog colorDialog = new ColorDialog(dialog);
            colorDialog.setRGB(color.foreground);
            RGB rgb = colorDialog.open();
            if (rgb != null)
            {
              color.foreground = rgb;
              widgetColorForeground.setForeground(new Color(null,rgb));
              widgetColorForeground.setBackground(new Color(null,rgb));
              widgetValueForeground.setText(String.format("#%02x%02x%02x",color.foreground.red,color.foreground.green,color.foreground.blue));
            }
          }
        });
      }

      label = Widgets.newLabel(composite,"Background:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,2,1,TableLayoutData.WE);
      {
        widgetColorBackground = Widgets.newCanvas(subComposite,SWT.BORDER);
        widgetColorBackground.setForeground((color.background != null) ? new Color(null,color.background) : null);
        widgetColorBackground.setBackground((color.background != null) ? new Color(null,color.background) : null);
        Widgets.layout(widgetColorBackground,0,0,TableLayoutData.WE,0,0,0,0,60,20);
        String colorName = (color.background != null)
                             ? String.format("#%02x%02x%02x",color.background.red,color.background.green,color.background.blue)
                             : "";
        widgetValueBackground = Widgets.newLabel(subComposite,colorName);
        Widgets.layout(widgetValueBackground,0,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
        widgetColorBackground.addMouseListener(new MouseListener()
        {
          @Override
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          @Override
          public void mouseUp(MouseEvent mouseEvent)
          {
            ColorDialog colorDialog = new ColorDialog(dialog);
            colorDialog.setRGB(color.background);
            RGB rgb = colorDialog.open();
            if (rgb != null)
            {
              color.background = rgb;
              widgetColorBackground.setForeground(new Color(null,rgb));
              widgetColorBackground.setBackground(new Color(null,rgb));
              widgetValueBackground.setText(String.format("#%02x%02x%02x",color.background.red,color.background.green,color.background.blue));
            }
          }
        });
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      button = Widgets.newButton(composite,"Save");
      Widgets.layout(button,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Default");
      Widgets.layout(button,0,1,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          color.foreground = color.DEFAULT_FOREGROUND;
          color.background = color.DEFAULT_BACKGROUND;

          widgetColorForeground.setForeground((color.foreground != null) ? new Color(null,color.foreground) : null);
          widgetColorForeground.setBackground((color.foreground != null) ? new Color(null,color.foreground) : null);
          widgetColorBackground.setForeground((color.background != null) ? new Color(null,color.background) : null);
          widgetColorBackground.setBackground((color.background != null) ? new Color(null,color.background) : null);

          String foregroundColorName = (color.foreground != null) ? String.format("#%02x%02x%02x",color.foreground.red,color.foreground.green,color.foreground.blue) : "";
          String backgroundColorName = (color.background != null) ? String.format("#%02x%02x%02x",color.background.red,color.background.green,color.background.blue) : "";
          widgetValueForeground.setText(foregroundColorName);
          widgetValueBackground.setText(backgroundColorName);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,2,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        @Override
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        @Override
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // add listeners

    return (Boolean)Dialogs.run(dialog,false);
  }
}

/* end of file */
